PRO G_GET_MAP_COORD, X_sky, Y_sky, coord

;	Program to get coordinates from GLESP fits maps
;	using output from g_read_fits_map.pro

;	Per Rex Christensen, September 2004

;	INPUT:	X_sky obtained from g_read_fits_map.pro
;		Y-sky obtained from g_read_fits_map.pro

;	OUTPUT:	coord[*,0] == theta {radian} (pi --> 0)  (0 = NORTH-POLE)
;		coord[*,1] == phi {radian} (0 --> 2*pi)

;=================

if N_params() NE 3 then begin
  print,'Syntax : G_GET_MAP_COORD, X_sky, Y_sky, coord'
  goto, Exit
endif

n_y	= n_elements(y_sky)
n_x	= n_elements(x_sky)

if n_y ne n_x then begin
	print,' '
	print,'X_sky and Y_sky  do not belong to the same map'
	goto,Exit
endif

n	= long(total(y_sky,/double) +0.5 )
print,'NUMBER of PIXELS: ', n

coord	= fltarr(n,2)

k	= -1L
for j = 0, n_x -1 do begin
	theta	= acos(x_sky[j])
	delta_phi	= 2.0D*!DPI/y_sky[j]
	for i = 0, y_sky[j] -1 do begin
		k	= k +1L
		coord[k,1]	= i*delta_phi
		coord[k,0]	= theta
	endfor
endfor

Exit:
end