#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define PR      fprintf
#define ER      stderr

/*
 * subrotines to calculate power spectra
 *
 *  created  by Oleg Verkhodanov 2002-2004
 *  last changes 10.05.2007
 */


/* calculate C_l by a_lm  */

void alm2cl (cl, lmax, alm, nalm)
double *cl;      /* output C_l array                           */
double *alm;     /* input a_lm array - complex (nalm/2, nalm)  */
int lmax;        /* maximum value of l                         */
int nalm;        /* dimension of a_lm array (Nmx, nalm=2*Nmx)  */
{
	register int l, m;

	for(l=0; l<=lmax; l++) {
		int ll = nalm * l;
		double s_alm = alm[ll] * alm[ll];
		for(m=1; m<=l; m++) {
			double v_r = alm[ll+m];
			double v_i = alm[ll+lmax+m+1];
			s_alm += 2. * (v_r * v_r + v_i * v_i);
		}
		cl[l] = s_alm / (double)(2*l + 1);
	}
	return;
}

void alm2cl_T (cl, lmax, alm, nbuf)
double *cl;      /* output C_l array            */
double *alm;     /* input a_lm array - complex  */
int lmax;        /* maximum value of l          */
int nbuf;        /* dimension of a_lm array     */
{
	register int l, m;

	for(l=0; l<=lmax; l++) {
		double s_alm = alm[l] * alm[l];
		for(m=1; m<=l; m++) {
			double v_r = alm[m*nbuf+l];
			double v_i = alm[(m+lmax+1)*nbuf+l];
			s_alm += 2. * (v_r * v_r + v_i * v_i);
		}
		cl[l] = s_alm / (double)(2*l + 1);
	}
	return;
}

void alm2dl (dl, lmax, alm, nalm)
double *dl;      /* output D_l array                           */
double *alm;     /* input a_lm array - complex (nalm/2, nalm)  */
int lmax;        /* maximum value of l                         */
int nalm;        /* dimension of a_lm array (nbuf=2*(lmax+1))  */
{
	register int l, m;
	double cl, pi2 = 2 * M_PI;
	double T_CMB = 2.726;

	T_CMB *= T_CMB;
	for(l=0; l<=lmax; l++) {
		int ll = nalm * l;
		double s_alm = alm[ll] * alm[ll];
		for(m=1; m<=l; m++) {
			double v_r = alm[ll+m];
			double v_i = alm[ll+lmax+m+1];
			s_alm += 2. * (v_r * v_r + v_i * v_i);
		}
		cl = s_alm / (double)(2*l + 1);
		dl[l] = cl * (double)(l*(l+1)) / pi2 / T_CMB;
	}

	return;
}

void alm2dl_T (dl, lmax, alm, nbuf)
double *dl;      /* output D_l array             */
double *alm;     /* input a_lm array - complex   */
int lmax;        /* maximum value of l           */
int nbuf;        /* dimension of a_lm array      */
{
	register int l, m;
	double cl, pi2 = 2 * M_PI;
	double T_CMB = 2.726;

	T_CMB *= T_CMB;
	for(l=0; l<=lmax; l++) {
		double s_alm = alm[l] * alm[l];
		for(m=1; m<=l; m++) {
			double v_r = alm[l+nbuf*m];
			double v_i = alm[l+nbuf*(m+lmax+1)];
			s_alm += 2. * (v_r * v_r + v_i * v_i);
		}
		cl = s_alm / (double)(2*l + 1);
		dl[l] = cl * (double)(l*(l+1)) / pi2 / T_CMB;
	}

	return;
}

void alm2clz (cl, lmax, alm, nbuf, zero_flag)
double *cl;      /* output C_l array                           */
double *alm;     /* input a_lm array - complex (nalm/2, nalm)  */
int lmax;        /* maximum value of l                         */
int nbuf;        /* dimension of a_lm array                    */
int zero_flag;
{
	register int l, m;
	int n_0=0;

	for(l=0; l<=lmax; l++) {
		int ll = nbuf * l;
		double s_alm = alm[ll] * alm[ll];
		n_0=0;
		for(m=1; m<=l; m++) {
		    double v_r = alm[ll+m];
		    double v_i = alm[ll+lmax+m+1];
		    if (zero_flag && fabs(v_r) < 1e-9 && fabs(v_r) < 1.e-9)
			 n_0++;
		    else
			 s_alm += 2. * (v_r * v_r + v_i * v_i);
		}
		cl[l] = s_alm / (double)(2*l + 1 - n_0);
	}
	return;
}

void alm2dlz (dl, lmax, alm, nalm, zero_flag)
double *dl;      /* output C_l array                           */
double *alm;     /* input a_lm array - complex (nalm/2, nalm)  */
int lmax;        /* maximum value of l                         */
int nalm;        /* dimension of a_lm array (Nmx, nalm=2*Nmx)  */
int zero_flag;
{
	register int l, m;
	double cl, pi2 = 2 * M_PI;
	double T_CMB = 2.726;
	int n_0=0;

	T_CMB *= T_CMB;
	for(l=0; l<=lmax; l++) {
		int ll = nalm * l;
		double s_alm = alm[ll] * alm[ll];
		n_0=0;
		for(m=1; m<=l; m++) {
		    double v_r = alm[ll+m];
		    double v_i = alm[ll+lmax+m+1];
		    if (zero_flag && fabs(v_r) < 1e-9 && fabs(v_r) < 1.e-9)
			 n_0++;
		    else
			 s_alm += 2. * (v_r * v_r + v_i * v_i);
		}
		cl = s_alm / (double)(2*l + 1 - n_0);
		dl[l] = cl * (double)(l*(l+1)) / pi2 / T_CMB;
	}

	return;
}

/* correlation spectra for polarization */
void alm2pcl (cle, clb, lmax, alm_t, alm_e, alm_b, nbuf)
double *cle,     /* output C_l-TE array                      */
       *clb;     /* output C_l-TB array                      */
double *alm_t,   /* input a_lmT array -complex (nbuf, nalm)  */
       *alm_e,   /* input a_lmE array -complex (nbuf, nalm)  */
       *alm_b;   /* input a_lmB array -complex (nbuf, nalm)  */
int lmax;        /* maximum value of l                       */
int nbuf;        /* dimension of a_lm array                  */
{
	register int l, m;

	for(l=0; l<=lmax; l++) {
		int ll = nbuf * l;
		double se_alm = alm_e == NULL ? 0. : alm_t[ll] * alm_e[ll],
		       sb_alm = alm_b == NULL ? 0. : alm_t[ll] * alm_b[ll];
		for(m=1; m<=l; m++) {
			double a_r = alm_t[ll+m],
			       a_i = alm_t[ll+lmax+m+1];
			double e_r = alm_e == NULL ? 0. : alm_e[ll+m],
			       e_i = alm_e == NULL ? 0. : alm_e[ll+lmax+m+1];
			double b_r = alm_b == NULL ? 0. : alm_b[ll+m],
			       b_i = alm_b == NULL ? 0. : alm_b[ll+lmax+m+1];
			if(alm_e) se_alm += 2. * (a_r * e_r + a_i * e_i);
			if(alm_b) sb_alm += 2. * (a_r * b_r + a_i * b_i);
		}
		if(alm_e) cle[l] = se_alm / (double)(2*l + 1);
		if(alm_b) clb[l] = sb_alm / (double)(2*l + 1);
	}
	return;
}
