#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "glesp.h"

#define PR  fprintf
#define ER  stderr

/*
 * Program calculates maximum T(l) by alm-rotation (see de Olivera Costa et
 * al., astro-ph/0307282)
 *
 *  code by Oleg Verkhodanov 2006
 *  last changes 24/07/2006
 */

void alm2tl(alm, lmax, nbuf, Tl, vpsi, vtheta, d_step, lmax_print)
double *alm, *Tl, *vpsi, *vtheta;
double d_step;
int lmax, nbuf, lmax_print;
{
    register int i, j, l, m;
    int i_m,
	j_m,       /* theta, degs */
	nb;
    double *alm_r;
    double d_d = d_step < 0.0 ? 0.06 : d_step;  /* in degs */
    double dphi   = d_d*M_PI/180.,
	   dpsi   = d_d*M_PI/180.,
	   dtheta = d_d*M_PI/180.;
    double phi = 0., psi = 0., theta = 0.;
    double t_max = -1.e+30;
    register int ll, llp;

    i_m = 360/d_d+.5;
//    i_m /= 2;
    j_m = 360/d_d+.5;
//    j_m /= 2;

    PR(ER, "%d steps by theta, %d steps by phi\n", j_m, i_m);
    nb = 2*(lmax_print + 1);
    i = (lmax_print+1)*nb;
    if ((alm_r=(double *)calloc((unsigned)i, sizeof *alm_r)) == (double *)0) {
	PR(ER, "alm2tl: no memory for temporal array...\n");
	exit(1);
    }

    for(l=2; l<=lmax_print; l++) {
	  Tl[l] = t_max;
	  vpsi[l] = vtheta[l] = 0.;
    }

//  for(i=-i_m; i<i_m; i++) {
    for(i=0; i<i_m; i++) {
	psi = dpsi * i;
	for(j=0; j<j_m; j++) {
	    theta = dtheta * j;

	    /* copy alm data */
	    ll = 0;
	    llp = 0;
	    for(l=0; l<=lmax_print; l++, ll+=nbuf, llp+=nb) {
		for(m=1; m<=l; m++) {
		    alm_r[llp+m] = alm[ll+m];
		    alm_r[llp+m+lmax_print+1] = alm[ll+m+lmax+1];
		}
	    }

	    /* rotate */
	    alm_rotate(alm_r, lmax_print, nb, psi, theta, phi);

	    /* find max */
	    for(llp=l=0; l<=lmax_print; l++, llp+=nb) {
		double s=0.0;
		for(m=1; m<=l; m++) {
		    double re = alm_r[llp+m],
			   im = alm_r[llp+m+lmax_print+1];
		    s += 2 * m*m * (re*re + im*im);
		}
PR(ER, "%5.1f %5.1f %d  %12.5e\n", theta*180/M_PI, psi*180/M_PI, l,
s<1.e-10?0.:s);
		if (s > Tl[l]) {
		    Tl[l] = s;
		    vpsi[l] = psi;
		    vtheta[l] = theta;
		}
	    }
	}
    }

    free((void *)alm_r);
    return;
}
