#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define PR      fprintf
#define ER      stderr

extern void     rffti_(int *, double *);
extern void     rfftb_(int *, double *, double *);
extern void     rfftf_(int *, double *, double *);
extern void     set_phi_(int *, double *, double *, double *);
extern void     fftw_b(int, double *, double *);

extern int  sym_grid;    /* symmetric grid over x */
extern int  fill_0;      /* fill arrays by 0      */
/*
 *  fast loop over m - SUM_m[a_{lm}e^{i m phi}]
 *      IMPLICIT REAL*8(a-h,o-y)
 *      DIMENSION f(Nmx,Nmp),Nph(Nmx)
 *      DIMENSION c(Nmp),wsave(2*Nmp+15)
 *      DIMENSION apm(0:Jmx,Nmx),bpm(0:Jmx,Nmx)
 *
 *
 *  input: map f(i,j), Nph - number of cells vs. x,
 *          Lmx - upper limit for m
 *  service: c, wsave
 *
 */
void alm_fast3_(Nmx,Nmp,Nx,Lmx,Vnyq,  Nph,f,c,apm,bpm,wsave)
int *Nmx;
int *Nmp, *Nx, *Lmx;
double *Vnyq;
int *Nph;
double *f;
double *apm, *bpm;
double *c, *wsave;
{
	register int ix, iz, ip;
	int nphi_max = 0, nx1 = *Nx-1;
	int Kx = sym_grid > 0 ? (*Nx + 1)/2 :  *Nx;

	if (fill_0) {
		for(ix=0; ix < *Nx; ix++)
			if (nphi_max < Nph[ix])
				nphi_max = Nph[ix];
	}

	for(ix=0; ix < Kx; ix++) {     /* loop over x */
		register int iix = ix * *Nmp, iiz;
		int nphi_work = fill_0 ? nphi_max : Nph[ix];
		int Mx = *Vnyq * nphi_work + .5;

		for(ip=0; ip<nphi_work; ip++)  c[ip] = f[iix+ip];

		fftw_b(nphi_work, c, wsave);
		apm[ix] = -c[0];

		if(Mx > *Lmx) Mx = *Lmx;
		if(2*Mx+1 > nphi_work)  Mx = (nphi_work-1)/2;

		for(iz=1,iiz=*Nmx+ix; iz <= Mx; iz++, iiz+=*Nmx) {
			register int iz2 = iz + iz - 1;
			apm[iiz] = -c[iz2];
			bpm[iiz] =  c[iz2+1];
		}
		if (sym_grid > 0) { /* then calculation for south pole */
		       register int ix_s = nx1 - ix;
		       if (ix_s == ix)
				continue;
		       iix = ix_s * *Nmp;
		       for(ip=0; ip<nphi_work; ip++)  c[ip] = f[iix+ip];

		       fftw_b(nphi_work, c, wsave);

		       apm[ix_s] = -c[0];
		       for(iz=1, iiz=*Nmx+ix_s; iz <= Mx; iz++, iiz+=*Nmx) {
			       register int iz2 = iz + iz - 1;
			       apm[iiz] = -c[iz2];
			       bpm[iiz] =  c[iz2+1];
		       }
		}
	}  /* END loop over ix */

	return;
}

/*
 *   loop over m - SUM_m[a_{lm}e^{i m phi}]
 *
 *     IMPLICIT REAL*8(a-h,o-y), COMPLEX*16(z)
 *     DIMENSION phi(Nmp)
 *     DIMENSION f(Nmx,Nmp),Nph(Nmx)
 *     DIMENSION apm(0:Jmx,Nmx),bpm(0:Jmx,Nmx)
 */
void alm_phi_(Nmx,Nmp,Lmx,Vnyq,Nx, Nph,f,phi2,phi,apm,bpm)
int *Nmx, *Nmp, *Lmx;
double *Vnyq;
int *Nx,
    *Nph;
double *phi2;
double *f,
       *apm, *bpm;
double *phi;
{
	register int ix=0, iq, m;
	int Nphi = Nph[0];
	int Mx = *Vnyq * Nphi + .5;
	double dphi;

	set_phi_(&Nphi, phi2, phi, &dphi);

	if ( Mx > *Lmx)  Mx = *Lmx;
	for(m=0; m<=Mx; m++) {
		register int mm = ix * *Nmx;
		for(iq=0; iq < Nphi; iq++) {
			double phase = m * phi[iq];
			double re = cos(phase);
			double im = sin(phase);

			for (ix=0; ix < *Nx; ix++) {
				register int iix = ix * Nphi;
				apm[mm+ix] -= f[iix+iq]*re;
				bpm[mm+ix] += f[iix+iq]*im;
			} /*  end of the loop over ix  */
		}         /* end of the loop over phi */
	}   /* end of the loop over m */

	return;
}


/*
 *   set of phi, 0< phi< phi2 through Nphi points
 */
void set_phi_(Nphi,phi2,phi,dph)
int *Nphi;
double *phi;
double *phi2, *dph;
{
	register int i;
	double pp = 0.0;

	*dph = *phi2 / (*Nphi - 0.0);

	for (i=0; i < *Nphi; i++) {
		phi[i] = pp;
		pp += *dph;
	}

	if ( phi[*Nphi-1] > *phi2)  phi[*Nphi-1] = *phi2;

	return;
}
