#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define PR  fprintf
#define ER  stderr
#define MOD(a,p)   ((a) - (p) * (int)((a)/(p)))

extern char *myname;
/*
  C-procedure, index accelarated
  conversion of alm, following Eric Hivon (2005) from HEALPIx:
      the recursion on the Wigner d matrix is inspired from the very stable
      double sided one described in Risbo (1996, J. of Geodesy, 70, 383)

*/
void alm_rotate(alm, lmax, nbuf, psi, theta, phi)
double *alm;
int lmax, nbuf;             /* dimension of alm-array   */
double psi, theta, phi;     /* rotation angles, radians */
{
    register int k, l, m, l2;
    double *alm1 = (double *)0, *alm2 = (double *)0;
    double *sqt = (double *)0,  *rsqt = (double *)0;
    double *d_  = (double *)0,  *dd_ = (double *)0;
    double *d   = (double *)0,  *dd  = (double *)0;
    double *sg  = (double *)0;
    double *sin_psi = (double *)0, *cos_psi = (double *)0;
    double *sin_phi = (double *)0, *cos_phi = (double *)0;
    int lma2 = lmax*2, lmax_2 = lmax + 2;
    double p, q, pc, qc;


    /* memory allocation */
    if((alm1=(double*)calloc((unsigned)(lma2+2),sizeof *alm1))==(double*)0){
nomem:
	PR(ER, "%s: no memory for temporal arrays...\n", myname);
	exit(1);
    }
    if((alm2=(double*)calloc((unsigned)(lma2+2),sizeof *alm2))==(double*)0)
	    goto nomem;
    if ((d_ = (double*)calloc((unsigned)((lma2+2)*lmax_2), sizeof *d_))==
	(double *)0)
	    goto nomem;
    d = d_ + lmax+3;
    if ((dd_ =(double*)calloc((unsigned)((lma2+2)*lmax_2),sizeof *dd_))==
	(double *)0)
	    goto nomem;
    dd = dd_ + lmax+3;
    if ((sqt = (double *)calloc((unsigned)(lma2+1), sizeof *sqt)) ==
	(double *)0)
	    goto nomem;
    if ((rsqt = (double *)calloc((unsigned)(lma2+1), sizeof *rsqt)) ==
	(double *)0)
	    goto nomem;
    if ((sg = (double *)calloc((unsigned)lmax_2, sizeof *sg)) == (double *)0)
	    goto nomem;
    if ((sin_psi = (double *)calloc((unsigned)(lmax+1), sizeof *sin_psi))==
	(double *)0)
	    goto nomem;
    if ((cos_psi = (double *)calloc((unsigned)(lmax+1), sizeof *cos_psi))==
	(double *)0)
	    goto nomem;
    if ((sin_phi = (double *)calloc((unsigned)(lmax+1), sizeof *sin_phi))==
	(double *)0)
	    goto nomem;
    if ((cos_phi = (double *)calloc((unsigned)(lmax+1), sizeof *cos_phi))==
	(double *)0)
	    goto nomem;

    for(m=0; m<=lmax; m+=2) sg[m+1] = -(sg[m]=1.0);
    for(m=0; m<=lma2; m++)  sqt[m]  = sqrt((double)m);
    for(m=0; m<=lmax; m++)  cos_psi[m] = cos_phi[m] = 1.0;
    for(m=1; m<=lmax; m++)  {
       sin_psi[m] = -sin(psi*m);
       cos_psi[m] =  cos(psi*m);
       sin_phi[m] = -sin(phi*m);
       cos_phi[m] =  cos(phi*m);
    }
    p = sin(theta/2);
    q = cos(theta/2);

    for(l=0; l<=lmax; l++) {
	register int ll;
	/* d-matrix of order l */
	if (!l) {
	    d[0] = 1.0;
	} else {
	    if (l == 1) {
		/* initialize d-matrix degree 1/2 */
		dd[0]         =  q;
		dd[1]         =  p;
		dd[lmax_2+0]  = -p;
		dd[lmax_2+1]  =  q;
	    }  else {

		/*  l-1 --> l-1/2  */
		l2 = 2*l - 1;
		for(m=0; m<=l2; m++) rsqt[m] = sqt[l2-m];
		qc = q / l2;
		pc = p / l2;

		ll = l2/2;
		for(k=0; k<=ll; k++) { /* keep only m' <= -1/2  */
		    register int lk = l2 - k;
		    for(m=0; m<=l2; m++) {
		     register int ml  = m* lmax_2, mlk  = ml +k;
		     register int ml1 = ml-lmax_2, mlk1 = ml1+k;
		     dd[mlk] = rsqt[m] *
			 (d[mlk]    * sqt[lk]*qc + d[mlk-1] * sqt[k] *pc)
		       + sqt[m] *
			 (d[mlk1-1] * sqt[k] *qc - d[mlk1]  * sqt[lk]*pc);
		    }
		}

		/* l=half-integer, reconstruct m'= 1/2 by symmetry */
		if ( ! MOD(l,2) ) {
		    register int k1 = l2-1, l1 = l-1;
		    for(k = 0; k<=k1; k+=2) {
			register int mk = k*lmax_2,
				     mlk = mk+l, l2k=(l2-k)*lmax_2 +l1;
			dd[mlk] =         dd[l2k];
			dd[mlk+lmax_2] = -dd[l2k-lmax_2];
		    }
		} else {
		    int k1 = l2-1, l1 = l-1;
		    for(k = 0; k<=k1; k+=2) {
			register int mk = k*lmax_2,
				     mlk = mk+l, l2k=(l2-k)*lmax_2 +l1;
			dd[mlk] =       -dd[l2k];
			dd[mlk+lmax_2] = dd[l2k-lmax_2];
		    }
		}
	    }

	    /*  l-1/2 --> l */
	    l2 = 2*l;
	    for(m = 0; m<=l2; m++)  rsqt[m] = sqt[l2-m];
	    qc = q / l2;
	    pc = p / l2;

	    ll = l2/2;
	    for(k = 0; k<=ll; k++) { /* keep only m' <= 0 */
		register int lk = l2 - k;
		for(m=0; m<=l2; m++) {
		    register int ml  = m *lmax_2,  mlk  = ml +k;
		    register int ml1 = ml-lmax_2,  mlk1 = ml1+k;
		    d[mlk] = rsqt[m] *
			(dd[mlk]   * sqt[lk] *qc + dd[mlk-1] * sqt[k]  *pc)
		      + sqt[m] *
			(dd[mlk1-1] * sqt[k] *qc - dd[mlk1]  * sqt[lk] *pc);
		}
	    } /* loop over k */
	}

	/* ------- apply rotation matrix ------- */
	for (m=0; m<=l; m++) {
	    register int r = m, i = m + lmax + 1;
	    register int lb = l*nbuf, lm_r = lb+m, lm_i = lm_r + lmax + 1;
	    register double c = cos_psi[m], s = sin_psi[m];
	    alm1[r] = alm[lm_r] * c - alm[lm_i] * s;
	    alm1[i] = alm[lm_i] * c + alm[lm_r] * s;
	}

	for (m=0; m<=l; m++) {
	    register int r = m, i = m + lmax + 1;
	    register double dl = d[(l+m)*lmax_2+l];
	    alm2[r] = alm1[0]      * dl;
	    alm2[i] = alm1[lmax+1] * dl;
	}

	for(m=0; m<=l; m++) {
	    register int r = m, i = m + lmax + 1, lm = l-m;
	    register double s_r1 = 0.0, s_i1 = 0.0;
	    register double s_r2 = 0.0, s_i2 = 0.0;

	    /* summing */
	    for(k=1; k<=l; k++) {
		register int mr = k, mi = k + lmax + 1;
		register double dl1 = d[(l-k)*lmax_2 + lm];
		register double dl2 = d[(l+k)*lmax_2 + lm] * sg[k];
		s_r1 += alm1[mr] * dl1;
		s_i1 += alm1[mi] * dl1;
		s_r2 += alm1[mr] * dl2;
		s_i2 += alm1[mi] * dl2;
	    }
	    alm2[r] +=  s_r1 + s_r2;
	    alm2[i] +=  s_i1 - s_i2;
	}

	/* output alm  */
	for(m=0; m<=l; m++) {
	    register int r = m, i = m + lmax + 1;
	    register double c = cos_phi[m], s = sin_phi[m];
	    alm[l*nbuf+m]        = alm2[r] * c - alm2[i] * s;
	    alm[l*nbuf+m+lmax+1] = alm2[i] * c + alm2[r] * s;
	}
    } /* loop over l */

    /* free memory */
    free((void *)d_);
    free((void *)dd_);
    free((void *)sqt);
    free((void *)rsqt);
    free((void *)alm1);
    free((void *)alm2);
    free((void *)sin_psi);
    free((void *)sin_phi);
    free((void *)cos_psi);
    free((void *)cos_phi);
    free((void *)sg);

    return;
}

extern void f_e2q(double *, double *, double);
extern void f_g2e(double *, double *, double);
extern void f_q2e(double *, double *, double);
extern void f_e2g(double *, double *, double);
extern void f_precess(double *, double *, double, double);
void  f_convert(vi,ei,eo,s1,s2,vo)
double *vi, *vo;
double ei, eo;
int s1, s2;
{
    register int i;
    double b[3];

    /* from this system */
    switch(s1) {
    case 'e':   /* ecliptic  */
	    for(i=0; i<3; i++)  b[i]=vi[i];
	    break;
    default:
    case 'g':   /* galactic */
	    f_g2e(vi, b, ei);
	    break;
    case 'q':   /* equatorial */
	    f_q2e(vi, b, ei);
	    break;
    }

    /* to this system */
    switch(s2) {
    case 'e':   /* ecliptic  */
	     for(i=0; i<3; i++)  vo[i]=b[i];
	     break;
    default:
    case 'g':   /* galactic */
	     f_e2g(b, vo, eo);
	     break;
    case 'q':   /* equatorial */
	     f_e2q(b, vo, eo);
	     break;
    }
    return;
}
double dot_product(v1,v2,n)
double *v1, *v2;
int n;
{
    register int i;
    double s=0.0;
    for (i=0; i<n; i++)  s += v1[i] * v2[i];
    return s;
}
/*
  computes the angular distance dist (in rad) between 2 vectors v1 and v2
  in general dist = acos ( v1 . v2 )
  except if the 2 vectors are almost aligned.
  HPix procedure
*/
void angdist(v1,v2, dist)
double *v1, *v2, *dist;
{
    register int i;
    double r1[3], r2[3], vd[3];
    double d, s;
    double dv1, dv2;

    /* normalize both vectors */
    dv1 = sqrt(dot_product(v1,v1,3));
    dv2 = sqrt(dot_product(v2,v2,3));
    for(i=0; i<3; i++) {
	r1[i] = v1[i] / dv1;
	r2[i] = v2[i] / dv2;
    }
    s = dot_product(r1, r2, 3);

    if (s > 0.999) {
       /* almost colinear vectors */
       for(i=0; i<3; i++)  vd[i] = r1[i] - r2[i];
       d = sqrt(dot_product(vd, vd, 3)); /* norm of difference */
       *dist = 2.0 * asin(d/2.);

    } else if (s < -0.999) {
       /* almost anti-colinear vectors */
       for(i=0; i<3; i++)  vd[i] = r1[i] + r2[i];
       d = sqrt(dot_product(vd, vd, 3)); /* norm of sum */
       *dist = M_PI - 2.0 * asin(d/2.);
    } else {
       /* other cases */
       *dist = acos( s );
    }
    return;
}

void cr2cr_alm(alm, lmax, nbuf, cr1, cr2)
double *alm;
int lmax, nbuf;
int cr1, cr2;
{
    register int i;
    double psi, phi, theta;
    double s1, s2, s3;
    double ep1 = 2000.0, ep2 = 2000.0;

    /* basis vector */
    double v1[3] = {1.0, 0.0, 0.0};
    double v2[3] = {0.0, 1.0, 0.0};
    double v3[3] = {0.0, 0.0, 1.0};
    double v1p[3], v2p[3], v3p[3];

    /* rotate them */
    f_convert(v1, ep1, ep2, cr1,cr2, v1p);
    f_convert(v2, ep1, ep2, cr1,cr2, v2p);
    f_convert(v3, ep1, ep2, cr1,cr2, v3p);

    /* normalize vectors */
    s1 = sqrt(dot_product(v1p,v1p,3));
    s2 = sqrt(dot_product(v2p,v2p,3));
    s3 = sqrt(dot_product(v3p,v3p,3));
    for (i=0; i<3; i++) {
	v1p[i] /= s1;
	v2p[i] /= s2;
	v3p[i] /= s3;
    }

    /* Euler angles */
    angdist(v3,v3p, &theta);
    psi = atan2(v2p[2], -v1p[2]);
    phi = atan2(v3p[1],  v3p[0]);

    alm_rotate(alm, lmax, nbuf, psi, theta, phi);
    return;
}


/*
    Routine to convert galactic coordinates to ecliptic
    coordinates at the given epoch.
    03/08/88, a.c.raugh, STX.
    11/16/01, C.Burigana, D.Maino, OAT
*/
void f_g2e(v_in, v_out, epoch)
double  *v_in,      /* Input coordinate vector  */
	*v_out,     /* Output coordinate vector */
	epoch;      /* Input coordinate epoch   */
{
    double b[3];  /* Temporary holding place for coordinates   */
    /* Galactic coordinate transformation matrix */
    double t[3][3] = {
	  {-0.054882486,  0.494116468, -0.867661702},
	  {-0.993821033, -0.110993846, -0.000346354},
	  {-0.096476249,  0.862281440,  0.497154957}
    };
    register int i,j;

    /*  Multiply by transformation matrix */
    for(i=0; i<3; i++) {
	b[i] = 0.0;
	for(j=0; j<3; j++)  b[i] += v_in[j]*t[i][j];
    }

    /*  Precess coordinates to epoch if necesssary */
    if (fabs(epoch - 2000.0) > .001)
	f_precess(b, v_out, 2000.0, epoch);
    else
	for(j=0; j<3; j++) v_out[j] = b[j];

    return;
}


/*
    Routine to convert from ecliptic coordinates to equatorial (celestial)
  coordinates at the given epoch.
  Adapted from the COBLIB routine by Dr. E. Wright.
  02/01/88, a.c.raugh, STX.
  11/16/01, C.Burigana,D.Maino, OAT
*/
void f_e2q(v_in,v_out,epoch)
double  *v_in,      /* Input coordinate vector  */
	*v_out,     /* Output coordinate vector */
	epoch;      /* Input coordinate epoch   */
{
    register int j;
    double t;            /* Julian centuries since 1900.0              */
    double epsilon;      /* Obliquity of the ecliptic                  */
    double b[3];         /* Temporary holding place for output vector  */
    double c = M_PI/180., dc, ds;

    /* Set-up */
    t = (epoch - 1900.0) / 100.0;
    epsilon = 23.452294 - 0.0130125*t - 1.63889e-6*t*t + 5.02778e-7*t*t*t;

    /* Conversion */
    dc = cos(c * epsilon);
    ds = sin(c * epsilon);
    b[0] = v_in[0];
    b[1] = dc*v_in[1] - ds*v_in[2];
    b[2] = dc*v_in[2] + ds*v_in[1];

    /* output vector */
    for(j=0; j<3; j++)  v_out[j] = b[j];

    return;
}

/*
    Routine to convert equatorial (celestial) coordinates at the given
  epoch to ecliptic coordinates.  Adapted from the COBLIB routine by Dr.
  E. Wright.

  02/01/88, a.c.raugh, STX.
  11/16/01, C.Burigana, D.Maino, OAT
*/
void f_q2e(v_in, v_out, epoch)
double *v_in, *v_out;
double epoch;
{
    register int i;
    double t;             /* Centuries since 1900.0                    */
    double epsilon;       /* Obliquity of the ecliptic, in degrees     */
    double b[3];          /* Temporary holding place for output vector */
    double dc, ds, c = M_PI/180.;

    t = (epoch - 1900.0) / 100.0;
    epsilon = 23.452294 - 0.0130125*t - 1.63889e-6*t*t + 5.02778e-7*t*t*t;

    dc = cos(c * epsilon);
    ds = sin(c * epsilon);
    b[0] = v_in[0];
    b[1] = dc*v_in[1] + ds*v_in[2];
    b[2] = dc*v_in[2] - ds*v_in[1];

    for(i=0; i<3; i++)  v_out[i] = b[i];

    return;
}


/*
    Routine to convert ecliptic coordinates at the given
  epoch to galactic coordinates.  The ecliptic coordinates are first
  precessed to 2000.0, then converted.
			  --------------
  Galactic Coordinate System Definition (from Zombeck, "Handbook of
  Space Astronomy and Astrophysics", page 71):
	 North Pole:  12:49       hours right ascension (1950.0)
		     +27.4        degrees declination   (1950.0)
    Reference Point:  17:42.6     hours right ascension (1950.0)
		     -28 55'      degrees declination   (1950.0)
			  --------------
  03/08/88, a.c.raugh, STX.
  11/16/01, C.Burigana, D.Maino, OAT
*/

void f_e2g(v_in, v_out, epoch)
double *v_in, *v_out;
double epoch;
{
    register int i, j;
    double b[3];  /* Temporary holding place for coordinates   */
    /* Galactic coordinate transformation matrix */
    double t[3][3] = {
	  {-0.054882486,  0.494116468, -0.867661702},
	  {-0.993821033, -0.110993846, -0.000346354},
	  {-0.096476249,  0.862281440,  0.497154957}
    };
    if (fabs(epoch - 2000.0) > .001)
	f_precess(v_in, b, epoch, 2000.0);
    else
	for(i=0; i<3; i++)  b[i] = v_in[i];

    /*  Multiply vector by transpose of transformation matrix */
    for(i=0; i<3; i++) {
	v_out[i] = 0.0;
	for(j=0; j<3; j++)  v_out[i] += b[j]*t[j][i];
    }
    return;
}

/*
    Subroutine to precess the input ecliptic rectangluar coordinates from
  the initial epoch to the final epoch.  First the precession around the
  Z-axis is done by a simple rotation.  A second rotation about the
  X-axis is then performed to account for the changing obliquity of the
  ecliptic.

  02/25/88, a.c.raugh, STX.
  03/22/88, acr: modified to include changing epsilon
  11/16/01, Carlo Burigana & Davide Maino, OAT
*/
void f_precess(v_in, v_out,  iepoch, oepoch)
double *v_in, *v_out;
double iepoch, oepoch;
{
    double temp;      /* Holding place for rotation results               */
    double gp_long;   /* General precession in longitude                  */
    double de;        /* Change in the obliquity of the ecliptic          */
    double obl_long;  /* -Longitude about which the obliquity is changing */
    double dl;        /* Longitude difference                             */
    double tm;        /* Mid-epoch, in tropical centuries since 1900      */
    double tmp_vec[3];/* Temporary holding vector                         */
    double dco, dso, dce, dse, dcl, dsl;
    double c = M_PI/180;
    register int i;

    tm = ((oepoch+iepoch)/2.0 - 1900.0) / 100.0;
    gp_long  = (oepoch-iepoch) * (50.2564+0.0222*tm) / 3600.;
    de       = (oepoch-iepoch) * (0.4711-0.0007*tm) / 3600.0;
    obl_long = 180.0 - (173. + (57.06+54.77*tm) / 60.0) + gp_long / 2.0;
    dl       = gp_long - obl_long;

    /*  Z-axis rotation by OBL_LONG */
    dco = cos(c * obl_long);
    dso = sin(c * obl_long);
    tmp_vec[0] = v_in[0]*dco - v_in[1]*dso;
    tmp_vec[1] = v_in[0]*dso + v_in[1]*dco;
    tmp_vec[2] = v_in[2];

    /*  X-axis rotation by dE */
    dce = cos(c * de);
    dse = sin(c * de);
    temp       = tmp_vec[1]*dce - tmp_vec[2]*dse;
    tmp_vec[2] = tmp_vec[1]*dse + tmp_vec[2]*dce;
    tmp_vec[1] = temp;

    /* Z-axis rotation by GP_LONG - OBL_LONG */
    dcl = cos(c * dl);
    dsl = sin(c * dl);
    temp       = tmp_vec[0]*dcl - tmp_vec[1]*dsl;
    tmp_vec[1] = tmp_vec[0]*dsl + tmp_vec[1]*dcl;
    tmp_vec[0] = temp;

    /* output */
    for(i=0; i<3; i++) v_out[i] = tmp_vec[i];

    return;
}


void f_h2q(v_in, v_out, epoch)
double *v_in, *v_out;
double epoch;
{
    register int i, j;
    double b[3];  /* Temporary holding place for coordinates   */
    /* Galactic coordinate transformation matrix */
    double t[3][3] = {
	  {-0.054882486,  0.494116468, -0.867661702},
	  {-0.993821033, -0.110993846, -0.000346354},
	  {-0.096476249,  0.862281440,  0.497154957}
    };
    if (fabs(epoch - 2000.0) > .001)
	f_precess(v_in, b, epoch, 2000.0);
    else
	for(i=0; i<3; i++)  b[i] = v_in[i];

    /*  Multiply vector by transpose of transformation matrix */
    for(i=0; i<3; i++) {
	v_out[i] = 0.0;
	for(j=0; j<3; j++)  v_out[i] += b[j]*t[j][i];
    }
    return;
}
