#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "glesp.h"
#define PR  fprintf
#define ER  stderr

/*
 *   procedure of mapping of correlation
 *
 *   Oleg Verkhodanov, 2014, vo@sao.ru
 */

void area_stack(f, x, nx, np, nph, file_src, area, nrx, nry, dx,dy, box_size,
							     flag_coor)
double *f, *x;     /* input map parameters           */
int nx, np, *nph;  /*                                */
char *file_src;    /* file of sources and zones      */
double **area;     /* output averaged (stacked) area */
int *nrx, *nry;    /* dimensions of area (nx, ny)    */
double *dx, *dy;   /* pixel size of area (nx, ny) for 1st area   */
double box_size;   /* default box side length        */
int flag_coor;     /* flag of coordinate system: 0 - gal, 1 - eq */
{
    struct _stacked_zone_ {
	double ra,   dec;       /* area center in eq. coor-s */
	double glon, glat;      /* area center in gal.coor-s */
	double theta, phi;      /* area center in pol.coor-s */
	double x_size, y_size;  /* size of area              */
	double xr1,xr2, yr1,yr2; /* limits of coors          */
	int proj;               /* type of area coors: 0-gal */
	double *ar;             /* area data                 */
    } *szo = NULL;
    FILE *fp = (FILE *)0;
    register int i, j, k, n_src=0;
    int na=0, nb=0;
    double r1=0.0, r2=0.0;
    char str[81], sc1[30], sc2[30], sz[30];
    double p = 180/M_PI;
    double fnx=2, fny=2;
    int median_flag = 0;

    if ((fp = fopen(file_src, "r")) == (FILE *)0) {
	PR(ER, "%s: can not open file '%s'\n", myname, file_src);
	exit(1);
    }
    /* read data about stacked zones from file */
    while (fgets(str, 80, fp) != NULL) {
	if (*str == '#' || *str == '\n' || *str == '\r')
	    continue;
	n_src++;
    }
    if ((szo=(struct _stacked_zone_*)calloc((unsigned)n_src, sizeof *szo))==
	(struct _stacked_zone_ *)0) {
	    PR(ER, "%s: no memory to process %d stacking areas.\n",
		myname, n_src);
	    exit(1);
    }
    i = 0;
    fseek(fp, 0l, 0);
    while (fgets(str, 80, fp) != NULL) {
	double b0=0.0, l0=0.0, th0=0.0, ph0=0.0;
	sc1[0] = sc2[0] = sz[0] = '\0';
	if (*str == '#' || *str == '\n' || *str == '\r')
	    continue;
	sscanf(str, "%s%s%s", sc1, sc2, sz);

	if (flag_coor) { /* flag of coordinate system: 0 - gal, 1 - eq */
	    szo[i].ra  = str2deg(sc1);         /* RA, hours */
	    szo[i].dec = str2deg(sc2);         /* Dec, degs */
	    equat2gal(szo[i].ra, szo[i].dec, &b0, &l0); /* b,l   in degs */
	    gal2pol(b0/p, l0/p, &th0, &ph0);   /* th,ph in rad  */
	    szo[i].glat = b0;
	    szo[i].glon = l0;
	} else {
	    szo[i].glat = str2deg(sc1);
	    szo[i].glon = str2deg(sc2);
	    gal2pol(szo[i].glat/p,szo[i].glon/p, &th0,&ph0); /* th,ph in rad*/
	}
	szo[i].theta = th0;
	szo[i].phi   = ph0;
	/* size of box */
	if (sz[0] != '\0')  {
	    r1 = r2 =  str2deg(sz);
	    szo[i].y_size = r1;
	    szo[i].x_size = r2;
	} else if (box_size > 0.) {
	    szo[i].y_size = r1 = box_size;
	    szo[i].x_size = r2 = box_size;
	}
	if (flag_coor) { /* flag of coordinate system: 0 - gal, 1 - eq */
	    szo[i].yr1 = szo[i].dec - r1/2;
	    szo[i].yr2 = szo[i].dec + r1/2;
	    szo[i].xr1 = szo[i].ra  - r2/2 / 15.;
	    szo[i].xr2 = szo[i].ra  + r2/2 / 15.;
	} else {
//            szo[i].yr1 = szo[i].theta - r1/2;
//            szo[i].yr2 = szo[i].theta + r1/2;
//            szo[i].xr1 = szo[i].phi   - r2/2;
//            szo[i].xr2 = szo[i].phi   + r2/2;
	    szo[i].yr1 = szo[i].glat - r1/2;
	    szo[i].yr2 = szo[i].glat + r1/2;
	    szo[i].xr1 = szo[i].glon - r2/2;
	    szo[i].xr2 = szo[i].glon + r2/2;
	}
	i++;
    }

    if (na == 0 && nb == 0) {
	double dphi = 360. / np;
	na = (szo[0].yr2 - szo[0].yr1) / dphi + 0.5;
	nb = (szo[0].xr2 - szo[0].xr1) / dphi * (flag_coor ? 15 : 1) + 0.5;
    }
    /* change size of pixels if fnx!=1 and fny!=1 */
    na = na * fny + .5;
    if (na < 2)  na = 2;
    nb = nb * fnx + .5;
    if (nb < 2)  nb= 2;


    /* find memory */
    i = na * nb;
    if ((*area = (double *)calloc((unsigned)i, sizeof **area))==(double*)0) {
	PR(ER, "%s: no memory for temporal array.\n", myname);
	exit(1);
    }
    *nry = na;    /* for Dec or GLat */
    *nrx = nb;    /* for RA  or GLon */
    *dx = r2 / nb / (flag_coor ? 15 : 1) ;
    *dy = r1 / na;

if (glesp_comments)
PR(ER, "number of fields: %d,  1st field size = (%10.3e x %10.3e) degs^2\n",
	 n_src, szo[0].x_size, szo[0].y_size);

    if (flag_coor) {    /* selection in equatorial coordinates */
	for(k=0; k<n_src; k++) {  /* loop over object fields */
	    double dra  = (szo[k].xr2 - szo[k].xr1) / nb,
		   ddec = (szo[k].yr2 - szo[k].yr1) / na;
	    for(i=0; i<na; i++) { /* loop over Dec */
		double tn = szo[k].yr1 + i * ddec, pn;
		for(j=0; j<nb; j++) { /* loop over RA */
		    double b0,l0, th0,ph0;
		    int n;
		    pn = szo[k].xr1 + j * dra;  /* in hours */
		    /* (pn; tn) = (RA,hours; Dec,degs) -> (b,l) */
		    equat2gal(pn, tn, &b0, &l0);
		    /* (b0/p; l0/p) = (glat,rad; glon,rad) -> (th,phi) */
		    gal2pol(b0/p, l0/p, &th0, &ph0);
		    n = ang2n(th0, ph0, x, nx, np, nph);
		    if (n<0) continue;
		    (*area)[i*nb+j] += f[n];
		}
	    }
	}
    } else {            /* selection in galactic coordinates */
	for(k=0; k<n_src; k++) {
	    double dlat  = (szo[k].yr2 - szo[k].yr1) / na,
		   dlon  = (szo[k].xr2 - szo[k].xr1) / nb;
	    for(i=0; i<na; i++) {
		double tn = szo[k].yr1 + i * dlat, pn;
		for(j=0; j<nb; j++) {
		    pn = szo[k].xr1 + j * dlon;  /* GLon, in degs */
		    double th0,ph0;
		    int n;
		    /* (b0/p; l0/p) = (glat,rad; glon,rad) -> (th,phi) */
		    gal2pol(tn/p, pn/p, &th0, &ph0);
		    n = ang2n(th0, ph0, x, nx, np, nph);
		    if (n<0) continue;
		    (*area)[i*nb+j] += f[n];
		}
	    }
	}
    }

    for(i=0; i<na; i++) {
	  for(j=0; j<nb; j++) {
		(*area)[i*nb+j] /= n_src;
	  }
    }

    if (szo != (struct _stacked_zone_*)0) free((void *)szo);

    if (median_flag) {
	double *marea = (double *)0;
	double median;
	extern int i_cmp_double(void *, void *);
	i=na*nb;
	if ((marea = (double*)calloc((unsigned)i,sizeof *marea))==(double*)0){
	    PR(ER, "%s: no memory for temporal array.\n", myname);
	    exit(1);
	}
	for(i=0; i<na; i++) {
	      for(j=0; j<nb; j++) {
		    marea[i*nb+j] = (*area)[i*nb+j];
	      }
	}

	qsort(marea, na*nb, sizeof(*marea), (int (*)())i_cmp_double);
	median = marea [ na*nb/2 ];

	for(i=0; i<na; i++) {
	      for(j=0; j<nb; j++) {
		    (*area)[i*nb+j] -= median;
	      }
	}

	if (marea != (double*)0) free((void *)marea);
    }

    return;
}
int i_cmp_double(p1, p2)
void  *p1, *p2;
{
	double *ph1 = (double *)p1,
	       *ph2 = (double *)p2;
	return (*ph1 < *ph2 ? -1 : (*ph1 > *ph2 ? 1 : 0));
}
