#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "glesp.h"

#define  PR     fprintf
#define  ER     stderr

extern void spline (double *, double *, int, double, double, double *);
extern void splint (double *, double *, double *, int, double, double *);
extern void gauleg (double, double, double *, double *, int);

/*
 * recalcultion of map to Gauss-Legendre quadrature grid
 *
 */
void calc2gl(f, nx, nphi, nph, x, x1, x2)
double *f;
double *x;
int *nph;
int nx, nphi;
double x1, x2;
{
    register int i, j;
    double *xl = (double *)0, *y = (double *)0, *w = (double *)0;
    double nphi_eq_flag = 1;
    int np, npmax;
    int x_c_flag = 0;

    i = nphi > nx ? nphi : nx;
    if ((xl = (double *)calloc((unsigned)i, sizeof *xl)) == (double *)0) {
nomem:
	   PR(ER, "calc2lgrid: no memory for temporal arrays...\n");
	   exit(1);
    }
    if ((y = (double *)calloc((unsigned)i, sizeof *y)) == (double *)0)
	   goto nomem;
    if ((w = (double *)calloc((unsigned)i, sizeof *w)) == (double *)0)
	   goto nomem;

    /* check equal or different numbers of phi */
    for(npmax=np=nph[0],i=1; i<nx; i++) {
	   if (np  !=  nph[i])  nphi_eq_flag = 0;
	   if (npmax < nph[i])  npmax = nph[i];
    }

    /* interpolate by phi: restore equal step by phi */
    if ( ! nphi_eq_flag ) {
       for(i=0; i<nx; i++) {
	   int ii = i * nphi;

	   if (nph[i] != npmax) {
		double dp = 2*M_PI / nph[i], dpn = 2*M_PI / npmax;
		for(j=0; j<nph[i]; j++) {
		     xl[j] = j * dp;
		     y[j] = f[ii+j];
		}

		spline (xl, y, nph[i], 2.e30, 2.e30, w);
		for(j=0; j<npmax; j++)
			splint (xl, y, w, nph[i], j * dpn, f+ii+j);

		nph[i] = npmax;
	   }
       }
    }

    /* set Legendre grid */
    gauleg(x1, x2, xl, (double *)0, nx);

    if (x[0] > x[1]) {
	    x_c_flag = 1;
	    for(i=0; i<nx; i++) { x[i] = -x[i]; xl[i] = -xl[i]; }
    }

    /*-----------------------------------------------*/
    /* interpolate by x: transition to Legendre grid */
    /*-----------------------------------------------*/
    for(j=0; j<npmax; j++) {
	    for(i=0; i<nx; i++)  y[i] = f[i*nphi+j];

	    spline (x, y, nx, 2.e30, 2.e30, w);
	    for(i=0; i<nx; i++)
		splint (x, y, w, nx, xl[i], f+i*nphi+j);
    }

    for(i=0; i<nx; i++)   x[i] = x_c_flag ? -xl[i] : xl[i];

    /* free memory */
    if(xl != (double *)0) free((void *)xl);
    if(y  != (double *)0) free((void *)y);
    if(w  != (double *)0) free((void *)w);

    return;
}
