#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "glesp.h"

#define PR      fprintf
#define ER      stderr
#define ABS(x)  ((x) < 0 ? -(x) : (x))

int naxis2;
char filename1[161], filename2[161];
char out_file[161];
int part_flag = 0;        /* type of value with relative error or */
			  /* output number flag */
int  lmax_print = -1, lmin_print = -1;

int cut_alm_flag=0;
int kolmogorov_smirnov_test = 0; /* flag to execute Kolmogorov-Smirnov test */

extern void alm_ks_test(double *, int, int, int, int, int, FILE *, int);
extern void alm_correl(double *, double *, int, int, int, int, FILE *);

extern void setopt(int, char **), usage(void);
extern void set_sig_fatal(void);

int  l_calc = -1, m_calc = -1,
     n_low = 100;
int sp_flag = 0;
int correl_flag = 0;
int kl_flag = 0;
int m0_off = 0;

char *myname="checkalm";

int main(argc, argv)
int argc;
char **argv;
{
    register int l, m;
    double *alm1 = (double *)0, *alm2 = (double *)0;
    int lmax=0, nalm=0, nbuf=0,
	lmx1=0, lmx2=0, nbuf1=0, nbuf2=0, nalm1=0, nalm2=0;
    int l_r_m=0, m_r_m=0, l_i_m=0, m_i_m=0;
    double sum1=0.0, sum2=0.0, re_max_dif=0.0,  im_max_dif=0.0;
    double ds;
    FILE *fp_d = (FILE *)0;

    out_file[0] = '\0';

    /* catch signal and finish program */
    set_sig_fatal();

    /* loop over external parameters */
    setopt(argc, argv);

    read_alm_table(filename1, &alm1, &nalm1, &nbuf1, &lmx1);
    lmax = lmax_print < lmx1 ? lmax_print : lmx1;
    if (!(kolmogorov_smirnov_test||cut_alm_flag)) {
	  read_alm_table(filename2, &alm2, &nalm2, &nbuf2, &lmx2);
	  lmax = lmax < lmx2 ? lmx2 : lmax;
    }
    nbuf = 2*(lmax+2);
    nalm = lmax + 1;
    alm2alm(&alm1, &lmx1, &nbuf1, lmax, nbuf);

    if ( *filename2 != '\0' )
	  alm2alm(&alm2, &lmx2, &nbuf2, lmax, nbuf);

    if (lmax_print <= 0 || lmax_print > lmax)
	    lmax_print = lmax;
    if (lmin_print <  0 || lmin_print > lmax)
	    lmin_print = 0;

    /*------------------*/
    /* open output file */
    /*------------------*/
    /* compulsory opening */
    if ( *out_file == '\0')
       fp_d = stdout;
    else
       if(strcmp(out_file, "stdout")=='\0' || strcmp(out_file, "-")=='\0') {
	    fp_d = stdout;
    } else {
       if((fp_d = fopen(out_file, "w")) == (FILE *)0) {
	    PR(ER,"%s: can not create file '%s'\n", myname, out_file);
	    exit(1);
       }
    }

    /***********************/
    /* a_{lm} - operations */
    /***********************/
    if (kolmogorov_smirnov_test) {
	alm_ks_test(alm1, nbuf, lmax, l_calc, m_calc, n_low, fp_d,part_flag);
    } else if (sp_flag) {
	alm2sp(alm1, nbuf, lmax, l_calc, m_calc, fp_d);
    } else if (cut_alm_flag) {
	alm2ffile(alm1, nbuf, lmax, l_calc, m_calc, fp_d, part_flag);
    } else if (correl_flag) {
	alm_correl(alm1, alm2, nbuf, lmax, l_calc, m_calc, fp_d);
    } else if (kl_flag) {
	extern void k_l();
	if ( m0_off )
	    for(l=2; l<=lmax; l++) {
		register int ll = l * nbuf;
		alm1[ll] = alm2[ll] = 0.0;
	    }
	k_l(alm1, alm2, lmax, nbuf, fp_d);
    } else {
	for (l=lmin_print; l<=lmax_print; l++) {
		register int ll = l * nbuf;
		for (m=0; m<=l; m++) {
			double alm_re1, alm_im1;
			double alm_re2, alm_im2;

			alm_re1 = alm1[ll+m];
			alm_im1 = alm1[ll+m+lmax+1];
			alm_re2 = alm2[ll+m];
			alm_im2 = alm2[ll+m+lmax+1];

			if ( out_file[0] != '\0' ) {
			    double v1, v2, dr;
			    switch (part_flag) {
			    case 1:
				v1 = alm_re1; v2 = alm_re2;
				break;
			    case 2:
				v1 = alm_im1; v2 = alm_im2;
				break;
			    default:
				v1 = sqrt(alm_re1*alm_re1+alm_im1*alm_im1);
				v2 = sqrt(alm_re2*alm_re2+alm_im2*alm_im2);
				break;
			    }
			    if ( ABS(v1) < 1.e-13 ) {
				    dr = -1000.;
			    } else
				    dr = (v1-v2) / v1;
			    PR(fp_d, "%5d %5d %12.5e\n", l, m, dr);
			}

			sum1 += alm_re1 * alm_re1 + alm_im1 * alm_im1;
			sum2 += alm_re2 * alm_re2 + alm_im2 * alm_im2;

			alm_re1 -= alm_re2;
			alm_im1 -= alm_im2;
			if (ABS(alm_re1) > ABS(re_max_dif)) {
				re_max_dif = alm_re1;
				l_r_m = l; m_r_m = m;
			}
			if (ABS(alm_im1) > ABS(im_max_dif)) {
				im_max_dif = alm_im1;
				l_i_m = l; m_i_m = m;
			}
		}
	}
	ds = sqrt(sum1) - sqrt(sum2);
	PR(ER,
	  "sum1^2=%15.8e, sum2^2=%15.8e, ds/sum1=%15.8e, sum1/sum2=%15.8e\n",
		sum1, sum2, ABS(ds)/sqrt(sum1), sqrt(sum1/sum2));
	PR(ER,
	  "max re(a1_lm - a2_lm) = %15.8e, re_a1_lm=%15.8e, l=%d, m=%d\n",
		re_max_dif, alm1[l_r_m*nbuf+m_r_m], l_r_m, m_r_m);
	PR(ER,
	  "max im(a1_lm - a2_lm) = %15.8e, im_a1_lm=%15.8e, l=%d, m=%d\n",
		im_max_dif, alm1[l_i_m*nbuf+m_i_m+lmax+1], l_i_m, m_i_m);
    }
    if (fp_d != (FILE *)0 && fp_d != stdout) fclose(fp_d);

    exit(0);
}

void setopt(argc, argv)
int argc;
char **argv;
{
	register int i, n=0;
	char buf[10];
	myname = argv[0];

	for(i=1; i<argc; i++) {
	    switch(argv[i][0]) {
	    case '-':
		switch(argv[i][1]) {
		case 'c':
			/* correlation of 2 sets of a_lm */
			correl_flag = 1;
			break;
		case 'e':
			/* file for relative errors */
			strcpy(out_file, argv[++i]);
			break;
		case 'f':
			cut_alm_flag = 1;
			strcpy(out_file, argv[++i]);
			break;
		case 'h':
			usage();
		case 'k':
			if (strncmp(argv[i]+1, "kl", 2) == 0)
			    kl_flag = 1;
			else
			    kolmogorov_smirnov_test = 1;
			break;
		case 'l':
			if (strcmp(argv[i]+1, "l") == 0) {
			    l_calc = atoi(argv[++i]);
			} else if (strncmp(argv[i]+1, "lmx",3) == 0 ||
			    strncmp(argv[i]+1, "lma",3) == 0  )
				lmax_print = atoi(argv[++i]);
			else if (strncmp(argv[i]+1, "lmn",3) == 0 ||
			    strncmp(argv[i]+1, "lmi",3) == 0  )
				lmin_print = atoi(argv[++i]);
			else
			    usage();
			break;
		case 'L':
			if (strncmp(argv[i]+1, "L", 1) == 0) {
			    l_calc = atoi(argv[++i]);
			} else if (strncmp(argv[i]+1, "Lmx",3) == 0 ||
			    strncmp(argv[i]+1, "Lma", 3) == 0  )
				lmax_print = atoi(argv[++i]);
			else if (strncmp(argv[i]+1, "Lmn", 3) == 0 ||
			    strncmp(argv[i]+1, "Lmi",3) == 0  )
				lmin_print = atoi(argv[++i]);
			else
			    usage();
			break;
		case 'm':
			if (strcmp(argv[i]+1, "m0") == 0) {
			    m0_off = 1;
			} else if (strncmp(argv[i]+1, "m",1) == 0) {
			    m_calc = atoi(argv[++i]);
			} else
			    usage();
			break;
		case 'M':
			if (strncmp(argv[i]+1, "M",1) == 0) {
			    m_calc = atoi(argv[++i]);
			} else
			    usage();
			break;
		case 'n':
			if (strncmp(argv[i]+1, "n",1) == 0) {
			    n_low = atoi(argv[++i]);
			} else
			    usage();
			break;
		case 'o':
			/* output file */
			strcpy(out_file, argv[++i]);
			break;
		case 's':
			sp_flag = 1;
			break;
		case 'v':
			buf[0] = argv[++i][0];
			switch (buf[0]) {
			case '1':
			case 'r':
				part_flag = 1;
				break;
			case '2':
			case 'i':
				part_flag = 2;
				break;
			case '0':
				part_flag = 0;
				break;
			}
			break;
		default :
			usage();
			break;
		}
		break;
	    case '+':
		switch(argv[i][1]) {
		case 'h':
			usage();
		default :
			usage();
			break;
		}
		break;
	    default :
		if (n == 0)
			strcpy(filename1, argv[i]);
		else
			strcpy(filename2, argv[i]);
		n++;
		break;
	    }
	}

	if (n < 2 && !(kolmogorov_smirnov_test||cut_alm_flag)) {
		PR(ER, "%s: must be 2 input files or flags '-k'/'-f'...\n",
			myname);
		usage();
	}
	return;
}

void usage()
{
PR(ER,"usage: %s file1 [file2] [-e err_file] [-v {012ri}]\n", myname);
PR(ER,"[-lmx lmax_print] [-lmn lmin_print]  [-k(Kolmogorov-Smirnov test)]\n");
PR(ER,"[-f ffile_out] [-l l_multipoles_sample] [-m m_multipoles_sample] [-n l_lim]\n");
PR(ER,"[-sp(spectrum)]\n");
exit(2);
}
