#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define ABS(x)  ((x) < 0 ? -(x) : (x))
#define PR      fprintf
#define ER      stderr

extern int    init_rand(void);
extern double norm_rand(void);
extern void   u_rand(double *, int, double, double);
extern int    seed_flag, random_seed;
extern char *myname;

/*   create a_lm by C_l   */
void cl2alm (cl, lmax, alm, nalm, fwhm)
double *cl;      /* input C_l array                            */
double *alm;     /* output a_lm array - complex (nalm/2, nalm) */
int lmax;        /* maximum value of l                         */
int nalm;        /* dimension of a_lm array (Nmx, nalm=2*Nmx)  */
double fwhm;     /* gaussian beam size in arcmin               */
{
	register int l, m, ll = lmax+2;
	double sq2 = M_SQRT1_2; /* 1/sqrt(2) = 0.70710678118654752440 */
	double *alm_r, *alm_i;

	ll *= ll;

	/* take memory for Real(a_lm) amd Imagin(a_lm) */
	if ((alm_r = (double*)calloc((unsigned)ll, sizeof *alm_r)) ==
	    (double*)0) {
nomem:
		PR(ER, "cl2alm: no memory for temporal arrays...\n");
		exit(1);
	}
	if ((alm_i = (double*)calloc((unsigned)ll, sizeof *alm_i)) ==
	    (double*)0)
		goto nomem;

	if ( seed_flag ) {
	      srand(random_seed);
	} else {
	      /* start random generator using system time */
	      random_seed = init_rand();
	}

	/* transform beam size to window size */
	fwhm *= M_PI/180. / 60. /  2.354820045;
	fwhm *= fwhm * 0.5;

	for(l=0; l<=lmax; l++) {
	    double C_l = cl[l];
	    double rms_cl;

	    if ( C_l < 1.e-50 )
		    continue;

	    rms_cl = sqrt(C_l);

	    if ( fwhm > 0. ) /* gaussian smoothing */
		    rms_cl *= exp(-l*(double)(l+1)*fwhm);

	    ll = (lmax+2) * l;

	    alm_r[ll+0] = rms_cl * norm_rand();
	    alm_i[ll+0] = 0.;

	    rms_cl *= sq2;
	    for(m=1; m<=l;    m++) {
		 /* real part    */
		 alm_r[ll+m] = rms_cl * norm_rand();
		 /* imagine part */
		 alm_i[ll+m] = rms_cl * norm_rand();
	    }
	}

	/* fill in 2-dim array of a_lm, real and imagine data */
	/*---------------------------------------------------------------*/
	/* in FORTRAN f(NMX,2*NMX): f(l,m), l=1,lmax, m=0,l, + real,imag */
	/* or map: f(x,phi), x=cos theta, x=-1,1, phi=0,2pi,             */
	/*---------------------------------------------------------------*/
	/* in C: double f[2*NMX][NMX]: alm[m][l]                         */
	/*                                                               */
	/*  f(2:Lmx,1:Lmx+1) - real a_lm (cos),                          */
	/*  f(2:Lmx,Lmx+2:2*Lmx+2) - imag a_lm (sin),                    */
	/* in C: f[m+1][i], f[Lmx+m+2][i]);                              */
	/*                                                               */
	/*---------------------------------------------------------------*/

	for(l=0; l<=lmax; l++) {
	    int ll1 = (lmax+2) * l;
	    ll = nalm * l;
	    for(m=0; m<=lmax; m++) {
		/* real part    */
		alm [ll+m] = alm_r[ll1+m];

		/* imagine part */
		alm [ll+m + lmax+1] = alm_i[ll1+m];
	    }
	}
	/* free memory */
	free((void *)alm_r);
	free((void *)alm_i);

	return;
}


/*   create a_lm by C_l   */
void cl2alm_T (cl, lmax, alm, nbuf, fwhm)
double *cl;      /* input C_l array                            */
double *alm;     /* output a_lm array - complex (nalm/2, nalm) */
int lmax;        /* maximum value of l                         */
int nbuf;        /* dimension of a_lm array (Nmx, nalm=2*Nmx)  */
double fwhm;     /* gaussian beam size in arcmin               */
{
	register int l, m, ll = 2 * (ABS(lmax)+1);
	double sq2 = M_SQRT1_2; /* 1/sqrt(2) = 0.70710678118654752440 */
	double *alm_r, *alm_i;

	nbuf = ABS(nbuf);
	ll *= nbuf;

	/* take memory for Real(a_lm) amd Imagin(a_lm) */
	if ((alm_r = (double*)calloc((unsigned)ll, sizeof *alm_r)) ==
	    (double*)0) {
nomem:
	    PR(ER, "cl2alm: no memory for temporal arrays of %d els...\n",ll);
	    exit(1);
	}
	if ((alm_i = (double*)calloc((unsigned)ll, sizeof *alm_i)) ==
	    (double*)0)
		goto nomem;

	if ( seed_flag ) {
	      srand(random_seed);
	} else {
	      /* start random generator using system time */
	      random_seed = init_rand();
	}

	/* transform beam size to window size */
	fwhm *= M_PI/180. / 60. /  2.354820045;
	fwhm *= fwhm * 0.5;

	for(l=0; l<=lmax; l++) {
	    double C_l = cl[l];
	    double rms_cl;

	    if ( C_l < 1.e-50 )
		    continue;

	    if ( fwhm > 0. ) /* gaussian smoothing */
		    C_l *= exp(-l*(double)(l+1)*fwhm);

	    rms_cl = sqrt(C_l);

	    ll = (lmax+2) * l;

	    alm_r[ll] = rms_cl * norm_rand();
	    alm_i[ll] = 0.;

	    rms_cl *= sq2;
	    for(m=1; m<=l;    m++) {
		 /* real part */
		 alm_r[ll+m] = rms_cl * norm_rand();
		 /* imagine   */
		 alm_i[ll+m] = rms_cl * norm_rand();
	    }
	}

	/* fill in 2-dim array of a_lm, real and imagine data */
	/*---------------------------------------------------------------*/
	/* in FORTRAN f(NMX,2*NMX): f(l,m), l=1,lmax, m=0,l, + real,imag */
	/* or map: f(x,phi), x=cos theta, x=-1,1, phi=0,2pi,             */
	/*---------------------------------------------------------------*/
	/* in C: double f[2*NMX][NMX]: alm[m][l]                         */
	/*                                                               */
	/*  f(2:Lmx,1:Lmx+1) - real a_lm (cos),                          */
	/*  f(2:Lmx,Lmx+2:2*Lmx+2) - imag a_lm (sin),                    */
	/* in C: f[m+1][i], f[Lmx+m+2][i]);                              */
	/*                                                               */
	/*---------------------------------------------------------------*/

	for(l=0; l<=lmax; l++) {
	    int ll1 = (lmax+2) * l;
	    for(m=0; m<=lmax; m++) {
		/* real part    */
		alm [m*nbuf+l] = alm_r[ll1+m];
		/* imagine part */
		alm [(m+lmax+1)*nbuf+l] = alm_i[ll1+m];
	    }
	}
	/* free memory */
	free((void *)alm_r);
	free((void *)alm_i);

	return;
}


void phase_rand(phase, n)
double *phase;
int n;
{
	double pi2 = 2 * M_PI, shift = 0.;

	if ( seed_flag ) {
	      /* start random generator using system time */
	      srand(random_seed);
	} else {
	      init_rand();
	}

	u_rand(phase, n, pi2, shift);

	return;
}


/* simulation a_lm by C_l with phase */
void cl2alm_ph(cl, lmax, alm, nalm, fwhm)
double *cl;      /* input C_l array              */
double *alm;     /* output a_lm array - complex  */
int lmax;        /* maximum value of l           */
int nalm;        /* dimension of a_lm array      */
double fwhm;     /* gaussian beam size in arcmin */
{
	register int l, m, ll = lmax+2;
	double *alm_r, *alm_i, *phase;
	double sq2 = M_SQRT1_2; /* 1/sqrt(2) = 0.70710678118654752440 */

	ll *= ll;

	if ((phase = (double *)calloc((unsigned)(2*(lmax+2)*(lmax+2)),
	      sizeof *phase)) == (double *)0) {
		PR(ER,"cl2alm_ph: no memory for phase array[%d]\n",
		    2*(lmax+2)*(lmax+2));
		exit(1);
	}
	/* take memory for Real(a_lm) amd Imagin(a_lm); */
	if ((alm_r = (double*)calloc((unsigned)ll, sizeof *alm_r)) ==
	   (double*)0) {
nomem:
		PR(ER, "cl2alm_phase: no memory for temporal arrays...\n");
		exit(1);
	}
	if ((alm_i = (double*)calloc((unsigned)ll, sizeof *alm_i)) ==
	   (double*)0)
		goto nomem;

	/* random generator start and simulation of phase  */
	phase_rand(phase, ll);

	if ( seed_flag ) {
	      srand(random_seed);
	} else {
	      /* start random generator using system time */
	      random_seed = init_rand();
	}

	/* transform beam size to window size */
	fwhm *= M_PI/180. / 60. /  2.354820045;
	fwhm *= fwhm * 0.5;

	for(l=0; l<=lmax; l++) {
	    /* conversion: C_l from CMBfast to actual C_l */
	    double C_l = cl[l];
	    double rms_cl;

	    if ( C_l < 1.e-50 )
		    continue;

	    if ( fwhm > 0. ) /* gaussian smoothing */
		    C_l *= exp(-l*(double)(l+1)*fwhm);

	    rms_cl=sqrt(C_l);


	    ll = (lmax+2) * l;

	    alm_r[ll+0] = rms_cl * norm_rand();
	    alm_i[ll+0] = 0.;

	    rms_cl *= sq2;
	    for(m=1; m<=l; m++) {
		 double rms = rms_cl * norm_rand();

		 /* real    part */
		 alm_r[ll+m] = rms * cos (phase[ll+m]);
		 /* imagine part */
		 alm_i[ll+m] = rms * sin (phase[ll+m]);
	    }
	}

	/* fill in 2-dim array of a_lm, real and imagine data */
	/*---------------------------------------------------------------*/
	/* in FORTRAN f(NMX,2*NMX): f(l,m), l=1,lmax, m=0,l, + real,imag */
	/* or map: f(x,phi), x=cos theta, x=-1,1, phi=0,2pi,             */
	/*---------------------------------------------------------------*/
	/* in C: double f[2*NMX][NMX]: alm[m][l]                         */
	/*                                                               */
	/*  f(2:Lmx,1:Lmx+1) - real a_lm (cos),                          */
	/*  f(2:Lmx,Lmx+2:2*Lmx+2) - imag a_lm (sin),                    */
	/* in C: f[m+1][i], f[Lmx+m+1][i]);                              */
	/*                                                               */
	/*---------------------------------------------------------------*/

	for(l=0; l<=lmax; l++) {
	    int ll1 = (lmax+2) * l;
	    ll = nalm * l;
	    for(m=0; m<=l; m++) {

		/* real part    */
		alm [ll+m] = alm_r[ll1+m];

		/* imagine part */
		alm [ll+m + lmax+1] = alm_i[ll1+m];
	    }
	}

	/* free memory */
	free((void *)alm_r);
	free((void *)alm_i);
	if (phase != (double *)0)
		{ free((void *)phase); phase = (double *)0; }


	return;
}

/*   create a_lm by amplitude and phase */
void amph2alm (amp, phase, lmax, alm, nalm)
double *amp,     /* amplitude   */
       *phase;   /* phase       */
double *alm;     /* output a_lm array - complex (nalm/2, nalm) */
int lmax;        /* maximum value of l                         */
int nalm;        /* dimension of a_lm array                    */
{
	register int l, m;

	for(l=2; l<=lmax; l++) {
		int ll = nalm * l;
		int la = (lmax+1) * l;
		for(m=0; m<=l; m++) {
			double r = amp[la+m], ph = phase[la+m];
			alm[ll+m]        = r * cos (ph);
			alm[ll+m+lmax+1] = r * sin (ph);
		}
	}
}

double *reform_tab(arr, old_lmx, new_lmx)
double **arr;
int old_lmx, new_lmx;
{
    register int l, m;
    double *v = (double *)0;

    if (old_lmx < new_lmx) {
	if((v=(double *)calloc((unsigned)new_lmx, sizeof *v))==(double *)0){
	    PR(ER,"reform_tab: no memory...\n");
	    exit(1);
	}
	for(l=0; l<=old_lmx; l++) {
	    int lo = (old_lmx + 1) * l,
		ln = (new_lmx + 1) * l;
	    for(m=0; m<=l; m++) v[ln+m] = (*arr)[lo+m];
	}

	free ((void *)*arr);
    } else if (old_lmx > new_lmx) {
	v = *arr;
	for(l=0; l<=old_lmx; l++) {
	    int lo = (old_lmx + 1) * l,
		ln = (new_lmx + 1) * l;
	    for(m=0; m<=l; m++) v[ln+m] = (*arr)[lo+m];
	}
    }

    return v;
}


void  read_random_alm_T(file_cl, alm,nalm,nbuf, lmax,rand_seed, fwhm, flag)
char *file_cl;
double **alm;
int *nalm, *nbuf, *lmax;
int rand_seed;
double fwhm;
int flag; /* 1:cl, 2:dl */
{
    int l;
    char str[81];
    double *cl = (double *)0;
    FILE *fp;

    random_seed = rand_seed;

    if (*lmax == 0) {
	fp = fopen(file_cl, "r");
	if (fp == (FILE *)0) {
	    PR(ER, "%s: can not open file '%s'\n", myname, file_cl);
	    exit(1);
	}
	while(fgets(str,80,fp)!=NULL) {
	    double v;
	    if(*str == '#' || *str == '\n' || *str == '\r')
		continue;
	    sscanf(str, "%d%lf", &l, &v);
	    if (l > *lmax) *lmax = l;
	}
	fclose(fp);
    }
    cl = (double *)calloc(*lmax+1, sizeof *cl);
    if (cl == (double *)0) {
	PR(ER, "%s: no memory for temporal array\n", myname);
	exit(1);
    }
    fp = fopen(file_cl, "r");
    while(fgets(str,80,fp)!=NULL) {
	double v;
	if(*str == '#' || *str == '\n' || *str == '\r')
	    continue;
	sscanf(str, "%d%lf", &l, &v);
	if (0<= l && l <= *lmax)  cl[l] = v;
    }
    fclose(fp);
    if (flag == 2) {
	/* convert Dl do Cl */
	extern void dl2cl(double *, int);
	dl2cl(cl, *lmax);
    }

    if (*nalm < 0) {
	if (*nbuf == 0)
	    *nbuf = 2*(*lmax + 1);
    }
    if (*nalm < 0 && *nbuf < 0 && *alm == (double*)0) {
	unsigned n = *nbuf * ( *nalm < 0 ? -(*nalm) : *nalm );
	if((*alm=(double*)calloc(n, sizeof **alm)) == (double *)0) {
	    PR(ER,"%s: no memory for arrays of %d elements...\n", myname, n);
	    exit(1);
	}
    }

    cl2alm_T (cl, *lmax, *alm, *nalm, fwhm);

    if (*nalm >= 0)
	*nalm = 2*(*lmax + 1);


    free((void *)cl);

    return;
}

int Cget_max(file_cl)
char *file_cl;
{
    int l;
    char str[81];
    FILE *fp;
    int lmax = 0;

    fp = fopen(file_cl, "r");
    if (fp == (FILE *)0) {
	PR(ER, "%s: can not open file '%s'\n", myname, file_cl);
	exit(1);
    }
    while(fgets(str,80,fp)!=NULL) {
	double v;
	if(*str == '#' || *str == '\n' || *str == '\r')
	    continue;
	sscanf(str, "%d%lf", &l, &v);
	if (l > lmax) lmax = l;
    }
    fclose(fp);

    return lmax;
}

void dl2cl(dl, lmax)
double *dl;
int lmax;
{
    register int l;
    double pi = M_PI;
    for(l=1; l<=lmax; l++)  dl[l] /= l*(l+1) / 2. / pi;
    return;
}
