#include <stdio.h>
#include <math.h>

#include "glesp.h"

#define PR      fprintf
#define ER      stderr
#define ABS(x)  ((x) < 0 ? -(x) : (x))

extern void gauleg(double, double, double *, double *, int);
extern void coord(int, int, double, double, double,
	   double *, double *, int, int *, int);
extern int mlim_max(double, double);
extern int get_nphi_by_x(double *, int, int);
extern int set_nphi_by_x(double *, int, int *);
extern int pget_nphi_by_x(double *, int, int, int);


/* select and calculate grid for Legendre polynomials */
void coord_(coord_flag, lx_add, x1, x2, phi2, x, pw, Nx, Nph, Nphi)
int *coord_flag;  /* flag of coordinate grid selection */
int *lx_add;      /* addition to Nphi numbers          */
double *x1, *x2,
      *phi2;
double *x,  *pw;
int    *Nx, *Nphi;
int *Nph;
{
  coord(*coord_flag, *lx_add, *x1, *x2, *phi2,
	  x, pw, *Nx, Nph, *Nphi);
}

void coord(coord_flag, lx_add, x1, x2, phi2, x, pw, Nx, Nph, Nphi)
int coord_flag;  /* flag of coordinate grid selection */
int lx_add;      /* addition to Nphi numbers          */
double x1, x2,
      phi2;
double *x,  *pw;
int    Nx, Nphi;
int *Nph;
{
	register int i;
	double xx;
	double dx=1.0, dphi;
	int NNphi = 0, Imx, lx;
	int npp = 0;

	if(coord_flag < 0) {
		/* uniform grid */
		dx = (x2-x1) / (Nx -1.0);
		xx = x1;
		for(i=0; i<Nx; i++) {
			pw[i] = dx / 1.5; /* for Sympson integral */
			x[i] = xx;
			xx += dx;
		}
		pw[0]    = pw[0]    / 2.;
		pw[Nx-1] = pw[Nx-1] / 2.;

		for(i=1; i<Nx-1; i+=2) {
			pw[i] += pw[i];
		}
	} else {  /* special grid for Legendre polynomials */
		gauleg(x1, x2, x, pw, Nx);
	}
	if( x[Nx-1] > 1.0)   x[Nx-1] = 1.0;

       /*
	*   arbitrary grid for sphere and fast-fourier
	*/
	Imx = (Nx+1)/2;
	if (coord_flag == 0) {
	    npp = lx_add;
	    lx = -20;
	} else if (coord_flag == 2) { /* polarization pixelization */
	    lx = 'P';
	} else {
	    lx = lx_add; /* old value:4, starting length, near Pole */
	}
	/* Nph[0] = Nphi; */
	switch (lx) {
	case 0:
		Nph[0] = Nphi;
		break;
	case -1:
		Nph[0] = 10;
		break;
	case -10: /* set standard Gauss-Legendre pixelization scheme */
		Nph[0] = get_nphi_by_x2(x, Nx, Nphi, 0);
		break;
	case -20: /* set pixelization from polar cap */
		Nph[0] = pget_nphi_by_x(x, Nx, npp, 0);
		break;
	case 'P': /* set pixelization for polarizaton scheme */
		Nph[0] = get_pol_nphi_by_x(x, Nx, Nphi, 0);
		break;
	default :
		Nph[0] = 10;
		break;
	}

	Nph[Nx-1] = Nph[0];
	NNphi = Nph[0] + Nph[Nx-1];
	dphi = phi2 / Nph[0];
	pw[0]    = dphi * pw[0];
	pw[Nx-1] = dphi * pw[Nx-1];
//PR(ER, "ni_0=%d, nj_0=%d\n", Nph[0], Nph[Nx-1]);

	for(i=1; i<Imx; i++) {
		int jx;
		/* Nph[i] = Nph[0]; */
		switch (lx) {
		case 0:
		       /*
			*   Number of cells over phi for a given x
			*   homogeneous grid for rectangular area
			*/
			Nph[i] = Nphi;
			break;
		case -10: /* set standard GLESP scheme */
			Nph[i] = get_nphi_by_x2(x, Nx, Nphi, i);
			break;
		case -20: /* set pixelization from polar cap */
			Nph[i] = pget_nphi_by_x(x, Nx, npp, i);
			break;
		case -1:
			Nph[i] = Nph[i-1] + lx;
			break;
		case 'P':
			Nph[i] = get_pol_nphi_by_x(x, Nx, Nphi, i);
			break;
		default :
			Nph[i] = Nph[i-1] + lx;
			break;
		}

		if(Nph[i] > Nphi) Nph[i] = Nphi;
		dphi  = phi2 / Nph[i];
		pw[i] *= dphi;
		NNphi += Nph[i];

		jx = Nx - i - 1;
		if(jx > i) {
			pw[jx]  *= dphi;
			Nph[jx] = Nph[i];
			NNphi += Nph[jx];
		}
//PR(ER, "i=%d, j=%d, ni=%d, nj=%d\n", i, jx, Nph[i], Nph[jx]);
	}

	return;
}

/* special grid for Legendre polynomials */
void gauleg(x1,x2,x,w,n)
double x1, x2;
double *x, *w;
register int n;
{
	double eps = 3.e-15,
	       pi = M_PI;
	double p1, p2, p3, pp, xm, xl, z, z1;
	register int i,j,m;

	m = (n+1) / 2;
	xm = 0.5 * (x2 + x1);
	xl = 0.5 * (x2 - x1);

	for (i=0; i<m; i++) {
		/* FORTRAN: z=cos(pi*(i-.25d0)/(n+.5d0)) */
		z = cos(pi * ((i+1)-.25) / (n+.50));
lab_1:
		p1 = 1.0;
		p2 = 0.0;
		for(j=0; j<n; j++) {
		    p3 = p2;
		    p2 = p1;
		    /* FORTRAN: p1=((2.d0*j-1.d0)*z*p2-(j-1.d0)*p3)/j */
		    p1 = ((2*j+1)*z*p2-j*p3)/(j+1);
		}
		pp = n*(z*p1-p2)/(z*z-1.0);
		z1 = z;
		z = z1-p1/pp;

		if(ABS(z-z1) > eps)  goto lab_1;

		x[i] = xm - xl * z;
		/* FORTAN: x(n+1-i)=xm+xl*z */
		x[n-1-i] = xm + xl * z;
		if (w != (double *)0) {
			w[i] = 2.0*xl / ((1. - z*z) * pp*pp);
			w[n-1-i] = w[i];
		}
	}

	return;
}

/*
 *   max M for a given x & eps
 */
int mlim_max(x, eps)
double x, eps;
{
	return (int)(2.*fabs(log(eps)/log(1.-x*x+1.e-15)));
}

/* select and calculate weights at grid for Legendre polynomials */
void weights_(coord_flag, x1, x2, phi2, x, pw, Nx, Nph)
double *x1, *x2,
      *phi2;
double *x,  *pw;
int *coord_flag; /* flag of coordinate grid selection */
int    *Nx;
int *Nph;
{
    weights(*coord_flag, *x1, *x2, *phi2, x, pw, *Nx, Nph);
}

void weights(coord_flag, x1, x2, phi2, x, pw, Nx, Nph)
double x1, x2,
      phi2;
double *x,  *pw;
int coord_flag; /* flag of coordinate grid selection */
int    Nx;
int *Nph;
{
	register int i, Imx;
	double xx;
	double dx, dphi;

	if(coord_flag < 0) {
		/* uniform grid */
		dx = (x2-x1) / (Nx -1.0);
		xx = x1;
		for(i=0; i<Nx; i++) {
			pw[i] = dx / 1.5; /* for Sympson integral */
			x[i] = xx;
			xx += dx;
		}
		pw[0]    = pw[0]    / 2.;
		pw[Nx-1] = pw[Nx-1] / 2.;

		for(i=1; i<Nx-1; i+=2) {
			pw[i] += pw[i];
		}
	} else {  /* special grid for Legendre polynomials */
		gauleg(x1, x2, x, pw, Nx);
	}

	Imx = (Nx+1)/2;
	dphi = phi2 / (Nph[0]-0.0);
	pw[0]    = dphi * pw[0];
	pw[Nx-1] = dphi * pw[Nx-1];

	for(i=1; i<Imx; i++) {
		int jx;
		dphi  = phi2 / (Nph[i]-0.0);
		pw[i] *= dphi;

		jx = Nx - i - 1;
		if(jx > i) {
			pw[jx]  *= dphi;
		}
	}

	return;
}

int set_nphi_by_x(x, nx, nph)
double *x;
int nx;
int *nph;
{
    register int i, ntot;
    int  np;
    double pi = M_PI, pi2=2.*M_PI;
    double s0;

    i = (nx + 1) / 2;
    /* np = 2pi / (width=(theta[i+1]-theta[i-1])/2) */
    np = fabs(4.*pi / (acos(x[i+1]) - acos(x[i-1]))) + 0.5;

    /* actual size of one pixel: */
    /*   s0 = dtheta * dphi      */
    s0 = pi2 / np * (0.5*fabs(acos(x[i+1]) - acos(x[i-1])));
/*
PR(ER, "pixel_size = %gx%g = %g sq.arcmin, nx=%d, np=%d\n",
0.5*fabs(acos(x[i+1]) - acos(x[i-1]))/pi*180*60,  2. / np * 180 * 60.,
s0/pi/pi*60*60*180*180,  nx, np);
*/
    /* set sizes by map */
    for(ntot=i=0; i< nx; i++) {
	double dth=0.0;
/*
	if ( i==0 )
		dth =  (acos(x[i+1]) + acos(x[i])) / 2;
	else if ( i==nx-1 )
		dth = pi - (acos(x[i]) + acos(x[i-1]))/2;
*/
	if ( i==0 || i==nx-1 ) {
	    if ( x[0] > 0. && i==0 ) {
		dth =  (acos(x[i+1]) + acos(x[i])) / 2;
	    } else if ( x[0] < 0. && i==0)  {
		dth = pi - (acos(x[i]) + acos(x[i+1]))/2;
	    } else if ( x[0] < 0. && i==nx-1 ) {
		dth =  (acos(x[i]) + acos(x[i-1])) / 2;
	    } else if ( x[0] > 0. && i==nx-1 ) {
		dth = pi - (acos(x[i]) + acos(x[i-1]))/2;
	    }
	    nph[i] = pi2 * sqrt(1.0-x[i]*x[i]) / s0 * fabs(dth)/2.+0.5;
	} else {
	    dth = (acos(x[i+1]) - acos(x[i-1])) / 2;
	    nph[i] = pi2 * sqrt(1.0-x[i]*x[i]) / s0 *fabs(dth)+0.5;
	}
	ntot += nph[i];
    }

    return ntot;
}

int get_nphi_by_x(x, nx, i)
double *x;
int nx;
int i;
{
	int  nphi, im;
	double pi = M_PI, pi2=2.*M_PI;
	double s0, dth=0.0;

	im = (nx + 1) / 2;
	/* nphi = 2pi / (width=(theta[im+1]-theta[im-1])/2) */
	nphi = fabs(4.*pi / (acos(x[im+1]) - acos(x[im-1]))) + 0.5;

	/* actual size of one pixel: */
	/*   s0 = dtheta * dphi      */
	s0 = pi2 / nphi * (0.5*fabs(acos(x[im+1]) - acos(x[im-1])));
/*
PR(ER, "pixel_size = %gx%g = %g sq.arcmin, nx=%d, np=%d\n",
0.5*fabs(acos(x[im+1]) - acos(x[im-1]))/pi*180*60,  2. / nphi * 180 * 60.,
s0/pi/pi*60*60*180*180,  nx, nphi);
*/
	/* get sizes by map */
	if ( i==0 || i==nx-1 ) {
	    if ( x[0] > 0. && i==0 ) {
		dth =  (acos(x[i+1]) + acos(x[i])) / 2;
	    } else if ( x[0] < 0. && i==0)  {
		dth = pi - (acos(x[i]) + acos(x[i+1]))/2;
	    } else if ( x[0] < 0. && i==nx-1 ) {
		dth =  (acos(x[i]) + acos(x[i-1])) / 2;
	    } else if ( x[0] > 0. && i==nx-1 ) {
		dth = pi - (acos(x[i]) + acos(x[i-1]))/2;
	    }
	    nphi = pi2 * sqrt(1.0-x[i]*x[i]) / s0 * fabs(dth)/2.+0.5;
	} else {
	    dth = (acos(x[i+1]) - acos(x[i-1])) / 2;
	    nphi = pi2 * sqrt(1.0-x[i]*x[i]) / s0 *fabs(dth)+0.5;
	}
	return nphi;
}

int set_nphi_by_x2p(x, nx, nphi, nph)
double *x;
int nx, nphi;
int *nph;
{
    register int i, ix, ntot=0;
    double dxx;

    i = (nx + 1) / 2;
    nph[i] = nphi;

    /* actual size of one pixel: */
    /*   s0 = dtheta * dphi      */
    dxx = x[i+1]-x[i];

    nph[i+1] = nph[i];
    nph[i-1] = nph[i];
    for (ix=i+1; ix < nx; ix++) {
	double fac=(x[ix]-x[ix-1]) / dxx;
	int kx;
	nph[ix] = nph[i]*fac;
	if(nph[ix] < 6)  nph[ix]=6;
	kx = nx-ix-1;
	if(kx >= 0) {
	    nph[kx] = nph[ix];
	}
    }

for (ix=0; ix < nx; ix++) ntot += nph[ix];
    return ntot;
}
int set_nphi_by_x2(x, nx, nphi, nph)
double *x;
int nx, nphi;
int *nph;
{
    register int i, ntot;
    int  np;
    double pi = M_PI, pi2=2.*M_PI;
    double s0;

    i = (nx + 1) / 2;
    /* better: np = 2pi / (width=(theta[i+1]-theta[i-1])/2), but */
    np = nphi;

    /* actual size of one pixel: */
    /*   s0 = dtheta * dphi      */
    s0 = pi2 / np * (0.5*fabs(acos(x[i+1]) - acos(x[i-1])));
/*
PR(ER, "pixel_size = %gx%g = %g sq.arcmin, nx=%d, np=%d\n",
0.5*fabs(acos(x[i+1]) - acos(x[i-1]))/pi*180*60,  2. / np * 180 * 60.,
s0/pi/pi*60*60*180*180,  nx, np);
*/
    /* set sizes by map */
    for(ntot=i=0; i< nx; i++) {
	double dth=0.0;
/*
	if ( i==0 )
		dth =  (acos(x[i+1]) + acos(x[i])) / 2;
	else if ( i==nx-1 )
		dth = pi - (acos(x[i]) + acos(x[i-1]))/2;
*/
	if ( i==0 || i==nx-1 ) {
	    if ( x[0] > 0. && i==0 ) {
		dth =  (acos(x[i+1]) + acos(x[i])) / 2;
	    } else if ( x[0] < 0. && i==0)  {
		dth = pi - (acos(x[i]) + acos(x[i+1]))/2;
	    } else if ( x[0] < 0. && i==nx-1 ) {
		dth =  (acos(x[i]) + acos(x[i-1])) / 2;
	    } else if ( x[0] > 0. && i==nx-1 ) {
		dth = pi - (acos(x[i]) + acos(x[i-1]))/2;
	    }
	    nph[i] = pi2 * sqrt(1.0-x[i]*x[i]) / s0 * fabs(dth)/2.+0.5;
	} else {
	    dth = (acos(x[i+1]) - acos(x[i-1])) / 2;
	    nph[i] = pi2 * sqrt(1.0-x[i]*x[i]) / s0 *fabs(dth)+0.5;
	}
	ntot += nph[i];
    }

    return ntot;
}

int set_nphi_by_x2pin4(x, nx, nphi, nph)
double *x;
int nx, nphi;
int *nph;
{
    register int i, ntot = 0;
    int nx2=(nx+1)/2;

    nph[nx-1] = nph[0] = 10;

    for(i=1; i<nx2; i++) {
	nph[i] = nph[0] + i*4;
	if (nph[i] > nphi)   nph[i] = nphi;
	nph[nx-1-i] = nph[i];
    }
    for(i=0; i<nx; i++)  ntot += nph[i];
    return ntot;
}

int get_nphi_by_x2(x, nx, np, i)
double *x;
int nx, np;
int i;
{
	int  nphi, im;
	double pi = M_PI, pi2=2.*M_PI;
	double s0, dth=0.0;

	im = (nx + 1) / 2;
	/* better: nphi = 2pi / (width=(theta[im+1]-theta[im-1])/2), but */
	nphi = np;

	/* actual size of one pixel: */
	/*   s0 = dtheta * dphi      */
	s0 = pi2 / nphi * (0.5*fabs(acos(x[im+1]) - acos(x[im-1])));
/*
PR(ER, "pixel_size = %gx%g = %g sq.arcmin, nx=%d, np=%d\n",
0.5*fabs(acos(x[im+1]) - acos(x[im-1]))/pi*180*60,  2. / nphi * 180 * 60.,
s0/pi/pi*60*60*180*180,  nx, nphi);
*/
	/* get sizes by map */
	if ( i==0 || i==nx-1 ) {
	    if ( x[0] > 0. && i==0 ) {
		dth =  (acos(x[i+1]) + acos(x[i])) / 2;
	    } else if ( x[0] < 0. && i==0)  {
		dth = pi - (acos(x[i]) + acos(x[i+1]))/2;
	    } else if ( x[0] < 0. && i==nx-1 ) {
		dth =  (acos(x[i]) + acos(x[i-1])) / 2;
	    } else if ( x[0] > 0. && i==nx-1 ) {
		dth = pi - (acos(x[i]) + acos(x[i-1]))/2;
	    }
	    nphi = pi2 * sqrt(1.0-x[i]*x[i]) / s0 * fabs(dth)/2.+0.5;
	} else {
	    dth = (acos(x[i+1]) - acos(x[i-1])) / 2;
	    nphi = pi2 * sqrt(1.0-x[i]*x[i]) / s0 *fabs(dth)+0.5;
	}
	return nphi;
}

int pget_nphi_by_x(x, nx, npp, i)
double *x;
int nx;
int npp;                          /* number of polar pixels */
int i;
{
	int  nphi;
	double pi = M_PI, pi2=2.*M_PI;
	double s0, dth=0.0;
	double cth;

	if ( x[0] > 0.) {
	    dth =  (acos(x[1]) + acos(x[0])) / 2;
	} else if ( x[0] < 0.) {
	    dth = pi - (acos(x[1]) + acos(x[0]))/2;
	}
	/* actual size of one pixel: 0.5 because of triangle */
	cth = cos(dth);
	s0 = pi2*(1-cth)/npp;

	/* get sizes by map */
	if ( i==0 || i==nx-1 ) {
	    if ( x[0] > 0. && i==0 ) {
		dth =  (acos(x[i+1]) + acos(x[i])) / 2;
	    } else if ( x[0] < 0. && i==0)  {
		dth = pi - (acos(x[i]) + acos(x[i+1]))/2;
	    } else if ( x[0] < 0. && i==nx-1 ) {
		dth =  (acos(x[i]) + acos(x[i-1])) / 2;
	    } else if ( x[0] > 0. && i==nx-1 ) {
		dth = pi - (acos(x[i]) + acos(x[i-1]))/2;
	    }
	    cth = cos(dth);
	    nphi = pi2 / s0 * (1-cth) + 0.5;
	} else {
	    double t0=acos(x[i-1]), t1=acos(x[i]), t2=acos(x[i+1]);
	    double cth0 = cos(t0 + 0.5*(t1-t0));
	    double cth1 = cos(t1 + 0.5*(t2-t1));
	    nphi = pi2 * fabs(cth0 - cth1) / s0 + 0.5;
	}

	return nphi;
}

int get_pol_nphi_by_x(x, nx, np, i)
double *x;
int nx;
int np;                          /* number of pixels in equator */
int i;
{
    int nx2 = (nx+1)/2 - 1, ii = i;
    int n = -1;
    double dx=x[nx2+1]-x[nx2];

    if (0 <= i && i < nx) {
	if (i <= nx2)  ii = nx-i-1;

	if (ii == nx2)
	    n = np;
	else if (ii == nx2-1 && nx%2 == 0)
	    n = np;
	else {
//          double fac = (x[ii]-x[ii-1]) / dx;
	    double fac = fabs(x[i]-x[i+1]) / dx;
	    n  = np * fac + .5;
	    if (n<6) n = 6;
	}
    }
    return n ;
}

