#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "glesp.h"

#define PR      fprintf
#define ER      stderr
#define ABS(x)  ((x) < 0 ? -(x) : (x))

char amp_in_file[161], alm2_in_file[161], alm3_in_file[161],
	 phase_in_file[161];
char file_C_l[161], alm_out_file[161], alm_in_file[161], outp_file[161];
char alm_str[51];     /* string with a_{lm},format as ASCII-file */
char ex_almfile[161], rm_almfile[161];
int  alm_input = 0;   /* alm input string                        */
int  print_alm_flag = 0;
int  print_alm_gnuplot_flag = 0;
int  outp_fflag = 0;  /* default output binary data in FITS format */
int  lmin_print = -1;
int  lmax_print = -1;
double p0=0.0;        /* phase selection value */
double eps=0.1;       /* environment of phase  */
int separ_flag = 0;   /* phase separation flag */
int hp_pol_flag= 0;   /* healpix polarization output */

/* flags for compatibility */
HEAD hd, *ph;

double fwhm = -1.;          /* if > 0, the smoothing beam in arcmin */
int Cl_input=0;             /* flag showing reading Cl  from file   */
int cl_output = 0;
int seed_flag=0,            /* flag to use user seed                */
    random_seed=0;          /* input random seed                    */

extern void setopt(int, char **), usage(void);

char *myname="create_alm";

int main(argc, argv)
int argc;
char **argv;
{
	double *alm = (double *)0,   *Cl = (double *)0,
	       *phase = (double *)0, *Cpl = (double *)0,
	       *amp = (double *)0;
	int lmax=0, Nalm=0, Nbuf=0;
	FILE *fp_out;

	amp_in_file[0] = alm2_in_file[0] = alm3_in_file[0] =
		phase_in_file[0] = '\0';
	file_C_l[0] = alm_out_file[0]=alm_in_file[0]=outp_file[0]='\0';
	ex_almfile[0] = rm_almfile[0] = '\0';

	/* pointer to head structure */
	ph = &hd;

	/* loop over external parameters */
	setopt(argc, argv);

	if (amp_in_file[0] != '\0' && phase_in_file[0] != '\0') {
	      /* if set amplitude and phase */
	      /* then generate a_lm         */

	      read_ph_table(phase_in_file, &phase, &lmax);
	      read_ph_table(amp_in_file,   &amp,   &lmax);

	      Nbuf = 2*(lmax+3);
	      if((alm  =(double*)calloc((unsigned)((lmax+2)*Nbuf),
		   sizeof *alm)) == (double *)0)
		      goto nomem;

	      amph2alm (amp, phase, lmax, alm, Nbuf);

	} else if (alm_input) {
	   if (lmax_print > 0)
		  lmax = lmax_print;
	   else
		  lmax = 3000;
	   Nalm = lmax+2;
	   Nbuf = 2*(lmax+2);

	   if((alm=(double *)calloc((unsigned)((lmax+2)*Nbuf),
		sizeof *alm)) == (double *)0)
		   goto nomem;

	   /* alm input string */
	   if (alm_input == 1) {
		 /* read alm from ASCII or binary file */
		 read_alm_table (alm_in_file, &alm, &Nalm, &Nbuf, &lmax);
	   } else {

		 /* read alm from ASCII string */
		 read_alm_str (alm_str, alm, Nbuf, lmax);
	   }
	   if((phase=(double*)calloc((unsigned)((lmax+2)*Nbuf),
		sizeof *phase)) == (double *)0)
		   goto nomem;


	} else {
try_again:
	   if (*file_C_l == '\0') {
		PR(ER,"%s: no file with Cl. Please, name of Cl-file ? => ",
			myname);
		fgets(file_C_l, 160, stdin);
		if ( *file_C_l == '\0') {
			usage();
		}
		if (file_C_l[strlen(file_C_l)-1] == '\n')
			file_C_l[strlen(file_C_l)-1] = '\0';
		if ( *file_C_l == '\0')
			goto try_again;
	   }

	   /* create C_l and Cp_l arrays and  */
	   /* reads data from file CMBfast    */
	   if (Cl_input==1) {
		   read_cl_table (file_C_l, &Cl, &Cpl, &lmax);
		   cl_output = 1;
	   } else
		   read_dl_table (file_C_l, &Cl, &Cpl, &lmax);

	   Nbuf = 2*(lmax+2);

	   if((alm=(double *)calloc((unsigned)((lmax+2)*Nbuf),
		sizeof *alm)) == (double *)0) {
nomem:
		   PR(ER,"%s: no memory for arrays[%d]\n",
		       myname, (lmax+2)*Nbuf);
		   exit(1);
	   }

	   /* simulation a_lm by C_l without phase */
	   cl2alm (Cl, lmax, alm, Nbuf, fwhm);
	}

	/* extract a_lm */
	if (*ex_almfile != '\0')
		alm_extract(alm, lmax, Nbuf, ex_almfile);

	/* remove a_lm */
	if (*rm_almfile != '\0')
		alm_remove(alm, lmax, Nbuf, rm_almfile);

	if (lmax_print < 0 || lmax_print > lmax)
		lmax_print = lmax;
	if (lmin_print <= 0 || lmin_print > lmax)
		lmin_print = 0;
	else {
		register int l, m;
		for(l=0; l<lmin_print; l++) {
		    int ll = l * Nbuf;
		    for(m=0; m<=l; m++)  alm[ll+m] = alm[ll+m+lmax+1] = 0.0;
		}
	}

	/* select phases */
	if (separ_flag) {
	    separ_phase(alm, lmax, Nbuf, p0, eps, separ_flag, lmax_print);
	}


	/**********/
	/* output */
	/**********/
	if (outp_file[0] == '\0' || strcmp(outp_file, "stdout") == 0)
	    fp_out = stdout;
	else
	    if ((fp_out = fopen(outp_file, "w")) == (FILE *)0) {
		PR(ER, "%s: can not create file '%s'\n", myname, outp_file);
		exit(1);
	    }

	/* create file and save a_lm data */
	if ( *alm_out_file != '\0' ) {
	    FILE *fp;
	    PR(ER, "%s: file '%s' of a_{lm} is created (lmax=%d)...",
		     myname, alm_out_file, lmax);
	    fp = fopen(alm_out_file, "w");
	    if (fp == (FILE *)0) {
		PR(ER,"\n%s: can not create this file...\n", myname);
	    } else {
		if (outp_fflag)
		    ff_write_alm(alm, Nbuf, lmax, lmax_print, fp);
		else {
		    /* write 1st head and data of a-lm */
		    out_glesp_alm(alm, lmax, Nbuf, lmax_print, fp);
		    /* write 2nd & 3d head and data of polarization */
		    if(hp_pol_flag) {
			double *alm2 = (double *)0;
			read_alm_table (alm2_in_file, &alm2,
				 &Nalm, &Nbuf, &lmax);
			if (*alm3_in_file != '\0') {
			  double *alm3 = (double *)0;
			  read_alm_table (alm3_in_file, &alm3,
				   &Nalm, &Nbuf, &lmax);
			  add_alm_ext(alm2,alm3,lmax,Nbuf,lmax_print,fp);
			  if (alm3 != (double *)0) free((void *)alm3);
			} else
			  add_alm_ext(alm2,alm2,lmax,Nbuf,lmax_print,fp);
			if (alm2 != (double *)0) free((void *)alm2);
		    }
		}
		fclose (fp);
		PR(ER, "Done.\n");
	    }
	}

	if (print_alm_flag) {
		/* calculate phase */
		phase_calc(alm, Nbuf, phase, lmax);

		print_alm_table (fp_out, alm, lmax, Nbuf, phase,
			lmin_print, lmax_print);
	} else if (print_alm_gnuplot_flag) {
		print_alm_plot (fp_out, alm, lmax, Nbuf,
			  lmin_print, lmax_print);
	} else {
		if((Cl=(double*)calloc((unsigned)((lmax+1)), sizeof *Cl)) ==
		   (double *)0)
			goto nomem;

		if (cl_output) {
/*
			PR(fp_out, "# Cl-s:\n");
			PR(fp_out, "#------------------------\n");
			PR(fp_out, "#   l      C_l      Cp_l:\n");
*/
			alm2cl (Cl, lmax, alm, Nbuf);
		} else {
			/* calculate Dl */
/*
			PR(fp_out, "# Dl-s:\n");
			PR(fp_out, "#------------------------\n");
			PR(fp_out, "#   l      D_l      Dp_l:\n");
*/
			alm2dl (Cl, lmax, alm, Nbuf);
		}

		/* print Cl/Dl     */
		print_cl_table (fp_out, Cl, Cpl, lmin_print, lmax_print);
	}

	if(fp_out != (FILE *)0 && fp_out != stdout) fclose(fp_out);

	exit(0);
}

void setopt(argc, argv)
int argc;
char **argv;
{
	register int i;
	myname = argv[0];

	for(i=1; i<argc; i++) {
	    switch(argv[i][0]) {
	    case '-':
		switch(argv[i][1]) {
		case 'a':
			switch(argv[i][2]) {
			case 'i':
				if (argv[i][3] == '2') {
				    strcpy(alm_in_file, argv[++i]);
				    strcpy(alm2_in_file, argv[++i]);
				    alm_input = 1; /* alm from file */
				} else if (argv[i][3] == '3') {
				    strcpy(alm_in_file,  argv[++i]);
				    strcpy(alm2_in_file, argv[++i]);
				    strcpy(alm3_in_file, argv[++i]);
				    alm_input = 1; /* alm from file */
				    hp_pol_flag = 1;
				} else {
				    strcpy(alm_in_file, argv[++i]);
				    alm_input = 1; /* alm from file */
				}
				break;
			case 'm':
				strcpy(amp_in_file, argv[++i]);
				break;
			case 's':
				strcpy(alm_str, argv[++i]);
				alm_input = 2; /* alm from string */
				break;
			case 'o':
			default :
				if (strncmp(argv[i]+2, "of", 2) == 0) {
				    outp_fflag = 1; /* output in F-format */
				}
				strcpy(alm_out_file, argv[++i]);
				break;
			}
			break;
		case 'd':
			if (strcmp(argv[i]+1, "double") == 0) {
				bitpix_map = DOUBLEPIX;
			} else {
				PR(ER,"%s: illegal flag '%s'\n",
				    myname, argv[i]);
				usage();
			}
			break;
		case 'D':
			strcpy(file_C_l, argv[++i]);
			Cl_input = 2;
			break;
		case 'c':
			/* print Cl */
			cl_output = 1;
			break;
		case 'C':
			strcpy(file_C_l, argv[++i]);
			Cl_input = 1;
			break;
		case 'e':
			if (strncmp(argv[i]+1, "ep", 2) == 0) {
				eps = atof(argv[++i]);
			} else {
				/* extract alm according to file */
				strcpy(ex_almfile, argv[++i]);
			}
			break;
		case 'f':
			if (strcmp(argv[i]+1, "fa") == 0) {
				strcpy(alm_in_file, argv[++i]);
				alm_input = 1; /* alm from file */
			} else if (strncmp(argv[i]+1, "flo",3) == 0) {
				bitpix_map = FLOATPIX;
			} else {
				PR(ER,"%s: illegal flag '%s'\n",
				    myname, argv[i]);
				usage();
			}
			break;
		case 'g': /* print alm in GNU Plot format */
			print_alm_gnuplot_flag = 1;
			break;
		case 'h':
			if (strncmp(argv[i]+1, "hp", 2) == 0) {
				hp_pol_flag = 1;
			} else
				usage();
			break;
		case 'i':
			switch(argv[i][2]) {
			default :
				usage();
				break;
			}
			break;
		case 'l':
			if (strncmp(argv[i]+1, "lmn",3) == 0 ||
			    strncmp(argv[i]+1, "lmi",3) == 0 )
				lmin_print = atoi(argv[++i]);
			else
				lmax_print = atoi(argv[++i]);
			break;
		case 'o':
			strcpy(outp_file, argv[++i]);
			break;
		case 'p':
			if (strncmp(argv[i]+1, "ph", 2) == 0) {
			      strcpy(phase_in_file, argv[++i]);
			} else if (strncmp(argv[i]+1, "p0", 2) == 0) {
			      if (argv[i][3] == 'r')  separ_flag = 2;
			      else separ_flag = 1;
			      p0 = atof(argv[++i]);
			} else
			      print_alm_flag = 1;
			break;
		case 'r':
			if (strncmp(argv[i]+1, "rm", 2) == 0) {
				/* remove alm from file */
				strcpy(rm_almfile, argv[++i]);
			} else {
				/* random seed */
				random_seed = atoi(argv[++i]);
				seed_flag = 1;
			}
			break;
		default :
			usage();
			break;
		}
		break;
	    case '+':
		switch(argv[i][1]) {
		case 'h':
			usage();
		default :
			usage();
			break;
		}
		break;
	    default :
		strcpy(alm_out_file, argv[i]);
		break;
	    }
	}

	return;
}

void usage()
{
PR(ER,"usage: %s -Cl/Dl inp_file_with_Cl -a alm_out_file [-o Cl_out_file] [-p]\n",
myname);
PR(ER,"[-g(GNU Plot output)] [-l lmax_print] [-r random_seed] [-cl]\n");
PR(ER,"[-ai[{2,3}] alm_inp_file] [-as alm_str] [-lmin lmin_print]\n");
PR(ER,"[-ph phase_file -amp amplitude_file] [-ep eps_phase] [-p0[r] phase_only]\n");
PR(ER,"[-ex extract_alm_file] [-rm remove_alm_file] [-{double,float}]\n");
exit(2);
}

