#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "glesp.h"

#define PR  fprintf
#define ER  stderr

extern void setopt(int, char **), usage(void);
extern void map_d2t2_(double *, int *, int *),
	    map_ddtt(double *, double *, int, int),
	    map_dt_(double *, int *, int *),
	    map_dt(double *, double *, int, int),
	    map_dp_(double *, int *, int *);

char alm_file[161], outp_file[161];
int ffile_flag = 0;    /* default output in FITS-format */
int  cut_lmin = -1;    /* cut value for a_lm        */
int  cut_lmax = 0;     /* cut value for a_lm        */
int  phi_deriv = 0;

HEAD *ph;

int naxis2=0;
int deriv = 1;
int lmax = -1;

char *myname="dalm";

int main(argc, argv)
int argc;
char **argv;
{
    double *alm = (double *)0;
    int nbuf=0, nalm=0, lmax_new = lmax, nbuf_new;
    FILE *fp_out = (FILE *)0;

    alm_file[0] = outp_file[0] = '\0';

    /* loop over external parameters */
    setopt(argc, argv);

    read_alm_table(alm_file, &alm, &nalm, &nbuf, &lmax);
    lmax_new = lmax;
    nbuf_new = 2 * (lmax + 4);
    alm2alm(&alm, &lmax, &nbuf, lmax_new, nbuf_new);


    switch (deriv) {
    case 'p':
	map_dp_(alm, &nbuf, &lmax);
	if (phi_deriv) map_dp_(alm, &nbuf, &lmax);
	lmax_new = lmax;
	break;
    case 1 :
	map_dt_(alm, &nbuf, &lmax);
	lmax_new = lmax+1;
	if (phi_deriv) map_dp_(alm, &nbuf, &lmax_new);
	break;
    default:
    case 2 :
	map_d2t2_(alm, &nbuf, &lmax);
	lmax_new = lmax+2;
	break;
    }
    if (cut_lmax <= 2)  cut_lmax = lmax_new;

    /* open output file */
    if (outp_file[0] == '\0' || strcmp(outp_file, "stdout") == 0)
	fp_out = stdout;
    else
	if ((fp_out = fopen(outp_file, "w")) == (FILE *)0) {
	    PR(ER, "%s: can not create file '%s'\n", myname, outp_file);
	    exit(1);
	}

    /* write new a_{lm}-s */
    if (ffile_flag)
	ff_write_alm(alm, nbuf, lmax, cut_lmax, fp_out);
    else
	out_glesp_alm(alm, lmax, nbuf, cut_lmax, fp_out);

    /* close file */
    if(fp_out != (FILE *)0 && fp_out != stdout) fclose(fp_out);

    free((void *)alm);

    exit(0);
}

void setopt(argc, argv)
int argc;
char **argv;
{
    register int i;
    myname = argv[0];

    for(i=1; i<argc; i++) {
	switch(argv[i][0]) {
	case '-':
		switch(argv[i][1]) {
		case 'a':
			strcpy(alm_file, argv[++i]);
			break;
		case 'd':
			switch (argv[i][2]) {
			case 'p':
				switch (argv[i][3]) {
				case '2':
				case 'p':
				       deriv = 'p';
				       phi_deriv = 1;
				       break;
				case 't':
				       deriv = 1;
				       phi_deriv = 1;
				       break;
				case '1':
				default :
				       deriv = 'p';
				       break;
				}
				break;
			case 't':
				switch (argv[i][3]) {
				case 'p':
				       deriv = 1;
				       phi_deriv = 1;
				       break;
				case 't':
				       deriv = 2;
				       break;
				case '2':
				       deriv = 2;
				       break;
				case '1':
				default :
				       deriv = 1;
				       break;
				}
				break;
			case '1':
				deriv = 1;
				break;
			case '2':
			default :
				deriv = 2;
				break;
			}
			break;
		case 'h':
			usage();
			break;
		case 'l':
		case 'L':
			lmax = atoi (argv[++i]);
			break;
		case 'o':
			/* output file */
			if (strncmp(argv[i]+1, "of", 2) == 0) {
			    ffile_flag = 1; /* output in F-format */
			}
			strcpy(outp_file, argv[++i]);
			break;
		default:
			break;
		}
		break;
	case '+':
		switch(argv[i][1]) {
		case 'h':
		default:
			usage();
			break;
		}
		break;
	default:
		strcpy(alm_file, argv[i]);
		break;
	}
    }

    return;
}

void usage()
{
PR(ER, "usage: %s -a alm_file [-d{tp12}] [-o output_alm_file]\n", myname);
exit(2);
}

void map_d2t2_(alm, nbuf, lmax)
double *alm;
int *nbuf, *lmax;
{
    register int l, m;
    double *alm2;
    int lm2 = *lmax + 2;

    if((alm2 = (double*)calloc(*nbuf*(*lmax+3),sizeof *alm2))==(double*)0) {
	    PR(ER, "map derivative: no memory for temporal arrays...\n");
	    exit(1);
    }
    map_ddtt(alm, alm2, *nbuf, *lmax);

    for (l=0; l <= lm2; l++) {
	int ll = l * *nbuf;
	for (m=0; m<=l; m++) {
	     alm[ll+m]            = alm2[ll+m];
	     alm[ll+m+ *lmax+2+1] = alm2[ll+m+ *lmax+2+1];
	}
    }

    free((void *)alm2);
}

# define  A(l,m)        ( (l)*(l-(m)+1)  / (double)(2*(l)+1))
# define  B(l,m)        ( -(l+1)*(l+m)   / (double)(2*(l)+1))

double alp_cf(l, m)
int l, m;
{
    double a = 0.0;

    if ( l != m && l != m+1) {
	a = (double)(l+1)/l * B(l, m) * B(l-1, m);
    }
    if (fabs(a) > 1.e-30) {
	 a *= 1./sqrt((double)(2*l-3)/(2*l+1)*(l+m)/(l-m)*(l+m-1)/(l-m-1));
    }
    return a;
}

double bet_cf(l, m)
int l, m;
{
    double b = 0.0;

    if ( l == m ) {
	b = (double)(m+3) / (m+2) * A(m,m) * B(m+1,m);
    } else if ( l == m+1) {
	b = (double)(m+4)/(m+3)*A(m+1,m)*B(m+2,m) +
		(double)(m-1)/(2*m+1)*B(m+1,m);
    } else {
	b = (double)(l+3)/(l+2) * A(l,m) * B(l+1, m) +
		(double)(l-2)/(l-1)*B(l,m)*A(l-1,m);
    }
    return b;
}

double gam_cf(l, m)
int l, m;
{
    double g = 0.0;

    if ( l == m ) {
	g = (double) m / (m+1) * A(m,m) * A(m+1,m);
    } else if ( l == m+1) {
	g = (double)(m+1) / (m+2) * A(m+1,m) * A(m+2,m);
    } else {
	g = (double)l/(l+1) * A(l,m) * A(l+1, m);
    }
    if (fabs(g) > 1.e-30)
	g *= 1./sqrt((double)(2*l+5)/(2*l+1)*(l-m+2)/(l+m+2)*(l-m+1)/(l+m+1));

    return g;
}

void map_ddtt(alm, alm2, nbuf, lmax)
double *alm, *alm2;
int lmax, nbuf;
{
    register int l, m;
    int lmax2 = lmax+2;

    for(m=0; m<=lmax; m++) {
	for(l=m; l<=lmax2; l++) {
	    int ll    = l * nbuf;
	    int ll_m2 = (l-2) * nbuf;
	    int ll_p2 = (l+2) * nbuf;
	    double re1=0.0, re2=0.0, re3=0.0,
		   im1=0.0, im2=0.0, im3=0.0;

	    if (l<lmax-1) {
		re1 = alp_cf(l+2, m) * alm[ll_p2+m];
		im1 = alp_cf(l+2, m) * alm[ll_p2+m+lmax+1];
	    }
	    re2 = bet_cf(l, m) * alm[ll+m];
	    im2 = bet_cf(l, m) * alm[ll+m+lmax+1];
	    if (l>1) {
		re3 = gam_cf(l-2, m) * alm[ll_m2+m];
		im3 = gam_cf(l-2, m) * alm[ll_m2+m+lmax+1];
	    }

	    alm2[ll+m]         += re1 + re2 + re3;
	    alm2[ll+m+lmax2+1] += im1 + im2 + im3;
	}
    }

    return;
}

void map_dp_(alm, nbuf, lmax)
double *alm;
int *lmax, *nbuf;
{
    register int l, m;

    for(l=0; l<= *lmax; l++) {
	int ll = l * *nbuf;
	for(m=0; m<= l; m++) {
	    double re, im;

	    re = alm[ll+m];
	    im = alm[ll+m+ *lmax+1];
	    alm[ll+m]          =  im * m;
	    alm[ll+m+ *lmax+1] = -re * m;
	}
    }

    return;
}
void map_dt(alm, alm2, nbuf, lmax)
double *alm, *alm2;
int lmax, nbuf;
{
    register int l, m;
    int lmax1 = lmax+1;

    for(m=0; m<=lmax; m++) {
	for(l=m; l<=lmax1; l++) {
	    int lm1 = l-1, lp1 = l+1;
	    int ll    = l * nbuf;
	    int ll_m1 = lm1 * nbuf;
	    int ll_p1 = lp1 * nbuf;
	    double re1=0.0, re2=0.0,
		   im1=0.0, im2=0.0;

	    if (l !=0 && lm1 != m-1 && !(lm1==0 && m==0)) {
		double k = sqrt((lm1+m+1.)/(lm1-m+1)*(2*lm1+1)/(2*lm1+3));
		re1 = A(lm1, m) * k * alm[ll_m1+m];
		im1 = A(lm1, m) * k * alm[ll_m1+m+lmax+1];
	    }
	    if (l < lmax) {
		double k=sqrt((double)(lp1-m)/(lp1+m)*(2*lp1+1)/(2*lp1-1));
		re2 = B(lp1, m) * k * alm[ll_p1+m];
		im2 = B(lp1, m) * k * alm[ll_p1+m+lmax+1];
	    }
	    alm2[ll+m]         += re1 + re2;
	    alm2[ll+m+lmax1+1] += im1 + im2;
	}
    }

    return;
}

void map_dt_(alm, nbuf, lmax)
double *alm;
int *nbuf, *lmax;
{
    register int l, m;
    double *alm2;
    int lm1 = *lmax + 1;

    if((alm2 = (double*)calloc(*nbuf*(*lmax+3),sizeof *alm2))==(double*)0) {
	    PR(ER, "map derivative: no memory for temporal arrays...\n");
	    exit(1);
    }

    map_dt(alm, alm2, *nbuf, *lmax);

    for (l=0; l <= lm1; l++) {
	int ll = l * *nbuf;
	for (m=0; m<=l; m++) {
	     alm[ll+m]        = alm2[ll+m];
	     alm[ll+m+ lm1+1] = alm2[ll+m+ lm1+1];
	}
    }

    free((void *)alm2);
}

