#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

/*
 *  program of arithmetic (and not only) operations over a_lm-s
 *
 *  created by Oleg Verkhodanov, 2002-2004
 *  last changes 12/04/2006,  21/07/2006,  26/01/2012, 18/04/2013
 *               2015
 */
#include "glesp.h"

#define PR      fprintf
#define ER      stderr
#define ST      stdout
#define ABS(x)  ((x) < 0 ? -(x) : (x))

HEAD *ph;

int naxis2;
char filename1[161], filename2[161], v_file[161];
char out_file[161];
int part_flag = 0;        /* type of value with relative error or */
			  /* output number flag */
int  lmax_print = -1, lmin_print = -1;

int action = 0;           /* action: 0 - difference, 1 - sum */
int cut_alm_flag=0;
int change_phase_flag = 0;
int mask_flag = 0;
int corr_phase_flag = 0;  /* flag to find phase correlations    */
int sin_flag = 0;
int alm_stat_flag = 0;    /* calculate statistics for alm-s     */
int out_flag = 1;         /* output alm F-file                  */
int flag_tl = 0;          /* calc Tl moments                    */
int coor_sys = 0;         /* galactic coordinates to equatorial */
int cl_compare_flag = 0;  /* flag to compare cl-s of 2 alm-s    */
int nbin = 0;             /* number of intervals for binnning   */
int max_deviation=0;
int gp_output=0;          /* hnuplot ASCII format output        */
int corr_shift_factor= -1;/* correlation shift dl               */

int sd_l=0, sd_m = 0;     /* value of shifting step for l & m   */
int shift_multip_flag = 0;/* flag to use multipole shifting     */
int phase_differ_flag = 0;/* calculate phase differences        */

extern void alm_ks_test(double *, int, int, int, int, int, FILE *, int);
extern void setopt(int, char **), usage(void), usage2(void);
extern void set_sig_fatal(void);

int rotate_flag = 0;
double dpsi = 0.0, dphi = 0.0, dtheta = 0.0;

int  l_calc = -1, m_calc = -1,
     n_low = 100;
int  sign = 0;
int abs_dif = 0;
double phase_ch = 0.0;
double c1=1.0, c2=1.0, cf = 1.0;
int amplit_1_flag = 0;    /* set amplitude = 1     */
int amplit_sc_flag = 0;   /* cos(dphi), sin(dphi)  */
int phase_conv_flag = 0;  /* phase conversion flag */
int l_vector_flag = 0;    /* multiply by vector(l) */
extern double ph_calc(double, double);
extern int sun_data_key;
int ffile_flag = 0;       /* default output in FITS-format */
int relative_differ =0;   /* (a1-a2) / (.5*(a1+a2))        */
int average_phase = 0;
int inverse_flag = 0;     /* inversion of a map            */
int abs_flag = 0;         /* take abs. values of Re & Im   */

int phase_set_limits = 0; /* set phase selection by values */
double phase_min=0, phase_max=2*M_PI;  /* phase limits     */


char *myname="difalm";

int main(argc, argv)
int argc;
char **argv;
{
    register int l=0, m=0;
    double *alm1=(double *)0, *alm2=(double *)0, *vect=(double *)0;
    int lmax=0, nalm=0, nbuf=0,
	lmx1=0, lmx2=0, nbuf1=0, nbuf2=0, nalm1=0, nalm2=0, v_lmax=0;
    FILE *fp_d = (FILE *)0;


    filename1[0]=filename2[0]=v_file[0]=out_file[0] = '\0';

    /* catch signal and finish program */
    set_sig_fatal();

    /* loop over external parameters */
    setopt(argc, argv);

    read_alm_table(filename1, &alm1, &nalm1, &nbuf1, &lmx1);
    lmax = lmax_print > lmx1 ? lmax_print : lmx1;

    if ( *filename2 != '\0' ) {
	  read_alm_table(filename2, &alm2, &nalm2, &nbuf2, &lmx2);
	  lmax = lmax < lmx2 ? lmx2 : lmax;
    }
    nbuf = 2*(lmax+2);
    nalm = lmax + 1;

    alm2alm(&alm1, &lmx1, &nbuf1, lmax, nbuf);
    if ( *filename2 != '\0' )
	  alm2alm(&alm2, &lmx2, &nbuf2, lmax, nbuf);
/*
{
for(l=0; l<=lmax; l++)
for(m=0; m<=l; m++) {
double re1 = alm1[nbuf*l+m];
double im1 = alm1[nbuf*l+m+lmax+1];
double re2 = alm2[nbuf*l+m];
double im2 = alm2[nbuf*l+m+lmax+1];
PR(stdout, "(%d,%d) %16.9e %16.9e %16.9e %16.9e\n", l,m,re1, im1, re2, im2);
}}
*/
    /* read vector of C(l) */
    if (l_vector_flag) {
	  double *cpl = (double *)0;
	  read_cl_table (v_file, &vect, &cpl, &v_lmax);
	  if (cpl != (double *)0) free((void *)cpl);
    }

    if (lmax_print <= 0 || lmax_print > lmax)
	    lmax_print = lmax;
    if (lmin_print <  0 || lmin_print > lmax)
	    lmin_print = 0;

    if (lmin_print) {
	for (l=0; l<lmin_print; l++) {
	    register int ll = l * nbuf;
	    for (m=0; m<=l; m++) {
		alm1[ll+m] = alm1[ll+m+lmx1+1] = 0.0;
		if ( *filename2 != '\0' ) {
		       alm2[ll+m] = alm2[ll+m+lmx2+1] = 0.0;
		}
	    }
	}
    }

    /*------------------*/
    /* open output file */
    /*------------------*/
    /* compulsory opening */
    if ( *out_file == '\0')  fp_d = stdout;
    else if (strcmp(out_file, "stdout") == '\0' ||
       strcmp(out_file, "-") == '\0') {
	    fp_d = stdout;
    } else {
       if((fp_d = fopen(out_file, "w")) == (FILE *)0) {
	    PR(ER,"%s: can not create file '%s'\n", myname, out_file);
	    exit(1);
       }
    }

    if (corr_phase_flag) { /* flag to find phase correlations */
	double *r;
	extern void corr_phase(double *, double *, double *, int, int, int);

	if ((r=(double*)calloc((unsigned)(lmax+1),sizeof *r))==(double *)0){
nomem:
	    PR(ER, "%s: no memory for arrays...\n", myname);
	    exit(1);
	}

	corr_phase(r, alm1, alm2, nbuf, lmax, sin_flag);

	for(l=0; l<=lmax; l++)
	    PR(fp_d, "%4d %12.5e\n", l, r[l]);

	if (fp_d != (FILE *)0 && fp_d != stdout) fclose(fp_d);
	exit(0);
    }

    /***********************/
    /* a_{lm} - operations */
    /***********************/
    if (phase_conv_flag) {
	    for (l=lmin_print; l<=lmax_print; l++) {
		register int ll = l * nbuf;
		for (m=0; m<=l; m++) {
		    double re1, im1;
		    double re2, im2;
		    double amp, ph;

		    re1 = alm1[ll+m];
		    re2 = alm2[ll+m];
		    im1 = alm1[ll+m+lmx1+1];
		    im2 = alm2[ll+m+lmx2+1];

		    amp = sqrt(re1*re1 + im1*im1);
		    if (m==0) {
			if (re2 < 0.0)  ph = M_PI; else ph = 0.0;
		    } else {
			ph  = ph_calc(re2, im2);
		    }
		    if (phase_conv_flag == 2) {
			alm1[ll+m]        -= amp * cos(ph);
			alm1[ll+m+lmax+1] -= amp * sin(ph);
		    } else {
			alm1[ll+m]        = amp * cos(ph);
			alm1[ll+m+lmax+1] = amp * sin(ph);
		    }
		}
	    }
    } else if(amplit_1_flag) {
	for (l=lmin_print; l<=lmax_print; l++) {
	    register int ll = l * nbuf;
	    alm1[ll] = 1.0;
	    for (m=1; m<=l; m++) {
		double re = alm1[ll+m], im = alm1[ll+m+lmax+1];
		double v = sqrt(re*re + im*im);
		alm1[ll+m]        /= v;
		alm1[ll+m+lmax+1] /= v;
	    }
	}
    } else if(amplit_sc_flag) {
	int l0 = lmin_print > 3 ? lmin_print+1 : 3;
	int l1 = lmax_print - 1;

	if ((alm2=(double*)calloc((unsigned)(nalm*nbuf),sizeof *alm2))==
	  (double *)0)
	      goto nomem;

	for (l=l0; l<=l1; l++) {
	    register int ll  = l * nbuf;
	    register int ll1 = (l-1) * nbuf;
	    register int ll2 = (l+1) * nbuf;
	    alm2[ll] = 1.0;
	    for (m=1; m<=l; m++) {
		double re1 = alm1[ll1+m], im1 = alm1[ll1+m+lmax+1];
		double re2 = alm1[ll2+m], im2 = alm1[ll2+m+lmax+1];
		double ph1 = ph_calc(re1, im1),
		       ph2 = ph_calc(re2, im2);
		alm2[ll+m]        = cos(ph1-ph2);
		alm2[ll+m+lmax+1] = sin(ph1-ph2);
	    }
	}
	free((void *)alm1);
	alm1 = alm2;
    } else if(phase_differ_flag) {
	int l0 = lmin_print;
	int l1 = lmax_print;
	double dph = 0.0;

	for (l=l0; l<=l1; l++) {
	    register int ll  = l * nbuf;
	    PR(fp_d, " %4d %4d %12.5e\n", l, 0, 0.0);
	    for (m=1; m<=l; m++) {
		double re1 = alm1[ll+m], im1 = alm1[ll+m+lmax+1];
		double re2 = alm2[ll+m], im2 = alm2[ll+m+lmax+1];
		double ph1 = ph_calc(re1, im1),
		       ph2 = ph_calc(re2, im2);
		dph = ph1 - ph2;
		if (dph < 0.)  dph += 2. * M_PI;
		PR(fp_d, " %4d %4d %12.5e\n", l, m, dph);
	    }
	}
	if (fp_d != (FILE *)0 && fp_d != stdout) fclose(fp_d);
	free((void *)alm1);
	free((void *)alm2);
	exit(0);
    } else if (cl_compare_flag) {
	/* compare of Cl-s of 2 alm samples */

	double *cl1 = (double *)0, *cl2 = (double *)0;

	if ( *filename2 == '\0' ) {
	    PR(ER, "%s: no second file with alm-s...\n", myname);
	    exit(1);
	}

	if ((cl1 = (double *)calloc(lmax+1, sizeof *cl1)) == (double *)0)
		goto nomem;
	if ((cl2 = (double *)calloc(lmax+1, sizeof *cl2)) == (double *)0)
		goto nomem;

	alm2cl (cl1, lmax, alm1, nbuf);
	alm2cl (cl2, lmax, alm2, nbuf);

	if (nbin != 0) {  /* using binning */
	    register int n;
	    double db = log10((double)lmax) / nbin; /* bin step, logarithm */
	    for(n=0; n<nbin;  n++) {
		double s = 0.0, dcl;
		int k=0, ls = 0;
		int l1, l2;

		l1 = pow(10., n * db);
		l2 = pow(10., (n+1) * db);
		for(l=l1; l<l2;  l++) {
		    if (cl2[l] > 1.e-40)  {
			if (max_deviation) {
			    dcl = cl1[l]/cl2[l] - 1.;
			    if (ABS(dcl) > s) {
				s = ABS(dcl);
				k = l;
			    }
			} else {
			    dcl = cl1[l]/cl2[l] - 1.;
			    k++; ls += l;
			    s += ABS(dcl);
			}
		    }
		}
		if (max_deviation) {
		      PR(fp_d, "%4d %12.5e, l:(%d,%d)\n", k, s, l1,l2-1);
		} else {
		    if (k) {
		      s /= k; ls = ls/(double)k + .5;
		      PR(fp_d, "%4d %12.5e, l:(%d,%d)\n", ls, s, l1,l2-1);
		    }
		}
	    }
	} else {
	    for(l=0; l<=lmax;  l++) {
		double dcl;
		if ( cl2[l] > 1.e-40)  {
		    dcl = cl1[l]/cl2[l] - 1.;
		    PR(fp_d, "%4d  %12.5e\n", l, ABS(dcl));
		}
	    }
	}
	if (cl1 != (double *)0)   free((void *)cl1);
	if (cl2 != (double *)0)   free((void *)cl2);
	exit(0);

    } if (corr_shift_factor >= 0) {  /* correlation shift dl */
	extern void k_l_nn(double *, double *, int, int, FILE *, int);
	int nn = corr_shift_factor;

	k_l_nn(alm1, alm2, lmax, nbuf, fp_d, nn);

	if (fp_d != (FILE *)0 && fp_d != stdout) fclose(fp_d);
	exit(0);
    } else if ( *filename2 != '\0' ) {
      int flag_i = 1, flag_r = 1;

      if (abs_dif) action = 20;

      c1 *= cf;
      c2 *= cf;
      if (part_flag == 1)      flag_i = 0; /* only real part is changed */
      else if (part_flag == 2) flag_r = 0; /* only imag part is changed */
      for(l=lmin_print; l<=lmax_print; l++) {
	  int ll = l * nbuf;
	  for(m=0; m<=l; m++) {
	    int mb = m+lmax+1;
	    double re1=alm1[ll+m], im1=alm1[ll+m+lmx1+1];
	    double re2=alm2[ll+m], im2=alm2[ll+m+lmx2+1];

	    switch (action) {
	    case 20: /* difference of absolute values */
		 re1 = sqrt(re1*re1+im1*im1) - sqrt(re2*re2+im2*im2);
		 im1 = 0.0;
		 break;
	    case 1 :  /* sum */
		 if(flag_r) re1 = c1*re1 + c2*re2;
		 if(flag_i) im1 = c1*im1 + c2*im2;
		 break;
	    case 2 :  /* multiplication */
		 if(flag_r) re1 = c1*re1 * c2*re2;
		 if(flag_i) im1 = c1*im1 * c2*im2;
		 break;
	    case 0 :
	    default:
		 { double re=re1, im=im1;
		    if(flag_r) re = c1*re1 - c2*re2;
		    if(flag_i) im = c1*im1 - c2*im2;

		    if (relative_differ) {
			if (relative_differ == 2) {
			    if (fabs(re1+re2) < 1.e-20)
				re1 = 0.;
			    else
				re1 = 2.* re / (re1+re2);
			    if (fabs(im1+im2) < 1.e-20)
				im1 = 0.;
			    else
				im1 = 2.* im / (im1+im2);
			} else {
			    if (fabs(c2*re2) < 1.e-20)
				re1 = 0.0;
			    else
				re1 = re / (c2*re2);
			    if (fabs(c2*im2) < 1.e-20)
				im1 = 0.0;
			    else
				im1 = im / (c2*im2);
			}
		    } else {
			re1 = re;
			im1 = im;
		    }
		 }
		 break;
	    }
	    alm1[ll+m]  = re1;
	    alm1[ll+mb] = im1;
	  }
      }

    } else if(mask_flag) {
	;
    }
    if (l_vector_flag) {
	for (l=lmin_print; l<=lmax_print; l++) {
	    register int ll = l * nbuf;
	    double v = v_lmax < l ? 1.:(l_vector_flag==1? 1./vect[l]:vect[l]);
	    alm1[ll] *= v;
	    for (m=1; m<=l; m++) {
		alm1[ll+m]        *= v;
		alm1[ll+m+lmax+1] *= v;
	    }
	}
    }

    /* rotation */
    if (rotate_flag) {
	alm_rotate(alm1, lmax, nbuf, dphi, dtheta, dpsi);
    }
    if (coor_sys) {
	switch (coor_sys) {
	case 'q':   /* galactic to equatorinal */
	    cr2cr_alm(alm1, lmax, nbuf, 'g', 'q');
	    break;
	case 'e':   /* galactic to ecliptic */
	    cr2cr_alm(alm1, lmax, nbuf, 'g', 'e');
	    break;
	case 3:     /* ecliptic to galactic */
	    cr2cr_alm(alm1, lmax, nbuf, 'e', 'g');
	    break;
	case 4:     /* equatorinal to galactic */
	    cr2cr_alm(alm1, lmax, nbuf, 'q', 'g');
	    break;
	case 5:     /* ecliptic to equatorinal */
	    cr2cr_alm(alm1, lmax, nbuf, 'e', 'q');
	    break;
	case 6:     /* equatorinal to ecliptic */
	    cr2cr_alm(alm1, lmax, nbuf, 'q', 'e');
	    break;
	}
    }
    if (abs_flag) {
	for(l=lmin_print; l<=lmax_print; l++) {
	    int ll = l * nbuf;
	    for(m=0; m<=l; m++) {
		int mb = m+lmx1+1;
		double re1=alm1[ll+m], im1=alm1[ll+m+lmx1+1];
		if (re1 < 0.)  alm1[ll+m]  = -re1;
		if (im1 < 0.)  alm1[ll+mb] = -im1;
	    }
	}
    }
    if (inverse_flag) {
	for(l=lmin_print; l<=lmax_print; l++) {
	    int ll = l * nbuf;
	    for(m=0; m<=l; m++) {
		int mb = m+lmx1+1;
		double re1=alm1[ll+m], im1=alm1[ll+m+lmx1+1];
		if (fabs(re1) > 1.e-30)  alm1[ll+m]  = 1./re1;
		if (fabs(im1) > 1.e-30)  alm1[ll+mb] = 1./im1;
	    }
	}
    }

    if (change_phase_flag) {
	for (l=lmin_print; l<=lmax_print; l++) {
	    register int ll = l * nbuf;
	    for (m=0; m<=l; m++) {
		double re=alm1[ll+m], im = alm1[ll+m+lmax+1];
		double amp = sqrt(re*re+im*im);
		double phase = ph_calc(re, im) + phase_ch;

		if (m==0) {
		    if (re < 0.0)  phase = M_PI; else phase = 0.0;
		    if(ABS(phase_ch-M_PI) < 0.001) {
			    phase += M_PI;
		    }
		} else {
		    phase = ph_calc(re, im) + phase_ch;
		}
		alm1[ll+m]        = amp * cos(phase);
		alm1[ll+m+lmax+1] = amp * sin(phase);
	    }
	}
    }

    /* shift mutlipoles by l and m */
    if (shift_multip_flag) {
	extern void shift_multipoles(double **, int *, int *, int, int);
	shift_multipoles(&alm1, &lmax, &nbuf, sd_l, sd_m);
    }

    if ( fabs(cf-1.) > 1.e-15 &&  *filename2 == '\0' ) {
	double cf_r = cf, cf_i = cf;

	if (part_flag == 1)      cf_i = 1.; /* only real part is changed */
	else if (part_flag == 2) cf_r = 1.; /* only imag part is changed */

	if (l_calc >=0 || m_calc >=0) {
	    if (l_calc >= 0 && m_calc < 0) {
		/* change for all m */
		int ll = l_calc * nbuf;
		for (m=0; m<=l_calc; m++) {
			alm1[ll+m]        *= cf_r;
			alm1[ll+m+lmax+1] *= cf_i;
		}
	    } else if (l_calc < 0 && m_calc >= 0) {
		/* change for all l */
		for (l=lmin_print; l<=lmax_print; l++) {
			int ll = l * nbuf;
			alm1[ll+m_calc]        *= cf_r;
			alm1[ll+m_calc+lmax+1] *= cf_i;
		}
	    } else {
		/* change for (l,m) only */
		int ll = l_calc * nbuf;
		alm1[ll+m_calc]        *= cf_r;
		alm1[ll+m_calc+lmax+1] *= cf_i;
	    }
	} else {
	    for (l=lmin_print; l<=lmax_print; l++) {
		register int ll = l * nbuf;
		for (m=0; m<=l; m++) {
		    alm1[ll+m]        *= cf_r;
		    alm1[ll+m+lmax+1] *= cf_i;
		}
	    }
	}
    }
    /* calculate average phase vs l */
    if (average_phase) {
	for (l=lmin_print; l<=lmax_print; l++) {
	    register int ll = l * nbuf;
	    double si_av = 0., co_av = 0., sum_w = 0.0;

	    if (l == 0) continue;

	    for (m=1; m<=l; m++) {
		double re = alm1[ll+m], im = alm1[ll+m+lmax+1];
		double phase = ph_calc(re, im) + phase_ch;
		double w = 1.0;

		if (average_phase == 2)  w = 1. / sqrt(re*re+im*im);
		sum_w += w;

		if (m==0) {
//                    if (re < 0.0)  phase = M_PI; else phase = 0.0;
//                    if(ABS(phase_ch-M_PI) < 0.001) {
//                            phase += M_PI;
//                    }
//                    si_av += sin (phase);
//                    co_av += cos (phase);
		} else {
		    phase = ph_calc(re, im) + phase_ch;
		    si_av +=  w * sin (phase);
		    co_av +=  w * cos (phase);
		}
	    }
	    si_av /= l+l+1;
	    co_av /= l+l+1;
	    printf("%5d %12.5e\n", l, atan2(si_av,co_av));
	}
	exit(0);
    }


    /* calculate statistics on alm data */
    if (alm_stat_flag) {
	    for (l=lmin_print; l<=lmax_print; l++) {
		register int ll = l * nbuf;
		double v=0.0, s=0.0, s_re=0.0, s_im=0.0,
		       sq=0.0, sq_im=0.0, sq_re=0.0;

		if (l<2) continue;

		 if (fabs(alm1[ll]) > 1.e-5 )
		      v = (1+alm1[ll])*(1+alm1[ll]);
		 if (fabs(alm1[ll+lmax+1]) > 1.e-5 )
		      v = (1+alm1[ll+lmax+1]) * (1+alm1[ll+lmax+1]);
		for (m=1; m<=l; m++)
{
//                    v += alm1[ll+m]*alm1[ll+m] +
//                         alm1[ll+m+lmax+1] * alm1[ll+m+lmax+1];
		      v += ((1+alm1[ll+m]) * (1+alm1[ll+m]) +
			     (1+alm1[ll+m+lmax+1]) * (1+alm1[ll+m+lmax+1]));
}
		for (m=1; m<=l; m++) {
			double re = alm1[ll+m];
			double im = alm1[ll+m+lmax+1];
			s += re + im;
			s_re += re;
			s_im += im;
		}
		s /= l+l;
		s_re /= l;
		s_im /= l;
		for (m=1; m<=l; m++) {
			double re = alm1[ll+m];
			double im = alm1[ll+m+lmax+1];
			sq += (re-s)*(re-s) + (im-s)*(im-s);
			sq_re += (re-s_re)*(re-s_re);
			sq_im += (im-s_im)*(im-s_im);
		}
		sq /= l+l;
		sq_re /= l;
		sq_im /= l;
		sq = sqrt(sq);
		sq_re = sqrt(sq_re);
		sq_im = sqrt(sq_im);

		v = sqrt(v);
		v /= (2*l+1);
//              PR(stdout, "%4d %12.5e\n", l, v);
		PR(stdout,
		"%4d  %9.2e,%9.2e,%8.1e  %9.2e,%9.2e,%8.1e  %9.2e,%9.2e,%8.1e\n",
		       l, s_re, sq_re, s_re/sq_re,
			  s_im, sq_im, s_im/sq_im, s, sq, s/sq);
	    }
    }
    if (phase_set_limits) {
	for (l=lmin_print; l<=lmax_print; l++) {
	    register int ll = l * nbuf;
	    for (m=0; m<=l; m++) {
		double re=alm1[ll+m], im = alm1[ll+m+lmax+1];
		double phase;

		if (m==0) {
		    if (re < 0.0)  phase = M_PI; else phase = 0.0;
		    if(ABS(phase_ch-M_PI) < 0.001)
			    phase += M_PI;
		} else {
		    phase = ph_calc(re, im);
		}
		if (phase < phase_min || phase_max < phase)
		    alm1[ll+m] =  alm1[ll+m+lmax+1] = 0.0;
	    }
	}
    }

    /* output alm-s data file */
    if (out_flag) {
	if (gp_output) {
	    print_alm_plot (fp_d, alm1, lmax, nbuf, lmin_print, lmax_print);
	} else {
	    /* write alm-file */
	    if (ffile_flag)
		ff_write_alm(alm1, nbuf, lmax, lmax_print, fp_d);
	    else
		out_glesp_alm(alm1, lmax, nbuf, lmax_print, fp_d);
	}
	if (fp_d != (FILE *)0 && fp_d != stdout) fclose(fp_d);
    }

    exit(0);
}

void setopt(argc, argv)
int argc;
char **argv;
{
	register int i, n=0;
	char buf[10];
	myname = argv[0];

	for(i=1; i<argc; i++) {
	    switch(argv[i][0]) {
	    case '-':
		switch(argv[i][1]) {
		case 'a':
			if (strncmp(argv[i]+1, "ab",2) == 0) {
			    if (strncmp(argv[i]+1, "abs",3) == 0)
				abs_flag = 1;
			    else
				abs_dif = 1;
			} else if (strncmp(argv[i]+1, "a1", 2) == 0) {
			    /* set amplitude = 1 */
			    amplit_1_flag = 1;
			} else if (strncmp(argv[i]+1, "apw", 3) == 0) {
			    /* calc weighted average phase vs l */
			    average_phase = 2;
			} else if (strncmp(argv[i]+1, "ap", 2) == 0) {
			    /* calc average phase vs l */
			    average_phase = 1;
			} else {
			    PR(ER, "%s: illegal flag: '%s'\n",
					myname, argv[i]);
			    usage();
			}
			break;
		case 'b':
			nbin = atoi(argv[++i]);
			break;
		case 'c':
			switch(argv[i][2]) {
			case '1':
				c1 = atof(argv[++i]);
				break;
			case '2':
				c2 = atof(argv[++i]);
				break;
			case 'd':
				cf = 1./atof(argv[++i]);
				if (argv[i][strlen(argv[i])-1] == 'r')
				    part_flag = 1;
				else if (argv[i][strlen(argv[i])-1] == 'i')
				    part_flag = 2;
				break;
			case 'm':
			case 'f':
				cf = atof(argv[++i]);
				if (argv[i][strlen(argv[i])-1] == 'r')
				    part_flag = 1;
				else if (argv[i][strlen(argv[i])-1] == 'i')
				    part_flag = 2;
				break;
			case 'l':
				cl_compare_flag = 1;
				break;
			case 'o': /* correlation flags */
				corr_shift_factor = atoi(argv[++i]);
				break;
			case 'r':
				if(strncmp(argv[i]+1, "crp", 3) == 0) {
				    corr_phase_flag = 1;
				} else {
				    PR(ER, "%s: illegal flag: '%s'\n",
						myname, argv[i]);
				    usage();
				}
				break;
			default :
				PR(ER, "%s: illegal flag: '%s'\n",
					    myname, argv[i]);
				usage();
				break;
			}
			break;
		case 'd':
			switch(argv[i][2]) {
			case 'o':
				if (strcmp(argv[i]+1, "double") == 0) {
					bitpix_map = DOUBLEPIX;
				} else {
					PR(ER,"%s: illegal flag '%s'\n",
					    myname, argv[i]);
					usage();
				}
				break;
			case 'p':
				if (strncmp(argv[i]+1, "dpha",4) == 0) {
				    phase_differ_flag = 1;
				} else {
				    dphi = deg2rad(argv[++i]);
				    rotate_flag = 1;
				}
				break;
			case 's':
				dpsi = deg2rad(argv[++i]);
				rotate_flag = 1;
				break;
			case 't':
				dtheta = deg2rad(argv[++i]);
				rotate_flag = 1;
				break;
			}
			break;
		case 'e':
			if (strncmp(argv[i]+1, "ec", 2) == 0) {
			   /* transform from galactic coors to ecliptic */
			   coor_sys='e';
			} else if (strncmp(argv[i]+1, "eq", 2) == 0) {
			   /* transform from galactic coors to equatorial */
			   coor_sys='q';
			} else if (strncmp(argv[i]+1, "e2g", 3) == 0 ||
				   strncmp(argv[i]+1, "e,g", 3) == 0) {
				coor_sys=3;
			} else if (strncmp(argv[i]+1, "e2q", 3) == 0 ||
				   strncmp(argv[i]+1, "e,q", 3) == 0) {
				coor_sys=5;
			} else {
			   /* file for relative errors */
			   strcpy(out_file, argv[++i]);
			}
			break;
		case 'f':
			if (strncmp(argv[i]+1, "flo", 3) == 0) {
				bitpix_map = FLOATPIX;
			} else {
				cut_alm_flag = 1;
				strcpy(out_file, argv[++i]);
			}
			break;
		case 'g':
			if (strcmp(argv[i]+1, "g") == 0) {
			    /* output in ASCII */
			    gp_output = 1;
			} else if (strncmp(argv[i]+1, "g2q", 3) == 0 ||
			    strncmp(argv[i]+1, "g,q", 3) == 0) {
				coor_sys='q';
			} else if (strncmp(argv[i]+1, "g2e", 3) == 0 ||
				   strncmp(argv[i]+1, "g,e", 3) == 0) {
				coor_sys='e';
			}
			break;
		case 'h':
			if (strcmp(argv[i]+1,"h2") == 0)
			    usage2();
			else
			    usage();
		case 'i':
			if (strncmp("inv", argv[i]+1, 3) == 0) {
			    inverse_flag = 1;
			} else {
			    PR(ER,"%s: illegal flag '%s'\n", myname,argv[i]);
			    usage();
			}
			break;
		case 'l':
			if (strcmp(argv[i]+1, "l") == 0) {
			    l_calc = atoi(argv[++i]);
			} else if (strncmp(argv[i]+1, "lmx",3) == 0 ||
			    strncmp(argv[i]+1, "lma",3) == 0  )
				lmax_print = atoi(argv[++i]);
			else if (strncmp(argv[i]+1, "lmn",3) == 0 ||
			    strncmp(argv[i]+1, "lmi",3) == 0  )
				lmin_print = atoi(argv[++i]);
			else
			    usage();
			break;
		case 'L':
			if (strncmp(argv[i]+1, "L",1) == 0) {
			    l_calc = atoi(argv[++i]);
			} else if (strncmp(argv[i]+1, "Lmx",3) == 0 ||
			    strncmp(argv[i]+1, "Lma",3) == 0  )
				lmax_print = atoi(argv[++i]);
			else if (strncmp(argv[i]+1, "Lmn",3) == 0 ||
			    strncmp(argv[i]+1, "Lmi",3) == 0  )
				lmin_print = atoi(argv[++i]);
			else
			    usage();
			break;
		case 'm':
			if (strncmp(argv[i]+1, "ma", 2) == 0) {
			    mask_flag = 1;
			} else if (strncmp(argv[i]+1, "md", 2) == 0) {
			    max_deviation = 1;
			} else if (strcmp(argv[i]+1, "mul") == 0) {
			    action = 2;
			} else
			    if (strncmp(argv[i]+1, "m",1) == 0) {
				m_calc = atoi(argv[++i]);
			} else
			    usage();
			break;
		case 'M':
			if (strncmp(argv[i]+1, "M",1) == 0) {
			    m_calc = atoi(argv[++i]);
			} else
			    usage();
			break;
		case 'n':
			if (strncmp(argv[i]+1, "noc",3) == 0) {
			    glesp_comments = 0;
			} else if (strncmp(argv[i]+1, "n",1) == 0) {
			    n_low = atoi(argv[++i]);
			} else
			    usage();
			break;
		case 'o':
			/* output file */
			if (strncmp(argv[i]+1, "of", 2) == 0) {
			    ffile_flag = 1; /* output in F-format */
			}
			strcpy(out_file, argv[++i]);
			break;
		case 'p':
			if (strncmp(argv[i]+1, "ph", 2) == 0) {
			    phase_ch = atof(argv[++i]) / 180 * M_PI;
			    change_phase_flag = 1;
			} else if (strncmp(argv[i]+1, "pc", 2) == 0) {
			    /* phase conversion flag */
			    phase_conv_flag = 1;
			} else if (strncmp(argv[i]+1, "ps", 2) == 0 ||
			      strncmp(argv[i]+1, "pds", 3) == 0) {
			    /* select phases */
			    int phase_sel_degr =  argv[i][2] == 'd' ? 1 : 0;
			    i++;
			    if (argv[i][0] != ',') {
				if (strncmp(argv[i], "none", 4) != 0) {
				    phase_min = atof(argv[i]);
				    if (phase_sel_degr) phase_min *= M_PI/180;
				}
			    }
			    {
				char *ps = strchr(argv[i], ',');
				if (ps != NULL && *ps && *(ps+1)) {
				     if (strncmp(ps+1, "none", 4) != 0) {
					phase_max = atof(ps+1);
					if (phase_sel_degr)
					    phase_max *= M_PI/180;
				     }
				}
			    }
			    phase_set_limits = 1;
			} else {
			    PR(ER, "%s: illegal flag: '%s'\n",
					myname, argv[i]);
			    usage();
			}
			break;
		case 'q':
			if (strncmp(argv[i]+1, "q2g", 3) == 0 ||
			    strncmp(argv[i]+1, "q,g", 3) == 0) {
				coor_sys=4;
			} else if (strncmp(argv[i]+1, "q2e", 3) == 0 ||
				   strncmp(argv[i]+1, "q,e", 3) == 0) {
				coor_sys=6;
			} else {
			    PR(ER, "%s: illegal flag: '%s'\n",
					myname, argv[i]);
			    usage();
			}
			break;
		case 'r':
			if (strcmp(argv[i]+1, "r1") == 0)
			    relative_differ = 2;
			else
			    relative_differ = 1;
			break;
		case 's':
			switch(argv[i][2]) {
			case 'u':   /* sum */
			    if( strncmp(argv[i], "-sum", 4) == 0) {
				 sign = 1; action = 1; break;
			    } else if( strncmp(argv[i], "-sun", 4) == 0) {
				 sun_data_key = 1;
			    }
			    break;
			case 'i':   /* si */
			    sin_flag = 1;  break;
			case 'c':   /* sc */
			    amplit_sc_flag = 1; break;
			case 't':   /* st */
			    alm_stat_flag = 1; out_flag = 0; break;
			case 'l': /* shift by step sl */
			    sd_l = atoi(argv[++i]);
			    shift_multip_flag = 1;
			    break;
			case 'm':
			    sd_m = atoi(argv[++i]);
			    shift_multip_flag = 1;
			    break;
			default:
			    PR(ER, "%s: illegal flag: '%s'\n",
					myname, argv[i]);
			    usage();
			}
			break;
		case 't':
			if (strcmp(argv[i]+1, "tl") == 0)
			     flag_tl = 1;
			break;
		case 'v':
		    switch (argv[i][2]) {
		    case '\0':
			printf("%s: version 0.7 (15/01/2016), GLESP 2.0\n",
			    myname);
			exit(0);
		    case '0':
			/* use only real or imaginary part */
			buf[0] = argv[++i][0];
			switch (buf[0]) {
			case '1':
			case 'r': /* real */
				part_flag = 1;
				break;
			case '2':
			case 'i': /* imagine */
				part_flag = 2;
				break;
			case '0':
				part_flag = 0;
				break;
			}
			break;
		    case 'd': /* divide by vector */
			    l_vector_flag = 1;
			    strcpy(v_file, argv[++i]);
			    break;
		    case 'm': /* multiply by vector */
			    l_vector_flag = 2;
			    strcpy(v_file, argv[++i]);
			    break;
		    }
		    break;
		default :
			usage();
			break;
		}
		break;
	    case '+':
		switch(argv[i][1]) {
		case 'h':
			usage();
		default :
			usage();
			break;
		}
		break;
	    default :
		if (n == 0)
			strcpy(filename1, argv[i]);
		else
			strcpy(filename2, argv[i]);
		n++;
		break;
	    }
	}

	return;
}

void usage()
{
PR(ER,"usage: %s file1 file2 [-e err_file] [-v {012ri}]\n", myname);
PR(ER,"[-lmx lmax_print] [-lmn lmin_print] [-o output_file] [-sum]\n");
PR(ER,"[-f ffile_out] [-l l_calc] [-m calc] [-n l_lim] [-cl] [-bn nbin] [-md]\n");
PR(ER,"[-ph phase_change(degs)] [-cf multipl_coeff] [-cd divide_coeff]\n");
PR(ER,"[-pc(phase_conversion)] [-v{md} v(l)_file(mult/div by vector) [-nocom]]\n");
PR(ER,"[-ab(dif.abs.value)] [-abs(both Re&Im)] [-a1(set amp=1)] [-g(ASCII out)]\n");
PR(ER,"[-sc(cos(dphi),sin(dphi))] [-ap{w}(calc.aver.phases)] [-{double,float}]\n");
PR(ER,"[-c1 c1] [-c2 c2] [-r(relat.differ.=(a-b)/b)] [-crp] [-sin] \n");
PR(ER,"[-r1(relat.differ.=2(a-b)/(a+b))] [-co corr.shift_dl] [-inv(inverse)]\n");
PR(ER,"[-dt rotate_theta_angle] [-dp rotate_phi_angle] [-ds dpsi] [-{eq,ec}]\n");
PR(ER,"[-{eq,ec}] [-{q2e,q2g,e2q,e2g}]\n");
PR(ER,"[-sl shift_dl] [-sm shift_dm] [-dphase(calc phase difference)]\n");
exit(2);
}

void usage2()
{
PR(ST, "usage: %s file1 [file2] [FLAGS]\n", myname);
PR(ST, "---------------------------------------------------------------\n");
PR(ST, "-a1 - sets amplitude of harmonics equal to 1: sqrt(re^2+im^2)=1\n");
PR(ST, "-ab - calculates difference in amplitudes from 2 sets of x_{lm},\n");
PR(ST, "       where the 1st column =|x1_lm| -|x2_lm| and the 2nd one = 0.\n");
PR(ST, "-abs - calculates absolute values both of Re & Im parts of a_lm-s: (|Re|,|Im|)\n");
PR(ST, "-ap  - calculates averaged phases\n");
PR(ST, "-apw - calculates weighted averaged phases\n");
PR(ST, "-c1 cf1 - multiplies the 1st x_{lm} by a factor cf1 (see '-c2')\n");
PR(ST, "-c2 cf2 - multiplies the 2nd x_{lm} by 'cf2' to have 'cf1*alm1-cf2*alm2'\n");
PR(ST, "-cl - compare C(l) of 2 sets x_{lm}:  R(l) = cl1[l]/cl2[l]-1\n");
PR(ST, "-cd K{r,i} - divides x_{lm} by a factor K. If a numerical K is followed by\n");
PR(ST, "        'r' or 'i', the division is performed on only Re- or Im-part.\n");
PR(ST, "-cf K{r,i} - multiplies x_{lm} by a factor K. If a numerical K is followed by\n");
PR(ST, "        'r' or 'i', the multiplication is performed on only Re- or Im-part.\n");
PR(ST, "-co n - correlation shift factor n=dl\n");
PR(ST, "-double - makes output in 64-bit double precision format\n");
PR(ST, "-dphase - calculates phase differences of x1_{lm} and x2_{lm}\n");
PR(ST, "-dp dphi   - rotates x_{lm} by phi-angle,radians\n");
PR(ST, "-ds dpsi   - rotates x_{lm} by psi-angle,radians\n");
PR(ST, "-dt dtheta - rotates x_{lm} by theta-angle,radians\n");
PR(ST, "-e2q - converts x_{lm} of ecliptic coord. system to equatorial one\n");
PR(ST, "-e2g - converts x_{lm} of ecliptic coord. system to galactic one\n");
PR(ST, "-ec  - converts x_{lm} of galactic coord. system to ecliptic one\n");
PR(ST, "-eq  - converts x_{lm} of galactic coord. system to equatorial one\n");
PR(ST, "-float  - makes output in 32-bit single precision format\n");
PR(ST, "-g - outputs x_{lm} coefficients in ASCII format\n");
PR(ST, "-inv - makes inversion both of Re & Im parts of a_{lm}-s: (1/Re,1/Im)\n");
PR(ST, "-l L - sets the l-multipole L for arithmetic operations (see '-cd')\n");
PR(ST, "-lmax LMAX - sets the upper limit of the l-multipole at LMAX for output\n");
PR(ST, "-lmin LMIN - sets the lower limit of the l-multipole at LMIN for output\n");
PR(ST, "-m M - sets the m-multipole M for arithmetic operations (see '-cd')\n");
PR(ST, "-nocom - no output comments\n");
PR(ST, "-o output_file - outputs resultant x_{lm} to file output_file\n");
PR(ST, "-pc - phase exchange: output contains amp(file1) + phase(file2)\n");
PR(ST, "-pds phase1,phase2 - selects a_lm with phases from interval in degs\n");
PR(ST, "-ph dphase - shifts phases of each x_{lm} by dphase,degs\n");
PR(ST, "-ps phase1,phase2 - selects a_lm with phases from interval in radians\n");
PR(ST, "-q2e - converts x_{lm} of equatorial coord. system to ecliptic one\n");
PR(ST, "-q2g - converts x_{lm} of equatorial coord. system to galactic one\n");
PR(ST, "-r - calculates relative difference of 2 sets of x_{lm} as:\n");
PR(ST, "       Re_out = 2*(c1*Re_inp1-c2*Re_inp2)/(c2*Re_inp2),\n");
PR(ST, "       Im_out = 2*(c1*Im_inp1-c2*Im_inp2)/(c2*Im_inp2)\n");
PR(ST, "-r1 - calculates relative difference of 2 sets of x_{lm} as:\n");
PR(ST, "       Re_out = 2*(c1*Re_inp1-c2*Re_inp2)/(c1*Re_inp1+c2*Re_inp2),\n");
PR(ST, "       Im_out = 2*(c1*Im_inp1-c2*Im_inp2)/(c1*Im_inp1+c2*Im_inp2)\n");
PR(ST, "-sl dl_to_shift - set value for shifting by l-multipole\n");
PR(ST, "-sm dm_to_shift - set value for shifting by m-multipole\n");
PR(ST, "-sum - adds up 2 input files\n");
PR(ST, "-v - print version number of '%s'\n", myname);
PR(ST, "-vd vector-file - reads data from ASCII 'vector-file' with 2 columns:\n");
PR(ST, "    l,v(l) and divides the input x_{lm} by v(l): aout_{lm}=x_{lm}/v(l).\n");
PR(ST, "-vm vector-file - reads data from ASCII 'vector-file' with 2 columns:\n");
PR(ST, "    l,v(l) and multiplies the input x_{lm} by v(l): aout_{lm}=x_{lm}*v(l).\n");
exit(0);
}
