#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <sys/types.h>
#include <dirent.h>
#include <sys/stat.h>

#include "glesp.h"
#include "gd/plot_gd.h"

extern void  marker2mol(struct _crm_ *,  int, double);

/*
 *  Put GLESP-file points on the sphere
 *   Oleg Verkhodanov, basic procedures: 2003-2005
 *   last changes 19/03/2006, 22/02/2008, 10/10/2008, 5/04/2009, 25/05/2009,
 *   28/12/2009, 19/10/2010, 29/11/2010, 21/03/2011, 15/04/2012
 *   05/04/2013, 16/07/2013, 02/09/2013, 10/2013, 05/2013, 0/2014, 2/2015
 */

#define PR      fprintf
#define ER      stderr
#define ST      stdout

extern void setopt(int, char **), usage(void), usage2(void);
extern void t2gif(double *, int, int, double *, int, FILE *);

int  xs=800, ys=320;
int  bar = 1;          /* flag to draw bar */
double min = -1.e+39,  /* limits           */
       max = +1.e+39;  /*                  */
int  set_limits=0;     /* set limits with force flag=2 */
int  col_map = 5;      /* color map */
int  coordinates_flag = 0;  /* (theta,phi) or (lat,lon)   */
int  flag_coor = 1; /* galactic or equatorial coordinates of source */
int  ffile_flag = 1;
double size_factor = 1.;
int font_size = 0;
double dummy = 0.0;
int flag_map = 3;   /* which map is extracted from 3map   */
int flag_pix = 'g';
int projection='m'; /* projection type, default Mollweide */
int log_flag = 0;

int inp_ff_flag = 0;   /* input F-file                    */
int inp_col=1;         /* input color for cross-cut       */
int inp_size=30;       /* input size for cross-cut pixel  */
int inp_mark='x';      /* input marker for cross-cut pixel*/
int bt_s=0, bt_m=0, bt_c=0;
int cats_input=0;      /* input sources from http://cats.sao.ru */
double d1_step = -1., d2_step = -1.; /* steps for grid in degs  */
int pixel_mask = 1;    /* use pixel mask (when pixel=GLESPMASK) */

char hp_mask_file[161];
char gl_mask_file[161];
char source_file[161];
char filename[161];
char outp_file[161];
char dir_out[161];          /* output directory */
char mark_str[161];
char inp_pol_file[161]; /* file with non-directed vectors */
char inp_amp_pol_f[161], inp_ang_pol_f[161]; /* files with pol.vectors */
int overlay_data = 0;  /* flag=1 => overlay mask data */
int overlay_color=0;
int overlay_flag=0;    /* =1 => mask outside, =0 => mask inside */
int overlay_width = 1;
char overlay_file[161];
int vect_overlay = 0;
int patt_overlay = 0; /* overlay some celestial patterns */
char plotfile[161];
int lmax = -1;        /* lmax for a_lm-s drawing         */

int format_ext_areas = 0;   /* flag to extract areas form list */
char file_ext_areas[80];    /* file with extracted areas       */
double b_size = 0.5;        /* default boxsize for areas: 0.5d */

double lat_c,  /* galactic coordinate of ring center or rectang. corner */
       lon_c,  /* galactic coordinate of ring center or rectang. corner*/
       lat1_c, /* galactic coordinate of another rectang. corner */
       lon1_c, /* galactic coordinate of another rectang. corner */
       r_c,    /* radius of inner circle */
       w_c;    /* radius of ring width   */
int flag_conc_fig = 0;  /* 1-circle, 2 - rectangle */
int phase_diagram = 0;  /* draw phase diagram for alm-file  */
int sin_flag=0;         /* flag to plot phase of sin(phase) */
int ampl_diagram = 0;   /* draw diagram for amplitude special file amp(l,m)*/

int z_mask_flag = 0;      /* set mask                                  */
double zmask_value = M_PI * 1.e+30;
int zone_circle = 0;
char zone_file[81];       /* file with list of zones to be processed   */
double zone_radius = -1.; /* radius of removed circle in degs          */
double zone_latR0 = -1000., zone_lonR0 = -1000.,
       zone_latR1 = -1000., zone_lonR1 = -1000.;
double zcut_latR1 = -1000., zcut_lonR1 = -1000.,
       zcut_latR2 = -1000., zcut_lonR2 = -1000.;
int    z_remove_flag = 1,
       z_remain_flag = 0;
int grid_type = -1;
int grid_color = 0;
int hsv_color = 0;   /* HSB-color grid */

double lon0 = 0.;

extern int sun_data_key;
extern int swap_write_key;

/* comments on the screen */
#define TXT_N 10
int txt_n = 0;   /* number of comments */
struct _txt_ {
	int  x, y;
	char text[301];
	int  color;
	int  size;
} txt[TXT_N];
/* limits for map values */
#define NV_INTR 10
struct _vv_  { double low, high; } v_intr [NV_INTR];
int nv_intr=0;

char *myname="f2fig";

int main(argc, argv)
int argc;
char **argv;
{
    register int i, j;
    FILE *fp_in = (FILE *)0, *fp_out = (FILE *)0;
    double *f = (double *)0,
	   *q = (double *)0,
	   *u = (double *)0,
	   *x = (double *)0,
	   *offs  = (double *)0,
	   *zmask = (double *)0,
	   *img = (double *)0,
	   *ar = (double *)0;
    int *nph = (int *)0;
    int nx=0, np=0;
    int *imgs    = (int *)0,
	*img_bar = (int *)0,
	*img_txt = (int *)0,
	*mask = (int *)0;
    struct _crm_ *crm = (struct _crm_ *)0;
    double fmin= 1.e+30, fmax= -1.e+30;
    int test_flag = 0;
    char output_name [301];
    FILE *fpl = (FILE *)0;   /* file with list of sources for areas */

    source_file[0] = filename[0] = outp_file[0] = mark_str[0] =
       inp_pol_file[0] = '\0';
    inp_amp_pol_f[0] = inp_ang_pol_f[0] = hp_mask_file[0] = '\0';
    dir_out[0] = plotfile[0] = gl_mask_file[0] = '\0';
    output_name[0] = '\0';

    strcpy(tmpcat, ".");

    ys = xs / 2;

    /* loop over external parameters */
    setopt(argc, argv);

    if (glesp_comments < 0) {
	if (getenv("GLESP_COMMENTS") != NULL) {
	    char str[21];
	    strcpy(str,getenv("GLESP_COMMENTS"));
	    glesp_comments = *str == 'N' || *str == 'n' ? 0 : 1;
	    if (glesp_comments) {
		 if ( strcmp(str, "OFF") == 0 || strcmp(str, "off") == 0)
		     glesp_comments = 0;
	    }
	} else
	    glesp_comments = 1;
    }

    if(outp_file[0] == '\0' && ! format_ext_areas ) {
	    PR(ER, "%s: no name of output file...\n", myname);
	    usage();
    }

    /* when we read source corrdinates from files */
    /* to prepare GIF-images of areas             */
    /* open in[ut file                            */
    if (format_ext_areas) {
	if ((fpl = fopen(file_ext_areas, "r")) == (FILE *)0) {
	    PR(ER, "%s: can not open input file with coordinates...\n",
		myname);
	    exit(1);
	}
    }

    if(*filename == '\0'  && *inp_amp_pol_f == '\0') {
	    PR(ER, "%s: no input file...\n", myname);
	    exit(1);
    }

    if(*filename != '\0' && strcmp(filename, "-") != 0)  {
	/* read input map */
	test_flag = test_glesp_map(filename);
	if ( test_flag == 1 || test_flag == 0)  {
	    read_glesp_map(&f, &x, &nx, &np, &nph, filename);
	    ar = f;
	    flag_map = 0;
	} else if ( test_flag == 3) {
	    in_glesp_3map(&f,&q,&u, &x, &nx, &np, &nph, filename);
	    if (flag_map == 'q') {
		 free((void *)f);
		 free((void *)u); u = (double *)0;
		 ar = q;
	    } else if (flag_map == 'u') {
		 free((void *)f);
		 free((void *)q); q = (double *)0;
		 ar = u;
	    } else if (flag_map != 3) {
		 ar = f;
		 free((void *)q); q = (double *)0;
		 free((void *)u); u = (double *)0;
	    }
	} else if (test_flag == 5) {
	    extern void in_hpix_map(double **, double **,  double **,
		int *, int *, int **, char *);

	    if (glesp_comments == 1)
		PR(ER, "%s: Hmm, the file '%s' is in the HEALPix format...\n",
		    myname, filename);
	    in_hpix_map(&f,&x, &offs, &nx, &np, &nph, filename);
	    ar = f;
	    flag_pix = 'h';
	    flag_map = 0;
	} else if (test_flag == 6) {
	    if (glesp_comments == 1)
		PR(ER, "%s: Hmm, the file '%s' is in the HEALPix format...\n",
		    myname, filename);
	    extern void in_hpix_3map(double **, double **,  double **,
		double **,  double **, int *, int *, int **, char *);
	    in_hpix_3map(&f,&q,&u, &x,&offs, &nx, &np, &nph, filename);
	    if (flag_map == 'q') {
		 free((void *)f);
		 free((void *)u); u = (double *)0;
		 ar = q;
	    } else if (flag_map == 'u') {
		 free((void *)f);
		 free((void *)q); q = (double *)0;
		 ar = u;
	    } else if (flag_map != 3) {
		 ar = f;
		 free((void *)q); q = (double *)0;
		 free((void *)u); u = (double *)0;
	    }
	    flag_pix = 'h';
	} else if (test_flag == 8) { /* 2-D Basic FITS */
	    if (glesp_comments == 1)
		PR(ER, "%s: Hmm, the file '%s' is in Basic FITS format...\n",
		    myname, filename);
	    flag_pix = 'f';
	    flag_map = 1;
	} else if ((test_flag == 2 || test_flag == 4)
	   && !phase_diagram && !ampl_diagram) {
	    /* alm-file input */
	    PR(ER, "%s: file '%s' is the alm-file.\n", myname, filename);
	    PR(ER,
		"To plot a phase diagram from alm-data, use flag '-ph'.\n");
	    exit(1);
	} else if (test_flag == 7 && !phase_diagram && !ampl_diagram) {
	    /* phase-file input */
	    PR(ER, "%s: file '%s' contains ASCII data.\n", myname, filename);
	    PR(ER,
       "If they are phases, then use flag '-ph' to plot a phase diagram .\n");
	    exit(1);

	} else if (!phase_diagram && !ampl_diagram) {
	    PR(ER, "%s: file '%s' has not acceptable format...\n",
		myname, filename);
	    exit(1);
	}
    } else {
	filename[0] = '\0';
	/* empty (zero) array */
	nx = 101; np = 202;
	if((f=(double*)calloc((unsigned)(nx*np),sizeof *f))==(double*)0){
nomem:
	    PR(ER, "%s: no memory for array...\n", myname);
	    exit(1);
	}
	if((x=(double *)calloc((unsigned)nx,sizeof *x))==(double *)0)
	    goto nomem;
	if((nph=(int *)calloc((unsigned)nx,sizeof *nph))==(int *)0)
	    goto nomem;
	for(i=0; i<nx; i++) {
	    x[i] = cos(M_PI / nx * (.5 + i));
	    nph[i] = np;
	    if (fabs(dummy) > 1.e-8) {
		for(j=0; j<nph[i]; j++)  f[i*np+j] = dummy;
	    }
	}
	ar = f;
	flag_map = 0;
    }

    /* create directory if required */
    if ( *dir_out != '\0' ) {
	 DIR *dr;
	 if ((dr = opendir(dir_out)) == NULL ) {
	     if (mkdir(dir_out, 0755) != 0) {
		PR(ER, "%s: problem with opening directory '%s'...\n",
		   myname, dir_out);
		PR(ER, "output directory will be './'\n");
		strcpy(dir_out, ".");
	     }
	 } else
	     closedir (dr);
    }
    if ( *dir_out == '\0' ) {
	strcpy(output_name, outp_file);
    } else {
	sprintf(output_name, "%s/%s", dir_out, outp_file);
    }

    /* here is work with alm-files, not maps */
    if (phase_diagram) {
	/* triangular a_lm-s phase diagram: color phase (l,m) */
	extern void
	    plot_phase_diagram(char *,int,int,int,int,int,char*,int),
	    plot_ascii_phase_diagram(char *,int,int,int,int,int,char*,int);
	int flag_col = 0;

	if (test_flag == 0 || test_flag == 1 || test_flag == 3 ||
	    test_flag == 5 || test_flag == 6) {
		PR(ER, "%s: file '%s' is the map-file. ", myname, filename);
		PR(ER, "To plot phase diagram, the program need alm-file\n.");
		exit(1);
	}
	/* set color table */
	setcol_i(col_map);
	font_size = 5;

	if (hsv_color) flag_col = -1;   /* HSB-color grid */
	if (sin_flag)  flag_col =  1;   /* use sin(phase) */

	if (test_flag == 7) {
	    /* read from 3-column ASCII table */
	    plot_ascii_phase_diagram(filename, lmax, xs,ys,bar,font_size,
		output_name, flag_col);
	} else {
	    /* read and calc from a_lm FITS file */
	    plot_phase_diagram(filename, lmax, xs,ys,bar,font_size,
		output_name, flag_col);
	}
	exit (0);
    }

    /* here is work with alm-files, not maps */
    if (ampl_diagram) {
	/* triangular a_lm-s amplitude diagram: color ampl (l,m) */
	extern void
	    plot_ampl_diagram(char *, int,int, int, int, int, char *,
	      double, double, int),
	    plot_ascii_ampl_diagram(char *, int,int, int, int, int, char *,
		 double, double, int);

	if (test_flag == 0 || test_flag == 1 || test_flag == 3 ||
	    test_flag == 5 || test_flag == 6) {
	      PR(ER, "%s: file '%s' is the map-file. To plot amplitude ",
		  myname, filename);
	      PR(ER, "diagram, the program needs the corresponding file\n.");
	      exit(1);
	}
	/* set color table */
	setcol_i(col_map);
	font_size = 5;

	if (test_flag == 7) {
	    /* read from 3-column ASCII table */
	    plot_ascii_ampl_diagram(filename, lmax, xs,ys,bar,font_size,
		output_name, min, max, log_flag);
	} else {
	    /* read and calc from a_lm FITS file */
	    plot_ampl_diagram(filename, lmax, xs,ys,bar,font_size,
		output_name, min, max, log_flag);
	}
	exit (0);
    }

    /*------------------------------------*/
    /* set limits of values to be plotted */
    /*------------------------------------*/
    if (nv_intr) {   /* when we have several intervals for limits */
	for(i=0; i<nx; i++) {
	    int ii = i * np;
	    for(j=0; j<nph[i]; j++) {
		int n = ii+j;
		int flag = 0, k;
		for (k=0; k<nv_intr; k++) {
		    if (v_intr[k].low < ar[n] && ar[n] < v_intr[k].high){
			flag = 1;
			break;
		    }
		}
		if ( !flag ) ar[n] = dummy;
	    }
	}
    }

    /*-------------------------*/
    /* read region for masking */
    /*-------------------------*/
    if ( z_mask_flag ) {
       register int is=0;
       int ns = 1;
       struct _zone_  *zz = (struct _zone_ *)0;
       extern void map_zone(struct _zone_ *, int, int,
	       double *, double *, int, int, int *, double);
       int function_type= z_remain_flag ? 0:1; /* 1:remove_flag, 0:remain */
       int zone_type = zone_circle ? 0 : 1;
       double lim1 = zmask_value * (1. - 1.e-6),
	      lim2 = zmask_value * (1. + 1.e-6);

       if((zmask=(double*)calloc((unsigned)(nx*np),sizeof *zmask))
	 ==(double*)0)
	    goto nomem;

       /* if to read zone-file */
       if (zone_file[0] != '\0') {
	    FILE *fz;
	    char str[81];
	    /* read data from file with zones */
	    if((fz = fopen(zone_file, "r")) == (FILE *)0) {
		PR(ER, "%s: I can not open file '%s'...\n",
		    myname, zone_file);
		exit(1);
	    }
	    /* get number of zones */
	    is =0;
	    while(fgets(str, 80, fz) != NULL) {
		if (*str == '#' || *str == '\n')
		    continue;
		is++;
	    }
	    /* allocate memory */
	    i = ns = is;
	    zz = (struct _zone_ *)calloc(i, sizeof *zz);
	    if (zz == (struct _zone_ *)0) goto nomem;

	    /* read zones  */
	    is=0;
	    fseek(fz, 0l, 0);
	    while(fgets(str, 80, fz) != NULL) {
		char s1[30], s2[30], s3[30], s4[30];
		char *cs = str;
		int n_par;  /* number of parameters per string */
		if (*str == '#' || *str == '\n')
			continue;
		while(*cs) if (*cs == '#') {*cs = '\0'; break;} else cs++;

		n_par = n_skip_wd(str);
		switch (n_par) {
		case 1:
		    PR(ER,"%s: there is only 1 parameter in the string:\n",
			    myname);
		    PR(ER,"%s", str);
		    PR(ER, "\n%s: I've skipped it...\n", myname);
		    break;
		case 2:
		    zz[is].latR0  = deg2rad((skip_wd(0,str,s1),s1));
		    zz[is].lonR0  = deg2rad((skip_wd(1,str,s2),s2));
		    zz[is].radius = zone_radius;
		    zz[is].type = 0;
		    break;
		case 3:
		    zz[is].latR0  = deg2rad((skip_wd(0,str,s1),s1));
		    zz[is].lonR0  = deg2rad((skip_wd(1,str,s2),s2));
		    if (zone_radius > 0.)
			zz[is].radius = zone_radius;
		    else
			zz[is].radius = deg2rad((skip_wd(2,str,s3),s3));
		    zz[is].type = 0;
		    break;
		case 4:
		    zz[is].latR0 = deg2rad((skip_wd(0,str,s1),s1));
		    zz[is].lonR0 = deg2rad((skip_wd(1,str,s2),s2));
		    zz[is].latR1 = deg2rad((skip_wd(2,str,s3),s3));
		    zz[is].lonR1 = deg2rad((skip_wd(3,str,s4),s4));
		    if (zone_type == 0) {
			    zz[is].radius = zz[is].latR1;
			    zz[is].type = 0;
		    } else  zz[is].type = 1;
		    break;
		}
		is++;
	    }
	    fclose(fz);
       } else {
	 /* allocate memory for 1 zone */
	 /* from command string */
	 i = ns = 1;
	 zz = (struct _zone_ *)calloc(i, sizeof *zz);
	 if (zz == (struct _zone_ *)0) goto nomem;

	 if ( flag_conc_fig ) {  /* 1-circle, 2 - rectangle */
	    extern void map_ring_zone();

	    if ( flag_conc_fig == 1) {
		zz[0].latR0 = lat_c;  zz[0].lonR0 = lon_c;
		zz[0].radius = r_c;
		zz[0].dr = w_c;
		zz[0].type = 0; /* circle */
	    } else {
		zz[0].latR0 = lat_c;   zz[0].lonR0 = lon_c;
		zz[0].latR1 = lat1_c;  zz[0].lonR1 = lon1_c;
		zz[0].dr = w_c;
		zz[0].type = 1; /* rectangule */
	    }
	    map_ring_zone(zz,ns,function_type,zmask,x,nx,np,nph, zmask_value);
	 } else {
	    zz[0].latR0 = zone_latR0;  zz[0].lonR0 = zone_lonR0;
	    zz[0].latR1 = zone_latR1;  zz[0].lonR1 = zone_lonR1;
	    zz[0].radius = zone_radius;
	    if (zone_latR1 < -900.) {
		    zz[0].type = 0;
	    } if (zone_lonR1 < -900.) {
		    if (zone_radius > 0.)
			zz[is].radius = zone_radius;
		    else
			zz[is].radius = zone_latR1;
		    zz[0].type = 0;  /* circle */
	    } else {
		    zz[0].type = 1;  /* rectangle */
	    }
	 }
       }
       if ( !flag_conc_fig ) {  /* 1-circle, 2 - rectangle */
	  if (flag_pix == 'h') {
	     extern void hmap_zone(struct _zone_ *, int, int, double *,
		  double *, double *, int, int, int *, double);
	     hmap_zone(zz,ns,function_type, zmask, x, offs, nx,np,nph,
		  zmask_value);
	  } else
	     map_zone(zz, ns,function_type, zmask, x, nx, np, nph,
		  zmask_value);
       }
       /* free memory from zone-array */
       if (zz != (struct _zone_ *)0) free((void *)zz);

       /* set mask given with external data in plotted map */
       /*--------------------------------------------------*/
       for(i=0; i<nx; i++) {
	    int ii = i * np;
	    for(j=0; j<nph[i]; j++) {
		int n = ii+j;
		if (z_mask_flag) {
		  if (z_remove_flag && (lim1 < zmask[n] && zmask[n] < lim2))
			 ar[n] = dummy;
		  if (z_remain_flag && ( !(zmask[n] < lim1 || lim2<zmask[n])))
			 ar[n] = dummy;
		}
	    }
       }
       free((void *)zmask);
       zmask = (double *)0;
    }

    /* start to plot for polarization arrays */
    if (flag_map==3)  ar = f;

    /* set mask given with internal data of plotted map */
    /*--------------------------------------------------*/
    if (pixel_mask) {
	for(i=0; i<nx; i++) {
	    int ii = i * np;
	    for(j=0; j<nph[i]; j++) {
		int n = ii+j;
		if (ABS(ar[n] - GLESPMASK) < 500)  ar[n] = dummy;
	    }
	}
    }



if_all_maps:

    /* set mask from HPix-file */
    if (*hp_mask_file != '\0') {
	    extern void mask_hp(double*,double*,int,int,int*, char*, double);
	    mask_hp(ar,x,nx,np,nph,hp_mask_file,dummy);
    }
    /* set mask from GLESP-file */
    if (*gl_mask_file != '\0') {
	    extern void mask_gl(double*,double*,int,int,int*, char*, double);
	    mask_gl(ar,x,nx,np,nph,gl_mask_file,dummy);
    }

when_use_file_list:

    /* use sources from list of areas to be extracted into Basic FITS */
    if (format_ext_areas) {
	  char str_dat[81];
	  char name[51], sra[31], sdec[31], ssize[21];
	  double ra =0.0, dec=0.0, glat=0.0, glon=0.0, bb=0.0;
	  double box_size =0.5, p=180./M_PI;

get_area_str_again:
	  if (fgets (str_dat, 81, fpl) == NULL) {
		fclose(fpl);
		if (fp_in  != (FILE *)0 && fp_in  != stdin)  fclose(fp_in);
		if (fp_out != (FILE *)0 && fp_out != stdout) fclose(fp_out);
		exit(0);
	  }
	  if (*str_dat == '#')
		goto get_area_str_again;

	  name[0] = sra[0] = sdec[0] = ssize[0] = '\0';
	  switch(format_ext_areas) {
	  case 'q': /* equatorial coordinates */
		{
		  double s_ra; int h_ra, m_ra;
		  double ds, s_dec; int d_dec, m_dec, c;
		  sscanf(str_dat, "%s%s%s", sra, sdec, ssize);
		  ra = str2deg(sra);
		  dec = ds = str2deg(sdec);
		  bb = str2deg(ssize);
		  h_ra = (int)ra;
		  m_ra = (int)((ra - h_ra) * 60.);
		  s_ra = (ra - h_ra - m_ra/60.) * 3600;
		  if ( ds < 0. ) {ds = ABS(dec); c='-';} else c='+';
		  d_dec = (int)ds;
		  m_dec = (int)((ds - d_dec) * 60.);
		  s_dec = (ds - d_dec - m_dec/60.) * 3600;
		  sprintf(name, "%02d%02d%02.0f%c%02d%02d%02.0f",
		      h_ra, m_ra, s_ra, c, d_dec, m_dec, s_dec);
		}
		break;
	  case 'Q': /* equatorial coordinates */
		  sscanf(str_dat, "%s%s%s%s", name, sra, sdec, ssize);
		  ra = str2deg(sra);
		  dec = str2deg(sdec);
		  bb = str2deg(ssize);
		  break;
	  case 'G': /* galactic coordinates */
		  sscanf(str_dat, "%s%s%s%s", name, sra, sdec, ssize);
		  glat = str2deg(sra);
		  glon = str2deg(sdec);
		  bb = str2deg(ssize);
		  break;
	  case 'g': /* galactic coordinates */
	  default :
		{
		  double ds=0.0, s_lat; int d_lat, m_lat, c;
		  double s_lon; int d_lon, m_lon;
		  sscanf(str_dat, "%s%s%s", sra, sdec, ssize);
		  glat = ds = str2deg(sra);
		  glon = str2deg(sdec);
		  bb = str2deg(ssize);
		  d_lon = (int)glon;
		  m_lon = (int)((glon - d_lon) * 60.);
		  s_lon = (glon - d_lon - m_lon/60.) * 3600;
		  if ( ds < 0. ) {ds = ABS(glat); c='-';} else c='+';
		  d_lat = (int)ds;
		  m_lat = (int)((ds - d_lat) * 60.);
		  s_lat = (ds - d_lat - m_lat/60.) * 3600;
		  sprintf(name, "G%03d%02d%02.0f%c%02d%02d%02.0f",
		       d_lon, m_lon, s_lon, c, d_lat, m_lat, s_lat);
		}
		break;
	  }

	  /* box size in degs */
	  if (bb > 0.) box_size = bb;
	  else if (b_size > 0.) box_size = b_size;
	  box_size /= p;
	  /* same box side --> same x,y-size */
	  xs = ys;

	  switch(format_ext_areas) {
	  case 'q': /* equatorial coordinates */
	  case 'Q':
		ra *= 15/p;
		dec /= p;

		/* RA limits in radians /*/
		zcut_latR1 = ra - .5 * box_size / cos (dec);
		zcut_latR2 = zcut_latR1 + box_size / cos (dec);
		/* Dec limits in rad /*/
		zcut_lonR1 = dec - .5 * box_size;
		zcut_lonR2 = zcut_lonR1 + box_size;

		projection = 'Q';
		break;
	  case 'g': /* galactic coordinates */
	  case 'G':
		glat /= p;
		glon /= p;

		/* GLON in radians /*/
		zcut_lonR1 = glon - .5 * box_size / cos (glat);
		zcut_lonR2 = zcut_lonR1 + box_size / cos (glat);
		/* GLAT limits in rad /*/
		zcut_latR1 = glat - .5 * box_size;
		zcut_latR2 = zcut_latR1 + box_size;

		projection = 's';
		break;
	  }

	  /* create output file for GIF data of area */
	  if ( *dir_out == '\0' )
	      sprintf (output_name, "%s.gif", name);
	  else
	      sprintf (output_name, "%s/%s.gif", dir_out, name);
    }

    /* set new limits */
    /*----------------*/
    if (flag_pix == 'f') {  /* Basic FITS data  */
	/* find limits for 2-D image */
	int nax1=0, nax2=0;
	double crv1=0.0, crv2=0.0, cdlt1=0.0, cdlt2=0.0;
	double xf, yf;
	extern void read_basic_fits(char *, double **ar, int *, int *,
	     double *, double *, double *, double *);

	if (f != (double *)0) {
	    PR(ER, "%s: memory for input map is occupied...\n", myname);
	    exit(1);
	}
	/* read Basic FITS data */
	read_basic_fits(filename, &f, &nax1, &nax2, &crv1, &crv2,
		&cdlt1, &cdlt2);
	ar = f;

	if ( nax2 == nax1 && nax2 < ys)  xs = ys;
	else {
	    if (xs < nax1) xs = nax1;
	    if (ys < nax2) ys = nax2;
	}
	zcut_latR1 = crv1;
	zcut_latR2 = crv1 + cdlt1*(nax1 - 1);
	zcut_lonR1 = crv2;
	zcut_lonR2 = crv2 + cdlt2*(nax2 - 1);

	if (set_limits) {
	    for(i=0; i<nax1; i++) {
		int ii = i * nax2;
		for(j=0; j<nax2; j++) {
		    int n = ii+j;
		    if(ar[n] < min)  ar[n] = min;
		    if(ar[n] > max)  ar[n] = max;
		}
	    }
	} else {
	    for(i=0; i<nax1; i++) {
		int ii = i * nax2;
		for(j=0; j<nax2; j++) {
		    int n = ii+j;
		    if(ar[n] < fmin)  fmin = ar[n];
		    if(ar[n] > fmax)  fmax = ar[n];
		}
	    }
	    set_gdlim(fmin, fmax);
	}

	/* create img-array for plotting */
	if((img=(double*)calloc((unsigned)(xs*ys),sizeof *img))==(double*)0){
	    PR(ER, "%s: no memory for image creation...\n", myname);
	    exit(1);
	}

	/* copy data array to img-array for plotting */
	xf = (double)(nax1-1)/(xs-1);
	yf = (double)(nax2-1)/(ys-1);
	for(i=0; i<ys; i++) {
	    register int ii = i * xs;
	    int ip = i*yf+.5;
	    int iip = ip * nax1;
	    for(j=0; j<xs; j++) {
		int jp = j*xf+.5;
		img[ii+j] = ar[iip+jp];
	    }
	}

	projection = 'f';
    } else {    /* analysis of data on a sphere */
	if (set_limits) {
	    for(i=0; i<nx; i++) {
		int ii = i * np;
		for(j=0; j<nph[i]; j++) {
		      int n = ii+j;
		      if (pixel_mask && ABS(ar[n] - GLESPMASK) < 500)
			    continue;
		      if(ar[n] < min)  ar[n] = min;
		      if(ar[n] > max)  ar[n] = max;
		}
	    }
	} else {
	    for(i=0; i<nx; i++) {
		int ii = i * np;
		for(j=0; j<nph[i]; j++) {
		      int n = ii+j;
		      if (pixel_mask && ABS(ar[n] - GLESPMASK) < 500)
			    continue;
		      if(ar[n] < fmin)  fmin = ar[n];
		      if(ar[n] > fmax)  fmax = ar[n];
		}
	    }
	    if(*filename == '\0') {
		fmin = -1.; fmax = 1.;
	    }
	    set_gdlim(fmin, fmax);
	}
    }

    /* set color table */
    setcol_i(col_map);

    /* create image and copy map to output file */
    /*------------------------------------------*/
    switch (projection) {
    case 'm':     /* Mollweide projection */
	if (flag_pix == 'h') {
	    extern void ht2mol(double *, int, int, double *, double *,
		int *, double **,  int, int, double, int **);
	    ht2mol(ar, nx, np, x, offs, nph, &img, xs, ys, lon0, &mask);
	} else
	    t2mol(ar, nx, np, x, nph, &img, xs, ys, lon0, &mask);
	break;
    case 'r': /* simple rectanular projection */
	if (flag_pix == 'h') {
	    extern void ht2rect(double *, int, int, double *, double *,
		int *, double **,  int, int, double, int **);
	    ht2rect(ar, nx, np, x, offs, nph, &img, xs, ys, lon0, &mask);
	} else {
	    extern void t2rect(double *, int, int, double *, int *,
		double **,  int, int, double, int **);
	    t2rect(ar, nx, np, x, nph, &img, xs, ys, lon0, &mask);
	}
	break;
    case 's': /* cut rectangular image in Galactic system   */
    case 'Q': /* cut rectangular image in equatorial system */
	 if (projection == 's') {
	  if (flag_pix == 'h') {
	    /* Hpix data */
	    extern void sq2hrect(double *, int, int, double *, double *,
		int *,double **, int,int,double,double,double,double, int **);
	    sq2hrect(ar, nx, np, x, offs, nph, &img, xs, ys,
		zcut_latR1,zcut_lonR1,zcut_latR2,zcut_lonR2, &mask);
	  } else {
	    /* GLESP data */
	    extern void sq2rect(double *, int, int, double *, int *,
		double **,  int, int, double,double,double,double, int **);
	    sq2rect(ar, nx,np, x, nph, &img, xs, ys,
		zcut_latR1,zcut_lonR1,zcut_latR2,zcut_lonR2, &mask);
	  }
	} else {  /* if (projection == 'Q') */
	  if (flag_pix == 'h') {
	    extern void eq2hrect(double *, int, int, double *, double *,
		int *,double **, int,int,double,double,double,double, int **);
	    eq2hrect(ar, nx, np, x, offs, nph, &img, xs, ys,
		zcut_latR1,zcut_lonR1,zcut_latR2,zcut_lonR2, &mask);
	  } else {
	    /* GLESP data */
	    extern void eq2rect(double *, int, int, double *, int *,
		double **,  int, int, double,double,double,double, int **);
	    eq2rect(ar, nx,np, x, nph, &img, xs, ys,
		zcut_latR1,zcut_lonR1,zcut_latR2,zcut_lonR2, &mask);
	  }
	}
	/* set natural (real) limits from image */
	if (set_limits == 3) {
	    int i, j;
	    double q_min = 1.e+30, q_max = -1.e+30;
	    for(i=0; i<ys; i++) {
		register int ii = i * xs;
		for(j=0; j<xs; j++) {
		    if (img[ii+j] < q_min)  q_min = img[ii+j];
		    if (img[ii+j] > q_max)  q_max = img[ii+j];
		}
	    }
	    set_gdlim(q_min, q_max);
	}
	break;
    case 'f':  /* image from the Basic FITS file */
	projection = 's';
	/* set natural (real) limits from image */
	if (set_limits == 3) {
	    int i, j;
	    double q_min = 1.e+30, q_max = -1.e+30;
	    for(i=0; i<ys; i++) {
		register int ii = i * xs;
		for(j=0; j<xs; j++) {
		    if (img[ii+j] < q_min)  q_min = img[ii+j];
		    if (img[ii+j] > q_max)  q_max = img[ii+j];
		}
	    }
	    set_gdlim(q_min, q_max);
	}

	break;
    }

    /* find memory for image */
    imgs = (int *)calloc(xs*ys, sizeof *imgs);
    if (imgs == (int *)0) {
	PR(ER, "%s: no memory for image...\n", myname);
	exit(1);
    }

    /* set limits by hands with force */
    if (set_limits == 2)
	    set_gdlim(min, max);

    /* converse float map to scaled integer image */
    map2img(img, imgs, xs, ys, mask);

    /* if sources are in figure to marked */
    if (*source_file != '\0' || *mark_str != '\0' ||
      *inp_amp_pol_f != '\0' || *inp_pol_file != '\0' ) {
	int nm = 0, nm2 = 0;
	struct _crm_ *crm2 = (struct _crm_ *)0;

	if (*mark_str != '\0') {
	    read_mark_str(&crm, &nm, mark_str, coordinates_flag);

	    if (crm[0].mark== -1 || bt_m) crm[0].mark  = inp_mark;
	    if (crm[0].size  < 0 || bt_s) crm[0].size  = inp_size;
	    if (crm[0].color < 0 || bt_c) crm[0].color = inp_col;
	} else if (*source_file != '\0') {
	    if (inp_ff_flag) {  /* input F-file */
	       struct _coor_  *cr = (struct _coor_ *)0;
	       double delta_dec = 0.0;

	       /* read (lat,lon) from F-file */
	       read_ff(&cr, &nm, source_file, flag_coor, delta_dec);
	       coordinates_flag = 1; /* galactic cooordinates */

	       /* find memory for arrays */
	       if ((crm = (struct _crm_ *)calloc((unsigned)nm,
		   sizeof *crm)) == (struct _crm_ *)0) {
		     PR(ER, "%s: can not find memory for array...\n",
			   myname);
		       exit(1);
	       }
	       for(i=0; i<nm; i++) {
		     gal2pol(cr[i].lat, cr[i].lon,
			    &crm[i].theta, &crm[i].phi);

		     if (crm[i].mark==-1|| bt_m) crm[i].mark  =inp_mark;
		     if (crm[i].size <0 || bt_s) crm[i].size  =inp_size;
		     if (crm[i].color<0 || bt_c) crm[i].color =inp_col;
	       }
	       free((void *)cr);
	    } else if (cats_input) {  /* input from the CATS database */
				 /* http://cats.sao.ru           */
	       struct _coor_  *cr = (struct _coor_ *)0;

	       /* read (lat,lon) from F-file */
	       read_cats(&cr, &nm, source_file, flag_coor);
	       coordinates_flag = 1; /* galactic coordinates */

	       /* find memory for arrays */
	       if ((crm = (struct _crm_ *)calloc((unsigned)nm,
		   sizeof *crm)) == (struct _crm_ *)0) {
		     PR(ER, "%s: can not find memory for array...\n",
			   myname);
		       exit(1);
	       }
	       for(i=0; i<nm; i++) {
		     gal2pol(cr[i].lat, cr[i].lon,
			    &crm[i].theta, &crm[i].phi);

		     if (crm[i].mark==-1|| bt_m) crm[i].mark  =inp_mark;
		     if (crm[i].size <0 || bt_s) crm[i].size  =inp_size;
		     if (crm[i].color<0 || bt_c) crm[i].color =inp_col;
	       }
	       free((void *)cr);
	    } else {
	       read_marker(&crm, &nm, source_file, coordinates_flag);
	    }
	}

	if ( *inp_pol_file != '\0' ) {
	    read_mark_pol(&crm2, &nm2, inp_pol_file, coordinates_flag);
	    for(i=0; i<nm2; i++) {
		  if (crm2[i].size <0 || bt_s) crm2[i].size  =inp_size;
		  if (crm2[i].color<0 || bt_c) crm2[i].color =inp_col;
	    }
	}

	/* plot-library operations */
	openpl_gd(imgs, xs, ys);        /* open function  */
	switch (projection) {
	case 'm':     /* Mollweide projection */
	    marker2mol(crm,  nm, lon0);      /* output markers */
	    if ( *inp_pol_file != '\0' )
		marker2mol(crm2, nm2, lon0); /* output polarization */
	    break;
	case 'r':     /* rectangular projection */
	case 's':     /* cut rectangular image  */
	case 'Q':     /* cut rectangular image in Equat.system */
	    break;
	}


	if(*inp_amp_pol_f != '\0') {
	      /* put directed polarization vectors */
	      register int k;
	      double *amp = (double *)0, *ang = (double *)0;
	      double *x1 = (double *)0, *x2 = (double *)0;
	      int *nph1 = (int *)0, *nph2 = (int *)0;
	      struct _crm_ *cr = (struct _crm_  *)0;
	      int nx1, nx2, np1, np2;
	      double pi2 = 2. * M_PI;

	      read_glesp_map(&amp, &x1, &nx1,&np1, &nph1, inp_amp_pol_f);
	      read_glesp_map(&ang, &x2, &nx2,&np2, &nph2, inp_ang_pol_f);
	      if (nx2 != nx1 && np2 != np1) {
		 PR(ER, "%s: different dimensions of data in files: \n",
		    myname);
		 PR(ER, "'%s' and '%s'\n", inp_amp_pol_f,inp_ang_pol_f);
		 exit(1);
	      }
	      free((void *)x2);
	      free((void *)nph2);

	      nm = 0;
	      for(i=0; i<nx; i++)   nm += nph[i];

	      if((cr=(struct _crm_ *)calloc(nm, sizeof *cr)) ==
		   (struct _crm_ *)0) {
		     PR(ER, "%s: no memory for coordinates array...\n",
			myname);
		     exit(1);
	      }

	      for(k=i=0; i<nx1; i++)  {
		   register int ii = i * np1;
		   double dphi = pi2 / nph1[i];
		   for(j=0; j<nph1[i]; j++) {
			double phi = dphi * j;
			cr[k].mark  = -20;
			cr[k].theta = acos(x1[i]);
			cr[k].phi   = phi;
			cr[k].color = inp_col;
			cr[k].size  = size_factor * 200. * amp[ii+j];
			cr[k].vpa   = ang[ii+j];
			k++;
		   }
	      }
	      marker2mol(cr,  nm, lon0);     /* output markers */
	      free((void *)cr);
	}
	closepl_gd();                   /* close function */
    }

    /* overlay coordinate grid */
    if (grid_type >= 0) {
	extern void grid2mol(int, int, double, double, double),
		    grid2rect(int, int, double, double, double),
		    grid2srect(int, int, double, double, double, double,
			   double, double);
	extern int check_gd_open(void);
	if (!check_gd_open())
	     openpl_gd(imgs, xs, ys);  /* initialize gd-window */
	switch (projection) {
	case 'm':     /* Mollweide projection */
		grid2mol(grid_type, grid_color, lon0, d1_step, d2_step);
		break;
	case 'r':     /* rectangular projection */
		grid2rect(grid_type, grid_color, lon0, d1_step, d2_step);
		break;
	case 's':    /* cut rectangular image in Galactic system */
	case 'Q':    /* cut in equatorial system */
		grid2srect(grid_type, grid_color,
		      zcut_latR1,zcut_lonR1,zcut_latR2,zcut_lonR2,
		      d1_step, d2_step);
		break;
	}
	closepl_gd();                   /* close function */
    }

    /* overlay data mask from file */
    if (overlay_data) {
	extern int check_gd_open(void);
	extern void overlay2mol(int,int,double,double*,double*,int,int,int*,
		int),
	    overlay2rect(int,int,double,double*,double*,int,int,int*, int),
	    overlay2sq(int,int,double,double,double,double,
		double*,double*,int,int,int*, int);
	double *fo = (double *)0, *xo = (double *)0;
	int nxo=0, npo=0, *npho=(int *)0;
	if (!check_gd_open())
	     openpl_gd(imgs, xs, ys);  /* initialize gd-window */
	read_glesp_map(&fo, &xo, &nxo, &npo, &npho, overlay_file);

	switch (projection) {
	case 'm':     /* Mollweide projection */
		overlay2mol(overlay_color, overlay_width,
			 lon0, fo,xo,nxo,npo,npho, overlay_flag);
		break;
	case 'r':     /* rectangular projection */
		overlay2rect(overlay_color, overlay_width,
			 lon0, fo,xo,nxo,npo,npho, overlay_flag);
		break;
	case 's': /* cut rectangular image in Galactic system */
	case 'Q': /* cut in equatorial system */
		overlay2sq(overlay_color, overlay_width,
		     zcut_latR1, zcut_lonR1, zcut_latR2, zcut_lonR2,
		     fo,xo,nxo,npo,npho, overlay_flag);
		break;
	}
	closepl_gd();                  /* close function */
	if (fo != (double *)0) free((void *)fo);
	if (xo != (double *)0) free((void *)xo);
	if (npho  != (int *)0) free((void *)npho);
    }

    /* overlay vector graphics with some patterns */
    if (patt_overlay > 0) {
	extern void patt2mol(int, double);
	 extern int check_gd_open(void);
	if (!check_gd_open())
	     openpl_gd(imgs, xs, ys);  /* initialize gd-window */
	patt2mol(patt_overlay, lon0);
	closepl_gd();                   /* close function */
    }

    /* overlay vector graphics */
    if (vect_overlay > 0) {
	extern void vect2mol(char *, double);
	 extern int check_gd_open(void);
	if (!check_gd_open())
	     openpl_gd(imgs, xs, ys);  /* initialize gd-window */
	vect2mol(plotfile, lon0);
	closepl_gd();                   /* close function */
    }

    /* add bar and rebuild image */
    if (bar) {
	set_gd_font_size(font_size);
	color_bar(imgs, &img_bar); /* add color bar */
	free((void *)imgs);        /* free memory from prev.map */
	imgs = img_bar;            /* change pointer to new map */
    }

    /* add text and rebuild image */
    if (txt_n) {
	extern void add_text(int *, int **, struct _txt_ *, int);

	set_gd_font_size(font_size);
	add_text(imgs, &img_txt, txt, txt_n); /* add text */
	free((void *)imgs);        /* free memory from prev.map */
	imgs = img_txt;            /* change pointer to new map */
    }

    /* convert fig to gif */
    /* and write to file  */
    map2file(imgs, output_name);

    if (flag_map == 3) { /* number of maps in 3-file */
	static char st_t[81], st_o[301];

	if (set_limits != 2 && set_limits != 3) {
	    fmin= 1.e+30, fmax= -1.e+30;
	}

	if (ar == f) {
	    strcpy(st_t, txt[0].text);     /* save text on a picture */
	    strcpy(st_o, output_name);     /* save basic name */
	    ar = q;
	    strcat(output_name, "_q");
	    strcat(txt[0].text, "_q");
	    goto if_all_maps;  /* loop over maps */
	} else if (ar == q) {
	    strcpy(output_name,  st_o);    /* restore basic name */
	    strcpy(txt[0].text,  st_t);
	    ar = u;
	    strcat(output_name,   "_u");
	    strcat(txt[0].text, "_u");
	    goto if_all_maps;  /* loop over maps */
	}
    }

    if (format_ext_areas) { /* use filenames from list */
	if (set_limits != 2 && set_limits != 3) {
	    fmin= 1.e+30, fmax= -1.e+30;
	}
	goto when_use_file_list; /* loop over files in list */
    }

    /* close output file  */
    if (fp_in  != (FILE *)0 && fp_in  != stdin)  fclose(fp_in);
    if (fp_out != (FILE *)0 && fp_out != stdout) fclose(fp_out);

    exit(0);
}

void setopt(argc, argv)
int argc;
char **argv;
{
    register int i;
    myname = argv[0];
    int no_title = 0, tf_flag=0;
    int x_set=0, y_set=0;
    int gr_set = 0;

    /* title: 1st string in a text: save place for element txt_n=0 */
    txt_n++;

    /* loop over external parameters */
    for(i=1; i<argc; i++) {
	switch(argv[i][0]) {
	case '-':
		switch(argv[i][1]) {
		case '\0':
			strcpy(filename, "-");
			break;
		case '0':
			strcpy(filename, "-");
			dummy = 0.0;
			break;
		case '1':
			strcpy(filename, "-");
			dummy = 1.0;
			break;
		case 'a':
			if (strncmp(argv[i]+1, "am", 2) == 0) {
			/* draw diagram for amplitude special file amp(l,m)*/
			  ampl_diagram = 1;
			}
			break;
		case 'b':
			bar = 0; /* no bar */
			break;
		case 'c':
			if (strncmp(argv[i]+1, "cr", 2)==0 ||
			    strncmp(argv[i]+1, "co", 2)==0)
				coordinates_flag = 1;
			else if (strncmp(argv[i]+1, "ca", 2)==0) {
				cats_input = 1;
				strcpy(source_file, argv[++i]);
			} else if (strncmp(argv[i]+1, "cc", 2)==0) {
			    switch(argv[i][2]) {
			    case 'c': /* concentric figure */
			      { char vstr[30];
			      switch(argv[i][3]) {
			      case 's':
				 break;
			      case 'r':  /* rectangle */
				 ++i;
				 skip_wd(0, argv[i], vstr);
				 lat_c = deg2rad(vstr);
				 skip_wd(1, argv[i], vstr);
				 lon_c = deg2rad(vstr);
				 skip_wd(2, argv[i], vstr);
				 lat1_c = deg2rad(vstr);
				 skip_wd(3, argv[i], vstr);
				 lon1_c = deg2rad(vstr);
				 /* radius of ring width */
				 skip_wd(4, argv[i], vstr);
				 w_c = deg2rad(vstr);
				 flag_conc_fig = 2;
				 break;
			      default :  /* circle */
				 ++i;
				 skip_wd(0, argv[i], vstr);
				 lat_c = deg2rad(vstr);
				 skip_wd(1, argv[i], vstr);
				 lon_c = deg2rad(vstr);
				 skip_wd(2, argv[i], vstr);
				 /* radius of inner circle */
				 r_c = deg2rad(vstr);
				 skip_wd(3, argv[i], vstr);
				 /* radius of ring width */
				 w_c = deg2rad(vstr);
				 flag_conc_fig = 1;
				 break;
			      }}
			    }
			    z_remain_flag = 1;
			    z_mask_flag = 1;
			} else if (strncmp(argv[i]+1, "cu", 2)==0 ||
				   strncmp(argv[i]+1, "cg", 2)==0) {
			    /* select a zone to be cut and plotted */
			    char s1[30], s2[30], s3[30], s4[30];
			    get_4_str(argv[++i], s1, s2, s3, s4);
			    if(*s1 !='\0') zcut_latR1=deg2rad(s1);
				else zcut_latR1= -1001.;
			    if(*s2 !='\0') zcut_lonR1=deg2rad(s2);
				else zcut_lonR1= -1001.;
			    if(*s3 !='\0') zcut_latR2=deg2rad(s3);
				else zcut_latR2= -1001.;
			    if(*s4 !='\0') zcut_lonR2=deg2rad(s4);
				else zcut_lonR2= -1001.;
			    projection = 's';
			} else if (strncmp(argv[i]+1, "cq", 2)==0) {
			    /* select a zone to be cut and plotted */
			    /* in equatorial coordinate limits     */
			    char s1[30], s2[30], s3[30], s4[30];
			    get_4_str(argv[++i], s1, s2, s3, s4);
			    if(*s1 !='\0') zcut_latR1=deg2rad(s1)*15.;
				else zcut_latR1= -1001.;
			    if(*s2 !='\0') zcut_lonR1=deg2rad(s2);
				else zcut_lonR1= -1001.;
			    if(*s3 !='\0') zcut_latR2=deg2rad(s3);
				else zcut_latR2= -1001.;

			    if (strncmp(argv[i]+1, "cqa", 3)==0) {
				/* s1,s2,s3: ra, deg, size_deg */
				/* set area bounaries */
				/* for declination */
				if (zcut_lonR1 > -1001.)
				    zcut_lonR1 -= .5 * zcut_latR2;
				zcut_lonR2 = zcut_lonR1 + zcut_latR2;

				/* for right ascension */
				if (zcut_latR1 > -1001.)
				    zcut_latR1 -= .5 * zcut_latR2/15.;
				zcut_latR2 = zcut_latR1 + zcut_latR2/15.;
			    } else {
				/* last coordinate for rectangular */
				if(*s3 !='\0') zcut_latR2 *= 15.;
				    else zcut_latR2= -1001.;
				if(*s4 !='\0') zcut_lonR2=deg2rad(s4);
				    else zcut_lonR2= -1001.;
			    }
			    projection = 'Q';
			} else
				col_map = atoi(argv[++i]);
			break;
		case 'C':
			i++;
			if (argv[i][0] != ',') {
			    if (strncmp(argv[i], "none", 4) != 0) {
				 if (strncmp(argv[i], "na", 2) == 0) {
				    set_limits = 3;
				 } else {
				    min = atof(argv[i]);
				    if (!set_limits) set_limits = 1;
				 }
			    }
			}
			{
			    char *ps = strchr(argv[i], ',');
			    if (ps != NULL && *ps && *(ps+1)) {
				 if (strncmp(ps+1, "none", 4) != 0) {
				      max = atof(ps+1);
				      if (!set_limits) set_limits = 1;
				 }
			    }
			}
			if (argv[i-1][2]=='s' && set_limits && set_limits!=3)
			    set_limits = 2;
			if (min > max) {
			    PR(ER, "%s: check limits: min=%g, max=%g\n",
				myname, min, max);
			    exit(2);
			}
			break;
		case 'd':
			switch(argv[i][2]) {
			case '1': /* step for grid by 1st direction, degs */
			    d1_step = atof(argv[++i]);
			    break;
			case '2': /* step for grid by 2nd direction, degs */
			    d2_step = atof(argv[++i]);
			    break;
			case 'i':  /* '-dir ' */
			    strcpy(dir_out, argv[++i]);
			    break;
			default :
			    dummy = atof(argv[++i]);
			    break;
			}
			break;
		case 'e':
		case 'E':
			if (argv[i][2] == 'q')
			    grid_type = 1; /* equatorial */
			else if (argv[i][2] == 'c')
			    grid_type = 2; /* ecliptic   */
			else {
			    PR(ER, "%s: illegal flag '%s'\n",myname,argv[i]);
			    usage();
			}
			break;
		case 'f':
			switch(argv[i][2]) {
			case 'f':
				inp_ff_flag = 1;
				strcpy(source_file, argv[++i]);
				break;
			case 'e':
				flag_coor = 0; /* equatorial coordinates */
				coordinates_flag = 2;
				break;
			case 'g':
				flag_coor = 0; /* galactic coordinates */
				coordinates_flag = 1;
				break;
			default :
				strcpy(source_file, argv[++i]);
				break;
			case 'z':
				if (strncmp("fz", argv[i]+1, 2) == 0) {
				    format_ext_areas = 'g';
				    if (argv[i][3] == 'q') {
					if (argv[i][4] == '1')
					    format_ext_areas = 'Q';
					else
					    format_ext_areas = 'q';
				    } else if (argv[i][3] == 'g') {
					if (argv[i][4] == '1')
					    format_ext_areas = 'G';
					else
					    format_ext_areas = 'g';
				    }
				    strcpy(file_ext_areas, argv[++i]);
				}
			}
			break;
		case 'g':
			if (strncmp(argv[i]+1, "gr", 2) == 0) {
			    /* set grid type for coordinate system */
			    if (argv[i][3] == 'q' || argv[i][3] == 'Q') {
				if (gr_set) {
				    switch (grid_type) {
				    case 0:  grid_type = 12;  break;
				    case 1:  grid_type = 12;  break;
				    case 3:  grid_type = 32;  break;
				    case 13: case 31:
					     grid_type = 123; break;
				    default: grid_type = 2;   break;
				    }
				} else grid_type = 2; /* equatorial */
			    } else if(argv[i][3] == 'e' || argv[i][3] == 'E'){
				if (gr_set) {
				    switch (grid_type) {
				    case 0:  grid_type = 13;  break;
				    case 1:  grid_type = 13;  break;
				    case 3:  grid_type = 23;  break;
				    case 12: case 21:
					     grid_type = 123; break;
				    default: grid_type = 3;   break;
				    }
				} else grid_type = 3; /* ecliptic   */
			    } else if(argv[i][3] == 'g' || argv[i][3] == 'G'){
				if (gr_set) {
				    switch (grid_type) {
				    case 1:  grid_type = 11;  break;
				    case 2:  grid_type = 21;  break;
				    case 23: case 32:
					     grid_type = 123; break;
				    default: grid_type = 1;   break;
				    }
				} else grid_type = 1; /* galactic   */
			    } else {
				switch(argv[++i][0]) {
				case 'g':  /* galactic */
					grid_type = 1;
					break;
				case 'e': /* ecliptic */
					if (argv[i][1] == 'q')
					    grid_type = 2; /* equatorial */
					else
					    grid_type = 3; /* ecliptic   */
					break;
				case 'q': /* equatorial */
					grid_type = 2;
					break;
				default :
					grid_type = atoi(argv[i]);
					break;
				}
			    }
			    gr_set = 1;
			} else if (strncmp(argv[i]+1, "gc", 2) == 0 ||
			     strncmp(argv[i]+1, "gc", 2) == 0) {
				grid_color = atoi (argv[++i]);
			} else
			    coordinates_flag = 1;
			break;
		case 'h':
			if (strncmp (argv[i]+1, "hm", 2) == 0) {
			    strcpy(hp_mask_file, argv[++i]);
			} else if (strncmp (argv[i]+1, "h2", 2) == 0) {
			    usage2();
			} else if (strncmp (argv[i]+1, "hsv", 2) == 0) {
			    hsv_color = 1;
			} else
			    usage();
			break;
		case 'i':
			switch(argv[i][2]) {
			case 'c': /* input color */
				inp_col = atoi(argv[++i]);
				bt_c = 1;
				break;
			case 'f': /* FITS-file map*/
				ffile_flag = 0;
				strcpy(filename, argv[++i]);
				break;
			case 'm': /* input marker */
			      {
				int i1 = '0', i2='9';
				inp_mark = argv[++i][0];
				if (i1 <= inp_mark && inp_mark <= i2)
					inp_mark = atoi(argv[i]);
				bt_m = 1;
			      }
			      break;
			case 'p': /* input polarization angles with file */
				strcpy(inp_pol_file, argv[++i]);
				break;
			case 's': /* input size of markers */
				inp_size = atoi(argv[++i]);
				bt_s = 1;
				break;
			case 'v': /* input pol. vectors with 2 files */
			      {
				char *cs;
				strcpy(inp_amp_pol_f, argv[++i]);
				cs = index(inp_amp_pol_f, ',');
				*cs = '\0';
				strcpy(inp_ang_pol_f,
				    index(argv[i], ',')+1);
			      }
			      break;
			default:
				strcpy(filename, argv[++i]);
				break;
			}
			break;
		case 'k':
			if (strncmp("ke", argv[i]+1, 2) == 0) {
			    char s1[30], s2[30], s3[30], s4[30];
			    get_4_str(argv[++i], s1, s2, s3, s4);
			    if(*s1 !='\0')
			      zone_latR0=deg2rad(s1);else zone_latR0= -1001.;
			    if(*s2 !='\0')
			      zone_lonR0=deg2rad(s2);else zone_lonR0= -1001.;
			    if(*s3 !='\0')
			      zone_latR1=deg2rad(s3);else zone_latR1= -1001.;
			    if(*s4 !='\0')
			      zone_lonR1=deg2rad(s4);else zone_lonR1= -1001.;
			    z_remain_flag = 1;
			    z_remove_flag = 0;
			    z_mask_flag = 1;
			} else {
			    PR(ER, "%s: illegal flag '%s'\n",
				 myname, argv[i]);
			    usage();
			}
			break;
		case 'K':  /* keep data in selected zones */
			z_remain_flag = 1;
			z_remove_flag = 0;
			break;
		case 'l':
			if (strncmp(argv[i]+1, "li", 2) == 0) {
			  /* set limits to be plotted */
			  char s[50], *cs=s, *ps=argv[i+1], *ss;
			  while (*ps) {
			    while ((*cs = *ps) != '/' && *cs) {cs++;ps++;}
			    *cs = '\0';
			    if (*ps == '/') ps++;
			    v_intr[nv_intr].low = *s!='\0' ? atof(s) : dummy;
			    if ( index(s,',') == NULL)
				ss = NULL;
			    else
				ss = index(s,',') + 1;
			    v_intr[nv_intr].high =
				(ss!=NULL && *ss != '\0') ? atof(ss) : dummy;
			    if (fabs(v_intr[nv_intr].low -dummy) > 1.e-10 ||
				fabs(v_intr[nv_intr].high-dummy) > 1.e-10 )
				    nv_intr++;
			    cs=s;
			    if (nv_intr == NV_INTR)
				break;
			  }
			} else if (strncmp(argv[i]+1, "lma", 3) == 0) {
			    lmax = atoi(argv[++i]);
			} else if (strncmp(argv[i]+1, "log", 3) == 0 ||
			    strncmp(argv[i]+1, "lg", 2) == 0) {
			    log_flag = 1;
			} else {
			    lon0 = deg2rad(argv[++i]);
			}
			break;
		case 'm':
			if (strncmp(argv[i]+1, "ma", 2) == 0) {
				max = atof(argv[++i]);
				if (!set_limits) set_limits = 1;
			} else if (strncmp(argv[i]+1, "mi", 2) == 0) {
				min = atof(argv[++i]);
				if (!set_limits) set_limits = 1;
			} else if (strncmp(argv[i]+1, "mr", 2) == 0) {
				strcpy(mark_str, argv[++i]);
				if (argv[i][3] == 'g')
					/* input in galactic ccordinates */
					coordinates_flag = 1;
				else if (argv[i][3] == 'q')
					/* input in equatorial ccordinates */
					coordinates_flag = 2;
			} else if (strncmp (argv[i]+1, "mh", 2) == 0) {
			    strcpy(hp_mask_file, argv[++i]);
			} else if (strncmp (argv[i]+1, "mf", 2) == 0) {
			    strcpy(source_file, argv[++i]);
			} else if (strncmp (argv[i]+1, "mg", 2) == 0) {
			    strcpy(gl_mask_file, argv[++i]);
			} else {
				PR(ER, "%s: illegal flag '%s'\n",
				    myname, argv[i]);
				usage();
			}
			break;
		case 'n':
			if (strncmp(argv[i]+1, "npm", 3) == 0) {
			    /* do not use pixel mask */
			    pixel_mask = 0;
			} else {
			    /* no default title */
			    no_title = 1;
			}
			break;
		case 'o': /* overlay */
			if (strncmp(argv[i]+1, "ov", 2) == 0) {
			    overlay_data = 1;  /* flag to overlay mask data */
			    strcpy(overlay_file, argv[++i]);
			} else if (strncmp(argv[i]+1, "oc", 2) == 0) {
			    overlay_color = atoi(argv[++i]);
			} else if (strncmp(argv[i]+1, "ow", 2) == 0) {
			    overlay_width = atoi(argv[++i]);
			} else {
			    void get_dir(char *);
			    char fdir[256];
			    strcpy(outp_file, argv[++i]);

			    /* copy name of file to a comment on a figure */
			    get_dir(fdir);
			    txt[0].x = -1; txt[0].y = -1;
			    sprintf(txt[0].text, "%s/%s", fdir, outp_file);
			    txt[0].color = 1;
			}
			break;
		case 'p':
			if (strcmp(argv[i]+1, "pf") == 0) {
			  strcpy(plotfile, argv[++i]);
			  vect_overlay = 1;
			} else if (strncmp(argv[i]+1, "pa", 2) == 0) {
			  patt_overlay = 1;
			} else if (strncmp(argv[i]+1, "ph", 2) == 0) {
			  /* plot phase diagram from input alm-file */
			  phase_diagram = 1;
			} else {
			  /* input polariz. vectors with 2 files */
			  char *cs;
			  strcpy(inp_amp_pol_f, argv[++i]);
			  cs = index(inp_amp_pol_f, ',');
			  *cs = '\0';
			  strcpy(inp_ang_pol_f,
			      index(argv[i], ',')+1);
			}
			break;
		case 'q':
		case 'Q':
			flag_map = 'q';
			break;
		case 'r':
			if (strncmp("rec", argv[i]+1, 3) == 0) {
			    projection = 'r';
			} else if (strcmp("rm", argv[i]+1) == 0 ||
			    strncmp("re", argv[i]+1, 2) == 0) {
			     char s1[30], s2[30], s3[30], s4[30];
			     get_4_str(argv[++i], s1, s2, s3, s4);
			     if(*s1 !='\0')
			       zone_latR0=deg2rad(s1);else zone_latR0= -1001.;
			     if(*s2 !='\0')
			       zone_lonR0=deg2rad(s2);else zone_lonR0= -1001.;
			     if(*s3 !='\0')
			       zone_latR1=deg2rad(s3);else zone_latR1= -1001.;
			     if(*s4 !='\0')
			       zone_lonR1=deg2rad(s4);else zone_lonR1= -1001.;
			     z_remove_flag = 1;
			     z_remain_flag = 0;
			     z_mask_flag = 1;
			} else if (strncmp ("ra", argv[i]+1, 2) == 0) {
			     zone_radius = deg2rad(argv[++i]);
			     zone_circle = 1;
			} else {
				PR(ER, "%s: illegal flag '%s'\n",
				    myname, argv[i]);
				usage();
			}
			break;
		case 'R':  /* keep data outside selected zones */
			z_remain_flag = 0;
			z_remove_flag = 1;
			break;
		case 's':
			switch(argv[i][2]) {
			case 'f':
			     if (strcmp(argv[i]+1, "sfo") == 0) {
				    font_size = atoi(argv[++i]);
			     } else {
				    size_factor = atof(argv[++i]);
			     }
			     break;
			case 'i':
			     sin_flag = 1;
			     break;
			case 'a':
			     if (strcmp(argv[i]+1, "sa")==0) {
				/* select a zone to be cut and plotted */
				/* in galactic coordinates */
				char s1[30], s2[30], s3[30], s4[30];
				get_4_str(argv[++i], s1, s2, s3, s4);
				if(*s1 !='\0') zcut_latR1=deg2rad(s1);
				    else zcut_latR1= -1001.;
				if(*s2 !='\0') zcut_lonR1=deg2rad(s2);
				    else zcut_lonR1= -1001.;
				if(*s3 !='\0') zcut_latR2=deg2rad(s3);
				    else zcut_latR2= -1001.;

				/* s1,s2,s3: lat, lon, size_deg */
				/* set area bounaries */
				/* for declination */
				if (zcut_lonR1 > -1001.)
				    zcut_lonR1 -= .5 * zcut_latR2;
				zcut_lonR2 = zcut_lonR1 + zcut_latR2;

				/* for right ascension */
				if (zcut_latR1 > -1001.)
				    zcut_latR1 -= .5 * zcut_latR2;
				zcut_latR2 = zcut_latR1 + zcut_latR2;

				projection = 's';
			     } else if (strcmp(argv[i]+1, "saq")==0) {
				/* select a zone to be cut and plotted */
				/* in equatorial coordinate limits     */
				char s1[30], s2[30], s3[30], s4[30];
				double cos_d;

				get_4_str(argv[++i], s1, s2, s3, s4);
				/* RA */
				if(*s1 !='\0') zcut_latR1=deg2rad(s1)*15.;
				    else zcut_latR1= -1001.;
				/* Dec */
				if(*s2 !='\0') zcut_lonR1=deg2rad(s2);
				    else zcut_lonR1= -1001.;
				/* size */
				if(*s3 !='\0') zcut_latR2=deg2rad(s3);
				    else zcut_latR2= -1001.;

				/* cosine for Dec=zcut_lonR1 */
				cos_d = cos (zcut_lonR1);

				/* s1,s2,s3: ra, deg, size_deg */
				/* set area bounaries */
				/* for declination */
				if (zcut_lonR1 > -1001.)
				    zcut_lonR1 -= .5 * zcut_latR2;
				zcut_lonR2 = zcut_lonR1 + zcut_latR2;

				/* for right ascension */
				if (zcut_latR1 > -1001.)
				    zcut_latR1 -= .5 * zcut_latR2/cos_d;
				zcut_latR2 = zcut_latR1 + zcut_latR2/cos_d;
				projection = 'Q';
			     }
			     break;
			case 'q':
			    /* select a zone to be masked */
			    if (strcmp(argv[i]+1, "sq")==0) {
				char s1[30], s2[30], s3[30], s4[30];
				get_4_str(argv[++i], s1, s2, s3, s4);
				if(*s1 !='\0') zone_latR0=deg2rad(s1);
				    else zone_latR0= -1001.;
				if(*s2 !='\0') zone_lonR0=deg2rad(s2);
				    else zone_lonR0= -1001.;
				if(*s3 !='\0') zone_latR1=deg2rad(s3);
				    else zone_latR1= -1001.;
				if(*s4 !='\0') zone_lonR1=deg2rad(s4);
				    else zone_lonR1= -1001.;
			    } else if (strncmp(argv[i]+1, "sqg", 3)==0) {
				/* select a zone to be cut and plotted */
				/* in galactic coordinates             */
				char s1[30], s2[30], s3[30], s4[30];
				get_4_str(argv[++i], s1, s2, s3, s4);
				if(*s1 !='\0') zcut_latR1=deg2rad(s1);
				    else zcut_latR1= -1001.;
				if(*s2 !='\0') zcut_lonR1=deg2rad(s2);
				    else zcut_lonR1= -1001.;
				if(*s3 !='\0') zcut_latR2=deg2rad(s3);
				    else zcut_latR2= -1001.;
				if(*s4 !='\0') zcut_lonR2=deg2rad(s4);
				    else zcut_lonR2= -1001.;
				projection = 's';
			    } else if (strncmp(argv[i]+1, "sqa", 3)==0
			       || strncmp(argv[i]+1, "sqq", 3)==0) {
				/* select a zone to be cut and plotted */
				/* in equatorial coordinate limits     */
				char s1[30], s2[30], s3[30], s4[30];
				get_4_str(argv[++i], s1, s2, s3, s4);
				if(*s1 !='\0') zcut_latR1=deg2rad(s1)*15.;
				    else zcut_latR1= -1001.;
				if(*s2 !='\0') zcut_lonR1=deg2rad(s2);
				    else zcut_lonR1= -1001.;
				if(*s3 !='\0') zcut_latR2=deg2rad(s3);
				    else zcut_latR2= -1001.;

				if (strncmp(argv[i]+1, "sqa", 3)==0) {
				    /* s1,s2,s3: ra, deg, size_deg */
				    /* set area bounaries */
				    /* for declination */
				    if (zcut_lonR1 > -1001.)
					zcut_lonR1 -= .5 * zcut_latR2;
				    zcut_lonR2 = zcut_lonR1 + zcut_latR2;

				    /* for right ascension */
				    if (zcut_latR1 > -1001.)
					zcut_latR1 -= .5 * zcut_latR2/15.;
				    zcut_latR2 = zcut_latR1 + zcut_latR2/15.;
				} else {
				    /* last coordinates for rectangular */
				    if(*s3 !='\0') zcut_latR2 *= 15.;
					else zcut_latR2= -1001.;
				    if(*s4 !='\0') zcut_lonR2=deg2rad(s4);
					else zcut_lonR2= -1001.;
				}
				projection = 'Q';
			     }
			     break;
			case 'u':
			     sun_data_key = 1;
			     break;
			default :
			     PR(ER, "%s: illegal flag '%s'\n",
				 myname, argv[i]);
			     usage();
			}
			break;
		case 't': /* text output */
			if (strcmp(argv[i]+1, "t") == 0 ) {
			   /* extract T-map */
			   flag_map = 't';
			} else if (strncmp(argv[i], "-te", 3) == 0 ||
			   strncmp(argv[i], "-tx", 3) == 0) {
			   extern void get_4_pars(char *, int *, int *,
				char *, int *);
			   i++;
			   if ( txt_n < TXT_N ) {
				get_4_pars(argv[i],
				    &txt[txt_n].x, &txt[txt_n].y,
				     txt[txt_n].text, &txt[txt_n].color);
				txt[txt_n].color = 1;
				txt_n++;
			   }
			} else {
			   PR(ER, "%s: illegal flag '%s'\n",
			       myname, argv[i]);
			   usage();
			}
			break;
		case 'T': /* text output */
			txt[txt_n].x = -1;  /* put in the middle */
			txt[txt_n].y = -1;  /* to resize */
			strcpy(txt[txt_n].text, argv[++i]);
			txt[txt_n].color = 1;
			txt_n++;
			tf_flag++;
			break;
		case 'u':
		case 'U':
			flag_map = 'u';
			break;
		case 'v':
			PR(stdout,
			    "%s: version 1.9 (16/12/2015), GLESP 2.0.\n",
			    myname);
			exit(0);
		case 'x':
			xs = atoi(argv[++i]);
			x_set = 1;
			break;
		case 'y':
			ys = atoi(argv[++i]);
			y_set = 1;
			break;
		case 'z':
		    switch (argv[i][2]) {
		    case 'c':
			{  char s1[30], s2[30];
			  /* center of a circle to be masked */
			  get_2_str(argv[++i], s1, s2);
			  if(*s1 !='\0')
			    zone_latR0=deg2rad(s1); else zone_latR0= -1001.;
			  if(*s2 !='\0')
			    zone_lonR0=deg2rad(s2); else zone_lonR0= -1001.;
			  zone_circle = 1;
			}
			z_mask_flag = 1;
			break;
		    case 'r':
			/* radius of a circle to be masked */
			zone_radius = deg2rad(argv[++i]);
			zone_circle = 1;
			b_size = 2. * zone_radius * 180./M_PI;
			z_mask_flag = 1;
			break;
		    case 'o':
			/* select a zone to be masked */
			{ char s1[30], s2[30], s3[30], s4[30];
			    get_4_str(argv[++i], s1, s2, s3, s4);
			    if(*s1 !='\0') zone_latR0=deg2rad(s1);
				else zone_latR0= -1001.;
			    if(*s2 !='\0') zone_lonR0=deg2rad(s2);
				else zone_lonR0= -1001.;
			    if(*s3 !='\0') zone_latR1=deg2rad(s3);
				else zone_latR1= -1001.;
			    if(*s4 !='\0') zone_lonR1=deg2rad(s4);
				else zone_lonR1= -1001.;
			}
			z_mask_flag = 1;
			break;
		    case 'd':
			/* if box_size selected for extr. areas from list */
			b_size = str2deg(argv[++i]);
			break;
		    default:
			/* file with list of zones to be processed */
			strcpy(zone_file, argv[++i]);
			z_mask_flag = 1;
			break;
		    }
		    break;
		default:
			break;
		}
		break;
	case '+':
		switch(argv[i][1]) {
		case 'h':
			usage();
			break;
		case 'o': /* overlay */
			if (strncmp(argv[i]+1, "ov", 2) == 0) {
			    overlay_data = 1;  /* flag to overlay mask data */
			    strcpy(overlay_file, argv[++i]);
			    overlay_flag = 1;
			} else {
			    PR(ER,"%s: illegal flag '%s'\n",myname,argv[i]);
			    usage();
			}
			break;
		case 'T': /* text output */
			txt[txt_n].x = -1;
			txt[txt_n].y = -1;
			strcpy(txt[txt_n].text, argv[++i]);
			txt[txt_n].color = 1;
			txt_n++;
			tf_flag++;
			break;
		case '\0':
			strcpy(filename, "-");
			break;
		default:
			usage();
			break;
		}
		break;
	default:
		strcpy(filename, argv[i]);
		break;
	}
    }

    /* check for title */
    for (i=0; i<txt_n; i++) {
	 if (no_title > 0) {
	    if (txt[i].x < 0 && txt[i].y < 0) {
		int j;
		txt_n--;
		for (j=i; j<txt_n; j++) {
			txt[j].x = txt[j+1].x;
			txt[j].y = txt[j+1].y;
			strcpy(txt[j].text, txt[j+1].text);
			txt[j].color = txt[j+1].color;
		}
		i--;
	    }
	 } else if (tf_flag) {
	    /* if another title instead of output filename */
	    /* remove the first one later;                 */
	    if (i==0) {
		int j;
		txt_n--;
		for (j=0; j<txt_n; j++) {
			txt[j].x = txt[j+1].x;
			txt[j].y = txt[j+1].y;
			strcpy(txt[j].text, txt[j+1].text);
			txt[j].color = txt[j+1].color;
		}
	    }
	 }
    }

    if (projection == 's' || projection == 'Q') {
	double dlat = fabs(zcut_latR2 - zcut_latR1),    /* or RA  */
	       dlon = fabs(zcut_lonR2 - zcut_lonR1);    /* or Dec */
	if (projection == 'Q') {
	       double cos_d = cos ((zcut_lonR1+zcut_lonR2)/2);
	       dlat *= cos_d;
	       if (cos_d < 0.) {
		  PR(ER, "%s: cos(delta) < 0... Wrong input coordinates ?\n",
		    myname);
	       }
	}

	/* draw part of sky */
	if (!y_set) {
	    if (!x_set) xs/=2;
	    ys = xs * dlat/dlon;
	} else if (!x_set && y_set) {
	    xs = ys * dlon/dlat;
	}
    }
    if (phase_diagram || ampl_diagram) {
	if ( !y_set && !x_set ) { x_set = 1; xs = 510; }
	if (!y_set) {
	    if (!x_set) xs/=2;
	    ys = xs;
	} else if (!x_set && y_set) {
	    xs = ys;
	}
    }

    return;
}

void usage()
{
PR(ER, "usage: %s [[-i] input_map_file] [-o output_gif_file] [-c colormap]\n",
    myname);
PR(ER, "[-rec(rectang.proj.)] [-patt(pattens overlay)] [-0] [-1]\n");
PR(ER, "[-xs xsize] [-ys ysize] [-C[s] min,max] [-{fs,ff} source_file]\n");
PR(ER, "[-mr mark_str] [-cr(galactic coordinates for marks)] [-feq(eq.coor.)]\n");
PR(ER, "[-is mark_size] [-im marker] [-ic marker_color] [-sun(sun format)]\n");
PR(ER, "[-lon lon] [-ip inp_pol_file] [-cats cats_source_file] [-sfo font_size]\n");
PR(ER, "[-pv amp_file,amg_file] [-sf vector_size_factor] [-d dummy] [-{t,q,u}]\n");
PR(ER, "[-text x,y,text] [-T text] [-notitle] [-gr {0g,1e,2q}] [-gc col]\n");
PR(ER, "[-zf file_with_zones] [-zo lat1,lon1,lat2,lon2(mask sky zone to remove)]\n");
PR(ER, "[-zc lat,lon] [-zr radius_in_degs] [-/+ov overlay_file] [-mh hpix_mask_file]\n");
PR(ER, "[-mg glesp_mask_file] [-ph(phase digarm from alm) [-sin] [-hsv] -lmax l]\n");
PR(ER, "[-{remove,keep} lat1,lon1,lat2,lon2(mask sky zone to remove/keep)]\n");
PR(ER, "[-li  a1,a2/b1,b2/c1,c2/... (set lim-s to)] [-cut lat1,lon1,lat2,lon2]\n");
PR(ST, "[-cc lat,lon,rad,w(ring mask in degs)]\n");
PR(ST, "[-ccr lat1,lon1,lat2,lon2,w(rectang.mask in degs)] [-log] [-pf plotfile]\n");
PR(ER, "[-fzq[1] list_source_equat.][-fzq[1] list_src_gal.] [-zd box_size_d]\n");
PR(ER, "[-dir output_directory[\n");
PR(ER, "[-v(version)]\n");
exit(2);
}

void usage2()
{
PR(ST, "usage: %s input_map_file -o output_gif_file  FLAGS\n", myname);
PR(ST, "----------------------------------------------------\n");
PR(ST, "-  - creates empty map and follows other instructions (see also '-d')\n");
PR(ST, "-0 - creates empty map and follows other instructions\n");
PR(ST, "-1 - creates unity map and follows other instructions\n");
PR(ST, "-amp - plots amplitude diagram from input alm-file\n");
PR(ST, "'-'          - sets empty (zero) file input\n");
PR(ST, "-c colormap  - sets colormap (0,1,2,3,4,5)\n");
PR(ST, "-cats cats_source_file - read radio source data from CATS output file\n");
PR(ST, "-cc   lat,lon,rad,w - set ring mask for sky zone to be kept:\n");
PR(ST, "      in center (lat,lon) of radius 'rad' and width 'w' in degs.\n");
PR(ST, "-ccr  lat1,lon1,lat2,lon2,w - set rectang. mask for sky zone to be kept\n");
PR(ST, "      in rectangle (lat1,lon1,lat2,lon2) of width 'w' in degs.\n");
PR(ST, "-C min,max   - cut-levels: set min & max, but not out of picture range\n");
PR(ST, "-Cs min,max  - cut-levels: set min & max by force\n");
PR(ST, "-Cs nat (and flag '-cut') - set real limits in selected area\n");
PR(ST, "-cr   - sets galactic coordinates for markers (see '-fs', '-mr')\n");
PR(ST, "-cut lat1,lon1,lat2,lon2 - selects zone by coordinates and draws this area\n");
PR(ST, "-cq  ra1,dec1,ra2,dec2 - selects zone by equat.coor-s and draws the area\n");
PR(ST, "-cqa ra,deg,size_deg - selects zone by equat.coor-s and draws the area\n");
PR(ST, "-d dummy - sets default value for pixels of a selected zone\n");
PR(ST, "-d1 d1_step  - sets step 'd1_step' for grid by latitude\n");
PR(ST, "-d2 d2_step  - sets step 'd2_step' for grid by latitude\n");
PR(ST, "-dir output_directory\n");
PR(ST, "-feq  - sets equatorial coordinates for markers (see '-fs', '-mr')\n");
PR(ST, "-ff source_file - reads sources from F-file (see FADPS)\n");
PR(ST, "-fs source_file - reads sources from marker file, format like for '-mr'\n");
PR(ST, "-fzg file - gets areas in GIF-files from a source list in gal.coors\n");
PR(ST, "            file format:  lat_DDD.DDDd lon_DDD.DDDd [DD.DDd]\n");
PR(ST, "-fzg1 file - gets areas in GIF-files from a source list in gal.coors\n");
PR(ST, "            file format:  NAME lat_DDD.DDDd lon_DDD.DDDd [DD.DDd]\n");
PR(ST, "-fzq file - gets areas in GIF-files from a source list in eq.coors [see '-fzq','-zd']\n");
PR(ST, "            file format:  hh:mm:ss.ss ddd:mm:ss.ss [DD.DDd]\n");
PR(ST, "-fzq1 file - gets areas in FITS-files from a source list in eq.coors [see '-fzq','-zd']\n");
PR(ST, "            file format:  GIF hh:mm:ss.ss ddd:mm:ss.ss [DD.DDd]\n");
PR(ST, "-gc col   - sets color of grid\n");
PR(ST, "-gr grid  - sets grid: {g,e,q} or {1,2,3} for galactic, ecliptic or equatorial\n");
PR(ST, "            system\n");
PR(ST, "-ic marker_color - sets marker color (0,1,2,3,...,207)\n");
PR(ST, "-im marker       - sets type of marker (1,2,3,4,5... or x,+,o,O,...)\n");
PR(ST, "-ip inp_pol_file -\n");
PR(ST, "-is mark_size    - sets marker size = mark_size\n");
PR(ST, "-keep   lat1,lon1,lat2,lon2 - sets mask for sky zone to be kept\n");
PR(ST, "-K - flag to keep selected zone on the figure\n");
PR(ST, "-li  a1,a2/b1,b2/c1,c2/... - keep pixels with with values from given intervals\n");
PR(ST, "-lmax lmax - upper l to plot phases (see flag '-ph')\n");
PR(ST, "-log - calculates logarithm from absolute values\n");
PR(ST, "-lon lon - sets a picture center point by longitude 'lon'\n");
PR(ST, "-mf source_file - reads sources from marker file, format like for '-mr'\n");
PR(ST, "-mg glesp_mask_file - sets mask from GLESP mask file\n");
PR(ST, "-mh hpix_mask_file - sets mask from HEALPix mask file\n");
PR(ST, "-mr mark_str   - marker's string: coord1,coord2,SHAPE,SIZE,COLOR\n");
PR(ST, "-mrg mark_str  - marker's string: same as '-mr mark_str' and '-cr'\n");
PR(ST, "-mrq mark_str  - marker's string: same as '-mr mark_str' and '-feq'\n");
PR(ST, "-notitle - (or '-n') do not output file name (at bottom)\n");
PR(ST, "-npm - do not use pixel mask from plotted map\n");
PR(ST, "-/+ov overlay_file - overalys mask data from file\n");
PR(ST, "-pa - overlays celestial patterns\n");
PR(ST, "-pf plotfile - to prepare picture with plot-vectors from 'plotfile'\n");
PR(ST, "-ph - plots phase diagram from input alm-file (see '-lmax')\n");
PR(ST, "-pv amp_file,amg_file\n");
PR(ST, "-q  - to prepare picture for Q-polarization map (see '-t', 'u')\n");
PR(ST, "-rec         - set rectangular projection (defaultrect: Molweide)\n");
PR(ST, "-remove lat1,lon1,lat2,lon2 - set mask for sky zone to be removed\n");
PR(ST, "-R - flag to remove selected zone on the figure\n");
PR(ST, "-sf vector_size_factor\n");
PR(ST, "-sfo font_size\n");
PR(ST, "-sin - to use values of sin(phase) for flag '-ph'\n");
PR(ST, "-sa lat,lon,size_deg    - selects zone by gal.coor-s and draws the area\n");
PR(ST, "-sqg lat1,lon1,lat2,lon2- selects zone by gal.coor-s and draws the area\n");
PR(ST, "-sqq ra1,dec1,ra2,dec2  - selects zone by equat.coor-s and draws the area\n");
PR(ST, "-saq ra,deg,size_deg    - selects zone by equat.coor-s and draws the area\n");
PR(ST, "-t  - to prepare picture for temperature map (see '-q', 'u')\n");
PR(ST, "-text x,y,text - outputs text 'text' in coordinates (x,y)\n");
PR(ST, "-T text  - outputs text 'text' as a picture title at a top\n");
PR(ST, "-u  - prepares picture for U-polarization map (see '-t', 'q')\n");
PR(ST, "-v - prints version of '%s'\n", myname);
PR(ST, "-xs xsize    - size of picture: xsize pixels by X\n");
PR(ST, "-ys ysize    - size of picture: ysize pixels by Y\n");
PR(ST, "-zc lat,lon - selects zone center (see '-K', '-R')\n");
PR(ST, "-zf file_with_zones - reads zones from file for masking\n");
PR(ST, "-zo lat1,lon1,lat2,lon2 - selects zone by coordinates: default to remove\n");
PR(ST, "                          (see '-K','-R', '-remove', '-keep')\n");
PR(ST, "-zr radius_in_degs - selects zone radius (see '-K', '-R')\n");
exit(0);
}

void get_4_pars(str, x, y, text, col)
char *str, *text;
int  *x, *y, *col;
{
	char s1[50], s2[50], s3[50], s4[50];
	register char *cs, *ps;

	s1[0] = s2[0] = s3[0] = s4[0] = '\0';
	ps = str;
	cs = s1;
	while ( ((*cs = *ps) != ',' && *cs != ';' ) && *cs ) {
		cs++; ps++;
	}
	*cs = '\0';
	if ( *ps ) {
		 ps++;
		 cs = s2;
		 while ( ((*cs = *ps) != ',' && *cs != ';' ) && *cs ) {
			 cs++; ps++;
		 }
		 *cs = '\0';
	}
	if ( *ps ) {
		 ps++;
		 cs = s3;
		 while ( ((*cs = *ps) != ',' && *cs != ';' ) && *cs ) {
			 cs++; ps++;
		 }
		 *cs = '\0';
	}
	if ( *ps ) {
		 ps++;
		 cs = s4;
		 while ( ((*cs = *ps) != ',' && *cs != ';' ) && *cs ) {
			 cs++; ps++;
		 }
		 *cs = '\0';
	}
	if (s1[0] != '\0' )  *x = atoi(s1);  else *x = 0;
	if (s2[0] != '\0' )  *y = atoi(s2);  else *y = 0;
	if (s3[0] != '\0' )  strcpy(text, s3);
	if (s4[0] != '\0' )  *col = atoi(s4);  else *col = 0;
	return;
}

/* read current directory */
#include <unistd.h>
void get_dir(fdir)
char *fdir;
{
    getcwd(fdir, 255);
    return;
}
