#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "glesp.h"

#define  PR     fprintf
#define  ER     stderr
/*
 * convert F-file Binary Table map to F-file uniform grid map or
 * to FITS Binary Table in HEALPix format
 *
 * Oleg Verkhodanov (vo@sao.ru, oleg@tac.dk),
 *         Theoretics Astrophysics Center
 * August, 2002
 * Last changes, VO:09/12/2004
 *
 */

extern void setopt(int, char **), usage(void);
extern void set_sig_fatal(void);
extern int sun_data_key;
extern int F_sun_data_key;

char inp_file[81], outp_file[81];    /* input F-file */
char q_inp_file[81], u_inp_file[81];
char tmpcat[161];

int debug_flag = 0;   /* debug flag                                 */
int hf_flag = 0;      /* if flag > 0, then output in HEALPix format */
int flag_int=0; /* make input of integer values (in some simulations) */
double nside_factor=1;   /* factor to decrease nside   */
int gp_flag = 0;
int reflection = 1;
int interpolation=1;    /* repixelization  by interpolation */
int ffile_flag = 0;     /* output to FITS file            */
int outp_type = 1;
int nside = -1;
int shift_flag = 0;


int pol_map = 0;        /* to write polarization map */
char *myname="f2map";
HEAD hd, *ph;

int main(argc, argv)
int argc;
char **argv;
{
    double *f = (double *)0, *x = (double *)0, *phi = (double *)0;
    double *q = (double *)0, *u = (double *)0;
    int *nph = (int *)0;
    int nx=0, nphi=0;
    register FILE *fp_out = stdout;
    double *f_new = (double *)0, *q_new = (double *)0, *u_new = (double *)0;
    register int i, j;

    /* processing fatal errors */
    set_sig_fatal();

    inp_file[0] = q_inp_file[0] = u_inp_file[0] = '\0';
    outp_file[0] = '\0';
    strcpy(tmpcat, ".");

    /* loop over external parameters */
    setopt(argc, argv);

    /* open output file */
    if (*outp_file=='\0' || (fp_out =fopen(outp_file,"w"))==(FILE *)0) {
	    PR(ER,"%s: can not open output file '%s'. I' use stdout.\n",
		    myname, outp_file);
	    fp_out = stdout;
    }

    /***********************/
    /* read map: f(x, phi) */
    /***********************/
    /* detect file type, read map and check size of arrays     */
    /* if arrays are short, then free old and catch new memory */
    /* It returns type of grid in input file                   */
    /* i = read_map2(&f, &nx, &nphi, &nph, &x, inp_file); */

    i = test_glesp_map(inp_file);

    if (i !=1 && i != 3) {
	PR(ER, "%s: input file '%s' does not contain the GLESP map\n",
	    myname, inp_file);
	exit(1);
    }

    if (i == 3)  {
	 in_glesp_3map(&f, &q, &u, &x, &nx, &nphi, &nph, inp_file);
	 pol_map = 1;
    } else {
	read_glesp_map(&f, &x, &nx, &nphi, &nph, inp_file);
	if (pol_map) {
	    double *x1   = (double *)0;
	    int    *nph1 = (int *)0;
	    int nx1 = 0, nphi1 = 0;
	    read_glesp_map(&q, &x1, &nx1, &nphi1, &nph1, q_inp_file);
	    if ( nx1 != nx || nphi1 != nphi) {
dif_resolution:
		PR(ER, "%s: input maps have different resolution...\n",
		    myname);
		PR(ER, "(nx,nphi)=(%d,%d),  (nx,nphi)_pol=(%d,%d)\n",
		    nx, nphi, nx1, nphi1);
		exit(1);
	    }
	    free((void *)x1);      x1   = (double *)0;
	    free((void *)nph1);    nph1 = (int *)0;
	    nx1 = 0, nphi1 = 0;
	    read_glesp_map(&u, &x1, &nx1, &nphi1, &nph1, u_inp_file);
	    if ( nx1 != nx || nphi1 != nphi)
		    goto dif_resolution;
	    free((void *)x1);
	    free((void *)nph1);
	}
    }
    if (shift_flag) {
	/* shift for 180 deg */
	for(i=0; i<nx; i++) {
	    int ii = i * nphi;
	    int np2 = (nph[i]+1)/2;
	    int n = np2-1;
	    for(j=0; j<np2; j++) {
		double v = f[ii+n+j];
		f[ii+n+j] = f[ii+j];
		f[ii+j] = v;
	    }
	}
    }


    /* memory for temporal arrays */
    f_new = (double *)calloc((unsigned)(nphi*nx), sizeof *f_new);
    if (f_new == (double *)0) {
	    PR(ER,"%s: no memory for output arrays.\n", myname);
	    exit(1);
    }
    if (pol_map) {
	q_new = (double *)calloc((unsigned)(nphi*nx), sizeof *q_new);
	if (q_new == (double *)0) {
		PR(ER,"%s: no memory for output arrays.\n", myname);
		exit(1);
	}
	u_new = (double *)calloc((unsigned)(nphi*nx), sizeof *u_new);
	if (u_new == (double *)0) {
		PR(ER,"%s: no memory for output arrays.\n", myname);
		exit(1);
	}
    }


    /* transposing here: f(nx, nphi) -> f_new(nphi, nx) */
    /* to be used in future calculations below          */
    for(j=0; j<nphi; j++) {
	    register int k = j * nx;
	    for(i=0; i<nx; i++) {
		   f_new[k+i] = f[i*nphi+j];
		   if (pol_map) {
			q_new[k+i] = q[i*nphi+j];
			u_new[k+i] = u[i*nphi+j];
		   }
	    }
    }
    /* change pointers for my perconal convenience (VO) */
    { double *t = f; f = f_new; f_new = t; }
    if (pol_map) {
	double *t = q; q = q_new; q_new = t;
		t = u; u = u_new; u_new = t;
    }
    if (!interpolation) {
	if (reflection==0) reflection=1;
	else if (reflection==1) reflection=0;
	else if (reflection==2) reflection=3;
	else if (reflection==3) reflection=2;
    }

    /* symmetric reflection */
    if( reflection ) {
	/* meridian reflection */
	if (reflection % 2 == 1) {  /* flag == 1 || == 3 */
	    for(i=0; i<nx; i++) {
		int np  = nph[i] - 1;
		int np2 = nph[i] / 2;
		for(j=0; j<np2; j++) {
		    int jj = j * nx;
		    int jn = (np-j) * nx;
		    double v = f[jj+i];
		    f[jj+i] = f[jn+i];
		    f[jn+i] = v;
		    if (pol_map) {
			v = q[jj+i]; q[jj+i] = q[jn+i]; q[jn+i] = v;
			v = u[jj+i]; u[jj+i] = u[jn+i]; u[jn+i] = v;
		    }
		}
	    }
	}
	/* equator reflection */
	if (reflection > 1)      {  /* flag == 2 || == 3 */
	    int nx2 = nx / 2;
	    for(j=0; j<nphi; j++) {
		register int jj = j * nx;
		for(i=0; i<nx2; i++) {
		    int ii = nx-1-i;
		    double v = f[jj+i];
		    f[jj+i]  = f[jj+ii];
		    f[jj+ii] = v;
		    if (pol_map) {
			v = q[jj+i]; q[jj+i]  = q[jj+ii]; q[jj+ii] = v;
			v = u[jj+i]; u[jj+i]  = u[jj+ii]; u[jj+ii] = v;
		    }
		}
	    }
	}
    }
    if (gp_flag) {  /* print  values of of map */
	for(i=0; i<nx; i++) {
	    double t = acos(x[i]);
	    double dp = 2 * M_PI / nph[i];
	    for(j=0; j<nph[i]; j++) {
		double p = j * dp;
		PR(fp_out, "%8.6f %8.6f %16.9e\n", t, p, f[j*nx+i]);
	    }
	}
	/* close files */
	if (fp_out != stdout && fp_out != (FILE *)0)   fclose(fp_out);

	/* free memory */
	if (f != (double *)0)     free((void *)f);
	if (pol_map) {
	    if (q != (double *)0)     free((void *)q);
	    if (u != (double *)0)     free((void *)u);
	}
	if (x != (double *)0)     free((void *)x);
	if (phi != (double *)0)   free((void *)phi);

	exit(0);
    }

    /*******************/
    /* re-pixelization */
    /*******************/
    if (interpolation) {
	if (pol_map) {
	   PR(ER,"%s: interpolation for polarization is not produced here.\n",
	       myname);
	   exit(1);
	}

	/*---------------*/
	/* interpolation */
	/*---------------*/
	switch (outp_type) {
	case '0': /* rectangular image */
	case  0 :
	   {
	    double x0= 1.0, dx=0.0;
	    int condition_flag = 1; /* for spline x0<x1<x2... */
	    int nmx = nx;  /*  f(nphi,nx): f(i*nmx+j), i=1,nphi */
	    double p0 = 0.0, phi2 = M_PI * 2;

	    /* if special grid, we do recalculation */
	    if (theta_grid == 1) {
		/* recalculate from cos(theta)-grid to theta-grid */
		x0 = acos(x[0]);
		dx = (acos(x[nx-1]) - x0) / (double)(nx-1);
		for(i=0; i<nx; i++)
			x[i] = acos(x[i]);
	    }
	    if (x[0] > x[1]) {
		condition_flag = 0;
		for(i=0; i<nx; i++)  x[i] = -x[i];
		x0 = -x0; dx = -dx;
	    }

	    /*-------------------------------------------*/
	    /* spline interpolation (modified input):    */
	    /*   non-uniform step by X ---> uniform step */
	    /*   and non-uniform step by Y               */
	    /*-------------------------------------------*/
	    /*
	     *
	     *       nx = nmx                          nx = nmx
	     *
	     *        ******                           ********
	     *        ******                           ********
	     * nph(i) ******           --------->      ********
	     *        *****                            ********  nphi
	     *        *****                            ********
	     * < nphi  ***                             ********
	     *          *                              ********
	     *
	     *         f()                               f_new()
	     *
	     */
	    intrpl2sm_Nnu_u(f, nx, nmx, f_new, nx, nmx, x, x0, dx,
		p0, phi2, nph, nphi);

	    if (condition_flag == 0) {
		condition_flag = 1;
		/* for(i=0; i<nx; i++)  x[i] = -x[i]; */
		x0 = -x0; dx = -dx;
	    }
	    /* set new x-array */
	    for(i=0; i<nx; i++)  {
		   x[i] = x0 + i * dx;
		   nph[i] = nphi;
	    }

	    theta_grid = 2;

	    /* transposing here:            */
	    /* f_new(nphi,nx) -> f(nx,nphi) */
	    for(j=0; j<nphi; j++) {
		    register int k = j * nmx;
		    for(i=0; i<nx; i++) {
			    f[i*nphi+j] = f_new[k+i];
		    }
	    }
	    {
		int flag_grid = -1;  /* homogeneous grid */
		int transposing = 0; /* no transposing   */
		int flag_phi_set = 0;

		/* set 'phi' array */
		if (phi == (double *)0) {
		    double dphi = phi2 / nphi;
		    flag_phi_set = 1;
		    /* take memory for phi */
		    phi = (double *)calloc(nphi, sizeof *phi);
		    if (phi == (double *)0) {
			PR(ER,"%s: no memory for output arrays.\n", myname);
			exit(1);
		    }
		    for(i=0; i<nphi; i++)  phi[i] = p0 + i * dphi;
		}
		/* output in basic F-format             */
		/* first nmx='nphi' for calculation of  */
		/* position f[i*nphi+j]                 */
		if (ffile_flag) {
			/* write ff-map */
			ff_write_map(f, nphi, x, nx, phi, nphi, flag_grid,
			       transposing, fp_out);
		} else {
			/* write FITS-map */
			fits_write_map(f, nphi, x, nx, phi, nphi, flag_grid,
			       transposing, fp_out);
		}

		if (flag_phi_set) { free((void *)phi); phi = (double *)0; }
	    }
	   }
	   break;
	case '1': /* HEALPix output */
	case  1:
	   {
	    /* to transpose matrix for output, what will be below, */
	    /* just copy from one to another  */
	    register int k = nx * nphi;
	    for(i=0; i<k; i++) f_new[i] = f[i];

	    /* transpose matrix for output  */
	    /* i.e. make it as in 'cl2map'  */
	    /* f_new(nphi,nx) -> f(nx,nphi) */

	    /* if output will be in HEALPix style     */
	    /* we have to keep f_new(nphi,nx) because */
	    /* of interpolation by theta (i.e. x)     */
	    /* and no transposing, just coping        */

	    /* consider only case when data were      */
	    /* interpolated                           */
	    /* and no transposing here:               */
	    /* f_new(nphi, nx) -> f(nphi, nx)         */

	    theta_grid = 2;

	   }
	   /* output data in HEALPix format */
	   /* input array: f(nx, nphi)      */
	   fhp_out2(&f, nx, nph, nphi, x, phi, -1, fp_out, flag_int);
	   break;
	} /* end switch */
    } else {
	/*---------------------------------------------------*/
	/* re-pixelization by simple coping values of pixels */
	/*---------------------------------------------------*/

	switch (outp_type) {
	case  0 :
	case '0': /* rectangular image */
	     /* transpose matrix:                   */
	    for(j=0; j<nphi; j++) {
		    register int k = j * nx;
		    for(i=0; i<nx; i++) {
			f_new[i*nphi+j] = f[k+i];
			if (pol_map) {
			    q_new[i*nphi+j] = q[k+i];
			    u_new[i*nphi+j] = u[k+i];
			}
		    }
	    }
	    if (ffile_flag) {
		    ff_write_map2d(f_new, x, nx, nphi, nph, fp_out);
	    } else {
		    /* write ff-map */
		    glesp_2_map2d(f_new, x, nx, nphi, nph, fp_out);
	    }
	    break;
	case  1 :
	case '1': /* HEALPix output */
	    if (pol_map)
		fhp_out2p_pol(f, q, u, x, nx, nph, nphi, nside, fp_out);
	    else
		fhp_out2p(f, x, nx, nph, nphi, nside, fp_out);
	    break;
	}
    }
    /* free memory */
    if (f_new != (double *)0) free((void *)f_new);
    if (pol_map) {
	if (q_new != (double *)0) free((void *)q_new);
	if (u_new != (double *)0) free((void *)u_new);
    }
    /* close files */
    if (fp_out != stdout && fp_out != (FILE *)0)   fclose(fp_out);

    /* free memory */
    if (f != (double *)0)     free((void *)f);
    if (pol_map) {
	if (q != (double *)0)     free((void *)q);
	if (u != (double *)0)     free((void *)u);
    }
    if (x != (double *)0)     free((void *)x);
    if (phi != (double *)0)   free((void *)phi);

    exit(0);
}

void setopt(argc, argv)
int argc;
char **argv;
{
	register int i;
	myname = argv[0];

	for(i=1; i<argc; i++) {
	    switch(argv[i][0]) {
	    case '-':
		switch(argv[i][1]) {
		case 'b':
			outp_type = 0;
			break;
		case 'd':
			debug_flag = 1;
			break;
		case 'g':
			gp_flag = 1;
			break;
		case 'i':
			if (strcmp(argv[i]+1, "int") == 0)
				flag_int = 1;
			else
				strcpy(inp_file, argv[++i]);
			break;
		case 'n':
			nside_factor = atof(argv[++i]);
			break;
		case 'h':
			switch(argv[i][2]) {
			case 'p':
				hf_flag = 1;
				outp_type = 1;
				break;
			default :
				PR(ER, "%s: illegal flag '%s'.\n",
				      myname, argv[i]);
				usage();
				break;
			}
			break;
		case 'o':
			if(strncmp("of", argv[i]+1, 2) == 0) {
				ffile_flag = 1;
				outp_type = 0;
			}
			else if(strncmp("o2", argv[i]+1, 2) == 0) {
				outp_type = 0;
			}
			strcpy(outp_file, argv[++i]);
			break;
		case 'q':
			strcpy(q_inp_file, argv[++i]);
			pol_map = 1;
			break;
		case 'r':
			switch(argv[i][2]) {
			case 'p':
			     if (strcmp(argv[i]+1, "rp") == 0)
				  interpolation = 0;
			     break;
			case 's':
			     shift_flag = 1;
			     break;
			case '0':
			     reflection = 0;
			     break;
			case '\0':
			case '1':
			     if (strcmp(argv[i]+1, "r12") == 0)
				    reflection = 3;
			     else {
				  if (reflection == 2)
					 reflection = 3;
				  else
					 reflection = 1;
			     }
			     break;
			case '2':
			     if (strcmp(argv[i]+1, "r21") == 0)
				    reflection = 3;
			     else {
				  if (reflection == 1)
					 reflection = 3;
				  else
					 reflection = 2;
			     }
			     break;
			default :
			     PR(ER,"%s: illegal flag '%s'\n",myname,argv[i]);
			     usage();
			     break;
			}
			break;
		case 's':
			switch(argv[i][2]) {
			case 'u':
			     sun_data_key = 1;
			     break;
			case 'f':
			     F_sun_data_key = !F_sun_data_key;
			     break;
			default :
			     PR(ER, "%s: illegal flag '%s'\n",
				 myname, argv[i]);
			     usage();
			}
			break;
		case 't':
			if (strcmp(argv[i]+1, "tmp") == 0) {
			     strcpy(tmpcat, argv[++i]);
			} else {
			     PR(ER,
				"%s: illegal flag '%s'.\n", myname, argv[i]);
			     usage();
			}
			break;
		case 'u':
			strcpy(u_inp_file, argv[++i]);
			pol_map = 1;
			break;
		default :
			PR(ER, "%s: illegal flag '%s'.\n", myname, argv[i]);
			usage();
			break;
		}
		break;
	    case '+':
		switch(argv[i][1]) {
		case 'h':
			usage();
		default :
			PR(ER, "%s: illegal flag '%s'.\n", myname, argv[i]);
			usage();
			break;
		}
		break;
	    default :
		strcpy(inp_file, argv[i]);
		break;
	    }
	}

	if ( inp_file[0] == '\0') {
		PR(ER, "%s: no input file ...\n", myname);
		usage();
	}

	return;
}

void usage()
{
PR(ER,"usage: %s [[-i] inp_file] [-o[{2f}] outp_file] [-bf(Basic FITS)] [-hp(HEALPix)]\n",
    myname);
PR(ER,"[-gp(print values)] [-tmp tmpcat] [-n nside_decrease_factor(2^n)] [-int]\n");
PR(ER,"[-r{012}(reflection)] [-rs(shift 180d longit.] [-rp(just repixel.)]\n");
PR(ER,"[-u u_glesp.fts] [-q q_glesp.fts] [-sun] [-sf]\n");
exit(2);
}
