# include <stdio.h>
# include <math.h>
# include <ff.h>

# include "glesp.h"

extern char *myname;
extern int flag_take_memory;  /* flag to take memory */

#ifdef _OSF_SOURCE
long long atoll(str)
char *str;
{
     return ((long long)atol(str));
}
#endif

void ff_read_ext2_map(f, x, nx, nph, np, filename)
double **f;      /* 2-dim map                           */
double **x;      /* argument, x = cos theta             */
int *nx;         /* number of elements in x-array       */
int *np;         /* max number of elements by phi-axis  */
int **nph;       /* number of elements in phi-array     */
char *filename;
{
	register int i,j;
	register FILE *fp_in;
	double fmax = -1.e+30, fmin = 1.e+30;
	long int nphi=0, nph_max=0;

	ph->tform1[0] = ph->tform2[0] = ph->tform3[0] =
	    ph->tdim1[0] = ph->tdim2[0] = ph->tdim3[0] = '\0';

	if ((fp_in = fopen(filename, "r")) == (FILE *)0) {
	    PR(ER, "%s: can not open file '%s'\n", myname, filename);
	    exit(1);
	}

	/****************************************/
	/* read basic keywords from the 1st head */
	/****************************************/
	Hinit(fp_in);

	if (EVget("EXTEND") != NULL)
	    strcpy(ph->extend, EVget("EXTEND"));

	if (strcmp(ph->extend, "T") != 0) {
	    PR(ER, "%s: bad extension in file '%s'\n", myname, filename);
	    exit(1);
	}

	Hfree();

	/****************************************/
	/* read basic keywords from the 2nd head */
	/****************************************/
	Hinit(fp_in);

	if (EVget("XTENSION") != NULL)
	    strcpy(ph->xtens, EVget("XTENSION"));

	if (strcmp(ph->xtens, "BINTABLE") != 0) {
	    PR(ER, "%s: can work only with BINTABLE extension\n", myname);
	    exit(1);
	}
	if (EVget("NAXIS") != NULL)
		ph->naxis  = Skey_get("NAXIS");
	if (EVget("NAXIS1") != NULL)
		ph->naxis1  = Lkey_get("NAXIS1");
	if (EVget("NAXIS2") != NULL)
		ph->naxis2  = Lkey_get("NAXIS2");

	if (EVget("TFIELDS") != NULL)
		ph->tfields = Lkey_get("TFIELDS");

	if( ph->tfields != 3) {
		PR(ER, "%s: bad number of fields (%d) for function ",
			myname, ph->tfields);
		PR(ER, "'ff_read_ext2_map'. It must be 3.\n");
		exit(1);
	}

	if (EVget("TFORM1") != NULL)
		strcpy(ph->tform1, EVget("TFORM1"));
	if (EVget("TDIM1") != NULL)
		strcpy(ph->tdim1, EVget("TDIM1"));
	if (EVget("TTYPE1") != NULL)
		strcpy(ph->ttype1, EVget("TTYPE1"));
	if (EVget("TUNIT1") != NULL)
		strcpy(ph->tunit1, EVget("TUNIT1"));

	if (EVget("TFORM2") != NULL)
		strcpy(ph->tform2, EVget("TFORM2"));
	if (EVget("TDIM2") != NULL)
		strcpy(ph->tdim2, EVget("TDIM2"));
	if (EVget("TTYPE2") != NULL)
		strcpy(ph->ttype2, EVget("TTYPE2"));
	if (EVget("TUNIT2") != NULL)
		strcpy(ph->tunit2, EVget("TUNIT2"));

	if (EVget("TFORM3") != NULL)
		strcpy(ph->tform3, EVget("TFORM3"));
	if (EVget("TDIM3") != NULL)
		strcpy(ph->tdim3, EVget("TDIM3"));
	if (EVget("TTYPE3") != NULL)
		strcpy(ph->ttype3, EVget("TTYPE3"));
	if (EVget("TUNIT3") != NULL)
		strcpy(ph->tunit3, EVget("TUNIT3"));

	if (EVget("OBJECT") != NULL)
		strcpy(ph->object, EVget("OBJECT"));
	if (EVget("ORIGIN") != NULL)
		strcpy(ph->origin, EVget("ORIGIN"));

	/* X-axis grid type */
	if (EVget("GRID") != NULL) {
		char str[51];
		strcpy(str, EVget("GRID"));
		switch(*str) {
		case 'P':
		case 'p':
			if (strcmp("PHI(X)+X", str) == 0)
				ph->grid = GRID_PHI_X;
			else
				ph->grid = GRID_PHI_X_LEGENDRE;
			break;
		case 'L':
		case 'l':
			ph->grid = GRID_LEGENDRE;
			break;
		case 'U':
		case 'u':
			ph->grid = GRID_UNIFORM_COS;
			break;
		case 'T':
		case 't':
			ph->grid = GRID_UNIFORM_THETA;
			break;
		default :
			ph->grid = GRID_NON_SET;
			break;
		}
	}

	if (*(ph->tdim1)=='\0' || *(ph->tdim2)=='\0' || *(ph->tdim3)=='\0') {
	     PR(ER, "%s: bad value of dimensions. Check TDIM# keywords.\n",
		myname);
	     exit(1);
	}

	/* extract dimension of the 1st array */
	i = atoi(ph->tdim1+1);

	/* extract dimensions of the 2nd array */
	*nx = atoi(ph->tdim2+1);
	if (*nx != i ) {
		PR(ER, "%s: bad dimension by X-axis...\n", myname);
		exit(1);
	}
	if (glesp_comments < 0) {
	    if (getenv("GLESP_COMMENTS") != NULL) {
		char str[21];
		strcpy(str,getenv("GLESP_COMMENTS"));
		glesp_comments = *str == 'N' || *str == 'n' ? 0 : 1;
		if (glesp_comments) {
		    if ( strcmp(str, "OFF") == 0 || strcmp(str, "off") == 0)
			glesp_comments = 0;
		}
	    } else
		glesp_comments = 1;
	}

	if (glesp_comments)
	    PR(ER,"%s: Reading map arrays...", myname);
	/**********************************/
	/* take memory and read 1st field */
	/**********************************/
	if ((*x) != (double *)1) {
	    if ((*x = (double *)calloc((unsigned)*nx, sizeof **x)) ==
		(double *)0)
			goto nomem;
	}
	/* read 1-dim array of  X-argument vector */
	bitpix = ph->bitpix1 = atoi(ph->tform1) / *nx * 8;
	for(i=0; i < *nx; i++) {
		(*x)[i] = get_pix(fp_in);
	}

	/**********************************/
	/* take memory and read 2nd field */
	/**********************************/
	if ((*nph) != (int *)1) {
	    if ((*nph = (int*)calloc((unsigned)*nx, sizeof **nph))==(int*)0)
		goto nomem;
	}
	/* read 1-dim array of Nph-vector */
	bitpix = ph->bitpix2 = atoi(ph->tform2) / *nx * 8;
	for(j=0; j < *nx; j++) {
		(*nph)[j] = get_pix(fp_in) + 0.5;
		if ( (*nph)[j] > nph_max )  nph_max = (*nph)[j];
		nphi += (*nph)[j];
	}

	ph->nx = *nx;

	/**********************************/
	/* take memory and read 3d field */
	/**********************************/
	/* find memory for arrays */
	if((*f =(double *)calloc((unsigned)(*nx * nph_max), sizeof **f)) ==
		(double *)0) {
nomem:
		PR(ER, "%s: no memory for arrays...\n", myname);
		exit(1);
	}

	/* in memory we keep transposed image:  */
	/*      f(nphi,nx)                      */
	/*      by X (in strings):  cos theta,  */
	/*      by Y (in columns):  phi.        */
	/* It gives a possibility to change     */
	/* step by X-axis with interpolation    */

	/* read 2-dim array */
	bitpix = ph->bitpix3 = atoll(ph->tform3) / nphi * 8;

	for(i=0; i < *nx; i++) {
		for(j=0; j< (*nph)[i]; j++) {
			double v;
			register int k = j * *nx;
			if (feof(fp_in)) {
			    PR(ER, "%s: unexpected end of file '%s'\n",
				myname, filename);
			    exit(1);
			}
			v = get_pix_d(fp_in);
			if (v > fmax)  fmax = v;
			if (v < fmin)  fmin = v;
			(*f)[k+i] = v;
		}
	}
	if (glesp_comments)
	    PR(ER,"Done.\nMin value=%15.8e, max value=%15.8e.\n", fmin, fmax);

	if (fp_in != stdin && fp_in != (FILE *)0)  fclose(fp_in);
	Hfree();

	*np = nph_max;

	return;
}

/* detection of file type */
/* flags:                 */
/*   F_BIN_TAB_PHI  Binary Table, x = cos theta, phi variable */
/*   F_BIN_TAB      Binary Table, x = cos theta */
/*   F_BFc          basic F-file, x = cos theta */
/*   F_BFc_TR       basic F-file, y = cos theta */
/*   F_BF           basic F-file, x = theta     */
/*   F_BF_TR        basic F-file, y = theta     */
/*   F_other        other type                  */
/*   F_notF         not F-file                  */
int ffile_type(filename)
char *filename;
{
	int file_type_flag;
	char extend[30], buf[161], grid[50];
	register FILE *fp_in;

	grid[0] = '\0';
	if ((fp_in = fopen(filename, "r")) == (FILE *)0) {
	    PR(ER, "%s: can not open file '%s'\n", myname, filename);
	    exit(1);
	}

	fgets(buf, 80, fp_in);

	/* check : is it FLEX-file ? */
	if (strncmp(buf, "SIMPLE", 6) != 0) {
	    file_type_flag = F_notF;

	} else {

	    /****************************************/
	    /* read basic keywords from the 1st head */
	    /****************************************/
	    Hinit(fp_in);

	    if (EVget("EXTEND") != NULL)
		strcpy(extend, EVget("EXTEND"));
	    else
		goto no_extension; /* see below */

	    if (strcmp(extend, "T") != 0) {
		Hfree();
		file_type_flag = F_other;
	    } else {
		Hfree();

		/****************************************/
		/* read basic keywords from the 2nd head */
		/****************************************/
		Hinit(fp_in);

		if (EVget("XTENSION") != NULL) {
		    strcpy(extend, EVget("XTENSION"));

		   if (strcmp(extend, "BINTABLE") != 0) {
		       /* we consider here only Binary Tables */
		       PR(ER, "%s: work only with BINTABLE extension\n",
			   myname);
		       file_type_flag = F_other;
		   } else {
		       /* default it's abinary table */
		       file_type_flag = F_BIN_TAB;

		       /* it is binary table, check other keys */
/*
		       if (EVget("NAXIS") != NULL) {
			   if (Skey_get("NAXIS") != 2) {
			       PR(ER, "%s: NAXIS must be = 2.\n", myname);
			       file_type_flag = F_other;
			   }
		       } else file_type_flag = F_other;
*/
		       if (EVget("GRID") != NULL)
			       strcpy(grid, EVget("GRID"));

		       if (EVget("TFIELDS") != NULL) {
			   if (Lkey_get("TFIELDS") == 2) {
			       if (*grid != 'L')
				  file_type_flag = F_other;
			   } else if (Lkey_get("TFIELDS") == 3) {
				  file_type_flag = F_BIN_TAB_PHI;
				  if (*grid != 'P') {
				     file_type_flag = F_other;
				  }
			   } else {
			       PR(ER, "%s: TFIELDS must be = 2 or 3.\n",
				  myname);
			       file_type_flag = F_other;
			   }
		       } else  file_type_flag = F_other;

		       if (EVget("TTYPE1") != NULL) {
			   strcpy(extend, EVget("TTYPE1"));
			   if(strcmp(extend,"COS(THETA)") != 0 &&
			      strcmp(extend,"COS(ANGLE)") != 0 )
				    file_type_flag = F_other;
			   if (*extend == 'I') {
			       PR(ER,"%s: may be file '%s' contains a_lm ?\n",
				    myname, filename);
			   }
		       } else  file_type_flag = F_other;
		   }
		} else { /* if (EVget("XTENSION") != NULL) */
no_extension:
		   /* if extension is absent it can be ordinary F-file */
		   file_type_flag = F_BFc;

		   /* it is basic F-file, check other keys */
		   if (EVget("NAXIS") != NULL) {
		       if (Skey_get("NAXIS") != 2) {
			   PR(ER, "%s: NAXIS must be = 2\n", myname);
			   file_type_flag = F_other;
		       }
		   } else file_type_flag = F_other;

		   if (EVget("CTYPE1") != NULL) {
		       strcpy(extend, EVget("CTYPE1"));
		       if(strcmp(extend,"THETA") == 0 )
				file_type_flag = F_BF;
		       else if(strcmp(extend,"COS(THETA)") == 0)
				file_type_flag = F_BFc;
		       else if(strcmp(extend,"PHI") == 0) {
			    if (EVget("CTYPE2") != NULL) {
				strcpy(extend, EVget("CTYPE2"));
				if(strcmp(extend,"THETA") == 0 )
					 file_type_flag = F_BF_TR;
				else if(strcmp(extend,"COS(THETA)") == 0)
					 file_type_flag = F_BFc_TR;
				else
				     file_type_flag = F_other;
			    } else
				file_type_flag = F_other;
		       } else
			    file_type_flag = F_other;
		   } else  file_type_flag = F_other;
		}
	    }
	}

	if (fp_in != stdin && fp_in != (FILE *)0)  fclose(fp_in);
	Hfree();

	return file_type_flag;
}

void ff_read_ext_map(f, x, nx, phi, nphi, filename)
double **f;      /* 2-dim map                           */
double **x;      /* argument, x = cos theta             */
int *nx;         /* number of elements in x-array       */
double **phi;    /* array of phi (Y-axis)               */
int *nphi;       /* number of elements in phi-array     */
char *filename;
{
	register int i,j;
	register FILE *fp_in;
	double fmax = -1.e+30, fmin = 1.e+30;

	ph->tdim1[0] = ph->tdim2[0] = '\0';

	if ((fp_in = fopen(filename, "r")) == (FILE *)0) {
	    PR(ER, "%s: can not open file '%s'\n", myname, filename);
	    exit(1);
	}

	/****************************************/
	/* read basic keywords from the 1st head */
	/****************************************/
	Hinit(fp_in);

	if (EVget("EXTEND") != NULL)
	    strcpy(ph->extend, EVget("EXTEND"));

	if (strcmp(ph->extend, "T") != 0) {
	    PR(ER, "%s: bad extension in file '%s'\n", myname, filename);
	    exit(1);
	}

	Hfree();

	/****************************************/
	/* read basic keywords from the 2nd head */
	/****************************************/
	Hinit(fp_in);


	if (EVget("XTENSION") != NULL)
	    strcpy(ph->xtens, EVget("XTENSION"));

	if (strcmp(ph->xtens, "BINTABLE") != 0) {
	    PR(ER, "%s: can work only with BINTABLE extension\n", myname);
	    exit(1);
	}
	if (EVget("NAXIS") != NULL)
		ph->naxis  = Skey_get("NAXIS");
	if (EVget("NAXIS1") != NULL)
		ph->naxis1  = Lkey_get("NAXIS1");
	if (EVget("NAXIS2") != NULL)
		ph->naxis2  = Lkey_get("NAXIS2");

	if (EVget("TFIELDS") != NULL)
		ph->tfields = Lkey_get("TFIELDS");

	if (EVget("TFORM1") != NULL)
		strcpy(ph->tform1, EVget("TFORM1"));
	if (EVget("TDIM1") != NULL)
		strcpy(ph->tdim1, EVget("TDIM1"));
	if (EVget("TTYPE1") != NULL)
		strcpy(ph->ttype1, EVget("TTYPE1"));
	if (EVget("TUNIT1") != NULL)
		strcpy(ph->tunit1, EVget("TUNIT1"));

	if (EVget("TFORM2") != NULL)
		strcpy(ph->tform2, EVget("TFORM2"));
	if (EVget("TDIM2") != NULL)
		strcpy(ph->tdim2, EVget("TDIM2"));
	if (EVget("TTYPE2") != NULL)
		strcpy(ph->ttype2, EVget("TTYPE2"));
	if (EVget("TUNIT2") != NULL)
		strcpy(ph->tunit2, EVget("TUNIT2"));

	if (EVget("CRVAL2") != NULL)
		ph->crval2 = Dkey_get("CRVAL2");
	if (EVget("CDELT2") != NULL)
		ph->cdelt2 = Dkey_get("CDELT2");
	if (EVget("CRPIX2") != NULL)
		ph->crpix2 = Dkey_get("CRPIX2");
	if (EVget("CTYPE2") != NULL)
		strcpy(ph->ctype2, EVget("CTYPE2"));

	if (EVget("OBJECT") != NULL)
		strcpy(ph->object, EVget("OBJECT"));
	if (EVget("ORIGIN") != NULL)
		strcpy(ph->origin, EVget("ORIGIN"));

	/* X-axis grid type */
	if (EVget("GRID") != NULL) {
		char str[51];
		strcpy(str, EVget("GRID"));
		switch(*str) {
		case 'P':
		case 'p':
			if (strcmp("PHI(X)+X", str) == 0)
				ph->grid = GRID_PHI_X;
			else
				ph->grid = GRID_PHI_X_LEGENDRE;
			break;
		case 'L':
		case 'l':
			ph->grid = GRID_LEGENDRE;
			break;
		case 'U':
		case 'u':
			ph->grid = GRID_UNIFORM_COS;
			break;
		case 'T':
		case 't':
			ph->grid = GRID_UNIFORM_THETA;
			break;
		default :
			ph->grid = GRID_NON_SET;
			break;
		}
	}

	if (*(ph->tdim1) == '\0' || *(ph->tdim2) == '\0') {
	     PR(ER, "%s: bad value of dimensions. Check TDIM# keywords.\n",
		myname);
	     exit(1);
	}

	/* extract dimension of the 1st array */
	i = atoi(ph->tdim1+1);

	/* extract dimensions of the 2nd array */
	*nx = atoi(ph->tdim2+1);
	if (*nx != i ) {
		PR(ER, "%s: bad dimension of by X-axis...\n", myname);
		exit(1);
	}
	*nphi = atoi(strchr(ph->tdim2, ',') + 1);
	ph->nx = *nx;
	ph->ny = *nphi;

	ph->bitpix1 = atoi(ph->tform1) / *nx;
	ph->bitpix2 = atoi(ph->tform2) / *nx / *nphi;

	/* find memory for arrays */
	if ((*f = (double *)calloc((unsigned)(*nx * *nphi), sizeof **f)) ==
		(double *)0) {
nomem:
		PR(ER, "%s: no memory for arrays...\n", myname);
		exit(1);
	}
	if ((*x = (double *)calloc((unsigned)*nx, sizeof **x)) ==
		(double *)0)
			goto nomem;
	if ((*phi = (double *)calloc((unsigned)*nphi, sizeof **phi)) ==
		(double *)0)
			goto nomem;

	if (glesp_comments < 0) {
	    if (getenv("GLESP_COMMENTS") != NULL) {
		char str[21];
		strcpy(str,getenv("GLESP_COMMENTS"));
		glesp_comments = *str == 'N' || *str == 'n' ? 0 : 1;
		if (glesp_comments) {
		    if ( strcmp(str, "OFF") == 0 || strcmp(str, "off") == 0)
			glesp_comments = 0;
		}
	    } else
		glesp_comments = 1;
	}

	if (glesp_comments)
	    PR(ER,"%s: Reading map arrays...", myname);
	/* fill array of phi */
	for(i=0; i < *nphi; i++)
		(*phi)[i] = ph->crval2 + (1. - ph->crpix2 + i) * ph->cdelt2;

	/* read 1-dim array of  X-argument vector */
	bitpix = ph->bitpix1;
	for(i=0; i < *nx; i++) {
		(*x)[i] = get_pix(fp_in);
	}
/*
for(i=0; i < *nx; i++) {
if (i%5 == 0) PR(stdout, "\n");
PR(stdout, " %12.5e", (*x)[i]);
}
PR(stdout, "\n");
*/
	/* in memory we keep transposed image:  */
	/*      f(nphi,nx)                      */
	/*      by X (in strings):  cos theta,  */
	/*      by Y (in columns):  phi.        */
	/* It gives a possibility to change     */
	/* step by X-axis with interpolation    */

	/* read 2-dim array */
	bitpix = ph->bitpix2;
	for(j=0; j < *nx; j++) {
		for(i=0; i< *nphi; i++) {
			double v;
			register int k = i * *nx;
			if (feof(fp_in)) {
			    PR(ER, "%s: unexpected end of file '%s'\n",
				myname, filename);
			    exit(1);
			}
			v = get_pix(fp_in);
			if (v > fmax)  fmax = v;
			if (v < fmin)  fmin = v;
			(*f)[k+j] = v;
		}
	}

	if (glesp_comments)
	    PR(ER,"Done.\nMin value=%12.5e, max value=%12.5e.\n", fmin, fmax);
/*
for(i=0; i < *nphi; i++) {
register int k = i * *nx;
for(j=0; j < *nx; j++) {
if (j%5 == 0) PR(stdout, "\n");
PR(stdout, " %12.5e", (*f)[k+j]);
}
PR(stdout, "\n");
}
*/
	if (fp_in != stdin && fp_in != (FILE *)0)  fclose(fp_in);
	Hfree();

	return;
}



/* read basic-format, create f-, x-, phi-arrays */
/* and catch memory                             */
void ff_read_stand_map(f, x, nx, phi, nphi, transp_flag, filename)
double **f;      /* 2-dim map                         */
double **x;      /* argument, x = theta or cos theta  */
int *nx;         /* number of elements in x-array     */
double **phi;    /* array of phi (Y-axis)             */
int *nphi;       /* number of elements in phi-array   */
int transp_flag;
char *filename;
{
	register int i,j;
	register FILE *fp_in;
	int naxis2=0;

	ph->tdim1[0] = ph->tdim2[0] = '\0';

	if ((fp_in = fopen(filename, "r")) == (FILE *)0) {
	    PR(ER, "%s: can not open file '%s'\n", myname, filename);
	    exit(1);
	}

	/************************************/
	/* read basic keywords from the head */
	/************************************/
	Hinit(fp_in);

	if (EVget("BITPIX") != NULL)
		bitpix  = Skey_get("BITPIX");
	ph->bitpix1 = ph->bitpix2 = bitpix;
	if (EVget("BSCALE") != NULL)
		bscale  = Dkey_get("BSCALE");
	if (EVget("BZERO") != NULL)
		b_zero  = Dkey_get("BZERO");

	if (EVget("NAXIS") != NULL)
		naxis  = Skey_get("NAXIS");
	if (EVget("NAXIS1") != NULL)
		naxis1  = Skey_get("NAXIS1");
	if (EVget("NAXIS2") != NULL)
		naxis2  = Skey_get("NAXIS2");
	if (transp_flag) {
		*nx   = ph->nx = naxis1;
		*nphi = ph->ny = naxis2;
	} else {
		*nphi = ph->ny = naxis1;
		*nx   = ph->nx = naxis2;
	}

	if (EVget("CRVAL1") != NULL)
		crval1 = Dkey_get("CRVAL1");
	if (EVget("CDELT1") != NULL)
		cdelt1 = Dkey_get("CDELT1");
	if (EVget("CRPIX1") != NULL)
		crpix1 = Dkey_get("CRPIX1");
	if (EVget("CTYPE1") != NULL)
		strcpy(ctype1, EVget("CTYPE1"));
	strcpy(ph->ttype1, ctype1);

	if (EVget("CRVAL2") != NULL)
		ph->crval2 = Dkey_get("CRVAL2");
	if (EVget("CDELT2") != NULL)
		ph->cdelt2 = Dkey_get("CDELT2");
	if (EVget("CRPIX2") != NULL)
		ph->crpix2 = Dkey_get("CRPIX2");
	if (EVget("CTYPE2") != NULL)
		strcpy(ph->ctype2, EVget("CTYPE2"));

	if (EVget("OBJECT") != NULL)
		strcpy(ph->object, EVget("OBJECT"));
	if (EVget("ORIGIN") != NULL)
		strcpy(ph->origin, EVget("ORIGIN"));

	/* X-axis grid type */
	if (EVget("GRID") != NULL) {
		char str[51];
		strcpy(str, EVget("GRID"));
		switch(*str) {
		case 'P':
		case 'p':
			if (strcmp("PHI(X)+X", str) == 0)
				ph->grid = GRID_PHI_X;
			else
				ph->grid = GRID_PHI_X_LEGENDRE;
			break;
		case 'L':
		case 'l':
			ph->grid = GRID_LEGENDRE;
			break;
		case 'U':
		case 'u':
			ph->grid = GRID_UNIFORM_COS;
			break;
		case 'T':
		case 't':
			ph->grid = GRID_UNIFORM_THETA;
			break;
		default :
			ph->grid = GRID_NON_SET;
			break;
		}
	}

	/* find memory for arrays */
	if ((*f = (double *)calloc((unsigned)(*nx * *nphi), sizeof **f)) ==
		(double *)0) {
nomem:
		PR(ER, "%s: no memory for arrays...\n", myname);
		exit(1);
	}
	if ((*x = (double *)calloc((unsigned)*nx, sizeof **x)) ==
		(double *)0)
			goto nomem;
	if ((*phi = (double *)calloc((unsigned)*nphi, sizeof **phi)) ==
		(double *)0)
			goto nomem;

	/* in memory we keep transposed image:  */
	/*      f(nphi,nx)                      */
	/*      by X (in strings):  cos theta,  */
	/*      by Y (in columns):  phi.        */
	/* It gives a possibility to change     */
	/* step by X-axis with interpolation    */

	if (transp_flag) {
	    /* fill array of phi */
	    for(i=0; i < *nphi; i++)
		(*phi)[i] = ph->crval2 + (i + 1. - ph->crpix2) * ph->cdelt2;

	    /* read 1-dim array of  X-argument vector */
	    /* theta or cos(theta)                    */
	    for(i=0; i < *nx; i++)
		(*x)[i] = crval1 + (i + 1. - crpix1) * cdelt1;

PR(ER,"nx=%d, nphi=%d, x0=%g, xN=%g\n", *nx, *nphi, (*x)[0], (*x)[*nx-1] );
	    /* read 2-dim array */
	    for(i=0; i < *nphi; i++) {      /* naxis2 */
		  register int k = i * *nx;
		  for(j=0; j< *nx; j++) { /* naxis1 */
			(*f)[k+j] = b_zero + bscale * get_pix(fp_in);
/*
if(i<2 && j<15) {
PR(ER, " %12.5e", (*f)[ii+j]);
if((j+1)%5==0) PR(ER, "\n");
}
*/
		  }
	    }
	} else {
	    /* fill array of phi */
	    for(i=0; i < *nphi; i++)
		(*phi)[i] = crval1 + (i + 1. - crpix1) * cdelt1;

	    /* read 1-dim array of  X-argument vector */
	    /* theta or cos(theta)                    */
	    for(i=0; i < *nx; i++)
		(*x)[i] = ph->crval2 + (i + 1. - ph->crpix2) * ph->cdelt2;

	    /* read 2-dim array */
	    for(j=0; j< *nx; j++) { /* naxis1 */
		for(i=0; i < *nphi; i++) {      /* naxis2 */
			register int k = i * *nx;
			(*f)[k+j] = b_zero + bscale * get_pix(fp_in);
		}
	    }

	    strcpy(ph->ttype1, ph->ctype2);
	    strcpy(ph->ctype2, ctype1);
	    ph->cdelt2 = cdelt1;
	    ph->crval2 = crval1;
	    ph->crpix2 = crpix1;
	}

	if (fp_in != stdin && fp_in != (FILE *)0)  fclose(fp_in);
	Hfree();

	return;
}

