#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include <errno.h>


#include "fits.h"
#include "glesp.h"

#define PR  fprintf
#define ER  stderr
#define  BLOCK_SIZE     2880

/*
 * created by Oleg Verkhodanov
 *    last correction: 14/05/2006
 */


extern int flag_take_memory;
extern int bitpix_map;

void outp_glesp_map(f, x, nx, np, nph, fp_out, x_type, grid_type)
double *f;     /* 2-dim map                           */
double *x;     /* argument, x = cos theta             */
int nx,        /* number of elements in x-array       */
    np;        /* 2-dim factor: f(i*nbuf+j)<->f[i][j] */
int *nph;      /* array of elements in phi-arrays     */
FILE *fp_out;
char *x_type, *grid_type;
{
	register int i,j,ii;
	char ttype1[30], tform1[30], tunit1[30],
	     ttype2[30], tform2[30], tunit2[30],
	     ttype3[30], tform3[30], tunit3[30];
	int tfields = 3;
	long nphi=0;     /* total number of elements in nph-arrays */
	int bitp1 = GLESPPIX,
	    bitp2 = LONGPIX,
	    bitp3;
	long long long_naxis1;
	char ctype1[30], bunit[30], object[30];
	int block_data = 0;

	/*****************/
	/* basic keywods */
	/*****************/
	bitp3 = bitpix_map; /* default: Float Point Pixels for map */
	bitp1 = bitp3;

	strcpy(ctype1, x_type);
	strcpy(bunit,  "K");
	strcpy(object, "CMB map");

	tfields = 3;

	if (bitp1 == FLOATPIX)
	    sprintf(tform1, "%dE", nx);
	else  if (bitp1 == DOUBLEPIX)
	    sprintf(tform1, "%dD", nx);
	strcpy(ttype1,  ctype1);
	if (strncmp(ctype1, "COS(", 4) == 0)
		strcpy(tunit1,  "COS(RAD)");
	else
		strcpy(tunit1,  "RAD");

	sprintf(tform2, "%dJ", nx);
	strcpy(ttype2,  "DIMS");
	strcpy(tunit2,  "UNITLESS");

	for (i=0; i<nx; i++)  nphi += nph[i];
	if (bitp3 == FLOATPIX)
	    sprintf(tform3, "%ldE", nphi);
	else  if (bitp3 == DOUBLEPIX)
	    sprintf(tform3, "%ldD", nphi);
	strcpy(ttype3,  "TEMPERATURE");
	strcpy(tunit3,  "K");

	/* Number of entries in table in bytes */
	long_naxis1 = (long long)nphi * (ABS(bitp3) / 8) +
	    nx * (ABS(bitp1) / 8) + nx * (ABS(bitp1) / 8);

	/************/
	/* set head */
	/************/
	/* sFEVset("SIMPLE", "T"); FEVexport("SIMPLE"); */
	Fckey_set("SIMPLE",    "T");
	FSkey_set("BITPIX",    8);
	FSkey_set("NAXIS",     0);
	Fckey_set("EXTEND",    "T");
	FCkey_set("COMMENT",  "GLESP MAP");

	/* 1st HEAD */
	Hupdate_fits(fp_out);
	FHfree();

	FCkey_set("XTENSION",  "BINTABLE");
	FSkey_set("BITPIX",  8);
	FSkey_set("NAXIS",   2);
	FLLkey_set("NAXIS1", long_naxis1);
	FLkey_set("NAXIS2",  1);
	FLkey_set("PCOUNT",  0);
	FLkey_set("GCOUNT",  1);

	FLkey_set("TFIELDS", tfields);

	FCkey_set("TFORM1",  tform1); /* 1st field: vector of X */
	/* FCkey_set("TDIM1",   tdim1); */
	FCkey_set("TTYPE1",  ttype1);
	FCkey_set("TUNIT1",  tunit1);

	FCkey_set("TFORM2",  tform2); /* 2nd field: vector nph(x) */
	/* FCkey_set("TDIM2",   tdim2); */
	FCkey_set("TTYPE2",  ttype2);
	FCkey_set("TUNIT2",  tunit2);

	FCkey_set("TFORM3",  tform3); /* 3d field: vector of V(X,Y)  */
	/* FCkey_set("TDIM3",   tdim3); */
	FCkey_set("TTYPE3",  ttype3);
	FCkey_set("TUNIT3",  tunit3);

	FCkey_set("OBJECT",  object);
	FCkey_set("ORIGIN",  "GLESP");
	FCkey_set("PIXTYPE", "GLESP");

	/* 2nd HEAD */
	Hupdate_fits(fp_out);

	/* recording 1-dim array of  X-argument vector */
	bitpix = bitp1;
	for(i=0; i<nx; i++) {
		double v = x[i];
		put_pix_fits(fp_out, v);
	}

	/* recording 1-dim array of dimensions of \phi-vector */
	bitpix = bitp2;
	for(i=0; i<nx; i++) {
		double v = nph[i];
		put_pix_fits(fp_out, v);
	}

	bitpix = bitp3;
	/* recording 2-dim array */
	for(ii=i=0; i<nx; i++,ii+=np) {
		for(j=0; j<nph[i]; j++) {
			double v = f[ii+j];
			put_pix_fits(fp_out, v);
		}
	}

	block_data = long_naxis1;
	if(block_data % BLOCK_SIZE != 0) {
	       long len = BLOCK_SIZE * ( block_data / BLOCK_SIZE + 1);
	       len -= block_data;
	       for(i=0; i<len; i++)   fputc('\0', fp_out);
	}

	/* free memory from head */
	FHfree();

	return;
}

void in_glesp_map(f, x, nx, np, nph, filename)
double **f;      /* 2-dim map                           */
double **x;      /* argument, x = cos theta             */
int *nx;         /* number of elements in x-array       */
int *np;         /* max number of elements by phi-axis  */
int **nph;       /* number of elements in phi-array     */
char *filename;
{
	register int i,j,ii;
	register FILE *fp_in;
	double fmax = -1.e+30, fmin = 1.e+30;
	long int nphi=0, nph_max=0;

	ph->tform1[0] = ph->tform2[0] = ph->tform3[0] =
	      ph->tdim1[0] = ph->tdim2[0] = ph->tdim3[0] = '\0';

	if ((fp_in = fopen(filename, "r")) == (FILE *)0) {
	    PR(ER, "%s: can not open file '%s'\n", myname, filename);
	    exit(1);
	}

	/****************************************/
	/* read basic keywords from the 1st head */
	/****************************************/
	/* Hinit_fits(fp_in); */
	FHinit_FITS(fp_in, 0);

	if (FEVget("EXTEND") != NULL)
	    strcpy(ph->extend, FEVget("EXTEND"));

	if (strcmp(ph->extend, "T") != 0) {
	    PR(ER, "%s: bad extension in file '%s'\n", myname, filename);
	    exit(1);
	}

	FHfree();

	/****************************************/
	/* read basic keywords from the 2nd head */
	/****************************************/
	/* Hinit(fp_in); */
	FHinit_FITS(fp_in, 0);

	if (FEVget("XTENSION") != NULL)
	    strcpy(ph->xtens, FEVget("XTENSION"));

	if (strcmp(ph->xtens, "BINTABLE") != 0) {
	    PR(ER, "%s: can work only with BINTABLE extension\n", myname);
	    exit(1);
	}
	if (FEVget("NAXIS") != NULL)
		ph->naxis  = FSkey_get("NAXIS");
	if (FEVget("NAXIS1") != NULL)
		ph->naxis1  = FLkey_get("NAXIS1");
	if (FEVget("NAXIS2") != NULL)
		ph->naxis2  = FLkey_get("NAXIS2");

	if (FEVget("TFIELDS") != NULL)
		ph->tfields = FLkey_get("TFIELDS");

	if( ph->tfields != 3) {
		PR(ER, "%s: bad number of fields (%d) for function ",
			myname, ph->tfields);
		PR(ER, "'ff_read_ext2_map'. It must be 3.\n");
		exit(1);
	}

	if (FEVget("TFORM1") != NULL)
		strcpy(ph->tform1, FEVget("TFORM1"));
	if (FEVget("TDIM1") != NULL)
		strcpy(ph->tdim1, FEVget("TDIM1"));
	if (FEVget("TTYPE1") != NULL)
		strcpy(ph->ttype1, FEVget("TTYPE1"));
	if (FEVget("TUNIT1") != NULL)
		strcpy(ph->tunit1, FEVget("TUNIT1"));

	if (FEVget("TFORM2") != NULL)
		strcpy(ph->tform2, FEVget("TFORM2"));
	if (FEVget("TDIM2") != NULL)
		strcpy(ph->tdim2, FEVget("TDIM2"));
	if (FEVget("TTYPE2") != NULL)
		strcpy(ph->ttype2, FEVget("TTYPE2"));
	if (FEVget("TUNIT2") != NULL)
		strcpy(ph->tunit2, FEVget("TUNIT2"));

	if (FEVget("TFORM3") != NULL)
		strcpy(ph->tform3, FEVget("TFORM3"));
	if (FEVget("TDIM3") != NULL)
		strcpy(ph->tdim3,  FEVget("TDIM3"));
	if (FEVget("TTYPE3") != NULL)
		strcpy(ph->ttype3, FEVget("TTYPE3"));
	if (FEVget("TUNIT3") != NULL)
		strcpy(ph->tunit3, FEVget("TUNIT3"));

	if (FEVget("OBJECT") != NULL)
		strcpy(ph->object, FEVget("OBJECT"));
	if (FEVget("ORIGIN") != NULL)
		strcpy(ph->origin, FEVget("ORIGIN"));

	/* set grid flag */

	/* X-axis grid type */
	if (FEVget("GRID") != NULL) {
		char str[51];
		strcpy(str, FEVget("GRID"));
		switch(*str) {
		case 'P':
		case 'p':
			if (strcmp("PHI(X)+X", str) == 0)
				ph->grid = GRID_PHI_X;
			else
				ph->grid = GRID_PHI_X_LEGENDRE;
			break;
		case 'L':
		case 'l':
			ph->grid = GRID_LEGENDRE;
			break;
		case 'U':
		case 'u':
			ph->grid = GRID_UNIFORM_COS;
			break;
		case 'T':
		case 't':
			ph->grid = GRID_UNIFORM_THETA;
			break;
		default :
			ph->grid = GRID_NON_SET;
			break;
		}
	}
	if(*(ph->tform1)=='\0'|| *(ph->tform2)=='\0'|| *(ph->tform3)=='\0') {
	     PR(ER, "%s: bad value of dimensions. Check TFORM# keywords.\n",
		myname);
	     exit(1);
	}

	/* extract dimension of the 1st array */
	i = atoi(ph->tform1);

	/* extract dimensions of the 2nd array */
	*nx = atoi(ph->tform2);
	if (*nx != i ) {
		PR(ER, "%s: bad dimension by X-axis...\n", myname);
		exit(1);
	}
	if (glesp_comments < 0) {
	    if (getenv("GLESP_COMMENTS") != NULL) {
		char str[21];
		strcpy(str,getenv("GLESP_COMMENTS"));
		glesp_comments = *str == 'N' || *str == 'n' ? 0 : 1;
		if (glesp_comments) {
		    if ( strcmp(str, "OFF") == 0 || strcmp(str, "off") == 0)
			glesp_comments = 0;
		}
	    } else
		glesp_comments = 1;
	}

	if (glesp_comments)
	    PR(ER,"%s: Reading map arrays...", myname);

	/**********************************/
	/* take memory and read 1st field */
	/**********************************/
	if (flag_take_memory) {
	    if ((*x = (double *)calloc((unsigned)*nx, sizeof **x)) ==
		    (double *)0)
			    goto nomem;
	}
	/* read 1-dim array of  X-argument vector */
	/* bitpix = ph->bitpix1 = atoi(ph->tform1) / *nx * 8; */
	i = ph->tform1[strlen(ph->tform1)-1];
	if ( i == 'D' )
		bitpix = ph->bitpix1 = DOUBLEPIX;
	else /* if ( i == 'E' )*/
		bitpix = ph->bitpix1 = FLOATPIX;
	for(i=0; i < *nx; i++) {
		(*x)[i] = get_pix_fits(fp_in);
	}

	/**********************************/
	/* take memory and read 2nd field */
	/**********************************/
	if (flag_take_memory) {
	    if ((*nph = (int*)calloc((unsigned)*nx, sizeof **nph))==(int *)0)
		goto nomem;
	}
	/* read 1-dim array of Nph-vector */
	/* bitpix = ph->bitpix2 = atoi(ph->tform2) / *nx * 8; */
	i = ph->tform2[strlen(ph->tform2)-1];
	switch (i) {
	case 'I':
		ph->bitpix2 = SHORTPIX;
		break;
	case 'J':
		ph->bitpix2 = LONGPIX;
		break;
	case 'D':
		ph->bitpix2 = DOUBLE;
		break;
	default:
	case 'E':
		ph->bitpix2 = FLOATPIX;
		break;
	}
	bitpix = ph->bitpix2;
	for(j=0; j < *nx; j++) {
		(*nph)[j] = get_pix_fits(fp_in) + 0.5;
		if ( (*nph)[j] > nph_max )  nph_max = (*nph)[j];
		nphi += (*nph)[j];
	}

	ph->nx = *nx;

	/**********************************/
	/* take memory and read 3d field */
	/**********************************/
	/* find memory for arrays */
	if (flag_take_memory) {
	   if ((*f =(double*)calloc((unsigned)(*nx *nph_max), sizeof **f)) ==
		(double *)0) {
nomem:
		PR(ER, "%s: no memory for arrays...\n", myname);
		exit(1);
	   }
	}

	/* in memory we keep transposed image:  */
	/*      f(nphi,nx)                      */
	/*      by X (in strings):  cos theta,  */
	/*      by Y (in columns):  phi.        */
	/* It gives a possibility to change     */
	/* step by X-axis with interpolation    */

	/* read 2-dim array */
	/* bitpix = ph->bitpix3 = atoll(ph->tform3) / nphi * 8; */
	i = ph->tform3[strlen(ph->tform3)-1];
	switch (i) {
	case 'D':
		ph->bitpix3 = DOUBLEPIX;
		break;
	default:
	case 'E':
		ph->bitpix3 = FLOATPIX;
		break;
	}
	bitpix = ph->bitpix3;

	for(ii=i=0; i < *nx; i++,ii+=nph_max) {
		for(j=0; j< (*nph)[i]; j++) {
			double v;
			if (feof(fp_in)) {
			    PR(ER, "%s: unexpected end of file '%s'\n",
				myname, filename);
			    exit(1);
			}
			v = get_pix_fits(fp_in);
			if (v > fmax)  fmax = v;
			if (v < fmin)  fmin = v;
			(*f)[ii+j] = v;
		}
	}
	if (glesp_comments)
	PR(ER,"Done.\nMin value=%15.8e, max value=%15.8e, (nx,np)=(%d,%ld)\n",
	    fmin, fmax, *nx, nph_max);

	if (fp_in != stdin && fp_in != (FILE *)0)  fclose(fp_in);
	FHfree();

	*np = nph_max;

	return;
}
void in_glesp_alm(alm, lmax, nbuf, filename)
double *alm;
int *lmax;
int nbuf;
char *filename;
{
	register int i, l, m;
	int tfields = 0;
	int bitp1, bitp2, bitp3;
	char tform1[30],
	     tform2[30],
	     tform3[30];
	char str[161];
	int  lmx = -1, lmax_input = *lmax;
	FILE *fp;
	int naxis=0, naxis1=0, naxis2=0;

	/* open file */
	if ((fp = fopen(filename, "r")) == (FILE *)0) {
		PR(ER, "%s: can not open file '%s'\n", myname, filename);
		exit(1);
	}
	/* read 1st head */
	FHinit(fp);

	if (FEVget("EXTEND") != NULL)
	    strcpy(str, FEVget("EXTEND"));

	if (strcmp(str, "T") != 0) {
	    PR(ER, "%s: bad extension in file '%s'\n", myname, filename);
	    exit(1);
	}

	/* Free memory from 1st head */
	FHfree();

	/****************************************/
	/* read basic keywods from the 2nd head */
	/****************************************/
	FHinit(fp);

	if (FEVget("XTENSION") != NULL)
	    strcpy(str, FEVget("XTENSION"));
	else {
	    PR(ER, "%s: no extension in file '%s'\n", myname, filename);
	    exit(1);
	}
	if (strcmp(str, "BINTABLE") != 0) {
	    PR(ER, "%s: can work only with BINTABLE extension\n", myname);
	    exit(1);
	}
	if (FEVget("NAXIS") != NULL)
		naxis  = FSkey_get("NAXIS");
	if (FEVget("NAXIS1") != NULL)
		naxis1  = FLkey_get("NAXIS1");
	if (FEVget("NAXIS2") != NULL)
		naxis2  = FLkey_get("NAXIS2");

	if (FEVget("TFIELDS") != NULL)
		tfields = FLkey_get("TFIELDS");
	if (tfields != 3) {
	    PR(ER, "%s: TFILEDS value must be equal to 3 in a_lm file '%s'\n",
		myname, filename);
	    exit(1);
	}

	if (FEVget("TFORM1") != NULL) {
		strcpy(tform1, FEVget("TFORM1"));
		if (strcmp(tform1, "1J") == 0) {
		    bitp1 = LONGPIX;
		} else if (strcmp(tform1, "1I") == 0) {
		    bitp1 = SHORTPIX;
		} else {
		    PR(ER, "%s: bad value for TFORM1. Sorry...\n", myname);
		    exit(1);
		}
	} else {
		PR(ER, "%s: no parameter TFORM1. Sorry...\n", myname);
		exit(1);
	}
	if (FEVget("TFORM2") != NULL) {
		strcpy(tform2, FEVget("TFORM2"));
		if (strcmp(tform2, "1E") == 0) {
		    bitp2 = FLOATPIX;
		} else if (strcmp(tform2, "1D") == 0) {
		    bitp2 = DOUBLEPIX;
		} else {
		    PR(ER, "%s: bad value for TFORM2. Sorry...\n", myname);
		    exit(1);
		}
	} else {
		PR(ER, "%s: no parameter TFORM2. Sorry...\n", myname);
		exit(1);
	}
	if (FEVget("TFORM3") != NULL) {
		strcpy(tform3, FEVget("TFORM3"));
		if (strcmp(tform2, "1E") == 0) {
		    bitp3 = FLOATPIX;
		} else if (strcmp(tform3, "1D") == 0) {
		    bitp3 = DOUBLEPIX;
		} else {
		    PR(ER, "%s: bad value for TFORM3. Sorry...\n", myname);
		    exit(1);
		}
	} else {
		PR(ER, "%s: no parameter TFORM3. Sorry...\n", myname);
		exit(1);
	}
	if (FEVget("LMAX") != NULL) {
		*lmax = atoi(FEVget("LMAX"));
	}

	if (lmax_input > 0 && lmax_input <= *lmax) {
	     *lmax = lmax_input;
	} else {
	     if (nbuf < (*lmax + 1)*2) {
		 if ( (lmax_input + 1) * 2 > nbuf)
		     lmax_input = nbuf / 2 - 1;
		 *lmax = lmax_input;
	     } else {
	       if (*lmax > lmax_input) {
		 PR(ER, "%s: check set of lmax: lmax=%d, nbuf=%d\n",
			myname, *lmax, nbuf);
		 exit(1);
	       }
	     }
	}

	/* Free memory from 1nd head */
	FHfree();

	/* get values from file */
	for (i=0; i<naxis2; i++) {
	    static int flag = 1;
	    register int bb, nlm;
	    double ddd;

	    bitpix = bitp1;
	    //nlm = get_pix_fits(fp)+0.5;
	    ddd = get_pix_fits(fp);
	    nlm = ddd + .5;
	    l = (int)sqrt((double)(nlm-1));

	    if (l > *lmax)
		break;

	    if (l*2+1 > nbuf) {
		if (flag) {
		    PR(ER, "%s: attention ! l(=%d) is larger ", myname, l);
		    PR(ER, "than array dimension (2*l+1>nbuf(%d)).\n", nbuf);
		}
		flag = 0;
		break;
	    }
	    m = nlm - 1 - l*l - l;
	    bb = l * nbuf;

	    /* Real part */
	    bitpix = bitp2;
	    alm [bb+m]        = get_pix_fits(fp);
	    if (feof(fp)) {
		 PR(ER, "%s: unexpected end of file at l=%d and m=%d\n",
		    myname, l, m);
		 break;
	    }
	    /* Imagine part */
	    bitpix = bitp3;
	    alm [bb+m+ *lmax+1] = get_pix_fits(fp);
	    if (feof(fp)) {
		 PR(ER, "%s: unexpected end of file at l=%d and m=%d\n",
		    myname, l, m);
		 break;
	    }
	    if (l > lmx)   lmx = l;
	}
	if (*lmax <= 0)  *lmax = lmx;

	fclose(fp);
	return;
}

void out_glesp_alm(alm, lmax, nbuf, lmax_print, fp_out)
double *alm;
int lmax, lmax_print;
int nbuf;
FILE *fp_out;
{
	register int l, m, bb;
	int nlm = (lmax_print+1)*(lmax_print+2) / 2;
	int tfields = 3, nsize = 0;
	char ttype1[30], tform1[30], tunit1[30],
	     ttype2[30], tform2[30], tunit2[30],
	     ttype3[30], tform3[30], tunit3[30];
	int bitp1, bitp2, bitp3;
	int block_data = 0;

	strcpy(tform1, "1J");
	strcpy(ttype1,  "INDEX");
	strcpy(tunit1,  "UNITLESS");

	bitp2=bitpix_map;
	bitp3=bitpix_map;
	if (bitpix_map == FLOATPIX) {
	    strcpy(tform2, "1E");
	    strcpy(tform3, "1E");
	} else if (bitpix_map == DOUBLEPIX) {
	    strcpy(tform2, "1D");
	    strcpy(tform3, "1D");
	}
	strcpy(ttype2,  "REAL");
	strcpy(tunit2,  "mK");
	strcpy(ttype3,  "IMAG");
	strcpy(tunit3,  "mK");

	/* TFORM 1 */
	bitp1=LONGPIX;
	if (strcmp(tform1, "1J") == 0) {
		nsize += 4;
		bitp1=LONG;
	} else if (strcmp(tform1, "1I") == 0) {
		nsize += 2;
		sprintf(tform1, "1I");
		bitp1=SHORTPIX;
	}
	/* TFORM 2 */
	if (strcmp(tform2, "1E") == 0) {
		nsize += 4;
		bitp2=FLOATPIX;
	} else if (strcmp(tform2, "1D") == 0) {
		nsize += 8;
		bitp2=DOUBLEPIX;
	}
	/* TFORM 3 */
	if (strcmp(tform3, "1E") == 0) {
		nsize += 4;
		bitp3=FLOATPIX;
	} else if (strcmp(tform3, "1D") == 0) {
		nsize += 8;
		bitp3=DOUBLEPIX;
	}
	/************/
	/* set head */
	/************/
	Fckey_set("SIMPLE",    "T");
	FSkey_set("BITPIX",    8);
	FSkey_set("NAXIS",     0);
	Fckey_set("EXTEND",    "T");
	FCkey_set("COMMENT",  "a_{lm} coefficients");

	FHupdate(fp_out);
	FHfree();

	FCkey_set("XTENSION",  "BINTABLE");
	FSkey_set("BITPIX",  8);
	FSkey_set("NAXIS",   2);
	FLkey_set("NAXIS1",  nsize);
	FLkey_set("NAXIS2",  nlm);
	FLkey_set("PCOUNT",  0);
	FLkey_set("GCOUNT",  1);

	FLkey_set("TFIELDS", tfields);
	FCkey_set("EXTNAME", "A_lm DATA");

	FCkey_set("TFORM1",  tform1); /* 1st field: index        */
	FCkey_set("TTYPE1",  ttype1);
	FCkey_set("TUNIT1",  tunit1);
	FCkey_set("TFORM2",  tform2); /* 2nd field: real a_lm    */
	FCkey_set("TTYPE2",  ttype2);
	FCkey_set("TUNIT2",  tunit2);
	FCkey_set("TFORM3",  tform3); /* 3d field:  imagine a_lm */
	FCkey_set("TTYPE3",  ttype3);
	FCkey_set("TUNIT3",  tunit3);

	FCkey_set("OBJECT",  "A_LM");
	FLkey_set("LMAX",    lmax_print);
	FCkey_set("ORIGIN",  "GLESP");

	FHupdate(fp_out);
	FHfree();

	for (bb=l=0; l<=lmax_print; l++,bb+=nbuf) {
		register int ll=l*l+l+1;
		for (m=0; m<=l; m++) {
			double v = ll+m;
			bitpix = bitp1;
			put_pix_fits(fp_out, v);
			bitpix = bitp2;
			v = alm [bb+m];
			put_pix_fits(fp_out, v);
			bitpix = bitp3;
			v = alm [bb+m+lmax+1];
			put_pix_fits(fp_out, v);
		}
	}

	block_data = nlm * nsize;
	if(block_data % BLOCK_SIZE != 0) {
	       register int i;
	       long len = BLOCK_SIZE * (block_data / BLOCK_SIZE + 1);
	       len -= block_data;
	       for(i=0; i<len; i++)   fputc('\0', fp_out);
	}

	return;
}

int Fget_lmax(filename)
char *filename;
{
	register int i, l, m;
	int tfields = 0;
	int bitp1, bitp2, bitp3;
	char tform1[30],
	     tform2[30],
	     tform3[30];
	char str[161];
	int  lmx = -1;
	FILE *fp;
	int naxis=0, naxis1=0, naxis2=0;

	/* open file */
	if ((fp = fopen(filename, "r")) == (FILE *)0) {
		PR(ER, "%s: can not open file '%s'\n", myname, filename);
		exit(1);
	}
	/* read 1st head */
	FHinit(fp);

	if (FEVget("EXTEND") != NULL)
	    strcpy(str, FEVget("EXTEND"));

	if (strcmp(str, "T") != 0) {
	    PR(ER, "%s: bad extension in file '%s'\n", myname, filename);
	    exit(1);
	}

	/* Free memory from 1st head */
	FHfree();

	/****************************************/
	/* read basic keywods from the 2nd head */
	/****************************************/
	FHinit(fp);

	if (FEVget("XTENSION") != NULL)
	    strcpy(str, FEVget("XTENSION"));
	else {
	    PR(ER, "%s: no extension in file '%s'\n", myname, filename);
	    exit(1);
	}
	if (strcmp(str, "BINTABLE") != 0) {
	    PR(ER, "%s: can work only with BINTABLE extension\n", myname);
	    exit(1);
	}
	if (FEVget("NAXIS") != NULL)
		naxis  = FSkey_get("NAXIS");
	if (FEVget("NAXIS1") != NULL)
		naxis1  = FLkey_get("NAXIS1");
	if (FEVget("NAXIS2") != NULL)
		naxis2  = FLkey_get("NAXIS2");

	if (FEVget("TFIELDS") != NULL)
		tfields = FLkey_get("TFIELDS");
	if (tfields != 3) {
	    PR(ER, "%s: TFILEDS value must be equal to 3 in a_lm file '%s'\n",
		myname, filename);
	    exit(1);
	}

	if (FEVget("TFORM1") != NULL) {
		strcpy(tform1, FEVget("TFORM1"));
		if (strcmp(tform1, "1J") == 0) {
		    bitp1 = LONGPIX;
		} else if (strcmp(tform1, "1I") == 0) {
		    bitp1 = SHORTPIX;
		} else {
		    PR(ER, "%s: bad value for TFORM1. Sorry...\n", myname);
		    exit(1);
		}
	} else {
		PR(ER, "%s: no parameter TFORM1. Sorry...\n", myname);
		exit(1);
	}
	if (FEVget("TFORM2") != NULL) {
		strcpy(tform2, FEVget("TFORM2"));
		if (strcmp(tform2, "1E") == 0) {
		    bitp2 = FLOATPIX;
		} else if (strcmp(tform2, "1D") == 0) {
		    bitp2 = DOUBLEPIX;
		} else {
		    PR(ER, "%s: bad value for TFORM2. Sorry...\n", myname);
		    exit(1);
		}
	} else {
		PR(ER, "%s: no parameter TFORM2. Sorry...\n", myname);
		exit(1);
	}
	if (FEVget("TFORM3") != NULL) {
		strcpy(tform3, FEVget("TFORM3"));
		if (strcmp(tform2, "1E") == 0) {
		    bitp3 = FLOATPIX;
		} else if (strcmp(tform3, "1D") == 0) {
		    bitp3 = DOUBLEPIX;
		} else {
		    PR(ER, "%s: bad value for TFORM3. Sorry...\n", myname);
		    exit(1);
		}
	} else {
		PR(ER, "%s: no parameter TFORM3. Sorry...\n", myname);
		exit(1);
	}
	/* Free memory from 1nd head */
	FHfree();

	/* get values from file */
	for (i=0; i<naxis2; i++) {
	    register int nlm;

	    bitpix = bitp1;
	    nlm = get_pix_fits(fp)+0.5;
	    l = (int)sqrt((double)(nlm-1));
	    m = nlm - 1 - l*l - l;

	    /* Real part */
	    bitpix = bitp2;
	    get_pix_fits(fp);
	    if (feof(fp)) {
		 PR(ER, "%s: unexpected end of file at l=%d and m=%d\n",
		    myname, l, m);
		 break;
	    }
	    /* Imagine part */
	    bitpix = bitp3;
	    get_pix_fits(fp);
	    if (feof(fp)) {
		 PR(ER, "%s: unexpected end of file at l=%d and m=%d\n",
		    myname, l, m);
		 break;
	    }
	    if (l > lmx)   lmx = l;
	}
	fclose(fp);
	return lmx;
}

/*
 * check file type:
 *      flag = -1       it is not our file
 *      flag = 0        F-file
 *      flag = -2       FITS-file, but not GLESP map
 *      flag = -3       ASCII-file
 *      flag = 1        GLESP FITS-file
 *      flag = 2        FITS-file alm-file
 *      flag = 3        GLESP FITS-file 3 maps (T,Q,U)
 *      flag = 4        FITS-file 3 alm-s (A,E,B)
 *      flag = 5        Hpix FITS-file T-map
 *      flag = 6        Hpix FITS-file 3 maps (T,Q,U)
 *      flag = 7        ASCII file with 3 parameters in a string
 *      flag = 8        2-D Basic FITS
 */
extern int test_ieb_alm(char *, int);

int test_glesp_map(file)
char *file;
{
	FILE *fp = (FILE *)0;
	char str[161];
	int flag = 0;

	/* open file */
	if ((fp = fopen(file, "r")) == (FILE *)0) {
	    PR(ER, "%s: can not open file '%s'\n", myname, file);
	    if ( errno == EOVERFLOW )
		PR(ER, "Too large value for defined data type...");
	    PR(ER, "(errno=%d)\n", errno);
	    exit(1);
	}

	fgets(str, 160, fp);
	if (strncmp(str, "SIMPLE  = ", 10) == 0)
		flag = 1;      /* FITS */
	else if (strncmp(str, "SIMPLE=", 7) == 0)
		flag = 0;      /* F */
	else
		flag = -1;   /* it is something else */

	if (flag == -1) {    /* it is something else */
	    register int i=0, k=0;
	    register char *cs = str;
	    while (*cs && *cs != '\n' && i<161) {
		i++;
		if (! isascii (*cs) ) { k=1; break; } /* k=1, when binary */
		cs++;
	    }
	    if (!k)
		flag = -3;   /* it is something else, in ASCII */
	}

	if (flag == -3) {
	    if (n_skip_wd(str) == 3) /* 3 parameters in a string */
		flag = 7;
	}

	if (flag == 1) {     /* it is FITS */
		fseek(fp, 0l, 0);
		/* read 1st head */
		FHinit(fp);
		/* read 2nd head */
		if (FEVget("EXTEND") != NULL) {
		    FHinit(fp);

		    if (flag == 1 &&
		      FEVget("TTYPE1") != NULL)  {
			strcpy(str, FEVget("TTYPE1"));
			if (strcmp(str, "INDEX") == 0)  {
			    flag = 2;  /* it's a-lm file */

			    if (test_ieb_alm(file, 0))
				flag = 4;  /* it's alm I,E,B-file */
			} else if (strcmp(str, "TEMPERATURE") == 0 ||
			       strncmp(str, "I_ST",3) == 0 ||
			       /* not "COS(THETA)" */
				*str != 'C' )  {
			    /* Check for Hpix-maps */
			    if (FEVget("PIXTYPE") != NULL) {
				strcpy(str, FEVget("PIXTYPE"));
				if (strcmp(str, "HEALPIX") == 0)  {
				     if (FEVget("TTYPE2")!= NULL) {
					strcpy(str, FEVget("TTYPE2"));
					if (strncmp(str,"Q-P",3)==0 ||
					    strncmp(str,"Q_P",3)==0 ||
					    strncmp(str,"Q_S",3)==0)
					   flag = 6;  /* 3-map in Hpix */
					else
					   flag = 5;  /* T-map in Hpix */
				     } else
					   flag = 5;  /* T-map in Hpix */
				} else
				     /* it is not our map  */
				     flag = -2;
			    } else
				flag = -2; /* it is not our map */
			} else if (strcmp(str, "COS(THETA)") != 0)
			    flag = -2; /* it is not a map */
		    } else
			    flag = -2; /* it is not a GLESP map */

		    if (flag == 1) { /* not '2' or '-2' */
			if (FEVget("TFIELDS") != NULL)  {
			    /* 3 maps (T,Q,U) file */
			    if (atoi(FEVget("TFIELDS")) == 5)
				flag = 3;
			}
		    }
		} else if (atoi(FEVget("NAXIS")) == 2 &&
		    atoi(FEVget("NAXIS1")) >1 && atoi(FEVget("NAXIS2")) >1) {
		    /* let it be Basic FITS */
		    flag = 8;
		} else
		    flag = -2; /* it is not a GLESP map */

		FHfree(); /* free memory */
	}

	fclose(fp);
	return flag;
}

void  add1_alm_ext(alm, lmax, nbuf, lmax_print, fp_out)
double *alm;
int lmax, lmax_print;
int nbuf;
FILE *fp_out;
{
	register int l, m, bb;
	int nlm = (lmax_print+1)*(lmax_print+2) / 2;
	int tfields = 3, nsize = 0;
	char ttype1[30], tform1[30], tunit1[30],
	     ttype2[30], tform2[30], tunit2[30],
	     ttype3[30], tform3[30], tunit3[30];
	int bitp1, bitp2, bitp3;
	int block_data = 0;

	strcpy(tform1, "1J");
	strcpy(ttype1,  "INDEX");
	strcpy(tunit1,  "UNITLESS");

	strcpy(tform2, "1E");
	strcpy(ttype2,  "REAL");
	strcpy(tunit2,  "mK");
	strcpy(tform3, "1E");
	strcpy(ttype3,  "IMAG");
	strcpy(tunit3,  "mK");

	/* TFORM 1 */
	bitp1=LONGPIX;
	if (strcmp(tform1, "1J") == 0) {
		nsize += 4;
	} else if (strcmp(tform1, "1I") == 0) {
		nsize += 2;
		sprintf(tform1, "1I");
		bitp1=SHORTPIX;
	}
	/* TFORM 2 */
	bitp2=bitpix_map;
	if (strcmp(tform2, "1E") == 0) {
		nsize += 4;
		bitp2=FLOATPIX;
	} else if (strcmp(tform2, "1D") == 0) {
		nsize += 8;
		bitp2=DOUBLEPIX;
	}
	/* TFORM 3 */
	bitp3=bitpix_map;
	if (strcmp(tform3, "1E") == 0) {
		nsize += 4;
		bitp2=FLOATPIX;
	} else if (strcmp(tform3, "1D") == 0) {
		nsize += 8;
		bitp3=DOUBLEPIX;
	}
	/***************************/
	/* first polarization mode */
	/***************************/
	FCkey_set("XTENSION",  "BINTABLE");
	FSkey_set("BITPIX",  8);
	FSkey_set("NAXIS",   2);
	FLkey_set("NAXIS1",  nsize);
	FLkey_set("NAXIS2",  nlm);
	FLkey_set("PCOUNT",  0);
	FLkey_set("GCOUNT",  1);

	FLkey_set("TFIELDS", tfields);
	FCkey_set("EXTNAME", "A_lm DATA");

	FCkey_set("TFORM1",  tform1); /* 1st field: index        */
	FCkey_set("TTYPE1",  ttype1);
	FCkey_set("TUNIT1",  tunit1);
	FCkey_set("TFORM2",  tform2); /* 2nd field: real a_lm    */
	FCkey_set("TTYPE2",  ttype2);
	FCkey_set("TUNIT2",  tunit2);
	FCkey_set("TFORM3",  tform3); /* 3d field:  imagine a_lm */
	FCkey_set("TTYPE3",  ttype3);
	FCkey_set("TUNIT3",  tunit3);

	FCkey_set("OBJECT",  "A_LM");
	FLkey_set("LMAX",    lmax_print);
	FCkey_set("ORIGIN",  "GLESP");

	FHupdate(fp_out);
	FHfree();

	for (bb=l=0; l<=lmax_print; l++,bb+=nbuf) {
		register int ll=l*l+l+1;
		for (m=0; m<=l; m++) {
			float v = ll+m;
			bitpix = bitp1;
			put_pix_fits(fp_out, v);
			bitpix = bitp2;
			v = alm [bb+m];
			put_pix_fits(fp_out, v);
			bitpix = bitp3;
			v = alm [bb+m+lmax+1];
			put_pix_fits(fp_out, v);
		}
	}

	block_data = nlm * nsize;
	if(block_data % BLOCK_SIZE != 0) {
	       register int i;
	       long len = BLOCK_SIZE * (block_data / BLOCK_SIZE + 1);
	       len -= block_data;
	       for(i=0; i<len; i++)   fputc('\0', fp_out);
	}


	return;
}
void  add_alm_ext(alm1, alm2, lmax, nbuf, lmax_print, fp_out)
double *alm1, *alm2;
int lmax, lmax_print;
int nbuf;
FILE *fp_out;
{
	register int l, m, bb;
	int nlm = (lmax_print+1)*(lmax_print+2) / 2;
	int tfields = 3, nsize = 0;
	char ttype1[30], tform1[30], tunit1[30],
	     ttype2[30], tform2[30], tunit2[30],
	     ttype3[30], tform3[30], tunit3[30];
	int bitp1, bitp2, bitp3;
	int block_data = 0;

	strcpy(tform1, "1J");
	strcpy(ttype1,  "INDEX");
	strcpy(tunit1,  "UNITLESS");

	strcpy(tform2, "1E");
	strcpy(ttype2,  "REAL");
	strcpy(tunit2,  "mK");
	strcpy(tform3, "1E");
	strcpy(ttype3,  "IMAG");
	strcpy(tunit3,  "mK");

	/* TFORM 1 */
	bitp1=LONGPIX;
	if (strcmp(tform1, "1J") == 0) {
		nsize += 4;
	} else if (strcmp(tform1, "1I") == 0) {
		nsize += 2;
		sprintf(tform1, "1I");
		bitp1=SHORTPIX;
	}
	/* TFORM 2 */
	bitp2=bitpix_map;
	if (strcmp(tform2, "1E") == 0) {
		nsize += 4;
		bitp2=FLOATPIX;
	} else if (strcmp(tform2, "1D") == 0) {
		nsize += 8;
		bitp2=DOUBLEPIX;
	}
	/* TFORM 3 */
	bitp3=bitpix_map;
	if (strcmp(tform3, "1E") == 0) {
		nsize += 4;
		bitp2=FLOATPIX;
	} else if (strcmp(tform3, "1D") == 0) {
		nsize += 8;
		bitp3=DOUBLEPIX;
	}
	/***************************/
	/* first polarization mode */
	/***************************/
	FCkey_set("XTENSION",  "BINTABLE");
	FSkey_set("BITPIX",  8);
	FSkey_set("NAXIS",   2);
	FLkey_set("NAXIS1",  nsize);
	FLkey_set("NAXIS2",  nlm);
	FLkey_set("PCOUNT",  0);
	FLkey_set("GCOUNT",  1);

	FLkey_set("TFIELDS", tfields);
	FCkey_set("EXTNAME", "A_lm DATA");

	FCkey_set("TFORM1",  tform1); /* 1st field: index        */
	FCkey_set("TTYPE1",  ttype1);
	FCkey_set("TUNIT1",  tunit1);
	FCkey_set("TFORM2",  tform2); /* 2nd field: real a_lm    */
	FCkey_set("TTYPE2",  ttype2);
	FCkey_set("TUNIT2",  tunit2);
	FCkey_set("TFORM3",  tform3); /* 3d field:  imagine a_lm */
	FCkey_set("TTYPE3",  ttype3);
	FCkey_set("TUNIT3",  tunit3);

	FCkey_set("OBJECT",  "A_LM");
	FLkey_set("LMAX",    lmax_print);
	FCkey_set("ORIGIN",  "GLESP");

	FHupdate(fp_out);
	FHfree();

	for (bb=l=0; l<=lmax_print; l++, bb+=nbuf) {
		register int ll=l*l+l+1;
		for (m=0; m<=l; m++) {
			float v = ll+m;
			bitpix = bitp1;
			put_pix_fits(fp_out, v);
			bitpix = bitp2;
			v = alm1 [bb+m];
			put_pix_fits(fp_out, v);
			bitpix = bitp3;
			v = alm1 [bb+m+lmax+1];
			put_pix_fits(fp_out, v);
		}
	}

	block_data = nlm * nsize;
	if(block_data % BLOCK_SIZE != 0) {
	       register int i;
	       long len = BLOCK_SIZE * (block_data / BLOCK_SIZE + 1);
	       len -= block_data;
	       for(i=0; i<len; i++)   fputc('\0', fp_out);
	}


	/***************************/
	/* second polarization mode */
	/***************************/
	FCkey_set("XTENSION",  "BINTABLE");
	FSkey_set("BITPIX",  8);
	FSkey_set("NAXIS",   2);
	FLkey_set("NAXIS1",  nsize);
	FLkey_set("NAXIS2",  nlm);
	FLkey_set("PCOUNT",  0);
	FLkey_set("GCOUNT",  1);

	FLkey_set("TFIELDS", tfields);
	FCkey_set("EXTNAME", "A_lm DATA");

	FCkey_set("TFORM1",  tform1); /* 1st field: index        */
	FCkey_set("TTYPE1",  ttype1);
	FCkey_set("TUNIT1",  tunit1);
	FCkey_set("TFORM2",  tform2); /* 2nd field: real a_lm    */
	FCkey_set("TTYPE2",  ttype2);
	FCkey_set("TUNIT2",  tunit2);
	FCkey_set("TFORM3",  tform3); /* 3d field:  imagine a_lm */
	FCkey_set("TTYPE3",  ttype3);
	FCkey_set("TUNIT3",  tunit3);

	FCkey_set("OBJECT",  "A_LM");
	FLkey_set("LMAX",    lmax_print);
	FCkey_set("ORIGIN",  "GLESP");

	FHupdate(fp_out);
	FHfree();

	for (bb=l=0; l<=lmax_print; l++, bb+=nbuf) {
		register int ll=l*l+l+1;
		for (m=0; m<=l; m++) {
			float v = ll+m;
			bitpix = bitp1;
			put_pix_fits(fp_out, v);
			bitpix = bitp2;
			v = alm2 [bb+m];
			put_pix_fits(fp_out, v);
			bitpix = bitp3;
			v = alm2 [bb+m+lmax+1];
			put_pix_fits(fp_out, v);
		}
	}

	block_data = nlm * nsize;
	if(block_data % BLOCK_SIZE != 0) {
	       register int i;
	       long len = BLOCK_SIZE * (block_data / BLOCK_SIZE + 1);
	       len -= block_data;
	       for(i=0; i<len; i++)   fputc('\0', fp_out);
	}

	return;
}

void read_hp_palm(alm, alm1, alm2, lmax, nbuf, filename)
double **alm, **alm1, **alm2;
int *lmax;
int *nbuf;
char *filename;
{
	register int i, l, m;
	int tfields = 0;
	int bitp1, bitp2, bitp3;
	char tform1[30],
	     tform2[30],
	     tform3[30];
	char str[161];
	int  lmx = -1, lmax_input = *lmax, nalm;
	FILE *fp;
	int naxis=0, naxis1=0, naxis2=0;
	int block_data = 0;

	/* detect maximum l */
	*lmax = Fget_lmax(filename);
	nalm = *lmax+2;
	*nbuf= 2*nalm;

	/* take memory */
	if ((*alm = (double *)calloc((unsigned)(*nbuf*nalm), sizeof **alm)) ==
	    (double *)0) {
nomem:
		PR(ER, "%s: no memory for arrays of %d elements...\n",
			myname, *nbuf*nalm);
		exit(1);
	}
	if ((*alm1 =(double*)calloc((unsigned)(*nbuf*nalm), sizeof **alm1))
	     == (double *)0) goto nomem;
	if ((*alm2 =(double*)calloc((unsigned)(*nbuf*nalm), sizeof **alm2))
	     == (double *)0) goto nomem;

	/* open file */
	if ((fp = fopen(filename, "r")) == (FILE *)0) {
		PR(ER, "%s: can not open file '%s'\n", myname, filename);
		exit(1);
	}
	/* read 1st head */
	FHinit(fp);

	if (FEVget("EXTEND") != NULL)
	    strcpy(str, FEVget("EXTEND"));

	if (strcmp(str, "T") != 0) {
	    PR(ER, "%s: bad extension in file '%s'\n", myname, filename);
	    exit(1);
	}

	/* Free memory from 1st head */
	FHfree();

	/****************************************/
	/* read basic keywods from the 2nd head */
	/****************************************/
	FHinit(fp);

	if (FEVget("XTENSION") != NULL)
	    strcpy(str, FEVget("XTENSION"));
	else {
	    PR(ER, "%s: no extension in file '%s'\n", myname, filename);
	    exit(1);
	}
	if (strcmp(str, "BINTABLE") != 0) {
	    PR(ER, "%s: can work only with BINTABLE extension\n", myname);
	    exit(1);
	}
	if (FEVget("NAXIS") != NULL)
		naxis  = FSkey_get("NAXIS");
	if (FEVget("NAXIS1") != NULL)
		naxis1  = FLkey_get("NAXIS1");
	if (FEVget("NAXIS2") != NULL)
		naxis2  = FLkey_get("NAXIS2");

	if (FEVget("TFIELDS") != NULL)
		tfields = FLkey_get("TFIELDS");
	if (tfields != 3) {
	    PR(ER, "%s: TFILEDS value must be equal to 3 in a_lm file '%s'\n",
		myname, filename);
	    exit(1);
	}

	if (FEVget("TFORM1") != NULL) {
		strcpy(tform1, FEVget("TFORM1"));
		if (strcmp(tform1, "1J") == 0) {
		    bitp1 = LONGPIX;
		} else if (strcmp(tform1, "1I") == 0) {
		    bitp1 = SHORTPIX;
		} else {
		    PR(ER, "%s: bad value for TFORM1. Sorry...\n", myname);
		    exit(1);
		}
	} else {
		PR(ER, "%s: no parameter TFORM1. Sorry...\n", myname);
		exit(1);
	}
	if (FEVget("TFORM2") != NULL) {
		strcpy(tform2, FEVget("TFORM2"));
		if (strcmp(tform2, "1E") == 0) {
		    bitp2 = FLOATPIX;
		} else if (strcmp(tform2, "1D") == 0) {
		    bitp2 = DOUBLEPIX;
		} else {
		    PR(ER, "%s: bad value for TFORM2. Sorry...\n", myname);
		    exit(1);
		}
	} else {
		PR(ER, "%s: no parameter TFORM2. Sorry...\n", myname);
		exit(1);
	}
	if (FEVget("TFORM3") != NULL) {
		strcpy(tform3, FEVget("TFORM3"));
		if (strcmp(tform2, "1E") == 0) {
		    bitp3 = FLOATPIX;
		} else if (strcmp(tform3, "1D") == 0) {
		    bitp3 = DOUBLEPIX;
		} else {
		    PR(ER, "%s: bad value for TFORM3. Sorry...\n", myname);
		    exit(1);
		}
	} else {
		PR(ER, "%s: no parameter TFORM3. Sorry...\n", myname);
		exit(1);
	}
	if (FEVget("LMAX") != NULL) {
		*lmax = atoi(FEVget("LMAX"));
	}

	if (lmax_input > 0 && lmax_input <= *lmax) {
	     *lmax = lmax_input;
	} else {
	     if (*nbuf < (*lmax + 1)*2) {
		 if ( (lmax_input + 1) * 2 > *nbuf)
		     lmax_input = *nbuf / 2 - 1;
		 *lmax = lmax_input;
//             } else {
//               if (*lmax > lmax_input) {
//                 PR(ER, "%s: check set of lmax: lmax=%d, nbuf=%d\n",
//                        myname, *lmax, *nbuf);
//                 exit(1);
//               }
	     }
	}

	/* Free memory from 1nd head */
	FHfree();

	/* get values from file */
	for (i=0; i<naxis2; i++) {
	    static int flag = 1;
	    register int bb, nlm;

	    bitpix = bitp1;
	    nlm = get_pix_fits(fp)+0.5;
	    l = (int)sqrt((double)(nlm-1));

//          if (nlm == 0) {
//              bitpix = bitp2;   get_pix_fits(fp);
//              bitpix = bitp3;   get_pix_fits(fp);
//              continue;
//          }
	    if (l > *lmax) {
		bitpix = bitp2;   get_pix_fits(fp);
		bitpix = bitp3;   get_pix_fits(fp);
		continue;
	    }

	    if (l*2+1 > *nbuf) {
		if (flag) {
		    PR(ER, "%s: attention ! l(=%d) is larger ", myname, l);
		    PR(ER, "than array dimension (2*l+1>nbuf(%d)).\n", *nbuf);
		}
		flag = 0;
		break;
	    }
	    m = nlm - 1 - l*l - l;
	    bb = l * *nbuf;

	    /* Real part */
	    bitpix = bitp2;
	    (*alm) [bb+m]        = get_pix_fits(fp);
	    if (feof(fp)) {
		 PR(ER, "%s: unexpected end of file at l=%d and m=%d\n",
		    myname, l, m);
		 break;
	    }
	    /* Imagine part */
	    bitpix = bitp3;
	    (*alm) [bb+m+ *lmax+1] = get_pix_fits(fp);
	    if (feof(fp)) {
		 PR(ER, "%s: unexpected end of file at l=%d and m=%d\n",
		    myname, l, m);
		 break;
	    }
	    if (l > lmx)   lmx = l;
	}
	if (*lmax <= 0)  *lmax = lmx;

	block_data = naxis1*naxis2;
	if(block_data % BLOCK_SIZE != 0) {
	       register int i;
	       long len = BLOCK_SIZE * (block_data / BLOCK_SIZE + 1);
	       len -= block_data;
	       for(i=0; i<len; i++)   fgetc(fp);
	}

	/**********************/
	/* read 2nd extension */
	/**********************/
	FHinit(fp);
	if (FEVget("XTENSION") != NULL)
	    strcpy(str, FEVget("XTENSION"));
	else {
	    PR(ER, "%s: no 2nd extension in file '%s'\n", myname, filename);
	    exit(1);
	}
	if (strcmp(str, "BINTABLE") != 0) {
	    PR(ER, "%s: can work only with BINTABLE extension\n", myname);
	    exit(1);
	}
	if (FEVget("NAXIS") != NULL)
		naxis  = FSkey_get("NAXIS");
	if (FEVget("NAXIS1") != NULL)
		naxis1  = FLkey_get("NAXIS1");
	if (FEVget("NAXIS2") != NULL)
		naxis2  = FLkey_get("NAXIS2");

	if (FEVget("TFIELDS") != NULL)
		tfields = FLkey_get("TFIELDS");
	if (tfields != 3) {
	    PR(ER, "%s: TFILEDS value must be equal to 3 in a_lm file '%s'\n",
		myname, filename);
	    exit(1);
	}

	if (FEVget("TFORM1") != NULL) {
		strcpy(tform1, FEVget("TFORM1"));
		if (strcmp(tform1, "1J") == 0) {
		    bitp1 = LONGPIX;
		} else if (strcmp(tform1, "1I") == 0) {
		    bitp1 = SHORTPIX;
		} else {
		    PR(ER, "%s: bad value for TFORM1. Sorry...\n", myname);
		    exit(1);
		}
	} else {
		PR(ER, "%s: no parameter TFORM1. Sorry...\n", myname);
		exit(1);
	}
	if (FEVget("TFORM2") != NULL) {
		strcpy(tform2, FEVget("TFORM2"));
		if (strcmp(tform2, "1E") == 0) {
		    bitp2 = FLOATPIX;
		} else if (strcmp(tform2, "1D") == 0) {
		    bitp2 = DOUBLEPIX;
		} else {
		    PR(ER, "%s: bad value for TFORM2. Sorry...\n", myname);
		    exit(1);
		}
	} else {
		PR(ER, "%s: no parameter TFORM2. Sorry...\n", myname);
		exit(1);
	}
	if (FEVget("TFORM3") != NULL) {
		strcpy(tform3, FEVget("TFORM3"));
		if (strcmp(tform2, "1E") == 0) {
		    bitp3 = FLOATPIX;
		} else if (strcmp(tform3, "1D") == 0) {
		    bitp3 = DOUBLEPIX;
		} else {
		    PR(ER, "%s: bad value for TFORM3. Sorry...\n", myname);
		    exit(1);
		}
	} else {
		PR(ER, "%s: no parameter TFORM3. Sorry...\n", myname);
		exit(1);
	}
	if (FEVget("LMAX") != NULL) {
		*lmax = atoi(FEVget("LMAX"));
	}

	if (lmax_input > 0 && lmax_input <= *lmax) {
	     *lmax = lmax_input;
	} else {
	     if (*nbuf < (*lmax + 1)*2) {
		 if ( (lmax_input + 1) * 2 > *nbuf)
		     lmax_input = *nbuf / 2 - 1;
		 *lmax = lmax_input;
//             } else {
//                if (*lmax > lmax_input) {
//                  PR(ER, "%s: check set of lmax: lmax=%d, nbuf=%d\n",
//                         myname, *lmax, *nbuf);
//                  exit(1);
//                }
	     }
	}

	/* Free memory from 1nd head */
	FHfree();

	/* get values from file */
	for (i=0; i<naxis2; i++) {
	    static int flag = 1;
	    register int bb, nlm;

	    bitpix = bitp1;
	    nlm = get_pix_fits(fp)+0.5;
	    l = (int)sqrt((double)(nlm-1));

	    if (l > *lmax) {
		bitpix = bitp2;   get_pix_fits(fp);
		bitpix = bitp3;   get_pix_fits(fp);
		continue;
	    }

	    if (l*2+1 > *nbuf) {
		if (flag) {
		    PR(ER, "%s: attention ! l(=%d) is larger ", myname, l);
		    PR(ER, "than array dimension (2*l+1>nbuf(%d)).\n", *nbuf);
		}
		flag = 0;
		break;
	    }
	    m = nlm - 1 - l*l - l;
	    bb = l * *nbuf;

	    /* Real part */
	    bitpix = bitp2;
	    (*alm1) [bb+m]        = get_pix_fits(fp);
	    if (feof(fp)) {
		 PR(ER, "%s: unexpected end of file at l=%d and m=%d\n",
		    myname, l, m);
		 break;
	    }
	    /* Imagine part */
	    bitpix = bitp3;
	    (*alm1) [bb+m+ *lmax+1] = get_pix_fits(fp);
	    if (feof(fp)) {
		 PR(ER, "%s: unexpected end of file at l=%d and m=%d\n",
		    myname, l, m);
		 break;
	    }
	    if (l > lmx)   lmx = l;
	}
	if (*lmax <= 0)  *lmax = lmx;

	block_data = naxis1*naxis2;
	if(block_data % BLOCK_SIZE != 0) {
	       register int i;
	       long len = BLOCK_SIZE * (block_data / BLOCK_SIZE + 1);
	       len -= block_data;
	       for(i=0; i<len; i++)   fgetc(fp);
	}

	/*********************/
	/* read 3d extension */
	/*********************/
	FHinit(fp);
	if (FEVget("XTENSION") != NULL)
	    strcpy(str, FEVget("XTENSION"));
	else {
	    PR(ER, "%s: no 3d extension in file '%s'\n", myname, filename);
	    exit(1);
	}
	if (strcmp(str, "BINTABLE") != 0) {
	    PR(ER, "%s: can work only with BINTABLE extension\n", myname);
	    exit(1);
	}
	if (FEVget("NAXIS") != NULL)
		naxis  = FSkey_get("NAXIS");
	if (FEVget("NAXIS1") != NULL)
		naxis1  = FLkey_get("NAXIS1");
	if (FEVget("NAXIS2") != NULL)
		naxis2  = FLkey_get("NAXIS2");

	if (FEVget("TFIELDS") != NULL)
		tfields = FLkey_get("TFIELDS");
	if (tfields != 3) {
	    PR(ER, "%s: TFILEDS value must be equal to 3 in a_lm file '%s'\n",
		myname, filename);
	    exit(1);
	}

	if (FEVget("TFORM1") != NULL) {
		strcpy(tform1, FEVget("TFORM1"));
		if (strcmp(tform1, "1J") == 0) {
		    bitp1 = LONGPIX;
		} else if (strcmp(tform1, "1I") == 0) {
		    bitp1 = SHORTPIX;
		} else {
		    PR(ER, "%s: bad value for TFORM1. Sorry...\n", myname);
		    exit(1);
		}
	} else {
		PR(ER, "%s: no parameter TFORM1. Sorry...\n", myname);
		exit(1);
	}
	if (FEVget("TFORM2") != NULL) {
		strcpy(tform2, FEVget("TFORM2"));
		if (strcmp(tform2, "1E") == 0) {
		    bitp2 = FLOATPIX;
		} else if (strcmp(tform2, "1D") == 0) {
		    bitp2 = LONGPIX;
		} else {
		    PR(ER, "%s: bad value for TFORM2. Sorry...\n", myname);
		    exit(1);
		}
	} else {
		PR(ER, "%s: no parameter TFORM2. Sorry...\n", myname);
		exit(1);
	}
	if (FEVget("TFORM3") != NULL) {
		strcpy(tform3, FEVget("TFORM3"));
		if (strcmp(tform2, "1E") == 0) {
		    bitp3 = FLOATPIX;
		} else if (strcmp(tform3, "1D") == 0) {
		    bitp3 = LONGPIX;
		} else {
		    PR(ER, "%s: bad value for TFORM3. Sorry...\n", myname);
		    exit(1);
		}
	} else {
		PR(ER, "%s: no parameter TFORM3. Sorry...\n", myname);
		exit(1);
	}
	if (FEVget("LMAX") != NULL) {
		*lmax = atoi(FEVget("LMAX"));
	}

	if (lmax_input > 0 && lmax_input <= *lmax) {
	     *lmax = lmax_input;
	} else {
	     if (*nbuf < (*lmax + 1)*2) {
		 if ( (lmax_input + 1) * 2 > *nbuf)
		     lmax_input = *nbuf / 2 - 1;
		 *lmax = lmax_input;
//             } else {
//               if (*lmax > lmax_input) {
//                 PR(ER, "%s: check set of lmax: lmax=%d, nbuf=%d\n",
//                        myname, *lmax, *nbuf);
//                 exit(1);
//               }
	     }
	}

	/* Free memory from 1nd head */
	FHfree();

	/* get values from file */
	for (i=0; i<naxis2; i++) {
	    static int flag = 1;
	    register int bb, nlm;

	    bitpix = bitp1;
	    nlm = get_pix_fits(fp)+0.5;
	    l = (int)sqrt((double)(nlm-1));

	    if (l > *lmax) {
		bitpix = bitp2;   get_pix_fits(fp);
		bitpix = bitp3;   get_pix_fits(fp);
		continue;
	    }

	    if (l*2+1 > *nbuf) {
		if (flag) {
		    PR(ER, "%s: attention ! l(=%d) is larger ", myname, l);
		    PR(ER, "than array dimension (2*l+1>nbuf(%d)).\n", *nbuf);
		}
		flag = 0;
		break;
	    }
	    m = nlm - 1 - l*l - l;
	    bb = l * *nbuf;

	    /* Real part */
	    bitpix = bitp2;
	    (*alm2) [bb+m]        = get_pix_fits(fp);
	    if (feof(fp)) {
		 PR(ER, "%s: unexpected end of file at l=%d and m=%d\n",
		    myname, l, m);
		 break;
	    }
	    /* Imagine part */
	    bitpix = bitp3;
	    (*alm2) [bb+m+ *lmax+1] = get_pix_fits(fp);
	    if (feof(fp)) {
		 PR(ER, "%s: unexpected end of file at l=%d and m=%d\n",
		    myname, l, m);
		 break;
	    }
	    if (l > lmx)   lmx = l;
	}
	if (*lmax <= 0)  *lmax = lmx;

	fclose(fp);
	return;
}


char *get_date(date_str_glesp)
char *date_str_glesp;
{
    strcpy(date_str_glesp, "01/12/11");
    return date_str_glesp;
}

void write_2d_fits(fp, f, nax1, nax2, crv1, crv2, cdlt1, cdlt2, ctp1, ctp2,
		bsc, bunit)
FILE *fp;
double *f;
int nax1, nax2;
double crv1, crv2;
double cdlt1, cdlt2;
char *ctp1, *ctp2, *bunit;
double bsc;
{
	register int i, j;
	char date_str[81];
	long length = 0;

	bitpix = FLOATPIX;
	Fckey_set("SIMPLE",    "T");
	FSkey_set("BITPIX",    bitpix);
	FSkey_set("NAXIS",     2);
	FLkey_set("NAXIS1",    (long)nax1);
	FLkey_set("NAXIS2",    (long)nax2);
	FDkey_set("CRVAL1",    (double)crv1);
	FDkey_set("CDELT1",    (double)cdlt1);
	FDkey_set("CRPIX1",    1.);
	FCkey_set("CTYPE1",    ctp1);
	FCkey_set("CUNIT1",    "deg");
	FDkey_set("CRVAL2",    (double)crv2);
	FDkey_set("CDELT2",    (double)cdlt2);
	FDkey_set("CRPIX2",    1.);
	FCkey_set("CTYPE2",    ctp2);
	FCkey_set("CUNIT2",    "deg");
	FDkey_set("EQUINOX",   2000.00);
	FCkey_set("BUNIT",     bunit);
	FDkey_set("BSCALE",    bsc);
	FSkey_set("BZERO",     0.);
	FCkey_set("OBJECT",    "map cut image");
	get_date(date_str);
	FCkey_set("DATE",      date_str);
	FCkey_set("HISTORY",   "GLESP mapcut");

	length = FHupdate(fp);

	for(i=0; i<nax1; i++) {
	    int ii = nax2 * i;
	    for(j=0; j<nax2; j++) {
		   put_pix_fits(fp, f[ii+j]/bsc);
	    }
	}
	length += (long)nax1 * (long)nax2 * ABS(bitpix) / 8;

	/* fill FITS-file to proportional to BLOCK_SIZE */
	if(length % BLOCK_SIZE != 0) {
	    long len = BLOCK_SIZE * ( length / BLOCK_SIZE + 1);
	    len -= length;
	    for(i=0; i<len; i++)   fputc('\0', fp);
	}


	return;
}

void glesp_2_map2d(f, x, nx, nphi, nph, fp_out)
double *f, *x;
int nx, nphi;
int *nph;
FILE *fp_out;
{
    register int i, j;
    double th0 = acos(x[0]);
    double dth = (acos(x[nx-1]) - th0) / (double)nx;
    double dphi = 2. * M_PI / nphi;
    double rad2ang = 180. / M_PI;
    double bsc = 1.0000;
    char bunit[30];

    bitpix = FLOATPIX;
    strcpy(bunit, "K");

    Fckey_set("SIMPLE",    "T");
    FSkey_set("BITPIX",    bitpix);
    FSkey_set("NAXIS",     2);
    FLkey_set("NAXIS1",    (long)nphi);
    FLkey_set("NAXIS2",    (long)nx);
    FDkey_set("CRVAL1",    th0*rad2ang);
    FDkey_set("CDELT1",    dth*rad2ang);
    FDkey_set("CRPIX1",    1.);
    FCkey_set("CTYPE1",    "GLON");
    FDkey_set("CRVAL2",    0.);
    FDkey_set("CDELT2",    dphi*rad2ang);
    FDkey_set("CRPIX2",    1.);
    FCkey_set("CTYPE2",    "GLAT");
    FDkey_set("EQUINOX",   2000.00);
    FCkey_set("BUNIT",     bunit);
    FDkey_set("BSCALE",    bsc);
    FCkey_set("OBJECT",    "CMB cut");

    FHupdate(fp_out);

    for(i=0; i<nx; i++) {
	double th = i * dth;
	for(j=0; j<nphi; j++) {
	       double phi = dphi * j;
	       register int nn = ang2n(th, phi, x, nx, nphi, nph);
	       float v = f[nn]/bsc;
	       put_pix_fits(fp_out, v);
	}
    }

    return;
}

void fits_write_map(f, nbuf, x, nx, phi, nphi, flag_grid, transp_flag, fp_out)
double *f;     /* 2-dim map                           */
int nbuf;      /* 2-dim factor: f(i*nbuf+j)<->f[i][j] */
double *x;     /* argument, x = cos theta             */
int nx;        /* number of elements in x-array       */
double *phi;   /* array of phi (Y-axis)               */
int nphi;      /* number of elements in phi-array     */
int flag_grid; /* <0 - dx=2/(Nx-1), >0 - special grid */
int transp_flag; /* transposing flag                  */
FILE *fp_out;
{
	register double *pd = f, *f_new = (double *)0;
	int  nx_new = nx;
	double x0=x[0], dx = (x[nx-1] - x[0]) / (double)(nx-1);
	register int i,j;
	int condition_flag = 1; /* for spline x0<x1<x2... */
	double bscale, b_zero, crval1,cdelt1,crpix1,crval2,cdelt2,crpix2;
	long naxis1, naxis2;
	int naxis;
	char bunit[30], object[30], ctype1[30], ctype2[30];

	if (flag_grid > 0) {
	    /***************************************/
	    /* if special grid we do recalculation */
	    /***************************************/
	    if ((f_new = (double *)calloc(nphi*nx, sizeof *f_new)) ==
		(double *)0) {
nomem:
		    PR(ER,"%s: no memory for output arrays.\n", myname);
		    exit(1);
	    }

	    if (theta_grid == 2) {
		/* recalculate from cos(theta)-grid to theta-grid */
		x0 = acos(x[0]);
		dx = (acos(x[nx-1]) - x0) / (double)nx;
		for(i=0; i<nx; i++)
			x[i] = acos(x[i]);
	    }
	    if (x[0] > x[1]) {
		condition_flag = 0;
		for(i=0; i<nx; i++)  x[i] = -x[i];
		x0 = -x0; dx = -dx;
	    }

	    /*-------------------------------------------*/
	    /* spline interpolation (modified input):    */
	    /*   non-uniform step by X ---> uniform step */
	    /*-------------------------------------------*/
	    intrpl2sm_nu_u(f, nx, nbuf, f_new,nx_new,nphi, x, x0, dx, nphi);

	    if (condition_flag == 0) {
		condition_flag = 1;
		for(i=0; i<nx; i++)  x[i] = -x[i];
		x0 = -x0; dx = -dx;
	    }

	    nbuf = nphi;
	    pd = f_new;
	}

	if (theta_grid == 1) {
	    /* recalculate from cos(theta)-grid to theta-grid */
	    x0 =  acos(x[0]);
	    dx = (acos(x[nx-1]) - x0) / (double)nx;
	    for(i=0; i<nx; i++)
		    x[i] = acos(x[i]);

	    if (x[0] > x[1]) {
		condition_flag = 0;
		for(i=0; i<nx; i++)  x[i] = -x[i];
		x0 = -x0; dx = -dx;
	    }

	    i= nphi * (nx > nx_new ? nx : nx_new);
	    /* if special grid we do recalculation */
	    if ((f_new = (double *)calloc(i, sizeof *f_new)) == (double *)0)
		    goto nomem;

	    /*-------------------------------------------*/
	    /* spline interpolation (modified input):    */
	    /*   non-uniform step by X ---> uniform step */
	    /*-------------------------------------------*/
	    intrpl2sm_nu_u(f, nx, nx, f_new, nx_new, nx_new, x, x0, dx,nphi);

	    if (condition_flag == 0) {
		condition_flag = 1;
		for(i=0; i<nx; i++)  x[i] = -x[i];
		x0 = -x0; dx = -dx;
	    }

	    nbuf = nphi;
	    pd = f_new;
	    ph->grid = GRID_UNIFORM_THETA;
	} else
	    ph->grid = GRID_UNIFORM_COS;

	/* just to be sure there nothing in memory */
	Fckey_set("SIMPLE",  "T");
	FHfree();

	/*****************/
	/* basic keywods */
	/*****************/
	bitpix = FLOATPIX;
	bscale = 1.000000000000;
	b_zero = 0.000000000000;
	naxis  = 2;

	/* just to save temporal value */
	if (theta_grid) {
		strcpy(bunit, "THETA");
	} else {
		strcpy(bunit, "COS(THETA)");
	}
	/**************************************/
	/* change axis if we need transposing */
	/**************************************/
	if ( transp_flag ) {
		naxis1 = nx;
		naxis2 = nphi;

		crval1 = x0;
		cdelt1 = dx;
		crpix1 = 1.;
		strcpy(ctype1, bunit);

		crval2 = phi[0];
		cdelt2 = (phi[nphi-1] - phi[0]) / (double)nphi;
		crpix2 = 1.;
		strcpy(ctype2, "PHI");
	} else {
		naxis2 = nx;
		naxis1 = nphi;

		crval1 = phi[0];
		cdelt1 = (phi[nphi-1] - phi[0]) / (double)nphi;
		crpix1 = 1.;
		strcpy(ctype1, "PHI");

		crval2 = x0;
		cdelt2 = dx;
		crpix2 = 1.;
		strcpy(ctype2, bunit);
	}
	strcpy(bunit,  "K");
	strcpy(object, "CMB map");

	/************/
	/* set head */
	/************/
	Fckey_set("SIMPLE",  "T");
	FSkey_set("BITPIX",  bitpix);
	FSkey_set("NAXIS",   naxis);
	FLkey_set("NAXIS1",  naxis1); /* transp_flag=0:|nx  |=1:|nphi */
	FLkey_set("NAXIS2",  naxis2); /*               |nphi|   |nx   */
	FCkey_set("BUNIT",   bunit);
	FDkey_set("BSCALE",  bscale);
	FDkey_set("BZERO",   b_zero);

	FCkey_set("CTYPE1",  ctype1);
	FDkey_set("CRVAL1",  crval1);
	FDkey_set("CRPIX1",  crpix1);
	FDkey_set("CDELT1",  cdelt1);

	FCkey_set("CTYPE2",  ctype2);
	FDkey_set("CRVAL2",  crval2);
	FDkey_set("CRPIX2",  crpix2);
	FDkey_set("CDELT2",  cdelt2);

	FCkey_set("OBJECT",  object);
	FCkey_set("ORIGIN",  "GLESP");

	switch(ph->grid) {
	case  GRID_LEGENDRE:
		FCkey_set("GRID",    "LEGENDRE");
		break;
	case  GRID_UNIFORM_COS:
		FCkey_set("GRID",    "UNIFORM COS");
		break;
	case  GRID_UNIFORM_THETA:
		FCkey_set("GRID",    "THETA UNIFORM");
		break;
	default :
		FCkey_set("GRID",    "NOT SET");
		break;
	}

	FHupdate(fp_out);

	if (  transp_flag ) {
/*
Good figure (circles): what's going on ?
such output, but nax1 and nax2 change each other... (nax2 = 2*nax1)
*/
	    for(j=0; j<nphi; j++) {
		for(i=0; i<nx_new; i++) {
			register int k = i * nbuf;
			float v = (float)((pd[k+j] - b_zero)/bscale);
			put_pix_fits(fp_out, v);
		}
	    }
	} else { /* write by standard way */
	    for(i=0; i<nx_new; i++) {
		register int k = (nx_new-i-1) * nbuf;
		for(j=0; j<nphi; j++) {
			double d = pd[k+(nphi-1-j)];
			float v = (float)((d - b_zero)/bscale);
			put_pix_fits(fp_out, v);
		}
	    }
	}
	if ( f_new != (double *)0 )  free ((void *)f_new);

	return;
}

/*
 * read alm, and e- & b- polarization modes
 */
void read_ieb_alm(alm_i, alm_e, alm_b, lmax, nbuf, filename)
double **alm_i, **alm_e, **alm_b;
int *lmax;
int *nbuf;      /* if ( *nbuf < 0)  *nbuf = -(*nbuf); */
char *filename;
{
    register int i, l, m;
    int tfields = 0;
    int bitp1, bitp2, bitp3;
    char tform1[30],
	 tform2[30],
	 tform3[30];
    char str[161];
    int  lmx = -1, lmax_input = *lmax, lmax_data, nalm;
    FILE *fp;
    int naxis=0, naxis1=0, naxis2=0;
    int block_data = 0;

//system("free");
//PR(ER, "rrrr1----------------------------------\n");
    /* detect maximum l */
    lmax_data = *lmax = Fget_lmax(filename);

    nalm = 2 * *lmax;
    if ( *nbuf < 0)  *nbuf = -(*nbuf); else *nbuf = 2 * nalm;

    /* take memory */
    if( *alm_i == (double*)0) {
	if((*alm_i=(double*)calloc((unsigned)(*nbuf*nalm),sizeof **alm_i))==
	    (double *)0) {
nomem:
		PR(ER, "%s: no memory for arrays of %d elements...\n",
			myname, *nbuf*nalm);
		exit(1);
	}
    }
    if( *alm_e == (double*)0) {
	if((*alm_e =(double*)calloc((unsigned)(*nbuf*nalm), sizeof **alm_e))
	     == (double *)0) goto nomem;
    }
    if( *alm_b == (double*)0) {
	if((*alm_b =(double*)calloc((unsigned)(*nbuf*nalm), sizeof **alm_b))
	     == (double *)0) goto nomem;
    }

    /* open file */
    if ((fp = fopen(filename, "r")) == (FILE *)0) {
	    PR(ER, "%s: can not open file '%s'\n", myname, filename);
	    exit(1);
    }
    /* read 1st head */
    FHinit(fp);

    if (FEVget("EXTEND") != NULL)
	strcpy(str, FEVget("EXTEND"));

    if (strcmp(str, "T") != 0) {
	PR(ER, "%s: bad extension in file '%s'\n", myname, filename);
	exit(1);
    }

    /* Free memory from 1st head */
    FHfree();

    /****************************************/
    /* read basic keywods from the 2nd head */
    /****************************************/
    FHinit(fp);

    if (FEVget("XTENSION") != NULL)
	strcpy(str, FEVget("XTENSION"));
    else {
	PR(ER, "%s: no extension in file '%s'\n", myname, filename);
	exit(1);
    }
    if (strcmp(str, "BINTABLE") != 0) {
	PR(ER, "%s: can work only with BINTABLE extension\n", myname);
	exit(1);
    }
    if (FEVget("NAXIS") != NULL)
	    naxis  = FSkey_get("NAXIS");
    if (FEVget("NAXIS1") != NULL)
	    naxis1  = FLkey_get("NAXIS1");
    if (FEVget("NAXIS2") != NULL)
	    naxis2  = FLkey_get("NAXIS2");

    if (FEVget("TFIELDS") != NULL)
	    tfields = FLkey_get("TFIELDS");
    if (tfields != 3) {
	PR(ER, "%s: TFILEDS value must be equal to 3 in a_lm file '%s'\n",
	    myname, filename);
	exit(1);
    }

    if (FEVget("TFORM1") != NULL) {
	    strcpy(tform1, FEVget("TFORM1"));
	    if (strcmp(tform1, "1J") == 0) {
		bitp1 = LONGPIX;
	    } else if (strcmp(tform1, "1I") == 0) {
		bitp1 = SHORTPIX;
	    } else {
		PR(ER, "%s: bad value for TFORM1. Sorry...\n", myname);
		exit(1);
	    }
    } else {
	    PR(ER, "%s: no parameter TFORM1. Sorry...\n", myname);
	    exit(1);
    }
    if (FEVget("TFORM2") != NULL) {
	    strcpy(tform2, FEVget("TFORM2"));
	    if (strcmp(tform2, "1E") == 0) {
		bitp2 = FLOATPIX;
	    } else if (strcmp(tform2, "1D") == 0) {
		bitp2 = LONGPIX;
	    } else {
		PR(ER, "%s: bad value for TFORM2. Sorry...\n", myname);
		exit(1);
	    }
    } else {
	    PR(ER, "%s: no parameter TFORM2. Sorry...\n", myname);
	    exit(1);
    }
    if (FEVget("TFORM3") != NULL) {
	    strcpy(tform3, FEVget("TFORM3"));
	    if (strcmp(tform2, "1E") == 0) {
		bitp3 = FLOATPIX;
	    } else if (strcmp(tform3, "1D") == 0) {
		bitp3 = LONGPIX;
	    } else {
		PR(ER, "%s: bad value for TFORM3. Sorry...\n", myname);
		exit(1);
	    }
    } else {
	    PR(ER, "%s: no parameter TFORM3. Sorry...\n", myname);
	    exit(1);
    }
    if (FEVget("LMAX") != NULL) {
	    *lmax = atoi(FEVget("LMAX"));
    }

    if (lmax_input > 0 && lmax_input <= *lmax) {
	 *lmax = lmax_input;
    } else {
	 if (*nbuf < (*lmax + 1)*2) {
	     if ( (lmax_input + 1) * 2 > *nbuf)
		 lmax_input = *nbuf / 2 - 1;
	     *lmax = lmax_input;
//         } else {
//           if (*lmax > lmax_input) {
//             PR(ER, "%s: check set of lmax: lmax=%d, nbuf=%d\n",
//                    myname, *lmax, *nbuf);
//             exit(1);
//           }
	 }
    }

    /* Free memory from 1nd head */
    FHfree();

    /* get values from file */
      for (i=0; i<naxis2; i++) {
	static int flag = 1;
	register int bb, nlm;

	bitpix = bitp1;
	nlm = get_pix_fits(fp)+0.5;
	l = (int)sqrt((double)(nlm-1));

	if (l > *lmax)
	    break;

	if (l*2+1 > *nbuf) {
	    if (flag) {
		PR(ER, "%s: attention ! l(=%d) is larger ", myname, l);
		PR(ER, "than array dimension (2*l+1>nbuf(%d)).\n", *nbuf);
	    }
	    flag = 0;
	    break;
	}
	m = nlm - 1 - l*l - l;
	bb = l * *nbuf;

	if (m < 0) {
	    PR(ER, "%s: attention ! bad number of m=%d for l=%d.\n",
		  myname, m, l);
	    exit(1);
	}
	if (m > l) {
	    PR(ER, "%s: attention ! m=%d > l=%d.\n", myname, m, l);
	    exit(1);
	}


	/* Real part */
	bitpix = bitp2;
	(*alm_i) [bb+m]        = get_pix_fits(fp);
	if (feof(fp)) {
	     PR(ER, "%s: unexpected end of file at l=%d and m=%d\n",
		myname, l, m);
	     break;
	}
	/* Imaginary part */
	bitpix = bitp3;
	(*alm_i) [bb+m+ *lmax+1] = get_pix_fits(fp);
	if (feof(fp)) {
	     PR(ER, "%s: unexpected end of file at l=%d and m=%d\n",
		myname, l, m);
	     break;
	}
	if (l > lmx)   lmx = l;
    }
    if (*lmax <= 0)  *lmax = lmx;

    /* skip remained */
    if (*lmax < lmax_data) {
	bitpix = bitp2; get_pix_fits(fp);
	bitpix = bitp3; get_pix_fits(fp);
	i++;
	for (; i<naxis2; i++) {
	    bitpix = bitp1; get_pix_fits(fp);
	    bitpix = bitp2; get_pix_fits(fp);
	    bitpix = bitp3; get_pix_fits(fp);
	}
    }

    block_data = naxis1*naxis2;
    if(block_data % BLOCK_SIZE != 0) {
	   register int i;
	   long len = BLOCK_SIZE * (block_data / BLOCK_SIZE + 1);
	   len -= block_data;
	   for(i=0; i<len; i++)   fgetc(fp);
    }


    /**********************/
    /* read 2nd extension */
    /**********************/
    FHinit(fp);
    if (FEVget("XTENSION") != NULL)
	strcpy(str, FEVget("XTENSION"));
    else {
	PR(ER, "%s: no 2nd extension in file '%s'\n", myname, filename);
	exit(1);
    }
    if (strcmp(str, "BINTABLE") != 0) {
	PR(ER, "%s: can work only with BINTABLE extension\n", myname);
	exit(1);
    }
    if (FEVget("NAXIS") != NULL)
	    naxis  = FSkey_get("NAXIS");
    if (FEVget("NAXIS1") != NULL)
	    naxis1  = FLkey_get("NAXIS1");
    if (FEVget("NAXIS2") != NULL)
	    naxis2  = FLkey_get("NAXIS2");

    if (FEVget("TFIELDS") != NULL)
	    tfields = FLkey_get("TFIELDS");
    if (tfields != 3) {
	PR(ER, "%s: TFILEDS value must be equal to 3 in a_lm file '%s'\n",
	    myname, filename);
	exit(1);
    }

    if (FEVget("TFORM1") != NULL) {
	    strcpy(tform1, FEVget("TFORM1"));
	    if (strcmp(tform1, "1J") == 0) {
		bitp1 = LONGPIX;
	    } else if (strcmp(tform1, "1I") == 0) {
		bitp1 = SHORTPIX;
	    } else {
		PR(ER, "%s: bad value for TFORM1. Sorry...\n", myname);
		exit(1);
	    }
    } else {
	    PR(ER, "%s: no parameter TFORM1. Sorry...\n", myname);
	    exit(1);
    }
    if (FEVget("TFORM2") != NULL) {
	    strcpy(tform2, FEVget("TFORM2"));
	    if (strcmp(tform2, "1E") == 0) {
		bitp2 = FLOATPIX;
	    } else if (strcmp(tform2, "1D") == 0) {
		bitp2 = LONGPIX;
	    } else {
		PR(ER, "%s: bad value for TFORM2. Sorry...\n", myname);
		exit(1);
	    }
    } else {
	    PR(ER, "%s: no parameter TFORM2. Sorry...\n", myname);
	    exit(1);
    }
    if (FEVget("TFORM3") != NULL) {
	    strcpy(tform3, FEVget("TFORM3"));
	    if (strcmp(tform2, "1E") == 0) {
		bitp3 = FLOATPIX;
	    } else if (strcmp(tform3, "1D") == 0) {
		bitp3 = LONGPIX;
	    } else {
		PR(ER, "%s: bad value for TFORM3. Sorry...\n", myname);
		exit(1);
	    }
    } else {
	    PR(ER, "%s: no parameter TFORM3. Sorry...\n", myname);
	    exit(1);
    }
    if (FEVget("LMAX") != NULL) {
	    *lmax = atoi(FEVget("LMAX"));
    }

    if (lmax_input > 0 && lmax_input <= *lmax) {
	 *lmax = lmax_input;
    } else {
	 if (*nbuf < (*lmax + 1)*2) {
	     if ( (lmax_input + 1) * 2 > *nbuf)
		 lmax_input = *nbuf / 2 - 1;
	     *lmax = lmax_input;
//         } else {
//           if (*lmax > lmax_input) {
//             PR(ER, "%s: check set of lmax: lmax=%d, nbuf=%d\n",
//                    myname, *lmax, *nbuf);
//             exit(1);
//           }
	 }
    }

    /* Free memory from 1nd head */
    FHfree();

    /* get values from file */
    for (i=0; i<naxis2; i++) {
	static int flag = 1;
	register int bb, nlm;

	bitpix = bitp1;
	nlm = get_pix_fits(fp)+0.5;
	l = (int)sqrt((double)(nlm-1));

	if (l > *lmax)
	    break;

	if (l*2+1 > *nbuf) {
	    if (flag) {
		PR(ER, "%s: attention ! l(=%d) is larger ", myname, l);
		PR(ER, "than array dimension (2*l+1>nbuf(%d)).\n", *nbuf);
	    }
	    flag = 0;
	    break;
	}
	m = nlm - 1 - l*l - l;
	bb = l * *nbuf;

	/* Real part */
	bitpix = bitp2;
	(*alm_e) [bb+m]        = get_pix_fits(fp);
	if (feof(fp)) {
	     PR(ER, "%s: unexpected end of file at l=%d and m=%d\n",
		myname, l, m);
	     break;
	}
	/* Imagine part */
	bitpix = bitp3;
	(*alm_e) [bb+m+ *lmax+1] = get_pix_fits(fp);
	if (feof(fp)) {
	     PR(ER, "%s: unexpected end of file at l=%d and m=%d\n",
		myname, l, m);
	     break;
	}
	if (l > lmx)   lmx = l;
    }
    if (*lmax <= 0)  *lmax = lmx;

    /* skip remained */
    if (*lmax < lmax_data) {
	bitpix = bitp2; get_pix_fits(fp);
	bitpix = bitp3; get_pix_fits(fp);
	i++;
	for (; i<naxis2; i++) {
	    bitpix = bitp1; get_pix_fits(fp);
	    bitpix = bitp2; get_pix_fits(fp);
	    bitpix = bitp3; get_pix_fits(fp);
	}
    }

    block_data = naxis1*naxis2;
    if(block_data % BLOCK_SIZE != 0) {
	   register int i;
	   long len = BLOCK_SIZE * (block_data / BLOCK_SIZE + 1);
	   len -= block_data;
	   for(i=0; i<len; i++)   fgetc(fp);
    }

    /*********************/
    /* read 3d extension */
    /*********************/
    FHinit(fp);
    if (FEVget("XTENSION") != NULL)
	strcpy(str, FEVget("XTENSION"));
    else {
	PR(ER, "%s: no 3d extension in file '%s'\n", myname, filename);
	exit(1);
    }
    if (strcmp(str, "BINTABLE") != 0) {
	PR(ER, "%s: can work only with BINTABLE extension\n", myname);
	exit(1);
    }
    if (FEVget("NAXIS") != NULL)
	    naxis  = FSkey_get("NAXIS");
    if (FEVget("NAXIS1") != NULL)
	    naxis1  = FLkey_get("NAXIS1");
    if (FEVget("NAXIS2") != NULL)
	    naxis2  = FLkey_get("NAXIS2");

    if (FEVget("TFIELDS") != NULL)
	    tfields = FLkey_get("TFIELDS");
    if (tfields != 3) {
	PR(ER, "%s: TFILEDS value must be equal to 3 in a_lm file '%s'\n",
	    myname, filename);
	exit(1);
    }

    if (FEVget("TFORM1") != NULL) {
	    strcpy(tform1, FEVget("TFORM1"));
	    if (strcmp(tform1, "1J") == 0) {
		bitp1 = LONGPIX;
	    } else if (strcmp(tform1, "1I") == 0) {
		bitp1 = SHORTPIX;
	    } else {
		PR(ER, "%s: bad value for TFORM1. Sorry...\n", myname);
		exit(1);
	    }
    } else {
	    PR(ER, "%s: no parameter TFORM1. Sorry...\n", myname);
	    exit(1);
    }
    if (FEVget("TFORM2") != NULL) {
	    strcpy(tform2, FEVget("TFORM2"));
	    if (strcmp(tform2, "1E") == 0) {
		bitp2 = FLOATPIX;
	    } else if (strcmp(tform2, "1D") == 0) {
		bitp2 = LONGPIX;
	    } else {
		PR(ER, "%s: bad value for TFORM2. Sorry...\n", myname);
		exit(1);
	    }
    } else {
	    PR(ER, "%s: no parameter TFORM2. Sorry...\n", myname);
	    exit(1);
    }
    if (FEVget("TFORM3") != NULL) {
	    strcpy(tform3, FEVget("TFORM3"));
	    if (strcmp(tform2, "1E") == 0) {
		bitp3 = FLOATPIX;
	    } else if (strcmp(tform3, "1D") == 0) {
		bitp3 = LONGPIX;
	    } else {
		PR(ER, "%s: bad value for TFORM3. Sorry...\n", myname);
		exit(1);
	    }
    } else {
	    PR(ER, "%s: no parameter TFORM3. Sorry...\n", myname);
	    exit(1);
    }
    if (FEVget("LMAX") != NULL) {
	    *lmax = atoi(FEVget("LMAX"));
    }

    if (lmax_input > 0 && lmax_input <= *lmax) {
	 *lmax = lmax_input;
    } else {
	 if (*nbuf < (*lmax + 1)*2) {
	     if ( (lmax_input + 1) * 2 > *nbuf)
		 lmax_input = *nbuf / 2 - 1;
	     *lmax = lmax_input;
//         } else {
//           if (*lmax > lmax_input) {
//             PR(ER, "%s: check set of lmax: lmax=%d, nbuf=%d\n",
//                    myname, *lmax, *nbuf);
//             exit(1);
//           }
	 }
    }

    /* Free memory from 1nd head */
    FHfree();

    /* get values from file */
    for (i=0; i<naxis2; i++) {
	static int flag = 1;
	register int bb, nlm;

	bitpix = bitp1;
	nlm = get_pix_fits(fp)+0.5;
	l = (int)sqrt((double)(nlm-1));

	if (l > *lmax)
	    break;

	if (l*2+1 > *nbuf) {
	    if (flag) {
		PR(ER, "%s: attention ! l(=%d) is larger ", myname, l);
		PR(ER, "than array dimension (2*l+1>nbuf(%d)).\n", *nbuf);
	    }
	    flag = 0;
	    break;
	}
	m = nlm - 1 - l*l - l;
	bb = l * *nbuf;

	/* Real part */
	bitpix = bitp2;
	(*alm_b) [bb+m]        = get_pix_fits(fp);
	if (feof(fp)) {
	     PR(ER, "%s: unexpected end of file at l=%d and m=%d\n",
		myname, l, m);
	     break;
	}
	/* Imagine part */
	bitpix = bitp3;
	(*alm_b) [bb+m+ *lmax+1] = get_pix_fits(fp);
	if (feof(fp)) {
	     PR(ER, "%s: unexpected end of file at l=%d and m=%d\n",
		myname, l, m);
	     break;
	}
	if (l > lmx)   lmx = l;
    }
    if (*lmax <= 0)  *lmax = lmx;

    fclose(fp);
//system("free");
//PR(ER, "rrrr10----------------------------------\n");
    return;
}
/*
 * read alm, and e- & b- polarization modes
 */
int test_ieb_alm(filename, comment)
char *filename;
int comment;
{
	register int i, l, m;
	int tfields = 0;
	int bitp1, bitp2, bitp3;
	char tform1[30],
	     tform2[30],
	     tform3[30];
	char str[161];
	int  lmx = -1, lmax, lmax_data;
	FILE *fp;
	int naxis=0, naxis1=0, naxis2=0;
	int block_data = 0;
	int flag = 0;       /* =1 when I,E,B - alm-file */
	double v;

	/* detect maximum l */
	lmax_data = lmax = Fget_lmax(filename);

	/* open file */
	if ((fp = fopen(filename, "r")) == (FILE *)0) {
		PR(ER, "%s: can not open file '%s'\n", myname, filename);
		exit(1);
	}
	/* read 1st head */
	FHinit(fp);

	if (FEVget("EXTEND") != NULL)
	    strcpy(str, FEVget("EXTEND"));

	if (strcmp(str, "T") != 0) {
	    fclose(fp);
	    return flag;
	}

	/* Free memory from 1st head */
	FHfree();

	/****************************************/
	/* read basic keywods from the 2nd head */
	/****************************************/
	FHinit(fp);

	if (FEVget("XTENSION") != NULL)
	    strcpy(str, FEVget("XTENSION"));
	else {
	    fclose(fp);
	    return flag;
	}
	if (strcmp(str, "BINTABLE") != 0) {
	    fclose(fp);
	    return flag;
	}
	if (FEVget("NAXIS") != NULL)
		naxis  = FSkey_get("NAXIS");
	if (FEVget("NAXIS1") != NULL)
		naxis1  = FLkey_get("NAXIS1");
	if (FEVget("NAXIS2") != NULL)
		naxis2  = FLkey_get("NAXIS2");

	if (FEVget("TFIELDS") != NULL)
		tfields = FLkey_get("TFIELDS");
	if (tfields != 3) {
	    fclose(fp);
	    return flag;
	}

	if (FEVget("POLAR") != NULL) {
	    strcpy(str, FEVget("POLAR"));
	    if (strcmp(str, "F") == 0) { /* false => no polarization */
		fclose(fp);
		return flag;
	    }
	}

	if (FEVget("TFORM1") != NULL) {
		strcpy(tform1, FEVget("TFORM1"));
		if (strcmp(tform1, "1J") == 0) {
		    bitp1 = LONGPIX;
		} else if (strcmp(tform1, "1I") == 0) {
		    bitp1 = SHORTPIX;
		} else {
		    if (comment)
		       PR(ER, "%s: bad value for TFORM1. Sorry...\n", myname);
		    fclose(fp);
		    return flag;
		}
	} else {
	    fclose(fp);
	    if (comment)
		PR(ER, "%s: no parameter TFORM1. Sorry...\n", myname);
	    return flag;
	}

	if (FEVget("TFORM2") != NULL) {
		strcpy(tform2, FEVget("TFORM2"));
		if (strcmp(tform2, "1E") == 0) {
		    bitp2 = FLOATPIX;
		} else if (strcmp(tform2, "1D") == 0) {
		    bitp2 = DOUBLEPIX;
		} else {
		    if (comment)
		       PR(ER, "%s: bad value for TFORM2. Sorry...\n", myname);
		    fclose(fp);
		    return flag;
		}
	} else {
		fclose(fp);
		return flag;
	}

	if (FEVget("TFORM3") != NULL) {
		strcpy(tform3, FEVget("TFORM3"));
		if (strcmp(tform2, "1E") == 0) {
		    bitp3 = FLOATPIX;
		} else if (strcmp(tform3, "1D") == 0) {
		    bitp3 = DOUBLEPIX;
		} else {
		    if (comment)
		       PR(ER, "%s: bad value for TFORM3. Sorry...\n", myname);
		    fclose(fp);
		    return flag;
		}
	} else {
		if (comment)
		     PR(ER, "%s: no parameter TFORM3. Sorry...\n", myname);
		fclose(fp);
		return flag;
	}
	if (FEVget("LMAX") != NULL) {
		lmax = atoi(FEVget("LMAX"));
	}

	/* Free memory from 1nd head */
	FHfree();

	/* get values from file */
	for (i=0; i<naxis2; i++) {
	    static int flag = 1;
	    register int nlm;
	    double ddd;

	    bitpix = bitp1;
	    //nlm = get_pix_fits(fp)+0.5;
	    ddd = get_pix_fits(fp);
	    nlm = ddd + .5;
	    l = (int)sqrt((double)(nlm-1));
	    m = nlm - 1 - l*l - l;

	    if (m > l) {
		if (comment)
		    PR(ER, "%s: attention ! m=%d > l=%d.\n", myname, m, l);
		fclose(fp);
		return flag;
	    }

	    /* Real part */
	    bitpix = bitp2;
	    v = get_pix_fits(fp);
	    if (feof(fp)) {
		 if (comment)
		    PR(ER, "%s: unexpected end of file at l=%d and m=%d\n",
			 myname, l, m);
		 fclose(fp);
		 return flag;
	    }
	    /* Imaginary part */
	    bitpix = bitp3;
	    v = get_pix_fits(fp);
	    if (feof(fp)) {
		 if (comment)
		    PR(ER, "%s: unexpected end of file at l=%d and m=%d\n",
			myname, l, m);
		 fclose(fp);
		 return flag;
	    }
	    if (l > lmx)   lmx = l;
	}

	block_data = naxis1*naxis2;
	if(block_data % BLOCK_SIZE != 0) {
	       register int i;
	       long len = BLOCK_SIZE * (block_data / BLOCK_SIZE + 1);
	       len -= block_data;
	       for(i=0; i<len; i++)   fgetc(fp);
	}

	/* check the end of file */
	/*************************/
	if (feof(fp)) {
	    fclose(fp);
	    return flag;
	}

	/**********************/
	/* read 2nd extension */
	/**********************/
	FHinit(fp);
	if (FEVget("XTENSION") != NULL)
	    strcpy(str, FEVget("XTENSION"));
	else {
	    if (comment)
		 PR(ER, "%s: no 2nd extension in file '%s'\n",
		     myname, filename);
	    fclose(fp);
	    return flag;
	}
	if (strcmp(str, "BINTABLE") != 0) {
	    if (comment)
		PR(ER, "%s: can work only with BINTABLE extension\n", myname);
	    fclose(fp);
	    return flag;
	}

	if (FEVget("NAXIS") != NULL)
		naxis  = FSkey_get("NAXIS");
	if (FEVget("NAXIS1") != NULL)
		naxis1  = FLkey_get("NAXIS1");
	if (FEVget("NAXIS2") != NULL)
		naxis2  = FLkey_get("NAXIS2");

	if (FEVget("TFIELDS") != NULL)
		tfields = FLkey_get("TFIELDS");
	if (tfields != 3) {
	    if (comment)
	    PR(ER,"%s: TFILEDS value must be equal to 3 in a_lm file '%s'\n",
		myname, filename);
	    fclose(fp);
	    return flag;
	}

	if (FEVget("TFORM1") != NULL) {
		strcpy(tform1, FEVget("TFORM1"));
		if (strcmp(tform1, "1J") == 0) {
		    bitp1 = LONGPIX;
		} else if (strcmp(tform1, "1I") == 0) {
		    bitp1 = SHORTPIX;
		} else {
		    if (comment)
		      PR(ER, "%s: bad value for TFORM1. Sorry...\n", myname);
		    fclose(fp);
		    return flag;
		}
	} else {
		if (comment)
		    PR(ER, "%s: no parameter TFORM1. Sorry...\n", myname);
		fclose(fp);
		return flag;
	}

	if (FEVget("TFORM2") != NULL) {
		strcpy(tform2, FEVget("TFORM2"));
		if (strcmp(tform2, "1E") == 0) {
		    bitp2 = FLOATPIX;
		} else if (strcmp(tform2, "1D") == 0) {
		    bitp2 = DOUBLEPIX;
		} else {
		    if (comment)
		       PR(ER, "%s: bad value for TFORM2. Sorry...\n", myname);
		    fclose(fp);
		    return flag;
		}
	} else {
		if (comment)
		    PR(ER, "%s: no parameter TFORM2. Sorry...\n", myname);
		fclose(fp);
		return flag;
	}

	if (FEVget("TFORM3") != NULL) {
		strcpy(tform3, FEVget("TFORM3"));
		if (strcmp(tform2, "1E") == 0) {
		    bitp3 = FLOATPIX;
		} else if (strcmp(tform3, "1D") == 0) {
		    bitp3 = DOUBLEPIX;
		} else {
		    if (comment)
		       PR(ER, "%s: bad value for TFORM3. Sorry...\n", myname);
		    fclose(fp);
		    return flag;
		}
	} else {
		if (comment)
		    PR(ER, "%s: no parameter TFORM3. Sorry...\n", myname);
		fclose(fp);
		return flag;
	}
	if (FEVget("LMAX") != NULL) {
		lmax = atoi(FEVget("LMAX"));
	}

	/* Free memory from 1st head */
	FHfree();

	/* get values from file */
	for (i=0; i<naxis2; i++) {
	    static int flag = 1;
	    register int nlm;

	    bitpix = bitp1;
	    nlm = get_pix_fits(fp)+0.5;
	    l = (int)sqrt((double)(nlm-1));
	    m = nlm - 1 - l*l - l;

	    /* Real part */
	    bitpix = bitp2;
	    v = get_pix_fits(fp);
	    if (feof(fp)) {
		 if (comment)
		    PR(ER, "%s: unexpected end of file at l=%d and m=%d\n",
			myname, l, m);
		 fclose(fp);
		 return flag;
	    }
	    /* Imagine part */
	    bitpix = bitp3;
	    v = get_pix_fits(fp);
	    if (feof(fp)) {
		 if (comment)
		    PR(ER, "%s: unexpected end of file at l=%d and m=%d\n",
			myname, l, m);
		 fclose(fp);
		 return flag;
	    }
	    if (l > lmx)   lmx = l;
	}

	block_data = naxis1*naxis2;
	if(block_data % BLOCK_SIZE != 0) {
	       register int i;
	       long len = BLOCK_SIZE * (block_data / BLOCK_SIZE + 1);
	       len -= block_data;
	       for(i=0; i<len; i++)   fgetc(fp);
	}

	/* check the end of file */
	/*************************/
	if (feof(fp)) {
	    fclose(fp);
	    return flag;
	}

	/*********************/
	/* read 3d extension */
	/*********************/
	FHinit(fp);
	if (FEVget("XTENSION") != NULL)
	    strcpy(str, FEVget("XTENSION"));
	else {
	    if (comment)
	       PR(ER, "%s: no 3d extension in file '%s'\n", myname, filename);
	    fclose(fp);
	    return flag;
	}
	if (strcmp(str, "BINTABLE") != 0) {
	    if (comment)
	       PR(ER, "%s: can work only with BINTABLE extension\n", myname);
	    fclose(fp);
	    return flag;
	}
	if (FEVget("NAXIS") != NULL)
		naxis  = FSkey_get("NAXIS");
	if (FEVget("NAXIS1") != NULL)
		naxis1  = FLkey_get("NAXIS1");
	if (FEVget("NAXIS2") != NULL)
		naxis2  = FLkey_get("NAXIS2");

	if (FEVget("TFIELDS") != NULL)
		tfields = FLkey_get("TFIELDS");
	if (tfields != 3) {
	    if (comment)
	    PR(ER, "%s: TFILEDS value must be equal to 3 in a_lm file '%s'\n",
		myname, filename);
	    fclose(fp);
	    return flag;
	}

	if (FEVget("TFORM1") != NULL) {
		strcpy(tform1, FEVget("TFORM1"));
		if (strcmp(tform1, "1J") == 0) {
		    bitp1 = LONGPIX;
		} else if (strcmp(tform1, "1I") == 0) {
		    bitp1 = SHORTPIX;
		} else {
		    if (comment)
		       PR(ER, "%s: bad value for TFORM1. Sorry...\n", myname);
		    fclose(fp);
		    return flag;
		}
	} else {
		if (comment)
		     PR(ER, "%s: no parameter TFORM1. Sorry...\n", myname);
		fclose(fp);
		return flag;
	}
	if (FEVget("TFORM2") != NULL) {
		strcpy(tform2, FEVget("TFORM2"));
		if (strcmp(tform2, "1E") == 0) {
		    bitp2 = FLOATPIX;
		} else if (strcmp(tform2, "1D") == 0) {
		    bitp2 = DOUBLEPIX;
		} else {
		    if (comment)
		       PR(ER, "%s: bad value for TFORM2. Sorry...\n", myname);
		    fclose(fp);
		    return flag;
		}
	} else {
		if (comment)
		     PR(ER, "%s: no parameter TFORM2. Sorry...\n", myname);
		fclose(fp);
		return flag;
	}
	if (FEVget("TFORM3") != NULL) {
		strcpy(tform3, FEVget("TFORM3"));
		if (strcmp(tform2, "1E") == 0) {
		    bitp3 = FLOATPIX;
		} else if (strcmp(tform3, "1D") == 0) {
		    bitp3 = DOUBLEPIX;
		} else {
		    if (comment)
		       PR(ER, "%s: bad value for TFORM3. Sorry...\n", myname);
		    fclose(fp);
		    return flag;
		}
	} else {
		if (comment)
		      PR(ER, "%s: no parameter TFORM3. Sorry...\n", myname);
		fclose(fp);
		return flag;
	}
	if (FEVget("LMAX") != NULL) {
		lmax = atoi(FEVget("LMAX"));
	}

	/* Free memory from 1nd head */
	FHfree();

	/* get values from file */
	for (i=0; i<naxis2; i++) {
	    static int flag = 1;
	    register int nlm;

	    bitpix = bitp1;
	    nlm = get_pix_fits(fp)+0.5;
	    l = (int)sqrt((double)(nlm-1));
	    m = nlm - 1 - l*l - l;

	    /* Real part */
	    bitpix = bitp2;
	    v = get_pix_fits(fp);
	    if (feof(fp)) {
		 if (comment)
		    PR(ER, "%s: unexpected end of file at l=%d and m=%d\n",
		       myname, l, m);
		 fclose(fp);
		 return flag;
	    }
	    /* Imagine part */
	    bitpix = bitp3;
	    v = get_pix_fits(fp);
	    if (feof(fp)) {
		 if (comment)
		    PR(ER, "%s: unexpected end of file at l=%d and m=%d\n",
		       myname, l, m);
		 fclose(fp);
		 return flag;
	    }
	    if (l > lmx)   lmx = l;
	}

	flag = 1;

	fclose(fp);
	return flag;
}


/*
 * write alm, and e- & b- polarization modes
 */
void write_ieb_alm(alm_i, alm_e, alm_b, lmax, nbuf, lmax_print, fp_out)
double *alm_i, *alm_e, *alm_b;
int lmax, lmax_print;
int nbuf;
FILE *fp_out;
{
	out_glesp_alm(alm_i, lmax, nbuf, lmax_print, fp_out);
	add_alm_ext(alm_e, alm_b, lmax, nbuf, lmax_print, fp_out);
	return;
}


void outp_glesp_3map(f, q, u, x, nx, np, nph, fp_out, x_type, grid_type)
double *f,*q,*u;/* 2-dim maps                          */
double *x;      /* argument, x = cos theta             */
int nx,         /* number of elements in x-array       */
    np;         /* 2-dim factor: f(i*nbuf+j)<->f[i][j] */
int *nph;       /* array of elements in phi-arrays     */
FILE *fp_out;
char *x_type, *grid_type;
{
	register int i,j,ii;
	char ttype1[30], tform1[30], tunit1[30],
	     ttype2[30], tform2[30], tunit2[30],
	     ttype3[30], tform3[30], tunit3[30],
	     ttype4[30], tform4[30], tunit4[30],
	     ttype5[30], tform5[30], tunit5[30];
	int tfields = 5;
	long nphi=0;     /* total number of elements in nph-arrays */
	int bitp1 = GLESPPIX,
	    bitp2 = LONGPIX,
	    bitp3,
	    bitp4,
	    bitp5;
	long long long_naxis1;
	char ctype1[30], bunit[30], object[30];
	int block_data = 0;

	/*****************/
	/* basic keywods */
	/*****************/
	bitp3 = bitpix_map; /* default: Float Point Pixels for map */
	bitp1 = bitp3;
	bitp4 = bitp5 = bitp3;

	strcpy(ctype1, x_type);
	strcpy(bunit,  "K");
	strcpy(object, "CMB T,Q,U maps");

	tfields = 5;

	if (bitp1 == FLOATPIX)
	    sprintf(tform1, "%dE", nx);
	else  if (bitp1 == DOUBLEPIX)
	    sprintf(tform1, "%dD", nx);
	strcpy(ttype1,  ctype1);
	if (strncmp(ctype1, "COS(", 4) == 0)
		strcpy(tunit1,  "COS(RAD)");
	else
		strcpy(tunit1,  "RAD");

	sprintf(tform2, "%dJ", nx);
	strcpy(ttype2,  "DIMS");
	strcpy(tunit2,  "UNITLESS");

	for (i=0; i<nx; i++)  nphi += nph[i];
	if (bitp3 == FLOATPIX)
	    sprintf(tform3, "%ldE", nphi);
	else  if (bitp3 == DOUBLEPIX)
	    sprintf(tform3, "%ldD", nphi);
	strcpy(ttype3,  "TEMPERATURE");
	strcpy(tunit3,  "K");
	strcpy(tform4,  tform3);
	strcpy(ttype4,  "Q-polarization");
	strcpy(tunit4,  "K");
	strcpy(tform5,  tform3);
	strcpy(ttype5,  "U-polarization");
	strcpy(tunit5,  "K");

	/* Number of entries in table in bytes */
	long_naxis1 = (long long)nphi * (ABS(bitp3) / 8) * 3 +
	    nx * (ABS(bitp2) / 8) + nx * (ABS(bitp1) / 8);

	/************/
	/* set head */
	/************/
	/* sFEVset("SIMPLE", "T"); FEVexport("SIMPLE"); */
	Fckey_set("SIMPLE",    "T");
	FSkey_set("BITPIX",    8);
	FSkey_set("NAXIS",     0);
	Fckey_set("EXTEND",    "T");
	FCkey_set("COMMENT",  "GLESP MAP");

	/* 1st HEAD */
	Hupdate_fits(fp_out);
	FHfree();

	FCkey_set("XTENSION",  "BINTABLE");
	FSkey_set("BITPIX",  8);
	FSkey_set("NAXIS",   2);
	FLLkey_set("NAXIS1", long_naxis1);
	FLkey_set("NAXIS2",  1);
	FLkey_set("PCOUNT",  0);
	FLkey_set("GCOUNT",  1);

	FLkey_set("TFIELDS", tfields);

	FCkey_set("TFORM1",  tform1); /* 1st field: vector of X */
	/* FCkey_set("TDIM1",   tdim1); */
	FCkey_set("TTYPE1",  ttype1);
	FCkey_set("TUNIT1",  tunit1);

	FCkey_set("TFORM2",  tform2); /* 2nd field: vector nph(x) */
	/* FCkey_set("TDIM2",   tdim2); */
	FCkey_set("TTYPE2",  ttype2);
	FCkey_set("TUNIT2",  tunit2);

	FCkey_set("TFORM3",  tform3); /* 3d field: vector of T(X,Y)  */
	/* FCkey_set("TDIM3",   tdim3); */
	FCkey_set("TTYPE3",  ttype3);
	FCkey_set("TUNIT3",  tunit3);

	FCkey_set("TFORM4",  tform4); /* 4th field: vector of Q(X,Y)  */
	/* FCkey_set("TDIM4",   tdim4); */
	FCkey_set("TTYPE4",  ttype4);
	FCkey_set("TUNIT4",  tunit4);

	FCkey_set("TFORM5",  tform5); /* 5th field: vector of U(X,Y)  */
	/* FCkey_set("TDIM5",   tdim5); */
	FCkey_set("TTYPE5",  ttype5);
	FCkey_set("TUNIT5",  tunit5);

	FCkey_set("OBJECT",  object);
	FCkey_set("ORIGIN",  "GLESP");
	FCkey_set("PIXTYPE", "GLESP");

	/* 2nd HEAD */
	Hupdate_fits(fp_out);

	/* recording 1-dim array of  X-argument vector */
	bitpix = bitp1;
	for(i=0; i<nx; i++) {
		double v = x[i];
		put_pix_fits(fp_out, v);
	}

	/* recording 1-dim array of dimensions of \phi-vector */
	bitpix = bitp2;
	for(i=0; i<nx; i++) {
		double v = nph[i];
		put_pix_fits(fp_out, v);
	}


	/*-------------*/
	/* temperature */
	/*-------------*/
	bitpix = bitp3;
	/* recording 2-dim array */
	for(ii=i=0; i<nx; i++,ii+=np) {
		register int n=nph[i];
		for(j=0; j<n; j++) {
			double v = f[ii+j];
			put_pix_fits(fp_out, v);
		}
	}
	/*----------------*/
	/* Q-polarization */
	/*----------------*/
	bitpix = bitp4;
	/* recording 2-dim array */
	for(ii=i=0; i<nx; i++,ii+=np) {
		register int n=nph[i];
		for(j=0; j<n; j++) {
			double v = q[ii+j];
			put_pix_fits(fp_out, v);
		}
	}
	/*----------------*/
	/* U-polarization */
	/*----------------*/
	bitpix = bitp5;
	/* recording 2-dim array */
	for(ii=i=0; i<nx; i++,ii+=np) {
		register int n=nph[i];
		for(j=0; j<n; j++) {
			double v = u[ii+j];
			put_pix_fits(fp_out, v);
		}
	}
	block_data = long_naxis1;
	if(block_data % BLOCK_SIZE != 0) {
	       long len = BLOCK_SIZE * ( block_data / BLOCK_SIZE + 1);
	       len -= block_data;
	       for(i=0; i<len; i++)   fputc('\0', fp_out);
	}

	/* free memory from head */
	FHfree();

	return;
    return;
}

void in_glesp_1map(f, x, nx, np, nph, filename)
double **f;         /* 2-dim maps: T,Q,U                   */
double **x;         /* argument, x = cos theta            */
int *nx;            /* number of elements in x-array      */
int *np;            /* max number of elements by phi-axis */
int **nph;          /* number of elements in phi-array    */
char *filename;
{
    double *q=(double *)0, *u=(double *)0; /* 2-dim maps:  Q,U   */

    in_glesp_3map(f, &q, &u, x, nx, np, nph, filename);

    if (q != (double *)0)  free((void *)q);
    if (u != (double *)0)  free((void *)u);

    if (glesp_comments < 0) {
	if (getenv("GLESP_COMMENTS") != NULL) {
	    char str[21];
	    strcpy(str,getenv("GLESP_COMMENTS"));
	    glesp_comments = *str == 'N' || *str == 'n' ? 0 : 1;
	    if (glesp_comments) {
		 if ( strcmp(str, "OFF") == 0 || strcmp(str, "off") == 0)
		     glesp_comments = 0;
	    }
	} else
	    glesp_comments = 1;
    }

    if (glesp_comments) {
      PR(ER, "%s:\n    input map contains polarization data too, \n", myname);
      PR(ER, "    but only anisotropy will be processed.\n\n");
    }
    return;
}

void in_glesp_3map(f, q, u, x, nx, np, nph, filename)
double **f,**q,**u; /* 2-dim map: T,Q,U                   */
double **x;         /* argument, x = cos theta            */
int *nx;            /* number of elements in x-array      */
int *np;            /* max number of elements by phi-axis */
int **nph;          /* number of elements in phi-array    */
char *filename;
{
	register int i,j,ii;
	register FILE *fp_in;
	double fmax = -1.e+30, fmin = 1.e+30;
	long int nphi=0, nph_max=0;

	ph->tform1[0] = ph->tform2[0] = ph->tform3[0] =
	      ph->tform4[0] = ph->tform5[0] =
	      ph->tdim1[0] = ph->tdim2[0] = ph->tdim3[0] =
	      ph->tdim4[0] = ph->tdim5[0] = '\0';

	if ((fp_in = fopen(filename, "r")) == (FILE *)0) {
	    PR(ER, "%s: can not open file '%s'\n", myname, filename);
	    exit(1);
	}

	/****************************************/
	/* read basic keywords from the 1st head */
	/****************************************/
	/* Hinit_fits(fp_in); */
	FHinit_FITS(fp_in, 0);

	if (FEVget("EXTEND") != NULL)
	    strcpy(ph->extend, FEVget("EXTEND"));

	if (strcmp(ph->extend, "T") != 0) {
	    PR(ER, "%s: bad extension in file '%s'\n", myname, filename);
	    exit(1);
	}

	FHfree();

	/****************************************/
	/* read basic keywords from the 2nd head */
	/****************************************/
	/* Hinit(fp_in); */
	FHinit_FITS(fp_in, 0);

	if (FEVget("XTENSION") != NULL)
	    strcpy(ph->xtens, FEVget("XTENSION"));

	if (strcmp(ph->xtens, "BINTABLE") != 0) {
	    PR(ER, "%s: can work only with BINTABLE extension\n", myname);
	    exit(1);
	}
	if (FEVget("NAXIS") != NULL)
		ph->naxis  = FSkey_get("NAXIS");
	if (FEVget("NAXIS1") != NULL)
		ph->naxis1  = FLkey_get("NAXIS1");
	if (FEVget("NAXIS2") != NULL)
		ph->naxis2  = FLkey_get("NAXIS2");

	if (FEVget("TFIELDS") != NULL)
		ph->tfields = FLkey_get("TFIELDS");

	if( ph->tfields != 5) {
		PR(ER, "%s: bad number of fields (%d) for function ",
			myname, ph->tfields);
		PR(ER, "'ff_read_ext2_map'. It must be 5.\n");
		exit(1);
	}

	if (FEVget("TFORM1") != NULL)
		strcpy(ph->tform1, FEVget("TFORM1"));
	if (FEVget("TDIM1") != NULL)
		strcpy(ph->tdim1, FEVget("TDIM1"));
	if (FEVget("TTYPE1") != NULL)
		strcpy(ph->ttype1, FEVget("TTYPE1"));
	if (FEVget("TUNIT1") != NULL)
		strcpy(ph->tunit1, FEVget("TUNIT1"));

	if (FEVget("TFORM2") != NULL)
		strcpy(ph->tform2, FEVget("TFORM2"));
	if (FEVget("TDIM2") != NULL)
		strcpy(ph->tdim2, FEVget("TDIM2"));
	if (FEVget("TTYPE2") != NULL)
		strcpy(ph->ttype2, FEVget("TTYPE2"));
	if (FEVget("TUNIT2") != NULL)
		strcpy(ph->tunit2, FEVget("TUNIT2"));

	if (FEVget("TFORM3") != NULL)
		strcpy(ph->tform3, FEVget("TFORM3"));
	if (FEVget("TDIM3") != NULL)
		strcpy(ph->tdim3,  FEVget("TDIM3"));
	if (FEVget("TTYPE3") != NULL)
		strcpy(ph->ttype3, FEVget("TTYPE3"));
	if (FEVget("TUNIT3") != NULL)
		strcpy(ph->tunit3, FEVget("TUNIT3"));

	if (FEVget("TFORM4") != NULL)
		strcpy(ph->tform4, FEVget("TFORM4"));
	if (FEVget("TDIM4") != NULL)
		strcpy(ph->tdim4,  FEVget("TDIM4"));
	if (FEVget("TTYPE4") != NULL)
		strcpy(ph->ttype4, FEVget("TTYPE4"));
	if (FEVget("TUNIT4") != NULL)
		strcpy(ph->tunit4, FEVget("TUNIT4"));

	if (FEVget("TFORM5") != NULL)
		strcpy(ph->tform5, FEVget("TFORM5"));
	if (FEVget("TDIM5") != NULL)
		strcpy(ph->tdim5,  FEVget("TDIM5"));
	if (FEVget("TTYPE5") != NULL)
		strcpy(ph->ttype5, FEVget("TTYPE5"));
	if (FEVget("TUNIT5") != NULL)
		strcpy(ph->tunit5, FEVget("TUNIT5"));

	if (FEVget("OBJECT") != NULL)
		strcpy(ph->object, FEVget("OBJECT"));
	if (FEVget("ORIGIN") != NULL)
		strcpy(ph->origin, FEVget("ORIGIN"));

	/* set grid flag */

	/* X-axis grid type */
	if (FEVget("GRID") != NULL) {
		char str[51];
		strcpy(str, FEVget("GRID"));
		switch(*str) {
		case 'P':
		case 'p':
			if (strcmp("PHI(X)+X", str) == 0)
				ph->grid = GRID_PHI_X;
			else
				ph->grid = GRID_PHI_X_LEGENDRE;
			break;
		case 'L':
		case 'l':
			ph->grid = GRID_LEGENDRE;
			break;
		case 'U':
		case 'u':
			ph->grid = GRID_UNIFORM_COS;
			break;
		case 'T':
		case 't':
			ph->grid = GRID_UNIFORM_THETA;
			break;
		default :
			ph->grid = GRID_NON_SET;
			break;
		}
	}
	if(*(ph->tform1)=='\0'|| *(ph->tform2)=='\0'|| *(ph->tform3)=='\0') {
	     PR(ER, "%s: bad value of dimensions. Check TFORM# keywords.\n",
		myname);
	     exit(1);
	}

	/* extract dimension of the 1st array */
	i = atoi(ph->tform1);

	/* extract dimensions of the 2nd array */
	*nx = atoi(ph->tform2);
	if (*nx != i ) {
		PR(ER, "%s: bad dimension by X-axis...\n", myname);
		exit(1);
	}
	if (glesp_comments < 0) {
	    if (getenv("GLESP_COMMENTS") != NULL) {
		char str[21];
		strcpy(str,getenv("GLESP_COMMENTS"));
		glesp_comments = *str == 'N' || *str == 'n' ? 0 : 1;
		if (glesp_comments) {
		    if ( strcmp(str, "OFF") == 0 || strcmp(str, "off") == 0)
			glesp_comments = 0;
		}
	    } else
		glesp_comments = 1;
	}
	if (glesp_comments)
	    PR(ER,"%s: Reading map arrays...", myname);

	/**********************************/
	/* take memory and read 1st field */
	/**********************************/
	if (flag_take_memory) {
	    if ((*x = (double *)calloc((unsigned)*nx, sizeof **x)) ==
		    (double *)0)
			    goto nomem;
	}
	/* read 1-dim array of  X-argument vector */
	/* bitpix = ph->bitpix1 = atoi(ph->tform1) / *nx * 8; */
	i = ph->tform1[strlen(ph->tform1)-1];
	if ( i == 'D' )
		bitpix = ph->bitpix1 = DOUBLEPIX;
	else /* if ( i == 'E' )*/
		bitpix = ph->bitpix1 = FLOATPIX;
	for(i=0; i < *nx; i++) {
		(*x)[i] = get_pix_fits(fp_in);
	}

	/**********************************/
	/* take memory and read 2nd field */
	/**********************************/
	if (flag_take_memory) {
	    if ((*nph = (int*)calloc((unsigned)*nx, sizeof **nph))==(int *)0)
		goto nomem;
	}
	/* read 1-dim array of Nph-vector */
	/* bitpix = ph->bitpix2 = atoi(ph->tform2) / *nx * 8; */
	i = ph->tform2[strlen(ph->tform2)-1];
	switch (i) {
	case 'I':
		ph->bitpix2 = SHORTPIX;
		break;
	case 'J':
		ph->bitpix2 = LONGPIX;
		break;
	default:
	case 'E':
		ph->bitpix2 = FLOATPIX;
		break;
	}
	bitpix = ph->bitpix2;
	for(j=0; j < *nx; j++) {
		(*nph)[j] = get_pix_fits(fp_in) + 0.5;
		if ( (*nph)[j] > nph_max )  nph_max = (*nph)[j];
		nphi += (*nph)[j];
	}

	ph->nx = *nx;

	/**********************************/
	/* take memory and read 3d field */
	/**********************************/
	/* find memory for arrays */
	if (flag_take_memory) {
	   if ((*f =(double*)calloc((unsigned)(*nx *nph_max), sizeof **f)) ==
		(double *)0) {
nomem:
		PR(ER, "%s: no memory for arrays...\n", myname);
		exit(1);
	   }
	   if ((*q =(double*)calloc((unsigned)(*nx *nph_max), sizeof **q)) ==
		(double *)0)
		    goto nomem;
	   if ((*u =(double*)calloc((unsigned)(*nx *nph_max), sizeof **u)) ==
		(double *)0)
		    goto nomem;
	}

	/* in memory we keep transposed image:  */
	/*      f(nphi,nx)                      */
	/*      by X (in strings):  cos theta,  */
	/*      by Y (in columns):  phi.        */
	/* It gives a possibility to change     */
	/* step by X-axis with interpolation    */

	/* read 2-dim array */
	/* bitpix = ph->bitpix3 = atoll(ph->tform3) / nphi * 8; */
	i = ph->tform3[strlen(ph->tform3)-1];
	switch (i) {
	case 'D':
		ph->bitpix3 = DOUBLEPIX;
		break;
	default:
	case 'E':
		ph->bitpix3 = FLOATPIX;
		break;
	}
	bitpix = ph->bitpix3;

	for(ii=i=0; i < *nx; i++,ii+=nph_max) {
		register int n = (*nph)[i];
		for(j=0; j< n; j++) {
			double v;
			if (feof(fp_in)) {
			    PR(ER, "%s: unexpected end of file '%s'\n",
				myname, filename);
			    exit(1);
			}
			v = get_pix_fits(fp_in);
			if (v > fmax)  fmax = v;
			if (v < fmin)  fmin = v;
			(*f)[ii+j] = v;
		}
	}
	if (glesp_comments)
	    PR(ER,"\n    t_min=%15.8e, t_max=%15.8e, (nx,np)=(%d,%ld)\n",
		  fmin, fmax, *nx, nph_max);

	/* read 2-dim array of Q-polarization */
	fmin =  1.e+30;
	fmax = -1.e+30;
	i = ph->tform4[strlen(ph->tform3)-1];
	switch (i) {
	case 'D':
		ph->bitpix4 = DOUBLEPIX;
		break;
	default:
	case 'E':
		ph->bitpix4 = FLOATPIX;
		break;
	}
	bitpix = ph->bitpix4;
	for(i=0; i < *nx; i++) {
		register int ii = i * nph_max;
		for(j=0; j< (*nph)[i]; j++) {
			double v;
			if (feof(fp_in)) {
			    PR(ER, "%s: unexpected end of file '%s'\n",
				myname, filename);
			    exit(1);
			}
			v = get_pix_fits(fp_in);
			if (v > fmax)  fmax = v;
			if (v < fmin)  fmin = v;
			(*q)[ii+j] = v;
		}
	}
	if (glesp_comments)
	    PR(ER,"    q_min=%15.8e, q_max=%15.8e;\n", fmin, fmax);


	/* read 2-dim array of U-polarization */
	fmin =  1.e+30;
	fmax = -1.e+30;
	i = ph->tform5[strlen(ph->tform3)-1];
	switch (i) {
	case 'D':
		ph->bitpix5 = DOUBLEPIX;
		break;
	default:
	case 'E':
		ph->bitpix5 = FLOATPIX;
		break;
	}
	bitpix = ph->bitpix5;
	for(ii=i=0; i < *nx; i++,ii+=nph_max) {
		register int n = (*nph)[i];
		for(j=0; j< n; j++) {
			double v;
			if (feof(fp_in)) {
			    PR(ER, "%s: unexpected end of file '%s'\n",
				myname, filename);
			    exit(1);
			}
			v = get_pix_fits(fp_in);
			if (v > fmax)  fmax = v;
			if (v < fmin)  fmin = v;
			(*u)[ii+j] = v;
		}
	}
	if (glesp_comments)
	    PR(ER,"    u_min=%15.8e, u_max=%15.8e\n", fmin, fmax);

	if (fp_in != stdin && fp_in != (FILE *)0)  fclose(fp_in);
	FHfree();

	*np = nph_max;

	return;
}

/*--------------------*/
void in_glesp_alm_T(alm, lmax, nbuf, filename)
double *alm;
int *lmax;
int nbuf;
char *filename;
{
	register int i, l, m;
	int tfields = 0;
	int bitp1, bitp2, bitp3;
	char tform1[30],
	     tform2[30],
	     tform3[30];
	char str[161];
	int  lmx = -1, lmax_input = *lmax;
	FILE *fp;
	int naxis=0, naxis1=0, naxis2=0;

	/* open file */
	if ((fp = fopen(filename, "r")) == (FILE *)0) {
		PR(ER, "%s: can not open file '%s'\n", myname, filename);
		exit(1);
	}
	/* read 1st head */
	FHinit(fp);

	if (FEVget("EXTEND") != NULL)
	    strcpy(str, FEVget("EXTEND"));

	if (strcmp(str, "T") != 0) {
	    PR(ER, "%s: bad extension in file '%s'\n", myname, filename);
	    exit(1);
	}

	/* Free memory from 1st head */
	FHfree();

	/****************************************/
	/* read basic keywods from the 2nd head */
	/****************************************/
	FHinit(fp);

	if (FEVget("XTENSION") != NULL)
	    strcpy(str, FEVget("XTENSION"));
	else {
	    PR(ER, "%s: no extension in file '%s'\n", myname, filename);
	    exit(1);
	}
	if (strcmp(str, "BINTABLE") != 0) {
	    PR(ER, "%s: can work only with BINTABLE extension\n", myname);
	    exit(1);
	}
	if (FEVget("NAXIS") != NULL)
		naxis  = FSkey_get("NAXIS");
	if (FEVget("NAXIS1") != NULL)
		naxis1  = FLkey_get("NAXIS1");
	if (FEVget("NAXIS2") != NULL)
		naxis2  = FLkey_get("NAXIS2");

	if (FEVget("TFIELDS") != NULL)
		tfields = FLkey_get("TFIELDS");
	if (tfields != 3) {
	    PR(ER, "%s: TFILEDS value must be equal to 3 in a_lm file '%s'\n",
		myname, filename);
	    exit(1);
	}

	if (FEVget("TFORM1") != NULL) {
		strcpy(tform1, FEVget("TFORM1"));
		if (strcmp(tform1, "1J") == 0) {
		    bitp1 = LONGPIX;
		} else if (strcmp(tform1, "1I") == 0) {
		    bitp1 = SHORTPIX;
		} else {
		    PR(ER, "%s: bad value for TFORM1. Sorry...\n", myname);
		    exit(1);
		}
	} else {
		PR(ER, "%s: no parameter TFORM1. Sorry...\n", myname);
		exit(1);
	}
	if (FEVget("TFORM2") != NULL) {
		strcpy(tform2, FEVget("TFORM2"));
		if (strcmp(tform2, "1E") == 0) {
		    bitp2 = FLOATPIX;
		} else if (strcmp(tform2, "1D") == 0) {
		    bitp2 = DOUBLEPIX;
		} else {
		    PR(ER, "%s: bad value for TFORM2. Sorry...\n", myname);
		    exit(1);
		}
	} else {
		PR(ER, "%s: no parameter TFORM2. Sorry...\n", myname);
		exit(1);
	}
	if (FEVget("TFORM3") != NULL) {
		strcpy(tform3, FEVget("TFORM3"));
		if (strcmp(tform2, "1E") == 0) {
		    bitp3 = FLOATPIX;
		} else if (strcmp(tform3, "1D") == 0) {
		    bitp3 = DOUBLEPIX;
		} else {
		    PR(ER, "%s: bad value for TFORM3. Sorry...\n", myname);
		    exit(1);
		}
	} else {
		PR(ER, "%s: no parameter TFORM3. Sorry...\n", myname);
		exit(1);
	}
	if (FEVget("LMAX") != NULL) {
		*lmax = atoi(FEVget("LMAX"));
	}

	if (lmax_input > 0 && lmax_input <= *lmax) {
	     *lmax = lmax_input;
	} else {
	     if (nbuf < (*lmax + 1)*2) {
		 if ( (lmax_input + 1) * 2 > nbuf)
		     lmax_input = nbuf / 2 - 1;
		 *lmax = lmax_input;
	     } else {
	       if (*lmax > lmax_input) {
		 PR(ER, "%s: check set of lmax: lmax=%d, nbuf=%d\n",
			myname, *lmax, nbuf);
		 exit(1);
	       }
	     }
	}

	/* Free memory from 1nd head */
	FHfree();

	/* get values from file */
	for (i=0; i<naxis2; i++) {
	    static int flag = 1;
	    register int bb, nlm;
	    double ddd;

	    bitpix = bitp1;
	    //nlm = get_pix_fits(fp)+0.5;
	    ddd = get_pix_fits(fp);
	    nlm = ddd + .5;
	    l = (int)sqrt((double)(nlm-1));

	    if (l > *lmax)
		break;

	    if (l*2+1 > nbuf) {
		if (flag) {
		    PR(ER, "%s: attention ! l(=%d) is larger ", myname, l);
		    PR(ER, "than array dimension (2*l+1>nbuf(%d)).\n", nbuf);
		}
		flag = 0;
		break;
	    }
	    m = nlm - 1 - l*l - l;
	    bb = m * nbuf;

	    /* Real part */
	    bitpix = bitp2;
	    alm [bb+l]        = get_pix_fits(fp);
	    if (feof(fp)) {
		 PR(ER, "%s: unexpected end of file at l=%d and m=%d\n",
		    myname, l, m);
		 break;
	    }
	    /* Imagine part */
	    bitpix = bitp3;
	    alm [bb+l+ (*lmax+1)*nbuf] = get_pix_fits(fp);
	    if (feof(fp)) {
		 PR(ER, "%s: unexpected end of file at l=%d and m=%d\n",
		    myname, l, m);
		 break;
	    }
	    if (l > lmx)   lmx = l;
	}
	if (*lmax <= 0)  *lmax = lmx;

	fclose(fp);
	return;
}

void out_glesp_alm_T(alm, lmax, nbuf, lmax_print, fp_out)
double *alm;
int lmax, lmax_print;
int nbuf;
FILE *fp_out;
{
	register int l, m;
	int nlm = (lmax_print+1)*(lmax_print+2) / 2;
	int tfields = 3, nsize = 0;
	char ttype1[30], tform1[30], tunit1[30],
	     ttype2[30], tform2[30], tunit2[30],
	     ttype3[30], tform3[30], tunit3[30];
	int bitp1, bitp2, bitp3;
	int block_data = 0;

	strcpy(tform1, "1J");
	strcpy(ttype1,  "INDEX");
	strcpy(tunit1,  "UNITLESS");

	bitp2=bitpix_map;
	bitp3=bitpix_map;
	if (bitpix_map == FLOATPIX) {
	    strcpy(tform2, "1E");
	    strcpy(tform3, "1E");
	} else if (bitpix_map == DOUBLEPIX) {
	    strcpy(tform2, "1D");
	    strcpy(tform3, "1D");
	}
	strcpy(ttype2,  "REAL");
	strcpy(tunit2,  "mK");
	strcpy(ttype3,  "IMAG");
	strcpy(tunit3,  "mK");

	/* TFORM 1 */
	bitp1=LONGPIX;
	if (strcmp(tform1, "1J") == 0) {
		nsize += 4;
		bitp1=LONG;
	} else if (strcmp(tform1, "1I") == 0) {
		nsize += 2;
		sprintf(tform1, "1I");
		bitp1=SHORTPIX;
	}
	/* TFORM 2 */
	if (strcmp(tform2, "1E") == 0) {
		nsize += 4;
		bitp2=FLOATPIX;
	} else if (strcmp(tform2, "1D") == 0) {
		nsize += 8;
		bitp2=DOUBLEPIX;
	}
	/* TFORM 3 */
	if (strcmp(tform3, "1E") == 0) {
		nsize += 4;
		bitp3=FLOATPIX;
	} else if (strcmp(tform3, "1D") == 0) {
		nsize += 8;
		bitp3=DOUBLEPIX;
	}
	/************/
	/* set head */
	/************/
	Fckey_set("SIMPLE",    "T");
	FSkey_set("BITPIX",    8);
	FSkey_set("NAXIS",     0);
	Fckey_set("EXTEND",    "T");
	FCkey_set("COMMENT",  "a_{lm} coefficients");

	FHupdate(fp_out);
	FHfree();

	FCkey_set("XTENSION",  "BINTABLE");
	FSkey_set("BITPIX",  8);
	FSkey_set("NAXIS",   2);
	FLkey_set("NAXIS1",  nsize);
	FLkey_set("NAXIS2",  nlm);
	FLkey_set("PCOUNT",  0);
	FLkey_set("GCOUNT",  1);

	FLkey_set("TFIELDS", tfields);
	FCkey_set("EXTNAME", "A_lm DATA");

	FCkey_set("TFORM1",  tform1); /* 1st field: index        */
	FCkey_set("TTYPE1",  ttype1);
	FCkey_set("TUNIT1",  tunit1);
	FCkey_set("TFORM2",  tform2); /* 2nd field: real a_lm    */
	FCkey_set("TTYPE2",  ttype2);
	FCkey_set("TUNIT2",  tunit2);
	FCkey_set("TFORM3",  tform3); /* 3d field:  imagine a_lm */
	FCkey_set("TTYPE3",  ttype3);
	FCkey_set("TUNIT3",  tunit3);

	FCkey_set("OBJECT",  "A_LM");
	FLkey_set("LMAX",    lmax_print);
	FCkey_set("ORIGIN",  "GLESP");

	FHupdate(fp_out);
	FHfree();

	for (l=0; l<=lmax_print; l++) {
		register int ll=l*l+l+1;
		for (m=0; m<=l; m++) {
			double v = ll+m;
			int bb_r = m*nbuf + l;
			int bb_i = (m+lmax+1)*nbuf + l;
			bitpix = bitp1;
			put_pix_fits(fp_out, v);
			bitpix = bitp2;
			v = alm [bb_r];
			put_pix_fits(fp_out, v);
			bitpix = bitp3;
			v = alm [bb_i];
			put_pix_fits(fp_out, v);
		}
	}

	block_data = nlm * nsize;
	if(block_data % BLOCK_SIZE != 0) {
	       register int i;
	       long len = BLOCK_SIZE * (block_data / BLOCK_SIZE + 1);
	       len -= block_data;
	       for(i=0; i<len; i++)   fputc('\0', fp_out);
	}

	return;
}


/*
 * read alm, and e- & b- polarization modes
 */
void read_ieb_alm_T(alm_i, alm_e, alm_b, lmax, nbuf, filename)
double **alm_i, **alm_e, **alm_b;
int *lmax;
int *nbuf;      /* if ( *nbuf < 0)  *nbuf = -(*nbuf); */
char *filename;
{
    register int i, l, m;
    int tfields = 0;
    int bitp1, bitp2, bitp3;
    char tform1[30],
	 tform2[30],
	 tform3[30];
    char str[161];
    int  lmx = -1, lmax_input = *lmax, lmax_data, nalm;
    FILE *fp;
    int naxis=0, naxis1=0, naxis2=0;
    int block_data = 0;

//system("free");
//PR(ER, "rrrr1----------------------------------\n");
    /* detect maximum l */
    lmax_data = *lmax = Fget_lmax(filename);

    nalm = 2 * *lmax;
    if ( *nbuf < 0)  *nbuf = -(*nbuf); else *nbuf = 2 * (*lmax + 1);

    /* take memory */
    if( *alm_i == (double*)0) {
	if((*alm_i=(double*)calloc((unsigned)(*nbuf*nalm),sizeof **alm_i))==
	    (double *)0) {
nomem:
		PR(ER, "%s: no memory for arrays of %d elements...\n",
			myname, *nbuf*nalm);
		exit(1);
	}
    }
    if( *alm_e == (double*)0) {
	if((*alm_e =(double*)calloc((unsigned)(*nbuf*nalm), sizeof **alm_e))
	     == (double *)0) goto nomem;
    }
    if( *alm_b == (double*)0) {
	if((*alm_b =(double*)calloc((unsigned)(*nbuf*nalm), sizeof **alm_b))
	     == (double *)0) goto nomem;
    }

    /* open file */
    if ((fp = fopen(filename, "r")) == (FILE *)0) {
	    PR(ER, "%s: can not open file '%s'\n", myname, filename);
	    exit(1);
    }
    /* read 1st head */
    FHinit(fp);

    if (FEVget("EXTEND") != NULL)
	strcpy(str, FEVget("EXTEND"));

    if (strcmp(str, "T") != 0) {
	PR(ER, "%s: bad extension in file '%s'\n", myname, filename);
	exit(1);
    }

    /* Free memory from 1st head */
    FHfree();

    /****************************************/
    /* read basic keywods from the 2nd head */
    /****************************************/
    FHinit(fp);

    if (FEVget("XTENSION") != NULL)
	strcpy(str, FEVget("XTENSION"));
    else {
	PR(ER, "%s: no extension in file '%s'\n", myname, filename);
	exit(1);
    }
    if (strcmp(str, "BINTABLE") != 0) {
	PR(ER, "%s: can work only with BINTABLE extension\n", myname);
	exit(1);
    }
    if (FEVget("NAXIS") != NULL)
	    naxis  = FSkey_get("NAXIS");
    if (FEVget("NAXIS1") != NULL)
	    naxis1  = FLkey_get("NAXIS1");
    if (FEVget("NAXIS2") != NULL)
	    naxis2  = FLkey_get("NAXIS2");

    if (FEVget("TFIELDS") != NULL)
	    tfields = FLkey_get("TFIELDS");
    if (tfields != 3) {
	PR(ER, "%s: TFILEDS value must be equal to 3 in a_lm file '%s'\n",
	    myname, filename);
	exit(1);
    }

    if (FEVget("TFORM1") != NULL) {
	    strcpy(tform1, FEVget("TFORM1"));
	    if (strcmp(tform1, "1J") == 0) {
		bitp1 = LONGPIX;
	    } else if (strcmp(tform1, "1I") == 0) {
		bitp1 = SHORTPIX;
	    } else {
		PR(ER, "%s: bad value for TFORM1. Sorry...\n", myname);
		exit(1);
	    }
    } else {
	    PR(ER, "%s: no parameter TFORM1. Sorry...\n", myname);
	    exit(1);
    }
    if (FEVget("TFORM2") != NULL) {
	    strcpy(tform2, FEVget("TFORM2"));
	    if (strcmp(tform2, "1E") == 0) {
		bitp2 = FLOATPIX;
	    } else if (strcmp(tform2, "1D") == 0) {
		bitp2 = DOUBLEPIX;
	    } else {
		PR(ER, "%s: bad value for TFORM2. Sorry...\n", myname);
		exit(1);
	    }
    } else {
	    PR(ER, "%s: no parameter TFORM2. Sorry...\n", myname);
	    exit(1);
    }
    if (FEVget("TFORM3") != NULL) {
	    strcpy(tform3, FEVget("TFORM3"));
	    if (strcmp(tform2, "1E") == 0) {
		bitp3 = FLOATPIX;
	    } else if (strcmp(tform3, "1D") == 0) {
		bitp3 = DOUBLEPIX;
	    } else {
		PR(ER, "%s: bad value for TFORM3. Sorry...\n", myname);
		exit(1);
	    }
    } else {
	    PR(ER, "%s: no parameter TFORM3. Sorry...\n", myname);
	    exit(1);
    }
    if (FEVget("LMAX") != NULL) {
	    *lmax = atoi(FEVget("LMAX"));
    }

    if (lmax_input > 0 && lmax_input <= *lmax) {
	 *lmax = lmax_input;
    } else {
	 if (*nbuf < (*lmax + 1)*2) {
	     if ( (lmax_input + 1) * 2 > *nbuf)
		 lmax_input = *nbuf / 2 - 1;
	     *lmax = lmax_input;
//         } else {
//           if (*lmax > lmax_input) {
//             PR(ER, "%s: check set of lmax: lmax=%d, nbuf=%d\n",
//                    myname, *lmax, *nbuf);
//             exit(1);
//           }
	 }
    }

    /* Free memory from 1nd head */
    FHfree();

    /* get values from file */
      for (i=0; i<naxis2; i++) {
	static int flag = 1;
	register int bb, nlm;

	bitpix = bitp1;
	nlm = get_pix_fits(fp)+0.5;
	l = (int)sqrt((double)(nlm-1));

	if (l > *lmax)
	    break;

	if (l*2+1 > *nbuf) {
	    if (flag) {
		PR(ER, "%s: attention ! l(=%d) is larger ", myname, l);
		PR(ER, "than array dimension (2*l+1>nbuf(%d)).\n", *nbuf);
	    }
	    flag = 0;
	    break;
	}
	m = nlm - 1 - l*l - l;
	bb = m * *nbuf;

	if (m < 0) {
	    PR(ER, "%s: attention ! bad number of m=%d for l=%d.\n",
		  myname, m, l);
	    exit(1);
	}
	if (m > l) {
	    PR(ER, "%s: attention ! m=%d > l=%d.\n", myname, m, l);
	    exit(1);
	}


	/* Real part */
	bitpix = bitp2;
	(*alm_i) [bb+l]        = get_pix_fits(fp);
	if (feof(fp)) {
	     PR(ER, "%s: unexpected end of file at l=%d and m=%d\n",
		myname, l, m);
	     break;
	}
	/* Imaginary part */
	bitpix = bitp3;
	(*alm_i) [bb+l+ (*lmax+1)* *nbuf] = get_pix_fits(fp);
	if (feof(fp)) {
	     PR(ER, "%s: unexpected end of file at l=%d and m=%d\n",
		myname, l, m);
	     break;
	}
	if (l > lmx)   lmx = l;
    }
    if (*lmax <= 0)  *lmax = lmx;

    /* skip remained */
    if (*lmax < lmax_data) {
	bitpix = bitp2; get_pix_fits(fp);
	bitpix = bitp3; get_pix_fits(fp);
	i++;
	for (; i<naxis2; i++) {
	    bitpix = bitp1; get_pix_fits(fp);
	    bitpix = bitp2; get_pix_fits(fp);
	    bitpix = bitp3; get_pix_fits(fp);
	}
    }

    block_data = naxis1*naxis2;
    if(block_data % BLOCK_SIZE != 0) {
	   register int i;
	   long len = BLOCK_SIZE * (block_data / BLOCK_SIZE + 1);
	   len -= block_data;
	   for(i=0; i<len; i++)   fgetc(fp);
    }


    /**********************/
    /* read 2nd extension */
    /**********************/
    FHinit(fp);
    if (FEVget("XTENSION") != NULL)
	strcpy(str, FEVget("XTENSION"));
    else {
	PR(ER, "%s: no 2nd extension in file '%s'\n", myname, filename);
	exit(1);
    }
    if (strcmp(str, "BINTABLE") != 0) {
	PR(ER, "%s: can work only with BINTABLE extension\n", myname);
	exit(1);
    }
    if (FEVget("NAXIS") != NULL)
	    naxis  = FSkey_get("NAXIS");
    if (FEVget("NAXIS1") != NULL)
	    naxis1  = FLkey_get("NAXIS1");
    if (FEVget("NAXIS2") != NULL)
	    naxis2  = FLkey_get("NAXIS2");

    if (FEVget("TFIELDS") != NULL)
	    tfields = FLkey_get("TFIELDS");
    if (tfields != 3) {
	PR(ER, "%s: TFILEDS value must be equal to 3 in a_lm file '%s'\n",
	    myname, filename);
	exit(1);
    }

    if (FEVget("TFORM1") != NULL) {
	    strcpy(tform1, FEVget("TFORM1"));
	    if (strcmp(tform1, "1J") == 0) {
		bitp1 = LONGPIX;
	    } else if (strcmp(tform1, "1I") == 0) {
		bitp1 = SHORTPIX;
	    } else {
		PR(ER, "%s: bad value for TFORM1. Sorry...\n", myname);
		exit(1);
	    }
    } else {
	    PR(ER, "%s: no parameter TFORM1. Sorry...\n", myname);
	    exit(1);
    }
    if (FEVget("TFORM2") != NULL) {
	    strcpy(tform2, FEVget("TFORM2"));
	    if (strcmp(tform2, "1E") == 0) {
		bitp2 = FLOATPIX;
	    } else if (strcmp(tform2, "1D") == 0) {
		bitp2 = DOUBLEPIX;
	    } else {
		PR(ER, "%s: bad value for TFORM2. Sorry...\n", myname);
		exit(1);
	    }
    } else {
	    PR(ER, "%s: no parameter TFORM2. Sorry...\n", myname);
	    exit(1);
    }
    if (FEVget("TFORM3") != NULL) {
	    strcpy(tform3, FEVget("TFORM3"));
	    if (strcmp(tform2, "1E") == 0) {
		bitp3 = FLOATPIX;
	    } else if (strcmp(tform3, "1D") == 0) {
		bitp3 = DOUBLEPIX;
	    } else {
		PR(ER, "%s: bad value for TFORM3. Sorry...\n", myname);
		exit(1);
	    }
    } else {
	    PR(ER, "%s: no parameter TFORM3. Sorry...\n", myname);
	    exit(1);
    }
    if (FEVget("LMAX") != NULL) {
	    *lmax = atoi(FEVget("LMAX"));
    }

    if (lmax_input > 0 && lmax_input <= *lmax) {
	 *lmax = lmax_input;
    } else {
	 if (*nbuf < (*lmax + 1)*2) {
	     if ( (lmax_input + 1) * 2 > *nbuf)
		 lmax_input = *nbuf / 2 - 1;
	     *lmax = lmax_input;
//         } else {
//           if (*lmax > lmax_input) {
//             PR(ER, "%s: check set of lmax: lmax=%d, nbuf=%d\n",
//                    myname, *lmax, *nbuf);
//             exit(1);
//           }
	 }
    }

    /* Free memory from 1nd head */
    FHfree();

    /* get values from file */
    for (i=0; i<naxis2; i++) {
	static int flag = 1;
	register int bb, nlm;

	bitpix = bitp1;
	nlm = get_pix_fits(fp)+0.5;
	l = (int)sqrt((double)(nlm-1));

	if (l > *lmax)
	    break;

	if (l*2+1 > *nbuf) {
	    if (flag) {
		PR(ER, "%s: attention ! l(=%d) is larger ", myname, l);
		PR(ER, "than array dimension (2*l+1>nbuf(%d)).\n", *nbuf);
	    }
	    flag = 0;
	    break;
	}
	m = nlm - 1 - l*l - l;
	bb = m * *nbuf;

	/* Real part */
	bitpix = bitp2;
	(*alm_e) [bb+l]        = get_pix_fits(fp);
	if (feof(fp)) {
	     PR(ER, "%s: unexpected end of file at l=%d and m=%d\n",
		myname, l, m);
	     break;
	}
	/* Imagine part */
	bitpix = bitp3;
	(*alm_e) [bb+l+ (*lmax+1)* *nbuf] = get_pix_fits(fp);
	if (feof(fp)) {
	     PR(ER, "%s: unexpected end of file at l=%d and m=%d\n",
		myname, l, m);
	     break;
	}
	if (l > lmx)   lmx = l;
    }
    if (*lmax <= 0)  *lmax = lmx;

    /* skip remained */
    if (*lmax < lmax_data) {
	bitpix = bitp2; get_pix_fits(fp);
	bitpix = bitp3; get_pix_fits(fp);
	i++;
	for (; i<naxis2; i++) {
	    bitpix = bitp1; get_pix_fits(fp);
	    bitpix = bitp2; get_pix_fits(fp);
	    bitpix = bitp3; get_pix_fits(fp);
	}
    }

    block_data = naxis1*naxis2;
    if(block_data % BLOCK_SIZE != 0) {
	   register int i;
	   long len = BLOCK_SIZE * (block_data / BLOCK_SIZE + 1);
	   len -= block_data;
	   for(i=0; i<len; i++)   fgetc(fp);
    }

    /*********************/
    /* read 3d extension */
    /*********************/
    FHinit(fp);
    if (FEVget("XTENSION") != NULL)
	strcpy(str, FEVget("XTENSION"));
    else {
	PR(ER, "%s: no 3d extension in file '%s'\n", myname, filename);
	exit(1);
    }
    if (strcmp(str, "BINTABLE") != 0) {
	PR(ER, "%s: can work only with BINTABLE extension\n", myname);
	exit(1);
    }
    if (FEVget("NAXIS") != NULL)
	    naxis  = FSkey_get("NAXIS");
    if (FEVget("NAXIS1") != NULL)
	    naxis1  = FLkey_get("NAXIS1");
    if (FEVget("NAXIS2") != NULL)
	    naxis2  = FLkey_get("NAXIS2");

    if (FEVget("TFIELDS") != NULL)
	    tfields = FLkey_get("TFIELDS");
    if (tfields != 3) {
	PR(ER, "%s: TFILEDS value must be equal to 3 in a_lm file '%s'\n",
	    myname, filename);
	exit(1);
    }

    if (FEVget("TFORM1") != NULL) {
	    strcpy(tform1, FEVget("TFORM1"));
	    if (strcmp(tform1, "1J") == 0) {
		bitp1 = LONGPIX;
	    } else if (strcmp(tform1, "1I") == 0) {
		bitp1 = SHORTPIX;
	    } else {
		PR(ER, "%s: bad value for TFORM1. Sorry...\n", myname);
		exit(1);
	    }
    } else {
	    PR(ER, "%s: no parameter TFORM1. Sorry...\n", myname);
	    exit(1);
    }
    if (FEVget("TFORM2") != NULL) {
	    strcpy(tform2, FEVget("TFORM2"));
	    if (strcmp(tform2, "1E") == 0) {
		bitp2 = FLOATPIX;
	    } else if (strcmp(tform2, "1D") == 0) {
		bitp2 = DOUBLEPIX;
	    } else {
		PR(ER, "%s: bad value for TFORM2. Sorry...\n", myname);
		exit(1);
	    }
    } else {
	    PR(ER, "%s: no parameter TFORM2. Sorry...\n", myname);
	    exit(1);
    }
    if (FEVget("TFORM3") != NULL) {
	    strcpy(tform3, FEVget("TFORM3"));
	    if (strcmp(tform2, "1E") == 0) {
		bitp3 = FLOATPIX;
	    } else if (strcmp(tform3, "1D") == 0) {
		bitp3 = DOUBLEPIX;
	    } else {
		PR(ER, "%s: bad value for TFORM3. Sorry...\n", myname);
		exit(1);
	    }
    } else {
	    PR(ER, "%s: no parameter TFORM3. Sorry...\n", myname);
	    exit(1);
    }
    if (FEVget("LMAX") != NULL) {
	    *lmax = atoi(FEVget("LMAX"));
    }

    if (lmax_input > 0 && lmax_input <= *lmax) {
	 *lmax = lmax_input;
    } else {
	 if (*nbuf < (*lmax + 1)*2) {
	     if ( (lmax_input + 1) * 2 > *nbuf)
		 lmax_input = *nbuf / 2 - 1;
	     *lmax = lmax_input;
//         } else {
//           if (*lmax > lmax_input) {
//             PR(ER, "%s: check set of lmax: lmax=%d, nbuf=%d\n",
//                    myname, *lmax, *nbuf);
//             exit(1);
//           }
	 }
    }

    /* Free memory from 1nd head */
    FHfree();

    /* get values from file */
    for (i=0; i<naxis2; i++) {
	static int flag = 1;
	register int bb, nlm;

	bitpix = bitp1;
	nlm = get_pix_fits(fp)+0.5;
	l = (int)sqrt((double)(nlm-1));

	if (l > *lmax)
	    break;

	if (l*2+1 > *nbuf) {
	    if (flag) {
		PR(ER, "%s: attention ! l(=%d) is larger ", myname, l);
		PR(ER, "than array dimension (2*l+1>nbuf(%d)).\n", *nbuf);
	    }
	    flag = 0;
	    break;
	}
	m = nlm - 1 - l*l - l;
	bb = m * *nbuf;

	/* Real part */
	bitpix = bitp2;
	(*alm_b) [bb+l]        = get_pix_fits(fp);
	if (feof(fp)) {
	     PR(ER, "%s: unexpected end of file at l=%d and m=%d\n",
		myname, l, m);
	     break;
	}
	/* Imagine part */
	bitpix = bitp3;
	(*alm_b) [bb+l+ (*lmax+1) * *nbuf] = get_pix_fits(fp);
	if (feof(fp)) {
	     PR(ER, "%s: unexpected end of file at l=%d and m=%d\n",
		myname, l, m);
	     break;
	}
	if (l > lmx)   lmx = l;
    }
    if (*lmax <= 0)  *lmax = lmx;

    fclose(fp);
    return;
}

void  add_alm_ext_T(alm1, alm2, lmax, nbuf, lmax_print, fp_out)
double *alm1, *alm2;
int lmax, lmax_print;
int nbuf;
FILE *fp_out;
{
	register int l, m, bb;
	int nlm = (lmax_print+1)*(lmax_print+2) / 2;
	int tfields = 3, nsize = 0;
	char ttype1[30], tform1[30], tunit1[30],
	     ttype2[30], tform2[30], tunit2[30],
	     ttype3[30], tform3[30], tunit3[30];
	int bitp1, bitp2, bitp3;
	int block_data = 0;

	strcpy(tform1, "1J");
	strcpy(ttype1,  "INDEX");
	strcpy(tunit1,  "UNITLESS");

	strcpy(tform2, "1E");
	strcpy(ttype2,  "REAL");
	strcpy(tunit2,  "mK");
	strcpy(tform3, "1E");
	strcpy(ttype3,  "IMAG");
	strcpy(tunit3,  "mK");

	/* TFORM 1 */
	bitp1=LONGPIX;
	if (strcmp(tform1, "1J") == 0) {
		nsize += 4;
	} else if (strcmp(tform1, "1I") == 0) {
		nsize += 2;
		sprintf(tform1, "1I");
		bitp1=SHORTPIX;
	}
	/* TFORM 2 */
	bitp2=bitpix_map;
	if (strcmp(tform2, "1E") == 0) {
		nsize += 4;
		bitp2=FLOATPIX;
	} else if (strcmp(tform2, "1D") == 0) {
		nsize += 8;
		bitp2=DOUBLEPIX;
	}
	/* TFORM 3 */
	bitp3=bitpix_map;
	if (strcmp(tform3, "1E") == 0) {
		nsize += 4;
		bitp2=FLOATPIX;
	} else if (strcmp(tform3, "1D") == 0) {
		nsize += 8;
		bitp3=DOUBLEPIX;
	}
	/***************************/
	/* first polarization mode */
	/***************************/
	FCkey_set("XTENSION",  "BINTABLE");
	FSkey_set("BITPIX",  8);
	FSkey_set("NAXIS",   2);
	FLkey_set("NAXIS1",  nsize);
	FLkey_set("NAXIS2",  nlm);
	FLkey_set("PCOUNT",  0);
	FLkey_set("GCOUNT",  1);

	FLkey_set("TFIELDS", tfields);
	FCkey_set("EXTNAME", "A_lm DATA");

	FCkey_set("TFORM1",  tform1); /* 1st field: index        */
	FCkey_set("TTYPE1",  ttype1);
	FCkey_set("TUNIT1",  tunit1);
	FCkey_set("TFORM2",  tform2); /* 2nd field: real a_lm    */
	FCkey_set("TTYPE2",  ttype2);
	FCkey_set("TUNIT2",  tunit2);
	FCkey_set("TFORM3",  tform3); /* 3d field:  imagine a_lm */
	FCkey_set("TTYPE3",  ttype3);
	FCkey_set("TUNIT3",  tunit3);

	FCkey_set("OBJECT",  "A_LM");
	FLkey_set("LMAX",    lmax_print);
	FCkey_set("ORIGIN",  "GLESP");

	FHupdate(fp_out);
	FHfree();

	for (l=0; l<=lmax_print; l++) {
		register int ll=l*l+l+1;
		for (m=0; m<=l; m++) {
			float v = ll+m;
			int bb_r = m * nbuf + l;
			int bb_i = (m + lmax + 1) * nbuf + l;
			bitpix = bitp1;
			put_pix_fits(fp_out, v);
			bitpix = bitp2;
			v = alm1 [bb_r];
			put_pix_fits(fp_out, v);
			bitpix = bitp3;
			v = alm1 [bb_i];
			put_pix_fits(fp_out, v);
		}
	}

	block_data = nlm * nsize;
	if(block_data % BLOCK_SIZE != 0) {
	       register int i;
	       long len = BLOCK_SIZE * (block_data / BLOCK_SIZE + 1);
	       len -= block_data;
	       for(i=0; i<len; i++)   fputc('\0', fp_out);
	}


	/***************************/
	/* second polarization mode */
	/***************************/
	FCkey_set("XTENSION",  "BINTABLE");
	FSkey_set("BITPIX",  8);
	FSkey_set("NAXIS",   2);
	FLkey_set("NAXIS1",  nsize);
	FLkey_set("NAXIS2",  nlm);
	FLkey_set("PCOUNT",  0);
	FLkey_set("GCOUNT",  1);

	FLkey_set("TFIELDS", tfields);
	FCkey_set("EXTNAME", "A_lm DATA");

	FCkey_set("TFORM1",  tform1); /* 1st field: index        */
	FCkey_set("TTYPE1",  ttype1);
	FCkey_set("TUNIT1",  tunit1);
	FCkey_set("TFORM2",  tform2); /* 2nd field: real a_lm    */
	FCkey_set("TTYPE2",  ttype2);
	FCkey_set("TUNIT2",  tunit2);
	FCkey_set("TFORM3",  tform3); /* 3d field:  imagine a_lm */
	FCkey_set("TTYPE3",  ttype3);
	FCkey_set("TUNIT3",  tunit3);

	FCkey_set("OBJECT",  "A_LM");
	FLkey_set("LMAX",    lmax_print);
	FCkey_set("ORIGIN",  "GLESP");

	FHupdate(fp_out);
	FHfree();

	for (bb=l=0; l<=lmax_print; l++, bb+=nbuf) {
		register int ll=l*l+l+1;
		for (m=0; m<=l; m++) {
			float v = ll+m;
			int bb_r = m * nbuf + l;
			int bb_i = (m + lmax + 1) * nbuf + l;
			bitpix = bitp1;
			put_pix_fits(fp_out, v);
			bitpix = bitp2;
			v = alm2 [bb_r];
			put_pix_fits(fp_out, v);
			bitpix = bitp3;
			v = alm2 [bb_i];
			put_pix_fits(fp_out, v);
		}
	}

	block_data = nlm * nsize;
	if(block_data % BLOCK_SIZE != 0) {
	       register int i;
	       long len = BLOCK_SIZE * (block_data / BLOCK_SIZE + 1);
	       len -= block_data;
	       for(i=0; i<len; i++)   fputc('\0', fp_out);
	}

	return;
}

void write_ieb_alm_T(alm_i, alm_e, alm_b, lmax, nbuf, lmax_print, fp_out)
double *alm_i, *alm_e, *alm_b;
int lmax, lmax_print;
int nbuf;
FILE *fp_out;
{
	out_glesp_alm_T(alm_i, lmax, nbuf, lmax_print, fp_out);
	add_alm_ext_T(alm_e, alm_b, lmax, nbuf, lmax_print, fp_out);
	return;
}



void read_basic_fits(bfits_file, arr, nax1, nax2, crv1, crv2, cdlt1, cdlt2)
char *bfits_file;
double **arr;
double *crv1, *crv2, *cdlt1, *cdlt2;
int *nax1, *nax2;
{
    register int i, j;
    int layout = 1;
    int nx1=0, nx2=0, nx3=0;
    double cv1=0.0, cv2=0.0, cp1=0.0, cp2=0.0, cdt1=0.0, cdt2=0.0;
    char ct1[51], ct2[51], bunit[51];
    double b_scale = 1.0, b_zero = 0.0;
    FILE *fp_bf;

    /* open input Basic FITS file */
    if ((fp_bf = fopen(bfits_file, "r")) == (FILE *)0) {
	PR(ER, "%s: can not open file '%s'.\n", myname, bfits_file);
	exit(1);
    }
    /*****************/
    /* read keywords */
    /*****************/
    FHinit(fp_bf);
    if (FEVget("BITPIX") != NULL)
	bitpix = FLkey_get("BITPIX");
    else {
	PR(ER, "%s: bad FITS file '%s', no keyword BITPIX.\n",
	    myname, bfits_file);
	exit(1);
    }
    if (FEVget("NAXIS1") != NULL)
	nx1 = FLkey_get("NAXIS1");
    else {
	PR(ER, "%s: bad FITS file '%s', no keyword NAXIS1.\n",
	    myname, bfits_file);
	exit(1);
    }
    if (FEVget("NAXIS2") != NULL)
	nx2 = FLkey_get("NAXIS2");
    else {
	PR(ER, "%s: bad FITS file '%s', no keyword NAXIS2.\n",
	    myname, bfits_file);
	exit(1);
    }
    if (FEVget("NAXIS3") != NULL)
	nx3 = FLkey_get("NAXIS3");
    else {
//      PR(ER, "%s: bad FITS file '%s', no keyword NAXIS3.\n",
//          myname, bfits_file);
//      exit(1);
	nx3=1;
    }
    if (FEVget("CRVAL1") != NULL)
	cv1 = FDkey_get("CRVAL1");
    else {
	PR(ER, "%s: bad FITS file '%s', no keyword CRVAL1.\n",
	    myname, bfits_file);
	exit(1);
    }
    if (FEVget("CRVAL2") != NULL)
	cv2 = FDkey_get("CRVAL2");
    else {
	PR(ER, "%s: bad FITS file '%s', no keyword CRVAL2.\n",
	    myname, bfits_file);
	exit(1);
    }
    if (FEVget("CRPIX1") != NULL)
	cp1 = FDkey_get("CRPIX1");
    else {
	PR(ER, "%s: bad FITS file '%s', no keyword CRPIX1.\n",
	    myname, bfits_file);
	exit(1);
    }
    if (FEVget("CRPIX2") != NULL)
	cp2 = FDkey_get("CRPIX2");
    else {
	PR(ER, "%s: bad FITS file '%s', no keyword CRPIX2.\n",
	    myname, bfits_file);
	exit(1);
    }
    if (FEVget("CDELT1") != NULL)
	cdt1 = FDkey_get("CDELT1");
    else {
	PR(ER, "%s: bad FITS file '%s', no keyword CDELT1.\n",
	    myname, bfits_file);
	exit(1);
    }
    if (FEVget("CDELT2") != NULL)
	cdt2 = FDkey_get("CDELT2");
    else {
	PR(ER, "%s: bad FITS file '%s', no keyword CDELT2.\n",
	    myname, bfits_file);
	exit(1);
    }
    if (FEVget("CTYPE1") != NULL)
	strcpy(ct1, FEVget("CTYPE1"));
    else  ct1[0] = '\0';
    if (FEVget("CTYPE2") != NULL)
	strcpy(ct2, FEVget("CTYPE2"));
    else  ct2[0] = '\0';
    if (FEVget("BSCALE") != NULL)
	b_scale = FDkey_get("BSCALE");
    if (FEVget("BZERO") != NULL)
	b_zero = FDkey_get("BZERO");
    if (FEVget("BUNIT") != NULL)
	strcpy(bunit, FEVget("BUNIT"));
    else  bunit[0] = '\0';

    if (layout > nx3) {
	PR(ER, "%s: no layout %d in file '%s'.\n", myname,layout,bfits_file);
	exit(1);
    }

    /* catch memory */
    if((*arr = (double *)calloc((unsigned)(nx1*nx2), sizeof **arr)) ==
     (double *)0) {
	PR(ER, "%s: no memory for data array of (%dx%d).\n",myname,nx1,nx2);
	exit(1);
    }

    if (nx3 == 0)  nx3 = 1;

    /* skip layouts */
    if (layout > 1) {
	int k, kl=layout-1;
	for(k=0; k<kl; k++)
	    for(i=0; i<nx2; i++)
		for(j=0; j<nx1; j++)
		    get_pix_fits(fp_bf);
    }

    /* read data */
    for(i=0; i<nx2; i++) {
	int ii = nx1 * i;
	for(j=0; j<nx1; j++) {
	    (*arr)[ii+j] = b_zero + b_scale * get_pix_fits(fp_bf);
	}
    }

    /* close input file */
    fclose(fp_bf);

    /* free memory from header */
    FHfree();

    *crv1 = cv1 + (1.  - cp1) * cdt1;
    *crv2 = cv2 + (1.  - cp2) * cdt2;
    *cdlt1 = cdt1;
    *cdlt2 = cdt2;
    *nax1 = nx1;
    *nax2 = nx2;

    return;
}
