#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define PR      fprintf
#define ER      stderr
#define MAX(x,y)    ((x) > (y) ?  (x):(y))
#define ABS(x)      ((x) <  0  ? -(x):(x))

#include "gd.h"
#include "plot_gd.h"

/* PLOT library for GD-functions      */
/* Oleg Verkhodanov, TAC & SAO RAN    */
/*     last modification: 26/08/2005, */
/*     22/03/2009                     */

struct _crm_ {                 /* marker on map figure */
     double theta, phi;  /* coordinates of marker */
     int mark, size, color;
     double vpa;     /* angle  of polarization vector, when mark=-10 */
};


static gdImagePtr im_gd;
static int gd_color = 0;   /* current color */
static int *ct_gd;         /* color table   */
static int *inp_img;
static int nx_gd=0, ny_gd=0;
static int x_gd_cur=0, y_gd_cur=0;

void marker_gd(int, int);

void openpl_gd(x, nx, ny)
int *x;
int nx, ny;
{
  register int i, j;
  int *r = (int *)get_r_color(),
      *g = (int *)get_g_color(),
      *b = (int *)get_b_color();
  int nc = n_colors();

  inp_img = x;
  nx_gd = nx;
  ny_gd = ny;

  /* create a new image with appropriate dimensions */
  im_gd = gdImageCreate(nx, ny);

  /* allocate color table */
  ct_gd = (int *)malloc(nc*sizeof(int));
  for (i = 0; i < nc; i++) {
      ct_gd[i] = gdImageColorAllocate(im_gd, r[i], g[i], b[i]);
  }

  /* set pixel colors in image */
  for (i = 0; i < nx; i++) {
    for (j = 0; j < ny; j++) {
      int c = x[i + j*nx];
      gdImageSetPixel(im_gd, i, j, ct_gd[c]);
    }
  }
}

int get_nx_gd() {return nx_gd;}
int get_ny_gd() {return ny_gd;}

int check_gd_open()
{
   return (nx_gd > 0 ? 1 : 0);
}

void color_gd(color)
int color;
{
	gd_color = color;
}

void circle_gd(x, y, r)
int x, y, r;
{
	int s = 0, e = 360;
	int w = r, h = r;

	gdImageArc(im_gd, x, y, w, h, s, e, gd_color);

	x_gd_cur = x;
	y_gd_cur = y;
}

void closepl_gd()
{
  register int i, j;

  /* free color table */
  free((void *)ct_gd);
  ct_gd = (int *)0;

  for (i = 0; i < nx_gd; i++) {
    for (j = 0; j < ny_gd; j++) {
      inp_img[i + j * nx_gd] = gdImageGetPixel(im_gd, i, j);
    }
  }

  /* destroy image */
  gdImageDestroy(im_gd);
  im_gd = 0;
  nx_gd = ny_gd = 0;
}

void line_gd(x1, y1, x2, y2)
int x1, y1, x2, y2;
{
	gdImageLine(im_gd, x1, y1, x2, y2, gd_color);
	x_gd_cur = x2;
	y_gd_cur = y2;
}
void line_angle_gd(x, y, l, phi)
int x, y, l;
double phi;
{
	double r = 0.5*l, cp=r*cos(phi), sp=r*sin(phi);
	int x1 = x - cp, y1 = y - sp,
	    x2 = x + cp, y2 = y + sp;
	gdImageLine(im_gd, x1, y1, x2, y2, gd_color);
	x_gd_cur = x;
	y_gd_cur = y;
}

void line_ang2ang_gd(p1,t1, p2,t2)
double p1,t1, p2,t2;
{
}

extern void move_gd(int, int), cont_gd(int, int);
void vector_angle_gd(x, y, l, phi, r)
int x, y, l,
	  r; /* length of arrow */
double phi;
{
	double rl = 0.5*l, cp=rl*cos(phi), sp=rl*sin(phi);
	int x1 = x, y1 = y,
	    x2 = x + 2*cp, y2 = y + 2*sp;
	int dx1, dy1, dx2, dy2;
	double pi=M_PI, arrow_ang = 15 * pi/180;

	gdImageLine(im_gd, x1, y1, x2, y2, gd_color);

	/* arrow marker */
	dx1 = r * cos (phi - arrow_ang);
	dy1 = r * sin (phi - arrow_ang);
	dx2 = r * sin (pi/2 - phi - arrow_ang);
	dy2 = r * cos (pi/2 - phi - arrow_ang);
	move_gd(x2, y2);    cont_gd(x2-dx1, y2-dy1);
	move_gd(x2, y2);    cont_gd(x2-dx2, y2-dy2);

	x_gd_cur = x;
	y_gd_cur = y;
}

void move_gd(x, y)
int x, y;
{
	x_gd_cur = x;
	y_gd_cur = y;
}

void cont_gd(x, y)
int x, y;
{
	line_gd(x_gd_cur, y_gd_cur, x, y);
}

/* create GD-image and copy map to output file */
void marker2mol(crm,  n, lon_0)
/* coordinates and type of markers */
struct _crm_ *crm;
int n;
double lon_0;
{
    register int i, j, k;
    struct _cr_  {double theta, phi; int x, y;} *cr;
    int nn = nx_gd*ny_gd;
    int y_c, x_c, x_r, y_r, x_m, y_m;
    int flag_theta_direction = 1;
    double theta, phi;
    double pi = M_PI;
    double pmin= 1.e+30, pmax= -1.e+30;

    cr = (struct _cr_ *)calloc((unsigned)nn, sizeof *cr);
    if (cr == (struct _cr_ *)0) {
	PR(ER, "marker2mol: no memory for temporal array...\n");
	exit(1);
    }

    y_c = ny_gd / 2;
    x_c = nx_gd / 2;
    x_r = nx_gd / 2;
    y_r = ny_gd / 2;

    for(k=i=0; i<ny_gd; i++) {
	double v, u;

	v = (i - y_c) / (y_r/1.02);

	for(j=0; j<nx_gd; j++) {
	    u = 2. * (j - x_c) / (x_r/1.02);

	    theta = phi = 1.e+30;

	    if ( u*u/4. + v*v <= 1.) {
		double sqv = sqrt(1.-v*v);

		/* theta in [0, pi] */
		theta = pi/2. - asin(2./pi * (asin(v) + v*sqv));
		/* lon in [-pi,pi], the minus sign for astro convention */
		phi = -lon_0 - pi/2 * u / MAX(sqv, 1.e-6);
		/* phi in [0, 2pi] */
		if ( phi < 0.0 )     phi += 2. * pi;
		if ( phi > 2.*pi )   phi -= 2. * pi;

	    }

	    if (phi < 1.e+3){
		if (pmin > phi) pmin = phi;
		if (pmax < phi) pmax = phi;
	    }
	    cr[k].theta = theta;
	    cr[k].phi   = phi;
	    cr[k].y     = i;
	    cr[k].x     = j;
	    k++;
	}
    }

    for(k=0; k<n; k++) {
	x_m = y_m = -1;
	theta = crm[k].theta;           /*  coordinates of markers */
	phi   = crm[k].phi;             /*                         */
	if (phi < 0.)     phi += 2.*pi;
	if (phi > 2.*pi)  phi -= 2.*pi;

	if (phi > pmax)  phi = 0.999 * pmax;

	/* find closest pixel on image */
	for(i=0; i<ny_gd; i++) {
	    int ii = i * nx_gd;        /* beginning of the row */
	    int ic = i * nx_gd + x_c;  /* middle of the row    */
	    int ic1= ic + nx_gd;

	    if (i+1 >= ny_gd)   /* last line */
		   break;

	    /* we are still out of picture ellipse */
	    if (cr[ic].theta > 1.e5)  continue;

	    if (flag_theta_direction) {
		 /* we did not reach our line yet */
		 /* ...t0...>...t1...>...theta....*/
		 if (cr[ic].theta > theta && cr[ic1].theta > theta)
			  /* we are not yet in theta range */
			  continue;

		 /* we are out of our line range */
		 /* theta...>...t0...>...t1...   */
		 if (theta > cr[ic].theta)
			  break;
	    } else {
		 /* we did not reach our line yet */
		 if (cr[ic1].theta < theta && cr[ic].theta < theta)
			  /* we are not yet in theta range */
			  continue;

		 /* we are out of our line range */
		 if (theta < cr[ic].theta)
			  break;
	    }
	    /* we now in theta range: */
	    /* if (cr[ic].theta < theta && theta < cr[ic1].theta) */

	    for(j=0; j<nx_gd; j++) {
	       int jj = ii+j;
	       if (cr[jj].phi > 1.e5)  continue;

//               /* we are not yet in theta range */
//               if (phi <= pi) {  /* [pi, 0] */
//                        if (j > x_c) {
//                              break;
//                        }
//               } else {         /* [2pi, pi] */
//                        if (j < x_c) {
//                              j = x_c-1;
//                              continue;
//                        }
//               }
	       if (cr[jj].phi > phi)
			continue;

	       if ((cr[jj].phi <= phi && phi < cr[jj-1].phi) ||
		   (cr[jj].phi >= phi && phi > cr[jj-1].phi)) {
		    /* position of marker */
		    x_m= j;
		    y_m= fabs(cr[ic].theta-theta)<fabs(cr[ic1].theta-theta) ?
			     i : i+1;
		    break;
	       }
	    }
	}

	/* put marker */
	if (x_m >= 0 && y_m >=0) {
		color_gd(crm[k].color);
		if ( crm[k].mark == -10 )
		    line_angle_gd(x_m, y_m, crm[k].size, crm[k].vpa);
		else if ( crm[k].mark == -20 ) {
		    vector_angle_gd(x_m, y_m, crm[k].size, crm[k].vpa, 5);
		} else {
		    move_gd(x_m, y_m);
		    marker_gd(crm[k].size, crm[k].mark);
		}
	}
    }

    free((void *)cr);

    return;
}

/*
 *  draw marker:
 *
 *    0    - point;                         XPM_POINT
 *    1,x  - X;                             XPM_X
 *    2,+  - plus '+';                      XPM_CROSS
 *    3,o  - empty  circle;                 XPM_CIRCLE
 *    4,O  - filled circle;                 XPM_F_CIRCLE
 *    5    - empty  rhomb;                  XPM_RHOMB
 *    6    - filled rhomb;                  XPM_F_RHOMB
 *    7,s  - empty  square;                 XPM_SQUARE
 *    8,S  - filled square;                 XPM_F_SQUARE
 *    9,v  - empty  triangular down;        XPM_TRIANG_DOWN
 *   10,V  - filled triangular down;        XPM_F_TRIANG_DOWN
 *   11,^  - empty  triangular up;          XPM_TRIANG_UP
 *   12,   - filled triangular up;          XPM_F_TRIANG_UP
 *   13,>  - empty  triangular left;        XPM_TRIANG_LEFT
 *   14,   - filled triangular left;        XPM_F_TRIANG_LEFT
 *   15,<  - empty  triangular right;       XPM_TRIANG_RIGHT
 *   16,   - filled triangular right;       XPM_F_TRIANG_RIGHT
 *   17    - arrow down;                    XPM_ARROW_DOWN
 *   18    - arrow up;                      XPM_ARROW_UP
 *   19    - arrow left;                    XPM_ARROW_LEFT
 *   20    - arrow right;                   XPM_ARROW_RIGHT
 *
 *   21    - pentagon star                  XPM_PENT_STAR
 *   22    - filled pentagon star           XPM_F_PENT_STAR
 *   23,*  - asterisk                       XPM_ASTERISK
 *   24    - Solomon's star                 XPM_SOLOMON_STAR
 *   25    - filled Solomon's star          XPM_F_SOLOMON_STAR
 *
 *   35    - fat right-down arrow           XPM_FAT_RIGHT_DOWN_ARROW
 *   36    - filled fat right-down arrow    XPM_F_FAT_RIGHT_DOWN_ARROW
 *   37    - fat left-down arrow            XPM_FAT_LEFT_DOWN_ARROW
 *   38    - filled fat left-down arrow     XPM_F_FAT_LEFT_DOWN_ARROW
 *   39    - fat down arrow                 XPM_FAT_DOWN_ARROW
 *   40    - filled fat down arrow          XPM_F_FAT_DOWN_ARROW
 *
 */

void marker_gd(r, mr)
int r, mr;
{
	pointm_gd(x_gd_cur, y_gd_cur, r, mr );
}
#include <ctype.h>
int st2mark(st)
char *st;
{
    int m = XPM_POINT;
    char str[30];
    register char *cs = str, *ps = st;

    switch ( *st ) {
    case '+':
	    m = XPM_CROSS;
	    break;
    case '*':
	    m = XPM_ASTERISK;
	    break;
    case '^':
	    m = XPM_TRIANG_UP;
	    break;
    case '>':
	    m = XPM_TRIANG_LEFT;
	    break;
    case '<':
	    m = XPM_TRIANG_RIGHT;
	    break;
    case 'o':
	    m = XPM_CIRCLE;
	    break;
    case 's':
	    m = XPM_SQUARE;
	    break;
    case 'v':
	    m = XPM_TRIANG_DOWN;
	    break;
    case '0':
    case '1':
    case '2':
    case '3':
    case '4':
    case '5':
    case '6':
    case '7':
    case '8':
    case '9':
	    m = atoi(st);
	    break;
    default:
	    while ((*cs++ = toupper(*ps++))) ;

	    switch ( *str ) {
	    case 'A':
		    if (strcmp(str, "ARROW_DOWN" ) == 0)
			    m = XPM_ARROW_DOWN;
		    else if (strcmp(str, "ARROW_UP" ) == 0)
			    m = XPM_ARROW_UP;
		    else if (strcmp(str, "ARROW_LEFT" ) == 0)
			    m = XPM_ARROW_LEFT;
		    else if (strcmp(str, "ARROW_RIGHT" ) == 0)
			    m = XPM_ARROW_RIGHT;
		    else if (strcmp(str, "ASTERISK" ) == 0)
			    m = XPM_ASTERISK;
		    break;
	    case 'C':
		    if (strcmp(str, "CIRCLE" ) == 0)
			    m = XPM_CIRCLE;
		    else if (strcmp(str, "CROSS" ) == 0)
			    m = XPM_CROSS;
		    break;
	    case 'F':
		    if (strcmp(str, "F_CIRCLE" ) == 0)
			    m = XPM_F_CIRCLE;
		    else if (strcmp(str, "F_RHOMB" ) == 0)
			    m = XPM_F_RHOMB;
		    else if (strcmp(str, "XPM_F_SQUARE" ) == 0)
			    m = XPM_F_SQUARE;
		    else if (strcmp(str, "F_TRIANG_DOWN" ) == 0)
			    m = XPM_F_TRIANG_DOWN;
		    else if (strcmp(str, "F_TRIANG_UP" ) == 0)
			    m = XPM_F_TRIANG_UP;
		    else if (strcmp(str, "F_TRIANG_LEFT" ) == 0)
			    m = XPM_F_TRIANG_LEFT;
		    else if (strcmp(str, "F_TRIANG_RIGHT" ) == 0)
			    m = XPM_F_TRIANG_RIGHT;
		    else if (strcmp(str, "F_PENT_STAR" ) == 0)
			    m = XPM_F_PENT_STAR;
		    else if (strcmp(str, "F_SOLOMON_STAR" ) == 0)
			    m = XPM_F_SOLOMON_STAR;
		    else if (strcmp(str, "FAT_RIGHT_DOWN_ARROW" ) == 0)
			    m = XPM_FAT_RIGHT_DOWN_ARROW;
		    else if (strcmp(str, "F_FAT_RIGHT_DOWN_ARROW" ) == 0)
			    m = XPM_F_FAT_RIGHT_DOWN_ARROW;
		    else if (strcmp(str, "FAT_LEFT_DOWN_ARROW" ) == 0)
			    m = XPM_FAT_LEFT_DOWN_ARROW;
		    else if (strcmp(str, "F_FAT_LEFT_DOWN_ARROW" ) == 0)
			    m = XPM_F_FAT_LEFT_DOWN_ARROW;
		    else if (strcmp(str, "FAT_DOWN_ARROW" ) == 0)
			    m = XPM_FAT_DOWN_ARROW;
		    else if (strcmp(str, "F_FAT_DOWN_ARROW" ) == 0)
			    m = XPM_F_FAT_DOWN_ARROW;
		    break;
	    case 'P':
		    if (strcmp(str, "POINT" ) == 0)
			    m = XPM_POINT;
		    else if (strcmp(str, "PENT_STAR" ) == 0)
			    m = XPM_PENT_STAR;
		    else if (strncmp(str, "PENTAGON", 8) == 0)
			    m = XPM_PENT_STAR;
		    break;
	    case 'R':
		    if (strcmp(str, "RHOMB" ) == 0)
			    m = XPM_RHOMB;
		    break;
	    case 'S':
		    if (strcmp(str, "SQUARE" ) == 0)
			    m = XPM_SQUARE;
		    else if (strncmp(str, "SOLOM",5) == 0)
			    m = XPM_SOLOMON_STAR;
		    break;
	    case 'T':
		    if (strcmp(str, "TRIANG_DOWN" ) == 0)
			    m = XPM_TRIANG_DOWN;
		    else if (strcmp(str, "TRIANG_UP" ) == 0)
			    m = XPM_TRIANG_UP;
		    else if (strcmp(str, "TRIANG_LEFT" ) == 0)
			    m = XPM_TRIANG_LEFT;
		    else if (strcmp(str, "TRIANG_RIGHT" ) == 0)
			    m = XPM_TRIANG_RIGHT;
		    else if (strcmp(str, "TRIANGLE_DOWN" ) == 0)
			    m = XPM_TRIANG_DOWN;
		    else if (strcmp(str, "TRIANGLE_UP" ) == 0)
			    m = XPM_TRIANG_UP;
		    else if (strcmp(str, "TRIANGLE_LEFT" ) == 0)
			    m = XPM_TRIANG_LEFT;
		    else if (strcmp(str, "TRIANGLE_RIGHT" ) == 0)
			    m = XPM_TRIANG_RIGHT;
		    break;
	    case 'X':
		    if (strcmp(str, "X" ) == 0)
			    m = XPM_X;
		    else if (strcmp(str, "XPM_POINT" ) == 0)
			    m = XPM_POINT;
		    else if (strcmp(str, "XPM_X" ) == 0)
			    m = XPM_X;
		    else if (strcmp(str, "XPM_CROSS" ) == 0)
			    m = XPM_CROSS;
		    else if (strcmp(str, "XPM_CIRCLE" ) == 0)
			    m = XPM_CIRCLE;
		    else if (strcmp(str, "XPM_F_CIRCLE" ) == 0)
			    m = XPM_F_CIRCLE;
		    else if (strcmp(str, "XPM_RHOMB" ) == 0)
			    m = XPM_RHOMB;
		    else if (strcmp(str, "XPM_F_RHOMB" ) == 0)
			    m = XPM_F_RHOMB;
		    else if (strcmp(str, "XPM_SQUARE" ) == 0)
			    m = XPM_SQUARE;
		    else if (strcmp(str, "XPM_F_SQUARE" ) == 0)
			    m = XPM_F_SQUARE;
		    else if (strcmp(str, "XPM_TRIANG_DOWN" ) == 0)
			    m = XPM_TRIANG_DOWN;
		    else if (strcmp(str, "XPM_F_TRIANG_DOWN" ) == 0)
			    m = XPM_F_TRIANG_DOWN;
		    else if (strcmp(str, "XPM_TRIANG_UP" ) == 0)
			    m = XPM_TRIANG_UP;
		    else if (strcmp(str, "XPM_F_TRIANG_UP" ) == 0)
			    m = XPM_F_TRIANG_UP;
		    else if (strcmp(str, "XPM_TRIANG_LEFT" ) == 0)
			    m = XPM_TRIANG_LEFT;
		    else if (strcmp(str, "XPM_F_TRIANG_LEFT" ) == 0)
			    m = XPM_F_TRIANG_LEFT;
		    else if (strcmp(str, "XPM_TRIANG_RIGHT" ) == 0)
			    m = XPM_TRIANG_RIGHT;
		    else if (strcmp(str, "XPM_F_TRIANG_RIGHT" ) == 0)
			    m = XPM_F_TRIANG_RIGHT;
		    else if (strcmp(str, "XPM_ARROW_DOWN" ) == 0)
			    m = XPM_ARROW_DOWN;
		    else if (strcmp(str, "XPM_ARROW_UP" ) == 0)
			    m = XPM_ARROW_UP;
		    else if (strcmp(str, "XPM_ARROW_LEFT" ) == 0)
			    m = XPM_ARROW_LEFT;
		    else if (strcmp(str, "XPM_ARROW_RIGHT" ) == 0)
			    m = XPM_ARROW_RIGHT;
		    else if (strcmp(str, "XPM_PENT_STAR" ) == 0)
			    m = XPM_PENT_STAR;
		    else if (strcmp(str, "XPM_F_PENT_STAR" ) == 0)
			    m = XPM_F_PENT_STAR;
		    else if (strcmp(str, "XPM_ASTERISK" ) == 0)
			    m = XPM_ASTERISK;
		    else if (strcmp(str, "XPM_SOLOMON_STAR" ) == 0)
			    m = XPM_SOLOMON_STAR;
		    else if (strcmp(str, "XPM_F_SOLOMON_STAR" ) == 0)
			    m = XPM_F_SOLOMON_STAR;
		    else if (strcmp(str, "XPM_FAT_RIGHT_DOWN_ARROW" ) == 0)
			    m = XPM_FAT_RIGHT_DOWN_ARROW;
		    else if (strcmp(str, "XPM_F_FAT_RIGHT_DOWN_ARROW" ) == 0)
			    m = XPM_F_FAT_RIGHT_DOWN_ARROW;
		    else if (strcmp(str, "XPM_FAT_LEFT_DOWN_ARROW" ) == 0)
			    m = XPM_FAT_LEFT_DOWN_ARROW;
		    else if (strcmp(str, "XPM_F_FAT_LEFT_DOWN_ARROW" ) == 0)
			    m = XPM_F_FAT_LEFT_DOWN_ARROW;
		    else if (strcmp(str, "XPM_FAT_DOWN_ARROW" ) == 0)
			    m = XPM_FAT_DOWN_ARROW;
		    else if (strcmp(str, "XPM_F_FAT_DOWN_ARROW" ) == 0)
			    m = XPM_F_FAT_DOWN_ARROW;
		    break;
	}
	break;
    }
    return m;
}

void pointm_gd(x, y, r, mr)
register int x,y;  /* coordinates of central point of marker */
register int r;    /* size of marker by X                    */
int mr;            /* number of marker                       */
{
	register int rr = r; /* * aspect_ratio_XP; */
	int xy[20];

	r  /= 2;
	rr /= 2;

	switch (mr) {
	case XPM_POINT:
		point_gd(x,y);
		if (r > 1) {
		    fcircle_gd(x, y, r);
		}
		break;
	case 'x': /* cross 'x' */
	case XPM_X:
		move_gd(x-r, y+rr);
		cont_gd(x+r, y-rr);
		move_gd(x-r, y-rr);
		cont_gd(x+r, y+rr);
		break;
	case XPM_CROSS:  /* plus '+' */
		move_gd(x, y-rr);
		cont_gd(x, y+rr);
		move_gd(x-r, y);
		cont_gd(x+r, y);
		break;
	case XPM_CIRCLE:  /* empty  circle */
/*
	    {
		int k = 0;
		for(i=0; i<=2*r; i++)
			for(j=0; j<=2*r; j++)
				if(-1<=(k=Q(i+j-r)-2*i*j) && k<=1)
					point(x-r+i,y-r+j);
		k = 1;
		break;
	    }
*/
		circle_gd(x, y, r);
		break;
	case XPM_F_CIRCLE:  /* filled circle */
/*
		for(i=0; i<=2*r; i++)
			for(j=0; j<=2*r; j++)
				if(Q(i+j-r)-2*i*j <= 0)
					point(x-r+i,y-r+j);
*/
		fcircle_gd(x, y, r);
		break;
	case XPM_RHOMB:    /* empty  rhomb */
	case XPM_F_RHOMB:  /* filled rhomb */
		xy[0] = x;   xy[1] = y-rr;
		xy[2] = x-r; xy[3] = y;
		xy[4] = x;   xy[5] = y+rr;
		xy[6] = x+r; xy[7] = y;
		xy[8] = x;   xy[9] = y-rr;
		if (mr == XPM_F_RHOMB)
			f_pol_gd (xy, 5);
		else
			p_line_gd (xy, 5);
		break;
	case XPM_SQUARE:    /* empty  square */
	case XPM_F_SQUARE:  /* filled square */
	      { int r1 = r, rr1 = rr;
		xy[0] = x-r1; xy[1] = y-rr1;
		xy[2] = x-r1; xy[3] = y+rr1;
		xy[4] = x+r1; xy[5] = y+rr1;
		xy[6] = x+r1; xy[7] = y-rr1;
		xy[8] = x-r1; xy[9] = y-rr1;
		if (mr == XPM_F_SQUARE)
			f_pol_gd (xy, 5);
		else
			p_line_gd (xy, 5);
	      }
		break;
	case XPM_TRIANG_DOWN:   /* empty  triangle down */
	case XPM_F_TRIANG_DOWN: /* filled triangle down */
		xy[0] = x-r; xy[1] = y+rr;
		xy[2] = x+r; xy[3] = y+rr;
		xy[4] = x;   xy[5] = y-rr;
		xy[6] = x-r; xy[7] = y+rr;
		if (mr == XPM_F_TRIANG_DOWN)
			f_pol_gd (xy, 4);
		else
			p_line_gd (xy, 4);
		break;
	case XPM_TRIANG_UP:   /* empty  triangle up */
	case XPM_F_TRIANG_UP: /* filled triangle up */
		xy[0] = x-r; xy[1] = y-rr;
		xy[2] = x+r; xy[3] = y-rr;
		xy[4] = x;   xy[5] = y+rr;
		xy[6] = x-r; xy[7] = y-rr;
		if (mr == XPM_F_TRIANG_UP)
			f_pol_gd (xy, 4);
		else
			p_line_gd (xy, 4);
		break;
	case XPM_TRIANG_RIGHT:  /* empty  triangle right */
	case XPM_F_TRIANG_RIGHT:/* filled triangle right */
		xy[0] = x-r; xy[1] = y;
		xy[2] = x+r; xy[3] = y+rr;
		xy[4] = x+r; xy[5] = y-rr;
		xy[6] = x-r; xy[7] = y;
		if (mr == XPM_F_TRIANG_RIGHT)
			f_pol_gd (xy, 4);
		else
			p_line_gd (xy, 4);
		break;
	case XPM_TRIANG_LEFT:    /* empty  triangle left  */
	case XPM_F_TRIANG_LEFT:  /* filled triangle left  */
		xy[0] = x+r; xy[1] = y;
		xy[2] = x-r; xy[3] = y+rr;
		xy[4] = x-r; xy[5] = y-rr;
		xy[6] = x+r; xy[7] = y;
		if (mr == XPM_F_TRIANG_LEFT)
			f_pol_gd(xy, 4);
		else
			p_line_gd(xy, 4);
		break;
	case XPM_ARROW_DOWN: /* arrow down */
		xy[0] = r/2;
		xy[1] = y - 2 * rr;
		xy[2] = y - 3 * rr;
		move_gd(x-r, y);    cont_gd(x+r, y);
		move_gd(x,   y);    cont_gd(x, xy[2]);
		move_gd(x-xy[0], xy[1]); cont_gd(x, xy[2]);
		move_gd(x+xy[0], xy[1]); cont_gd(x, xy[2]);
		break;
	case XPM_ARROW_UP: /* arrow   up */
		xy[0] = r/2;
		xy[1] = y + 2 * rr;
		xy[2] = y + 3 * rr;
		move_gd(x-r, y);    cont_gd(x+r, y);
		move_gd(x,   y);    cont_gd(x, xy[2]);
		move_gd(x-xy[0], xy[1]); cont_gd(x, xy[2]);
		move_gd(x+xy[0], xy[1]); cont_gd(x, xy[2]);
		break;
	case XPM_ARROW_LEFT: /* arrow left */
		xy[0] = rr/2;
		xy[1] = x - 2 * r;
		xy[2] = x - 3 * r;
		move_gd(x, y-rr);   cont_gd(x, y+rr);
		move_gd(x,   y);    cont_gd(xy[2], y);
		move_gd(xy[1], y-xy[0]); cont_gd(xy[2], y);
		move_gd(xy[1], y+xy[0]); cont_gd(xy[2], y);
		break;
	case XPM_ARROW_RIGHT: /* arrow right */
		xy[0] = rr/2;
		xy[1] = x + 2 * r;
		xy[2] = x + 3 * r;
		move_gd(x, y-rr);   cont_gd(x, y+rr);
		move_gd(x,   y);    cont_gd(xy[2], y);
		move_gd(xy[1], y-xy[0]); cont_gd(xy[2], y);
		move_gd(xy[1], y+xy[0]); cont_gd(xy[2], y);
		break;
	case XPM_PENT_STAR:   /* pentagon star        */
	case XPM_F_PENT_STAR: /* filled pentagon star */
	       {              /* sin */           /* cos */
		int r72 = r *   .951,   rr72 = rr *   .309 / 2,
		   r144 = r *   .588,  rr144 = rr * (-.809),
		   r216 = r * (-.588), rr216 = rr * (-.809),
		   r288 = r * (-.951), rr288 = rr *   .309 / 2;

		xy[0] = x;        xy[1] = y+r;
		xy[2] = x + r144; xy[3] = y + rr144;
		xy[4] = x + r288; xy[5] = y + rr288;
		xy[6] = x + r72;  xy[7] = y + rr72;
		xy[8] = x + r216; xy[9] = y + rr216;
		xy[10] = x;       xy[11] = y+r;
	       }
		if (mr == XPM_F_PENT_STAR)
			f_pol_gd (xy, 6);
		else
			p_line_gd (xy, 6);
		break;
	case XPM_ASTERISK: /* asterisk */
	      { int r2 = r*0.707, rr2= rr*0.707;
		move_gd(x-r2, y+rr2);
		cont_gd(x+r2, y-rr2);
		move_gd(x-r2, y-rr2);
		cont_gd(x+r2, y+rr2);
		move_gd(x, y-rr);
		cont_gd(x, y+rr);
		move_gd(x-r, y);
		cont_gd(x+r, y);
	      }
		break;
	case XPM_SOLOMON_STAR:   /* Solomon's star        */
	case XPM_F_SOLOMON_STAR: /* filled Solomon's star */
	    {
		int rr2 = rr / 2.5;

		xy[0] = x-r; xy[1] = y+rr-rr2;
		xy[2] = x+r; xy[3] = y+rr-rr2;
		xy[4] = x;   xy[5] = y-rr-rr2;
		xy[6] = x-r; xy[7] = y+rr-rr2;
		if (mr == XPM_F_SOLOMON_STAR)
			f_pol_gd (xy, 4);
		else
			p_line_gd (xy, 4);
		xy[0] = x-r; xy[1] = y-rr+rr2;
		xy[2] = x+r; xy[3] = y-rr+rr2;
		xy[4] = x;   xy[5] = y+rr+rr2;
		xy[6] = x-r; xy[7] = y-rr+rr2;
		if (mr == XPM_F_SOLOMON_STAR)
			f_pol_gd (xy, 4);
		else
			p_line_gd (xy, 4);
	     }
	     break;
	case XPM_FAT_RIGHT_DOWN_ARROW:   /* fat left-down arrow */
	case XPM_F_FAT_RIGHT_DOWN_ARROW: /* filled fat left-down arrow */
	    {
		int r3 = r/3, rr3 = rr/3;
		xy[0]  = x;         xy[1]  = y;        /*            */
		xy[2]  = x-r;       xy[3]  = y;        /*            */
		xy[4]  = x-r+r3;    xy[5]  = y+rr3;    /*    / \     */
		xy[6]  = x-r-r3;    xy[7]  = y+rr;     /*    \   \/| */
		xy[8]  = x-r;       xy[9]  = y+rr+rr3; /*      \   | */
		xy[10] = x-r3;      xy[11] = y+rr-rr3; /*     /____| */
		xy[12] = x;         xy[13] = y+rr;     /*            */
		xy[14] = x;         xy[15] = y;        /*            */

		if (mr == XPM_F_FAT_RIGHT_DOWN_ARROW)
			f_pol_gd (xy, 8);
		else
			p_line_gd (xy, 8);
	     }
	     break;
	case XPM_FAT_LEFT_DOWN_ARROW:   /* fat left-down arrow */
	case XPM_F_FAT_LEFT_DOWN_ARROW: /* filled fat left-down arrow */
	    {
		int r3 = r/3, rr3 = rr/3;
		xy[0]  = x;         xy[1]  = y;        /*            */
		xy[2]  = x+r;       xy[3]  = y;        /*            */
		xy[4]  = x+r-r3;    xy[5]  = y+rr3;    /*      / \   */
		xy[6]  = x+r+r3;    xy[7]  = y+rr;     /*  |\/   /   */
		xy[8]  = x+r;       xy[9]  = y+rr+rr3; /*  |   /     */
		xy[10] = x+r3;      xy[11] = y+rr-rr3; /*  |____\    */
		xy[12] = x;         xy[13] = y+rr;     /*            */
		xy[14] = x;         xy[15] = y;        /*            */

		if (mr == XPM_F_FAT_LEFT_DOWN_ARROW)
			f_pol_gd (xy, 8);
		else
			p_line_gd (xy, 8);
	     }
	     break;
	case XPM_FAT_DOWN_ARROW:   /* fat down arrow */
	case XPM_F_FAT_DOWN_ARROW: /* filled fat down arrow */
	    {
		int r2 = r/1.5;
		xy[0]  = x;         xy[1]  = y;        /*            */
		xy[2]  = x-r;       xy[3]  = y+rr;     /*     __     */
		xy[4]  = x-r+r2;    xy[5]  = y+rr;     /*    |  |    */
		xy[6]  = x-r+r2;    xy[7]  = y+rr+rr;  /*   _|  |_   */
		xy[8]  = x+r-r2;    xy[9]  = y+rr+rr;  /*   \    /   */
		xy[10] = x+r-r2;    xy[11] = y+rr;     /*     \/     */
		xy[12] = x+r;       xy[13] = y+rr;     /*            */
		xy[14] = x;         xy[15] = y;        /*            */

		if (mr == XPM_F_FAT_DOWN_ARROW)
			f_pol_gd (xy, 8);
		else
			p_line_gd (xy, 8);
	     }
	     break;
	default:
		point_gd(x, y);
		break;
	}
	move_gd(x,y);
	return;
}

void point_gd(x, y)
int x, y;
{
    gdImageSetPixel(im_gd, x, y, gd_color);
    return;
}

void f_pol_gd (xy, n)
int *xy, n;
{
    register int i;
    move_gd(xy[0], xy[1]);
    for(i=0; i<n; i++)
	  cont_gd(xy[2*i], xy[2*i+1]);
    return;
}

void p_line_gd (xy, n)
int *xy, n;
{
    register int i;
    int x=0, y=0;

    move_gd(xy[0], xy[1]);
    for(i=0; i<n; i++) {
	  int i2 = i*2;
	  x += xy[i2];
	  y += xy[i2+1];
	  cont_gd(xy[i2], xy[i2+1]);
    }
    x /= n;
    y /= n;
    gdImageFillToBorder(im_gd, x, y, gd_color, gd_color);
    return;
}

void fcircle_gd(x, y, r)
int x, y, r;
{
    int s = 0, e = 360;
    int w = r, h = r;

    gdImageArc(im_gd, x, y, w, h, s, e, gd_color);
    gdImageFillToBorder(im_gd, x, y, gd_color, gd_color);

    x_gd_cur = x;
    y_gd_cur = y;
    return;
}


/*
 *   color conversion procedures
 *   O.Verkhodanov, 19/07/2011
 *   conversion: RGB -> HSV
 *   H in [0, 360); S,V,R,G,B  in [0, 1]
 */

void rgb2hsv (r,g,b, h,s,v, min, max)
double r,g,b;
double *h, *s, *v;
double min, max;
{
    if (ABS(max-min) < 1.e-8)  /* max == min */
	*h = 0.0;
    else if (ABS(max-r) < 1.e-8)  /* max == r &&  g >= b */
	*h = 60. * (g-b)/(max-min);
    else if (ABS(max-r) < 1.e-8)  /* max == r &&  g < b */
	*h = 60. * (g-b)/(max-min) + 360.;
    else if (ABS(max-g) < 1.e-8)  /* max == g */
	*h = 60. * (b-r)/(max-min) + 120.;
    else if (ABS(max-b) < 1.e-8)  /* max == b */
	*h = 60. * (r-g)/(max-min) + 240.;

    if ( max < 1.e-8 ) /* max == 0 */
	*s = 0.0;
    else
	*s = 1 - min/max;

    *v = max;

    return;
}

/*
    conversion: RGB -> HSV
    H in [0, 360); S,V,R,G,B  in [0, 1]
*/
void hsv2rgb (h,s,v, r,g,b)
double h, s, v;
double *r, *g, *b;
{
    int h_i;
    double f, p, q, t;

    h_i = (int)(h/60) % 6;
    f = h/60. - (int)(h/60.);
    p = v * (1-s);
    q = v * (1-s*f);
    t = v * (1-(1-f)*s);

    switch (h_i) {
    case 0:
	    *r = v,  *g = t,  *b = p;
	    break;
    case 1:
	    *r = q,  *g = v,  *b = p;
	    break;
    case 2:
	    *r = p,  *g = v,  *b = t;
	    break;
    case 3:
	    *r = p,  *g = q,  *b = v;
	    break;
    case 4:
	    *r = t,  *g = p,  *b = v;
	    break;
    case 5:
	    *r = v,  *g = p,  *b = q;
	    break;
    }

    return;
}

int dindex_color(r, g, b)
double r, g, b;
{
    int ir = r * 255, ig = r * 255, ib = r * 255;
    extern int index_color(int, int, int);
    if (ir < 0)   ir = 0;
    if (ir > 255) ir = 255;
    if (ig < 0)   ig = 0;
    if (ig > 255) ig = 255;
    if (ib < 0)   ib = 0;
    if (ib > 255) ib = 255;
    return index_color(ir, ig, ib);
}
