#include <stdio.h>

#define FMAX    1.e+30
#define ABS(x)  ((x) < 0 ? -(x) : (x))

/**************************************/
/* some flags to analyze file formats */
/**************************************/
#define  F_BIN_TAB_PHI  6    /* Binary Table, x = cos theta, phi variable */
#define  F_BIN_TAB      5    /* Binary Table, x = cos theta */
#define  F_BFc          4    /* basic F-file, x = cos theta */
#define  F_BFc_TR       3    /* basic F-file, y = cos theta */
#define  F_BF           2    /* basic F-file, x = theta     */
#define  F_BF_TR        1    /* basic F-file, y = theta     */
#define  F_other       -1    /* other type                  */
#define  F_notF        -2    /* not F-file                  */

#define  GRID_PHI_X             5
#define  GRID_PHI_X_LEGENDRE    4
#define  GRID_LEGENDRE          3
#define  GRID_UNIFORM_COS       2
#define  GRID_UNIFORM_THETA     1
#define  GRID_NON_SET           0

#ifndef  FLOATPIX
/*
    #undef BYTEPIX
    #undef SHORTPIX
    #undef LONGPIX
    #undef FLOATPIX
    #undef DOUBLEPIX
*/
    #define BYTEPIX      8
    #define SHORTPIX     16
    #define LONGPIX      32
    #define FLOATPIX   (-32)
    #define DOUBLEPIX    64
#endif
#define     GLESPPIX    FLOATPIX
#define     GLESPMASK   (M_PI*M_E*1.0e9)

#ifdef   sun
#define  MAX_RAND   2147483647.
#define  MAX_RAND1  4294967295.
#else
#define  MAX_RAND   32767.
#endif

#ifdef __GNUC__
#undef   MAX_RAND
#define  MAX_RAND   RAND_MAX
#endif


extern int    _lmin_gl;
extern double _enorm_gl;

/**************/
/* FFILE head */
/**************/
struct _head_ {
    int naxis, naxis1, naxis2;
    char extend[10];
    char xtens[20];
    int tfields;
    char ttype1[30], tform1[30], tunit1[30], tdim1[30],
	 ttype2[30], tform2[30], tunit2[30], tdim2[30],
	 ttype3[30], tform3[30], tunit3[30], tdim3[30],
	 ttype4[30], tform4[30], tunit4[30], tdim4[30],
	 ttype5[30], tform5[30], tunit5[30], tdim5[30];
    int bitpix1, bitpix2, bitpix3, bitpix4, bitpix5;
    int nx, ny;
    double crpix2, crval2, cdelt2;
    char ctype2[30];
    char origin[30], object[30];
    int  grid;
};

#define HEAD struct _head_

extern HEAD *ph;                /* pointer to head structure */

struct _obs_ {                  /* observational data struct */
     float theta;
     float phi;
     float t;
};

struct _coor_ { double lat, lon; };

struct _crm_ {                 /* marker on map figure */
     double theta, phi;  /* coordinates of marker */
     int mark, size, color;
     double vpa;     /* angle  of polarization vector, when mark=-10 */
};


struct _zone_ {
     double latR0, lonR0;
     double latR1, lonR1;
     double radius;
     double dr;
     int type;
};

extern char *myname;
extern int theta_grid;   /* flag to restore grid for theta  */
extern int bitpix_map;   /* GLESP bitpix for float data     */
extern char tmpcat[161]; /* temporal catalog                */

extern int    Cl_max;        /* Max multipole l used in map synthesis */
extern int    alm_rand_seed; /* Random generator seed */
extern double fwhm;          /* FWHM in DEGREES       */

extern int glesp_comments;   /* comments output flag  */

/*************************/
/* function descriptions */
/*************************/
extern void ff_read_ext_map(double **, double **, int *, double **, int *,
		char *);
extern int  ffile_type(char *);
extern void ff_read_stand_map(double **, double **, int *, double **, int *,
		int, char *);
extern char *take_cur_date(char *);
extern char *skip_wd(int, char *, char *);
extern int  n_skip_wd(char *);
extern void transpose_map(double *, int, int);
extern void equat2gal(double, double, double *, double *),
	    gal2equat(double, double, double *, double *),
	    gal2pol(double, double, double *, double *),
	    pol2gal(double, double, double *, double *);
extern int map_res_cmp(int,int,int *, int, int, int *);

extern void ff_write_alm(double *, int, int, int, FILE *);
extern void alm2dl (double *, int, double *, int),
	    alm2cl (double *, int, double *, int),
	    alm2clz(double *, int, double *, int, int),
	    alm2dlz(double *, int, double *, int, int);

extern void print_cl_table (FILE *, double *, double *, int, int);
extern void print_alm_table (FILE *, double *, int, int, double *, int, int);
extern void print_alm_plot  (FILE *, double *, int, int, int, int);
extern int  read_cl_table (char *, double **, double **, int *);
extern int  read_dl_table (char *, double **, double **, int *);
extern void ff_read_ext2_map(double ** , double **, int *, int **, int *,
	      char *);

extern int  read_map2(double **, int *, int *, int **, double **, char *);
extern int  read_map1(double **, int *, int *, int **, double **, char *);
extern void recalc_map2(double **, int,int, int,int, int *, double *,double *,
	      int, int);

extern void read_hp_file(double **, int **, int *, int *, double **, char *),
	    read_hp_pol(double **,double **,double **,
				 int **, int *, int *, double **, char *);
extern void calc2gl(double *, int, int, int *, double *, double, double);

extern void alm_ks_test(double *, int, int, int, int, int, FILE *, int);
extern void alm2ffile(double *, int, int, int, int, FILE *, int);
extern void phase_calc(double *, int, double *, int);
extern double ph_calc(double, double);
extern double w_exp(double, double), w_eq(double, double);
extern void approp_hpix(double *, int, int, int *, double *, double, double);

extern void gauleg(double, double, double *, double *, int);
extern void coord(int, int, double, double, double, double *, double *,
	       int, int *,  int);
extern void weights(int, double, double, double, double *, double *,
	       int, int *);

extern void rotate_map(double *, double *, int, int, int *, double *,
	       double, double);

extern int set_nphi_by_x(double *, int, int *);
extern int get_nphi_by_x(double *, int, int),
	   get_pol_nphi_by_x(double *, int, int, int);
extern int set_nphi_by_x2(double *, int, int, int *),
	   set_nphi_by_x2p(double *, int, int, int *);
extern int get_nphi_by_x2(double *, int, int, int);
extern void cop_grid(double *, int, int, int *, double *,
		    double *, int, int, int *, double *);
extern int convert_nested2ring(int, int);
extern void alm2sp(double *, int, int, int, int, FILE *);
extern void gal2eq_map(double *, double *, int, int, int *, double *);

extern void   get_val_str(char *, double *,  double *);
extern void   get_2_str(char *, char *, char *),
	      get_4_str(char *, char *, char *, char *, char *);
extern double deg2rad(char *), str2deg(char *);
extern void  read_data_gal(struct _coor_ **, int *, char *),
	     read_data_pol(struct _coor_ **, int *, char *),
	     read_data_eq2gal(struct _coor_ **, int *, char *);
extern void  str2sph(char *, double *,  double *, double *, int);

extern void read_marker(struct _crm_ **, int *, char *, int);
extern void read_mark_str(struct _crm_ **, int *, char *, int);
extern void read_mark_pol(struct _crm_ **, int *, char *, int);

/* fits_glesp.c */
extern void outp_glesp_map(double *, double *, int, int, int *, FILE *,
		char *, char *);
extern void in_glesp_map(double **, double **, int*, int*, int **, char *);
extern void read_glesp_map(double **,double **, int *,int *, int **, char *),
	   read_glesp_tmap(double **,double **, int *,int *, int **, char *),
	   read_glesp_qmap(double **,double **, int *,int *, int **, char *),
	   read_glesp_umap(double **,double **, int *,int *, int **, char *),
	   read_glesp_3map(double **,double **,double **,double **,
	       int *,int *, int **, char *);
extern int test_glesp_map(char *);
extern int Fget_lmax(char *);
extern void in_glesp_alm(double *, int *, int, char *);
extern void in_glesp_alm_T(double *, int *, int, char *);
extern void out_glesp_alm(double *, int, int, int, FILE *);
extern void out_glesp_al_Tm(double *, int, int, int, FILE *);
extern void add_alm_ext(double *, double *, int, int, int, FILE *),
	    add_alm_ext_T(double *, double *, int, int, int, FILE *),
	    add1_alm_ext(double *, int, int, int, FILE *);
extern void read_hp_palm(double **,double **,double **, int *, int *, char *);
extern void write_2d_fits(FILE *, double *, int, int, double, double,
		double, double, char *, char *, double, char *);
extern void glesp_2_map2d(double *, double *, int, int, int *, FILE *);
extern void fits_write_map(double *, int, double *, int, double *,
		int, int, int, FILE *);
extern void outp_glesp_3map(double *, double *, double *, double *, int, int,
		int *, FILE *, char *, char *);
extern void in_glesp_3map(double **, double **, double **, double **,
		int*, int*, int **, char *);
extern void read_ieb_alm(double **, double **, double **,
		int *, int *, char *);
extern void read_ieb_alm_T(double **, double **, double **,
		int *, int *, char *);
extern void write_ieb_alm(double *, double *, double *, int,int,int,FILE *);
extern void write_ieb_alm_T(double*, double*, double *, int,int,int,FILE *);

/* cl2alm.c */
extern void cl2alm (double *, int, double *, int, double);
extern void phase_rand(double *, int);
extern void cl2alm_ph(double *, int, double *, int, double);
extern void amph2alm (double *, double *, int, double *, int);

/*  read_alm.c */
extern int  getLmx(char *);
extern int  get_lmax(char *);
extern int  get_lmax_str(char *);
extern int Sget_lmax(char *);
extern int  read_alm_table(char *, double **, int *, int *, int *);
extern void read_alm_str  (char *, double *, int, int);
extern void str2alm(double *, int, int, char *);
extern void ff_read_alm(double *, int, int *, char *);
extern int  read_ph_table(char *, double **, int *);
extern void str2phase(double *, int, char *);
extern void alm2alm(double **, int *, int *, int, int);
extern void realloc_alm(double *, int, int, int);
extern void alm_remove (double *, int, int, char *);
extern void alm_extract(double *, int, int, char *);
extern void separ_phase(double *, int, int, double, double, int, int);

/* cosbell.c */
extern void area_pos(double *, int, int, int *, double *,  int *, int *);
extern void area_size(double *, int, int, int *, double *, int, int,
	      double *, double *);
extern void cosbell(double *, int, int, int *, double *, int, int,
	      double, double, double *, int);

/* fhp_in.c */
extern void fhp_in(double **, int *, int *, FILE *);
extern void fhp2tac(double *, int, int, double **, int **, int *, int *,
		double **);
extern void read_mask(int **, int **, int *, int *, double **, char *, int);
extern void read_hp_file_nobs(int **, int **, int *, int *, double **,
		 char *);
extern void hp_arrays(double *, double *, int *, int);

/* fhp_out.c */
extern void fhp_out(double **, int, int, double *, double *, int, FILE *);
extern void fhp_out2(double **, int, int *, int, double *, double *, int,
		  FILE *, int);
extern void fhp_out2p(double *, double *, int, int *, int, int, FILE *),
	    fhp_out2p_pol(double *, double *, double *, double *,
		  int, int *, int, int, FILE *);

/* fitsfun.c */
extern int FHinit_FITS(FILE *, int);
extern int Fits_update(FILE *);
extern void put_dvec_fits(FILE *, double *, int);
extern void put_dvec_fits_pol(FILE *, double *, double *, double *,
	      int, int);

/* intrp.c */
extern void spline (double *, double *, int, double, double, double *);
extern void splint (double *, double *, double *, int, double, double *);
extern int intrpl2Fsm_nu_nu(FILE *, int, double *, int, int, double *,
		double *,int);
extern int intrpl2Fsm_nu_u (FILE *, int, double *, int, int, double *,double,
		double, int);
extern int  intrpl2sm_Nnu_u(double *, int, int, double *, int, int,
		double *, double, double,  double, double, int *, int);
extern int  intrpl2sm_Nnu_nu(double *, int, int, double *, int, int,
		double *, double *, double, double, int *, int);
extern int  intrpl2Fsm_Nnu_nu(FILE *, int, int, double *, int, int,
		double *, double *, double, double, int *, int);
extern int  intrpl2_u_nu(double *, int, double, double, double *, double *,
		int, int);
extern int  intrpl_nu_u(double *, int, double *, int, int, double *, double,
		double);
extern int  intrpl2m_nu_u(double *, int, int, double *, int, int,
		double *, double, double, int);
extern int  intrpl2sm_nu_u(double *, int, int, double *, int, int,
		double *, double, double, int);
extern int intrpl2sm_nu_nu(double *, int, int, double *, int, int,
		double *, double *, int);
extern int indx(double *, int, double);
extern int copy2sm_Nnu_nu(double *, int, int, double *, int, int,
	double *, double *, double, double, int *, int);

/* ff_out.c */
extern void ff_write_ext2_map(double *, double *, int, int, int *,
	      FILE *, char *, char *);
extern void ff_write_ext_map(double *, int, double *, int, double *,
	     int, FILE *);
extern void ff_write_map(double *, int, double *, int, double *,
		int, int, int, FILE *);
extern void write_simple_map(FILE *, double *, int, int, double, double,
		double, double, char *, char *, double, char *);
extern void ff_write_map2d(double *, double *, int, int, int *, FILE *);

/* pix2ang.c */
extern int  pix2pix(double *, int,int,int *, int, double *, int,int,int *);
extern void n2ang(int, double *, double *, double *, int, int, int *),
	    na2ang(int, double *, double *, double *, int, int, int *);
extern int  ang2n(double, double, double *, int, int, int *),
	    ang2na(double, double, double *, int, int, int *);
extern void pixel2angle(double *, int, int *, int, int, double *, double *);
extern void angle2pixel(double *, int, int *, int *, int *, double, double);
extern void pixel2hangle(double *, int, int *, int, int, double *, double *,
	       double *);
extern void hangle2pixel(double *, int, int *, int *, int *, double, double,
	       double *);
extern void n2hang(int, double *, double *, double *, int, int, int *,
	       double *);
extern int  hang2n(double, double, double *, int, int, int *, double *);
extern void pix2sphere(int, int, int, int *, double *,
	       double *, double *, double *);
extern int sphere2pix(double, double, double, int, int, int *, double *);
extern int ind_x(int, double *, double);
extern int hpix2pix(double *, int, int, int *, double *, int, double *, int,
	       int, int *);
extern int pix_eq2gal(double,double, double *, int, int, int *);

/* rw_ff.c */
extern void  read_ff(struct _coor_ **, int *, char *, int, double);
extern void  write_ff(double *, int, FILE *);

/* read_cats.c */
extern void  read_cats(struct _coor_ **, int *, char *, int);
extern void  cats_str2sph(char *, double *,  double *, double *, int);

/* beam_oper.c */
extern void beam_tikh(double *, int, int, double *, double);
extern void beam_tikh_gauss(double *, int, int, double, double);
extern void beam_smooth(double *, int, int, double *);
extern void beam_gauss_smooth(double *, int, int, double);

/* rw_ff.c */
extern void  create_ff(int, long, char *, double, double, double, double,
	  double, char *);

/* map_isol.c */
extern void map_isol(double *, int, int, int *, double *, int);

/* alm_rotate.c */
extern void alm_rotate(double *, int, int, double, double, double);
extern void cr2cr_alm (double *, int, int, int, int);


/* map_grad.c */
extern void map_grad_phi  (double *, double *, int, int, int *),
	    map_grad_theta(double *, double *, int, int, int *);
extern void map_grad_pa(double *, double *, double *, double *, double *,
		int, int, int *);


/* alm2tl.c */
extern void alm2tl(double *,int,int, double*,double*,double*, double, int);
