#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "glesp.h"
#include "fits.h"

#define PR      fprintf
#define ER      stderr

/* Conversion of Basic FITS to GLESP FITS map */
/* if flag_bf=1, then converse FITS map to the full sphere */
/* if flag_bf=0, then converse FITS map as it is (part of sphere) */

extern int bitpix;

void map_basic_fits0(f,x,nx,np,nph, bfits_file, flag_bf, layout, glon0,glat0)
double *f;
double *x;
int nx, np, *nph;
char *bfits_file;
int flag_bf;           /* type of expansion of map onto sphere */
int layout;            /* layout <= NAXIS3 */
double glon0, glat0;
{
    register int i, j;
    int nx1=0, nx2=0, nx3=0;
    double crv1=0.0, crv2=0.0, crp1=0.0, crp2=0.0, cdlt1=0.0, cdlt2=0.0;
    char ct1[51], ct2[51], bunit[51];
    double b_scale = 1.0, b_zero = 0.0;
    double *data;
    FILE *fp_bf;
    double glon_min, glon_max, glat_min, glat_max;

    /* open input Basic FITS file */
    if ((fp_bf = fopen(bfits_file, "r")) == (FILE *)0) {
	PR(ER, "%s: can not open file '%s'.\n", myname, bfits_file);
	exit(1);
    }
    /*****************/
    /* read keywords */
    /*****************/
    FHinit(fp_bf);
    if (FEVget("BITPIX") != NULL)
	bitpix = FLkey_get("BITPIX");
    else {
	PR(ER, "%s: bad FITS file '%s', no keyword BITPIX.\n",
	    myname, bfits_file);
	exit(1);
    }
    if (FEVget("NAXIS1") != NULL)
	nx1 = FLkey_get("NAXIS1");
    else {
	PR(ER, "%s: bad FITS file '%s', no keyword NAXIS1.\n",
	    myname, bfits_file);
	exit(1);
    }
    if (FEVget("NAXIS2") != NULL)
	nx2 = FLkey_get("NAXIS2");
    else {
	PR(ER, "%s: bad FITS file '%s', no keyword NAXIS2.\n",
	    myname, bfits_file);
	exit(1);
    }
    if (FEVget("NAXIS3") != NULL)
	nx3 = FLkey_get("NAXIS3");
    else {
//      PR(ER, "%s: bad FITS file '%s', no keyword NAXIS3.\n",
//          myname, bfits_file);
//      exit(1);
	nx3=1;
    }
    if (FEVget("CRVAL1") != NULL)
	crv1 = FDkey_get("CRVAL1");
    else {
	PR(ER, "%s: bad FITS file '%s', no keyword CRVAL1.\n",
	    myname, bfits_file);
	exit(1);
    }
    if (FEVget("CRVAL2") != NULL)
	crv2 = FDkey_get("CRVAL2");
    else {
	PR(ER, "%s: bad FITS file '%s', no keyword CRVAL2.\n",
	    myname, bfits_file);
	exit(1);
    }
    if (FEVget("CRPIX1") != NULL)
	crp1 = FDkey_get("CRPIX1");
    else {
	PR(ER, "%s: bad FITS file '%s', no keyword CRPIX1.\n",
	    myname, bfits_file);
	exit(1);
    }
    if (FEVget("CRPIX2") != NULL)
	crp2 = FDkey_get("CRPIX2");
    else {
	PR(ER, "%s: bad FITS file '%s', no keyword CRPIX2.\n",
	    myname, bfits_file);
	exit(1);
    }
    if (FEVget("CDELT1") != NULL)
	cdlt1 = FDkey_get("CDELT1");
    else {
	PR(ER, "%s: bad FITS file '%s', no keyword CDELT1.\n",
	    myname, bfits_file);
	exit(1);
    }
    if (FEVget("CDELT2") != NULL)
	cdlt2 = FDkey_get("CDELT2");
    else {
	PR(ER, "%s: bad FITS file '%s', no keyword CDELT2.\n",
	    myname, bfits_file);
	exit(1);
    }
    if (FEVget("CTYPE1") != NULL)
	strcpy(ct1, FEVget("CTYPE1"));
    else  ct1[0] = '\0';
    if (FEVget("CTYPE2") != NULL)
	strcpy(ct2, FEVget("CTYPE2"));
    else  ct2[0] = '\0';
    if (FEVget("BSCALE") != NULL)
	b_scale = FDkey_get("BSCALE");
    if (FEVget("BZERO") != NULL)
	b_zero = FDkey_get("BZERO");
    if (FEVget("BUNIT") != NULL)
	strcpy(bunit, FEVget("BUNIT"));
    else  bunit[0] = '\0';

    if (layout > nx3) {
	PR(ER, "%s: no layout %d in file '%s'.\n", myname,layout,bfits_file);
	exit(1);
    }

    /* catch memory */
    if((data=(double*)calloc((unsigned)(nx1*nx2),sizeof *data))==(double*)0) {
	PR(ER, "%s: no memory for data array of (%dx%d).\n",myname,nx1,nx2);
	exit(1);
    }

    if (nx3 == 0)  nx3 = 1;

    /* skip layouts */
    if (layout > 1) {
	int k, kl=layout-1;
	for(k=0; k<kl; k++)
	    for(i=0; i<nx2; i++)
		for(j=0; j<nx1; j++)
		    get_pix_fits(fp_bf);
    }

    /* read data */
    for(i=0; i<nx2; i++) {
	int ii = nx1 * i;
	for(j=0; j<nx1; j++) {
	    data[ii+j] = b_zero + b_scale * get_pix_fits(fp_bf);
	}
    }

    /* close input file */
    fclose(fp_bf);

    /* free memory from header */
    FHfree();

    /* min & max coordinates in input map */
    glon_min = crv1 + (1.  - crp1) * cdlt1;
    glon_max = crv1 + (nx1 - crp1) * cdlt1;
    glat_min = crv2 + (1.  - crp2) * cdlt2;
    glat_max = crv2 + (nx2 - crp2) * cdlt2;

    if (glon0 > -1000.) {
	    glon_min = glon0 - (glon_max - glon_min) / 2.;
	    glon_max = glon_min + (nx1 - 1) * cdlt1;
	    crv1 = glon_min;
	    crp1 = 1.;
    }
    if (glat0 > -1000.) {
	    glat_min = glat0 - (glat_max - glat_min) / 2.;
	    glat_max = glat_min + (nx2 - 1) * cdlt2;
	    crv2 = glat_min;
	    crp2 = 1.;
    }

    if (glon_min>glon_max) {double v=glon_min;glon_min=glon_max;glon_max=v;}
    if (glat_min>glat_max) {double v=glat_min;glat_min=glat_max;glat_max=v;}

PR(ER, "glon_min=%g, glon_max=%g, glat_min=%g, glat_max=%g\n",
glon_min,glon_max,glat_min,glat_max);
    /*********************************************/
    /* get pixel from the GLESP map and          */
    /* find corresponding pixel in the input map */
    /*********************************************/
    for(i=0; i<nx; i++) {
	int ii=i*np;
	for(j=0; j<nph[i]; j++) {
	    double theta, phi, glat, glon, aa;
	    int pi, pj;

	    /* find (theta,phi) by pixel number */
	    pixel2angle(x, nx, nph, i, j, &theta, &phi);
	    /* coversion (theta,phi) -> (glat,glon) */
	    pol2gal(theta, phi, &glat, &glon);
	    /* radians -> degrees */
	    glat *= 180/M_PI;
	    glon *= 180/M_PI;

	    /* check limits */
	    if (180. < glon_max)  {
		if (glon < 0.)  glon += 360.;
	    } else {
		if (glon > 180.0)  glon -= 360.;
	    }
	    if (glon < glon_min || glon > glon_max)
		continue;
	    if (glat < glat_min || glat > glat_max)
		continue;


	    /* find corresponding pixel in input map */
	    aa = (glon - crv1) / cdlt1 + crp1;
	    if ( aa > 0. ) pj = aa + 0.5; else pj = aa - 0.5;
	    aa = (glat - crv2) / cdlt2 + crp2;
	    if ( aa > 0. ) pi = aa + 0.5; else pi = aa - 0.5;

	    /* appropriate values */
	    f[ii+j] = data[(pi-1)*nx1 + (pj-1)];
	}
    }
    free((void *)data);
    return;
}



void map_basic_fits(f,x,nx,np,nph, bfits_file, flag_bf, layout, glon0,glat0)
double *f;
double *x;
int nx, np, *nph;
char *bfits_file;
int flag_bf;           /* type of expansion of map onto sphere */
int layout;            /* layout <= NAXIS3 */
double glon0, glat0;
{
    register int i, j;
    int nx1=0, nx2=0, nx3=0;
    double crv1=0.0, crv2=0.0, crp1=0.0, crp2=0.0, cdlt1=0.0, cdlt2=0.0;
    char ct1[51], ct2[51], bunit[51];
    double b_scale = 1.0, b_zero = 0.0;
    double *data;
    FILE *fp_bf;
    double glon_min, glon_max, glat_min, glat_max;

    /* open input Basic FITS file */
    if ((fp_bf = fopen(bfits_file, "r")) == (FILE *)0) {
	PR(ER, "%s: can not open file '%s'.\n", myname, bfits_file);
	exit(1);
    }
    /*****************/
    /* read keywords */
    /*****************/
    FHinit(fp_bf);
    if (FEVget("BITPIX") != NULL)
	bitpix = FLkey_get("BITPIX");
    else {
	PR(ER, "%s: bad FITS file '%s', no keyword BITPIX.\n",
	    myname, bfits_file);
	exit(1);
    }
    if (FEVget("NAXIS1") != NULL)
	nx1 = FLkey_get("NAXIS1");
    else {
	PR(ER, "%s: bad FITS file '%s', no keyword NAXIS1.\n",
	    myname, bfits_file);
	exit(1);
    }
    if (FEVget("NAXIS2") != NULL)
	nx2 = FLkey_get("NAXIS2");
    else {
	PR(ER, "%s: bad FITS file '%s', no keyword NAXIS2.\n",
	    myname, bfits_file);
	exit(1);
    }
    if (FEVget("NAXIS3") != NULL)
	nx3 = FLkey_get("NAXIS3");
    else {
//      PR(ER, "%s: bad FITS file '%s', no keyword NAXIS3.\n",
//          myname, bfits_file);
//      exit(1);
	nx3=1;
    }
    if (FEVget("CRVAL1") != NULL)
	crv1 = FDkey_get("CRVAL1");
    else {
	PR(ER, "%s: bad FITS file '%s', no keyword CRVAL1.\n",
	    myname, bfits_file);
	exit(1);
    }
    if (FEVget("CRVAL2") != NULL)
	crv2 = FDkey_get("CRVAL2");
    else {
	PR(ER, "%s: bad FITS file '%s', no keyword CRVAL2.\n",
	    myname, bfits_file);
	exit(1);
    }
    if (FEVget("CRPIX1") != NULL)
	crp1 = FDkey_get("CRPIX1");
    else {
	PR(ER, "%s: bad FITS file '%s', no keyword CRPIX1.\n",
	    myname, bfits_file);
	exit(1);
    }
    if (FEVget("CRPIX2") != NULL)
	crp2 = FDkey_get("CRPIX2");
    else {
	PR(ER, "%s: bad FITS file '%s', no keyword CRPIX2.\n",
	    myname, bfits_file);
	exit(1);
    }
    if (FEVget("CDELT1") != NULL)
	cdlt1 = FDkey_get("CDELT1");
    else {
	PR(ER, "%s: bad FITS file '%s', no keyword CDELT1.\n",
	    myname, bfits_file);
	exit(1);
    }
    if (FEVget("CDELT2") != NULL)
	cdlt2 = FDkey_get("CDELT2");
    else {
	PR(ER, "%s: bad FITS file '%s', no keyword CDELT2.\n",
	    myname, bfits_file);
	exit(1);
    }
    if (FEVget("CTYPE1") != NULL)
	strcpy(ct1, FEVget("CTYPE1"));
    else  ct1[0] = '\0';
    if (FEVget("CTYPE2") != NULL)
	strcpy(ct2, FEVget("CTYPE2"));
    else  ct2[0] = '\0';
    if (FEVget("BSCALE") != NULL)
	b_scale = FDkey_get("BSCALE");
    if (FEVget("BZERO") != NULL)
	b_zero = FDkey_get("BZERO");
    if (FEVget("BUNIT") != NULL)
	strcpy(bunit, FEVget("BUNIT"));
    else  bunit[0] = '\0';

    if (layout > nx3) {
	PR(ER, "%s: no layout %d in file '%s'.\n", myname,layout,bfits_file);
	exit(1);
    }

    /* catch memory */
    if((data=(double*)calloc((unsigned)(nx1*nx2),sizeof *data))==(double*)0) {
	PR(ER, "%s: no memory for data array of (%dx%d).\n",myname,nx1,nx2);
	exit(1);
    }

    if (nx3 == 0)  nx3 = 1;

    /* skip layouts */
    if (layout > 1) {
	int k, kl=layout-1;
	for(k=0; k<kl; k++)
	    for(i=0; i<nx2; i++)
		for(j=0; j<nx1; j++)
		    get_pix_fits(fp_bf);
    }

    /* read data */
    for(i=0; i<nx2; i++) {
	int ii = nx1 * i;
	for(j=0; j<nx1; j++) {
	    data[ii+j] = b_zero + b_scale * get_pix_fits(fp_bf);
	}
    }

    /* close input file */
    fclose(fp_bf);

    /* free memory from header */
    FHfree();

    /* min & max coordinates in input map */
    glon_min = crv1 + (1.  - crp1) * cdlt1;
    glon_max = crv1 + (nx1 - crp1) * cdlt1;
    glat_min = crv2 + (1.  - crp2) * cdlt2;
    glat_max = crv2 + (nx2 - crp2) * cdlt2;

    if (glon_min>glon_max) {
	double v=glon_min; glon_min=glon_max; glon_max=v;
	crv1 = -crv1;
    }
    if (glat_min>glat_max) {
	double v=glat_min; glat_min=glat_max; glat_max=v;
	crv2 = -crv2;
    }

    if (glon0 > -1000.) {
	    glon_min = - (glon_max - glon_min) / 2.;
	    glon_max = glon_min + (nx1 - 1) * cdlt1;
	    crv1 = glon_min;
	    crp1 = 1.;
	    if (cdlt1 < 0) {
		int nn = (nx2-1)/2;
		double v= glon_min;
		glon_min=glon_max; glon_max=v;
		cdlt1= -cdlt1;
		for(i=0; i<nx2; i++) {
		    int ii = nx1 * i;
		    for(j=0; j<=nn; j++) {
			int i1= ii+j, i2 = ii+ nx2-1-j;
			v = data[i1];
			data[i1] = data[i2];
			data[i2] = v;
		    }
		}
	    }
    } else  glon0 = 0.;
    if (glat0 > -1000.) {
	    glat_min = - (glat_max - glat_min) / 2.;
	    glat_max = glat_min + (nx2 - 1) * cdlt2;
	    crv2 = glat_min;
	    crp2 = 1.;
    }  else  glat0 = 0.;

  {
    double euler_m[3][3];
    double rr = M_PI/180.;
    extern void euler_matrix(double, double, double, double [3][3]);
    extern void  xyz_euler_transform(double, double, double,
		 double *, double *, double *, double [3][3]);

    /* sign '-' because we use new pixel to find correspondig old one */
    /* i.e. from new position, we obtain old one                      */
    euler_matrix(-glon0*rr, -glat0*rr, 0.0, euler_m);

    /*********************************************/
    /* get pixel from the GLESP map and          */
    /* find corresponding pixel in the input map */
    /*********************************************/
    for(i=0; i<nx; i++) {
	int ii=i*np;
	for(j=0; j<nph[i]; j++) {
	    double theta, phi, glat, glon, aa;
	    int pi, pj;
	    int n_s, n_p;
	    double x_s, y_s, z_s, x_p, y_p, z_p;

	    /* get coordinates of a map (rotated) */
	    n_s = ii + j;
	    pix2sphere(n_s, nx,np,nph,x, &x_s, &y_s, &z_s);

	    /* get initial coordinates with Euler matrix              */
	    /* (rotate pixel from a sphere to check if it's in area ) */
	    xyz_euler_transform(x_s, y_s, z_s,
		 &x_p, &y_p, &z_p, euler_m);

	    /* corresponding pixel value */
	    n_p = sphere2pix(x_p, y_p, z_p, nx,np,nph,x);

	    /* find (theta,phi) by pixel number */
	    n2ang(n_p, &theta, &phi, x, nx, np, nph);

	    /* conversion (theta,phi) -> (glat,glon) */
	    pol2gal(theta, phi, &glat, &glon);
	    glat /= rr;
	    glon /= rr;

	    /* check limits */
	    if (180. < glon_max)  {
		if (glon < 0.)  glon += 360.;
	    } else {
		if (glon > 180.0)  glon -= 360.;
	    }
	    if (glon < glon_min || glon > glon_max)
		continue;
	    if (glat < glat_min || glat > glat_max)
		continue;


	    /* find corresponding pixel in input map */
	    aa = (glon - crv1) / cdlt1 + crp1;
	    if ( aa > 0. ) pj = aa + 0.5; else pj = aa - 0.5;
	    aa = (glat - crv2) / cdlt2 + crp2;
	    if ( aa > 0. ) pi = aa + 0.5; else pi = aa - 0.5;

	     /* appropriate values */
	    f[n_s] = data[(pi-1)*nx1 + (pj-1)];
	}
    }
  }
    free((void *)data);
    return;
}



void map_basic_files(f,x,nx,np,nph, bfits_list, lay)
double *f;
double *x;
int nx, np, *nph;
char *bfits_list;
int lay;            /* layout <= NAXIS3 */
{
    register int i,j, k, nf=0;
    int *cn = (int *)0;
    double *b=(double *)0;
    struct _ffb_ { char name[81]; double lat, lon;}  *ffb;
    double qq = M_PI + M_E;
    FILE *fp;
    char str[81];

    if ((cn=(int *)calloc((unsigned)(nx*np), sizeof *cn))==(int *)0) {
nomem:
	PR(ER, "%s: no memory for temporal data for basic FITS.\n", myname);
	exit(1);
    }
    if ((b=(double *)calloc((unsigned)(nx*np), sizeof *b))==(double *)0)
	    goto nomem;

    /* count file names */
    if ((fp = fopen(bfits_list, "r")) == (FILE *)0) {
	PR(ER, "%s: can not open file '%s'.\n", myname, bfits_list);
	exit(1);
    }
    while (fgets(str, 80, fp) != NULL) {
	if (*str == '#' || *str == '\n' || *str == '\r') continue;
	nf++;
    }
    if ((ffb=(struct _ffb_ *)calloc(nf, sizeof *ffb))==(struct _ffb_ *)0)
	    goto nomem;
    k=0;
    fseek(fp, 0l, 0);
    while (fgets(str, 80, fp) != NULL) {
	if (*str == '#' || *str == '\n' || *str == '\r') continue;
	sscanf(str, "%s%lf%lf", ffb[k].name, &ffb[k].lon, &ffb[k].lat);
	k++;
    }
    fclose(fp);

    for (k=0; k<nf; k++) {
	for(i=0; i<nx; i++) {
	    int ii=i*np;
	    for(j=0; j<nph[i]; j++) b[ii+j] = qq;
	}

	/* incorporate  Basic FITS to GLESP FITS map */
	map_basic_fits(b,x,nx,np,nph,ffb[k].name,1,lay,ffb[k].lon,ffb[k].lat);

	/* count */
	for(i=0; i<nx; i++) {
	    int ii=i*np;
	    for(j=0; j<nph[i]; j++)
		if (ABS(b[ii+j]-qq) > 1.e-20) {
		    f[ii+j] += b[ii+j];
		    cn[ii+j] += 1;
		}
	}
    }
    /* average */
    for(i=0; i<nx; i++) {
	int ii=i*np;
	for(j=0; j<nph[i]; j++)
	    if (cn[ii+j]) f[ii+j] /= cn[ii+j];
    }


    free((void *)cn);
    free((void *)b);
    free((void *)ffb);
    return;
}
