#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "glesp.h"
#define PR  fprintf
#define ER  stderr

/*
 *   procedure of mapping of correlation
 *
 *   Oleg Verkhodanov, 20/06/2008, vo@sao.ru
 */

extern double corr_area(double *,double *,double *,int,int,int *,
	  double,double,double,double);
extern double sig_area(double *,double *,int,int,int *,
	  double,double,double,double);
extern void  map_sig(double *, double *, int, int, int *,
		 double, double **, double **, int *, int *, int **);
extern void  map_rsig(double *, double *, int, int, int *, double);


void  map_cor(f1, f2, x, nx, np, nph, corr_win, fcor, xc, nxc, npc, nphc)
double *f1, *f2, *x;
int nx, np, *nph;
double corr_win;         /* correlation window, min */
double **fcor,           /* output array            */
       **xc;             /* output X-array          */
int *nxc, *npc, **nphc;  /* output nx,np,nph        */
{
    register int i, j;
    int n;
    double th1, th2, ph1, ph2;
    double x1, x2;
    int pin_flag = nph[0] == np ? 0 : 1;

    *nxc = 180 * 60 / corr_win + 0.5;
    corr_win = 180 * 60 / *nxc;

    if (glesp_comments) {
	PR(ER, "\nactual correlation window of %gx%g (%g) sq.arcmin.\n",
		corr_win, corr_win, corr_win*corr_win);
    }
    *npc = 2 * *nxc;

    n = *nxc * *npc;
    if ((*fcor = (double *)calloc((unsigned)n, sizeof **fcor))==(double *)0){
nomem:
	PR(ER, "%s: no memory for output arrays...\n", myname);
	exit(1);
    }
    if ((*xc = (double *)calloc((unsigned)(*nxc), sizeof **xc))==(double *)0)
	goto nomem;
    if ((*nphc = (int *)calloc((unsigned)(*nxc), sizeof **nphc))==(int *)0)
	goto nomem;

    x1 = -1.0; x2 = 1.0;
    gauleg(x1, x2, *xc, (double *)0, *nxc);
    if (pin_flag) {
	/* GLESP pixelization */
//      for (i=0; i < *nxc; i++)  (*nphc)[i] = get_nphi_by_x(*xc, *nxc, i);
	for (i=0; i < *nxc; i++)
	    (*nphc)[i] = get_nphi_by_x2(*xc, *nxc, *npc, i);
    } else {
	/* pin = 0 */
	for (i=0; i < *nxc; i++)  (*nphc)[i] = *npc;
    }


    n = *nxc - 1;
    for (i=0; i < *nxc; i++) {
	register int ii = i * *npc;
	double dphi = 2*M_PI / (*nphc)[i];

	if (i == 0) {
	    if (x[i] < 0.0)  {
		th1 = fabs(acos((*xc)[i])+acos((*xc)[i+1]))/2.;
		th2 = M_PI;
	    } else {
		th1 = 0.0;
		th2 = fabs(acos((*xc)[i])+acos((*xc)[i+1]))/2.;
	    }
	} else if (i == *nxc - 1) {
	    if (x[i] < 0.0)  {
		th1 = 0.0;
		th2 = fabs(acos((*xc)[i])+acos((*xc)[i-1]))/2.;
	    } else {
		th1 = fabs(acos((*xc)[i])+acos((*xc)[i-1]))/2.;
		th2 = M_PI;
	    }
	} else {
	    double t1 = acos((*xc)[i-1]),
		   t2 = acos((*xc)[i]),
		   t3 = acos((*xc)[i+1]);
	    th1 = (t1 + t2) / 2;
	    th2 = (t2 + t3) / 2;
	    if (th1 > th2)  {t1 = th1; th1 = th2; th2 = t1; }
	}

	for (j=0; j < (*nphc)[i]; j++) {
	    ph1 = j * dphi;
	    ph2 = (j+1) * dphi;

	    (*fcor)[ii+j] = corr_area(f1,f2,x,nx,np,nph, th1,th2,ph1,ph2);
	}
    }

    return;
}

/* calculate correlation in selected area */
double corr_area(f1,f2,x,nx,np,nph, th1,th2,ph1,ph2)
double *f1, *f2, *x;
int nx, np, *nph;
double th1,th2,ph1,ph2;
{
    double corr=0.0, sum1 = 0.0, sum2 = 0.0,
	   cov= 0.0, sig1 = 0.0, sig2 = 0.0;
    register int i, j;
    double pi2 = M_PI * 2;
    int i1, i2, i3, i4, j1, j2, j3, j4,
	i_min, i_max, j_min, j_max, n;

    /* find limits of area in pixels */
    angle2pixel(x, nx, nph, &i1, &j1, th1, ph1);
    angle2pixel(x, nx, nph, &i2, &j2, th2, ph2);
    if (ABS(ph2-pi2) < .00001)
	    j2 = nph[i1] - 1;
    angle2pixel(x, nx, nph, &i3, &j3, th1, ph2);
    if (ABS(ph2-pi2) < .00001)
	    j3 = nph[i3] - 1;
    angle2pixel(x, nx, nph, &i4, &j4, th2, ph1);

    i_min = i1 < i2 ? i1 : i2;
    i_min = i_min < i3 ? i_min : i3;
    i_min = i_min < i4 ? i_min : i4;
    i_max = i1 > i2 ? i1 : i2;
    i_max = i_max > i3 ? i_max : i3;
    i_max = i_max > i4 ? i_max : i4;
    j_min = j1 < j2 ? j1 : j2;
    j_min = j_min < j3 ? j_min : j3;
    j_min = j_min < j4 ? j_min : j4;
    j_max = j1 > j2 ? j1 : j2;
    j_max = j_max > j3 ? j_max : j3;
    j_max = j_max > j4 ? j_max : j4;

    n = 0;
    for(i=i_min; i<=i_max; i++) {
	int ii = i * np;
	double dphi = pi2/nph[i];
	double th = acos(x[i]);
	if (!((th1 <= th && th <= th2) || (th2 <= th && th <= th1)))
	    continue;
	for(j=j_min; j<=j_max; j++) {
	    double phi = j * dphi;
	    if (phi < ph1) continue;
	    if (phi > ph2 || j>=nph[i]) break;
	    sum1 += f1[ii+j];
	    sum2 += f2[ii+j];
	    n++;
	}
    }
    if (n == 0)
	   return 0.0;

    sum1 /= (double)n;
    sum2 /= (double)n;

    for(i=i_min; i<=i_max; i++) {
	int ii = i * np;
	double dphi = pi2/nph[i];
	double th = acos(x[i]);
	if (!((th1 <= th && th <= th2) || (th2 <= th && th <= th1)))
	    continue;
	for(j=j_min; j<=j_max; j++) {
	    int ij = ii + j;
	    double s1 = f1[ij] - sum1, s2 = f2[ij] - sum2;
	    double phi = j * dphi;
	    if (phi < ph1) continue;
	    if (phi > ph2 || j>=nph[i]) break;
	    sig1 += s1 * s1;
	    sig2 += s2 * s2;
	    cov  += s1 * s2;
	}
    }
    sig1 /= (double)n;
    sig2 /= (double)n;
    cov  /= (double)n;

    if (fabs(sig1) < 1.e-30  ||  fabs(sig2) < 1.e-30)
	corr = cov;
    else
	corr = cov / sqrt(sig1*sig2);

    return corr;
}



void map_sig(f,x,nx,np,nph, sig_win, fsig, xs, nxs, nps, nphs)
double *f, *x;
int nx, np, *nph;
double sig_win;          /* window  for sigma calc, min */
double **fsig,           /* output array            */
       **xs;             /* output X-array          */
int *nxs, *nps, **nphs;  /* output nx,np,nph        */
{
    register int i, j;
    int n;
    double th1, th2, ph1, ph2;
    double x1, x2;
    int pin_flag = nph[0] == np ? 0 : 1;

    *nxs = 180 * 60 / sig_win + 0.5;
    sig_win = 180 * 60 / *nxs;

    if (glesp_comments) {
     PR(ER,"\nactual window for sigma-calculation is %gx%g (%g) sq.arcmin.\n",
		sig_win, sig_win, sig_win*sig_win);
    }
    *nps = 2 * *nxs;

    n = *nxs * *nps;
    if ((*fsig = (double *)calloc((unsigned)n, sizeof **fsig))==(double *)0){
nomem:
	PR(ER, "%s: no memory for output arrays...\n", myname);
	exit(1);
    }
    if ((*xs = (double *)calloc((unsigned)(*nxs), sizeof **xs))==(double *)0)
	goto nomem;
    if ((*nphs = (int *)calloc((unsigned)(*nxs), sizeof **nphs))==(int *)0)
	goto nomem;

    x1 = -1.0; x2 = 1.0;
    gauleg(x1, x2, *xs, (double *)0, *nxs);
    if (pin_flag) {
	/* GLESP pixelization */
	for (i=0; i < *nxs; i++)
	    (*nphs)[i] = get_nphi_by_x2(*xs, *nxs, *nps, i);
    } else {
	/* pin = 0 */
	for (i=0; i < *nxs; i++)  (*nphs)[i] = *nps;
    }


    n = *nxs - 1;
    for (i=0; i < *nxs; i++) {
	register int ii = i * *nps;
	double dphi = 2*M_PI / (*nphs)[i];

	if (i == 0) {
	    if (x[i] < 0.0)  {
		th1 = fabs(acos((*xs)[i])+acos((*xs)[i+1]))/2.;
		th2 = M_PI;
	    } else {
		th1 = 0.0;
		th2 = fabs(acos((*xs)[i])+acos((*xs)[i+1]))/2.;
	    }
	} else if (i == *nxs - 1) {
	    if (x[i] < 0.0)  {
		th1 = 0.0;
		th2 = fabs(acos((*xs)[i])+acos((*xs)[i-1]))/2.;
	    } else {
		th1 = fabs(acos((*xs)[i])+acos((*xs)[i-1]))/2.;
		th2 = M_PI;
	    }
	} else {
	    double t1 = acos((*xs)[i-1]),
		   t2 = acos((*xs)[i]),
		   t3 = acos((*xs)[i+1]);
	    th1 = (t1 + t2) / 2;
	    th2 = (t2 + t3) / 2;
	    if (th1 > th2)  {t1 = th1; th1 = th2; th2 = t1; }
	}

	for (j=0; j < (*nphs)[i]; j++) {
	    ph1 = j * dphi;
	    ph2 = (j+1) * dphi;

	    (*fsig)[ii+j] = sig_area(f,x,nx,np,nph, th1,th2,ph1,ph2);
	}
    }

    return;
}

/* calculate correlation in selected area */
double sig_area(f,x,nx,np,nph, th1,th2,ph1,ph2)
double *f, *x;
int nx, np, *nph;
double th1,th2,ph1,ph2;
{
    double sum1 = 0.0, sig1 = 0.0;
    register int i, j;
    double pi2 = M_PI * 2;
    int i1, i2, i3, i4, j1, j2, j3, j4,
	i_min, i_max, j_min, j_max, n;

    /* find limits of area in pixels */
    angle2pixel(x, nx, nph, &i1, &j1, th1, ph1);
    angle2pixel(x, nx, nph, &i2, &j2, th2, ph2);
    if (ABS(ph2-pi2) < .00001)
	    j2 = nph[i1] - 1;
    angle2pixel(x, nx, nph, &i3, &j3, th1, ph2);
    if (ABS(ph2-pi2) < .00001)
	    j3 = nph[i3] - 1;
    angle2pixel(x, nx, nph, &i4, &j4, th2, ph1);

    i_min = i1 < i2 ? i1 : i2;
    i_min = i_min < i3 ? i_min : i3;
    i_min = i_min < i4 ? i_min : i4;
    i_max = i1 > i2 ? i1 : i2;
    i_max = i_max > i3 ? i_max : i3;
    i_max = i_max > i4 ? i_max : i4;
    j_min = j1 < j2 ? j1 : j2;
    j_min = j_min < j3 ? j_min : j3;
    j_min = j_min < j4 ? j_min : j4;
    j_max = j1 > j2 ? j1 : j2;
    j_max = j_max > j3 ? j_max : j3;
    j_max = j_max > j4 ? j_max : j4;

    n = 0;
    for(i=i_min; i<=i_max; i++) {
	int ii = i * np;
	double dphi = pi2/nph[i];
	double th = acos(x[i]);
	if (!((th1 <= th && th <= th2) || (th2 <= th && th <= th1)))
	    continue;
	for(j=j_min; j<=j_max; j++) {
	    double phi = j * dphi;
	    if (phi < ph1) continue;
	    if (phi > ph2 || j>=nph[i]) break;
	    sum1 += f[ii+j];
	    n++;
	}
    }
    if (n == 0)
	   return 0.0;

    sum1 /= (double)n;

    for(i=i_min; i<=i_max; i++) {
	int ii = i * np;
	double dphi = pi2/nph[i];
	double th = acos(x[i]);
	if (!((th1 <= th && th <= th2) || (th2 <= th && th <= th1)))
	    continue;
	for(j=j_min; j<=j_max; j++) {
	    int ij = ii + j;
	    double s1 = f[ij] - sum1;
	    double phi = j * dphi;
	    if (phi < ph1) continue;
	    if (phi > ph2 || j>=nph[i]) break;
	    sig1 += s1 * s1;
	}
    }
    sig1 /= (double)n;

    return sig1;
}


void map_rsig(f,x,nx,np,nph, diameter)
double *f, *x;
int nx, np, *nph;
double diameter;    /* diameter of window(arcmin) for sigma calculation */
{
    register int i, j, ii, jj;
    int n;
    double *fsig = (double *)0, *ss = (double *)0;
    double radius = diameter/2./60.*M_PI/180.;/* diam,arcmin -->rad,radian */

    n = nx * np;
    if ((fsig = (double *)calloc((unsigned)n, sizeof *fsig))==(double *)0){
nomem:
	PR(ER, "%s: no memory for output arrays...\n", myname);
	exit(1);
    }
    if ((ss = (double *)calloc((unsigned)n, sizeof *fsig))==(double *)0)
	goto nomem;

    for (ii=0; ii < nx; ii++) {
	register int is = ii * np;
	double dphi0 = 2*M_PI / nph[ii];
	double theta0 = acos(x[ii]);
	double cth0 = x[ii], sth0 = sqrt(1-cth0*cth0);
	int nx0, nx1, np0, np1;

	for (jj=0; jj < nph[ii]; jj++) {
	    double phi0 = jj * dphi0;
	    int k = 0, n_k=0;
	    double ss_aver = 0.0, sig = 0.0;

	    /* extract circle */
	    /*----------------*/
	    nx0 = ind_x(nx, x, cos(theta0-radius));
	    nx1 = ind_x(nx, x, cos(theta0+radius));
	    if ( nx0 > nx1 ) {int a=nx0; nx0=nx1; nx1=a; }

	    for(i=0; i<nx; i++) {
		int ik = i * np;
		double cth = x[i], sth = sqrt(1.-cth*cth);
		double dphi = 2.*M_PI / (double)nph[i];
		double dp;   /* dt = theta - acos(cth), */
		double v;

		if (theta0-radius < 0. || theta0+radius > M_PI ) {
		   /* process pixels near pole */
		   if (!((theta0+radius > M_PI &&
		     ((x[0]>x[1] && i < nx0) || (x[0]<x[1] && i > nx1))) ||
		     (theta0-radius < 0.   &&
		     ((x[0]>x[1] && i > nx1) || (x[0]<x[1] && i < nx0))))) {
			for(j=0; j<nph[i]; j++) {
			    dp = phi0 - j*dphi;
			    v = acos(cos(dp)*sth0*sth + cth0*cth);
			    if (v <= radius)  {
				    ss[k] = f[ik+j];
				    ss_aver += ss[k];
				    k++;
				    continue;
			    }
			}
		   }
		} else {
		    /* process areas far from poles */
		    /* make checking interval a little larger */
		    double rl = radius/sth0/sth;

		    if (i >= nx0 && i <= nx1) {
			np0 = (phi0-rl) / dphi;
			np1 = (phi0+rl) / dphi;
			if ( np1 < np0 ) {int a=np0; np0=np1; np1=a; }

			if (np0 < 0) {
			     int npR = nph[i] + np0;
			     for(j=0; j<nph[i]; j++) {
			       if (npR < j || j < np1) {
				 dp = (2*M_PI+phi0) - j*dphi;
				 v = acos(cos(dp)*sth0*sth + cth0*cth);
				 if (v <= radius) {
					ss[k] = f[ik+j];
					ss_aver += ss[k];
					k++;
					continue;
				 }
			       }
			     }
			}
			if (np1 > nph[i]) {
			    int npR = np1 - nph[i];
			    for(j=0; j<nph[i]; j++) {
			      if (npR < j || j < np0) {
				dp = -phi0 + j*dphi + 2*M_PI;
				v = acos(cos(dp)*sth0*sth + cth0*cth);
				if (v <= radius) {
					ss[k] = f[ik+j];
					ss_aver += ss[k];
					k++;
					continue;
				}
			      }
			    }
			}
			for(j=0; j<nph[i]; j++) {
			    if (j >= np0 && j <= np1) {
				  dp = phi0 - j*dphi;
				  v = acos(cos(dp)*sth0*sth + cth0*cth);
				  if (v <= radius) {
					ss[k] = f[ik+j];
					ss_aver += ss[k];
					k++;
					continue;
				  }
			    }
			}
		    }
		}
	    }
	    n_k = k;
	    ss_aver /= (double)n_k;
	    for(k=0; k<n_k; k++) {
		double sv = ss[k] - ss_aver;
		sig += sv * sv;
		ss[k] = 0.0;
	    }
	    fsig[is+jj] = sqrt(sig/(double)n_k) ;
	}
    }

    for (ii=0; ii < nx; ii++) {
	register int is = ii * np;
	for (jj=0; jj < nph[ii]; jj++) f[is+jj] = fsig[is+jj];
    }

    free((void *)fsig);
    free((void *)ss);
    return;
}


void  map_rcor(f1, f2, x, nx, np, nph, diameter, fcor, xc, nxc, npc, nphc)
double *f1, *f2, *x;
int nx, np, *nph;
double diameter;         /* correlation window, min */
double **fcor,           /* output array            */
       **xc;             /* output X-array          */
int *nxc, *npc, **nphc;  /* output nx,np,nph        */
{
    register int i, j, ii, jj;
    int n;
    double *ss1 = (double *)0, *ss2 = (double *)0;
    double radius = diameter/2./60.*M_PI/180.;/*diam,arcmin-->radius,radian*/

    /* dimension of output arrays */
    *nxc = nx;
    *npc = np;
    n = *nxc * *npc;
    /* output arrays */
    if ((*fcor = (double *)calloc((unsigned)n, sizeof **fcor))==(double *)0){
nomem:
	PR(ER, "%s: no memory for output arrays...\n", myname);
	exit(1);
    }
    if ((*xc = (double *)calloc((unsigned)(*nxc), sizeof **xc))==(double *)0)
	goto nomem;
    if ((*nphc = (int *)calloc((unsigned)(*nxc), sizeof **nphc))==(int *)0)
	goto nomem;

    /* temporal buffers */
    if ((ss1 = (double *)calloc((unsigned)n, sizeof *ss1))==(double *)0)
	goto nomem;
    if ((ss2 = (double *)calloc((unsigned)n, sizeof *ss2))==(double *)0)
	goto nomem;

    for (ii=0; ii < nx; ii++) {
	(*nphc)[ii] = nph[ii];
	(*xc)[ii] = x[ii];
    }

    /* loop over rings of output arrays */
    for (ii=0; ii < nx; ii++) {
	register int is = ii * *npc;
	double dphi0 = 2*M_PI / (*nphc)[ii];
	double theta0 = acos((*xc)[ii]);
	double cth0 = (*xc)[ii], sth0 = sqrt(1-cth0*cth0);
	int nx0, nx1, np0, np1;

PR(ER, "%4d\b\b\b\b", ii);
	for (jj=0; jj < nph[ii]; jj++) {
	    double phi0 = jj * dphi0;
	    int k = 0, n_k=0;
	    double ss_aver1 = 0.0, sig1 = 0.0;
	    double ss_aver2 = 0.0, sig2 = 0.0;
	    double cov = 0.0, corr = 0.0;

	    /* limits in rings to extract circle */
	    /*-----------------------------------*/
	    nx0 = ind_x(nx, x, cos(theta0-radius));
	    nx1 = ind_x(nx, x, cos(theta0+radius));
	    if ( nx0 > nx1 ) {int a=nx0; nx0=nx1; nx1=a; }
	    /*----------------*/
	    /* extract circle */
	    /*----------------*/
	    for(i=0; i<nx; i++) {
		int ik = i * np;
		double cth = x[i], sth = sqrt(1.-cth*cth);
		double dphi = 2.*M_PI / (double)nph[i];
		double dp;   /* dt = theta - acos(cth), */
		double v;

		if (theta0-radius < 0. || theta0+radius > M_PI ) {
		   /* process pixels near poles */
		   /*--------------------------*/
		   if (!((theta0+radius > M_PI &&
		     ((x[0]>x[1] && i < nx0) || (x[0]<x[1] && i > nx1))) ||
		     (theta0-radius < 0.   &&
		     ((x[0]>x[1] && i > nx1) || (x[0]<x[1] && i < nx0))))) {
			for(j=0; j<nph[i]; j++) {
			    dp = phi0 - j*dphi;
			    v = acos(cos(dp)*sth0*sth + cth0*cth);
			    if (v <= radius)  {
				    ss1[k] = f1[ik+j];
				    ss_aver1 += ss1[k];
				    ss2[k] = f2[ik+j];
				    ss_aver2 += ss2[k];
				    k++;
				    continue;
			    }
			}
		   }
		} else {
		    /* process areas far from poles;          */
		    /* make checking interval a little larger */
		    /* ---------------------------------------*/
		    double rl = radius/sth0/sth;

		    if (i >= nx0 && i <= nx1) {
			/* limits in pixels to extract circle */
			np0 = (phi0-rl) / dphi;
			np1 = (phi0+rl) / dphi;
			if ( np1 < np0 ) {int a=np0; np0=np1; np1=a; }

			if (np0 < 0) {
			     int npR = nph[i] + np0;
			     for(j=0; j<nph[i]; j++) {
			       if (npR < j || j < np1) {
				 dp = (2*M_PI+phi0) - j*dphi;
				 v = acos(cos(dp)*sth0*sth + cth0*cth);
				 if (v <= radius) {
					ss1[k] = f1[ik+j];
					ss_aver1 += ss1[k];
					ss2[k] = f2[ik+j];
					ss_aver2 += ss2[k];
					k++;
					continue;
				 }
			       }
			     }
			}
			if (np1 > nph[i]) {
			    int npR = np1 - nph[i];
			    for(j=0; j<nph[i]; j++) {
			      if (npR < j || j < np0) {
				dp = -phi0 + j*dphi + 2*M_PI;
				v = acos(cos(dp)*sth0*sth + cth0*cth);
				if (v <= radius) {
					ss1[k] = f1[ik+j];
					ss_aver1 += ss1[k];
					ss2[k] = f2[ik+j];
					ss_aver2 += ss2[k];
					k++;
					continue;
				}
			      }
			    }
			}
			for(j=0; j<nph[i]; j++) {
			    if (j >= np0 && j <= np1) {
				  dp = phi0 - j*dphi;
				  v = acos(cos(dp)*sth0*sth + cth0*cth);
				  if (v <= radius) {
					ss1[k] = f1[ik+j];
					ss_aver1 += ss1[k];
					ss2[k] = f2[ik+j];
					ss_aver2 += ss2[k];
					k++;
					continue;
				  }
			    }
			}
		    }
		}
	    }  /* end of circle extracting, circle of 2 maps in ss1 & ss2 */
	    n_k = k;
	    ss_aver1 /= (double)n_k;
	    ss_aver2 /= (double)n_k;

	    for(k=0; k<n_k; k++) {
		double sv1 = ss1[k] - ss_aver1;
		double sv2 = ss2[k] - ss_aver2;
		sig1 += sv1 * sv1;
		sig2 += sv2 * sv2;
		cov  += sv1 * sv2;
		ss1[k] = 0.0;
		ss2[k] = 0.0;
	    }
	    sig1 /= (double)n_k;
	    sig2 /= (double)n_k;
	    cov  /= (double)n_k;

	    if (fabs(sig1) < 1.e-30  ||  fabs(sig2) < 1.e-30)
		corr = cov;
	    else
		corr = cov / sqrt(sig1*sig2);

	    (*fcor)[is+jj] = corr ;
	}
    }

    free((void *)ss1);
    free((void *)ss2);

    return;
}


void  map_cor_mask(f1, f2, mask, x, nx, np, nph, corr_win, fcor, xc, nxc, npc,
		    nphc, dummy)
double *f1, *f2, *x;
int *mask;
int nx, np, *nph;
double corr_win;         /* correlation window, min */
double **fcor,           /* output array            */
       **xc;             /* output X-array          */
int *nxc, *npc, **nphc;  /* output nx,np,nph        */
double dummy;
{
    register int i, j;
    int n;
    double th1, th2, ph1, ph2;
    double x1, x2;
    int pin_flag = nph[0] == np ? 0 : 1;
    extern double corr_area_mask(double *, double *, int *, double *, int,int,
	int *, double,double,double,double,double);

    *nxc = 180 * 60 / corr_win + 0.5;
    corr_win = 180 * 60 / *nxc;

    if (glesp_comments) {
	PR(ER, "\nactual correlation window of %gx%g (%g) sq.arcmin.\n",
		corr_win, corr_win, corr_win*corr_win);
    }
    *npc = 2 * *nxc;

    n = *nxc * *npc;
    if ((*fcor = (double *)calloc((unsigned)n, sizeof **fcor))==(double *)0){
nomem:
	PR(ER, "%s: no memory for output arrays...\n", myname);
	exit(1);
    }
    if ((*xc = (double *)calloc((unsigned)(*nxc), sizeof **xc))==(double *)0)
	goto nomem;
    if ((*nphc = (int *)calloc((unsigned)(*nxc), sizeof **nphc))==(int *)0)
	goto nomem;

    x1 = -1.0; x2 = 1.0;
    gauleg(x1, x2, *xc, (double *)0, *nxc);
    if (pin_flag) {
	/* GLESP pixelization */
//      for (i=0; i < *nxc; i++)  (*nphc)[i] = get_nphi_by_x(*xc, *nxc, i);
	for (i=0; i < *nxc; i++)
	    (*nphc)[i] = get_nphi_by_x2(*xc, *nxc, *npc, i);
    } else {
	/* pin = 0 */
	for (i=0; i < *nxc; i++)  (*nphc)[i] = *npc;
    }


    n = *nxc - 1;
    for (i=0; i < *nxc; i++) {
	register int ii = i * *npc;
	double dphi = 2*M_PI / (*nphc)[i];

	if (i == 0) {
	    if (x[i] < 0.0)  {
		th1 = fabs(acos((*xc)[i])+acos((*xc)[i+1]))/2.;
		th2 = M_PI;
	    } else {
		th1 = 0.0;
		th2 = fabs(acos((*xc)[i])+acos((*xc)[i+1]))/2.;
	    }
	} else if (i == *nxc - 1) {
	    if (x[i] < 0.0)  {
		th1 = 0.0;
		th2 = fabs(acos((*xc)[i])+acos((*xc)[i-1]))/2.;
	    } else {
		th1 = fabs(acos((*xc)[i])+acos((*xc)[i-1]))/2.;
		th2 = M_PI;
	    }
	} else {
	    double t1 = acos((*xc)[i-1]),
		   t2 = acos((*xc)[i]),
		   t3 = acos((*xc)[i+1]);
	    th1 = (t1 + t2) / 2;
	    th2 = (t2 + t3) / 2;
	    if (th1 > th2)  {t1 = th1; th1 = th2; th2 = t1; }
	}

	for (j=0; j < (*nphc)[i]; j++) {
	    ph1 = j * dphi;
	    ph2 = (j+1) * dphi;

	    (*fcor)[ii+j] =
		corr_area_mask(f1,f2,mask,x,nx,np,nph, th1,th2,ph1,ph2,dummy);
	}
    }

    return;
}

/* calculate correlation in selected area */
double corr_area_mask(f1,f2,mask,x,nx,np,nph, th1,th2,ph1,ph2,dummy)
double *f1, *f2, *x;
int *mask;
int nx, np, *nph;
double th1,th2,ph1,ph2;
double dummy;
{
    double corr=0.0, sum1 = 0.0, sum2 = 0.0,
	   cov= 0.0, sig1 = 0.0, sig2 = 0.0;
    register int i, j;
    double pi2 = M_PI * 2;
    int i1, i2, i3, i4, j1, j2, j3, j4,
	i_min, i_max, j_min, j_max, n;
    int mask_counts = 0;

    /* find limits of area in pixels */
    angle2pixel(x, nx, nph, &i1, &j1, th1, ph1);
    angle2pixel(x, nx, nph, &i2, &j2, th2, ph2);
    if (ABS(ph2-pi2) < .00001)
	    j2 = nph[i1] - 1;
    angle2pixel(x, nx, nph, &i3, &j3, th1, ph2);
    if (ABS(ph2-pi2) < .00001)
	    j3 = nph[i3] - 1;
    angle2pixel(x, nx, nph, &i4, &j4, th2, ph1);

    i_min = i1 < i2 ? i1 : i2;
    i_min = i_min < i3 ? i_min : i3;
    i_min = i_min < i4 ? i_min : i4;
    i_max = i1 > i2 ? i1 : i2;
    i_max = i_max > i3 ? i_max : i3;
    i_max = i_max > i4 ? i_max : i4;
    j_min = j1 < j2 ? j1 : j2;
    j_min = j_min < j3 ? j_min : j3;
    j_min = j_min < j4 ? j_min : j4;
    j_max = j1 > j2 ? j1 : j2;
    j_max = j_max > j3 ? j_max : j3;
    j_max = j_max > j4 ? j_max : j4;

    n = 0;
    for(i=i_min; i<=i_max; i++) {
	int flag_break = 0;
	int ii = i * np;
	double dphi = pi2/nph[i];
	double th = acos(x[i]);
	if (!((th1 <= th && th <= th2) || (th2 <= th && th <= th1)))
	    continue;
	for(j=j_min; j<=j_max; j++) {
	    double phi = j * dphi;
	    if (phi < ph1) continue;
	    if (phi > ph2 || j>=nph[i]) break;
	    if ( mask[ii+j] ) {
		mask_counts++;
		flag_break = 1;
		continue;
	    }
	    if (flag_break)  continue;
	    sum1 += f1[ii+j];
	    sum2 += f2[ii+j];
	    n++;
	}
    }
    if (mask_counts > n/3)
	   return dummy;

    if (n == 0)
	   return 0.0;


    sum1 /= (double)n;
    sum2 /= (double)n;

    for(i=i_min; i<=i_max; i++) {
	int flag_break = 0;
	int ii = i * np;
	double dphi = pi2/nph[i];
	double th = acos(x[i]);
	if (!((th1 <= th && th <= th2) || (th2 <= th && th <= th1)))
	    continue;
	for(j=j_min; j<=j_max; j++) {
	    int ij = ii + j;
	    double s1 = f1[ij] - sum1, s2 = f2[ij] - sum2;
	    double phi = j * dphi;
	    if (phi < ph1) continue;
	    if (phi > ph2 || j>=nph[i]) break;
	    if ( mask[ii+j] ) {
		flag_break = 1;
		continue;
	    }
	    if (flag_break)  continue;
	    sig1 += s1 * s1;
	    sig2 += s2 * s2;
	    cov  += s1 * s2;
	}
    }
    sig1 /= (double)n;
    sig2 /= (double)n;
    cov  /= (double)n;

    if (fabs(sig1) < 1.e-30  ||  fabs(sig2) < 1.e-30)
	corr = cov;
    else
	corr = cov / sqrt(sig1*sig2);

    return corr;
}



