#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "glesp.h"
#define PR  fprintf
#define ER  stderr

/*
 *   find extrema on a map
 *
 *   Oleg Verkhodanov, 07/02/2010, vo@sao.ru
 */
struct _pixel_ {
    int i, j;
    double theta, phi;
    double v;
};
extern int circle_select_pixels(double *, double *, int, int, int *,
	      double, double, double, struct _pixel_ **);

int map_extrem_mask(f, x, nx, np, nph, imask, mex, dp)
double *f, *x;
int nx, np, *nph;
int *imask;
int *mex;
double dp;           /* radius to search for extrema */
{
    int n_extr = 0;             /* number of extrema */
    register int i,j,k;
    double r = 2*M_PI / np * 2;  /* defult radius: of 2 pixels */
    struct _pixel_ *cp = (struct _pixel_ *)0;
int pc=0;

    if (dp > 0.)  r = dp;

    for (i=0; i<nx; i++) {
	int ii = np * i;
	double dp = 2*M_PI / nph[i];
	double t0 = acos (x[i]); /* theta */
	for (j=0; j<nph[i]; j++) {
	    double min = 1e+30, max = -1e+30;
	    int i_min=-1, i_max=-1, j_min=-1, j_max= -1;
	    int nn;
	    /* coordinates (t0,p0) */
	    double p0 = j * dp;  /* phi */

	    if (imask[ii+j])  continue;
	    /* select all pixels in a circle of radius 'r' around (t0,p0) */
	    nn = circle_select_pixels(f, x, nx, np, nph, t0, p0, r, &cp);

	    /* check for extrema in this circle */
	    for (k=0; k<nn; k++) {
		if (cp[k].v >max) {max=cp[k].v; i_max=cp[k].i;j_max=cp[k].j;}
		if (cp[k].v <min) {min=cp[k].v; i_min=cp[k].i;j_min=cp[k].j;}
	    }
	    if (i_min == i && j_min == j)  {mex[ii+j] = -1; n_extr++;}
	    else if(i_max==i&&j_max == j)  {mex[ii+j] =  1; n_extr++;}
	}
    }
    free((void *)cp);
    return n_extr;
}

int map_extrem(f, x, nx, np, nph, mex, dp)
double *f, *x;
int nx, np, *nph;
int *mex;
double dp;           /* radius to search for extrema */
{
    int n_extr = 0;             /* number of extrema */
    register int i,j,k;
    double r = 2*M_PI / np * 2;  /* defult radius: of 2 pixels */
    struct _pixel_ *cp = (struct _pixel_ *)0;

    if (dp > 0.)  r = dp;

    for (i=0; i<nx; i++) {
	int ii = np * i;
	double dp = 2*M_PI / nph[i];
	double t0 = acos (x[i]); /* theta */
	for (j=0; j<nph[i]; j++) {
	    double min = 1e+30, max = -1e+30;
	    int i_min=-1, i_max=-1, j_min=-1, j_max= -1;
	    int nn;
	    /* coordinates (t0,p0) */
	    double p0 = j * dp;  /* phi */

	    /* select all pixels in a circle of radius 'r' around (t0,p0) */
	    nn = circle_select_pixels(f, x, nx, np, nph, t0, p0, r, &cp);

	    /* check for extrema in this circle */
	    for (k=0; k<nn; k++) {
		if (cp[k].v >max) {max=cp[k].v; i_max=cp[k].i;j_max=cp[k].j;}
		if (cp[k].v <min) {min=cp[k].v; i_min=cp[k].i;j_min=cp[k].j;}
	    }
	    if (i_min == i && j_min == j)  {mex[ii+j] = -1; n_extr++;}
	    else if(i_max==i&&j_max == j)  {mex[ii+j] =  1; n_extr++;}
	}
    }

    free((void *)cp);
    return n_extr;
}

int circle_select_pixels(f, x, nx, np, nph, theta0, phi0, radius, cp)
double *f, *x;
int nx, np, *nph;
double phi0, theta0, radius;  /* center:(theta0,phi0) of radius */
struct _pixel_ **cp;
{
    register int i, j;
    int nn=0;
    int flag = 0;
    unsigned int nbuf;
    double cth0, sth0, v;
    int nx_c, np_c, nx0, nx1, np0, np1;
    double pi2 = 2. * M_PI;

    /* upper limit of nbuf: = np * (int)(2.*radius / (pix = 2*M_PI/np)) */
    nbuf = np * np * radius / M_PI;

    if (*cp == (struct _pixel_ *)0)
	*cp = (struct _pixel_ *)calloc(nbuf, sizeof **cp);
    if (*cp == (struct _pixel_ *)0) {
	flag = 1;
nomem:
	PR(ER, "%s: no memory for array of extrema pixels (err=%d)\n",
	    myname, flag);
	exit(1);
    }

    cth0 = cos(theta0);
    sth0 = sqrt(1-cth0*cth0);
    nx_c = ind_x(nx, x, cth0);
    np_c = phi0 / pi2 * (double)nph[nx_c];

    nx0 = ind_x(nx, x, cos(theta0-radius));
    nx1 = ind_x(nx, x, cos(theta0+radius));
    if ( nx0 > nx1 ) { j=nx0; nx0=nx1; nx1=j; }

    if (theta0-radius < 0. || theta0+radius > M_PI ) {
      /* process pixels near pole */
      for(i=0; i<nx; i++) {
	int ii = i * np;
	double cth = x[i], sth = sqrt(1.-cth*cth);
	double dphi = pi2 / (double)nph[i];
	double dp;   /* dt = theta0 - acos(cth), */

	if (!((theta0+radius > M_PI &&
	  ((x[0]>x[1] && i < nx0) || (x[0]<x[1] && i > nx1))) ||
	  (theta0-radius < 0.   &&
	  ((x[0]>x[1] && i > nx1) || (x[0]<x[1] && i < nx0))))) {
	     for(j=0; j<nph[i]; j++) {
		 dp = phi0 - j*dphi;
		 v = acos(cos(dp)*sth0*sth + cth0*cth);
		 if (v <= radius)  {
			 (*cp)[nn].i = i;
			 (*cp)[nn].j = j;
			 (*cp)[nn].theta = acos(cth);
			 (*cp)[nn].phi = j * dphi;
			 (*cp)[nn].v = f[ii+j];
			 nn++;
			 if (nn >= nbuf) {
				flag = 2;
				goto nomem;
			 }
			 continue;
		 }
	     }
	}
      }
    } else {
       for(i=0; i<nx; i++) {
	 int ii = i * np;
	 double cth = x[i],  sth = sqrt(1-cth*cth);
	 double dphi = pi2 / (double)nph[i];
	 double dp; /* dt = theta0 - acos(cth), */
	 /* make checking interval a little larger */
	 double rl = radius/sth0/sth;

	 if (i >= nx0 && i <= nx1) {
	     np0 = (phi0-rl) / dphi;
	     np1 = (phi0+rl) / dphi;
	     if ( np1 < np0 ) {j=np0; np0=np1; np1=j; }

	     if (np0 < 0) {
		  int npR = nph[i] + np0;
		  for(j=0; j<nph[i]; j++) {
		    if (npR < j || j < np1) {
		      dp = (2*M_PI+phi0) - j*dphi;
		      v = acos(cos(dp)*sth0*sth + cth0*cth);
		      if (v <= radius) {
			 (*cp)[nn].i = i;
			 (*cp)[nn].j = j;
			 (*cp)[nn].theta = acos(cth);
			 (*cp)[nn].phi = j * dphi;
			 (*cp)[nn].v = f[ii+j];
			 nn++;
			 if (nn >= nbuf) {
				flag = 3;
				goto nomem;
			 }
			 continue;
		      }
		    }
		  }
	     }
	     if (np1 > nph[i]) {
		 int npR = np1 - nph[i];
		 for(j=0; j<nph[i]; j++) {
		   if (npR < j || j < np0) {
		     dp = -phi0 + j*dphi + 2*M_PI;
		     v = acos(cos(dp)*sth0*sth + cth0*cth);
		     if (v <= radius) {
			 (*cp)[nn].i = i;
			 (*cp)[nn].j = j;
			 (*cp)[nn].theta = acos(cth);
			 (*cp)[nn].phi = j * dphi;
			 (*cp)[nn].v = f[ii+j];
			 nn++;
			 if (nn >= nbuf) {
				flag = 3;
				goto nomem;
			 }
			 continue;
		     }
		   }
		 }
	     }
	     for(j=0; j<nph[i]; j++) {
		 if (j >= np0 && j <= np1) {
		       dp = phi0 - j*dphi;
		       v = acos(cos(dp)*sth0*sth + cth0*cth);
		       if (v <= radius) {
			    (*cp)[nn].i = i;
			    (*cp)[nn].j = j;
			    (*cp)[nn].theta = acos(cth);
			    (*cp)[nn].phi = j * dphi;
			    (*cp)[nn].v = f[ii+j];
			    nn++;
			    if (nn >= nbuf) {
				   flag = 4;
				   goto nomem;
			    }
			    continue;
		       }
		 }
	     }
	 }
       }
    }
    return nn;
}
