#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "glesp.h"

#define PR  fprintf
#define ER  stderr

void map_grad_phi(f, x, nx, np, nph)
double *f, *x;
int nx, np;
int *nph;
{
    register int i, j;
    double *df = (double *)0;
    double pi2 = 2 * M_PI;

    if ((df = (double*)calloc((unsigned)(nx*np), sizeof *df))==(double*)0) {
	PR(ER, "%s: no memory for temporal array...\n", myname);
	exit(1);
    }

    for(i=0; i<nx; i++) {
	int ii = i * np;
	double dphi = pi2 / nph[i];
	for(j=0; j<nph[i]; j++) {
	    if (j==0)
//                df[ii+j] = f[ii+j] - .5 * (f[ii+nph[i]-1 ] + f[ii+j+1]);
		  df[ii+j] = f[ii+j] - f[ii+nph[i]-1 ];
//            else if (j==nph[i]-1)
//                df[ii+j] = f[ii+j] - .5 * (f[ii+j-1] + f[ii]);
	    else
//                df[ii+j] = f[ii+j] - .5 * (f[ii+j-1] + f[ii+j+1]);
		df[ii+j] = f[ii+j] - f[ii+j-1];
	}
	for(j=0; j<nph[i]; j++)    f[ii+j] = df[ii+j]/dphi;
    }

    free((void *)df);
    return;
}


void map_grad_theta(f, x, nx, np, nph)
double *f, *x;
int nx, np;
int *nph;
{
    register int i, j;
    double *df = (double *)0;
    double pi2 = 2 * M_PI;

    if ((df = (double*)calloc((unsigned)(nx*np), sizeof *df))==(double*)0) {
	PR(ER, "%s: no memory for temporal array...\n", myname);
	exit(1);
    }

    for(i=0; i<nx; i++) {
	int ii = i * np;
	double f_up, f_down;
	double phi, theta_up, theta_down;
	double dphi = pi2 / nph[i];

	theta_up   = i ? acos(x[i-1]) : -1.;
	theta_down = i != nx-1 ?  acos(x[i+1]) : -1.;

	for(j=0; j<nph[i]; j++) {
	    phi = j * dphi;
	    if (i==0) {
		  int n;
//                int n = ang2n(theta_down, phi, x, nx, np, nph);
//                f_up = 0.;
//                f_down = 2. * f[n];
		  n = j - nph[0]/2;
		  if (n < 0) n += nph[0];
		  f_up = f[n];
//            } else if (i == nx-1) {
//                int n = ang2n(theta_up, phi, x, nx, np, nph);
//                f_up = 2. * f[n];
//                f_down = 0.;
	    } else {
		int n_up   = ang2n(theta_up,   phi, x, nx, np, nph);
		int n_down = ang2n(theta_down, phi, x, nx, np, nph);
		f_up   = f[n_up];
		f_down = f[n_down];
/*
PR(ER, "i=%d, j=%d, x=%g, phi=%g\n", i, j, x[i], phi);
PR(ER, "    theta_up=%g, theta_down=%g\n", theta_up, theta_down);
PR(ER, "    n_up=%d, n_down=%d\n", n_up, n_down);
PR(ER, "    f_up=%g, f_down=%g\n", f_up, f_down);
*/
	    }
//            df[ii+j] = f[ii+j] - .5 * (f_up + f_down);
	    df[ii+j] = f[ii+j] - f_up;
	}
    }

    for(i=0; i<nx; i++) {
	int ii = i * np;
	double dx;

	if (i==0) {
	    dx = x[i+1] -x [i];
	} else if (i == nx-1) {
	    dx = x[i] - x[i-1];
	} else {
	    dx = 0.5 * (x[i+1] - x[i-1]);
	}

	for(j=0; j<nph[i]; j++)  f[ii+j] = df[ii+j] / dx;
    }

    free((void *)df);
    return;
}

void map_grad_pa(q, u, p, pa, x, nx, np, nph)
double *u, *q, *p, *pa, *x;
int nx, np, *nph;
{
    register int i, j;
    for(i=0; i<nx; i++) {
	int ii = i * np;
	for(j=0; j<nph[i]; j++)  {
	    p[ii+j] = sqrt(q[ii+j]*q[ii+j] + u[ii+j]*u[ii+j]);
	    pa[ii+j] = ph_calc (q[ii+j], u[ii+j]);
	}
    }
    return;
}
