#include <stdio.h>

#define ABS(x)  ((x) < 0 ? -(x) : (x))
#define PR  fprintf
#define ER  stderr

void map_separation(ch1, ch2, x, nx, np, nph, mask, f)
double *ch1, *ch2, *x, *f;
int nx, np, *nph;
int *mask;
{
    register int i,j;
    double sa=0.0, sb=0.0, w=0.0;

    for(i=0; i<nx; i++) {
	int ii = i * np;
	for(j=0; j<nph[i]; j++) {
	    int jj = ii + j;
	    double ss = ch1[jj] - ch2[jj];
	    if (mask[jj]) continue;
	    sa += ch2[jj] * ss;
	    sb += ss * ss;
	}
    }
    if (ABS(sb) > 1.e-30)
	w = -sa/sb;
    else if (ABS(sa) < 1.e-30 && ABS(sb) < 1.e-30 && ABS(sb) > 1.e-60)
	w = -sa/sb;
    else
	w = 0.;

//PR(ER, ".................w=%g\n", w);
    for(i=0; i<nx; i++) {
	int ii = i * np;
	for(j=0; j<nph[i]; j++) {
	    int jj = ii + j;
	    if (mask[jj]) {
		f[jj] = 0.0;
		continue;
	    }
	    f[jj] = ch2[jj] + w * (ch1[jj] - ch2[jj]);
	}
    }
    return;
}

void calc_alpha(ch1, ch2, x, nx, np, nph, mask, nt, alpha)
double *ch1, *ch2, *x;
int nx, np, *nph;
int *mask;
int *nt;
double *alpha;
{
    register int i,j;
    double sa=0.0, sb=0.0, w=0.0;
    int n = 0;

    for(i=0; i<nx; i++) {
	int ii = i * np;
	for(j=0; j<nph[i]; j++) {
	    int jj = ii + j;
	    double ss = ch1[jj] - ch2[jj];
	    if (mask[jj]) continue;
	    n++;
	    sa += ch2[jj] * ss;
	    sb += ss * ss;
	}
    }
PR(ER, "sa=%g, sb=%g\n", sa, sb);
    if (ABS(sb) > 1.e-30)
	w = -sa/sb;
    else if (ABS(sa) < 1.e-30 && ABS(sb) < 1.e-30 && ABS(sb) > 1.e-60)
	w = -sa/sb;
    else
	w = 0.;

    *alpha = -w;
    *nt = n;
    return;
}


void map_separation2(ch1, ch2, fg1, fg2, x, nx, np, nph, mask, f)
double *ch1, *ch2, *fg1, *fg2, *x, *f;
int nx, np, *nph;
int *mask;
{
    register int i,j;
    double sa=0.0, sb=0.0, w=0.0;

    for(i=0; i<nx; i++) {
	int ii = i * np;
	for(j=0; j<nph[i]; j++) {
	    int jj = ii + j;
	    double f_betta1 = fg2[jj] * (1. - fg1[jj] / fg2[jj]);
	    if (mask[jj]) continue;
	    sa += ch2[jj] * f_betta1;
	    sb += f_betta1 * f_betta1;
	}
    }
    if (ABS(sb) > 1.e-30)
	w = -sa/sb;
    else if (ABS(sa) < 1.e-30 && ABS(sb) < 1.e-30 && ABS(sb) > 1.e-60)
	w = -sa/sb;
    else
	w = 0.;

//PR(ER, ".................w=%g\n", w);
    for(i=0; i<nx; i++) {
	int ii = i * np;
	for(j=0; j<nph[i]; j++) {
	    int jj = ii + j;
	    if (mask[jj]) {
		f[jj] = 0.0;
		continue;
	    }
	    f[jj] = ch2[jj] + w * (ch1[jj] - ch2[jj]);
	}
    }
    return;
}


void map_separ_alpha(ch1, ch2, x, nx, np, nph, mask, f, alpha)
double *ch1, *ch2, *x, *f;
int nx, np, *nph;
int *mask;
double alpha;
{
    register int i,j;
    double  w= alpha;

    for(i=0; i<nx; i++) {
	int ii = i * np;
	for(j=0; j<nph[i]; j++) {
	    int jj = ii + j;
	    if (mask[jj]) {
		f[jj] = 0.0;
		continue;
	    }
	    f[jj] = ch2[jj] + w * (ch1[jj] - ch2[jj]);
	}
    }

    return;
}



void map_separ_sq(ch1, ch2, x, nx, np, nph, mask, f)
double *ch1, *ch2, *x, *f;
int nx, np, *nph;
int *mask;
{
    register int i,j;
    double s2_aver=0.0, s12_aver=0.0, q2=0.0, q12=0.0, w=0.0;
    int ntot = 0;

    for(i=0; i<nx; i++) {
	int ii = i * np;
	for(j=0; j<nph[i]; j++) {
	    int jj = ii + j;
	    if (mask[jj]) continue;
	    ntot++;
	    s2_aver  += ch2[jj];
	    s12_aver += ch1[jj] - ch2[jj];
	}
    }
    s2_aver  /= ntot;
    s12_aver /= ntot;

    for(i=0; i<nx; i++) {
	int ii = i * np;
	for(j=0; j<nph[i]; j++) {
	    double v_q2, v_q12;
	    int jj = ii + j;
	    if (mask[jj]) continue;
	    v_q2  =  ch2[jj] - s2_aver;
	    v_q12 = (ch1[jj] - ch2[jj]) - s12_aver;
	    q2  += v_q2  * v_q2;
	    q12 += v_q12 * v_q12;
	}
    }
    q2  /= ntot;
    q12 /= ntot;


    if (ABS(q12) > 1.e-30)
	w = q2/q12;
    else
	w = 0.;

//PR(ER, "--------------w=%g\n", w);
    for(i=0; i<nx; i++) {
	int ii = i * np;
	for(j=0; j<nph[i]; j++) {
	    int jj = ii + j;
	    if (mask[jj]) {
		f[jj] = 0.0;
		continue;
	    }
	    f[jj] = ch2[jj] + w * (ch1[jj] - ch2[jj]);
	}
    }
    return;
}



void map_a_separation(ch1, ch2, x, nx, np, nph, mask, f)
double *ch1, *ch2, *x, *f;
int nx, np, *nph;
int *mask;
{
    register int i,j;
    double sa=0.0, sb=0.0, w=0.0;
    double s2 = 0.0, ds =0.0;

    for(i=0; i<nx; i++) {
	int ii = i * np;
	for(j=0; j<nph[i]; j++) {
	    int jj = ii + j;
	    double ss = ch1[jj] - ch2[jj];
	    if (mask[jj]) continue;
	    sa += ch2[jj] * ss;
	    sb += ss * ss;
	    s2 +=  ch2[jj];
	    ds += ss;
	}
    }
    if (ABS(sb) > 1.e-30)
	w = - (sa - s2*ds) / (sb - ds*ds);
    else if (ABS(sa) < 1.e-30 && ABS(sb) < 1.e-30 && ABS(sb) > 1.e-60)
	w = - (sa - s2*ds) / (sb - ds*ds);
    else
	w = 0.;

//PR(ER, ".................w=%g\n", w);
    for(i=0; i<nx; i++) {
	int ii = i * np;
	for(j=0; j<nph[i]; j++) {
	    int jj = ii + j;
	    if (mask[jj]) {
		f[jj] = 0.0;
		continue;
	    }
	    f[jj] = ch2[jj] + w * (ch1[jj] - ch2[jj]);
	}
    }
    return;
}

void calc_a_alpha(ch1, ch2, x, nx, np, nph, mask, nt, alpha)
double *ch1, *ch2, *x;
int nx, np, *nph;
int *mask;
int *nt;
double *alpha;
{
    register int i,j;
    double sa=0.0, sb=0.0, w=0.0;
    double s2 = 0.0, ds =0.0;
    int n = 0;

    for(i=0; i<nx; i++) {
	int ii = i * np;
	for(j=0; j<nph[i]; j++) {
	    int jj = ii + j;
	    double ss = ch1[jj] - ch2[jj];
	    if (mask[jj]) continue;
	    n++;
	    sa += ch2[jj] * ss;
	    sb += ss * ss;
	    s2 +=  ch2[jj];
	    ds += ss;
	}
    }
    if (ABS(sb) > 1.e-30)
	w = - (sa - s2*ds) / (sb - ds*ds);
    else if (ABS(sa) < 1.e-30 && ABS(sb) < 1.e-30 && ABS(sb) > 1.e-60)
	w = - (sa - s2*ds) / (sb - ds*ds);
    else
	w = 0.;

    *alpha = -w;
    *nt = n;
    return;
}


void calcd_a_alpha(ch1, ch2, x, nx, np, nph, mask, nt, alpha)
double *ch1, *ch2, *x;
int nx, np, *nph;
int *mask;
int *nt;
double *alpha;
{
    register int i,j;
    double sa=0.0, sb=0.0, w=0.0;
    double s2 = 0.0, ds =0.0;
    int n = 0;

    for(i=0; i<nx; i++) {
	int ii = i * np;
	for(j=0; j<nph[i]; j++) {
	    int jj = ii + j;
	    double ss = ch1[jj] - ch2[jj];
	    double sss = ch1[jj] + ch2[jj];
	    if (mask[jj]) continue;
	    n++;
	    sa += ch2[jj] * ss;
	    sb += ss * ss;
	    s2 +=  ch2[jj];
	    ds += ss;
	}
    }
    if (ABS(sb) > 1.e-30)
	w = - (sa - s2*ds) / (sb - ds*ds);
    else if (ABS(sa) < 1.e-30 && ABS(sb) < 1.e-30 && ABS(sb) > 1.e-60)
	w = - (sa - s2*ds) / (sb - ds*ds);
    else
	w = 0.;

    *alpha = -w;
    *nt = n;
    return;
}
