#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <unistd.h>

#include "glesp.h"

#define PR  fprintf
#define ER  stderr

extern void setopt(int, char **), usage(void);

int z_mask_flag = 0;      /* set mask                                  */
double zmask_value = M_PI * 1.e+30;
int zone_circle = 0;
char zone_file[81];       /* file with list of zones to be processed   */
double zone_radius = -1.; /* radius of removed circle in degs          */
double zone_latR0 = -1000., zone_lonR0 = -1000.,
       zone_latR1 = -1000., zone_lonR1 = -1000.;
int    z_remove_flag = 1,
       z_remain_flag = 0;

#define N_BIN_MIN   10
int n_bin_min = N_BIN_MIN;
int n_hist = 0;
int divi_flag=0;
int log_bin=0;
struct _lu_ {double l,u;} *lim = (struct _lu_ *)0;

int flag_hist_peak = 0;

char buffile[161];      /* file containing map of number of bins       */
int flag_buf = 0;       /* if flag=1 then output, if flag=2 then input */

double low_lim= -1.e+10, up_lim=1.e+10;
double hist_up_lim = -1.0001e+30, hist_down_lim = 1.0001e+30;
char map1_file[161], map2_file[161], outfile[161];
char file_limits[161];
char div_file[161];
int print_flag = 0;
double bin_size = -1.0;
char acf_file[161];
char adf_file[161];
int  no_squeeze=0;
int no_calc=0;
double alpha_cf = -2.7*M_PI;
int bin_zero = 0;
int set_1bin=0;
double bin1_min= -1.e+30, bin1_max = 1.e-30;
char acf_map_file[161];
char ncf_map_file[161];
char *myname="mapsep";

int main(argc, argv)
int argc;
char **argv;
{
    register int i, j, n;
    double *f1 = (double *)0, *f2 = (double *)0;
    double *x1 = (double *)0, *x2 = (double *)0;
    double *cmap = (double *)0, *t = (double *)0;
    double *zmask = (double *)0, *div_map = (double *)0,
	   *pix_map = (double *)0,
	   *hist  = (double *)0, s_hist=0.;
    int *nph1 = (int *)0, *nph2 = (int *)0;
    int nx1, nx2, np1, np2, n_tot=0, n_act=0;
    int *mask;
    char x_type[40], grid_type[40];
    FILE *fp_out;
    FILE *fp_acf = (FILE *)0;
    FILE *fp_adf = (FILE *)0;
    FILE *fp_buf = (FILE *)0;
    struct _zone_ {
	 double latR0, lonR0;
	 double latR1, lonR1;
	 double radius;
	 int type;               /* zone type */
    } *zz = (struct _zone_ *)0;
    extern void map_separation(double *, double *, double *,
	     int, int, int *, int *, double *);
    extern void map_a_separation(double *, double *, double *,
	     int, int, int *, int *, double *);
    extern void map_separ_alpha(double *, double *, double *,
	     int, int, int *, int *, double *, double);
    extern void map_separ_sq(double *, double *, double *,
	     int, int, int *, int *, double *);

    double *acf_map = (double *)0, *ncf_map = (double *)0;


    map1_file[0] = map2_file[0] = outfile[0] = '\0';
    file_limits[0] = div_file[0] = buffile[0] = '\0';
    acf_map_file[0] = ncf_map_file[0] = '\0';

    /* select grid output keyword values */
    strcpy(x_type, "COS(THETA)");
    strcpy(grid_type, "PHI(X)+LEGENDRE");

    /* loop over external parameters */
    setopt(argc, argv);

    if (glesp_comments < 0) {
	if (getenv("GLESP_COMMENTS") != NULL) {
	    char str[21];
	    strcpy(str,getenv("GLESP_COMMENTS"));
	    glesp_comments = *str == 'N' || *str == 'n' ? 0 : 1;
	    if (glesp_comments) {
		 if ( strcmp(str, "OFF") == 0 || strcmp(str, "off") == 0)
		     glesp_comments = 0;
	    }
	} else
	    glesp_comments = 1;
    }

    read_glesp_map(&f1, &x1, &nx1, &np1, &nph1, map1_file);
    read_glesp_map(&f2, &x2, &nx2, &np2, &nph2, map2_file);

    if (map_res_cmp(nx1,np1,nph1,nx2,np2, nph2)) {
	PR(ER,"%s: maps '%s' and '%s' have different resolution. ",
	       myname, map1_file, map2_file);
	PR(ER, "Sorry...\n");
	exit(1);
    }
    if (*acf_file != '\0') {
	if ((fp_acf=fopen(acf_file, "w")) == (FILE *)0) {
	    PR(ER, "%s: can not create file '%s'\n", myname, acf_file);
	    exit(2);
	}
    }
    /* get memory for buffer keeping bin numbers in pixels as a map */
    if (flag_buf) {
      if (flag_buf == 2) {
	double *xa= (double *)0;
	int nxa, npa, *npha = (int *)0;
	/* read map containing bin numbers */
	read_glesp_map(&pix_map, &xa, &nxa, &npa, &npha, buffile);
	if (map_res_cmp(nx1,np1,nph1,nxa,npa, npha)) {
	    PR(ER,
	      "%s: map '%s' and buffer map '%s' have different resolution. ",
		   myname, map1_file, buffile);
	    PR(ER, "Sorry...\n");
	    exit(1);
	}
	/* get number of interavsl of bins */
	n_hist = 0;
	for(i=0; i<nx1; i++) {
	    int ii = i * np1;
	    for(j=0; j<nph1[i]; j++) {
		int n = pix_map[ii+j] + .5;
		if (n_hist < n) n_hist = n;
	    }
	}
	free((void *)xa);
	free((void *)npha);
      } else
	/* buffer map to be filled with bin numbers to save to disk */
	if((pix_map =(double*)calloc((unsigned)(nx1*np1), sizeof *pix_map))==
	    (double*)0)
		goto nomem1;
	if ((fp_buf = fopen(buffile, "w")) == (FILE *)0) {
	    PR(ER, "%s: can not create file '%s'\n", myname, buffile);
	    exit(2);
	}
    }

    if (*div_file != '\0') {
	/* read map file to built a histohram of pixels */
	read_glesp_map(&div_map, &x2, &nx2, &np2, &nph2, div_file);
	if (map_res_cmp(nx1,np1,nph1,nx2,np2, nph2)) {
	    PR(ER,"%s: maps '%s' and '%s' have different resolution. ",
		   myname, map1_file, div_file);
	    PR(ER, "Sorry...\n");
	    exit(1);
	}
    } else  {
      /* get memory for map to built a histohram of pixels */
      if((div_map=(double*)calloc((unsigned)(nx1*np1),sizeof *div_map))==
	(double *)0)
	    goto nomem1;
      for(i=0; i<nx1; i++)  {
	  int ii = i * np1;
	  for(j=0; j<nph1[i]; j++) {
	      int jj = ii+j;
	      div_map[jj] = fabs(f2[jj])< 1.e-30 ? 1.e-30 : f1[jj]/f2[jj];
	  }
      }
    }

    /* if flag_buf == 2, then use prepared bin numbers in mapped data */
    /* else use divisions by bins                                     */
    if (flag_buf != 2) {
      /* division flag */
      if (divi_flag && set_1bin) {
	/* only bin, manually set */
	n_hist = 1;
	if((hist = (double *)calloc((unsigned)(n_hist), sizeof *hist)) ==
	    (double*)0)
		goto nomem1;
	if((lim = (struct _lu_ *)calloc((unsigned)(n_hist), sizeof *lim)) ==
	    (struct _lu_ *)0)
		goto nomem1;
	lim[0].l = bin1_min;
	lim[0].u = bin1_max;
      } else if (divi_flag) {
	double h_min = 1.e+30, h_max = -1.e+30;
	double lh_min = 1.e+30, lh_max = -1.e+30;
	int nt=0, k=0;
	double ls_hist=0.;
	int k_start=0, k_stop=0;
	double actual_max=0., actual_min=0.;
	double sig_aver = 0., sig_peak = -1.e+30;

	/* detect min and maximum of histogram */
	for(i=0; i<nx1; i++)  {
	    int ii = i * np1;
	    for(j=0; j<nph1[i]; j++) {
		int jj = ii+j;
		double c, dm = div_map[jj];
		if (h_min > dm)  h_min = dm;
		if (h_max < dm)  h_max = dm;
		if (log_bin) {
		    c = dm < 0.? log(-dm) : log(dm);
		    if (lh_min > c)  lh_min = c;
		    if (lh_max < c)  lh_max = c;
		}
		sig_aver += dm;
		if (sig_peak < dm) sig_peak = dm;
		nt++;
	    }
	}
	sig_aver /= nt;   /* average value of 'div_map' */

	/* check equal values for min=max */
	/* and make small interval in this case by hand */
	if (fabs(h_max) > 1.e-30 &&  (h_max - h_min)/h_max < 1.e-5) {
		h_max += 1.e-7 * h_max;
		h_min -= 1.e-7 * h_max;
		if (log_bin) {
		    lh_min -= 0.3;
		    lh_max += 0.3;
		}
	}

	/*------------------------------*/
	/* set new limits for histogram */
	/*------------------------------*/
	if (hist_up_lim > -1.e+30) {
	    actual_max = h_max; /* keep real value */
	    h_max = hist_up_lim;
	}
	if (hist_down_lim < 1.e+30) {
	    actual_min = h_min; /* keep real value */
	    h_min = hist_down_lim;
	}

	if (! n_hist && bin_size < 0.) {
	    /* calculate by default */
	    int nn = (int) (log((double)nt)/log (2.)) + 2;
	    n_hist = nn * 1000;
	    s_hist = (h_max - h_min) / n_hist;
	    if (log_bin)
		ls_hist = (lh_max - lh_min) / n_hist * 2;
	} else if (bin_size > 0.) {
	  /* when bin_size is set */
	  s_hist = bin_size;
	  n_hist = (h_max - h_min) / s_hist + .5;
	  if (log_bin)
	      ls_hist = (lh_max - lh_min) / n_hist * 2;
	  if (hist_down_lim < 1.e+30)  n_hist++; /* if limits are set */
	  if (hist_up_lim > -1.e+30)   n_hist++; /*                   */
	} else {
	  /* when number of pixels is set */
	  s_hist = (h_max - h_min) / n_hist;
	  if (log_bin)
	      ls_hist = (lh_max - lh_min) / n_hist * 2;
	  if (hist_down_lim <  1.e+30)  n_hist++; /* if limits are set */
	  if (hist_up_lim   > -1.e+30)  n_hist++; /*                   */
	}

	if((hist = (double *)calloc((unsigned)(n_hist), sizeof *hist)) ==
	    (double*)0)
		goto nomem1;
	if((lim = (struct _lu_ *)calloc((unsigned)(n_hist), sizeof *lim)) ==
	    (struct _lu_ *)0)
		goto nomem1;

	/* fill in histogram table */
	if (flag_hist_peak) {
	    if (n_hist) {
		double rp = sig_aver;
		if (n_hist == 2) {
		    /* exmaple of 2 intervals */
		    lim[0].l = h_min;
		    lim[0].u = rp;
		    lim[1].l = lim[0].u;
		    lim[1].u = h_max;
		} else if (n_hist == 3) {
		    /* exmaple of 3 intervals */
		    lim[0].l = h_min;
		    lim[0].u = rp - s_hist/2;
		    lim[1].l = lim[0].u;
		    lim[1].u = lim[1].l + s_hist;
		    lim[2].l = lim[1].u;
		    lim[2].u = h_max;
		} else {
		    /* all other intervals */
		    int nk0, nk1, k;

		    if (n_hist%2 == 0) { /* even */
			nk0 = (rp - h_min) / s_hist;
			nk1 = (h_max - rp) / s_hist + 0.5;
		    } else { /* odd */
			nk0 = (rp-s_hist/2 - h_min) / s_hist;
			nk1 = (h_max - (rp-s_hist/2)) / s_hist + 0.5;
		    }

		    if (nk1-nk0 > n_hist) {
		      PR(ER,"%s: wrong memory volume calculation: ",myname);
		      PR(ER,"array for %d elements (real: %d)\n",
			    n_hist, nk1-nk0);
		      exit(1);
		    }
		    for(k=nk0; k>=0; k--) {
			lim[k].l = rp - s_hist*k;
			lim[k].u = lim[k].l + s_hist;
			if (lim[k].l < h_min) lim[k].l = h_min;
		    }
		    for(k=nk0+1; k<nk1; k++) {
			lim[k].l = rp + s_hist*k;
			lim[k].u = lim[k].l + s_hist;
			if (lim[k].u > h_max) lim[k].u = h_max;
		    }
		}
	    }
	} else {
	    k_start = 0; k_stop = n_hist-1;
	    if (hist_down_lim < 1.e+30)  k_start = 1;
	    if (hist_up_lim > 1.e+30)    k_stop = n_hist-1;
	    for(i=0; i<nx1; i++)  {
		int ii = i * np1;
		for(j=0; j<nph1[i]; j++) {
		  int jj = ii+j;
		  if (log_bin) {
		    double c=div_map[jj]<0?log(-div_map[jj]):log(div_map[jj]);
		    if (div_map[jj] < 0.0) {
			k = (c - lh_min)  / ls_hist + .5;
			k = n_hist/2 - k;
		    } else {
			k = (c - lh_min)  / ls_hist + .5;
			k = n_hist/2 + k;
		    }
		  } else {
		      if ( div_map[jj] < h_min )  k = 0;
		      else if ( div_map[jj] > h_max )  k = k_stop;
		      else
			  k = k_start + (div_map[jj]-h_min) / s_hist + .5;
		  }
		  if (k < 0)        k = 0;
		  if (k >= n_hist)  k = n_hist-1;
		  hist[k] += 1.;  /* number of pixels in k-th bin */
		}
	    }

	    /* check histogram bins */
	    /* integrating in bins to do min_bin_n > N_BIN_MIN */
	    for (k=0; k<n_hist; k++) {
		if (log_bin) {
		    if (k < n_hist/2 )  {
			int kk = n_hist/2 - k - 1;
			lim[k].l = -exp(lh_min + (kk+1) * ls_hist);
			lim[k].u = -exp(lh_min + kk *ls_hist);
		    } else if (k == n_hist/2 )  {
			int kk = k - n_hist/2;
			lim[k].l = -exp(lh_min + kk * ls_hist);
			lim[k].u =  exp(lh_min + kk * ls_hist);
		    } else {
			int kk = k - n_hist/2 - 1;
			lim[k].l = exp(lh_min + kk * ls_hist);
			lim[k].u = exp(lh_min + (kk+1)*ls_hist);
		    }
		} else {
		    if (k == 0 && hist_down_lim < 1.e+30)  {
			lim[k].l = actual_min;
			lim[k].u = h_min;
		    } else if (k == n_hist-1 && hist_up_lim > -1.e+30)  {
			lim[k].l = lim[n_hist-2].u;
			lim[k].u = actual_max;
		    } else {
			lim[k].l = h_min + (k-k_start) * s_hist;
			lim[k].u = lim[k].l + s_hist;
		    }
//PR(ER, "limits: %g < v < %g, %g pixels\n", lim[k].l, lim[k].u, hist[k]);
		}
	    }
	}
	k = n_hist - 1;

	/* squeeze number of active bins */
	if ( !no_squeeze ) {
	    for (i=0; i<k; i++) {
		if ( hist[i] < n_bin_min ) {
//PR(ER, "squeezed: %g < v < %g, %g pixels,", lim[i].l, lim[i].u, hist[i]);
		     hist[i] += hist[i+1];
		     lim[i].u = lim[i+1].u;
		     for (j=i+1; j<k; j++) {
			hist[j] = hist[j+1];
			lim[j].l = lim[j+1].l;
			lim[j].u = lim[j+1].u;
		     }
//PR(ER, "res: %g < v < %g, %g pixels\n", lim[i].l, lim[i].u, hist[i]);
		     i--; k--; n_hist--;
		}
	    }
	}
	/* check last bin */
	if ( hist[k] < n_bin_min ) {
	    hist[k-1] += hist[k];
	    lim[k-1].u = lim[k].u;
	    n_hist--;
	}
      } else if (file_limits[0] != '\0') {
	FILE *fp;
	char s1[30], s2[30], str[161];
	if ((fp = fopen(file_limits, "r")) == (FILE *)0) {
	    PR(ER, "%s: problem to open file '%s'\n", myname, file_limits);
	    exit(1);
	}
	n_hist = 0;
	while (fgets(str, 160, fp) != NULL) {
	    if ( *str == '\n' || *str == '\r' || *str == '#')
		continue;
	    n_hist++;
	}
	if((hist = (double *)calloc((unsigned)(n_hist), sizeof *hist)) ==
	    (double*)0)
		goto nomem1;
	if((lim = (struct _lu_ *)calloc((unsigned)(n_hist), sizeof *lim)) ==
	    (struct _lu_ *)0)
		goto nomem1;
	fseek(fp, 0l, 0);
	i = 0;
	while (fgets(str, 160, fp) != NULL) {
	    if ( *str == '\n' || *str == '\r' || *str == '#')
		continue;
	    sscanf(str, "%s%s", s1, s2);

	    if (strncmp(s1, "-inf", 4) == 0)
		    lim[i].l = -1.e+30;
	    else if (strncmp(s1, "inf", 3) == 0)
		    lim[i].l = 1.e+30;
	    else
		    lim[i].l = atof(s1);

	    if (strncmp(s2, "-inf", 4) == 0)
		    lim[i].u = -1.e+30;
	    else if (strncmp(s2, "inf", 3) == 0)
		    lim[i].u = 1.e+30;
	    else
		    lim[i].u = atof(s2);
	    i++;
	}
	fclose(fp);
      } else {
	n_hist = 1;
	if((lim = (struct _lu_ *)calloc((unsigned)(n_hist), sizeof *lim)) ==
	    (struct _lu_ *)0)
		goto nomem1;
	if((hist = (double *)calloc((unsigned)(n_hist), sizeof *hist)) ==
	    (double*)0)
		goto nomem1;
	lim[0].l = low_lim;
	lim[0].u = up_lim;
	hist[0] = 0;
	if (fabs(alpha_cf+2.7*M_PI) > 1.e-8) {
	    /* common alpha for map */
	    lim[0].l = -1.e+30;
	    lim[0].u =  1.e+30;
	}
      }

      /* filling in pixel bin map */
      /*--------------------------*/
      if (flag_buf == 1) {
	for(i=0; i<nx1; i++)  {
	    int ii = i * np1;
	    for(j=0; j<nph1[i]; j++) {
		int jj = ii + j;
		double dm = div_map[jj];
		register int k;
		for(k=0; k<n_hist; k++) {
		   if (lim[k].l <=  dm && dm < lim[k].u)
			pix_map[jj] = k;
		}
	    }
	}
      }

      if (print_flag) {
	  for (n = 0; n<n_hist; n++) {
	      printf("%15.10g %15.10g %8g\n", lim[n].l, lim[n].u, hist[n]);
	  }
      }
    }

    if ((mask = (int*)calloc((unsigned)(nx1*np1), sizeof *mask)) == (int*)0){
nomem1:
	PR(ER, "%s: no memory...\n", myname);
	exit(1);
    }
    if ((cmap=(double*)calloc((unsigned)(nx1*np1),sizeof *cmap))==(double*)0)
	goto nomem1;
    if ((t=(double*)calloc((unsigned)(nx1*np1),sizeof *t))==(double*)0)
	goto nomem1;

    /*-------------------------*/
    /* read region for masking */
    /*-------------------------*/
    if ( z_mask_flag ) {
	register int is=0;
	int ns = 1;

	extern void map_zone(struct _zone_ *, int, int,
		double *, double *, int, int, int *, double);
	int function_type= z_remain_flag ? 0:1;/*1:remove_flag,0:remain*/
	int zone_type = zone_circle ? 0 : 1;
//        double lim1 = zmask_value * (1. - 1.e-6),
//               lim2 = zmask_value * (1. + 1.e-6);

	if((zmask=(double*)calloc((unsigned)(nx1*np1),sizeof *zmask))
	  ==(double*)0)
	    goto nomem1;

	/* if to read zone-file */
	if (zone_file[0] != '\0') {
	     FILE *fz;
	     char str[81];
	     /* read data from file with zones */
	     if((fz = fopen(zone_file, "r")) == (FILE *)0) {
		 PR(ER, "%s: I can not open file '%s'...\n",
		     myname, zone_file);
		 exit(1);
	     }
	     /* get number of zones */
	     is =0;
	     while(fgets(str, 80, fz) != NULL) {
		 if (*str == '#' || *str == '\n')
		     continue;
		 is++;
	     }
	     /* allocate memory */
	     i = ns = is;
	     zz = (struct _zone_ *)calloc(i, sizeof *zz);
	     if (zz == (struct _zone_ *)0) goto nomem1;

	     /* read zones  */
	     is=0;
	     fseek(fz, 0l, 0);
	     while(fgets(str, 80, fz) != NULL) {
		 char s1[30], s2[30], s3[30], s4[30];
		 char *cs = str;
		 int n_par;  /* number of parameters per string */
		 if (*str == '#' || *str == '\n')
			 continue;
		 while(*cs)
			 if (*cs == '#') {*cs = '\0'; break;} else cs++;

		 n_par = n_skip_wd(str);
		 switch (n_par) {
		 case 1:
		     PR(ER,
		       "%s: there is only 1 parameter in the string:\n",
			     myname);
		     PR(ER,"%s", str);
		     PR(ER, "\n%s: I've skipped it...\n", myname);
		     break;
		 case 2:
		     zz[is].latR0  = deg2rad((skip_wd(0,str,s1),s1));
		     zz[is].lonR0  = deg2rad((skip_wd(1,str,s2),s2));
		     zz[is].radius = zone_radius;
		     zz[is].type = 0;
		     break;
		 case 3:
		     zz[is].latR0  = deg2rad((skip_wd(0,str,s1),s1));
		     zz[is].lonR0  = deg2rad((skip_wd(1,str,s2),s2));
		     if (zone_radius > 0.)
			 zz[is].radius = zone_radius;
		     else
			 zz[is].radius = deg2rad((skip_wd(2,str,s3),s3));
		     zz[is].type = 0;
		     break;
		 case 4:
		     zz[is].latR0 = deg2rad((skip_wd(0,str,s1),s1));
		     zz[is].lonR0 = deg2rad((skip_wd(1,str,s2),s2));
		     zz[is].latR1 = deg2rad((skip_wd(2,str,s3),s3));
		     zz[is].lonR1 = deg2rad((skip_wd(3,str,s4),s4));
		     if (zone_type == 0) {
			     zz[is].radius = zz[is].latR1;
			     zz[is].type = 0;
		     } else  zz[is].type = 1;
		     break;
		 }
		 is++;
	     }
	     fclose(fz);
	} else {
	    /* allocate memory for 1 zone */
	    /* from command string */
	    i = ns = 1;
	    zz = (struct _zone_ *)calloc(i, sizeof *zz);
	    if (zz == (struct _zone_ *)0) goto nomem1;
	    zz[0].latR0 = zone_latR0;  zz[0].lonR0 = zone_lonR0;
	    zz[0].latR1 = zone_latR1;  zz[0].lonR1 = zone_lonR1;
	    zz[0].radius = zone_radius;
	    if (zone_latR1 < -900.) {
		    zz[0].type = 0;
	    } if (zone_lonR1 < -900.) {
		    if (zone_radius > 0.)
			zz[is].radius = zone_radius;
		    else
			zz[is].radius = zone_latR1;
		    zz[0].type = 0;  /* circle */
	    } else {
		    zz[0].type = 1;  /* rectangular */
	    }
	}
//      if (flag_pix == 'h') {
//         extern void hmap_zone(struct _zone_ *, int, int, double *,
//              double *, double *, int, int, int *, double);
//         hmap_zone(zz,ns,function_type,zmask,x,offs,nx,np,nph,
//                  zmask_value);
//      } else
	   map_zone(zz,ns,function_type, zmask, x1,nx1,np1,nph1,
		    zmask_value);
    }

    /* find memory for alpha-coeff. map */
    if (*acf_map_file != '\0') {
	if((acf_map = (double *)calloc((unsigned)(nx1*np1),
	    sizeof *acf_map)) == (double *)0)
		   goto nomem1;
    }
    if (*ncf_map_file != '\0') {
	if((ncf_map = (double *)calloc((unsigned)(nx1*np1),
	    sizeof *ncf_map)) == (double *)0)
		   goto nomem1;
    }

    /*******************************/
    /* loop over histograms levels */
    /*******************************/
//PR(ER, "n_hist=%d\n", n_hist);
    for (n = 0; n<n_hist; n++) {
	/* set mask */
	for(i=0; i<nx1; i++)  {
	    int ii = i * np1;
	    for(j=0; j<nph1[i]; j++) {
		int jj = ii+j;

		if (!n) n_tot++;  /* to count all pixels one time */

		if (flag_buf == 2) {
		    if (n != (int)(pix_map[jj]+.5))     mask[jj] = 1;
		} else {
		    double c = div_map[jj];
		    if (c < lim[n].l || lim[n].u <= c)  mask[jj] = 1;
		}
	    }
	}

	/* use masked pixels to fill in structure */


//if(glesp_comments)
//PR(ER, "n=%d, l=%g, u=%g\n", n, lim[n].l, lim[n].u);
	/*-------------------------*/
	/* read region for masking */
	/*-------------------------*/
	if ( z_mask_flag ) {
	    /* set mask in map */
	    for(i=0; i<nx1; i++) {
		 int ii = i * np1;
		 for(j=0; j<nph1[i]; j++) {
		     int nn = ii+j;
		     if (zmask[nn] > 0.5) {
			 mask[nn] = 1;
		     } else
			 if (!n) n_act++;
		 }
	    }
	} /* end of processing a region for masking */


	/* map separation for given mask */
	/*-------------------------------*/
	if (! no_calc) {
	    if (fabs(alpha_cf+2.7*M_PI) > 1.e-8) {
		map_separ_alpha(f1,f2, x1, nx1, np1, nph1, mask, t,alpha_cf);
	    } else {
		if (bin_zero)   /* -e */
		    map_a_separation(f1, f2, x1, nx1, np1, nph1, mask, t);
		else
		    map_separation(f1, f2, x1, nx1, np1, nph1, mask, t);
	    }
	}
if(set_1bin == 1){
int i,j;
double sc1=0.0,sc2=0.0,sf=0.0;
double ac1=0.0,ac2=0.0,af=0.0;
int n=0, nt=0;
for(i=0; i<nx1; i++) {
  int ii = i * np1;
  for(j=0; j<nph1[i]; j++) {
    int jj = ii + j;
    nt++;
    if (mask[jj])  continue;
    ac1 += f1[jj];
    ac2 += f2[jj];
    af  += t[jj];
    n++;
  }
}
ac1 /= n;
ac2 /= n;
af  /= n;
for(i=0; i<nx1; i++) {
  int ii = i * np1;
  for(j=0; j<nph1[i]; j++) {
    int jj = ii + j;
    if (mask[jj])  continue;
    sc1 += (f1[jj] - ac1) * (f1[jj] - ac1);
    sc2 += (f2[jj] - ac2) * (f2[jj] - ac2);
    sf  += (t[jj]   - af) * (t[jj]   - af) ;
  }
}
sc1 = sqrt (sc1 / n);
sc2 = sqrt (sc2 / n);
sf =  sqrt (sf  / n);
PR(ER, "nt=%d, n=%d, ac1=%g, ac2=%g, af=%g, sc1=%g, sc2=%g, sf=%g\n",
nt, n, ac1, ac2, af, sc1, sc2, sf);
}

	/* output separation coefficients from bins */
	if (fp_acf != (FILE *)0) {
	    extern void calc_alpha(double *, double *, double *, int, int,
		    int *, int *, int *, double *),
		calc_a_alpha(double *, double *, double *, int, int,
		    int *, int *, int *, double *);
	    int nt; double alpha;
	    if (bin_zero)   /* -e */
		calc_a_alpha(f1, f2, x1, nx1, np1, nph1, mask, &nt, &alpha);
	    else
		calc_alpha(f1, f2, x1, nx1, np1, nph1, mask, &nt, &alpha);

	    PR(fp_acf, "%12.5e %3d %12.5e %12.5e\n",
		alpha, nt, lim[n].l, lim[n].u);

	    /* output map of separation coefficients from bins */
	    if ( acf_map != (double *)0 || ncf_map != (double *)0) {
		for(i=0; i<nx1; i++)  {
		    int ii = i * np1;
		    for(j=0; j<nph1[i]; j++) {
			int ij = ii+j;
			if (!mask[ij]) {
			    if (acf_map != (double *)0) acf_map[ij] = alpha;
			    if (ncf_map != (double *)0) ncf_map[ij] = n;
			}
		    }
		}
	    }
	}
	/* output separation coefficients from bins */
	if (fp_adf != (FILE *)0) {
	    extern void
		calcd_a_alpha(double *, double *, double *, int, int,
		    int *, int *, int *, double *);
	    int nt; double alpha;
	    calcd_a_alpha(f1, f2, x1, nx1, np1, nph1, mask, &nt, &alpha);

	    PR(fp_adf, "%12.5e %3d %12.5e %12.5e\n",
		alpha, nt, lim[n].l, lim[n].u);
	}

	/* summing current layout with output map */
	/*----------------------------------------*/
	for(i=0; i<nx1; i++)  {
	    int ii = i * np1;
	    for(j=0; j<nph1[i]; j++) cmap[ii+j] += t[ii+j];
	}

	/* make mask epmpty */
	/*------------------*/
	if (n < n_hist-1 ) {
	    for(i=0; i<nx1; i++)  {
		int ii = i * np1;
		for(j=0; j<nph1[i]; j++) mask[ii+j] = 0;
	    }
	}
    }

    /*--------------*/
    /* output block */
    /*--------------*/
    /* output map of bin separation coefficients to file */
    if (*acf_map_file != '\0') {
	FILE *fcf;
	if ((fcf = fopen(acf_map_file, "w")) == (FILE *)0) {
	    PR(ER, "%s: I can not create file '%s'\n", myname, acf_map_file);
	    exit(1);
	}
	/*  write map  */
	outp_glesp_map(acf_map, x1, nx1, np1, nph1, fcf, x_type, grid_type);
	/* close file and free memory */
	if (fcf != stdout) fclose(fcf);
	free((void *)acf_map);
    }
    if (*ncf_map_file != '\0') {
	FILE *fcf;
	if ((fcf = fopen(ncf_map_file, "w")) == (FILE *)0) {
	    PR(ER, "%s: I can not create file '%s'\n", myname, ncf_map_file);
	    exit(1);
	}
	/*  write map  */
	outp_glesp_map(ncf_map, x1, nx1, np1, nph1, fcf, x_type, grid_type);
	/* close file and free memory */
	if (fcf != stdout) fclose(fcf);
	free((void *)ncf_map);
    }
    /* output map of bin numbers to file */
    if (flag_buf == 1) {
	/*  write map  */
	outp_glesp_map(pix_map, x1, nx1,np1,nph1, fp_buf, x_type, grid_type);
	/* close file and free memory */
	if (fp_buf != stdout) fclose(fp_buf);
	free((void *)acf_map);
    }

    if ( z_mask_flag ) {
	/* free memory from zone-array */
	free((void *)zz);

	free((void *)zmask);
	zmask = (double *)0;
    }

    if (glesp_comments) {
      if (n_act)
	PR (ER, "number of pixels: n_actual=%d, n_total=%d\n", n_act, n_tot);
      else
	PR (ER, "number of pixels: n_total=%d\n", n_tot);
    }

    if (!no_calc) {
	if (strcmp(outfile,"-") == 0  || strcmp(outfile,"stdout") == 0) {
	    fp_out = stdout;
	} else {
	    /* open output file */
	    if((fp_out = fopen(outfile, "w")) == (FILE *)0) {
		PR(ER, "%s: can not create file '%s'\n",
		    myname, outfile);
		exit(1);
	    }
	}

	/*  write map  */
	outp_glesp_map(cmap, x1, nx1, np1, nph1, fp_out, x_type, grid_type);

	if (fp_out != stdout) fclose(fp_out);
    }

    /* free memory */
    if (f1 != (double *)0)      free((void *)f1);
    if (t != (double *)0)       free((void *)t);
    if (cmap != (double *)0)    free((void *)cmap);
    if (div_map != (double *)0) free((void *)div_map);

    exit(0);
}

void setopt(argc, argv)
int argc;
char **argv;
{
    register int i;
    myname = argv[0];

    for(i=1; i<argc; i++) {
	switch(argv[i][0]) {
	case '-':
	    switch(argv[i][1]) {
	    case 'a':
		if (strncmp(argv[i]+1, "af", 2) == 0) {
		    strcpy(acf_file, argv[++i]);
		} else if (strncmp(argv[i]+1, "ad", 2) == 0) {
		    /* for some various output */
		    strcpy(adf_file, argv[++i]);
		} else if (strcmp(argv[i]+1, "a") == 0) {
		    alpha_cf = atof(argv[++i]);
		}
		break;
	    case 'b':
		i++;
		if (argv[i][0] != ',') {
		    if (strncmp(argv[i], "none", 4) != 0) {
			 bin1_min = atof(argv[i]);
		    }
		}
		{
		    char *ps = strchr(argv[i], ',');
		    if (ps != NULL && *ps && *(ps+1)) {
			 if (strncmp(ps+1, "none", 4) != 0) {
			      bin1_max = atof(ps+1);
			 }
		    }
		}

		set_1bin = 1;
		break;
	    case 'c':
	    case 'C':
		i++;
		if (argv[i][0] != ',') {
		    if (strncmp(argv[i], "none", 4) != 0) {
			 low_lim = atof(argv[i]);
		    }
		}
		{
		    char *ps = strchr(argv[i], ',');
		    if (ps != NULL && *ps && *(ps+1)) {
			 if (strncmp(ps+1, "none", 4) != 0) {
			      up_lim = atof(ps+1);
			 }
		    }
		}
		break;
	    case 'd': /* division using historgram */
		if (strncmp(argv[i]+1, "df", 2) == 0) {
		    strcpy(div_file, argv[++i]);
		}
		divi_flag=1;
		break;
	    case 'f':
		switch(argv[i][2]) {
		case 'a':
		    strcpy(acf_map_file,argv[++i]);
		    break;
		case 'n':
		    strcpy(ncf_map_file,argv[++i]);
		    break;
		default :
		    strcpy(file_limits, argv[++i]);
		    break;
		}
		break;
	    case 'h':
		if (strncmp(argv[i]+1, "h0", 2) == 0)
		    hist_down_lim = atof(argv[++i]);
		else if (strncmp(argv[i]+1, "he", 2) == 0)
		    hist_up_lim = atof(argv[++i]);
		else
		    usage();
		break;
	    case 'i':
		if ( strncmp(argv[i]+1, "ib", 2) == 0) {
		    /* read buffer of used pixels */
		    flag_buf = 2;
		    strcpy(buffile, argv[++i]);
		} else {
		    PR(ER, "%s: illegal flag '%s'\n", myname, argv[i]);
		    usage();
		}
		break;
	    case 'k':
		if (strncmp("ke", argv[i]+1, 2) == 0) {
		    char s1[30], s2[30], s3[30], s4[30];
		    get_4_str(argv[++i], s1, s2, s3, s4);
		    if(*s1 !='\0')
		      zone_latR0=deg2rad(s1);else zone_latR0= -1001.;
		    if(*s2 !='\0')
		      zone_lonR0=deg2rad(s2);else zone_lonR0= -1001.;
		    if(*s3 !='\0')
		      zone_latR1=deg2rad(s3);else zone_latR1= -1001.;
		    if(*s4 !='\0')
		      zone_lonR1=deg2rad(s4);else zone_lonR1= -1001.;
		    z_remain_flag = 1;
		    z_remove_flag = 0;
		    z_mask_flag = 1;
		} else {
		    PR(ER, "%s: illegal flag '%s'\n", myname, argv[i]);
		    usage();
		}
		break;
	    case 'K':  /* keep data in selected zones */
		z_remain_flag = 1;
		z_remove_flag = 0;
		break;
	    case 'l':
		log_bin = 1;
		break;
	    case 'n':
		if (strncmp(argv[i]+1, "nb",2) == 0) {
		    /* minimal pixels number per bin */
		    n_bin_min = atoi(argv[++i]);
		} else if (strncmp(argv[i]+1, "noc",3) == 0) {
		    no_calc=1;
		} else if (strncmp(argv[i]+1, "nsq",3) == 0) {
		    no_squeeze=1;
		} else if (strncmp(argv[i]+1, "nosq",4) == 0) {
		    no_squeeze=1;
		} else if (strncmp(argv[i]+1, "ns",2) == 0) {
		    /* bin size */
		    bin_size = atof(argv[++i]);
		    divi_flag=1;
		} else {
		    /* set number of bins for '-div' */
		    n_hist = atoi(argv[++i]);
		    divi_flag=1;
		}
		break;
	    case 'o':
		if ( strncmp(argv[i]+1, "ob", 2) == 0) {
		    /* write buffer of used pixels */
		    flag_buf = 1;
		    strcpy(buffile, argv[++i]);
		} else
		    strcpy(outfile, argv[++i]);
		break;
	    case 'p':
		/* print flag */
		print_flag = 1;
		break;
	    case 'r':
		if (strcmp("rm", argv[i]+1) == 0 ||
		    strncmp("re", argv[i]+1, 2) == 0) {
		     char s1[30], s2[30], s3[30], s4[30];
		     get_4_str(argv[++i], s1, s2, s3, s4);
		     if(*s1 !='\0')
		       zone_latR0=deg2rad(s1);else zone_latR0= -1001.;
		     if(*s2 !='\0')
		       zone_lonR0=deg2rad(s2);else zone_lonR0= -1001.;
		     if(*s3 !='\0')
		       zone_latR1=deg2rad(s3);else zone_latR1= -1001.;
		     if(*s4 !='\0')
		       zone_lonR1=deg2rad(s4);else zone_lonR1= -1001.;
		     z_remove_flag = 1;
		     z_remain_flag = 0;
		     z_mask_flag = 1;
		} else if (strncmp ("ra", argv[i]+1, 2) == 0) {
		     zone_radius = deg2rad(argv[++i]);
		     zone_circle = 1;
		} else {
			PR(ER, "%s: illegal flag '%s'\n",
			    myname, argv[i]);
			usage();
		}
		break;
	    case 'R':  /* keep data outside selected zones */
		z_remain_flag = 0;
		z_remove_flag = 1;
		break;
	    case 'v': /* version */
//              PR(ER, "%s: GLESP 2.0, version 0.2 (25.11.2009)\n",
		PR(ER, "%s: GLESP 2.0, version 0.3 (06.06.2010)\n",
		    myname);
		exit(0);
	    case 'z':
		switch (argv[i][2]) {
		case 'c':
		    {  char s1[30], s2[30];
		      /* center of a circle to be masked */
		      get_2_str(argv[++i], s1, s2);
		      if(*s1 !='\0')
			zone_latR0=deg2rad(s1); else zone_latR0= -1001.;
		      if(*s2 !='\0')
			zone_lonR0=deg2rad(s2); else zone_lonR0= -1001.;
		      zone_circle = 1;
		    }
		    break;
		case 'e':
		    bin_zero = 1;
		    break;
		case 'r':
		    /* radius of a circle to be masked */
		    zone_radius = deg2rad(argv[++i]);
		    zone_circle = 1;
		    break;
		case 'o':
		    /* select a zone to be masked */
		    { char s1[30], s2[30], s3[30], s4[30];
			get_4_str(argv[++i], s1, s2, s3, s4);
			if(*s1 !='\0') zone_latR0=deg2rad(s1);
			    else zone_latR0= -1001.;
			if(*s2 !='\0') zone_lonR0=deg2rad(s2);
			    else zone_lonR0= -1001.;
			if(*s3 !='\0') zone_latR1=deg2rad(s3);
			    else zone_latR1= -1001.;
			if(*s4 !='\0') zone_lonR1=deg2rad(s4);
			    else zone_lonR1= -1001.;
		    }
		    break;
		default:
		    /* file with list of zones to be processed */
		    strcpy(zone_file, argv[++i]);
		    break;
		}
		z_mask_flag = 1;
		break;
	    default:
		    PR(ER,"%s: illegal flag '%s'\n", myname, argv[i]);
		    usage();
	    }
	    break;
	case '+':
		switch(argv[i][1]) {
		case 'h':
		default:
			usage();
			break;
		}
		break;
	default:
		if (*map1_file == '\0')      strcpy(map1_file, argv[i]);
		else if (*map2_file == '\0') strcpy(map2_file, argv[i]);
		break;
	}
    }

    return;
}

void usage()
{
PR(ER, "usage: %s file1 file2 -o output_file [-C ll,up] [-div(divison)]\n",
myname);
PR(ER, "[-n n_hist(set n_of bins)] [-nb min_pix_bin] [-ns bin_size] [-a alp]\n");
PR(ER, "[-bh lim1,lim2_from histogram] [-ib/ob buffile_with_pixel_numbers]\n");
PR(ER, "[-fl limits_from_file] [-df div_file] [-log(set log scales)]\n");
PR(ER, "[-h0 start_level_for hist] [-he stop_level_for hist]\n");
PR(ER, "[-af a_coeff_file] [-nosqueeze] [-nocalc] [-ze] [-fa alp_cf_map]\n");
exit(2);
}
