#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include "glesp.h"
#define PR  fprintf
#define ER  stderr

/*
 *  Oleg Verkhodanov (SAO),  2005-2012
 *      masking with Hpix and GLESP data
 *
 */


/* masking with Hpix-file */
void mask_hp(f,x,nx,np,nph,hp_mask_file,dummy_value)
double *f, *x;
int nx, np, *nph;
char *hp_mask_file;
double dummy_value;
{
    register int i, j, k;
    double *map_ring = (double *)0;
    int nside=0, npixtot=0, nx_hp, np_hp;
    double *x_hp = (double *)0, *offs = (double *)0;
    double *map_hp = (double *)0;
    int *nph_hp = (int *)0;
    FILE *fp = (FILE *)0;
    /* mask (remove) inside zone: 1; mask outsize zone (keep): 0 */
    int flag_mask_type = nx > 0 ? 1 : 0;
    if (nx < 0) nx = -nx;

    if ((fp = fopen(hp_mask_file, "r")) == (FILE *)0) {
	PR(ER,"%s: can not open file '%s' with mask.\n",myname,hp_mask_file);
	exit(1);
    }
    /* read Hpix-map with mask */
    fhp_in(&map_ring, &nside, &npixtot, fp);
    if (fp != (FILE *)0) fclose(fp);

    np_hp = nside * 4;
    nx_hp = np_hp - 1;
    if((offs=(double*)calloc((unsigned)nx_hp, sizeof *offs))==(double*)0) {
nomem_e:
	PR(ER, "%s: no memory for mask array...\n", myname);
	exit(1);
    }
    if((x_hp=(double*)calloc((unsigned)nx_hp, sizeof *x_hp))==(double*)0)
	goto nomem_e;
    if((nph_hp=(int*)calloc((unsigned)nx_hp, sizeof *nph_hp))==(int*)0)
	goto nomem_e;
    if((map_hp=(double *)calloc((unsigned)(nx_hp*np_hp), sizeof *map_hp))==
	(double*)0)
	    goto nomem_e;

    /* calculate arrays: x_hp & nph_hp (again), offs (new) */
    hp_arrays(x_hp, offs, nph_hp, nside);

    /* copy map_ring to square matrix */
    for(k=i=0; i<nx_hp; i++)  {
	int ii=i*np_hp;
	for(j=0; j<nph_hp[i]; j++)  {
	    map_hp[ii+j] = map_ring[k];
	    k++;
	}
    }
    free ((void *)map_ring);

    /* read GLESP map pixel and copy to them data of mask from Hpix-map */
    for(i=0; i<nx; i++) {
	int ii = i * np;
	for(j=0; j<nph[i]; j++) {
	    int v, n;
	    double theta, phi;

	    /* read angle from GLESP map */
	    pixel2angle(x, nx, nph, i, j, &theta, &phi);

	    /* find number of pixel in Hpix-map */
	    n = hang2n(theta, phi, x_hp, nx_hp, np_hp, nph_hp, offs);

	    /* take int value of pixel */
	    v = map_hp[n] + 0.3;

	    if (flag_mask_type) {
		/* remove inside zone */
		if ( !v ) { /* if zero then mask */
		    f[ii+j] = dummy_value;
		}
	    } else {
		/* remove outside zone, keep inside zone */
		if ( v ) { /* if non-zero then mask */
		    f[ii+j] = dummy_value;
		}
	    }
	}
    }
    free ((void *)offs);
    free ((void *)x_hp);
    free ((void *)nph_hp);
    free ((void *)map_hp);
    return;
}

/* mask from GLESP-file */
void mask_gl(f,x,nx,np,nph,gl_mask_file,dummy_value)
double *f, *x;
int nx, np, *nph;
char *gl_mask_file;
double dummy_value;
{
    register int i, j;
    int nx_g, np_g;
    double *x_g = (double *)0;
    double *map_g = (double *)0;
    int *nph_g = (int *)0;
    FILE *fp = (FILE *)0;
    /* mask (remove) inside zone: 1; mask outsize zone (keep): 0 */
    int flag_mask_type = nx > 0 ? 1 : 0;
    if (nx < 0) nx = -nx;

    if ((fp = fopen(gl_mask_file, "r")) == (FILE *)0) {
	PR(ER,"%s: can not open file '%s' with mask.\n",myname,gl_mask_file);
	exit(1);
    }
    /* read GLESP-map with mask */
    read_glesp_tmap(&map_g, &x_g, &nx_g, &np_g, &nph_g, gl_mask_file);

    /* read GLESP map pixel and copy to them data of mask from Hpix-map */
    for(i=0; i<nx; i++) {
	int ii = i * np;
	for(j=0; j<nph[i]; j++) {
	    int v, n;
	    double theta, phi;

	    /* read angle from GLESP map */
	    pixel2angle(x, nx, nph, i, j, &theta, &phi);

	    /* find number of pixel in Hpix-map */
	    n = ang2n(theta, phi, x_g, nx_g, np_g, nph_g);

	    /* take int value of pixel */
	    v = map_g[n] + 0.3;

	    if (flag_mask_type) {
		if ( !v ) { /* if zero then mask */
		    f[ii+j] = dummy_value;
		}
	    } else {
		if ( v ) { /* if non-zero then mask */
		    f[ii+j] = dummy_value;
		}
	    }
	}
    }
    free ((void *)x_g);
    free ((void *)nph_g);
    free ((void *)map_g);
    return;
}


