#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "glesp.h"

#define PR      fprintf
#define ER      stderr

extern void setopt(int, char **), usage(void);
extern double w_exp(double, double), w_eq(double, double);
extern void   gauleg(double, double, double *, double *, int);
extern void   read_obs(double *, int, int, double *, int *, char *);

double pixel_size = -1.0;
int lmax = -1;
char obs_file[161];
char outp_file[161];
double (*func_weight)();          /* pointer to weighting function */

int repix_map = 1;
int theta_grid=1;     /* flag to restore grid for theta from cos()  */
int bit_int=0, log_int=0;
int bitpix_map = -32; /* Float Point Pixels for map                 */
int lx_add=4;
extern HEAD hd, *ph;
char tmpcat[161];
int test_flag = 0;

char *myname="o2t";

int main(argc, argv)
int argc;
char **argv;
{
    FILE *fp_out;
    double *f=(double *)0, *x=(double *)0;
    int *nph = (int *)0;
    int nx=0, np=0, ntot=0;
    double x1 = 1.0, x2 = -1.0;
    double s0=1.0, pi = M_PI;
    char grid_type[30], x_type[30];
    char str[50];

    outp_file[0] = obs_file[0] = '\0';
    strcpy(tmpcat, ".");

    /* pointer to head structure */
    ph = &hd;

    /* loop over external parameters */
    setopt(argc, argv);

    /* set pointer onto weighting function */
    func_weight = w_eq;

    /* detect resolution of map */
    if (lmax > 0) {
	    nx = lmax * 2;
	    if (nx%2 == 0)  nx++;
    } else if (pixel_size > 0) {
	    pixel_size = sqrt(pixel_size);
PR(ER, "inp_pix=%g\n",  pixel_size);
	    nx = pi / pixel_size;
	    lmax = nx / 2;
	    if (nx%2 == 0)  nx++;
	    /* np = pi * 2 / pixel_size; */
    } else {
label_lmax:
	    str[0] = '\0';
	    PR(ER, "%s: Could you set l_max, please ? =>", myname);
	    fgets(str, 20, stdin);
	    if(*str == '\0' || *str == 'n')  exit(0);
	    if(*str == '\n')  goto label_lmax;
	    if(*str == 'y') {
		   str[0] = '\0';
		   PR(ER, "your value, please ? =>");
		   fgets(str, 20, stdin);
		   if(*str == '\0' || *str == 'n')  exit(0);
		   if(*str == '\n')  goto label_lmax;
	    }
	    lmax = atoi(str);
	    nx = lmax * 2;
	    if (nx%2 == 0)  nx++;
    }
    if (np <= 0)  np = 2*nx;

    /* take memory for map-arrays */
    if ((x = (double *)calloc(nx, sizeof *x)) == (double *)0) {
nomem:
	    PR(ER,"%s: no memory for main arrays...\n", myname);
	    exit(1);
    }
    if ((nph = (int *)calloc(nx, sizeof *nph)) == (int *)0)
	    goto nomem;

    /* set grid Gauss-Legendre grid */
    gauleg(x1,x2, x, (double *)0, nx);

    /* set values by nphi */
    ntot = set_nphi_by_x2(x, nx, np, nph);

{
int i = (nx+1)/2;
double dt= 0.5*(acos(x[i+1]) - acos(x[i-1]))/pi*180*60,
       dp= 2. / np * 60. * 180.;
s0 = dt * dp;
PR(ER, "pix_size=%gx%g=%g sq.arcmin, nx=%d,np=%d, lmax=%d,\nntot=%d(%g%%)\n",
dt, dp, s0, nx, np, lmax, ntot, ntot * s0/(60.*60*180*180/pi/pi)/pi/4.*100);
}

    /* take memory for map-array */
    if ((f = (double *)calloc((unsigned)(nx*np), sizeof *f))==(double *)0)
	    goto nomem;

    if (repix_map) {
	/* read old map */
	double *f_old = (double *)0,
	       *x_old = (double *)0;
	int *nph_old = (int *)0;
	int nx_old, np_old;

	/* read initial map */
	read_map2(&f_old, &nx_old, &np_old, &nph_old, &x_old, obs_file);

	/* transpose image */
	/* f(nb, na) -> f(na, nb) */
	transpose_map(f_old, np_old, nx_old);

	/* appropriate pixels -> 1st grid to 2nd grid */
	repix_grid(f_old, nx_old,np_old,nph_old, x_old,   f, nx,np,nph, x);

	/* free memory */
	free((void *)f_old);
	free((void *)x_old);
	free((void *)nph_old);
    } else {
	/* read observational data */
	read_obs(f, nx, np, x, nph, obs_file);
    }

    /* output map */
    /* open output file */
    if ((fp_out = fopen(outp_file, "w")) == (FILE *)0) {
	    PR(ER, "%s: can not create file '%s'\n", myname, outp_file);
	    exit(1);
    }

    /* select grid output keyword values */
    strcpy(x_type, "COS(THETA)");
    strcpy(grid_type, "PHI(X)+LEGENDRE");

    /* write tac2cmb-file */
    ff_write_ext2_map(f, x, nx, np, nph, fp_out, x_type, grid_type);

    /* free memory */
    if (f != (double *)0)  free((void *)f);
    if (x != (double *)0)  free((void *)x);
    if (nph != (int *)0)   free((int *)nph);

    exit(0);
}

void setopt(argc, argv)
int argc;
char **argv;
{
    register int i;
    myname = argv[0];

    for(i=1; i<argc; i++) {
	switch(argv[i][0]) {
	case '-':
	    switch(argv[i][1]) {
	    case 'h':
		    usage();
		    break;
	    case 'i':
		    strcpy(obs_file, argv[++i]);
		    break;
	    case 'l':
		    lmax = atoi(argv[++i]);
		    break;
	    case 'm':
		    repix_map = 1;
		    strcpy(obs_file, argv[++i]);
		    break;
	    case 'o':
		    strcpy(outp_file, argv[++i]);
		    break;
	    case 'p':
		    pixel_size = atof(argv[++i]);
		    pixel_size /= 1.1664e+8/(M_PI*M_PI); /* sq.min->sr*/
		    break;
	    default:
		    break;
	    }
	    break;
	case '+':
	    switch(argv[i][1]) {
	    case 'h':
	    default:
		    usage();
		    break;
	    }
	    break;
	default:
	    usage();
	    break;
	}
    }

    return;
}

void usage()
{
PR(ER,"usage: %s -i obs_file -l lmax -o file_out [-p pix_size(sq.arcmin)]\n",
myname);
PR(ER,"[-m old_map_file]\n");
exit(2);
}
