#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "glesp.h"

#define PR      fprintf
#define ER      stderr

struct _source_ {
    float min_axis, maj_axis;   /* half power width in ellipse */
    double x_angle; /* angle to X-axis, deg  */
    double ra,      /* right ascension, hour */
	   dec;     /* declination, deg      */
    float amp;
    double theta, phi;
};

extern void pat_src(double *, double *, int, int, int *, struct _source_ *,
	  int);


/*
 *
 *  Oleg Verkhodanov, vo@sao.ru, 2008
 *      last modification, 20/06/2008, 28/07/2008
 */

void map_ellip_src(f,x,nx,np,nph,filename)
double *f, *x;
int nx, np;
int *nph;
char *filename;
{
    FILE *fp;
    struct _source_ *src;
    int ns = 0;
    char str[100];
    double p = M_PI/180.;

    if ((fp = fopen(filename, "r")) == (FILE *)0) {
	    PR(ER, "%s: I can not open file '%s'.\n",
		myname, filename);
	    exit(1);
    }

    while (fgets(str, 99, fp) != NULL ) {
	if (*str == '#' || *str == '%' || *str == '\n' || *str == '\r')
		continue;
	ns++;
    }

    if ((src = (struct _source_ *)calloc ((unsigned)ns, sizeof *src)) ==
	(struct _source_ *)0) {
	    PR(ER, "%s: no memory for source structure array.\n", myname);
	    exit(1);
    }

    ns = 0;
    fseek(fp, 0l, 0);
    while (fgets(str, 99, fp) != NULL ) {
	char sra[21], sdec[21];
	double lat, lon;
	if (*str == '#' || *str == '%' || *str == '\n' || *str == '\r')
		continue;
	sscanf(str, "%s%s%f%f%f%lf", sra, sdec, &src[ns].amp,
	      &src[ns].min_axis, &src[ns].maj_axis, &src[ns].x_angle);

//PR(ER, "min_ax=%g, maj_ax=%g, amp=%g, ang=%g\n",
//src[ns].min_axis, src[ns].maj_axis, src[ns].amp, src[ns].x_angle);

	/* major axis should be larger then minor axis */
	if ( src[ns].min_axis > src[ns].maj_axis) {
	    lat = src[ns].min_axis;
	    src[ns].min_axis = src[ns].maj_axis;
	    src[ns].maj_axis = lat;
	}

	src[ns].ra  = str2deg(sra);     /* string to hours   */
	src[ns].dec = str2deg(sdec);    /* string to degrees */

	/* galactic coordinates */
	equat2gal(src[ns].ra, src[ns].dec, &lat, &lon);
//PR(ER, "lat=%g, lon=%g\n", lat, lon);
	lat *= p;   /* in radians */
	lon *= p;   /* in radians */

	/* polar coordinates */
	src[ns].theta  = M_PI/2 - lat;
	src[ns].phi    = lon;

	ns++;
    }
//PR(ER, "ns=%d\n", ns);
    fclose(fp);

    /* read source data  and put them on a map */
    pat_src(f, x, nx, np, nph, src, ns);

    free ((void *)src);
    return;
}


void pat_src(f, xx, nx, np, nph, src, ns)
double *f, *xx;
int nx, np;
int *nph;
struct _source_ *src;
int ns;
{
    register int i,j,n;
    int ii, mx, my, i1, j1;
    double sig1=1., sig2=1., x, y, x0, y0, dx = 1., dy = 1., ang;
    double *r = (double *)0, *q = (double *)0;
    float max_ax, min_ax;
    double ln2 = sqrt (8.*log(2));
    double k_min2rad = M_PI / 180. / 60.;
    extern void source_pars(double *, int,int, double,double,double,double);

    /* find maximum and minimum axis */
    max_ax = 0;
    min_ax = 1.e+30;
    for(n=0; n<ns; n++) {
	if (max_ax < src[n].maj_axis)  max_ax = src[n].maj_axis;
	if (min_ax > src[n].min_axis)  min_ax = src[n].min_axis;
    }
    /* step in minutes */
    /* let minimal number of points on minimal axis be 11 */
    if (min_ax < 11.)
	dx = min_ax / 11.;
    else if (min_ax > 25.)
	dx = min_ax / 25.;
    else dx = 1.;

    /* compare with input map step 360*60/np */
    /* and make step in area at least twice less than on a map */
    if (21600./np < 2*dx)  dx = 21600./np /2;

    dy = dx;
//PR(ER, "steps in input area with source: dx = dy = %g arcmin\n", dx);
//PR(ER, "minor axis = %g, major axis = %g arcmin\n",
//src[0].min_axis, src[0].maj_axis);

    /* let side size of square area be 5*maximum axis */
    my = 5 * (int)(max_ax / dy);    /* size in pixels */
    if (! my%2) my++; /* make odd */
    mx = my;

PR(ER, "ns=%d, mx=%d, my=%d\n", ns, mx, my);
    /* let center of source be in the center of area */
    x0 = mx*dx/2.;
    y0 = my*dy/2.;

    /* take memory for area */
    /* r is array with input source, q is rotated r */
    if ((r = (double *)calloc((unsigned)(mx*my), sizeof *r)) == (double *)0) {
nomem:
	PR(ER, "%s: no memory for source area...\n", myname);
	exit(1);
    }
    if ((q = (double *)calloc((unsigned)(mx*my), sizeof *q)) == (double *)0)
	goto nomem;

    /* loop over sources in a list */
    for(n=0; n<ns; n++) {

	/* sigmas for gaussian */
	sig1 = src[n].min_axis / ln2;
	sig2 = src[n].maj_axis / ln2;
	sig1 *= sig1;
	sig2 *= sig2;

	/* put elliptic gaussian into area */
	for(i=0; i<my; i++) {
	    ii = i * mx;
	    y = i * dy;
	    for(j=0; j<mx; j++) {
		double xx, yy;
		x = j * dx;
		xx = x - x0, yy = y - y0;
		r[ii+j] = src[n].amp * exp (-0.5*(xx*xx/sig1 + yy*yy/sig2));
//              r[ii+j] += 1.e-15 * src[n].amp * sin((i+j)*M_PI/8);
	    }
	}

//source_pars(r, mx, my, 0., 0., dx*k_min2rad,dy*k_min2rad);


	/* rotate area with gaussian */
	ang = (180. - src[n].x_angle) * M_PI / 180.;
//      ang = src[n].x_angle * M_PI / 180.;
	for(i=0; i<my; i++) {
	    int ii = i * mx;
	    y = i * dy - y0;        /* use a center of area */
	    for(j=0; j<mx; j++) {
		double x1,y1;
		x = j * dx - x0;    /* use a center of area */

//              /* new x and y */
//              x1 = x * cos(ang) + y * sin(ang);
//              y1 = y * cos(ang) - x * sin(ang);
//              /* convert new (x,y) to new (i,j) */
//              i1 = (y0+y1) / dy;
//              j1 = (x0+x1) / dx;

		/* old x1 and y1 */
		x1 = x * cos(ang) - y * sin(ang);
		y1 = y * cos(ang) + x * sin(ang);

		/* convert old (x,y) to old (i,j) */
		i1 = (y0+y1) / dy;
		j1 = (x0+x1) / dx;

		/* nothing to do if out of limits */
		if (i1 < 0 || j1 < 0 || i1 >= my || j1 >= mx)
		    continue;

		/* copy old pixel to new area */
//              q[i1*mx+j1] = r[ii+j];
		q[ii+j] = r[i1*mx+j1];
	    }
	}

//source_pars(q, mx, my, 0., 0., dx*k_min2rad,dy*k_min2rad);

	/* copy pixel by pixel from area onto a map */
PR(ER, "n=%d, theta=%g, phi=%g, amp=%g\n",
n, src[n].theta, src[n].phi,  src[n].amp);
	for(i=0; i<my; i++) {
	    int ii = i * mx;
	    double phi, theta;
	    theta = src[n].theta + (i*dy - y0) * k_min2rad; /* in radian */
	    for(j=0; j<mx; j++) {
		phi = src[n].phi + (j*dx - x0) * k_min2rad; /* in radian */

		/* coordinates on a sphere */
		angle2pixel(xx, nx, nph, &i1, &j1, theta, phi);

		/* copy pixels */
		f[i1*np+j1] = q[ii+j];
	    }
	}

	/* make zero in rotated array */
	if (n < ns-1) {
	    for(i=0; i<my; i++) {
		int ii = i * mx;
		for(j=0; j<mx; j++) q[ii+j] = 0.0;
	    }
	}
    }

    free((void *)r);
    free((void *)q);

    return;
}
