#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "glesp.h"
#include "gd/plot_gd.h"

#define PR  fprintf
#define ER  stderr
#define MAX(x,y)    ((x) > (y) ? (x) : (y))
#define ABS(x)      ((x) <  0 ? -(x) : (x))

/*
 *  Put phases of alm-file as color points on the triangular diagram:
    phase(l,m)
 *      Oleg Verkhodanov, 2011
 *      last changes  22/11/2011, 13/06/2011
 */

void plot_phase_diagram(alm_file, lmax_inp, xs,ys,bar,font_size,
				 outp_file,col_flag)
char *alm_file;
int lmax_inp;
int xs, ys;     /* size of window */
int bar;
int font_size;
char *outp_file;
int col_flag;   /* if flag=1, to use values of sin(phase) instead of phase */
		/* if flag=-1, to use HSV-colors */
{
    double *img = (double *)0;
    int *imgs = (int *)0, *img_bar = (int *)0, *img_axes = (int *)0;
    double *alm = (double *)0;
    int nbuf = 0, lmin = 0, lmax = 0, nalm = 0;
    int axis_flag = 1;
    extern void phase2triang(double *, int, int, double **, int, int, int);
    extern void alm2img(double *, int *, int, int, double, double),
		alm2img_hsv(double *, int *, int, int, double, double);

    /* read alm-table */
    read_alm_table(alm_file, &alm, &nalm, &nbuf, &lmax);

    /* set new limit of lmax  */
    if (lmax_inp > 0) {
	int lmax_old = lmax, nbuf_old = nbuf;
	int nbuf_new = 2 * (lmax+1);
	alm2alm(&alm, &lmax_old, &nbuf_old, lmax_inp, nbuf_new);
	nbuf = nbuf_new;
	lmax = lmax_inp;
    }


    /* find memory for image */
    imgs = (int *)calloc(xs*ys, sizeof *imgs);
    if (imgs == (int *)0) {
	PR(ER, "%s: no memory for image...\n", myname);
	exit(1);
    }

    /* set color table */
//  setcol_i(col_map);

    phase2triang(alm, lmax, nbuf, &img, xs, ys, col_flag>0 ? 1 : 0);

    /* converse float map to scaled integer image */
    if (col_flag > 0)
	alm2img(img, imgs, xs, ys, -1., 1.);     /* sin instead of phase */
    else if (col_flag < 0)
	alm2img_hsv(img, imgs, xs, ys, 0., 2*M_PI); /* phase in HSV-color */
    else
	alm2img(img, imgs, xs, ys, 0., 2*M_PI);

    if (axis_flag) {
	char sax1[30], sax2[30];
	extern void addaxes(int *, int **, char *, char *, int, int);
	strcpy (sax1, "l");
	strcpy (sax2, "m");
	set_gd_font_size(font_size);
	addaxes(imgs, &img_axes, sax1, sax2, lmin, lmax);
	free((void *)imgs);        /* free memory from prev.map */
	imgs = img_axes;            /* change pointer to new map */
    }

    /* add bar and rebuild image */
    if (bar) {
	set_gd_font_size(font_size);
	color_bar(imgs, &img_bar); /* add axis titles */
	free((void *)imgs);        /* free memory from prev.map */
	imgs = img_bar;            /* change pointer to new map */
    }

    /* convert fig to gif */
    /* and write to file  */
    map2file(imgs, outp_file);

    /* free memory */
    if (img  != (double *)0) free ((void *)img);
    if (imgs != (int *)0)    free ((void *)imgs);
    if (alm  != (double *)0) free ((void *)alm);

    return;
}

void plot_ascii_phase_diagram(ph_file, lmax_inp, xs,ys,bar,font_size,
				 outp_file,col_flag)
char *ph_file;
int lmax_inp;
int xs, ys;     /* size of window */
int bar;
int font_size;
char *outp_file;
int col_flag;   /* if flag=1, to use values of sin(phase) instead of phase */
		/* if flag=-1, to use HSV-colors */
{
    double *img = (double *)0;
    int *imgs = (int *)0, *img_bar = (int *)0, *img_axes = (int *)0;
    double *phase = (double *)0;
    int lmin = 0, lmax = 0;
    int axis_flag = 1;
    extern void phasetab2triang(double *, int, double **, int, int, int);
    extern void alm2img(double *, int *, int, int, double, double),
		alm2img_hsv(double *, int *, int, int, double, double);

    /* read alm-table */
    read_ph_table(ph_file, &phase, &lmax);

    /* set new limit of lmax  */
    if (lmax_inp > 0 && lmax_inp != lmax) {
	register int l, m;
	double *phs = (double *)0;
	unsigned int bb = (lmax_inp + 1) * (lmax_inp + 1);
	if ((phs = (double*)calloc(bb, sizeof *phs)) == (double *)0) {
	    PR(ER, "%s: no memory for temporal array...\n", myname);
	    exit(1);
	}
	for (l=0; l<=lmax_inp; l++) {
	    int ll = (lmax_inp + 1) * l;
	    int llo = (lmax + 1) * l;
	    if (l <= lmax)
		for (m=0; m<=l; m++)
		    phs[ll+m] = phase[llo+m];
	    else
		break;
	}
	lmax = lmax_inp;
	free((void *)phase);
	phase = phs;
    }


    /* find memory for image */
    imgs = (int *)calloc(xs*ys, sizeof *imgs);
    if (imgs == (int *)0) {
	PR(ER, "%s: no memory for image...\n", myname);
	exit(1);
    }

    /* set color table */
//  setcol_i(col_map);

    phasetab2triang(phase, lmax, &img, xs, ys, col_flag>0 ? 1 : 0);

    /* converse float map to scaled integer image */
    if (col_flag > 0)
	alm2img(img, imgs, xs, ys, -1., 1.);     /* sin instead of phase */
    else if (col_flag < 0)
	alm2img_hsv(img, imgs, xs, ys, 0., 2*M_PI); /* phase in HSV-color */
    else
	alm2img(img, imgs, xs, ys, 0., 2*M_PI);

    if (axis_flag) {
	char sax1[30], sax2[30];
	extern void addaxes(int *, int **, char *, char *, int, int);
	strcpy (sax1, "l");
	strcpy (sax2, "m");
	set_gd_font_size(font_size);
	addaxes(imgs, &img_axes, sax1, sax2, lmin, lmax);
	free((void *)imgs);        /* free memory from prev.map */
	imgs = img_axes;            /* change pointer to new map */
    }

    /* add bar and rebuild image */
    if (bar) {
	set_gd_font_size(font_size);
	color_bar(imgs, &img_bar); /* add axis titles */
	free((void *)imgs);        /* free memory from prev.map */
	imgs = img_bar;            /* change pointer to new map */
    }

    /* convert fig to gif */
    /* and write to file  */
    map2file(imgs, outp_file);

    /* free memory */
    if (img  != (double *)0) free ((void *)img);
    if (imgs != (int *)0)    free ((void *)imgs);
    if (phase  != (double *)0) free ((void *)phase);

    return;
}

#define N_SPEC  22.45915771    /*  pi^e  */

void phase2triang(alm, lmax, nbuf, img, xs, ys, sin_flag)
double *alm;
int lmax, nbuf;
double **img;
int xs, ys;     /* window size */
int sin_flag;   /* flag to use values of sin(phase) instead of phase */
{
    register int i, j, l, m;
    unsigned int n = xs*ys;
    double dl, dm;
    double x_c, y_c, x_r, y_r;

    /* find memorry */
    if ( (*img = (double *)calloc(n, sizeof **img)) == (double *)0 ) {
	PR(ER, "phase2triang: no memory for temporal array...\n");
	exit(1);
    }
    for(i=0; i<n; i++)  (*img)[i] = N_SPEC;

    /* center of image */
    y_c = ys / 2;
    x_c = xs / 2;
    x_r = xs / 2;
    y_r = ys / 2;
    dl = (double)(lmax) / ys;
    dm = (double)(lmax) / xs;

    /* fill in image */
    for(i=0; i<ys; i++) {
	register int ii = i * xs, ll;
	l = i * dl + .5;
	ll = l * nbuf;
	for(j=0; j<xs; j++) {
	    double re, im, phase;
	    (*img)[ii+j] = 0.0;
	    m = j * dm + .5;

	    if (m > l) break;

	    re = alm[ll+m];
	    im = alm[ll+m+lmax+1];
	    phase = ph_calc(re, im);

	    if (m==0) {
		if (re < 0.0)  phase = M_PI; else phase = 0.0;
	    } else {
		phase = ph_calc(re, im);
	    }

	    if (sin_flag)
		(*img)[ii+j] = sin(phase);
	    else
		(*img)[ii+j] = phase;
	}
    }

    return;
}

void phasetab2triang(phase, lmax, img, xs, ys, sin_flag)
double *phase;
int lmax;
double **img;
int xs, ys;     /* window size */
int sin_flag;   /* flag to use values of sin(phase) instead of phase */
{
    register int i, j, l, m;
    unsigned int n = xs*ys;
    double dl, dm;
    double x_c, y_c, x_r, y_r;

    /* find memorry */
    if ( (*img = (double *)calloc(n, sizeof **img)) == (double *)0 ) {
	PR(ER, "phase2triang: no memory for temporal array...\n");
	exit(1);
    }
    for(i=0; i<n; i++)  (*img)[i] = N_SPEC;

    /* center of image */
    y_c = ys / 2;
    x_c = xs / 2;
    x_r = xs / 2;
    y_r = ys / 2;
    dl = (double)(lmax+1) / ys;
    dm = (double)(lmax+1) / xs;

    /* fill in image */
    for(i=0; i<ys; i++) {
	register int ii = i * xs, ll;
	l = i * dl + .5;
	ll = l * (lmax + 1);
	for(j=0; j<xs; j++) {
	    double ph;
	    (*img)[ii+j] = 0.0;
	    m = j * dm + .5;

	    if (m > l) break;

	    ph = phase[ll+m];

	    if (sin_flag)
		(*img)[ii+j] = sin(ph);
	    else
		(*img)[ii+j] = ph;
	}
    }

    return;
}

void alm2img(img, imgs, xs, ys, amp_min, amp_max)
double *img;
int *imgs;
int xs, ys;
double amp_min, amp_max;
{
    register int i;
    double da, amin=amp_min, amax=amp_max, tiny = 1.0e-20;
    int nn, nc = n_colors();

    set_img_size(xs,ys);
    nn = xs * ys;

    set_gdlim(amin, amax);
    da = MAX(amax - amin, tiny);

    for(i=0; i<nn; i++) {
	if (ABS(img[i]-N_SPEC) > 1.e-7) {
	    double v = img[i];
	    v = v < amin ? amin : v;  /* cut values */
	    v = v > amax ? amax : v;  /*            */
	    imgs[i] = (int)((v - amin) / da * (double)(nc-4)) + 3;
	} else
	    imgs[i] = 1;  /* white color for empty triangle */
    }

    return;
}

void alm2img_hsv(img, imgs, xs, ys, amp_min, amp_max)
double *img;
int *imgs;
int xs, ys;
double amp_min, amp_max;
{
    register int i;
    double da, amin=amp_min, amax=amp_max, tiny = 1.0e-20;
    int nn, nc = n_colors();
    extern void hsv2rgb (double,double,double, double*,double*,double*);
    extern int  dindex_color(double, double, double);

    set_img_size(xs,ys);
    nn = xs * ys;

    set_gdlim(amin, amax);
    da = MAX(amax - amin, tiny);

    for(i=0; i<nn; i++) {
	double h,s,v, r,g,b;
	int cc;

	if (ABS(img[i]-N_SPEC) > 1.e-7) {
	    double v = img[i];
	    v = v < amin ? amin : v;  /* cut values */
	    v = v > amax ? amax : v;  /*            */
	    cc = (int)((v - amin) / da * (double)(nc-4)) + 3;

	    /* convert phase to color of hue in HSV-scale and then to RGB */
	    h = (double)cc / (nc-4) * 360.;
//          h = img[i] / 2./M_PI *360.;
	    s = 1.0;
	    v = 1.0;
	    hsv2rgb (h,s,v, &r, &g, &b);
//PR(ER, "cc=%d, h=%g, r=%g, g=%g, b=%g, ", cc,h,r,g,b);
	    cc =  dindex_color(r, g, b);
//PR(ER, "c=%d\n", cc);
	} else
	    cc = 1;  /* white color for empty triangle */


	imgs[i] = cc;
    }
    return;
}


void plot_ampl_diagram(alm_file, lmax_in, xs, ys, bar, font_size, outp_file,
		 ami,ama,log_flag)
char *alm_file;
int lmax_in;
int xs, ys;
int bar;
int font_size;
char *outp_file;
double ami, ama; /* external limits for amplitude value */
int log_flag;
{
    double *img = (double *)0;
    int *imgs = (int *)0, *img_bar = (int *)0, *img_axes = (int *)0;
    double *alm = (double *)0;
    int nbuf = 0, lmin=0, lmax = 0, nalm = 0;
    double  amp_min=0., amp_max=0.;
    int axis_flag = 1;
    extern void ampl2triang(double *, int, int, double **, int, int, int);
    extern void alm2img(double *, int *, int, int, double, double);
    extern void ampl_minmax(double *, int, int, double *, double *, int);

    /* read alm-table */
    read_alm_table(alm_file, &alm, &nalm, &nbuf, &lmax);

    /* find memory for image */
    imgs = (int *)calloc(xs*ys, sizeof *imgs);
    if (imgs == (int *)0) {
	PR(ER, "%s: no memory for image...\n", myname);
	exit(1);
    }

    /* set color table */
//  setcol_i(col_map);

    ampl2triang(alm, lmax, nbuf, &img, xs, ys, log_flag);

    /* find minimum and maximum in amplitude */
    ampl_minmax(alm, lmax, nbuf, &amp_min, &amp_max, log_flag);

    /* reset limits */
    if (ami > -1.5e+38)   amp_min = ami;
    if (ama < +1.5e+38)   amp_max = ama;

    /* converse float map to scaled integer image */
    alm2img(img, imgs, xs, ys, amp_min, amp_max);

    if (axis_flag) {
	char sax1[30], sax2[30];
	extern void addaxes(int *, int **, char *, char *, int, int);
	strcpy (sax1, "L");
	strcpy (sax2, "M");
	set_gd_font_size(font_size);
	addaxes(imgs, &img_axes, sax1, sax2, lmin, lmax);
	free((void *)imgs);        /* free memory from prev.map */
	imgs = img_axes;            /* change pointer to new map */
    }

    /* add bar and rebuild image */
    if (bar) {
	set_gd_font_size(font_size);
	color_bar(imgs, &img_bar); /* add color bar */
	free((void *)imgs);        /* free memory from prev.map */
	imgs = img_bar;            /* change pointer to new map */
    }

    /* convert fig to gif */
    /* and write to file  */
    map2file(imgs, outp_file);

    /* free memory */
    if (img  != (double *)0) free ((void *)img);
    if (imgs != (int *)0)    free ((void *)imgs);
    if (alm  != (double *)0) free ((void *)alm);

    return;
}

void plot_ascii_ampl_diagram(ampfile, lmax_in, xs, ys, bar, font_size,
	outp_file, ami,ama,log_flag)
char *ampfile;
int lmax_in;
int xs, ys;
int bar;
int font_size;
char *outp_file;
double ami, ama; /* external limits for amplitude value */
int log_flag;
{
    double *img = (double *)0;
    int *imgs = (int *)0, *img_bar = (int *)0, *img_axes = (int *)0;
    double *amp = (double *)0;
    int lmin=0, lmax = 0;
    double  amp_min=0., amp_max=0.;
    int axis_flag = 1;
    extern void amptab2triang(double *, int, double **, int, int, int);
    extern void alm2img(double *, int *, int, int, double, double);
    extern void amptab_minmax(double *, int, double *, double *, int);

    /* read alm-table */
    read_ph_table(ampfile, &amp, &lmax);

    /* find memory for image */
    imgs = (int *)calloc(xs*ys, sizeof *imgs);
    if (imgs == (int *)0) {
	PR(ER, "%s: no memory for image...\n", myname);
	exit(1);
    }

    /* set color table */
//  setcol_i(col_map);

    amptab2triang(amp, lmax, &img, xs, ys, log_flag);

    /* find minimum and maximum in amplitude */
    amptab_minmax(amp, lmax, &amp_min, &amp_max, log_flag);

    /* reset limits */
    if (ami > -1.5e+38)   amp_min = ami;
    if (ama < +1.5e+38)   amp_max = ama;

    /* converse float map to scaled integer image */
    alm2img(img, imgs, xs, ys, amp_min, amp_max);

    if (axis_flag) {
	char sax1[30], sax2[30];
	extern void addaxes(int *, int **, char *, char *, int, int);
	strcpy (sax1, "L");
	strcpy (sax2, "M");
	set_gd_font_size(font_size);
	addaxes(imgs, &img_axes, sax1, sax2, lmin, lmax);
	free((void *)imgs);        /* free memory from prev.map */
	imgs = img_axes;            /* change pointer to new map */
    }

    /* add bar and rebuild image */
    if (bar) {
	set_gd_font_size(font_size);
	color_bar(imgs, &img_bar); /* add color bar */
	free((void *)imgs);        /* free memory from prev.map */
	imgs = img_bar;            /* change pointer to new map */
    }

    /* convert fig to gif */
    /* and write to file  */
    map2file(imgs, outp_file);

    /* free memory */
    if (img  != (double *)0) free ((void *)img);
    if (imgs != (int *)0)    free ((void *)imgs);
    if (amp  != (double *)0) free ((void *)amp);

    return;
}

#define N_SPEC  22.45915771    /*  pi^e  */

void ampl2triang(alm, lmax, nbuf, img, xs, ys, log_flag)
double *alm;
int lmax, nbuf;
double **img;
int xs, ys;
{
    register int i, j, l, m;
    unsigned int n = xs*ys;
    double dl, dm;
    double x_c, y_c, x_r, y_r;

    /* find memorry */
    if ( (*img = (double *)calloc(n, sizeof **img)) == (double *)0 ) {
	PR(ER, "phase2triang: no memory for temporal array...\n");
	exit(1);
    }
    for(i=0; i<n; i++)  (*img)[i] = N_SPEC;

    /* center of image */
    y_c = ys / 2;
    x_c = xs / 2;
    x_r = xs / 2;
    y_r = ys / 2;
    dl = (double)(lmax+1) / ys;
    dm = (double)(lmax+1) / xs;

    /* fill in image */
    for(i=0; i<ys; i++) {
	register int ii = i * xs, ll;
	l = i * dl + .5;
	ll = l * nbuf;
	for(j=0; j<xs; j++) {
	    double re, im, phase, amp;
	    (*img)[ii+j] = 0.0;
	    m = j * dm + .5;

	    if (m > l) break;

	    re = alm[ll+m];
	    im = alm[ll+m+lmax+1];

	    amp = sqrt(re*re+im*im);
	    phase = ph_calc(re, im);

	    if (m==0) {
		if (re < 0.0)  phase = M_PI; else phase = 0.0;
	    } else {
		phase = ph_calc(re, im);
	    }

	    if (log_flag) {
	       amp = log10(amp);
	    } else {
		if (ABS(phase-M_PI) < 1.e-15 || (re < 0 && ABS(im) < 1.e-15))
			amp = -amp;
	    }
	    (*img)[ii+j] = amp;
	}
    }

    return;
}

void amptab2triang(amp, lmax, img, xs, ys, log_flag)
double *amp;
int lmax;
double **img;
int xs, ys;
{
    register int i, j, l, m;
    unsigned int n = xs*ys;
    double dl, dm;
    double x_c, y_c, x_r, y_r;

    /* find memorry */
    if ( (*img = (double *)calloc(n, sizeof **img)) == (double *)0 ) {
	PR(ER, "phase2triang: no memory for temporal array...\n");
	exit(1);
    }
    for(i=0; i<n; i++)  (*img)[i] = N_SPEC;

    /* center of image */
    y_c = ys / 2;
    x_c = xs / 2;
    x_r = xs / 2;
    y_r = ys / 2;
    dl = (double)(lmax+1) / ys;
    dm = (double)(lmax+1) / xs;

    /* fill in image */
    for(i=0; i<ys; i++) {
	register int ii = i * xs, ll;
	l = i * dl + .5;
	ll = l * (lmax + 1);
	for(j=0; j<xs; j++) {
	    double a;
	    (*img)[ii+j] = 0.0;
	    m = j * dm + .5;

	    if (m > l) break;

	    a = amp[ll+m];

	    if (log_flag) {
	       a = log10(fabs(a));
	    }
	    (*img)[ii+j] = a;
	}
    }

    return;
}


void ampl_minmax(alm, lmax, nbuf, amp_min, amp_max, log_flag)
double *alm;
int lmax, nbuf;
double *amp_min, *amp_max;
int log_flag;
{
    register int l, m;
    double amin = 1e+30, amax = -1e+30;

    for(l=0; l<=lmax; l++) {
	register int ll = l * nbuf;
	for(m=0; m<=l; m++) {
	    double re = alm[ll+m],  im = alm[ll+m+lmax+1];
	    double amp = sqrt(re*re+im*im),  phase = ph_calc(re, im);

	    if (m==0) {
		if (re < 0.0)  phase = M_PI; else phase = 0.0;
	    } else {
		phase = ph_calc(re, im);
	    }

	    if (log_flag) {
		if (ABS(amp) < 1.e-30 ) amp = 1.e-30;
		amp = log10(amp);
	    } else {
		if(ABS(phase - M_PI) < 1.e-15 || (re < 0 && ABS(im) < 1.e-15))
		    amp = -amp;
	    }

	    if (amp < amin) amin = amp;
	    if (amp > amax) amax = amp;
	}
    }

    *amp_min = amin;
    *amp_max = amax;

    return;
}


void amptab_minmax(amp, lmax, amp_min, amp_max, log_flag)
double *amp;
int lmax;
double *amp_min, *amp_max;
int log_flag;
{
    register int l, m;
    double amin = 1e+30, amax = -1e+30;

    for(l=0; l<=lmax; l++) {
	register int ll = l * (lmax + 1);
	for(m=0; m<=l; m++) {
	    double a = amp[ll+m];

	    if (log_flag) {
		if (ABS(a) < 1.e-30 ) a = 1.e-30;
		a = log10(a);
	    }

	    if (a < amin) amin = a;
	    if (a > amax) amax = a;
	}
    }

    *amp_min = amin;
    *amp_max = amax;

    return;
}


