#include <stdio.h>
#include <math.h>

#define PR      fprintf
#define ER      stderr
#define ABS(x)  ((x) < 0. ? -(x) : (x))

extern int    ind_x(int, double *, double);
static double pi2 = 2. * M_PI;

void pixel2angle(x, nx, nph, i_x, j_phi, theta, phi)
double *x;
int nx;
int *nph;
int i_x, j_phi;
double *theta, *phi;
{
    double bad_value = -1010.;
    if (i_x < 0 && i_x >= nx) {
/*
	PR(ER,
	    "pixel2angle: number of pixel by theta-axis is out of range.\n");
*/
	*theta = *phi = bad_value;
    } else if (j_phi < 0 && j_phi >= nph[i_x]) {
/*
	PR(ER,
	    "pixel2angle: number of pixel by phi-axis is out of range.\n");
*/
	*theta = *phi = bad_value;
    } else {
	double dphi = pi2 / nph[i_x];
	*phi = dphi * j_phi;
	*theta = acos(x[i_x]);
    }

    return;
}
void pixel2hangle(x, nx, nph, i_x, j_phi, theta, phi, p_offset)
double *x;
int nx;
int *nph;
int i_x, j_phi;
double *theta, *phi;
double *p_offset;
{
    double bad_value = -1010.;
    if (i_x < 0 && i_x >= nx) {
	*theta = *phi = bad_value;
    } else if (j_phi < 0 && j_phi >= nph[i_x]) {
	*theta = *phi = bad_value;
    } else {
	double dphi = pi2 / nph[i_x];
	*phi = dphi * j_phi + p_offset[i_x];
	*theta = acos(x[i_x]);
    }

    return;
}

void hangle2pixel0(x, nx, nph, i_x, j_phi, theta, phi, p_offset)
double *x;
int nx;
int *nph;
int *i_x, *j_phi;
double theta, phi;
double *p_offset;
{
    *i_x   = ind_x(nx, x, cos(theta));
    if (*i_x == nx)  *i_x = -nx;
    if (*i_x  < 0)
	*j_phi = -1;
    else {
	double p = -p_offset[*i_x] + phi;
	double dphi = pi2 / nph[*i_x];
/*
PR(ER, "nx=%d, np=%d, i=%d, j=%d, offs=%f, p=%g\n",
nx, nph[*i_x], *i_x, (int)(p/dphi + 0.5), p_offset[*i_x], p);
*/
	if (p > pi2)  p -= pi2;
	if (p < 0.0)  p += pi2;

	if(ABS(p-pi2) < 1e-8*dphi)  p = 0.0;
/*
PR(ER, "1,nx=%d, np=%d, i=%d, j=%d(%g), offs=%f, p=%g\n",
nx, nph[*i_x], *i_x, (int)(p/dphi + 0.5), p/dphi, p_offset[*i_x], p);
*/
	*j_phi = p / dphi + 0.5;
	if (*j_phi >= nph[*i_x]) *j_phi -= 1;
    }
    return;
}

void hangle2pixel(x, nx, nph, i_x, j_phi, theta, phi, p_offset)
double *x;
int nx;
int *nph;
int *i_x, *j_phi;
double theta, phi;
double *p_offset;
{
    extern int hangle2pix(int, double, double, int *, int *);
    int n, np = nx+1, nside = np / 4;
    n = hangle2pix(nside, theta, phi, i_x, j_phi);
    return;
}

/* conversed from Hpix procedure */
int hangle2pix(nside, theta, phi, i_x, j_phi)
int nside;
double theta, phi;
int *i_x, *j_phi;
{
    int n;
    double cth = cos(theta),  cth_abs = ABS(cth);
    double r = fmod(phi, 2*M_PI) * 2/M_PI;

    if ( cth_abs <= 2./3. ) {
       /* in central region */

       double temp1 = nside*(.5 + r);
       double temp2 = nside*.75 * cth;
       int jp = (int)(temp1-temp2); /* index of  ascending edge line */
       int jm = (int)(temp1+temp2); /* index of descending edge line */

       int ir = nside+1 + jp - jm ; /* in {1,2n+1} (ring No from z=2/3) */
       int kshift = 1 - ir % 2;     /* kshift=1 if ir even, 0 otherwise */

       int nl4 = 4*nside;
       int ip = (int)((jp+jm - nside+kshift+1) / 2); /* in {0,4n-1} */
       if (ip >= nl4)  ip -= nl4;
       n = 2*nside*(nside-1) + nl4*(ir-1) + ip;

       *i_x = nside-1 + ir-1;
       *j_phi = ip;
    } else {
	/* for North and South polar caps */
       double tp = r - (int)r;
       double tmp = nside * sqrt(3.*(1.-cth_abs));
       int jp = (int)(tp * tmp );       /* increasing edge line index */
       int jm = (int)((1. - tp) * tmp); /* decreasing edge line index */

       int ir = jp + jm + 1;      /* ring No from the closest pole */
       int ip = (int)(r * ir);   /* in {0,4*ir-1}                 */
       if (ip >= 4*ir) ip -= 4*ir;
       if (cth > 0)
	  n = 2*ir*(ir-1) + ip;
       else
	  n = 12*nside*nside - 2*ir*(ir+1) + ip;

       if (cth > 0)
	    *i_x = ir-1;
       else
	    *i_x = 4*nside - ir-1;
       *j_phi = ip;
    }
    return n;
}

void angle2pixel(x, nx, nph, i_x, j_phi, theta, phi)
double *x;
int nx;
int *nph;
int *i_x, *j_phi;
double theta, phi;
{
    *i_x   = ind_x(nx, x, cos(theta));
    if (*i_x == nx)  *i_x = -nx;
    if (*i_x  < 0)
	*j_phi = -1;
    else {
	double dp = pi2 / nph[*i_x];
	*j_phi = phi / dp + 0.5;
	if ( *j_phi >= nph[*i_x] )  *j_phi=0;
    }
    return;
}

void angle2pixel_(x, nx, nph, i_x, j_phi, theta, phi)
double *x;
int *nx;
int *nph;
int *i_x, *j_phi;
double *theta, *phi;
{
    angle2pixel(x, *nx, nph, i_x, j_phi, *theta, *phi);
    return;
}
void pixel2angle_(x, nx, nph, i_x, j_phi, theta, phi)
double *x;
int *nx;
int *nph;
int *i_x, *j_phi;
double *theta, *phi;
{
    pixel2angle(x, *nx, nph, *i_x, *j_phi, theta, phi);
    return;
}

/* detect index of x in ordered X-table */
/*
 * given an array x[0,...,n-1], returns a value i such that x_in is
 * between  x[i] and x[i+1]. x must be monotonic, either increasing or
 * decreasing. i=-1 or i=n-1 is returned to indicate that x is out of range
 *  (Press et al., 2002)
 *
 */
int x_locate(n, x, x_in)
int n;
double *x, x_in;
{
	register int iu, im, il;
	int ascnd, n1=n-1, i;

	il = -1;        /*  initialize lower and  */
	iu = n;         /*  upper limits          */

	ascnd=(x[n1] >= x[0]);

	/* if we are not yet done compute a midpoint */
	while (iu-il > 1) {
		im = (iu+il) >> 1;
		if ((x_in >= x[im]) == ascnd)
			il = im; /* and replace either the lower limit */
		else
			iu = im; /* or the upper limit, as appropriate */
	} /* repeat until the test condition is satified*/

	/* Then set the output */
	if (x_in == x[0])       i = 0;
	else if (x_in == x[n1]) i = n1;
	else i = il;

	return i;
}

int ind_x(n, x, x_in)
int n;
double *x, x_in;
{
	int i = x_locate(n, x, x_in);
	if (i == -1)  i = 0;
	/* thanks to Martin Reinecke for finding bug in 'i=n-1' */
	else  if (i == n || i==n-1)  i = n-1;
	else  {
	     i = (fabs(x_in-x[i]) < fabs(x[i+1]-x_in) ? i : i+1);
	}
	return i;
}

void set_pix_scheme1(n, x, pw)
int n;
double *x, *pw;
{

}

/* return (i,j) by actual number of n-pixel */
void rn2ij(nx,nph,n,i,j)
int nx, *nph, n, *i, *j;
{
    register int k,nc=0;

    *i = *j = -1;

    k = 0;
    while (nc < n && k < nx)  {nc += nph[k]; k++;};
    if (k < nx)  {
	*i = k-1;
	nc -= nph[*i];
	k=0;
	while (nc < n && k < nph[*i])  {nc ++; k++;};
	*j = k-1;
    }

    return;
}

void set_pix_scheme2(nx, nphi, nph, x, pw)
int nx, nphi;
{

}

/* convert actual number of pixel to angle */
void na2ang(n, theta, phi, x, nx, nphi, nph)
int n;
double *x;
int nx, nphi, *nph;
double *theta, *phi;
{
	int i_x = 0, j_phi;
	int ns = 0;

	while (ns <= n && i_x < nx) { ns += nph[i_x]; i_x++; }
	if (i_x <= nx) {
	       i_x--;
	       ns -= nph[i_x];
	       j_phi = n - ns;
	} else {
		i_x = j_phi = -1;
	}
	pixel2angle(x, nx, nph, i_x, j_phi, theta, phi);
	return;
}

void n2ang(n, theta, phi, x, nx, nphi, nph)
int n;
double *x;
int nx, nphi, *nph;
double *theta, *phi;
{
	int i_x = n/nphi, j_phi = n - nphi * i_x;
	pixel2angle(x, nx, nph, i_x, j_phi, theta, phi);
	return;
}

void n2hang(n, theta, phi, x, nx, nphi, nph, p_offset)
int n;
double *x;
int nx, nphi, *nph;
double *theta, *phi;
double *p_offset;
{
	int i_x = n/nphi, j_phi = n - nphi * i_x;

	pixel2hangle(x, nx, nph, i_x, j_phi, theta, phi, p_offset);
	return;
}

int hang2n(theta, phi, x, nx, nphi, nph, p_offset)
double *x;
int nx, nphi, *nph;
double theta, phi;
double *p_offset;
{
	int n = 0, i_x=0, j_phi=0;

	hangle2pixel(x, nx, nph, &i_x, &j_phi, theta, phi, p_offset);
	if (i_x < 0 || j_phi < 0)
		n = -10;
	else
		n = i_x * nphi + j_phi;

	return n;
}

/* convert angle to pixel number in rectangular grid */
int ang2n(theta, phi, x, nx, nphi, nph)
double *x;
int nx, nphi, *nph;
double theta, phi;
{
	int n = 0, i_x=0, j_phi=0;

	angle2pixel(x, nx, nph, &i_x, &j_phi, theta, phi);
	if (i_x < 0 || j_phi < 0)
		n = -10;
	else
		n = i_x * nphi + j_phi;

	return n;
}

/* angle to actual number of pixel in spherical grid */
int ang2na(theta, phi, x, nx, nphi, nph)
double *x;
int nx, nphi, *nph;
double theta, phi;
{
	register int i;
	int n = 0, i_x=0, j_phi=0;

	angle2pixel(x, nx, nph, &i_x, &j_phi, theta, phi);

	if (i_x < 0 || j_phi < 0)
		n = -10;
	else {
	    for (i=0; i<i_x; i++)  n += nph[i];
	    n += j_phi;
	}

	return n;
}

void n2ang_(n, theta, phi, x, nx, nphi, nph)
int *n;
double *x;
int *nx, *nphi, *nph;
double *theta, *phi;
{
	n2ang(*n-1, theta, phi, x, *nx, *nphi, nph);
	return;
}
void ang2n_(n, theta, phi, x, nx, nphi, nph)
int *n;
double *x;
int *nx, *nphi, *nph;
double *theta, *phi;
{
	*n = 1 + ang2n(*theta, *phi, x, *nx, *nphi, nph);
	return;
}

/* get a pixel number of one scheme by a number in another one */
/* returns a number of pixel                                   */
int pix2pix(x1, nx1, nphi1, nph1, n2, x2, nx2, nphi2, nph2)
double *x1, *x2;
int *nph1, *nph2;
int nx1, nphi1,
    nx2, nphi2;
int n2;           /* input pixel number */
{
	int n1;
	double theta, phi;
	n2ang(n2, &theta, &phi, x2, nx2, nphi2, nph2);
	if (theta < -900. || phi < -900.)
		n1 = -1;
	else
		n1 = ang2n(theta, phi, x1, nx1, nphi1, nph1);

	return n1;
}

/* get a pixel number of one scheme by a number in another one */
/* returns a number of pixel                                   */
int hpix2pix(x1, nx1, nphi1, nph1, p_offset, n2, x2, nx2, nphi2, nph2)
double *x1, *x2;
int *nph1, *nph2;
double *p_offset;
int nx1, nphi1,
    nx2, nphi2;
int n2;           /* input pixel number */
{
	int n1;
	double theta, phi;

	if (p_offset == (double *)0) {
	      n1 = pix2pix(x1, nx1, nphi1, nph1, n2, x2, nx2, nphi2, nph2);
	      return n1;
	}

	n2hang(n2, &theta, &phi, x2, nx2, nphi2, nph2, p_offset);
	if (theta < -900. || phi < -900.)
		n1 = -1;
	else
		n1 = ang2n(theta, phi, x1, nx1, nphi1, nph1);

	return n1;
}

/* all pixels around given */
/* returns array of pixel numbers */
int *ang_around(theta, phi, x, nx, nphi, nph)
double theta, phi;
double *x;
int *nph;
int nx, nphi;
{
	int *a = (int *)0;
	int i_x, j_phi;

	angle2pixel(x, nx, nph, &i_x, &j_phi, theta, phi);

	return a;
}

int *pix_circle(theta, phi, r_ang)
double theta, phi;
double r_ang;      /* angle in radians */
{
	int *a = (int *)0;
	return a;
}

int *pix_box(theta, phi, dt, dp)
double theta, phi;
double dt, dp;
{
	int *a = (int *)0;
	return a;
}

void pix2sphere(n,nx,np,nph,xc, x,y,z)
int n;  /* input number of pixel */
int nx, np;
int *nph;
double *xc; /* array of cos theta_i */
double *x, *y, *z;
{
    double theta, phi;
    double r = 1.;

    n2ang(n, &theta, &phi, xc, nx, np, nph);

    *x = r * sin (theta) * cos (phi);
    *y = r * sin (theta) * sin (phi);
    *z = r * cos (theta);
}

int sphere2pix(x,y,z, nx,np,nph,xc)
int nx, np;
int *nph;
double *xc; /* array of cos theta_i */
double x, y, z;
{
    int n;        /* input number of pixel */
    double r = sqrt(x*x + y*y + z*z);
    double cos_theta, phi=0.0, theta;

    /*****************************************/
    /* convert (x,y,z) to angle (theta, phi) */
    /*****************************************/
    cos_theta = z / r;
    theta = acos ( cos_theta );

    if (ABS(x) > 1.e-30)   phi = atan2(y, x);
    if (phi < 0.)  phi += pi2;


    /* angle to pixel */
    n = ang2n(theta, phi, xc, nx, np, nph);

    return n;
}

int pix_eq2gal(alp,del, x, nx, np, nph)
double alp, del;
double *x;
int nx, np, *nph;
{
    int n;
    double lat, lon, theta, phi;
    double p = pi2/360.;
    extern void equat2gal(double, double, double *, double *);

    equat2gal(alp, del, &lat, &lon);

    theta = (90. - lat) * p;
    phi   = lon * p;

    /* angle to pixel */
    n = ang2n(theta, phi, x, nx, np, nph);

    return n;
}


int pix_ec2gal(bet,lam, x, nx, np, nph)
double bet, lam;
double *x;
int nx, np, *nph;
{
    int n;
    double alp, del, lat, lon, theta, phi;
    double p = pi2/360.;
    extern void equat2gal(double, double, double *, double *),
		eclip2equat(double, double, double *, double *);

    eclip2equat(bet, lam, &alp, &del);
    equat2gal(alp, del, &lat, &lon);

    theta = (90. - lat) * p;
    phi   = lon * p;

    /* angle to pixel */
    n = ang2n(theta, phi, x, nx, np, nph);

    return n;
}


