#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "glesp.h"

#define PR  fprintf
#define ER  stderr

#define IMPOSIBLE_VALUE     1234567890.0987654321

extern int x_locate(int, double *, double);
extern double line_pos (double, double, double, double, double);
extern int    impo(double);


/* select pixels in a zone  in a given area */
int *pix_in_sq_zone(x, nx, np, nph, t1, p1, t2, p2, t3, p3, t4, p4, n)
double *x;                              /* position of knots */
int nx, np, *nph;                       /* dimensions        */
double t1,p1, t2,p2, t3,p3, t4,p4;      /* limits of zone            */
int *n;                                 /* number of returned pixels */
{
    register int i, j;
    int *arr = (int *)0;                /* array or returned pixels  */
    int narr = 0;
    unsigned int nd=0;
    double t_min, t_max, p_min, p_max;
    double v, pi2=2*M_PI;
    double s = fabs(acos(x[nx/2])-acos(x[nx/2+1])); /* pixel size */
    int i_min, i_max, j_min, j_max;
    struct _tp_ {double t,p;} tp[4];

    /* step size */
    if (s > pi2/nph[nx/2])  s = pi2/nph[nx/2];
    s /= 2;  /* let it be half of pixel */

    /* simple bubble sorting by phi */
    tp[0].t = t1;  tp[1].t = t2;  tp[2].t = t3;  tp[3].t = t4;
    tp[0].p = p1;  tp[1].p = p2;  tp[2].p = p3;  tp[3].p = p4;
    for(i=0; i<3; i++)
	for(j=i+1; j<4; j++) {
	    if (tp[i].p > tp[j].p) {
		v = tp[i].p; tp[i].p = tp[j].p; tp[j].p = v;
		v = tp[i].t; tp[i].t = tp[j].t; tp[j].t = v;
	    }
	}
    p_min = tp[0].p;
    p_max = tp[3].p;

    /* limits by theta */
    if (t1 < t2) t_max = t2; else t_max = t1;
    if (t3 < t4) v = t4; else v = t3;
    t_max = t_max > v ? t_max : v;

    if (t1 > t2) t_min = t2; else t_min = t1;
    if (t3 > t4) v = t4; else v = t3;
    t_min = t_min < v ? t_max : v;

    i_min = x_locate(nx, x, cos(t_min));
    i_max = x_locate(nx, x, cos(t_max));

    /* calculate initial number of pixels in environment */
    for(i=i_min; i<=i_max; i++) {
	    double dp = pi2 / nph[i];
	    nd += (p_max - p_min) / dp;
    }
    if ((arr = (int *)calloc(2*nd, sizeof *arr)) == (int *)0) {
	PR(ER, "pix_in_zone: no memory for temporal arrays of %d elements\n",
		4*nd);
	exit(1);
    }

    /* change positions in sorted array */
    v = tp[3].p; tp[3].p = tp[2].p; tp[2].p = v; /* last point is the 3d */
    v = tp[3].t; tp[3].t = tp[2].t; tp[2].p = v; /*                      */

    if (tp[1].t > tp[2].t) {   /* the smallest theta is in highest angle */
	    v = tp[1].t; tp[1].t = tp[2].t; tp[2].t = v;
	    v = tp[1].p; tp[1].p = tp[2].p; tp[2].p = v;
    }

    /* select pixels whose centers inside the zone */
    /*               (t1,p1)
     *           r1  / \ r2             --------->
     *             /    \               |        phi
     *  (t0,p0)  /  *    / (t2,p2)      |
     *           \     /r3              |
     *         r4  \ /                  Y theta
     *            (t3,p3)
     */
    for (i=i_min; i<=i_max; i++) {
	double p = pi2/nph[i], t = acos(x[i]);
	for(j=j_min; j<=j_max; j++) {
	    double r1 = line_pos (tp[0].p, tp[0].t, tp[1].p, tp[1].t, p),
		   r2 = line_pos (tp[1].p, tp[1].t, tp[2].p, tp[2].t, p),
		   r3 = line_pos (tp[2].p, tp[2].t, tp[3].p, tp[3].t, p),
		   r4 = line_pos (tp[3].p, tp[3].t, tp[0].p, tp[0].t, p);

	    if ((r1 < t ||impo(r1)) && r2<t && (t>r3 || impo(3)) && t>r4) {
		    arr[narr] = ang2n(t, p, x, nx, np, nph);
		    narr++;
	    }
	}
    }

    *n = narr;
    return arr;
}

double line_pos (x1, y1, x2, y2, x)
double x1, y1, x2, y2, x;
{
    double y;
    if (fabs(x2-x1) > 1.e-20)
	    y = y1 + (y2-y1) / (x2-x1) * (x-x1);
    else y = IMPOSIBLE_VALUE;   /* x-s are the same */
    return y;
}

int impo(r)
double r;
{
    return fabs(r-IMPOSIBLE_VALUE) < 1.e-10 ? 1 : 0;
}


















