#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "glesp.h"
#include "glespol.h"

#define ABS(x)  ((x) < 0 ? -(x) : (x))
#define PR      fprintf
#define ER      stderr

extern void setopt(int, char **), usage(void), set_sig_fatal(void);
extern void eb2qu(double *, double *, double *, double *, int, int, int *,
	       double *, double *, int, int, double);
extern void sp2qu_n(double *, double *, double *, double *, double *,
	  double *, int, int, int *, double *, double *, double *,
	  int, int, int,
	  double, int *);

char alm_bfile[161], alm_efile[161], out_qfile[161], out_ufile[161];
char alm_file[161], out_tfile[161];
int  cut_lmin = -1;    /* cut value for a_lm        */
int  cut_lmax  = 0;    /* cut value for a_lm        */
int  phi_deriv = 0;
double eps = 1.e-20;
int  sym_grid = 1;    /* symmetric grid over x                         */
int  fill_0 = 0;      /* fill arrays by 0 till common length for FFT   */
HEAD hd, *ph;
int ffile_flag = 0;   /* output to FITS file                           */
//int lx_add = 'r';   /* cap pixels number in Pol-GLESP                */
int lx_add =  0;      /* cap pixels number in Pol-GLESP                */

char alm_hp_file[161]; /* file of alm, E- & B-modes in HEALPix format  */
char map3_file[161];   /* file with 3 maps                             */
double vnyq = 0.5;
int legendre_calc_scheme = 0;  /* if flag=-1, then direct calculations, */
			       /* else from largest M to lower          */
int N_pol1 = -1, N_pol2 = -1, N_pol3 = -1; /* number of pixels in P1,2,3 */

int lmax = -1;
int lmin = -1;
int nx= -1, np= -1;
double np_factor = -1.0;
char np_formula[121];

char *myname="polmap";

int main(argc, argv)
int argc;
char **argv;
{
    double *alm = (double *)0, *b_alm = (double *)0, *e_alm = (double *)0;
    double *x = (double *)0, *w = (double *)0, *wqu = (double *)0;
    int *nph = (int *)0, *mx = (int *)0;
    double *f = (double *)0, *q = (double *)0, *u = (double *)0;
    int nbuf, lmax0 = -1, lmax1 = lmax, lmax2 = lmax, nbuf0, nbuf1, nbuf2;
    FILE *fp_out = (FILE *)0;
    char x_type[40], grid_type[40];
    double x1 = -1.0, x2 = 1.0, phi2 = M_PI*2.;

    alm_bfile[0] = alm_efile[0] = out_qfile[0] = out_ufile[0] = '\0';
    alm_file[0] = out_tfile[0] = '\0';
    alm_hp_file[0] = map3_file[0] = '\0';
    np_formula[0] = '\0';

    /* select grid output keyword values */
    strcpy(x_type, "COS(THETA)");
    strcpy(grid_type, "PHI(X)+LEGENDRE");

    /* control signal interruption */
    set_sig_fatal();

    /* loop over external parameters */
    setopt(argc, argv);

    if (lmax > 0) {
	if (lmax < 100) {
	    if (nx <= 0)  nx = 201;
	    if (np <= 0)  np = 2 * nx;
	} else {
	    if (nx <= 0)  nx = lmax * 2 + 1;
	    if (np <= 0)  np = 2 * nx;
	}
    }

    if (*alm_hp_file != '\0') {
	/*  read 3-alm file */
	read_ieb_alm(&alm, &e_alm, &b_alm, &lmax, &nbuf, alm_hp_file);
	if (nx <= 0)  nx = lmax * 2 + 1;
	if (np <= 0)  np = 2 * nx;
    } else {
	int flag=0;
	int nalm;
	/* temperature if given */
	if (*alm_file != '\0') {
	    flag = test_glesp_map(alm_file);
	    switch(flag) {
	    case 0: /* F-format file */
		    lmax0 = get_lmax(alm_file);
		    break;
	    case 2: /* FITS alm-format file */
		    lmax0 = Fget_lmax(alm_file);
		    break;
	    case -1: /* ASCII format */
		    PR(ER, "%s file '%s' has bad format.\n",
			    myname, alm_file);
		    exit(1);
	    default :
	    case -3 : /* ASCII format */
		    lmax0 = Sget_lmax(alm_file);
		    break;
	    }
	    if (lmax0 <= 0)  *alm_file = '\0';
	}

	/* polarization */
	flag = test_glesp_map(alm_bfile);
	switch(flag) {
	case 0: /* F-format file */
		lmax1 = get_lmax(alm_bfile);
		break;
	case 2: /* FITS alm-format file */
		lmax1 = Fget_lmax(alm_bfile);
		break;
	case -1: /* ASCII format */
		PR(ER, "%s file '%s' has bad format.\n",
			myname, alm_bfile);
		exit(1);
	default :
	case -3 : /* ASCII format */
		lmax1 = Sget_lmax(alm_bfile);
		break;
	}


	flag = test_glesp_map(alm_efile);
	switch(flag) {
	case 0: /* F-format file */
		lmax2 = get_lmax(alm_efile);
		break;
	case 2: /* FITS alm-format file */
		lmax2 = Fget_lmax(alm_efile);
		break;
	case -1: /* ASCII format */
		PR(ER, "%s file '%s' has bad format.\n",
			myname, alm_efile);
		exit(1);
	default :
	case -3 : /* ASCII format */
		lmax2 = Sget_lmax(alm_efile);
		break;
	}

	nbuf1 = (lmax1 + 1) * 2;
	nbuf2 = (lmax2 + 1) * 2;
	if (lmax < 0) {
	    lmax = lmax1 < lmax2 ? lmax1 : lmax2;
	} else {
	    int l = lmax1 < lmax2 ? lmax1 : lmax2;
	    lmax = lmax < l ? lmax : l;
	}
	if (*alm_file != '\0') {
	    lmax = lmax < lmax0 ? lmax : lmax0;
	    nbuf0 = (lmax0 + 1) * 2;
	}

	if (nx <= 0)  nx = lmax * 2 + 1;
	if (np <= 0)  np = 2 * nx;
	nbuf = (lmax + 1) * 2;

	if (nx < 2*lmax+1) {
	    PR(ER,
	      "%s: resolution parameter nx=%d is less than 2*lmax+1=%d\n",
	      myname, nx, 2*lmax+1);
	    exit(1);
	}
	if (np < nx) {
	    PR(ER, "%s: resolution parameter np=%d is less than nx=%d\n",
		 myname, np, nx);
	    exit(1);
	}

	if((b_alm=(double*)calloc((unsigned)((lmax1+3)*nbuf1),sizeof *b_alm))
	    == (double *)0) {
nomem:
		PR(ER, "%s: no memory for arrays...\n", myname);
		exit(1);
	}

	nalm = lmax1 + 1;
	read_alm_table (alm_bfile, &b_alm, &nalm, &nbuf1, &lmax1);
	nalm = lmax2 + 1;
	read_alm_table (alm_efile, &e_alm, &nalm, &nbuf2, &lmax2);

	/* temperature if we need */
	if (*alm_file != '\0') {
	    if((alm=(double*)calloc((unsigned)((lmax2+3)*nbuf0),sizeof *alm))
		== (double *)0)
			goto nomem;
	    nalm = lmax0 + 1;
	    read_alm_table (alm_file, &alm, &nalm, &nbuf0, &lmax0);
	}

	if (lmax != lmax1)
	    alm2alm(&b_alm, &lmax1, &nbuf1, lmax, nbuf);
	if (lmax != lmax2)
	    alm2alm(&e_alm, &lmax2, &nbuf2, lmax, nbuf);
	if (*alm_file != '\0') {
	    if (lmax != lmax0)
		alm2alm(&alm, &lmax0, &nbuf0, lmax, nbuf);
	}
	lmax = lmax1 < lmax2 ? lmax1 : lmax2;
    }

    if (*map3_file != '\0') {
	if (alm == (double *)0) {
	  PR(ER,
	  "%s: can not contsruct file with 3 maps: no temperature data...\n",
	    myname);
	  PR(ER, "Use '-q' and '-u' flags for output\n");
	  exit(1);
	}
    }

    if ((x = (double *)calloc((unsigned)nx, sizeof *x)) == (double *)0)
		goto nomem;
    if ((w = (double *)calloc((unsigned)nx, sizeof *w)) == (double *)0)
		goto nomem;
    if ((mx = (int *)calloc((unsigned)nx, sizeof *mx)) == (int *)0)
		goto nomem;
    if ((wqu = (double *)calloc((unsigned)nx, sizeof *wqu)) == (double *)0)
		goto nomem;
    if ((nph=(int *)calloc((unsigned)nx, sizeof *nph))  == (int *)0)
		goto nomem;

    nbuf1 = nx * np * 2;  /* double arrays for polarization */
    if ((q = (double *)calloc((unsigned)nbuf1, sizeof *q)) == (double *)0)
		goto nomem;
    if ((u = (double *)calloc((unsigned)nbuf1, sizeof *u)) == (double *)0)
		goto nomem;
    if (*alm_file != '\0' || *alm_hp_file != '\0') {
	if((f = (double *)calloc((unsigned)nbuf1, sizeof *f)) == (double *)0)
		goto nomem;
    }

    /* coordinates and weights */
    {
	/* if c_flag >=0,  special grid for Legendre polynomials */
	int c_flag,         /* flag of coordinate grid selection */
	    j=lx_add;       /* addition to Nphi numbers          */
	int npp = nx*2;

	switch (lx_add) {
	case 0: /* pin = 0 => square matrix */
		c_flag = 0;  /* => npp = lx_add; lx = -20; */
		/* calc maximum nph */
		j = npp = np;
		coord_pol(c_flag, j, x1, x2, phi2, x, w, wqu, nx, nph, npp);
		npp = nph[(nx+1)/2];
		break;
	case 6:
		c_flag = 0;  /* => npp = lx_add; lx = -20; */
		break;
	case 4:
	case -10:  /* pin = 'o' */
		c_flag = 1;  /* lx=lx_add; old value:n0=4, near Pole */
		j = lx_add;
		break;
	case -11:
	case 'r': /* pin = 'r' */
		c_flag = 1;  /* lx=lx_add; old value:n0=4, near Pole */
		j = lx_add = -11;
		break;
	case -12: /* pin = 'r' + N_pol */
		c_flag = 1;  /* lx=lx_add; old value:n0=4, near Pole */
		j = lx_add = -12;
		break;
	default :
		if (lx_add < -10000) {
		   c_flag = 1; /* lx=lx_add; old value:n0=4, near Pole */
		   j = -(lx_add + 10000);
		} else if (lx_add < -1000) {
		   /* construct GLESP grid from polar cap */
		   c_flag = 0;  /* => npp = lx_add; lx = -20; */
		   /* number of pixels near Pole */
		   j = -(lx_add + 1000);
		} else {
		    c_flag = 2; /* polarization pixelization */
		}
		break;
	}

	/* 1st par - flag of coordinate grid selection */
	/* 2nd par - addition to Nphi numbers          */
	coord_pol(c_flag, j, x1, x2, phi2, x, w, wqu, nx, nph, npp);
	if (c_flag == 0)
	    np = nph[(nx+1)/2];
    }
    if (np_factor > 0.0)  {
	register int i;
	for(i=0; i<nx; i++) {
		nph[i] *= np_factor;
		if (nph[i] > np)   nph[i] = np;
	}
	weights_pol(0, 0, x1, x2, phi2, x, w, wqu, nx, nph, np);
    }
    if (np_formula[0] != '\0') {
	extern int nf_formu(double *, int, int, int *, char *);
	nf_formu(x, nx, np, nph, np_formula);
	weights_pol(0, 0, x1, x2, phi2, x, w, wqu, nx, nph, np);
    }
    if (fabs(np_factor+10.0) < 1.e-4) {
	/* healpix */
	extern int np_hpix(double *, int, int, int *);
	np_hpix(x, nx, np, nph);
	weights_pol(0, 0, x1, x2, phi2, x, w, wqu, nx, nph, np);
    }

//PR(ER, "np_factor=%g\n", np_factor);
//PR(ER, "nx=%d, np=%d, nph[nx/2]=%d\n", nx, np, nph[(nx+1)/2]);
    {
      register int i;
      for(i=0; i<nx; i++)  mx[i] = nph[i] * vnyq + 0.5;
    }

    /* calculation of Q & U maps */
    sp2qu_n(f, q, u, x, w, wqu, nx, np, nph,
	  alm, e_alm, b_alm, lmin, lmax, nbuf, eps, mx);

    /*-------------*/
    /* output data */
    /*-------------*/
    if (*map3_file != '\0') {   /* file with 3 maps */
	/* open output file with t,Q,U */
	if (map3_file[0] == '\0' || strcmp(map3_file, "stdout") == 0)
	    fp_out = stdout;
	else
	    if ((fp_out = fopen(map3_file, "w")) == (FILE *)0) {
		PR(ER,"%s: can not create file '%s'\n",myname,map3_file);
		exit(1);
	    }

	/*  write map  */
	outp_glesp_3map(f,q,u, x, nx, np, nph, fp_out,x_type,grid_type);

	/* close file */
	if(fp_out != (FILE *)0 && fp_out != stdout) fclose(fp_out);
    } else {
	if (q != (double *)0) {
	    /* open output file, Q */
	    if (out_qfile[0] == '\0' || strcmp(out_qfile, "stdout") == 0)
		fp_out = stdout;
	    else
		if ((fp_out = fopen(out_qfile, "w")) == (FILE *)0) {
		    PR(ER,"%s: can not create file '%s'\n",myname,out_qfile);
		    exit(1);
		}

	    /*  write map  */
	    if (ffile_flag) {
		/* map output in F-format */
		ff_write_ext2_map(q, x,nx,np,nph, fp_out, x_type, grid_type);
	    } else {
		/* map output in FITS-format */
		outp_glesp_map(q, x, nx, np, nph, fp_out, x_type, grid_type);
	    }

	    /* close file */
	    if(fp_out != (FILE *)0 && fp_out != stdout) fclose(fp_out);
	}

	if (u != (double *)0) {
	    /* open output file, U */
	    if (out_ufile[0] == '\0' || strcmp(out_ufile, "stdout") == 0)
		fp_out = stdout;
	    else
		if ((fp_out = fopen(out_ufile, "w")) == (FILE *)0) {
		    PR(ER,"%s: can not create file '%s'\n",myname,out_ufile);
		    exit(1);
		}

	    /*  write map  */
	    if (ffile_flag) {
		/* map output in F-format */
		ff_write_ext2_map(u, x,nx,np,nph,fp_out,x_type,grid_type);
	    } else {
		/* map output in FITS-format */
		outp_glesp_map(u, x, nx, np, nph, fp_out,x_type,grid_type);
	    }
	    /* close file */
	    if(fp_out != (FILE *)0 && fp_out != stdout) fclose(fp_out);
	}
	if (*alm_file != '\0' || *alm_hp_file != '\0') {
	    /* open output file, T */
	    if (out_tfile[0] == '\0' || strcmp(out_tfile, "stdout") == 0)
		fp_out = stdout;
	    else
		if ((fp_out = fopen(out_tfile, "w")) == (FILE *)0) {
		    PR(ER,"%s: can not create file '%s'\n",myname,out_tfile);
		    exit(1);
		}
	    /*  write map  */
	    if (ffile_flag) {
		/* map output in F-format */
		ff_write_ext2_map(f,x,nx,np,nph, fp_out, x_type, grid_type);
	    } else {
		/* map output in FITS-format */
		outp_glesp_map(f, x,nx, np,nph, fp_out, x_type, grid_type);
	    }
	    /* close file */
	    if(fp_out != (FILE *)0 && fp_out != stdout) fclose(fp_out);
	}
    }
    free((void *)nph);
    free((void *)x);
    free((void *)w);
    free((void *)wqu);

    free((void *)e_alm);
    free((void *)b_alm);
    free((void *)q);
    free((void *)u);

    if (*alm_file != '\0') {
	free((void *)alm);
	free((void *)f);
    }

    exit(0);
}

void setopt(argc, argv)
int argc;
char **argv;
{
    register int i;
    myname = argv[0];

    for(i=1; i<argc; i++) {
	switch(argv[i][0]) {
	case '-':
		switch(argv[i][1]) {
		case 'a':
			strcpy(alm_file, argv[++i]);
			break;
		case 'b':
			strcpy(alm_bfile, argv[++i]);
			break;
		case 'd':
			if (strncmp(argv[i]+1, "do", 2) == 0) {
				bitpix_map = DOUBLEPIX;
			} else {
				PR(ER,"%s: illegal flag '%s'\n",
				    myname, argv[i]);
				usage();
			}
			break;
		case 'f':
			if (strncmp(argv[i]+1, "fl", 2) == 0) {
				bitpix_map = FLOATPIX;
			} else {
				PR(ER,"%s: illegal flag '%s'\n",
				    myname, argv[i]);
				usage();
			}
			break;
		case 'e':
			if (strncmp(argv[i]+1, "ep", 2) == 0) {
			      /* accuracy for Legendre plynomials */
			      eps = atof(argv[++i]);
			} else {
			      strcpy(alm_efile, argv[++i]);
			}
			break;
		case 'h':
			usage();
			break;
		case 'l':
			switch(argv[i][2]) {
			case 'm':
			    if (strncmp(argv[i]+2,"mi",2) == 0)
				lmin = atoi(argv[++i]); /* Lmin */
			    else if (strncmp(argv[i]+2,"ma",2) == 0)
				lmax = atoi(argv[++i]); /* Lmax */
			    else {
				PR(ER,"%s: illegal flag '%s'\n",
				    myname, argv[i]);
				usage();
			    }
			}
			break;
		case 'L':
			lmax = lmin = atoi(argv[++i]);
			break;
		case 'n':
			switch(argv[i][2]) {
			case 'p':
				np = atoi(argv[++i]);
				break;
			case 'x':
				nx = atoi(argv[++i]);
				break;
			case 'q':
				vnyq = atof(argv[++i]);
				break;
			case 'f':
				i++;
				if (strncmp(argv[i], "hp", 2) == 0)
				    np_factor = -10;
				else
				    np_factor = atof(argv[i]);
				break;
			case 'F':
				strcpy(np_formula, argv[++i]);
				break;
			default:
				usage();
				break;
			}
			break;
		case 'o':
			switch(argv[i][2]) {
			case 'u':
				if(strncmp("ouf", argv[i]+1, 3) == 0) {
					ffile_flag = 1;
				}
				strcpy(out_ufile, argv[++i]);
				break;
			case 'q':
				if(strncmp("oqf", argv[i]+1, 3) == 0) {
					ffile_flag = 1;
				}
				strcpy(out_qfile, argv[++i]);
				break;
			case 't':
				if(strncmp("otf", argv[i]+1, 3) == 0) {
					ffile_flag = 1;
				}
				strcpy(out_tfile, argv[++i]);
				break;
			default:
				strcpy(map3_file, argv[++i]);
				break;
			}
			break;
		case 'p':
			if (strncmp(argv[i]+1, "pin", 3) == 0) {
				/* phi increment number :       */
				/* addition to Nphi numbers for */
				/* 'coord'-function             */
				int v = atoi(argv[++i]);
				if (argv[i][0] == 'o' ||
				    argv[i][0] == 'O' ) {
				     v = atoi(argv[i]+1);
				     if ( v > 0 )
					lx_add = -10000 - v;
				     else
					lx_add = -10;
				} else if (argv[i][0] == 'P')
				     /* polarization */
				     lx_add = v = 'P';
				else if (argv[i][0] == 'r')
				     lx_add = v = -11;
				else if (v < -1)
				     lx_add = -1000 - v;
				else lx_add = v;
			}
			break;
		case 'P':
			switch (argv[i][1]) {
			case 'P':
			default :
			    switch (argv[i][2]) {
			    default :
			    case '1':
				      N_pol1 = atoi(argv[++i]);
				      break;
			    case '2':
				      N_pol2 = atoi(argv[++i]);
				      break;
			    case '3':
				      N_pol3 = atoi(argv[++i]);
				      break;
			    }
			}
			lx_add = -12;
			break;
		case 't':
			strcpy(out_tfile, argv[++i]);
			break;
		case 'q':
			strcpy(out_qfile, argv[++i]);
			break;
		case 's':
			/* if flag=1, then direct calculations, */
			/* else from largest M to lower         */
			if (i < argc-1) {
			    if (argv[i+1][0] != '-') {
				i++;
				if(argv[i][0] == 'L' || argv[i][0] == '1') {
					legendre_calc_scheme = 1;
				} else
					legendre_calc_scheme = -1;
			    } else {
				legendre_calc_scheme = !legendre_calc_scheme;
			    }
			} else /* flag '-s' is the last parameter */
				legendre_calc_scheme = !legendre_calc_scheme;
			break;
		case 'u':
			strcpy(out_ufile, argv[++i]);
			break;
		case 'v':
			PR(ER, "%s: version 0.2\n", myname);
			exit(0);
		default:
			break;
		}
		break;
	case '+':
		switch(argv[i][1]) {
		case 'h':
		default:
			usage();
			break;
		}
		break;
	default:
		strcpy(alm_hp_file, argv[i]);
		break;
	}
    }

    return;
}

void usage()
{
PR(ER, "usage: %s [-b alm_bfile -e alm_efile] [-q out_qmap_file] [-u out_umap_file]\n",
myname);
PR(ER, "[-a alm_file] [-t out_temp_file] [3alm_file] [-o out_3map_file]\n");
PR(ER, "[-lmax lmax] [-nx nx] [-np np] [-pin {0,P,o,4}] [-nq nyq_coeff] [-ep eps]\n");
PR(ER, "[-{float,double}] [-s {M,L}] [-P{123} N{123}]\n");
PR(ER, "[-nf {np_factor,hpix}] [-nF np_formula]\n");
exit(2);
}
