#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <unistd.h>
#include <string.h>

#include "glesp.h"
#include "glespol.h"
#define PR      fprintf
#define ER      stderr

#define MOD(a,p)   ((a) - (p) * (int)((a)/(p)))
#define ABS(x)     ((x) < 0   ? -(x) : (x))
#define MAX(x,y)   ((x) > (y) ?  (x) : (y))
#define MIN(x,y)   ((x) < (y) ?  (x) : (y))


extern int lmin_ylm_hpix(int, double);
extern void start_time(void), print_time(void);
extern double return_time(void);
int _lmin_;
double _enorm_;
extern int max_double_(void);

extern int N_pol1, N_pol2, N_pol3;

extern void coord_pol(int, int, double, double, double, double *,
	      double *, double *, int, int *, int);
extern int pget_nphi_by_x(double *, int, int, int);

void coord_pol(coord_flag, lx_add, x1, x2, phi2, x, pw, pqu, Nx, Nph, Nphi)
int coord_flag;  /* flag of coordinate grid selection */
int lx_add;      /* addition to Nphi numbers          */
double x1, x2,
      phi2;
double *x, *pw, *pqu;
int    Nx, Nphi;
int *Nph;
{
	register int i;
	double xx;
	double dx=1.0, dphi;
	int NNphi = 0, Imx, lx;
	int npp = 0;

	if(coord_flag < 0) {
		/* uniform grid */
		dx = (x2-x1) / (Nx -1.0);
		xx = x1;
		for(i=0; i<Nx; i++) {
			pqu[i] = dx / 1.5; /* for Sympson integral */
			x[i] = xx;
			xx += dx;
		}
		pqu[0]    = pqu[0]    / 2.;
		pqu[Nx-1] = pqu[Nx-1] / 2.;

		for(i=1; i<Nx-1; i+=2) {
			pqu[i] += pqu[i];
		}
	} else {  /* special grid for Legendre polynomials */
		gauleg(x1, x2, x, pqu, Nx);
	}
	if( x[Nx-1] > 1.0)   x[Nx-1] = 1.0;

       /*
	*   arbitrary grid for sphere and fast-fourier
	*/
	Imx = (Nx+1)/2;
	if (coord_flag == 0) {
	    npp = lx_add;
	    lx = -20;
	} else if (coord_flag == 2) { /* polarization pixelization */
	    lx = 'P';
	} else {
	    lx = lx_add; /* old value:4, starting length, near Pole */
	}
	/* Nph[0] = Nphi; */
	switch (lx) {
	case 0:
		Nph[0] = Nphi;
		break;
	case -1:
		Nph[0] = 10;
		break;
	case -10: /* set standard Gauss-Legendre pixelization scheme */
	case 's': /* set standard Gauss-Legendre pixelization scheme */
	case 'o':
	case 't': /* as for temperature */
		Nph[0] = get_nphi_by_x2(x, Nx, Nphi, 0);
		lx = -10;
		break;
	case -11: /* set ~standard Gauss-Legendre pixelization scheme */
		Nph[0] = get_nphi_by_x2(x, Nx, Nphi, 0);
		// if (Nph[0] < 6) Nph[0] *= 2;
		if (Nph[0] < 9) Nph[0] = 9;
		break;
	case -12: /* set ~standard Gauss-Legendre pixelization scheme */
		Nph[0] = get_nphi_by_x2(x, Nx, Nphi, 0);
		if (N_pol1 > 0) Nph[0] = N_pol1;
		break;
	case -20: /* set pixelization from polar cap */
		Nph[0] = pget_nphi_by_x(x, Nx, npp, 0);
		break;
	case 'P': /* set pixelization for polarizaton scheme */
		Nph[0] = get_pol_nphi_by_x(x, Nx, Nphi, 0);
		break;
	default :
		Nph[0] = 10;
		break;
	}

	Nph[Nx-1] = Nph[0];
	NNphi = Nph[0] + Nph[Nx-1];
	dphi = phi2 / Nph[0];
	pw[0]    = dphi * pqu[0];
	pw[Nx-1] = dphi * pqu[Nx-1];

	for(i=1; i<Imx; i++) {
		int jx;
		/* Nph[i] = Nph[0]; */
		switch (lx) {
		case 0:
		       /*
			*   Number of cells over phi for a given x
			*   homogeneous grid for rectangular area
			*/
			Nph[i] = Nphi;
			break;
		case -11: /* set standard GLESP scheme */
			Nph[i] = get_nphi_by_x2(x, Nx, Nphi, i);
			if (i<5) Nph[i]+=2;
			break;
		case -12: /* set ~standard GLESP scheme */
			Nph[i] = get_nphi_by_x2(x, Nx, Nphi, i);
			if (i==1 && N_pol2 > 0) Nph[i] = N_pol2;
			else if (i==2 && N_pol3 > 0) Nph[i] = N_pol3;
			break;
		case -10: /* set standard GLESP scheme */
			Nph[i] = get_nphi_by_x2(x, Nx, Nphi, i);
			break;
		case -20: /* set pixelization from polar cap */
			Nph[i] = pget_nphi_by_x(x, Nx, npp, i);
			break;
		case -1:
			Nph[i] = Nph[i-1] + lx;
			break;
		case 'P':
			Nph[i] = get_pol_nphi_by_x(x, Nx, Nphi, i);
			break;
		default :
			Nph[i] = Nph[i-1] + lx;
			break;
		}

		if(Nph[i] > Nphi) Nph[i] = Nphi;
		dphi  = phi2 / Nph[i];
		pw[i] = dphi * pqu[i];
		NNphi += Nph[i];

		jx = Nx - i - 1;
		if(jx > i) {
			pw[jx]  = dphi * pqu[jx];
			Nph[jx] = Nph[i];
			NNphi += Nph[jx];
		}
	}

	return;
}

void weights_pol(coord_flag, lx_add, x1, x2, phi2, x, pw, pqu, Nx, Nph, Nphi)
int coord_flag;  /* flag of coordinate grid selection */
int lx_add;      /* addition to Nphi numbers          */
double x1, x2,
      phi2;
double *x,  *pw, *pqu;
int    Nx, Nphi;
int *Nph;
{
    register int i;

    gauleg(x1,x2,x,pqu,Nx);
/*
C   Number of cells over phi for a given x
c   homogeneous grid for rectangular area
*/

    for (i=0; i< Nx; i++)
	pw[i] =pqu[i]*phi2/Nph[i];

    return;
}

/*
 *  measure user and system time for calculation
 */
#include <sys/times.h>
#include <time.h>
static clock_t ut_0, ut_1;
static clock_t st_0, st_1;
static long tic_sec;
void start_time()
{
    struct tms t;
    times(&t);                      /* get time in ticks    */
    ut_0 = ut_1 = t.tms_utime;
    st_0 = st_1 = t.tms_stime;
    tic_sec = sysconf(_SC_CLK_TCK); /* get ticks per second */

    return;
}
void print_time()
{
    struct tms t;
    clock_t ut_2;
    clock_t st_2;

    times(&t);                      /* get time in ticks    */
    ut_2 = t.tms_utime;
    st_2 = t.tms_stime;

    PR(ER, "ut0=%g, st0=%g, ut1=%g, st1=%g\n",
	((double)(ut_2-ut_0)/tic_sec),
	((double)(st_2-st_0)/tic_sec),
	((double)(ut_2-ut_1)/tic_sec),
	((double)(st_2-st_1)/tic_sec)
    );
    ut_1 = ut_2;
    st_1 = st_2;

    return;
}
double return_time()
{
    struct tms t;
    clock_t ut_2;
    clock_t st_2;
    double tt;

    times(&t);                      /* get time in ticks    */
    ut_2 = t.tms_utime;
    st_2 = t.tms_stime;
    tt = (double)(ut_2-ut_1)/tic_sec;

    ut_1 = ut_2;
    st_1 = st_2;

    return tt;
}


#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define PR  fprintf
#define ER  stderr
#define ABS(x)     ((x) < 0 ? -(x) : (x))
#define MOD(a,p)   ((a) - (p) * (int)((a)/(p)))

/*
C   Legendre functions F^m_k(x), k=m,m+1,...,Lmx
      DIMENSION glm(Lmx+1),c(2*Lmx+1)
c   Lim=Log(eps)< 0, ~ -20
*/
void lgndr_Mn(Mlim, En, Lmx,mv,cmm,c,x_i,glm,NTlm)
int *Mlim, *Lmx, *mv;
double *En, *x_i;
double *cmm, *c, *glm;
int *NTlm;
{
    register int m= *mv, m1 = m + 1, m2 = m + 2;
    register int k, Imax=0, lmx= *Lmx;
    double x = *x_i, Enorm = *En;
    double dfm = 0.5 * m * log10(1.-x*x + 1.e-50);
    double *gm1 = glm + m1, *c2 = c + lmx;

    if(ABS(dfm) > *Mlim)  Imax = -dfm / *Mlim;

    dfm += Imax * *Mlim;
    *gm1 = *cmm * pow(10., dfm);          /* P_m^m */

    (*NTlm)++;
    if(m == lmx) {
	while(Imax > 0) { *gm1 /= Enorm; Imax--; }
	return;
    }

    glm[m2] = x * c[m1] * *gm1;    /* P_m+1^m  */
    (*NTlm)++;

    if(m1 == lmx) {
	while(Imax>0) { *gm1 /= Enorm; glm[m2]/=Enorm; Imax--;}
	return;
    }

    for(k = m+2; k<= lmx; k++) {  /* P_m+l^m */
	register int kk, k1=k+1;
	register double *ck = c+k, *c2k = c2 + k,  *g = glm + k;

	(*NTlm)++;
	/* recursional relation */
//      glm[k1] = x *  c[k] * glm[k] - c[k+ *Lmx] * glm[k-1];
	*(g+1) = x *  *ck  *  *g - *c2k  *  *(g-1);

	if(ABS(glm[k1]) > Enorm) {
	       for(kk=m1; kk <= k1; kk++) {
			glm[kk] /= Enorm;
			if(Imax >= 2) glm[kk] /= Enorm;
	       }
	       if(Imax >= 2) Imax -= 2; else Imax--;
	}
    }
    if(Imax >= 1) {
	    register int i, l1 = lmx + 1;
	    for(i=1; i<=Imax; i++) {
		    for(k= m+1; k <= l1; k++)
			    glm[k] /= Enorm;
	    }
    }

    return;
}

/*
 c       SUBROUTINE Plm(Jmx,Lmx,L0,m,e,ax,x,glm,Kxf,xlbd,NPlm)
 C   Spin-Legeandr functions lbd^m_k(x), k=m,m+1,...,Kmx
	DIMENSION glm(Lmx+1),xlbd(2*Jmx+3),cpM(4*Jmx,Jmx+2)
*/
void Plm_n(Jmx,Lmx,L0,m,cpM,ax,x,glm,Kxf,xlbd,NPlm)
int Jmx,Lmx,L0,m,Kxf, *NPlm;
int *ax;
double *cpM, x, *glm, *xlbd;
{
    register int l;
    register double dfx=1./(1.-x*x+1.e-20);
    int ll=Jmx+1;
    int m1=m+1, mm =Jmx+2;
    register double sx, sy, mdfx = m * dfx;

//    for (l=L0; l<=lmx1; l++) {
    for (l=L0; l<=Lmx; l++) {
	register int i=4*l-3, j, imm = i * mm + m1;
	register double gl1 = glm[l+1], gl = glm[l], cgl = cpM[imm]*gl;
	(*NPlm)++;

	sx = cgl*x  + cpM[imm+mm]   * gl1;
	sx = sx*dfx - cpM[imm+3*mm] * gl1;                  /* F_lm */
	sy = cgl  - x*cpM[imm+mm+mm]* gl1;
	sy = mdfx * sy;                                     /* Phi_lm */

	j=ll+l;
	if(Kxf > 0) {
	  xlbd[l] = sx+sy;                  /* F_lm+Phi_lm=lbd^+ */
	  if (*ax > 0)
		xlbd[j] = sx-sy;           /* F_lm-Phi_lm=lbd^- */
	  else
		xlbd[j] = sy-sx;           /* F_lm-Phi_lm=lbd^- */
	  *ax = -(*ax);
	} else {
	   /* if(Kxf < 0) { */
	   xlbd[l] = sx;              /* F_lm              */
	   xlbd[j] = sy;              /* Phi_lm            */
	}
    }
    return;
}

/* Kxf > 0 */
void Plm_n1b(Jmx,Lmx,L0,m,cpM,ax,x,glm,xlbd,NPlm)
int Jmx,Lmx,L0,m, *NPlm;
int *ax;
register double *cpM, x, *glm, *xlbd;
{
    register int m1=m+1, mm =Jmx+2;
    register int l=L0, imm = (4*l-3) * mm + m1;
    int mm2 = mm+mm, m4 = mm2 + mm2;
    register double dfx=1./(1.-x*x + 1.e-20);
    double sx, sy, mdfx = m * dfx,
	     *cpmm = cpM + mm, *xlb = xlbd + Jmx+1;

    for (l=L0; l<=Lmx; l++) {
	register double gl1 = glm[l+1], gl = glm[l], cgl = cpM[imm]*gl;
	register double *cpi = cpmm + imm;
	register double *xj  = xlb + l;
	(*NPlm)++;

	sx = cgl*x  +  *cpi       * gl1;
	sx = sx*dfx -  *(cpi+mm2) * gl1;                  /* F_lm */
	sy = (cgl - x* *(cpi+mm)  * gl1) * mdfx;          /* Phi_lm */

	/* if(Kxf > 0) {  */
	xlbd[l] = sx+sy;             /* F_lm+Phi_lm=lbd^+ */
	if (*ax > 0)
	      *xj = sx-sy;           /* F_lm-Phi_lm=lbd^- */
	else
	      *xj = sy-sx;           /* F_lm-Phi_lm=lbd^- */

	*ax = -(*ax);

	imm += m4;
    }
    return;
}

/* Kxf > 0 */
void Plm_n1(Jmx,Lmx,L0,m,cpM,ax,x,glm,xlbd,NPlm)
int Jmx,Lmx,L0,m, *NPlm;
int *ax;
register double *cpM, x, *glm, *xlbd;
{
    register int l; // lmx1=Lmx+1;
    register double dfx=1./(1.-x*x + 1.e-20);
    register int m1=m+1, mm =Jmx+2;
    register double sx, sy, mdfx = m * dfx,
	     *cpmm = cpM + mm, *xlb = xlbd + Jmx+1;

//    for (l=L0; l<=lmx1; l++) {
    for (l=L0; l<=Lmx; l++) {
	register int i=4*l-3, imm = i * mm + m1;
	register double gl1 = glm[l+1], gl = glm[l], cgl = cpM[imm]*gl;
	register double *cpi = cpmm + imm;
	register double *xj  = xlb + l;
	(*NPlm)++;

	sx = cgl*x  +  *cpi         * gl1;
	sx = sx*dfx -  *(cpi+mm+mm) * gl1;                  /* F_lm */
	sy = cgl  - x* *(cpi+mm)    * gl1;
	sy = mdfx * sy;                                     /* Phi_lm */

//PR(ER, "l=%d, m=%d, x=%f, ax=%d\n", l, m, x, *ax);
	/* if(Kxf > 0) {  */
	xlbd[l] = sx+sy;                  /* F_lm+Phi_lm=lbd^+ */
	if (*ax > 0)
	      *xj = sx-sy;           /* F_lm-Phi_lm=lbd^- */
	else
	      *xj = sy-sx;           /* F_lm-Phi_lm=lbd^- */

	*ax = -(*ax);
    }
    return;
}

/* Kxf<0: direct spin-weighted harmonic calculation */
void Plm_n2(Jmx,Lmx,L0,m,cpM,x,glm,xlbd, NPlm)
int Jmx,Lmx,L0,m;
double *cpM, x, *glm, *xlbd;
int *NPlm;
{
    register int m1=m+1, mm =Jmx+2, mm2=mm+mm;
    register int l=L0, imm = (4*l-3) * mm + m1;
    int m4 = mm2 + mm2;
    double dfx=1./(1.-x*x + 1.e-20);
    double sx, sy, mdfx = m * dfx,
	    *cpmm = cpM + mm;
    double  *xlb = xlbd + Jmx+1;

    for (l=L0; l<=Lmx; l++) {
	double gl1 = glm[l+1], gl = glm[l], cgl = cpM[imm]*gl;
	double *cpi = cpmm + imm;
	(*NPlm)++;

	sx = cgl*x  +     *cpi      * gl1;
	sx = sx*dfx -    *(cpi+mm2) * gl1;                /* F_lm   */
	sy = (cgl  - x * *(cpi+mm)  * gl1) * mdfx;        /* Phi_lm */

	/* if(Kxf < 0) { */
	*(xlbd+l) = sx;              /* F_lm    */
	*(xlb +l) = sy;              /* Phi_lm  */

	imm += m4;
    }
    return;
}


/*
C   Legendre polynomials F^0_l(x)
      IMPLICIT REAL*8(a-h,o-y)
      DIMENSION glm0(4*Jmx+4)
*/
void lgndr_0(Jmx,Lmx,x,glm0)
register int Jmx, Lmx;
double x, *glm0;
{
    double x2 = x*x;
    register int i,j,k;
    int j2 = 2*Jmx;

    glm0[Jmx+1] = glm0[j2+3];                  /* P_0 */
    glm0[Jmx+2] = glm0[j2+4]*x;                /* P_1 */
    glm0[Jmx+3] = glm0[j2+5]*(3.*x2-1);        /* P_2 */
    glm0[Jmx+4] = glm0[j2+6]*x*(5.0*x2-3.0);   /* P_3 */
    k=j2+6;

    for(i=4; i<=Lmx; i++) {
       j=Jmx+i+1;
       k += 2;
       glm0[j] = x * glm0[k-1] * glm0[j-1] - glm0[k] * glm0[j-2];
    }
    return;
}

/*
 c       Legendre function F_m^m
 */
double F_mm(m,x,sqx)
register int m;
double x, sqx;
{
    register int k;
    register double cmm=.5, f;
    register int m2 = 2 * m;

    for(k=2; k<=m2; k+=2) cmm *= (double)(k+1)/k;
    cmm = (1-2*MOD(m,2)) * sqrt(cmm);
    f = cmm * pow(sqx, (double)m);
    return f;
}

/*
ccc    gam=abc(1:n),a=abc(n+1:2n),b=abc(2n+1:3n),
ccc                 c=abc(3n+1:4n)
        dimension abc(4*n),u(n)
*/
void  tridag2_(abc,u,n,r1,rn)
double *abc,    /* numerated from 0 */
       *u;      /* here 'u' is 'glm' numerated from 1 to Lmax */
int n;
double r1, rn;
{
	register int j;
	double bet;
	int na = n,
	    nb = n + n,
	    nc = nb + n;

	if ( ABS(abc[nb]) < 1.e-50)  {
		PR(ER, "tridag: rewrite equations or ");
		PR(ER, "use even number of rings...\n");
		exit(1);
	}
	bet = abc[nb];
	u[1] = r1 / bet;
	for(j=2; j <= n; j++) {
		abc[j-1] = abc[nc+j-2] / bet;
		bet = abc[nb+j-1] - abc[na+j-1] * abc[j-1];
		if(ABS(bet) < 1.e-50) {
			PR(ER, "tridag: failed...\n");
			exit(1);
		}
		u[j] = -abc[na+j-1] * u[j-1] / bet;
	}
	u[n] = (rn - abc[na+ n -1] * u[n-1]) / bet;
	for(j= n-1; j>=1; j--)  u[j] -= abc[j] * u[j+1];

	return;
}


/*
C   coefficients abc 
      DIMENSION cc(4*Jmx+4)
*/
void start_L0(Jmx,Lmn,Lmx,cc)
int Jmx, Lmn, Lmx;
double *cc;
{
    register int i, j, j2 = 2 * Jmx;

    cc[j2+3] = sqrt(0.5);                /* P_0^0 */
    cc[j2+4] = sqrt(1.5);                /* P_1^0 */
    cc[j2+5] = 0.5*sqrt(2.5);            /* P_2^0 */
    cc[j2+6] = 0.5*sqrt(3.5);            /* P_3   */
    j=j2+6;

    for(i=4; i<= Lmx; i++) {
       double eps = 1. - 1./i;                 /* i=l */
       double eps1 = i+0.5;
       double eps2 = sqrt(eps1/(eps1-1.));
       eps1 = sqrt(eps1/(eps1-2.));
       j++;
       cc[j] = (1.+eps)*eps2;
       j++;
       cc[j] = eps*eps1;
    }

    return;
}


/*
C   coefficients abc 
      IMPLICIT REAL*8(a-h,o-y)
      DIMENSION abc(4*Jmx),cc(4*Jmx+4),acL(2*Jmx*(Jmx+1))

C     abc(n+i)=a,abc(2n+i)=b,abc(3n+i)=c,
C   parameters of the matrix
*/
void  start_ll(Jmx,Lmx,l,abc,cc,acL,cll)
int Jmx,Lmx, l;
double *abc, *cc, *acL, *cll;
{
    register int m;
    int n = l-1, n1 = n -1;
    int nc = 3 * n;
    int ik = l*(l+1);

    abc[n]=0;                        /* a(1)=0 */
    ik+=2;

    cc[1]=acL[ik];                     /* b(1)  */
    abc[nc]=1.0;                     /* c(1)=1 */

    for (m=1; m<=n1; m++) {
       ik++;
       abc[n+m] = acL[ik];          /* a(m)   */
       ik++;
       cc[m+1] = acL[ik];             /* b(m)   */
       abc[nc+m] = 1.0;             /* c(m)=1 */
    }
    abc[nc+n-1]=0.0;                  /* c(L)=0 */
    *cll = 1.0 / sqrt((1.0-1.0/l)*(1.0+1.0/(1+l)));

    return;
}

/*
C   Legendre functions F^m_L(x), m=1,2,...,L
      IMPLICIT REAL*8(a-h,o-y)
      DIMENSION glm(Lmx+1),abc(4*Lmx),cc(3*Jmx+3)
*/
void lgndr_Ln(Jmx,Lmx,l,x,sqx,abc,cc,glm,cll,NT)
int Jmx, Lmx, l;
double x, sqx, *abc, *cc, *glm, *cll;
int *NT;
{
    register int i, m1;

    if(l == 2) {
       glm[1] =  cc[Jmx+1+l];
       glm[2] = -sqrt(3.75)*x*sqx;
       glm[3] =  sqrt(0.9375)*sqx*sqx;
       return;
    }

    if(l > 2) {
	int n=l-1;
	int nb = 2*n;
	double xs = 2.*x/sqx;

	/*   abc(n+i)=a,abc(2n+i)=b,abc(3n+i)=c,  */
	/*   parameters of the matrix             */
	double r1 = -cc[Jmx+1+l] * *cll;
	double rf = -F_mm(l,x,sqx);
	abc[nb] = xs * cc[1];                    /* b(1)=0 */
	for(m1=2; m1<=n; m1++)
		 abc[nb+m1-1] = xs * cc[m1];     /* b(m)   */

	tridag2_(abc,glm,n,r1,rf);
	glm[l+1] = -rf;
//      (*NT) += l+1;

	for (i=l; i>=2; i--)    glm[i] = glm[i-1];
	glm[1] = cc[Jmx+l+1];
    }
    return;
}

/*
 C   Spin-Legendre functions lbd^m_k(x), m=k,k+1,...,Lmx
      IMPLICIT REAL*8(a-h,o-y)
      DIMENSION glm(Jmx+1),xlbd(2*Jmx+3),cpL(3*Jmx*(Jmx+1))
 c      DIMENSION glm(Jmx+1),xlbd(2*Jmx+3),e(3*Jmx+3)
 c   glm - f_l^m for 1<m<L
 */
void  Pll_n(Jmx,Kkm,l,cpL,ax,x,dsx,glm,Kxf,xlbd,NP)
int Jmx, Kkm, l, Kxf, *NP;
double *cpL, x,dsx, *glm, *xlbd;
int ax;
{
    int kl=3*l*(l-1);
    int ll = Jmx+1;
    double dfx=1.0/dsx/dsx, s1, s2, sx, sy;
    register int i, j, m, m1, km;

    for(m1=1; m1<=Kkm; m1++) {   /* 1<m<L */
       register double gl1 = glm[m1];
       (*NP)++;
       m=m1-1;
       i=m+m+m1;
       km = kl+i;
       s1=cpL[km] * dfx*gl1;
       s2=cpL[km+2] / dsx*glm[m1+1];
       sx =  s1-x*s2-cpL[km+1]*gl1;          /* F_lm */
       sy = -s1*x-m*s2;

       if(Kxf > 0) {
	    j=Jmx+m1;
	    xlbd[m1] = sx+sy;              /* F_lm+Phi_lm=lbd^+ */
	    xlbd[j] = ax * (sx-sy);        /* F_lm-Phi_lm=lbd^- */
	    ax = -ax;
       }
       if(Kxf < 0) {
	  j=ll+m1;
	  xlbd[m1] = sx;               /* F_lm(m+1)         */
	  xlbd[j]  = sy;               /* Phi_lm(Jmx+1+m1)  */
       }
    }

    return;
}

void start_clm(Lmx,m_i,cmm,c)
int *Lmx;
int *m_i;
double *cmm,
       *c;
{
    register int k, k1, k2, km2;
    register int m = *m_i, mm = m+m, m2=m*m;

    *cmm = 0.5;
    if(m > 0) {
	    for(k=2; k<=mm; k+=2)  *cmm *= (double)(k+1)/k;
    }
    *cmm = (1-2*MOD(m,2)) * sqrt(*cmm);

    k = m+1;
    k2 = k*k;
    c[k] = sqrt((double)(4*k2-1)/(k2-m2));
    c[*Lmx+k] = 0.;

    if(m != *Lmx) {
       for(k=m+2; k <= *Lmx; k++) {
	  double *ck = c + k;
	  k1 = k-1; k2 = k*k; km2 = k2 - m2;
	  *ck = sqrt((double)(4*k2-1)/km2);
	  *(ck + *Lmx) = sqrt((double)(k+k+1)*(k1*k1-m2)/(double)(k+k-3)/km2);
       }
    }
    return;
}

#include <stdio.h>
#include <math.h>

#define PR  fprintf
#define ER  stderr
/*
      normalized coefficients for spin-weighted functions
      both M & L versions
      double cpM(4*Jmx,Jmx+2),cpL(3*Jmx*(Jmx+1))
 */

void norma_M(Jmx,Lmn,Lmx,cpL,cpM)
int *Jmx, *Lmn, *Lmx;
register double *cpM, *cpL;
{
    register int l, m;
    register double cmm = 0.5;
    register int mm = *Jmx + 2, mm2 = mm+mm;

    /* for cpL */
    for (l= *Lmn; l<= *Lmx; l++) {
	register int l1=l-1, kl=3*l*l1,
		     i=4*l-3, im = i * mm,
		     l2 = l*l;
	register double ss = (l+2)*l*(double)(l2-1);
	register double sl0 = (l+cmm)/(l-cmm), sl1, sl2;
	ss = 1.0 / sqrt(ss);
	sl1 = l1 * ss;
	sl2 = l*l1*ss*cmm;

	for (m=0; m<=l; m++) {  /* m=0<m<L */
	    register int km=kl+3*m, m2 = m*m,
			 imm = im + m + 1;
	    register double sx;
	    register double *cpm_im = cpM + imm + mm;
	    register double *cpl_km = cpL + km + 1;

	    *cpl_km = (m2-m)*ss;
	    *(cpl_km+1) = cmm*(l2+l-m-m)*ss;
	    sx= (l+m+1) * (l-m);
	    *(cpl_km+2) = sqrt(sx)*ss;
	    sx = sl0 * (l2-m2);

	    *(cpm_im-mm)   = sqrt(sx) * ss;
	    *cpm_im        = (m2-l) * ss;
	    *(cpm_im+mm)   = sl1;
	    *(cpm_im+mm2)  = sl2;
	}
    }

    return;
}

/*
	dimension acL(2*Jmx*(Jmx+1))
c	preparation of coefficients for recurrent 
c	relation for Legender functions P_L^M, L-version
*/

void start_Labc(Jmx,Lmx,Lmn,acL)
int Jmx,Lmx,Lmn;
double *acL;
{
    register int l,m;
    register int lmx1 = Lmx + 1;

    for (l= Lmn; l<=lmx1; l++) {
	register int i = l*(l+1), ll = l-2;
	for (m=0; m<=ll; m++) {
	    double e1 = sqrt((double)((l-m-1)*(l+m+2)));
	    i++;
	    acL[i] = sqrt((l-m+1.e-20)*(l+m+1)) / e1;
	    i++;
	    acL[i] = (m+1) / e1;
	}
    }
    return;
}

/*
 C     normalized coefficients for spin-weighted functions
 C     both M & L versions (cpM & cpL)

       DIMENSION cpM(4*Jmx,Jmx+2),cpL(3*Jmx*(Jmx+1))
 */
void norma_L(Jmx,Lmn,Lmx,cpL,cpM)
int *Jmx, *Lmn, *Lmx;
double *cpL, *cpM;
{
    register int l, m;
    double  cmm=0.5;
    int mm = *Jmx + 2;

    for (l= *Lmn; l<= *Lmx; l++) {
	register int i=4*l-3, im = i*mm;
	register int l2 = l * l;
	register int kl=3*l*(l-1);
	double ss = (double)(l+2)*l*(l2-1);
	ss=1./ sqrt(ss);

	for (m=0; m <= l; m++) {
	    register int km=kl+3*m;
	    register int m2 = m * m;
	    register int imm = im + m + 1;
	    double sx = (l+cmm)/(l-cmm)*(l2-m2);

	    cpM[imm]       = sqrt(sx)*ss;
	    cpM[imm+mm]    = (m2-l)*ss;
	    cpM[imm+mm+mm] = (l-1)*ss;
	    cpM[imm+3*mm]  = cmm*l*(l-1)*ss;

	    cpL[km+1] = m*(m-1)*ss;             /* e(i)   */
	    cpL[km+2] = cmm*(l2+l-2*m)*ss;      /* e(i+1) */
	    sx = (l+m+1) * (l-m);
	    cpL[km+3] = sqrt(sx)*ss;            /* e(i+2) */
	}
    }
    return;
}


extern double calc_nph(double, double, int, int);
extern int start_calc(char *);

int  nf_formu(x, nx, np, nph, fstr)
double *x;
int nx, np, *nph;
char *fstr;
{
    register int i;
    int l = 0;
    char str[181];

    sprintf(str, "x=0.0; p=0.0; t=0.0; i=0.0; l=0.0; %s",  fstr);
    start_calc(str);

    for(i=0; i<nx; i++) {
	double t = acos(x[i]);
	nph[i] = calc_nph(x[i],t,i,l);
    }

    return 1;
}

int  np_hpix(x, nx, np, nph)
double *x;
int nx, np, *nph;
{
    register int i;
    int nside = (nx + 1) / 4, nx2 = nx/2;
    double fc2 = 2./3./nside;
    double cth = (nx2 - nside) * fc2;
    int nf = 4.*np/nx + .5;

//PR(ER, "nside=%d, nx=%d, np=%d, nf=%d\n", nside, nx, np, nf);
    for(i=0; i<nx2; i++) {
	int i1 = i+1;
	nph[i] = nf * (i1 < nside ? i1 : nside);
	if (i==0)  nph[i] = 12;
	if (i==1)  nph[i] = 18;
	if (i==2)  nph[i] = 24;

	if (x[i] < 0 && x[i] > -cth)
		nph[i] = np;
	if (x[i] > 0 && x[i] < cth)
		nph[i] = np;
//        if (nph[i] > np)  nph[i] = np;
	nph[nx-1-i] = nph[i];
//PR(ER, "i=%d, ik=%d, n=%d, x[i]=%g, cth=%g\n", i, nx-1-i, nph[i], x[i], cth);
    }

    return 1;
}


/*
 * returns minimal order l at which to keep Ylm
 * |Ylm| < eps * Y00 ==>
 * m_cut(theta, l) = theta * l * e / 2 + | ln(eps)| + ln(l)/2
 * if eps = 1.e-15 and l < 1.e4
 * m_cut(theta, l) = theta * l * 1.35 + 40
 * the choice of 1.35 (or larger)
 * also insures that the equatorial rings will have all their Ylm's computed
 * default parameters are HPX_MXL0 = 40 and HPX_MXL1 = 1.35_DP
 *
 */
#define HPX_MXL0 40
#define HPX_MXL1 1.35
int l_min_ylm(m, x)
int m;
double x;
{
    int lmin = m;

    if (HPX_MXL0 > 0) {
	int ll = (int)((m - HPX_MXL0)/(HPX_MXL1 * sqrt(1-x*x)));
	lmin = MAX(lmin, ll);
    }

    return lmin;
}
