#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <unistd.h>

#include "glesp.h"
#include "glespol.h"

/*
 * polmap:
 * calculates anisotropy temperature and Q,U-polarization maps
 * by {a,e,b}-lm coefficients or by C_TT, C_EE, C_BB spectra with simulations
 *
 *
 * Last modification by V.O. 09/2015
 *
 */
#define ABS(x)  ((x) < 0 ? -(x) : (x))
#define PR      fprintf
#define ER      stderr
#define ST      stdout

extern void setopt(int, char **),usage(void),usage2(void),set_sig_fatal(void);
extern void eb2qu(double *, double *, double *, double *, int, int, int *,
	       double *, double *, int, int, double);
extern int sp2qu_n(double *, double *, double *, double *, double *,
	  double *, int, int, int *, double *, double *, double *,
	  int, int, int,
	  double, double, int *);

char alm_bfile[161], alm_efile[161], out_qfile[161], out_ufile[161];
char alm_file[161],  out_tfile[161];
char alm_out_file[161];/* alm out file when simulation */
char elm_out_file[161];/* elm out file when simulation */
char blm_out_file[161];/* blm out file when simulation */
char alm_str[81];      /* string for input of alm      */
char blm_str[81], elm_str[81];  /* input of blm & elm  */
int alm_input =  0;    /*1:alm from file,2:alm from str*/
int blm_input =  0;    /*1:blm from file,2:blm from str*/
int elm_input =  0;    /*1:elm from file,2:elm from str*/
char file_cl[161];     /* file of Cl for simulation    */
int cl_input = 0;      /* flag to show that Cl-input   */
char file_cle[161];    /* file of C_EE for simulation  */
char file_clb[161];    /* file of Cl_BB for simulation */
int clp_input = 0;     /* flag to show that Cpol-input */
int random_seed = 0;   /* random seed for simulation   */
int random_seed_e = 0;
int random_seed_b = 0;
int seed_flag = 0;     /* flag to use the user seed    */
double fwhm = -1.;     /* if > 0, the smoothing beam in arcmin */

int  cut_lmin = -1;    /* cut value for a_lm           */
int  cut_lmax  = 0;    /* cut value for a_lm           */
int  phi_deriv = 0;
double eps = 1.e-20;
double eps1 = 1.e-20,
       eps2 = 1.e-20;
//double eps1 = 1.e-12,
//       eps2 = 1.e-6;
double vnyq = 0.5;
int  sym_grid = 1;    /* symmetric grid over x                         */
int  fill_0 = 0;      /* fill arrays by 0 till common length for FFT   */
HEAD hd, *ph;
int ffile_flag = 0;   /* output to FITS file                           */
//int lx_add = 'r';   /* cap pixels number in Pol-GLESP                */
int lx_add =  0;      /* cap pixels number in Pol-GLESP                */

char alm_hp_file[161]; /* file of alm, E- & B-modes in HEALPix format  */
char map3_file[161];   /* file with 3 maps                             */
int legendre_calc_scheme = 0;  /* if flag=-1, then direct calculations, */
			       /* else from largest M to lower          */
int N_pol1 = -1, N_pol2 = -1, N_pol3 = -1; /* number of pixels in P1,2,3*/
int no_pol = 0,                /* if flag = 1, no polarization calc-s   */
    no_ani = 0,                /* if flag = 1, no anisotropy calc-s     */
    memory_separation=0;       /* separate calculattions of pol. & anis.*/
char tmp_dir[100];             /* directory for temporal files          */
int nomap=0;                   /* stop calc-s after alm-simulation      */

int lmax = -1;
int lmin = -1;
int mmax = -1;
int mmin = -1;
int nx= -1, np= -1;
double np_factor = -1.0;
char np_formula[121];

char *myname="polmap";

int main(argc, argv)
int argc;
char **argv;
{
    double *alm = (double *)0, *blm = (double *)0, *elm = (double *)0;
    double *x = (double *)0, *w = (double *)0, *wqu = (double *)0;
    int *nph = (int *)0, *mx = (int *)0;
    double *f = (double *)0, *q = (double *)0, *u = (double *)0;
    int nbuf, lmax0 = -1, lmax1 = lmax, lmax2 = lmax, nbuf0, nbuf1, nbuf2;
    FILE *fp_out = (FILE *)0;
    char x_type[40], grid_type[40];
    double x1 = -1.0, x2 = 1.0, phi2 = M_PI*2.;
    int flag_sp2qu = 0;
    char tmp_f[100], tmp_u[100], tmp_q[100];
    int mem_sep_flag = 0;

    alm_bfile[0] = alm_efile[0] = out_qfile[0] = out_ufile[0] = '\0';
    alm_file[0] = out_tfile[0] = '\0';
    alm_hp_file[0] = map3_file[0] = '\0';
    elm_out_file[0] = blm_out_file[0] = '\0';
    np_formula[0] = '\0';

    /* select grid output keyword values */
    strcpy(x_type, "COS(THETA)");
    strcpy(grid_type, "PHI(X)+LEGENDRE");

    strcpy(tmp_dir, "/tmp"); /* default directory for temporal files */

    /* control signal interruption */
    set_sig_fatal();

    /* loop over external parameters */
    setopt(argc, argv);

    if (glesp_comments < 0) {
	if (getenv("GLESP_COMMENTS") != NULL) {
	    char str[21];
	    strcpy(str,getenv("GLESP_COMMENTS"));
	    glesp_comments = *str == 'N' || *str == 'n' ? 0 : 1;
	    if (glesp_comments) {
		 if ( strcmp(str, "OFF") == 0 || strcmp(str, "off") == 0)
		     glesp_comments = 0;
	    }
	} else
	    glesp_comments = 1;
    }

    if (lmax > 0) {
	if (lmax < 100) {
	    if (nx <= 0)  nx = 201;
	    if (np <= 0)  np = 2 * nx;
	} else {
	    if (nx <= 0)  nx = lmax * 2 + 1;
	    if (np <= 0)  np = 2 * nx;
	}
    }
    nbuf1 = nx * np;  /* double arrays for polarization */

    if (*alm_efile != '\0' || *alm_hp_file != '\0' || elm_input||clp_input) {
      /* thanks to Martin Reinecke: addition for 'elm_input' */
      if ((q = (double *)calloc((unsigned)nbuf1, sizeof *q)) == (double *)0) {
nomem:
		PR(ER, "%s: no memory for arrays...\n", myname);
		exit(1);
      }
      if ((u = (double *)calloc((unsigned)nbuf1, sizeof *u)) == (double *)0)
		goto nomem;
    }

    if (alm_input || *alm_hp_file != '\0' || cl_input) {
       if((f = (double *)calloc((unsigned)nbuf1, sizeof *f)) == (double *)0)
		goto nomem;
    }

    if (q == (double *)0 && u == (double *)0 && f == (double *)0) {
	PR(ER, "%s: no input files...\n", myname);
	PR(ER, "---------------------------------\n");
	usage();
    }

    nbuf = -(2*(lmax+1)); /* make negative means use old value */
    alm = f;    /* save memory in f-array */
    elm = q;    /* save memory in q-array */
    blm = u;    /* save memory in u-array */

    if (*alm_hp_file != '\0') {
	/* check input file */
	int flag = test_glesp_map(alm_hp_file);
	switch (flag) {
	case 2: /* FITS alm-file */
	    PR(ER,
	     "%s: the input file '%s' contains a single set of a_{lm}-s.\n",
		myname, alm_hp_file);
	    PR(ER, "Use flags '-a' or '-e/-b' for input.\n");
	    exit (2);
	case -1: /* it is not our file */
	case  0: /* F-file             */
	case -2: /* FITS-file, but not GLESP map */
	    PR(ER,
	     "%s: the input file '%s' contains data of illegal format.\n",
		myname, alm_hp_file);
	    exit (1);
	case -3: /* ASCII-file         */
	    PR(ER,
	     "%s: the input file '%s' contains ASCII data.\n",
		myname, alm_hp_file);
	    PR(ER,
	     "Use flags '-a' or '-e/-b' for input, if they are a_{lm}-s.\n");
	    exit(1);
	case 1:  /* GLESP FITS-file                 */
	case 3:  /* GLESP FITS-file 3 maps (T,Q,U)  */
	    PR(ER,
	     "%s: the input file '%s' contains a map, not a_{lm}-s.\n",
		myname, alm_hp_file);
	    exit(1);
	case 4:  /* FITS-file of 3 alm-s (A,E,B) */
		 /* That is Ok */
		break;
	case 5: /*  Hpix FITS-file T-map  */
	case 6: /*  Hpix FITS-file 3 maps (T,Q,U) */
		PR(ER, "%s: the file '%s' contains Hpix-map. ",
		    myname, alm_hp_file);
		PR(ER, "I can not process it.\n");
		exit (1);
	}

	/*  read 3-alm file */
	/* not create alm,elm,blm because they are not =0, and nbuf < 0 */
	read_ieb_alm_T(&alm, &elm, &blm, &lmax, &nbuf, alm_hp_file);

mem_separ:
	if (memory_separation) { /* if no memory */
	    FILE *fp_tmp;
	    if (mem_sep_flag == 0) {  /* first stage */
		extern void out_glesp_alm_T();
		sprintf(tmp_f, "%s/polmap_fXXXXXX", tmp_dir);
		mktemp(tmp_f);

		if ((fp_tmp = fopen(tmp_f, "w")) == (FILE *)0) {
		    PR(ER,"%s: can not create temoral file '%s'\n",
			myname, tmp_f);
		    exit(1);
		}

		/* save alm-s to disk */
		out_glesp_alm_T(alm, lmax, nbuf, lmax, fp_tmp);
		fclose(fp_tmp);

		free((void *)f);
		f = (double *)0;
		mem_sep_flag = 1;
	    } else if (mem_sep_flag == 1) { /* second stage */
		extern int read_alm_table_T ();
		/* save Q,U maps after calculations */
		int nalm;
		/* Q-map temporal output in FITS-format */
		sprintf(tmp_q, "%s/polmap_qXXXXXX", tmp_dir);
		mktemp(tmp_q);
		if ((fp_tmp = fopen(tmp_q, "w")) == (FILE *)0) {
		    PR(ER,"%s: can not create temporal file '%s'\n",
			myname, tmp_q);
		    unlink(tmp_f);
		    exit(1);
		}
		outp_glesp_map(q, x, nx, np, nph, fp_tmp, x_type, grid_type);
		fclose(fp_tmp);

		/* U-map temporal output in FITS-format */
		sprintf(tmp_u, "%s/polmap_uXXXXXX", tmp_dir);
		mktemp(tmp_u);
		if ((fp_tmp = fopen(tmp_u, "w")) == (FILE *)0) {
		    PR(ER,"%s: can not create temporal file '%s'\n",
			myname, tmp_u);
		    unlink(tmp_f);
		    unlink(tmp_q);
		    exit(1);
		}
		outp_glesp_map(u, x, nx, np, nph, fp_tmp, x_type, grid_type);
		fclose(fp_tmp);

		free((void *)q);
		free((void *)u);
		q = (double *)0;
		u = (double *)0;

		/* read saved alm-s from disk */
		if((f = (double *)calloc((unsigned)(np*nx), sizeof *f))
		       == (double *)0)
			  goto nomem;
		alm = f;
		nbuf = -2*(lmax+1); /* make negative means use old value */
		read_alm_table_T (tmp_f, &alm, &nalm, &nbuf, &lmax);
		mem_sep_flag = 2;
	    }
	}
	/* calculation without anisotropy */
	if (no_ani) {
	    free((void *)f);
	    f = (double *)0;
	}
	/* calculation without polarization */
	if (no_pol) {
	    free((void *)q);
	    free((void *)u);
	    q = (double *)0;
	    u = (double *)0;
	}
	if ( nbuf < 0 ) nbuf = -nbuf;
    } else {
	/* 1-par. files (T,Q,U) input */
	int flag=0;
	int nalm;
	extern void realloc_alm_T();

	/* temperature if given */
	if (*alm_file != '\0') {
	    flag = test_glesp_map(alm_file);
	    switch(flag) {
	    case 0: /* F-format file */
		    lmax0 = get_lmax(alm_file);
		    break;
	    case 2: /* FITS alm-format file */
		    lmax0 = Fget_lmax(alm_file);
		    break;
	    case -1: /* ASCII format */
		    PR(ER, "%s file '%s' has bad format.\n",
			    myname, alm_file);
		    exit(1);
	    default :
	    case -3 : /* ASCII format */
		    lmax0 = Sget_lmax(alm_file);
		    break;
	    }
	    if (lmax0 <= 0)  *alm_file = '\0';
	}  else if (alm_input == 2) { /* alm input from string */
	    lmax0 = get_lmax_str(alm_str);
	}  else if (cl_input) {
	    extern int Cget_max(char *);
	    lmax0 = Cget_max(file_cl);
	}

	/* polarization */
	if (*alm_bfile != '\0' && *alm_efile != '\0') {
	    flag = test_glesp_map(alm_bfile);
	    switch(flag) {
	    case 0: /* F-format file */
		    lmax1 = get_lmax(alm_bfile);
		    break;
	    case 2: /* FITS alm-format file */
		    lmax1 = Fget_lmax(alm_bfile);
		    break;
	    case -1: /* ASCII format */
		    PR(ER, "%s file '%s' has bad format.\n",
			    myname, alm_bfile);
		    exit(1);
	    default :
	    case -3 : /* ASCII format */
		    lmax1 = Sget_lmax(alm_bfile);
		    break;
	    }

	    flag = test_glesp_map(alm_efile);
	    switch(flag) {
	    case 0: /* F-format file */
		    lmax2 = get_lmax(alm_efile);
		    break;
	    case 2: /* FITS alm-format file */
		    lmax2 = Fget_lmax(alm_efile);
		    break;
	    case -1: /* ASCII format */
		    PR(ER, "%s file '%s' has bad format.\n",
			    myname, alm_efile);
		    exit(1);
	    default :
	    case -3 : /* ASCII format */
		    lmax2 = Sget_lmax(alm_efile);
		    break;
	    }
	} else if (blm_input == 2 && elm_input == 2) {
	    lmax1 = get_lmax_str(blm_str);
	    lmax2 = get_lmax_str(elm_str);
	} else {
	    if (!alm_input && !cl_input && !clp_input) {
		PR(ER, "%s: at least, it should be one of samples: ", myname);
		PR(ER, "'a_lm' or 'e_lmb & b_lm' data files...\n");
		exit(1);
	    }
	}
	/* set same resolution */
	if (blm_input && elm_input) {
	    nbuf1 = (lmax1 + 1) * 2;
	    nbuf2 = (lmax2 + 1) * 2;
	    if (lmax < 0) {
		lmax = lmax1 < lmax2 ? lmax1 : lmax2;
	    } else {
		int l = lmax1 < lmax2 ? lmax1 : lmax2;
		lmax = lmax < l ? lmax : l;
	    }
	}

	if (alm_input || cl_input) {
	    if (lmax < 0)
		lmax = lmax0;
	    else
		lmax = lmax < lmax0 ? lmax : lmax0;
	    nbuf0 = (lmax0 + 1) * 2;
	}

	if (nx <= 0)  nx = lmax * 2 + 1;
	if (np <= 0)  np = 2 * nx;
//      if (nbuf ==0 ) nbuf = (lmax + 1) * 2;

	if (nx < 2*lmax+1) {
	    PR(ER,
	      "%s: resolution parameter nx=%d is less than 2*lmax+1=%d\n",
	      myname, nx, 2*lmax+1);
	    exit(1);
	}
	if (np < nx) {
	    PR(ER, "%s: resolution parameter np=%d is less than nx=%d\n",
		 myname, np, nx);
	    exit(1);
	}

	/* set negative values to pass them to subrotines as flags */
	nbuf = -(lmax + 1) * 2;
	nbuf0 = nbuf1 = nbuf2 = nbuf;
	nalm = -(lmax+1) * 2;

	/* polarization input data */
//      if (*alm_bfile != '\0' && *alm_efile != '\0') {
	if (blm_input && elm_input) {
	    extern int read_alm_table_T ();
	    extern void read_alm_str_T (char *, double *, int, int);
	    /* not create blm because it is not =0, and nbuf < 0 */
	    if (blm_input == 1)
		read_alm_table_T (alm_bfile, &blm, &nalm, &nbuf1, &lmax);
	    else if (blm_input == 2)
		read_alm_str_T (blm_str, blm, nbuf1, lmax);
	    /* not create elm because it is not =0, and nbuf < 0 */
	    if (elm_input == 1)
		read_alm_table_T (alm_efile, &elm, &nalm, &nbuf2, &lmax);
	    else if (elm_input == 2)
		read_alm_str_T (elm_str, elm, nbuf2, lmax);
	    lmax1 = lmax2 = lmax;
	} else if (clp_input) {
	    /* simulation of data according to Cl */
	    extern void read_random_alm_T(char *, double **, int *, int *,
		    int *, int, double,int);

	    read_random_alm_T(file_cle, &elm,&nalm,&nbuf0,&lmax,
		    random_seed_e, fwhm, clp_input);
	    read_random_alm_T(file_clb, &blm,&nalm,&nbuf0,&lmax,
		    random_seed_b, fwhm, clp_input);

	    /* output of elm if we need this */
	    if (elm_out_file[0] != '\0') {
		extern void out_glesp_alm_T(double *, int, int, int, FILE *);
		if ((fp_out = fopen(elm_out_file, "w")) == (FILE *)0) {
		    PR(ER, "%s: can not create file '%s'\n",
			myname, elm_out_file);
		    exit(1);
		}
		/* write new a_{lm}-s */
		out_glesp_alm_T(elm, lmax, ABS(nbuf0), lmax, fp_out);
		/* close file */
		if(fp_out != (FILE *)0 && fp_out != stdout) fclose(fp_out);
		fp_out = (FILE *)0;

		if (nomap) {
		  PR(ER, "\nE_lm simulation is done.");
		}
	    }
	    /* output of blm if we need this */
	    if (blm_out_file[0] != '\0') {
		extern void out_glesp_alm_T(double *, int, int, int, FILE *);
		if ((fp_out = fopen(blm_out_file, "w")) == (FILE *)0) {
		    PR(ER, "%s: can not create file '%s'\n",
			myname, blm_out_file);
		    exit(1);
		}
		/* write new a_{lm}-s */
		out_glesp_alm_T(blm, lmax, ABS(nbuf0), lmax, fp_out);
		/* close file */
		if(fp_out != (FILE *)0 && fp_out != stdout) fclose(fp_out);
		fp_out = (FILE *)0;

		if (nomap) {
		  PR(ER, "\nB_lm simulation is done.");
		}
	    }
	    if (nomap) {
		PR(ER, "No map calculations.\n");
		exit(0);
	    }
	}

	/* temperature if we need */
	if (*alm_file != '\0') {
	    extern int read_alm_table_T ();
	    /* not create alm because it is not =0, and nbuf < 0 */
	    read_alm_table_T (alm_file, &alm, &nalm, &nbuf0, &lmax);
	    lmax0 = lmax;
	} else if (alm_input==2) {
	    /* read alm from ASCII string */
	    extern void read_alm_str_T (char *, double *, int, int);
	    read_alm_str_T (alm_str, alm, nbuf, lmax);
	} else if (cl_input) {
	    /* simulation of data according to Cl */
	    extern void read_random_alm_T(char *, double **, int *, int *,
		    int *, int, double,int);
	    read_random_alm_T(file_cl, &alm,&nalm,&nbuf0,&lmax, random_seed,
		    fwhm, cl_input);
	}
	/* output of alm if we need this */
	if (alm_out_file[0] != '\0') {
	    extern void out_glesp_alm_T(double *, int, int, int, FILE *);
	    if ((fp_out = fopen(alm_out_file, "w")) == (FILE *)0) {
		PR(ER, "%s: can not create file '%s'\n",
		    myname, alm_out_file);
		exit(1);
	    }
	    /* write new a_{lm}-s */
	    out_glesp_alm_T(alm, lmax, ABS(nbuf0), lmax, fp_out);
	    /* close file */
	    if(fp_out != (FILE *)0 && fp_out != stdout) fclose(fp_out);
	    fp_out = (FILE *)0;

	    if (nomap) {
		PR(ER, "\nA_lm simulation is done. No map calculations.\n");
		exit(0);
	    }
	}

	/* return values back after subrotines */
	nbuf = (lmax + 1) * 2;
	nalm = (lmax + 1) * 2;

	if (blm != (double *)0 && lmax != lmax1)
	    realloc_alm_T(blm, nbuf, lmax1, lmax);
	if (elm != (double *)0 && lmax != lmax2)
	    realloc_alm_T(elm, nbuf, lmax2, lmax);

	if (*alm_file != '\0') {
	    if (lmax != lmax0)
		realloc_alm_T(alm, nbuf, lmax0, lmax);
	}
	if (nbuf < 0) nbuf = -nbuf;
    }

    if (*map3_file != '\0') {
	if (alm == (double *)0) {
	  PR(ER,
	  "%s: can not contsruct file with 3 maps: no temperature data...\n",
	    myname);
	  PR(ER, "Use '-q' and '-u' flags for output\n");
	  exit(1);
	}
    }

    /* take memory for coordinate arrays the only time */
    if (mem_sep_flag != 2) {
	if ((x = (double *)calloc((unsigned)nx, sizeof *x)) == (double *)0)
		    goto nomem;
	if ((w = (double *)calloc((unsigned)nx, sizeof *w)) == (double *)0)
		    goto nomem;
	if ((mx = (int *)calloc((unsigned)nx, sizeof *mx)) == (int *)0)
		    goto nomem;
	if ((wqu = (double *)calloc((unsigned)nx, sizeof *wqu))==(double *)0)
		    goto nomem;
	if ((nph=(int *)calloc((unsigned)nx, sizeof *nph))  == (int *)0)
		    goto nomem;
    }

    /* coordinates and weights */
    {
	/* if c_flag >=0,  special grid for Legendre polynomials */
	int c_flag,         /* flag of coordinate grid selection */
	    j=lx_add;       /* addition to Nphi numbers          */
	int npp = nx*2;

	switch (lx_add) {
	case 0: /* pin = 0 => square matrix */
		c_flag = 0;  /* => npp = lx_add; lx = -20; */
		/* calc maximum nph */
		j = npp = np;
		coord_pol(c_flag, j, x1, x2, phi2, x, w, wqu, nx, nph, npp);
		npp = nph[(nx+1)/2];
		break;
	case 6:
		c_flag = 0;  /* => npp = lx_add; lx = -20; */
		break;
	case 4:    /* pin = '4' */
		c_flag = 1;  /* lx=lx_add; old value:n0=4, near Pole */
		j = lx_add;
		break;
	case -10:
	case 'o':  /* pin = 'o' */
		c_flag = 1;  /* standard old pixelization */
		j = 't';
		break;
	case -11:
	case 'r': /* pin = 'r' */
		c_flag = 1;  /* lx=lx_add; old value:n0=4, near Pole */
		j = lx_add = -11;
		break;
	case -12: /* pin = 'r' + N_pol */
		c_flag = 1;  /* lx=lx_add; old value:n0=4, near Pole */
		j = lx_add = -12;
		break;
	default :
		if (lx_add < -10000) {
		   c_flag = 1; /* lx=lx_add; old value:n0=4, near Pole */
		   j = -(lx_add + 10000);
		} else if (lx_add < -1000) {
		   /* construct GLESP grid from polar cap */
		   c_flag = 0;  /* => npp = lx_add; lx = -20; */
		   /* number of pixels near Pole */
		   j = -(lx_add + 1000);
		} else {
		    c_flag = 2; /* polarization pixelization */
		}
		break;
	}

	/* 1st par (here 'c_flag') - flag of coordinate grid selection */
	/* 2nd par (here 'j')- addition (increment) to Nphi numbers    */
	coord_pol(c_flag, j, x1, x2, phi2, x, w, wqu, nx, nph, npp);
	if (c_flag == 0)
	    np = nph[(nx+1)/2];
    }
    if (np_factor > 0.0)  {
	register int i;
	for(i=0; i<nx; i++) {
		nph[i] *= np_factor;
		if (nph[i] > np)   nph[i] = np;
	}
	weights_pol(0, 0, x1, x2, phi2, x, w, wqu, nx, nph, np);
    }
    if (np_formula[0] != '\0') {
	extern int nf_formu(double *, int, int, int *, char *);
	nf_formu(x, nx, np, nph, np_formula);
	weights_pol(0, 0, x1, x2, phi2, x, w, wqu, nx, nph, np);
    }
    if (fabs(np_factor+10.0) < 1.e-4) {
	/* healpix */
	extern int np_hpix(double *, int, int, int *);
	np_hpix(x, nx, np, nph);
	weights_pol(0, 0, x1, x2, phi2, x, w, wqu, nx, nph, np);
    }

    /* number of pixels per ring to be used with Nyquist theorem */
    {
      register int i;
      for(i=0; i<nx; i++)  mx[i] = nph[i] * vnyq + 0.5;
    }

//PR(ER, "2----------------------------------\n");
//system("free");
//PR(ER, "2----------------------------------\n");

    /********************************************/
    /* calculation of anisotropy and Q & U maps */
    /********************************************/
    flag_sp2qu = sp2qu_n(f, q, u, x, w, wqu, nx, np, nph,
	  alm, elm, blm, lmin, lmax, nbuf, eps1, eps2, mx);

    /* check free memory flag */
    if (flag_sp2qu < 0) {
	if ( flag_sp2qu == -1 && f != (double *)0 ) {
	    PR(ER,
		"\n%s: I can not find enough memory for temporal arrays,\n",
		 myname);
	    PR(ER, "and continue calculation separately ");
	    PR(ER, "for anisotropy and Q,U-maps.\n");
	    memory_separation = 1;
	    goto mem_separ;
	} else {
	    PR(ER, "\n%s: sorry, ", myname);
	    PR(ER, " actually no memory for temporal arrays...\n");
	    unlink(tmp_f);
	    exit(1);
	}
    }

    /* return back and                         */
    /* calculate separately anisotropy and Q&U */
    if (mem_sep_flag == 1)  {
	if (glesp_comments)
	    PR(ER, "And anisotropy calculation now\n");
	goto mem_separ;
    }

    /* read Q,U-maps calculated earlier to unify in output file */
    if (mem_sep_flag == 2) {
	/* read polarization data from saved arrays from disk */
	read_glesp_map(&q, &x,  &nx,  &np,  &nph,  tmp_q);
	read_glesp_map(&u, &x,  &nx,  &np,  &nph,  tmp_u);
    }

    /***************/
    /* output data */
    /***************/
    if (*map3_file != '\0') {   /* file with 3 maps */
	/* open output file with t,Q,U */
	if (strcmp(map3_file, "stdout") == 0 || strcmp(map3_file, "-") == 0)
	    fp_out = stdout;
	else  {
	    if (q == (double *)0) {
	      strcat(map3_file, "_t");
	      if (glesp_comments)
	       PR(ER,"\nThe file '%s' with a single GLESP map is created.\n",
		    map3_file);
	    } else if (f == (double *)0) {
	      strcat(map3_file, "_q");
	      if (glesp_comments)
		 PR(ER,"\nThe file '%s' with a single GLESP map is created.",
		    map3_file);
	    }
	    if ((fp_out = fopen(map3_file, "w")) == (FILE *)0) {
		PR(ER,"%s: can not create file '%s'\n",myname,map3_file);
		if(mem_sep_flag) {unlink(tmp_f);unlink(tmp_q);unlink(tmp_u);}
		exit(1);
	    }
	}

	if (f == (double *)0) {
	    /* output without anisotropy into 2 files */
	    /* map output in FITS-format */
	    outp_glesp_map(q, x, nx, np, nph, fp_out, x_type, grid_type);
	    if(fp_out != (FILE *)0 && fp_out != stdout) fclose(fp_out);

	    map3_file[strlen(map3_file)-1] = 'u';
	    if (glesp_comments)
		PR(ER,"And the file '%s' too...\n", map3_file);
	    if ((fp_out = fopen(map3_file, "w")) == (FILE *)0) {
		PR(ER,"%s: can not create file '%s'\n",myname,map3_file);
		if(mem_sep_flag) {unlink(tmp_f);unlink(tmp_q);unlink(tmp_u);}
		exit(1);
	    }
	    outp_glesp_map(u, x, nx, np, nph, fp_out, x_type, grid_type);
	} else if (q == (double *)0) {
	    /* output without polarization into 1 file */
	    outp_glesp_map(f, x, nx, np, nph, fp_out, x_type, grid_type);
	} else {
	    /*  write 3-map  */
	    outp_glesp_3map(f,q,u, x, nx, np, nph, fp_out,x_type,grid_type);
	}

	/* close file */
	if(fp_out != (FILE *)0 && fp_out != stdout) fclose(fp_out);
    } else {
	/*----------------------------*/
	/* output into separate files */
	/*----------------------------*/
	if (q != (double *)0) {
	    /* open output file, Q */
	    if (strcmp(out_qfile, "-")==0 || strcmp(out_qfile, "stdout")==0)
		fp_out = stdout;
	    else
		if ((fp_out = fopen(out_qfile, "w")) == (FILE *)0) {
		    PR(ER,"%s: can not create file '%s'\n",myname,out_qfile);
		    exit(1);
		}

	    /*  write map  */
	    if (ffile_flag) {
		/* map output in F-format */
		ff_write_ext2_map(q, x,nx,np,nph, fp_out, x_type, grid_type);
	    } else {
		/* map output in FITS-format */
		outp_glesp_map(q, x, nx, np, nph, fp_out, x_type, grid_type);
	    }

	    /* close file */
	    if(fp_out != (FILE *)0 && fp_out != stdout) fclose(fp_out);
	}

	if (u != (double *)0) {
	    /* open output file, U */
	    if (strcmp(out_ufile, "-")==0 || strcmp(out_ufile, "stdout")==0)
		fp_out = stdout;
	    else
		if ((fp_out = fopen(out_ufile, "w")) == (FILE *)0) {
		    PR(ER,"%s: can not create file '%s'\n",myname,out_ufile);
		    exit(1);
		}

	    /*  write map  */
	    if (ffile_flag) {
		/* map output in F-format */
		ff_write_ext2_map(u, x,nx,np,nph,fp_out,x_type,grid_type);
	    } else {
		/* map output in FITS-format */
		outp_glesp_map(u, x, nx, np, nph, fp_out,x_type,grid_type);
	    }
	    /* close file */
	    if(fp_out != (FILE *)0 && fp_out != stdout) fclose(fp_out);
	}
	if (alm_input || cl_input || *alm_hp_file != '\0') {
	    /* open output file, T */
	    if (out_tfile[0] == '-' || strcmp(out_tfile, "stdout") == 0) {
		fp_out = stdout;
	    } else if (out_tfile[0] == '\0') {
		/* default file */
		sprintf(out_tfile, "%s_t.fts", myname);
		PR(ER,"%s: default file '%s' for temperature map was created",
		       myname, out_tfile);
		if ((fp_out = fopen(out_tfile, "w")) == (FILE *)0) {
		    PR(ER,"%s: can not create file '%s'\n",myname,out_tfile);
		    exit(1);
		}
	    } else
		if ((fp_out = fopen(out_tfile, "w")) == (FILE *)0) {
		    PR(ER,"%s: can not create file '%s'\n",myname,out_tfile);
		    exit(1);
		}
	    /*  write map  */
	    if (ffile_flag) {
		/* map output in F-format */
		ff_write_ext2_map(f,x,nx,np,nph, fp_out, x_type, grid_type);
	    } else {
		/* map output in FITS-format */
		outp_glesp_map(f, x,nx, np,nph, fp_out, x_type, grid_type);
	    }
	    /* close file */
	    if(fp_out != (FILE *)0 && fp_out != stdout) fclose(fp_out);
	}
    }
    free((void *)nph);
    free((void *)x);
    free((void *)w);
    free((void *)wqu);

    free((void *)q);
    free((void *)u);

    if (*alm_file != '\0')   free((void *)f);

    if (mem_sep_flag) {
	unlink(tmp_f);
	unlink(tmp_q);
	unlink(tmp_u);
    }
    exit(0);
}

void setopt(argc, argv)
int argc;
char **argv;
{
    register int i;
    int eps_flag = 0;
    myname = argv[0];

    for(i=1; i<argc; i++) {
	switch(argv[i][0]) {
	case '-':
	    switch(argv[i][1]) {
	    case 'a':
		    if (strncmp(argv[i]+1, "alm", 3) == 0 ||
			strncmp(argv[i]+1, "as",  2) == 0) {
			strcpy(alm_str, argv[++i]);
			alm_input = 2; /* alm from string */
		    } else if (strncmp(argv[i]+1, "ao", 2) == 0) {
			strcpy(alm_out_file, argv[++i]);
		    } else if (strncmp(argv[i]+1, "ai", 2) == 0) {
			strcpy(alm_file, argv[++i]);
			alm_input = 1; /* alm from file */
		    } else  {
			strcpy(alm_file, argv[++i]);
			alm_input = 1; /* alm from file */
		    }
		    break;
	    case 'b':
		    if (strncmp(argv[i]+1, "blm", 3) == 0 ||
			strncmp(argv[i]+1, "bs",  2) == 0) {
			strcpy(blm_str, argv[++i]);
			blm_input = 2; /* blm from string */
		    } else {
			strcpy(alm_bfile, argv[++i]);
			blm_input = 1; /* blm from file */
		    }
		    break;
	    case 'c':
	    case 'C':
		    if (strcmp(argv[i]+1, "Ce") == 0 ||
			    strcmp(argv[i]+1, "ce") == 0 ||
			    strcmp(argv[i]+1, "Cle") == 0 ||
			    strcmp(argv[i]+1, "cle") == 0 ) {
			    clp_input += 1;
			    strcpy(file_cle, argv[++i]);
		    } else if (strcmp(argv[i]+1, "Cb") == 0 ||
			    strcmp(argv[i]+1, "cb") == 0 ||
			    strcmp(argv[i]+1, "Clb") == 0 ||
			    strcmp(argv[i]+1, "clb") == 0 ) {
			    clp_input += 2;
			    strcpy(file_clb, argv[++i]);
		    } else {
			/* read Cl-file */
			strcpy(file_cl, argv[++i]);
			cl_input = 1;
		    }
		    break;
	    case 'd':
		    if (strncmp(argv[i]+1, "do", 2) == 0) {
			    bitpix_map = DOUBLEPIX;
		    } else if (strncmp(argv[i]+1, "di", 2) == 0) {
			    /* dipole calculation only */
			    lmin = lmax = 1;
		    } else if (strncmp(argv[i]+1, "dl", 2) == 0) {
			    /* D(l)-spectrum is in input */
			    strcpy(file_cl, argv[++i]);
			    cl_input = 2;
		    } else {
			    PR(ER,"%s: illegal flag '%s'\n",
				myname, argv[i]);
			    usage();
		    }
		    break;
	    case 'D':
		    /* D(l)-spectrum is in input */
		    strcpy(file_cl, argv[++i]);
		    cl_input = 2;
		    break;
	    case 'e':
		    if (strncmp(argv[i]+1, "ep1", 3) == 0) {
			  eps1 = atof(argv[++i]);
		    } else if (strncmp(argv[i]+1, "ep2", 3) == 0) {
			  eps2 = atof(argv[++i]);
		    } else if (strncmp(argv[i]+1, "ep", 2) == 0) {
			  /* accuracy for Legendre plynomials */
			  eps = atof(argv[++i]);
			  eps_flag = 1;
		    } else if (strncmp(argv[i]+1, "elm", 3) == 0 ||
			strncmp(argv[i]+1, "es",  2) == 0) {
			strcpy(elm_str, argv[++i]);
			elm_input = 2; /* blm from string */
		    } else {
			strcpy(alm_efile, argv[++i]);
			elm_input = 1; /* elm from file */
		    }
		    break;
	    case 'f':
		    if (strncmp(argv[i]+1, "fl", 2) == 0) {
			    bitpix_map = FLOATPIX;
		    } else if (strncmp(argv[i]+1, "fa", 2) == 0) {
			    strcpy(alm_file, argv[++i]);
		    } else if (strncmp(argv[i]+1, "fw", 2) == 0) {
			    fwhm = atof(argv[++i]);
			    if (fwhm < 1.e-8) {
			      PR(ER, "%s: you can not smooth data with ",
					 myname);
			      PR(ER, "such beam size: '-fw %g'\n", fwhm);
			      PR(ER, "Please, set positive value.\n");
			      exit(1);
			    }
			    if (argv[i-1][strlen(argv[i-1])-1] == '2') {
				/* last symbol after 'fw' is '2' */
				fwhm *= M_SQRT2;
			    }
		    } else {
			    PR(ER,"%s: illegal flag '%s'\n",
				myname, argv[i]);
			    usage();
		    }
		    break;
	    case 'g':
		    if (strncmp(argv[i]+1, "gr", 2) == 0) {
		      if (argv[i][3] != '\0') {
			int c = argv[i][3];
			switch (c) {
			case 'a': case 'A': lx_add = -10; break;
			case 'n': case 'N': lx_add =   0; break;
			case 's': case 'S': lx_add =   4; break;
			case 'i': goto lab1;
			default :
			    PR(ER,"%s: illegal flag '%s'\n",
				myname, argv[i]);
			    usage();
			}
		      } else {
lab1:
			i++;
			if (argv[i][0] == 'a' || argv[i][0] == 'A') {
			    /* quasi-equal area pixel */
			    lx_add = -10;  /* old scheme */
			} else if (argv[i][0]=='n' || argv[i][0]=='N') {
			    /* equal pixel number in rings */
			    lx_add = 0; /* pin 0 */
			} else if (argv[i][0]=='s' || argv[i][0]=='S') {
			    /* equal pixel number in rings */
			    lx_add = 4; /* pin 4 */
			}
		      }
		    }
		    break;
	    case 'h':
		    if (strncmp(argv[i]+1,"h2",2) == 0)
			usage2();
		    else
			usage();
		    break;
	    case 'l':
		    switch(argv[i][2]) {
		    case 'm':
			if (strncmp(argv[i]+2,"mi",2) == 0)
			    lmin = atoi(argv[++i]); /* Lmin */
			else if (strncmp(argv[i]+2,"ma",2) == 0)
			    lmax = atoi(argv[++i]); /* Lmax */
			else {
			    PR(ER,"%s: illegal flag '%s'\n",
				myname, argv[i]);
			    usage();
			}
		    }
		    break;
	    case 'L':
		    if ('0' <= argv[i][2] && argv[i][2] <= '9')
			lmax = lmin = atoi(argv[i]+2);
		    else
			lmax = lmin = atoi(argv[++i]);
		    break;
	    case 'm':
		    switch(argv[i][2]) {
		    case 'm':
			if (strncmp(argv[i]+2,"mi",2) == 0)
			    mmin = atoi(argv[++i]); /* Mmin */
			else if (strncmp(argv[i]+2,"ma",2) == 0)
			    mmax = atoi(argv[++i]); /* Mmax */
			else {
			    PR(ER,"%s: illegal flag '%s'\n",
				myname, argv[i]);
			    usage();
			}
		    case 's':
			if (strncmp(argv[i]+1,"ms", 2) == 0) {
			    /* separation of calculation of anisotropy */
			    /* and polarization */
			    memory_separation = 1;
			}
			break;
		    default:
			PR(ER,"%s: illegal flag '%s'\n",
			    myname, argv[i]);
			usage();
		    }
		    break;
	    case 'M':
		    mmax = mmin = atoi(argv[++i]);
		    break;
	    case 'n':
		    switch(argv[i][2]) {
		    case 'f':
			    i++;
			    if (strncmp(argv[i], "hp", 2) == 0)
				np_factor = -10;
			    else
				np_factor = atof(argv[i]);
			    break;
		    case 'F':
			    strcpy(np_formula, argv[++i]);
			    break;
		    case 'o':
			    if (strncmp(argv[i]+1, "nop", 3) == 0) {
				no_pol = 1; /* no polarization calc */
			    } else if (strncmp(argv[i]+1, "noa", 3)==0){
				no_ani = 1; /* no anisotropy calc */
			    } else if (strncmp(argv[i]+1, "nom", 3)==0){
				nomap = 1; /* no anisotropy calc */
			    } else {
				PR(ER,"%s: illegal flag '%s'\n",
				    myname, argv[i]);
				usage();
			    }
			    break;
		    case 'p':
			    np = atoi(argv[++i]);
			    break;
		    case 'x':
			    nx = atoi(argv[++i]);
			    break;
		    case 'q':
			    vnyq = atof(argv[++i]);
			    break;
		    default:
			    usage();
			    break;
		    }
		    break;
	    case 'o':
		    switch(argv[i][2]) {
		    case 'u':
			    if(strncmp("ouf", argv[i]+1, 3) == 0) {
				    ffile_flag = 1;
			    }
			    strcpy(out_ufile, argv[++i]);
			    break;
		    case 'q':
			    if(strncmp("oqf", argv[i]+1, 3) == 0) {
				    ffile_flag = 1;
			    }
			    strcpy(out_qfile, argv[++i]);
			    break;
		    case 't':
			    if(strncmp("otf", argv[i]+1, 3) == 0) {
				    ffile_flag = 1;
			    }
			    strcpy(out_tfile, argv[++i]);
			    break;
		    default:
			    strcpy(map3_file, argv[++i]);
			    break;
		    }
		    break;
	    case 'p':
		    if (strncmp(argv[i]+1, "pin", 3) == 0) {
			    /* phi increment number :       */
			    /* addition to Nphi numbers for */
			    /* 'coord'-function             */
			    int v = atoi(argv[++i]);
			    if (argv[i][0] == 'o' ||
				argv[i][0] == 'O' ||
				argv[i][0] == 'T' ||
				argv[i][0] == 't' ) {
				    lx_add = -10;
			    } else if( argv[i][0] == '4' ) {
				   /* special case for pin=4 */
				    lx_add = 4;
			    } else if (argv[i][0] == 'P')
				 /* polarization */
				 lx_add = v = 'P';
			    else if (argv[i][0] == 'r')
				 lx_add = v = -11;
			    else if (v < -1)
				 lx_add = -1000 - v;
			    else lx_add = v;
		    }
		    break;
	    case 'P':
		    switch (argv[i][1]) {
		    case 'P':
		    default :
			switch (argv[i][2]) {
			default :
			case '1':
				  N_pol1 = atoi(argv[++i]);
				  break;
			case '2':
				  N_pol2 = atoi(argv[++i]);
				  break;
			case '3':
				  N_pol3 = atoi(argv[++i]);
				  break;
			}
		    }
		    lx_add = -12;
		    break;
	    case 'q':
		    strcpy(out_qfile, argv[++i]);
		    break;
	    case 'r': /* random seed */
		    if (strcmp(argv[i]+1, "re") == 0) {
			random_seed_e = atoi(argv[++i]);
			seed_flag = 1; /* global flag */
		    } else if (strcmp(argv[i]+1, "rb") == 0) {
			random_seed_b = atoi(argv[++i]);
			seed_flag = 1; /* global flag */
		    } else {
			random_seed = atoi(argv[++i]);
			seed_flag = 1; /* global flag */
		    }
		    break;
	    case 's':
		    /* if flag=1, then direct calculations, */
		    /* else from largest M to lower         */
		    if (i < argc-1) {
			if (argv[i+1][0] != '-') {
			    i++;
			    if(argv[i][0] == 'L' || argv[i][0] == '1') {
				    legendre_calc_scheme = 1;
			    } else
				    legendre_calc_scheme = -1;
			} else {
			    legendre_calc_scheme = !legendre_calc_scheme;
			}
		    } else /* flag '-s' is the last parameter */
			    legendre_calc_scheme = !legendre_calc_scheme;
		    break;
	    case 't':
		    if (strcmp(argv[i]+1, "tmp") == 0) {
			    strcpy(tmp_dir, argv[++i]);
		    } else {
			    strcpy(out_tfile, argv[++i]);
		    }
		    break;
	    case 'u':
		    strcpy(out_ufile, argv[++i]);
		    break;
	    case 'v':
		    PR(ER, "%s: version 0.6 (07.09.2015), GLESP 2.0.\n",
			myname);
		    exit(0);
	    case 'w':
		    fwhm = atof(argv[++i]);
		    if (fwhm < 1.e-8) {
		      PR(ER, "%s: you can not smooth data with ",
				 myname);
		      PR(ER, "such beam size: '-fw %g'\n", fwhm);
		      PR(ER, "Please, set positive value.\n");
		      exit(1);
		    }
		    if (argv[i-1][strlen(argv[i-1])-1] == '2') {
			/* last symbol after 'fw' is '2' */
			fwhm *= M_SQRT2;
		    }
		    break;
	    default:
		    break;
	    }
	    break;
	case '+':
		switch(argv[i][1]) {
		case 'h':
		default:
			usage();
			break;
		}
		break;
	default:
		strcpy(alm_hp_file, argv[i]);
		break;
	}
    }

    if (eps_flag) {
	eps1 = eps2 = eps;
    }

    if (lmax > (nx+1) / 2 && !nomap) {
	PR(ER,"%s:\n'nx'-parameter (flag '-nx') should be at least \n",
	    myname);
	PR(ER,"two times + 1 larger than 'lmax' (flag '-lmax').\n");
	exit(1);
    }

    if (np < nx * 2 && !nomap) {
	PR(ER,"%s:\n'nphi'-parameter (flag '-np') should be at least \n",
	    myname);
	PR(ER,"two times larger than 'nx' (flag '-nx').\n");
	exit(1);
    }

    if (*map3_file == '\0' && *out_tfile == '\0' &&
	  *out_qfile == '\0' &&  *out_ufile == '\0' && !nomap)  {
	    PR(ER, "%s: please, set map output filename too...\n", myname);
	    if (alm_input || cl_input) {
		PR(ER, "Use flags '-o' or '-t'.\n");
	    }
	    exit(1);
    }
    return;
}

void usage()
{
PR(ER, "usage: %s [3alm_file] [-o out_3map_file] [-nx nx] [-np np] \n", myname);
PR(ER, "[-b alm_bfile -e alm_efile] [-q out_qmap_file] [-u out_umap_file]\n");
PR(ER, "[-a alm_file(format: l,m,re,im)] [-t out_anis_file] [-as l,m,re,im]\n");
PR(ER, "[-bs string:l,m,re,im] [-es string:l,m,re,im]\n");
PR(ER, "[-lmax lmax] [-pin {0,P,o,4}] [-ep eps] [-nq nyq_coeff] [-tmp tmp_dir]\n");
PR(ER, "[-Cl file_with_Cl] [-r random_seed] [-ao output_alm_file]\n");
PR(ER, "[-Cl{e,b} Cl{e_b}-file] [-r{e,b} rand_seed] [-o{e,b} outp_xlm_file]\n");
PR(ER, "[-nop(no polarization)] [-noa(no anisotropy)] [-ms(memory_sep)]\n");
PR(ER, "[-ep1 eps1] [-ep2 eps2][-{float,double}] [-s {M,L}] [-P{123} N{123}]\n");
PR(ER, "[-nf {np_factor,hpix}] [-nF np_formula] [-nomap]\n");
exit(2);
}

void usage2()
{
PR(ST,"usage: %s [FLAGS]\n", myname);
PR(ST,"-----------------------------------------------------------------\n");
//PR(ST, "-P{123} N{123}\n");
//PR(ST, "-nf {np_factor,hpix}\n");
//PR(ST, "-nF np_formula\n");
PR(ST,"[3alm_file] - input 3alm-file (T-alm and E,B-modes)\n");
PR(ST,"-a alm_file(l,m,re,im) - input temperature a_lm file  (see '-t','-q','-u')\n");
PR(ST,"-ao output_alm_file - file to save a_lm-s when calculation of Cl\n");
PR(ST,"-as 'l,m,re,im' - input string with temperature a_lm-s (see '-bs','-es')\n");
PR(ST,"-b alm_bfile - input B-polarization mode a_lm file\n");
PR(ST,"-bs 'l,m,re,im' - input string with B-polarizaion mode a_lm-s (see '-as','-es')\n");
PR(ST,"-Cl Cl-file - use Cl-file to simulate random map (see '-r','-ao')\n");
PR(ST,"-Clb Clb-file - use Clb-file to simulate random pol-map (see '-rb','-ob')\n");
PR(ST,"-Cle Cle-file - use Cl-file to simulate random pol-map (see '-re','-oe')\n");
PR(ST,"-Dl Dl-file - use Dl-file to simulate random map (see '-r','-ao')\n");
PR(ST,"-double - output file in 64-bit float format (default) (see '-float')\n");
PR(ST,"-e alm_efile - input E-polarization mode a_lm file\n");
PR(ST,"-es 'l,m,re,im' - input string with E-polarizaion mode a_lm-s (see '-as','-bs')\n");
PR(ST,"-ep eps   - sets accuracy for temperature calculations (dflt: 1.e-20)\n");
PR(ST,"-ep1 eps1 - sets accuracy for Q-pol. calculations (dflt: 1.e-20)\n");
PR(ST,"-ep2 eps2 - sets accuracy for U-pol. calculations (dflt: 1.e-20)\n");
PR(ST,"-float - output file in 32-bit float format (see '-double')\n");
PR(ST,"-fw{2} fwhm(') - smooth a map with gaussian beam of 'fwhm' arcmin\n");
PR(ST,"-gr a - selects GLESP-pixelization: equal area ('-pin o')\n");
PR(ST,"-gr n - selects GLESP-pixelization: equal ring number of pixels ('-pin 0')\n");
PR(ST,"-LN - same to '-lmin N -lmax N'\n");
PR(ST,"-lmax lmax - sets maximum l-multipole number to calculate\n");
PR(ST,"-lmin lmin - sets minimum l-multipole number to calculate\n");
PR(ST,"-MN - same to '-mmin N -mmax N'\n");
PR(ST,"-mmin mmin - sets minimum m-multipole number to calculate\n");
PR(ST,"-mmax mmax - sets maximum m-multipole number to calculate\n");
PR(ST,"-ms - separates memory for temperature and polarization calculation\n");
PR(ST,"-noa   - no calculation of temperature anisotropy for 3alm-file\n");
PR(ST,"-nomap - no map calculation when alm-calculation by Cl (see '-Cl','-ao')\n");
PR(ST,"-nop   - no calculation of polarization for 3alm-file\n");
PR(ST,"-np np - resolution parameter, number of pixels on equator (np=2nx)\n");
PR(ST,"-nx nx - resolution parameter, number of rings (nx=2lmax+1)\n");
PR(ST,"-nq nyq_coeff - sets Nyquist frequency coefficient (default: 0.5)\n");
PR(ST,"-o out_3map_file - output 3map-file (T,U,Q)\n");
PR(ST,"-oa output_alm_file - file to save a_lm-s when calculation of Cl\n");
PR(ST,"-ob output_blm_file - file to save b_lm-s when calculation of Clb\n");
PR(ST,"-oe output_elm_file - file to save b_lm-s when calculation of Cle\n");
PR(ST,"-pin {0,P,o,4} - selects type of GLESP-pixelization (default: '-pin 0')\n");
PR(ST,"-q out_qmap_file - output Q-map file (see '-a','-u')\n");
PR(ST,"-r random_seed - random seed for random map simulation (see '-Cl')\n");
PR(ST,"-r random_seed_b - random seed for random pol-map simulation (see '-Clb')\n");
PR(ST,"-r random_seed_e - random seed for random pol-map simulation (see '-Cle')\n");
PR(ST,"-s {M,L} - selects Legendre polynomial calculation scheme (dflt: M)\n");
PR(ST,"-t out_anis_file - output temperature anisotropy map file\n");
PR(ST,"-tmp tmpcat - changes temporal directory (default: /tmp)\n");
PR(ST,"-u out_umap_file - output U-map file (see '-a','-q')\n");
PR(ST,"-v - prints version of '%s'\n", myname);
exit(0);
}
