#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "glesp.h"

#define PR      fprintf
#define ER      stderr
#define ABS(x)  ((x) < 0 ? -(x) : (x))
#define FMAX    1.e+30

char alm1_file[161],  alm2_file[161],
     almG1_file[161], almG2_file[161],
     outp_file[161];
char log_file[161], file_wcf[161];

//int  bitpix_map = -32;
HEAD *ph;
int  print_w_flag = 0;
int  print_alm_flag = 0;
int  lmin_print = -1;
int  lmax_print = -1;
int  cut_lmin = -1;    /* cut value for a_lm        */
int  cut_lmax = 0;     /* cut value for a_lm        */
# if defined (IRIX)
int  lmax = 2000;
# else
int  lmax = 600;
# endif
double gm1 = 1., gm2 = 1.;
int filter_flag = 0;     /* flag to select among filters */
int sum_flag = 0;        /* if flag=1, summing by m      */
int tensor_flag=0;       /* tensor estimation of weights */
int remove_zero_flag=0;  /* flag to escape m=0           */
int old_flag=0;
int bw_flag = 0;
int m2_flag=0;

extern void setopt(int, char **), usage(void);
extern void get_val_str(char *, double *, double *);

char   tmpcat[161];
int    naxis2;
char *myname="alm2dl";

int main(argc, argv)
int argc;
char **argv;
{
    register int l, m;
    double *alm1 = (double *)0,
	   *alm2 = (double *)0;
    double *almG1= (double *)0,
	   *almG2= (double *)0;
    int Nmx, nbuf, nalm;
    double *da;

    FILE *fp_wcf = (FILE *)0;
    FILE *fp_out = (FILE *)0;
    double w1, w2, s1=0.0, s2=0.0, s3=0.0, v1=0.0, v2=0.0;

    alm1_file[0] = alm2_file[0] = outp_file[0] = '\0';
    almG1_file[0] = almG2_file[0] = log_file[0] = '\0';
    file_wcf[0] = '\0';

    /* loop over external parameters */
    setopt(argc, argv);

    nalm=Nmx=(lmax+1)*2+1;
    nbuf=2*Nmx;

    if ((alm1 = (double *)calloc((unsigned)(Nmx*nbuf), sizeof *alm1)) ==
	(double *)0) {
nomem:
	    PR(ER, "%s: no memory for arrays...\n", myname);
	    exit(1);
    }
    if ((alm2 = (double *)calloc((unsigned)(Nmx*nbuf), sizeof *alm2)) ==
	(double *)0)
	     goto nomem;
    if ((da = (double *)calloc((unsigned)(Nmx*nbuf), sizeof *da)) ==
	(double *)0)
	     goto nomem;

    if ( *almG1_file != '\0') {
	if((almG1=(double*)calloc((unsigned)(Nmx*nbuf),sizeof *almG1))==
	   (double *)0)
		goto nomem;
	if((almG2=(double*)calloc((unsigned)(Nmx*nbuf),sizeof *almG2))==
	    (double *)0)
		 goto nomem;
    }
    read_alm_table(alm1_file, &alm1, &nalm, &nbuf, &lmax);
    read_alm_table(alm2_file, &alm2, &nalm, &nbuf, &lmax);
//    ff_read_alm(alm1, nbuf, &lmax, alm1_file);
//    ff_read_alm(alm2, nbuf, &lmax, alm2_file);

    if ( *almG1_file != '\0') {
	    ff_read_alm(almG1, nbuf, &lmax, almG1_file);
	    ff_read_alm(almG2, nbuf, &lmax, almG2_file);
    }

    for(l=0; l<=lmax; l++) {
	int ll = l * nbuf;
	for(m=0; m<=l; m++) {
	    alm1 [ll+m]        /= gm1;
	    alm2 [ll+m]        /= gm2;
	    alm1 [ll+m+lmax+1] /= gm1;
	    alm2 [ll+m+lmax+1] /= gm2;
	    if ( *almG1_file != '\0') {
		    almG1 [ll+m]        /= gm1;
		    almG2 [ll+m]        /= gm2;
		    almG1 [ll+m+lmax+1] /= gm1;
		    almG2 [ll+m+lmax+1] /= gm2;
	    }
	}
    }
    if (cut_lmin > 0) {
	    for(l=0; l<=cut_lmin; l++) {
		int ll = l * nbuf;
		alm1[ll] = 0.0;
		alm2[ll] = 0.0;
		for(m=1; m<=l; m++) {
		    alm1[ll + m] = alm1[ll + m + lmax+1] = 0.0;
		    alm2[ll + m] = alm2[ll + m + lmax+1] = 0.0;
		    if ( *almG1_file != '\0') {
			almG1[ll + m] = almG1[ll + m + lmax+1] = 0.0;
			almG2[ll + m] = almG2[ll + m + lmax+1] = 0.0;
		    }
		}
	    }
    }

    if (lmax_print < 0 || lmax_print > lmax)
	    lmax_print = lmax;
    if (lmin_print < 0 || lmin_print > lmax)
	    lmin_print = 0;

    if (cut_lmax == 0)  cut_lmax = lmax;

    if (outp_file[0] == '\0' || strcmp(outp_file, "stdout") == 0)
	fp_out = stdout;
    else
	if ((fp_out = fopen(outp_file, "w")) == (FILE *)0) {
	    PR(ER, "%s: can not create file '%s'\n", myname, outp_file);
	    exit(1);
	}
/*
    da = alm1;
*/
    if (print_w_flag) {
	fp_wcf = fopen(file_wcf, "w");
	if (fp_wcf == (FILE *)0) {
	    PR(ER, "%s: can not open file '%s'\n", myname, file_wcf);
	    exit(1);
	}
    }

    /* calculation of da */
    if (filter_flag) {
	/********************************************************/
	/* w-filter: see Naselsky et al. 2003, astro-ph/0310235 */
	/********************************************************/
	FILE *fp_log;

	/* open log-file */
	if (log_file[0] == '\0')  sprintf(log_file, "%s.log", myname);
	fp_log = fopen(log_file, "w");
	if (fp_log == (FILE *)0) fp_log = stderr;

	for(l=2; l<=lmax; l++) {
	  int ll = l * nbuf;

	  w1=0.0, w2=0.0;

	  if (sum_flag) {
	      double s1=0.0, s2=0.0, s3=0.0;

	      for(m=0; m<=l; m++) {
		  double reG1 = almG1 [ll+m],
			 reG2 = almG2 [ll+m],
			 imG1 = almG1 [ll+m+lmax+1],
			 imG2 = almG2 [ll+m+lmax+1];
		  double ampG1 = sqrt(reG1*reG1 + imG1*imG1),
			 ampG2 = sqrt(reG2*reG2 + imG2*imG2);

		  s1 += ampG2;
		  s2 += ampG1;
		  s3 += fabs(ampG1-ampG2);

	      }
	      w1 = s1 / s3;
	      w2 = s2 / s3;
	  }
	  for(m=0; m<=l; m++) {
	    double re1 = alm1 [ll+m],
		   re2 = alm2 [ll+m],
		   im1 = alm1 [ll+m+lmax+1],
		   im2 = alm2 [ll+m+lmax+1];
	    double reG1 = almG1 [ll+m],
		   reG2 = almG2 [ll+m],
		   imG1 = almG1 [ll+m+lmax+1],
		   imG2 = almG2 [ll+m+lmax+1];
	    double amp1 = sqrt(re1*re1 + im1*im1),
		   amp2 = sqrt(re2*re2 + im2*im2);
	    double ampG1 = sqrt(reG1*reG1 + imG1*imG1),
		   ampG2 = sqrt(reG2*reG2 + imG2*imG2);
	    double psi1 = ph_calc(reG1, imG1),
		   psi2 = ph_calc(reG2, imG2);
	    double delta = psi2 - psi1,
		   c_d=cos(delta), s_d=sin(delta);
	    double sin_Omega1=0.00, sin_Omega2=0.00,
		   cos_Omega1=0.00, cos_Omega2=0.00;
	    double a_c, a_s;

	    if ( !sum_flag ) {
		    w1 = ampG2 / fabs(ampG1-ampG2);
		    w2 = ampG1 / fabs(ampG1-ampG2);
	    }
	    sin_Omega1 = w2 * sin(delta);

	    if (fabs(sin_Omega1) <= 1.) {
		    double Omega1 = asin(sin_Omega1);
		    if (M_PI/2 < delta && delta < 1.5*M_PI) {
			    Omega1 = Omega1;
		    }
		    sin_Omega2 = sin(Omega1 + delta);
	    }

	    if (fabs(sin_Omega1) > 1.)  {
	       da[ll+m] = da[ll+m+lmax+1] = 0.00000;

	       PR(fp_log,
	       "%4d %4d %12.5e %12.5e %12.5e %12.5e %12.5e %12.5e %12.5e\n",
		  l, m, delta, sin(delta), sin_Omega1,
		  ampG1, ampG2, fabs(ampG1-ampG2), w2);
	    } else {
	       cos_Omega1 = sqrt(1-sin_Omega1*sin_Omega1);
	       cos_Omega2 = sqrt(1-sin_Omega2*sin_Omega2);
	       if (0.5*M_PI < delta && delta < 1.5*M_PI) {
		    cos_Omega1 = -cos_Omega1;
		    cos_Omega2 = -cos_Omega2;
	       }

	       a_c = w1*re1 - w2 * (re2*c_d + im2*s_d);
	       a_s = w1*im1 + w2 * (re2*s_d - im2*c_d);
	       da[ll+m]        = cos_Omega1*a_c + sin_Omega1*a_s;
	       da[ll+m+lmax+1] = cos_Omega1*a_s - sin_Omega1*a_c;

	       da[ll+m]        *= -1.;
	       da[ll+m+lmax+1] *= -1.;

	    }
	  }
	}
	fclose(fp_log);
    } else {
	int m_min;
PR(ER, "G-filter, M\n");
	/********************************************************/
	/* g-filter: see Naselsky et al. 2003, astro-ph/0310235 */
	/********************************************************/
	for(m=0; m<=lmax; m++) {
double z1=0., z0=0., z3=0., z03=0.0;
	    s1 = s2 = s3 = 0.0;

	    for(l=m; l<=lmax; l++) {
		register int ll = l * nbuf;
		double  re1 = alm1 [ll+m],
			re2 = alm2 [ll+m],
			im1 = alm1 [ll+m+lmax+1],
			im2 = alm2 [ll+m+lmax+1];
		double  amp1 = sqrt(re1*re1+im1*im1),
			amp2 = sqrt(re2*re2+im2*im2);
		double  psi1 = ph_calc(re1, im1),
			psi2 = ph_calc(re2, im2);
		double  dpsi = psi2-psi1; /*, damp = amp2-amp1; */
		double  r1, r2, r3;
		int m2 = m*m;

		r1 = amp2 * (amp2 - amp1 * cos(dpsi));
		r2 = amp1 * (amp1 - amp2 * cos(dpsi));
		if ( m != 0 ) { r1 *= 2.; r2 *= 2.; }
		/* r3 = damp * damp; */
		if (m2_flag)  { r1 *= m2; r2 *= m2; }
		r3 = r1 + r2;

		if ( m == 0 ) {
		    if (!remove_zero_flag) {
			double ss=(re1-re2)*(re1-re2)+(im1-im2)*(im1-im2);

			if (old_flag) {
			    z1 += (amp2*amp2 - amp1*amp1);
			    z3 += ss;
			}
			z0  += amp2*amp2 - amp1*amp1;
			z03 += ss;

		    }
		} else {
		    double ss=(re1-re2)*(re1-re2)+(im1-im2)*(im1-im2);
		    z0  += 2 * (amp2*amp2 - amp1*amp1);
		    z03 += 2 * ss;
		    z1 += 2 * (amp2*amp2 - amp1*amp1);
		    z3 += 2 * ss;
		}

		s1 += r1;
		s2 += r2;
		s3 += r3;
	    }

	    w1 = 0.5 * ( 1 + z1/z3 );
	    w2 = 1.0 - w1;

	    if (fp_wcf != (FILE *)0)
		PR(fp_wcf, "%d  %12.5e, %12.5e\n", m, w1, w2);

//            /* m = 0 */
//            m_min = 0;
//            if (!remove_zero_flag) {
//                if (!old_flag) {
//                   z0 = 0.5 * ( 1 + z0/z03 );
//                   da[ll]  = z0 * alm1[ll] + (1.0-z0) * alm2[ll] ;
//                   m_min = 1;
//                }
//            }

	    for(l=m; l<=lmax; l++) {
		register int ll = l * nbuf;
		double re1 = alm1 [ll+m],
		       re2 = alm2 [ll+m],
		       im1 = alm1 [ll+m+lmax+1],
		       im2 = alm2 [ll+m+lmax+1];

		if (bw_flag) {
		    da[ll+m]        = w2 * (re1 - re2);
		    da[ll+m+lmax+1] = w2 * (im1 - im2);
		} else {
		    da[ll+m]        = w1 * re1 + w2 * re2;
		    da[ll+m+lmax+1] = w1 * im1 + w2 * im2;
		}
	    }
	}
    }
    if ((fp_out = fopen(outp_file, "w")) == (FILE *)0) {
	PR(ER, "%s: can not create file '%s'\n", myname, outp_file);
	exit(1);
    }
    /* write alm */
    //ff_write_alm(da, nbuf, lmax, cut_lmax, fp_out);
    out_glesp_alm(da, lmax, nbuf, cut_lmax, fp_out);

    /* close file */
    if(fp_out != (FILE *)0 && fp_out != stdout) fclose(fp_out);
    if(fp_wcf != (FILE *)0 && fp_wcf != stdout) fclose(fp_wcf);

    /* free memory */
    if (alm1!= (double *)0) free((void *)alm1);
    if (alm2!= (double *)0) free((void *)alm2);

    exit(0);
}

void setopt(argc, argv)
int argc;
char **argv;
{
	register int i;
	myname = argv[0];
	if (argc == 1)
		usage();
	for(i=1; i<argc; i++) {
	    switch(argv[i][0]) {
	    case '-':
		switch(argv[i][1]) {
		case '0':
			remove_zero_flag = 1;
			break;
		case 'a':
			if (alm1_file[0] == '\0')
			       strcpy(alm1_file, argv[++i]);
			else if (alm2_file[0] == '\0')
			       strcpy(alm2_file, argv[++i]);
			break;
		case 'c':
			if (strncmp(argv[i]+1, "clma", 4) == 0) {
				cut_lmax = atoi(argv[++i]);
			} else if (strncmp(argv[i]+1, "clmi", 4) == 0) {
				cut_lmin = atoi(argv[++i]);
			}
			break;
		case 'g':
			if (strcmp(argv[i]+1, "gm") == 0)
			       get_val_str(argv[++i], &gm1, &gm2);
			else if (almG1_file[0] == '\0')
			       strcpy(almG1_file, argv[++i]);
			else if (almG2_file[0] == '\0')
			       strcpy(almG2_file, argv[++i]);
			break;
		case 'G':
			filter_flag = 0;
			break;
		case 'b':
			bw_flag = 1;
			break;
		case 'h':
			usage();
			break;
		case 'l':
			if (strncmp(argv[i]+1, "lmn",3) == 0 ||
			    strncmp(argv[i]+1, "lmi",3) == 0 )
				lmin_print = atoi(argv[++i]);
			else if (strcmp(argv[i]+1, "log") == 0 )
				strcpy(log_file, argv[++i]);
			else
				lmax_print = atoi(argv[++i]);
			break;
		case 'L':
			lmax = atoi(argv[++i]);
			break;
		case 'm':
			m2_flag=1;
			break;
		case 'o':
			if (strncmp(argv[i]+1, "old", 2) == 0)
				old_flag = 1;
			else
				strcpy(outp_file, argv[++i]);
			break;
		case 's':
			if (strcmp(argv[i]+1, "sum") == 0)
			       sum_flag = 1;
			break;
		case 't':
			if (strcmp(argv[i]+1, "t") == 0)
			       tensor_flag = 1;
			break;
		case 'w':
			strcpy(file_wcf, argv[++i]);
			print_w_flag = 1;
			break;
		case 'W':
			filter_flag = 1;
			break;
		default :
			usage();
			break;
		}
		break;
	    case '+':
		switch(argv[i][1]) {
		case 'h':
			usage();
		default :
			usage();
			break;
		}
		break;
	    default :
		if (alm1_file[0] == '\0')
		       strcpy(alm1_file, argv[i]);
		else if (alm2_file[0] == '\0')
		       strcpy(alm2_file, argv[i]);
		break;
	    }
	}

	return;
}

void usage()
{
PR(ER,"usage: %s -a alm1_file alm2_file [-g Gal_file1 -g Gal_file2] [-o out_file]\n",
myname);
PR(ER,"[-l lmax_print] [-lmin lmin_print] [-cl]\n");
PR(ER,"[-clmin cut_lmin] [-clmax cut_lmax] [-{G,W}]\n");
PR(ER,"[-gm gm1,gm2] [-log log_file] [-tens(tensor weights)] [-0]\n");
PR(ER,"[-w file_wcf] [-m2]\n");
exit(2);
}
