#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#ifdef   sun
#define  MAX_RAND   2147483647.
#define  MAX_RAND1  4294967295.
#else
#define  MAX_RAND   32767.
#endif

#ifdef __GNUC__
#undef   MAX_RAND
#define  MAX_RAND   RAND_MAX
#endif

#define  ABS(x)     ((x) < 0 ? -(x) : (x))

int _mq_sum = 32;

#
/* - defined in stdlib.h
#ifndef __GNUC__
int rand();
void srand();
#endif
*/

#ifdef __GNUC__
extern int  init_rand(void);
extern void g_rand(double *, unsigned int, double);
extern void g_frand(float *, unsigned int, float);
extern void u_rand(double *, unsigned int, double, double);
extern void u_frand(float *, unsigned int, float, float);
extern double norm_rand(void), uniform_rand(void), poisson_rand(double);
extern double gammaln(double);
#else
extern int  init_rand();
extern void g_rand();
extern void g_frand();
extern void u_rand();
extern void u_frand();
extern double norm_rand(), uniform_rand(), poisson_rand();
extern double gammaln();
#endif

/*
 * init_rand() initiates random sequence by time point
 * and returns seed
 */
int init_rand()
{
	int start_point;
	long time(),tm;
	start_point = (long)((double)time(&tm) - 660912760.0) % 660912760l;
	srand(start_point);
	return start_point;
}

void g_rand(noise,n,lev)
double *noise;
double lev;
unsigned int n;
{
	register unsigned int i,j;
	double val;

	for(j=0 ; j < n ; j++ ) {
		val = 0.;
		for(i=0 ; i < _mq_sum; i++) {
		       val  += (double)rand() / MAX_RAND;
		}
		val = (val/_mq_sum - .5) * sqrt( _mq_sum * 12.);
		noise[j] = val * lev;
	}
}

/*
 *  Transformation random uniforms numbers to Gaussian
 */
void g_rand2(noise, n, level)
double *noise;          /* output array */
double level;           /* deviation    */
unsigned int n;
{
	register unsigned int j;
	double pi2 = 2 * M_PI;

	for(j=0 ; j < n ; j+=2 ) {
		double u1 = (double)rand() / MAX_RAND;
		double u2 = (double)rand() / MAX_RAND;
		double v1 = level * sqrt(-2. * log(u1));
		double v2 = pi2 * u2;
		noise[j]   = v1 * sin(v2);
		if(j+1 < n) noise[j+1] = v1 * cos(v2);
	}
	return;
}


#ifdef __GNUC__
void g_frand(float *noise, unsigned int n, float lev)
#else
void g_frand(noise, n, lev)
float *noise;
float lev;
unsigned int n;
#endif
{
	register unsigned int i,j;
	double val;

	for(j=0 ; j < n ; j++ ) {
		val = 0.;
		for(i=0 ; i < _mq_sum; i++) {
		       val  += (double)rand() / MAX_RAND;
		}
		val = (val/_mq_sum - .5) * sqrt( _mq_sum * 12.);
		noise[j] = val * lev;
	}
}

void u_rand(noise,n,lev, shift)
double *noise;
double lev, shift;
unsigned int n;
{
	register unsigned int j;

	for(j=0 ; j < n ; j++ )
		noise[j] = lev * ((double)rand() / MAX_RAND - shift);
}

#ifdef __GNUC__
void u_frand(float* noise, unsigned int n, float lev, float shift)
#else
void u_frand(noise,n,lev, shift)
float *noise;
float lev, shift;
unsigned int n;
#endif
{
	register unsigned int j;

	for(j=0 ; j < n ; j++ )
		noise[j] = lev * ((double)rand() / MAX_RAND - shift);
}

/* gaussian distribution [0, 1] */
double norm_rand()
{
	register unsigned int i;
	double val=0;

	for(i=0 ; i < _mq_sum; i++) {
	       val  += (double)rand() / MAX_RAND;
	}
	val = (val/_mq_sum - .5) * sqrt( _mq_sum * 12.);

	return val;
}

double norm_rand2()
{
	static int flag = 1;
	static double s = 0.0;
	double v;

	if (flag) {
		double pi2 = 2 * M_PI;
		double u1 = (double)rand() / MAX_RAND;
		double u2 = (double)rand() / MAX_RAND;
		double v1 = sqrt(-2. * log(u1));
		double v2 = pi2 * u2;
		v = v1 * sin(v2);
		s = v1 * cos(v2);
	} else {
		v = s;
	}
	flag = !flag;
	return v;
}


/* uniform distributed from 0 to 1 */
double uniform_rand()
{
	return  (double)rand() / MAX_RAND;
}

/*
 * Poisson distributed from 0 to 1
 *-----------------------------------------------------------
 *       Numerical Recipies in Fortran 77
 *       Second Edition.
 *       The Art of Scientific Computing
 *       W.H.Press, S.A.Teukoisky, W.T.Vetterling, Brian P.Flannery
 *       Cambridge University Press. 1992,
 *       V.1, Chapter 7.3
 *
 *               x^[m] e^-x
 *  q_x(m)dm =  ------------ dm, [m] - largest integer less than m.
 *                [m]!
 *
 *
 *
 *  usess gammln,ran1
 *
 *  Returns as a floating-point number an interger value that is a random
 *  deviate drawn from a Poisson distribution of mean xm, using ran1(idum)
 *  as a source of uniform random deviates
 *
 */

double poisson_rand(xm)
double xm;
{
    double m;
    double t, y;
    /* double gammln(), ran1() */
    static double alxm, g, sq;
    /* oldm: flag for whether xm has changed since last call */
    static double oldm = -1.00;

    if (xm > 12.)  {  /* Use direct method */
	if ( ABS(xm - oldm) < 1.e-5) {
	    oldm = xm;
	    g = exp (-xm);    /* If xm is new, compute the exponential  */
	}

	m = -1.;
	t =  1.;
lab_2:
	m += 1.;              /* Instead of adding exponential deviates */
	t *= uniform_rand();  /* it is equivavlent to multiply uniform  */
			      /* deviates. We never actually have to    */
	if (t > g)            /* take the log, merely compare to the    */
	     goto lab_2;      /* pre-computed exponential               */

    } else { /* Use rejection method */
	if(ABS(xm - oldm) > 1.e-5){ /* if xm has changed since the last */
	    oldm = xm ;             /* call, then pre-compute some      */
	    sq = sqrt(2.*xm);       /* functions that occur below       */
	    alxm = log(xm);

	    /* The function gammaln is the natural log of the */
	    /* the gamma function, as given in Paragraph 6.1  */
	    g = xm * alxm - gammaln(xm+1.);
	}
lab_1:
	/* y is a deviate from a Lorentzian comparison function */
	y = tan(M_PI * uniform_rand());

	/* m is ym shifted  and scaled */
	m = sq * y + xm;

	/* Reject if in regime of zero probablity */
	if(m <  0.)  goto lab_1;

	/* The trick for integer-valued distributions */
	m = (int)m;

       /*
	! The ratio of the desired distribution to the comaprison
	! function; we accept or reject by comparing it to another
	! uniform deviate. The factor 0.9 is chosed so that t never
	! exceeds 1.
	*/
	t=0.9 * (1.+y*y) * exp(m * alxm - gammaln(m+1.) - g);

	if (uniform_rand() > t)  goto lab_1;
    }

    return m;
}


#ifdef __GNUC__
void poi_rand(double* noise, unsigned int n, double lev)
#else
void poi_rand(noise,n,lev, shift)
double *noise;
double lev;
unsigned int n;
#endif
{
	register unsigned int j;

	for(j=0 ; j < n ; j++ )
		noise[j] = poisson_rand(lev);
	return;
}


/*
 *
 *-----------------------------------------------------------
 *       Numerical Recipies in Fortran 77
 *       Second Edition.
 *       The Art of Scientific Computing
 *       W.H.Press, S.A.Teukoisky, W.T.Vetterling, Brian P.Flannery
 *       Cambridge University Press. 1992,
 *       V.1, Chapter 6.1
 *
 *    returns the value ln[G(xx)] for xx > 0.
 */
double gammaln(xx)
double xx;
{
	double gln;
	register int j;
	double ser, tmp, x, y;
/*
 C Internal arithmetic will be done in double precision, a nicety that you
 C can omit if five-figure accuracy is good enough
 */
	static double stp = 2.506628274631000e0,
		      cof[6] = {
			 76.18009172947146e0, -86.505320322941677e0,
			 24.01409824083091e0,  -1.231739572450155e0,
			.1208650973866179e-2,  -.5395239384953e-5
		      };

	x = xx;
	y = x;
	tmp = x + 5.5;
	tmp = (x+0.5e0) * log(tmp) - tmp;
	ser = 1.000000000190015e0;
	for (j=0; j<6; j++) {
		y += 1.e0;
		ser += cof[j]/y;
	}

	gln = tmp + log(stp*ser/x);

	return gln;
}


int _nst(a,b)
int *a, *b;
{
	return ( *a < *b ? -1 : (*a > *b ? 1 : 0));
}
/* generate rand_lines unique numbers in rand_arr */
void random_numbers(arr, n_sel, max, seed)
int *arr;
int n_sel, max, seed;
{
    int j, ja=0, jb, start_point;
    long time(),tm;
    if (seed < 0)
	start_point = (long)((double)time(&tm) - 660912760.0) % 660912760l;
    else
	start_point = seed;
    srand(start_point);

again:
    for(j=ja ; j < n_sel ; j++ )
	    arr[j] = max * uniform_rand() + 0.5;

    qsort(arr, n_sel, sizeof(*arr), (int (*)())_nst);

    /* remove equal indices */
    jb = n_sel - 1;
    for(j=0; j < jb ; j++ ) {
	    if (arr[j] == arr[j+1]) {
		int k;
		/* squeezing */
		for(k=j; k<jb; k++)  arr[k] = arr[k+1];
		jb--;
	    }
    }
    if (jb+1 != n_sel) {
	ja = jb;
	goto again;
    }

    return;
}
