#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "glesp.h"

#define PR      fprintf
#define ER      stderr

extern void setopt(int, char **), usage(void);

int  bitpix_map = -32; /* default: Float Point Pixels for map */
char inp_file[161], outp_file[161];
char tmpcat[161];
int theta_grid=1;

int nx = 0,
    np = 0;

char *myname="rdmap";

int main(argc, argv)
int argc;
char **argv;
{
	register int count = 0;
	double *f = (double *)0,
	       *w  = (double *)0,
	       *x  = (double *)0;
	int *nph=(int *)0;
	FILE *fp_in, *fp_out;
	char x_type[40], grid_type[40];
	char str[161];

	/* select grid output keyword values */
	strcpy(x_type, "COS(THETA)");
	strcpy(grid_type, "PHI(X)+LEGENDRE");

	inp_file[0] = outp_file[0] = '\0';

	/* loop over external parameters */
	setopt(argc, argv);

	if ((f = (double *)calloc((unsigned)(nx*np), sizeof *f)) ==
	    (double *)0) {
nomem:
		PR(ER, "%s: no memory for arrays of dimension (%d,%d)...\n",
		    myname, nx, np);
		exit(1);
	}
	if ((x = (double *)calloc((unsigned)nx, sizeof *x)) == (double *)0)
		goto nomem;
	if ((w = (double *)calloc((unsigned)nx, sizeof *w)) == (double *)0)
		goto nomem;
	if ((nph = (int *)calloc((unsigned)nx, sizeof *nph)) == (int *)0)
		goto nomem;

	coord(1, -10, -1., 1., M_PI*2., x, w, nx, nph, np);

	/* open input file */
	if (*inp_file == '\0')  fp_in = stdin;
	else {
	    if (strcmp(inp_file,"-")==0 || strcmp(outp_file,"stdin")==0)
		fp_out = stdin;
	    else if ((fp_in = fopen(inp_file, "r")) == (FILE *)0) {
		PR(ER, "%s: can not open file '%s'\n", myname, inp_file);
		exit(1);
	    }
	}

	/* read input data */
	while (fgets(str, 160, fp_in) != NULL) {
		double theta, phi, t;
		int n_pix;

		if (*str == '#' || *str == '\n' || *str == '\r')
			continue;

		sscanf(str, "%lf %lf %lf", &theta, &phi, &t);
		n_pix = ang2n(theta, phi, x, nx, np, nph);

		f[n_pix] = t;
		count++;
	}

	/* open output file */
	if (*outp_file == '\0')  fp_out = stdout;
	else {
	    if (strcmp(outp_file,"-")==0 || strcmp(outp_file,"stdout")==0)
		fp_out = stdout;
	    else if ((fp_out = fopen(outp_file, "w")) == (FILE *)0) {
		PR(ER, "%s: can not create file '%s'\n", myname, outp_file);
		exit(1);
	    }
	}

	/*  write map  */
	ff_write_ext2_map(f, x, nx, np, nph, fp_out, x_type, grid_type);

	/* close files */
	if (fp_in  != stdin)   fclose(fp_in);
	if (fp_out != stdout)  fclose(fp_out);

	/* free memory */
	free((void *)f);
	free((void *)x);
	free((void *)nph);

	exit(0);
}

void setopt(argc, argv)
int argc;
char **argv;
{
    register int i;
    myname = argv[0];

    for(i=1; i<argc; i++) {
	switch(argv[i][0]) {
	case '-':
		switch(argv[i][1]) {
		case 'h':
			usage();
			break;
		case 'i':
			/* input file for F(x,phi) */
			strcpy(inp_file, argv[++i]);
			break;
		case 'n':
			switch(argv[i][2]) {
			case 'x': /* number of points along cos(theta) */
				nx = atoi(argv[++i]);
				break;
			case 'p': /* number of points along phi */
				np = atoi(argv[++i]);
				break;
			default :
				/* number of calculation  */
				nx = atoi(argv[++i]);
				break;
			}
			break;
		case 'o':
			/* input file for F(x,phi) */
			strcpy(outp_file, argv[++i]);
			break;
		default:
			break;
		}
		break;
	case '+':
		switch(argv[i][1]) {
		case 'h':
		default:
			usage();
			break;
		}
		break;
	default:
		strcpy(inp_file, argv[++i]);
		break;
	}
    }

    return;
}

void usage()
{
PR(ER, "usage: %s -i inp_file -o outp_file -nx nx -np np\n", myname);
exit(2);
}
