#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <ff.h>
#include "glesp.h"

#define PR      fprintf
#define ER      stderr

#ifdef  __DARWIN__
int    naxis2;
#else
extern int    naxis2;
#endif
extern double crval2;
extern double cdelt2;
extern double crpix2;
extern char   ctype2[61];

extern char *myname;
/*
 * read a_lm ASCII table
 * data format:
 *              L  M  Real_part Imagine_part
 * where
 *              L can be a number, or limits n-k
 *              M can be a number, or 'a' - all, 'e' - even, 'o' - odd
 *              Real_part    can be a number, or '-' - no value
 *              Imagine_part can be a number, or '-' - no value
 * examples:
 *              2-3000   all   1  1   - l=2,3000, m=0,l,         alm=(1,1)
 *              1000     e4-10 1 -1   - l=1000,   m=4,10,2(even) alm=(1,-1)
 *              1000     4-10o 1 -1   - l=1000,   m=5,10,2(odd)  alm=(1,-1)
 *              2-100    odd   0  1   - l=2,100,  m=1,l,2 (odd)  alm=(0,-1)
 *
 *  created by VO (vo@sao.ru, oleg@tac.dk), 14.11.2002, TAC
 *      last modification: 19-Oct-2004, SAO; 23-Oct-2005, ASC
 */

/* read Lmax from ASCII data */
int getLmx(s_l)
char *s_l;
{
    char *ps = s_l;
    int l_max;

    /* if numbers separated by '-' */
    /* they are low & high limits  */
    if ( (ps = strchr(s_l, '-')) != NULL ) {
	 l_max = atoi(++ps);
    } else {
	 l_max = atoi(s_l);
    }
    return l_max;
}

int read_alm_table(filename, alm, nalm, nbuf, lmax)
char *filename;
double **alm;
int *nalm;
int *nbuf;
int *lmax;
{
    register int counter = 0;
    int alm_ffile_flag = 0;
    int lmx = -1;
    FILE *fp = (FILE *)0;
    char str[161];

    /* open file */
    if ((fp = fopen(filename, "r")) == (FILE *)0) {
	    PR(ER, "%s: can not open file '%s'\n", myname, filename);
	    exit(1);
    }

    /* strings count */
    while (fgets(str, 160, fp) != NULL) {
	register int l;
	char s_l[20];

	/* check if this file is F-file with a_{lm}s */
	if (!counter && strncmp(str, "SIMPLE=", 7) == 0) {
		alm_ffile_flag = 1;
		fclose(fp);
		fp = (FILE *)0;
		break;
	}   else   {
		/* check if this file is FITS-file with a_{lm}s */
		if (!counter && strncmp(str, "SIMPLE  = ", 10) == 0) {
			alm_ffile_flag = 2;
			fclose(fp);
			fp = (FILE *)0;
			break;
		}
	}

	if (*str == '\n' || *str == '\r' || *str == '#'|| *str == '%')
		continue;

	sscanf(str, "%s", s_l);
	l = getLmx(s_l);
	if (l > lmx)   lmx = l;
	counter++;
    }

    if ( alm_ffile_flag ) {
	if (*alm == (double *)0) {
	    if (*lmax == 0 || *lmax == -1) {
		    if (alm_ffile_flag == 1)
			    *lmax = get_lmax(filename);
		    else
			    *lmax = Fget_lmax(filename);
		    *nbuf = 2 * (*lmax + 1);
		    *nalm = *lmax + 1;
	    } else if (*lmax < -1) {
		    int Lmax = -(*lmax);
		    *nbuf = 2 * (Lmax + 1);
		    *nalm = Lmax + 1;
		    *lmax = Lmax;
	    }
	    if((*alm=(double *)calloc((unsigned)(*nbuf * *nalm),
		 sizeof **alm)) == (double *)0) {
nomem:
		    PR(ER,"%s: no memory for arrays[%d]\n",
			myname, *nbuf * *nalm);
		    exit(1);
	    }
	}
	if (alm_ffile_flag == 1)
		ff_read_alm(*alm, *nbuf <0? -(*nbuf): *nbuf, lmax,filename);
	else
		in_glesp_alm(*alm, lmax, *nbuf <0? -(*nbuf): *nbuf,filename);
    } else {
	int Lmax = *lmax < -1 ? -(*lmax) : lmx;
	fseek(fp, 0l, 0);

	if ((*nalm < 0 ? -(*nalm) : *nalm) < Lmax + 2) {
	    if (*alm != (double *)0) {
		free((void *)*alm); /* free memory if buffer is small */
		*alm = (double *)0;
	    }
	}

	if ( *alm == (double *)0) {
	   *nalm = Lmax + 2;
	   *nbuf = *nalm * 2;

	   /* find memory for arrays */
	   if ((*alm = (double *)calloc((unsigned)(*nbuf * *nalm),
	       sizeof **alm)) == (double *)0)
		  goto nomem;
	}

	while (fgets(str, 160, fp) != NULL) {
	    if (*str == '\n'||*str == '\r'||*str == '#'||*str == '%')
		    continue;

	    str2alm(*alm, *nbuf < 0 ? -(*nbuf) : *nbuf, Lmax, str);
	}
	*lmax = Lmax;
    }
    if(fp != (FILE *)0) fclose(fp);

    return *lmax;
}

int read_alm_3table(filename, alm, nalm, nbuf, lmax)
char *filename;
double **alm;
int *nalm;     /* =Nmx */
int *nbuf;
int *lmax;
{
    double *alm_e = (double *)0, *alm_b = (double *)0;

    if (*nalm < 0) {
	if (*nbuf == 0)
	    *nbuf = 2 * *nalm;
    }
    if (*nalm < 0 && *nbuf < 0 && *alm == (double*)0) {
	if((*alm=(double*)calloc((unsigned)(*nbuf*(*nalm)),sizeof **alm))==
	    (double *)0) {
		PR(ER, "%s: no memory for arrays of %d elements...\n",
			myname, *nbuf*(*nalm));
		exit(1);
	}
    }

    read_ieb_alm(alm, &alm_e, &alm_b, lmax, nbuf, filename);

    if (*nalm >= 0)
	*nalm = *lmax + 2;

    if (alm_e != (double *)0)  free((void *)alm_e);
    if (alm_b != (double *)0)  free((void *)alm_b);

    PR(ER, "%s:\n    input file contains polarization data too, \n", myname);
    PR(ER, "    but only anisotropy will be processed.\n\n");

    return *lmax;
}

int Sget_lmax(filename)
char *filename;
{
    int lmx = -1;
    FILE *fp = (FILE *)0;
    char str[161];

    /* open file */
    if ((fp = fopen(filename, "r")) == (FILE *)0) {
	    PR(ER, "%s: can not open file '%s'\n", myname, filename);
	    exit(1);
    }

    /* strings count */
    while (fgets(str, 160, fp) != NULL) {
	register int l;
	char s_l[20];

	if (*str == '\n' || *str == '\r' || *str == '#'|| *str == '%')
		continue;

	sscanf(str, "%s", s_l);
	l = getLmx(s_l);
	if (l > lmx)   lmx = l;
    }
    fclose(fp);
    return lmx;
}


void str2alm_T(alm, nbuf, lmax, str)
double *alm;
int nbuf, lmax;
char *str;
{
    register int l=0,m;
    char s_l[20],s_m[20],s_re[50],s_im[50];
    int l_min, l_max, m_min, m_max, dm;
    double re, im;
    register char *ps;

    sscanf(str, "%s%s %s%s", s_l, s_m, s_re, s_im);

    /* if numbers separated by '-' */
    /* they are low & high limits  */
    if ( (ps = strchr(s_l, '-')) != NULL ) {
	 l_min = atoi(s_l);
	 l_max = atoi(++ps);
    } else {
	 l_min = l_max = atoi(s_l);
    }
    if (l_max > lmax)   l_max = lmax;
    if (l_min > lmax)
	    return;

    for(l=l_min; l<=l_max; l++) {
	re = atof(s_re);
	im = atof(s_im);

	/* if numbers separated by '-' */
	/* they are low & high limits  */
	ps = s_m;
	if ( *ps == '-' )  ps++;
	if ( *ps != 'a' ) {
		register char *cs;
		int odd_flag  = 0,
		    even_flag = 0;

		dm = strlen(ps) - 1;
		switch (ps[dm]) {
		case 'o':  odd_flag  = 1; dm=2; break;
		case 'e':  even_flag = 1; dm=2; break;
		default :  break;
		}
		switch (ps[0]) {
		case 'o':  odd_flag  = 1; dm=2; ps++; break;
		case 'e':  even_flag = 1; dm=2; ps++; break;
		default :  break;
		}
		if (dm != 2)  dm=1;

		if ( (cs = strchr(ps, '-')) != NULL ) {
		     m_min = atoi(ps);
		     cs++;
		     if (*cs == 'a')
			m_max = l;
		     else
			m_max = atoi(cs);
		     if (odd_flag)       {m_min = (m_min / 2) * 2 + 1;}
		     else if (even_flag) {m_min = (m_min + 1) / 2 * 2;}
		     if(m_min > l)
				continue;
		} else {
		     if (odd_flag)       {m_min = 1; m_max = l;}
		     else if (even_flag) {m_min = 2; m_max = l;}
		     else m_min = m_max = atoi(ps);
		}
	} else {
		/* if (*s_m == 'a') */
		/* all m            */
		m_min = 0; m_max = l; dm = 1;
	}
	for(m=m_min; m<=m_max; m+=dm) {
		if (strcmp(s_re,"-") != 0)
		     *(alm+m*nbuf+l) = re;
		if ( m == 0 ) {
			*(alm+(m+lmax+1)*nbuf+l) = 0;
		} else {
			if (strcmp(s_im,"-") != 0)
			     *(alm+(m+lmax+1)*nbuf+l) = im;
		}
	}
    }
    return;
}

void str2alm(alm, nbuf, lmax, str)
double *alm;
int nbuf, lmax;
char *str;
{
    register int l=0,m;
    char s_l[20],s_m[20],s_re[50],s_im[50];
    int l_min, l_max, m_min, m_max, dm;
    double re, im;
    register char *ps;

    sscanf(str, "%s%s %s%s", s_l, s_m, s_re, s_im);

    /* if numbers separated by '-' */
    /* they are low & high limits  */
    if ( (ps = strchr(s_l, '-')) != NULL ) {
	 l_min = atoi(s_l);
	 l_max = atoi(++ps);
    } else {
	 l_min = l_max = atoi(s_l);
    }
    if (l_max > lmax)   l_max = lmax;
    if (l_min > lmax)
	return;

    for(l=l_min; l<=l_max; l++) {
	register int ll = nbuf * l;

	re = atof(s_re);
	im = atof(s_im);
	/* if numbers separated by '-' */
	/* they are low & high limits  */
	ps = s_m;
	if ( *ps == '-' )  ps++;
	if ( *ps != 'a' ) {
		register char *cs;
		int odd_flag  = 0,
		    even_flag = 0;

		dm = strlen(ps) - 1;
		switch (ps[dm]) {
		case 'o':  odd_flag  = 1; dm=2; break;
		case 'e':  even_flag = 1; dm=2; break;
		default :  break;
		}
		switch (ps[0]) {
		case 'o':  odd_flag  = 1; dm=2; ps++; break;
		case 'e':  even_flag = 1; dm=2; ps++; break;
		default :  break;
		}
		if (dm != 2)  dm=1;

		if ( (cs = strchr(ps, '-')) != NULL ) {
		     m_min = atoi(ps);
		     cs++;
		     if (*cs == 'a')
			m_max = l;
		     else
			m_max = atoi(cs);
		     if (odd_flag)       {m_min = (m_min / 2) * 2 + 1;}
		     else if (even_flag) {m_min = (m_min + 1) / 2 * 2;}
		     if(m_min > l)
				continue;
		} else {
		     if (odd_flag)       {m_min = 1; m_max = l;}
		     else if (even_flag) {m_min = 2; m_max = l;}
		     else m_min = m_max = atoi(ps);
		}
	} else {
		/* if (*s_m == 'a') */
		/* all m            */
		m_min = 0; m_max = l; dm = 1;
	}

	for(m=m_min; m<=m_max; m+=dm) {
		if (strcmp(s_re,"-") != 0)
		     *(alm+ll+m) = re;
		if ( m == 0 ) {
			*(alm+ll+m + lmax+1) = 0;
		} else {
			if (strcmp(s_im,"-") != 0)
			     *(alm+ll+m + lmax+1) = im;
		}
	}
    }
    return;
}

void ff_read_alm(alm, nbuf, lmax, filename)
double *alm;
int *lmax;
int nbuf;
char *filename;
{
	register int i, l, m;
	int tfields = 0;
	int bitp1, bitp2, bitp3;
	char tform1[30],
	     tform2[30],
	     tform3[30];
	char str[161];
	int  lmx = -1, lmax_input = *lmax;
	FILE *fp;

	/* open file */
	if ((fp = fopen(filename, "r")) == (FILE *)0) {
		PR(ER, "%s: can not open file '%s'\n", myname, filename);
		exit(1);
	}
	/* read 1st head */
	Hinit(fp);

	if (EVget("EXTEND") != NULL)
	    strcpy(str, EVget("EXTEND"));

	if (strcmp(str, "T") != 0) {
	    PR(ER, "%s: bad extension in file '%s'\n", myname, filename);
	    exit(1);
	}

	/* Free memory from 1st head */
	Hfree();

	/****************************************/
	/* read basic keywods from the 2nd head */
	/****************************************/
	Hinit(fp);

	if (EVget("XTENSION") != NULL)
	    strcpy(str, EVget("XTENSION"));
	else {
	    PR(ER, "%s: no extension in file '%s'\n", myname, filename);
	    exit(1);
	}
	if (strcmp(str, "BINTABLE") != 0) {
	    PR(ER, "%s: can work only with BINTABLE extension\n", myname);
	    exit(1);
	}
	if (EVget("NAXIS") != NULL)
		naxis  = Skey_get("NAXIS");
	if (EVget("NAXIS1") != NULL)
		naxis1  = Lkey_get("NAXIS1");
	if (EVget("NAXIS2") != NULL)
		naxis2  = Lkey_get("NAXIS2");

	if (EVget("TFIELDS") != NULL)
		tfields = Lkey_get("TFIELDS");
	if (tfields != 3) {
	    PR(ER, "%s: TFILEDS value must be equal to 3 in a_lm file '%s'\n",
		myname, filename);
	    exit(1);
	}

	if (EVget("TFORM1") != NULL) {
		strcpy(tform1, EVget("TFORM1"));
		if (strcmp(tform1, "1L") == 0 || strcmp(tform1, "1J") == 0) {
		    bitp1 = LONGPIX;
		} else if (strcmp(tform1, "1S") == 0) {
		    bitp1 = SHORTPIX;
		} else {
		    PR(ER, "%s: bad value for TFORM1. Sorry...\n", myname);
		    exit(1);
		}
	} else {
		PR(ER, "%s: no parameter TFORM1. Sorry...\n", myname);
		exit(1);
	}
	if (EVget("TFORM2") != NULL) {
		strcpy(tform2, EVget("TFORM2"));
		if (strcmp(tform2, "1F") == 0 || strcmp(tform2, "1E") == 0) {
		    bitp2 = FLOATPIX;
		} else if (strcmp(tform2, "1D") == 0) {
		    bitp2 = LONGPIX;
		} else {
		    PR(ER, "%s: bad value for TFORM2. Sorry...\n", myname);
		    exit(1);
		}
	} else {
		PR(ER, "%s: no parameter TFORM2. Sorry...\n", myname);
		exit(1);
	}
	if (EVget("TFORM3") != NULL) {
		strcpy(tform3, EVget("TFORM3"));
		if (strcmp(tform2, "1F") == 0 || strcmp(tform2, "1E") == 0) {
		    bitp3 = FLOATPIX;
		} else if (strcmp(tform3, "1D") == 0) {
		    bitp3 = LONGPIX;
		} else {
		    PR(ER, "%s: bad value for TFORM3. Sorry...\n", myname);
		    exit(1);
		}
	} else {
		PR(ER, "%s: no parameter TFORM3. Sorry...\n", myname);
		exit(1);
	}
	if (EVget("LMAX") != NULL) {
		*lmax = atoi(EVget("LMAX"));
	}

	if (lmax_input > 0 && lmax_input <= *lmax) {
	     *lmax = lmax_input;
	} else {
	     if (nbuf < (*lmax + 1)*2) {
		 if ( (lmax_input + 1) * 2 > nbuf)
		     lmax_input = nbuf / 2 - 1;
		 *lmax = lmax_input;
	     } else {
	       if (*lmax > lmax_input) {
		 PR(ER, "%s: check set of lmax: lmax=%d, nbuf=%d\n",
			myname, *lmax, nbuf);
		 exit(1);
	       }
	     }
	}

	/* Free memory from 1nd head */
	Hfree();
	/* get values from file */
	for (i=0; i<naxis2; i++) {
	    static int flag = 1;
	    register int bb, nlm;

	    bitpix = bitp1;
	    nlm = get_pix(fp)+0.5;
	    l = (int)sqrt((double)(nlm-1));

	    if (l > *lmax)
		break;
	    if (l*2+1 > nbuf) {
		if (flag) {
		    PR(ER, "%s: attention ! l(=%d) is larger ", myname, l);
		    PR(ER, "than array dimension.\n");
		}
		flag = 0;
		break;
	    }
	    m = nlm - 1 - l*l - l;
	    bb = l * nbuf;

	    /* Real part */
	    bitpix = bitp2;
	    alm [bb+m]        = get_pix(fp);
	    if (feof(fp)) {
		 PR(ER, "%s: unexpected end of file at l=%d and m=%d\n",
		    myname, l, m);
		 break;
	    }
	    /* Imagine part */
	    bitpix = bitp3;
	    alm [bb+m+ *lmax+1] = get_pix(fp);
	    if (feof(fp)) {
		 PR(ER, "%s: unexpected end of file at l=%d and m=%d\n",
		    myname, l, m);
		 break;
	    }
	    if (l > lmx)   lmx = l;
	}
	if (*lmax <= 0)  *lmax = lmx;

	fclose(fp);
	return;
}

void read_alm_str (alm_str, alm, nbuf, lmax)
char *alm_str;
double *alm;
int nbuf, lmax;
{
	char buf_str[50], form[30];

	/* read l, m, alm_re, alm_im  strings */
	/* and write them in another format   */
	skip_wd(0, alm_str, form);
	strcpy(buf_str, form); strcat(buf_str, " ");
	skip_wd(1, alm_str, form);
	strcat(buf_str, form); strcat(buf_str, " ");
	skip_wd(2, alm_str, form);
	strcat(buf_str, form); strcat(buf_str, " ");
	skip_wd(3, alm_str, form);
	strcat(buf_str, form);

	/* fill in a_{lm} array */
	str2alm(alm, nbuf, lmax, buf_str);

	return;
}

int get_lmax_str(str)
char *str;
{
    char s_l[20];
    int lmax;
    register char *ps;

    skip_wd(0, str, s_l);

    /* if numbers separated by '-' */
    /* they are low & high limits  */
    if ( (ps = strchr(s_l, '-')) != NULL ) {
	 lmax = atoi(++ps);
    } else {
	 lmax = atoi(s_l);
    }
    return lmax;
}

/* get Lmax from F-file */
int get_lmax(filename)
char *filename;
{
	register int i, l, m;
	int tfields = 0;
	int bitp1, bitp2, bitp3;
	char tform1[30],
	     tform2[30],
	     tform3[30];
	char str[161];
	int  lmx = -1;
	FILE *fp;

	/* open file */
	if ((fp = fopen(filename, "r")) == (FILE *)0) {
		PR(ER, "%s: can not open file '%s'\n", myname, filename);
		exit(1);
	}
	/* read 1st head */
	Hinit(fp);

	if (EVget("EXTEND") != NULL)
	    strcpy(str, EVget("EXTEND"));

	if (strcmp(str, "T") != 0) {
	    PR(ER, "%s: bad extension in file '%s'\n", myname, filename);
	    exit(1);
	}

	/* Free memory from 1st head */
	Hfree();

	/****************************************/
	/* read basic keywods from the 2nd head */
	/****************************************/
	Hinit(fp);

	if (EVget("XTENSION") != NULL)
	    strcpy(str, EVget("XTENSION"));
	else {
	    PR(ER, "%s: no extension in file '%s'\n", myname, filename);
	    exit(1);
	}
	if (strcmp(str, "BINTABLE") != 0) {
	    PR(ER, "%s: can work only with BINTABLE extension\n", myname);
	    exit(1);
	}
	if (EVget("NAXIS") != NULL)
		naxis  = Skey_get("NAXIS");
	if (EVget("NAXIS1") != NULL)
		naxis1  = Lkey_get("NAXIS1");
	if (EVget("NAXIS2") != NULL)
		naxis2  = Lkey_get("NAXIS2");

	if (EVget("TFIELDS") != NULL)
		tfields = Lkey_get("TFIELDS");
	if (tfields != 3) {
	    PR(ER, "%s: TFILEDS value must be equal to 3 in a_lm file '%s'\n",
		myname, filename);
	    exit(1);
	}

	if (EVget("TFORM1") != NULL) {
		strcpy(tform1, EVget("TFORM1"));
		if (strcmp(tform1, "1L") == 0 || strcmp(tform1, "1J") == 0) {
		    bitp1 = LONGPIX;
		} else if (strcmp(tform1, "1S") == 0) {
		    bitp1 = SHORTPIX;
		} else {
		    PR(ER, "%s: bad value for TFORM1. Sorry...\n", myname);
		    exit(1);
		}
	} else {
		PR(ER, "%s: no parameter TFORM1. Sorry...\n", myname);
		exit(1);
	}
	if (EVget("TFORM2") != NULL) {
		strcpy(tform2, EVget("TFORM2"));
		if (strcmp(tform2, "1F") == 0 || strcmp(tform2, "1E") == 0) {
		    bitp2 = FLOATPIX;
		} else if (strcmp(tform2, "1D") == 0) {
		    bitp2 = LONGPIX;
		} else {
		    PR(ER, "%s: bad value for TFORM2. Sorry...\n", myname);
		    exit(1);
		}
	} else {
		PR(ER, "%s: no parameter TFORM2. Sorry...\n", myname);
		exit(1);
	}
	if (EVget("TFORM3") != NULL) {
		strcpy(tform3, EVget("TFORM3"));
		if (strcmp(tform2, "1F") == 0 || strcmp(tform2, "1E") == 0) {
		    bitp3 = FLOATPIX;
		} else if (strcmp(tform3, "1D") == 0) {
		    bitp3 = LONGPIX;
		} else {
		    PR(ER, "%s: bad value for TFORM3. Sorry...\n", myname);
		    exit(1);
		}
	} else {
		PR(ER, "%s: no parameter TFORM3. Sorry...\n", myname);
		exit(1);
	}
	/* Free memory from 1nd head */
	Hfree();

	/* get values from file */
	for (i=0; i<naxis2; i++) {
	    register int nlm;

	    bitpix = bitp1;
	    nlm = get_pix(fp)+0.5;
	    l = (int)sqrt((double)(nlm-1));
	    m = nlm - 1 - l*l - l;

	    /* Real part */
	    bitpix = bitp2;
	    get_pix(fp);
	    if (feof(fp)) {
		 PR(ER, "%s: unexpected end of file at l=%d and m=%d\n",
		    myname, l, m);
		 break;
	    }
	    /* Imagine part */
	    bitpix = bitp3;
	    get_pix(fp);
	    if (feof(fp)) {
		 PR(ER, "%s: unexpected end of file at l=%d and m=%d\n",
		    myname, l, m);
		 break;
	    }
	    if (l > lmx)   lmx = l;
	}
	fclose(fp);

	return lmx;
}

extern void  str2phase(double *, int, char *);

int read_ph_table(filename, phase, lmax)
char *filename;
double **phase;
int *lmax;
{
    register int counter = 0;
    int lmx = -1;
    FILE *fp;
    char str[161];

    /* open file */
    if ((fp = fopen(filename, "r")) == (FILE *)0) {
	    PR(ER, "%s: can not open file '%s'\n", myname, filename);
	    exit(1);
    }

    /* strings count */
    while (fgets(str, 160, fp) != NULL) {
	register int l;
	char s_l[20];

	if (*str == '\n' || *str == '\r' || *str == '#'|| *str == '%')
		continue;

	sscanf(str, "%s", s_l);
	l = getLmx(s_l);
	if (l > lmx)   lmx = l;
	counter++;
    }
    fseek(fp, 0l, 0);

    if (lmx > *lmax)  *lmax = lmx;

    if ( *phase == (double *)0) {
       /* find memory for arrays */
       if ((*phase = (double *)calloc((unsigned)((*lmax+2)*(*lmax+2)),
	   sizeof **phase)) == (double *)0) {
	       PR(ER,"%s: can not find memory for array...\n", myname);
	       exit(1);
       }
       PR(ER,"memory for %d: O'k.\n", (*lmax+2)*(*lmax+2));
    }

    while (fgets(str, 160, fp) != NULL) {
	if (*str == '\n'||*str == '\r'||*str == '#'||*str == '%')
		continue;

	str2phase(*phase, *lmax, str);

    }

    fclose(fp);

    return *lmax;
}
void  str2phase(phase, lmax, str)
double *phase;
int lmax;
char *str;
{
    register int l=0,m;
    char s_l[20], s_m[20], s_phase[50];
    int l_min, l_max, m_min, m_max, dm;
    register char *ps;

    sscanf(str, "%s%s %s", s_l, s_m, s_phase);

    /* if numbers separated by '-' */
    /* they are low & high limits  */
    if ( (ps = strchr(s_l, '-')) != NULL ) {
	 l_min = atoi(s_l);
	 l_max = atoi(++ps);
    } else {
	 l_min = l_max = atoi(s_l);
    }
    if (l_max > lmax)   l_max = lmax;
    if (l_min > lmax)   l_min = lmax;

    for(l=l_min; l<=l_max; l++) {
	register int ll = (lmax+1) * l;

	double ph = atof(s_phase);

	/* if numbers separated by '-' */
	/* they are low & high limits  */
	ps = s_m;
	if ( *ps == '-' )  ps++;
	if ( *ps != 'a' )  {
		register char *cs;
		int odd_flag  = 0,
		    even_flag = 0;

		dm = strlen(ps) - 1;
		switch (ps[dm]) {
		case 'o':  odd_flag  = 1; dm=2; break;
		case 'e':  even_flag = 1; dm=2; break;
		default :  break;
		}
		switch (ps[0]) {
		case 'o':  odd_flag  = 1; dm=2; ps++; break;
		case 'e':  even_flag = 1; dm=2; ps++; break;
		default :  break;
		}
		if (dm != 2)  dm=1;

		if ( (cs = strchr(ps, '-')) != NULL ) {
		     m_min = atoi(ps);
		     cs++;
		     if (*cs == 'a')
			m_max = l;
		     else
			m_max = atoi(cs);
		     if (odd_flag)       {m_min = (m_min / 2) * 2 + 1;}
		     else if (even_flag) {m_min = (m_min + 1) / 2 * 2;}
		     if(m_min > l)
			    continue;
		} else {
		     if (odd_flag)       {m_min = 1; m_max = l;}
		     else if (even_flag) {m_min = 2; m_max = l;}
		     else m_min = m_max = atoi(ps);
		}
	} else {
		/* if (*s_m == 'a') */
		/* all m            */
		m_min = 0; m_max = l; dm = 1;
	}

	for(m=m_min; m<=m_max; m+=dm) {
		if (strcmp(s_phase, "-") != 0)
		     *(phase+ll+m) = ph;
	}
    }
    return;
}

/*
 * shift alm-s inside one buffer
 */
void alm2alm(alm, lmax_old, nbuf_old, lmax_new, nbuf_new)
double **alm;
int *lmax_old, *nbuf_old;
int lmax_new, nbuf_new;
{
    register int l, m;
    int flag = 1;

    if (nbuf_new < 0) { nbuf_new = -nbuf_new; flag = 0; }

    if ( !flag ) {
	/* create temporal arrays */
	double *a;
	l = (lmax_new+1) * nbuf_new;
	if ((a = (double *)calloc((unsigned)l, sizeof *a)) == (double *)0) {
	    PR(ER, "%s: alm2alm: no memory for temporal buffer...\n", myname);
	    exit(1);
	}

	for(l=0; l<= lmax_new; l++) {
	    int lo = l * *nbuf_old, ln = l * nbuf_new;
	    for(m=0; m<=l; m++)  {
		a[ln+m] = (*alm)[lo+m];
		a[ln+m+lmax_new+1] = (*alm)[lo+m + *lmax_old+1];
	    }
	}
	for(l=0; l<= lmax_new; l++) {
	    int ln = l * nbuf_new;
	    for(m=0; m<=l; m++)  {
		(*alm)[ln+m]            = a[ln+m];
		(*alm)[ln+m+lmax_new+1] = a[ln+m + lmax_new+1];
	    }
	}
	free((void *)a);

//      for(l=lmax_new; l>= 0; l--) {
//          int lo = l * *nbuf_old, ln = l * nbuf_new;
//          for(m=0; m<=l; m++)  {
//              (*alm)[ln+m] = (*alm)[lo+m];
//              (*alm)[ln+m+lmax_new+1] = (*alm)[lo+m + *lmax_old+1];
//              if (lo > ln) {
//                  (*alm)[lo+m] = 0.0;
//                  if (*lmax_old > lmax_new)
//                      (*alm)[lo+m + *lmax_old+1] = 0.0;
//              }
//          }
//      }

    } else if ((*lmax_old+1) * *nbuf_old > (lmax_new+1) * nbuf_new) {
	for(l=0; l<= lmax_new; l++) {
	    int lo = l * *nbuf_old, ln = l * nbuf_new;
	    for(m=0; m<=l; m++)  {
		(*alm)[lo+m] = (*alm)[lo+m];
		(*alm)[lo+m+lmax_new+1] = (*alm)[lo+m + *lmax_old+1];
	    }
	    for(m=0; m<=l; m++)  {
		(*alm)[ln+m] = (*alm)[lo+m];
		(*alm)[ln+m+lmax_new+1] = (*alm)[lo+m + lmax_new+1];
	    }
	}
    } else if ((*lmax_old+1) * *nbuf_old < (lmax_new+1) * nbuf_new) {
	double *alm_new;
	/* find memory for arrays */
	if ((alm_new = (double *)calloc((unsigned)((lmax_new+2)*nbuf_new),
	    sizeof *alm_new)) == (double *)0) {
		PR(ER,"%s: can not find memory for alm-array...\n", myname);
		exit(1);
	}
	for(l=0; l<= *lmax_old; l++) {
	    int lo = l * *nbuf_old, ln = l * nbuf_new;
	    for(m=0; m<=l; m++)  {
		alm_new[ln+m] = (*alm)[lo+m];
		alm_new[ln+m+lmax_new+1] = (*alm)[lo+m + *lmax_old+1];
	    }
	}
	free((void *)(*alm));
	*alm = alm_new;
    }
    *lmax_old = lmax_new;
    *nbuf_old = nbuf_new;

    return;
}

/* re-allocate alm in old array with new lmax */
void realloc_alm(alm, nbuf, lmax_old, lmax_new)
double *alm;
int nbuf;
int lmax_old, lmax_new;
{
    register int l, m, lo1=lmax_old+1, ln1=lmax_new+1;

    if (lmax_old < lmax_new) {
	for(l=0; l<=lmax_new; l++) {
	    int ll = l * nbuf;
	    for(m=l; m>=0; m--)  alm[ll+m+ln1] = alm[ll+m+lo1];
	}
    } else if (lmax_old > lmax_new) {
	for(l=0; l<= lmax_new; l++) {
	    int ll = l * nbuf;
	    for(m=0; m<=l; m++)  alm[ll+m+ln1] = alm[ll+m+lo1];
	}
    }
    return;
}




struct _clm_ { int l, m; };
extern void sort_clm (struct _clm_ *, int);
extern void read_lm(struct _clm_ **, int *, char *);

void alm_remove(alm, lmax, nbuf, filename)
double *alm;
int lmax, nbuf;
char *filename;
{
    register int i;
    struct _clm_ *clm = (struct _clm_ *)0;
    int lm;

    read_lm(&clm, &lm, filename);

    for(i=0; i<lm; i++) {
	int l = clm[i].l, m = clm[i].m;
	int ll = nbuf * l;
	alm[ll+m] = alm[ll+m+lmax+1] = 0.00;
    }

    free((void *)clm);
    return;
}

void alm_extract(alm, lmax, nbuf, filename)
double *alm;
int lmax, nbuf;
char *filename;
{
    register int i, l, m, i_b=0;
    int lm=0;
    struct _clm_ *clm = (struct _clm_ *)0;

    read_lm(&clm, &lm, filename);
    sort_clm (clm, lm);

    for(l=0; l<=lmax; l++) {
	int flag = 1;
	int ll = nbuf * l;

	/* select corresponding l */
	for(i=i_b; i<lm; i++) {
	      if(clm[i].l == l)   { i_b  = i; break; }
	      if(clm[i].l  > l)   { flag = 0; break; }
	}
	if (i==lm)  flag = 0;

	/* select corresponding m, else a_lm = 0 */
	if ( flag ) {
	      for(m=0; m<=l; m++) {
		   int fflag = 1;
		   i = i_b;
		   while (clm[i].l == l)   {
			if (clm[i].m == m)  { fflag = 0; break; }
			i++;
			if (i>=lm)  break;
		   }
		   if ( fflag )  alm[ll+m] = alm[ll+m+lmax+1] = 0.00;
	      }
	} else {
	      /* all a_lm =0 at current l */
	      for(m=0; m<=l; m++)  alm[ll+m] = alm[ll+m+lmax+1] = 0.00;
	}
    }

    free((void *)clm);
    return;
}

void read_lm(clm, lm, filename)
struct _clm_ **clm;
int  *lm;
char *filename;
{
    register int i=0;
    int lmx=0;
    FILE *fp;
    char str[161];

    /* open file */
    if ((fp = fopen(filename, "r")) == (FILE *)0) {
	    PR(ER, "%s: can not open file '%s'\n", myname, filename);
	    exit(1);
    }

    /* strings count */
    while (fgets(str, 160, fp) != NULL) {
	register int l;
	char s_l[20];

	if (*str == '\n' || *str == '\r' || *str == '#'|| *str == '%')
		continue;

	sscanf(str, "%s", s_l);
	l = getLmx(s_l);
	if (l > lmx)   lmx = l;
	i++;
    }
    fseek(fp, 0l, 0);

    if ( *clm == (struct _clm_ *)0) {
       /* find memory for arrays */
       if ((*clm = (struct _clm_ *)calloc((unsigned)i, sizeof **clm)) ==
	  (struct _clm_ *)0) {
	       PR(ER,"%s: can not find memory (%d elements) for array...\n",
		     myname, i);
	       exit(1);
       }
    }
    i=0;
    while (fgets(str, 160, fp) != NULL) {
	if (*str == '\n'||*str == '\r'||*str == '#'||*str == '%')
		continue;
	sscanf(str, "%d%d", &(*clm)[i].l, &(*clm)[i].m);
	i++;
    }
    *lm = i;

    fclose(fp);

    return;

}

/* "Quicker-sort algorithm" from system library         */
int i_cmp(p1, p2)
void  *p1, *p2;
{
	int flag=0;
	struct _clm_ *ph1 = (struct _clm_ *)p1,
		     *ph2 = (struct _clm_ *)p2;
	flag = ph1->l < ph2->l ? -1 : (ph1->l > ph2->l ? 1 : 0);
	if (!flag)
	     flag = ph1->m < ph2->m ? -1 : (ph1->m > ph2->m ? 1 : 0);
	return flag;
}
void sort_clm (c, n)
struct _clm_ *c;
int n;
{
	qsort(c, n, sizeof *c, (int (*)())i_cmp);
	return;
}

void separ_phase(alm, lmax, nbuf, ph0, eps, separ_flag, lmax_print)
double *alm;
int lmax, nbuf;
double ph0, eps; /* phase +- epsilon        */
int separ_flag;  /* 1 - extract, 2 - remove */
int lmax_print;
{
    register int i, l, m;
    int lm=0;
    struct _clm_ *clm = (struct _clm_ *)0;
    double pi2 = 2. * M_PI;


    /* allocate memory */
    if (clm == (struct _clm_ *)0) {
       /* find memory for arrays */
       if ((clm = (struct _clm_ *)calloc((unsigned)(lmax*(lmax+1)/2),
	  sizeof *clm)) == (struct _clm_ *)0) {
	       PR(ER,"%s: can not find memory (%d elements) for array...\n",
		     myname, lmax*(lmax+1)/2);
	       exit(1);
       }
    }

    /* select phases */
    for(i=l=0; l<=lmax_print; l++) {
	register int ll = l * nbuf;
	for(m=1; m<=l; m++) {
	    double phase = ph_calc(alm[ll+m], alm[ll+m +lmax+1]);
	    if (ph0 - eps < 0. &&
	       ((0. < phase  && phase < ph0 + eps) ||
		(ph0-eps+pi2 < phase && phase < pi2)) ) {
		    clm[i].l = l;
		    clm[i].m = m;
		    i++;
	    }
	    if (ph0 + eps > pi2 &&
	       ((ph0-eps < phase  && phase < pi2) ||
		(0. < phase && phase < ph0+eps-pi2)) ) {
		    clm[i].l = l;
		    clm[i].m = m;
		    i++;
	    }
	    if (ph0 - eps < phase && phase < ph0 + eps) {
		    clm[i].l = l;
		    clm[i].m = m;
		    i++;
	    }
	}
    }

    lm = i;

    if (separ_flag == 1) {
	int i_b = 0;
	/**************************/
	/* sort selected a_{lm}-s */
	/**************************/
	sort_clm (clm, lm);

	/* extract a_{lm}-s */
	for(l=0; l<=lmax; l++) {
	    int flag = 1;
	    int ll = nbuf * l;

	    /* select corresponding l */
	    for(i=i_b; i<lm; i++) {
		  if(clm[i].l == l)   { i_b  = i; break; }
		  if(clm[i].l  > l)   { flag = 0; break; }
	    }
	    if (i==lm)  flag = 0;

	    /* select corresponding m, else a_lm = 0 */
	    if ( flag ) {
		  for(m=0; m<=l; m++) {
		       int fflag = 1;
		       i = i_b;
		       while (clm[i].l == l)   {
			    if (clm[i].m == m)  { fflag = 0; break; }
			    i++;
			    if (i>=lm)  break;
		       }
		       if ( fflag )  alm[ll+m] = alm[ll+m+lmax+1] = 0.00;
		  }
	    } else {
		  /* all a_lm =0 at current l */
		  for(m=0; m<=l; m++)  alm[ll+m] = alm[ll+m+lmax+1] = 0.00;
	    }
	}
    } else if (separ_flag == 2) {
	/*******************/
	/* remove a_{lm}-s */
	/*******************/
	for(i=0; i<lm; i++) {
	    int ll;
	    l = clm[i].l, m = clm[i].m;
	    ll = nbuf * l;
	    alm[ll+m] = alm[ll+m+lmax+1] = 0.00;
	}
    }

    free((void *)clm);

    return;
}


/* re-allocate alm in old array with new lmax */
void realloc_alm_T(alm, nbuf, lmax_old, lmax_new)
double *alm;
int nbuf;
int lmax_old, lmax_new;
{
    register int l, m, lo1=lmax_old+1, ln1=lmax_new+1;

    for(l=0; l<=lmax_new; l++) {
	for(m=0; m<=l; m++)
	    alm[(m+ln1)*nbuf+l] = alm[(m+lo1)*nbuf+l];
    }

    return;
}

int read_alm_table_T(filename, alm, nalm, nbuf, lmax)
char *filename;
double **alm;
int *nalm;
int *nbuf;
int *lmax;
{
    register int counter = 0;
    int alm_ffile_flag = 0;
    int lmx = -1;
    FILE *fp = (FILE *)0;
    char str[161];

    /* open file */
    if ((fp = fopen(filename, "r")) == (FILE *)0) {
	    PR(ER, "%s: can not open file '%s'\n", myname, filename);
	    exit(1);
    }

    /* strings count */
    while (fgets(str, 160, fp) != NULL) {
	register int l;
	char s_l[20];

	/* check if this file is F-file with a_{lm}s */
	if (!counter && strncmp(str, "SIMPLE=", 7) == 0) {
		alm_ffile_flag = 1;
		fclose(fp);
		fp = (FILE *)0;
		break;
	}   else   {
		/* check if this file is FITS-file with a_{lm}s */
		if (!counter && strncmp(str, "SIMPLE  = ", 10) == 0) {
			alm_ffile_flag = 2;
			fclose(fp);
			fp = (FILE *)0;
			break;
		}
	}

	if (*str == '\n' || *str == '\r' || *str == '#'|| *str == '%')
		continue;

	sscanf(str, "%s", s_l);
	l = getLmx(s_l);
	if (l > lmx)   lmx = l;
	counter++;
    }

    if ( alm_ffile_flag ) {
	if (*alm == (double *)0) {
	    if (*lmax == 0 || *lmax == -1) {
		    if (alm_ffile_flag == 1)
			    *lmax = get_lmax(filename);
		    else
			    *lmax = Fget_lmax(filename);
		    *nbuf = 2 * (*lmax + 1);
		    *nalm = *lmax + 1;
	    }
	    if((*alm=(double *)calloc((unsigned)(*nbuf * *nalm),
		 sizeof **alm)) == (double *)0) {
nomem:
		    PR(ER,"%s: no memory for arrays[%d]\n",
			myname, *nbuf * *nalm);
		    exit(1);
	    }
	}
	if (alm_ffile_flag == 1)
		ff_read_alm(*alm, *nbuf <0? -(*nbuf): *nbuf, lmax,filename);
	else
	    in_glesp_alm_T(*alm, lmax, *nbuf <0? -(*nbuf): *nbuf,filename);
    } else {
	fseek(fp, 0l, 0);

//        if (lmx > *lmax)  *lmax = lmx;
//        if ((*nalm < 0 ? -(*nalm) : *nalm) < (*lmax + 1)*2) {
//            if (*alm != (double *)0) {
//                free((void *)*alm); /* free memory if buffer is small */
//                *alm = (double *)0;
//            }
//        }

	if ( *alm == (double *)0) {
	   *nalm = 2*(*lmax + 1);
	   *nbuf = 2*(*nalm + 1);

	   /* find memory for arrays */
	   if ((*alm = (double *)calloc((unsigned)(*nbuf * *nalm),
	       sizeof **alm)) == (double *)0)
		  goto nomem;
	}

	while (fgets(str, 160, fp) != NULL) {
	    if (*str == '\n'||*str == '\r'||*str == '#'||*str == '%')
		    continue;

	    str2alm_T(*alm, *nbuf < 0 ? -(*nbuf) : *nbuf, *lmax, str);
	}
    }
    if(fp != (FILE *)0) fclose(fp);


    return *lmax;
}

void read_alm_str_T (alm_str, alm, nbuf, lmax)
char *alm_str;
double *alm;
int nbuf, lmax;
{
	char buf_str[50], form[30];

	/* read l, m, alm_re, alm_im  strings */
	/* and write them in another format   */
	skip_wd(0, alm_str, form);
	strcpy(buf_str, form); strcat(buf_str, " ");
	skip_wd(1, alm_str, form);
	strcat(buf_str, form); strcat(buf_str, " ");
	skip_wd(2, alm_str, form);
	strcat(buf_str, form); strcat(buf_str, " ");
	skip_wd(3, alm_str, form);
	strcat(buf_str, form);

	/* fill in a_{lm} array */
	str2alm_T(alm, ABS(nbuf), lmax, buf_str);
	return;
}

int read_alm_3table_T(filename, alm, nalm, nbuf, lmax)
char *filename;
double **alm;
int *nalm;     /* =Nmx */
int *nbuf;
int *lmax;
{
    double *alm_e = (double *)0, *alm_b = (double *)0;

    if (*nalm < 0) {
	if (*nbuf == 0)
	    *nbuf = 2*(*lmax + 1);
    }
    if (*nalm < 0 && *nbuf < 0 && *alm == (double*)0) {
	unsigned n = *nbuf * ( *nalm < 0 ? -(*nalm) : *nalm );
	if((*alm=(double*)calloc(n,sizeof **alm)) == (double *)0) {
	    PR(ER,"%s: no memory for arrays of %d elements...\n", myname, n);
	    exit(1);
	}
    }

    read_ieb_alm_T(alm, &alm_e, &alm_b, lmax, nbuf, filename);

    if (*nalm >= 0)
	*nalm = 2*(*lmax + 1);

    if (alm_e != (double *)0)  free((void *)alm_e);
    if (alm_b != (double *)0)  free((void *)alm_b);

    PR(ER, "%s:\n    input file contains polarization data too, \n", myname);
    PR(ER, "    but only anisotropy will be processed.\n\n");

    return *lmax;
}

