# include <stdio.h>
# include <stdlib.h>
# include <unistd.h>
# include <strings.h>
# include <math.h>

# include "glesp.h"

#define PR      fprintf
#define ER      stderr

/*
 *  read source information from http://cats.sao.ru
 *      Oleg Verkhodanov,
 *      20/03/2006
 */
extern char *myname;
extern char tmpcat[161];

/* read equatorial coordinates from CATS-file and convert them */
void  read_cats(cr, n, filename, flag_coor)
struct _coor_  **cr;
int *n;
char *filename;
int flag_coor;   /* 0 - equatorial, 1 - galactic */
{
    register int i=0;
    FILE *fp;
    char str[161];

    if((fp  = fopen(filename, "r")) == (FILE *)0) {
	PR(ER, "%s: can not open file '%s'...\n",  myname, filename);
	exit(1);
    }
    while(fgets(str, 160, fp) != NULL) {
	if (*str == '#' || *str == '\n' || *str == '\r')
	    continue;
	i++;
    }
    fseek(fp, 0l, 0);

    *n = i;
    /* find memory for arrays */
    if ((*cr = (struct _coor_ *)calloc((unsigned)i, sizeof **cr)) ==
	(struct _coor_ *)0) {
	    PR(ER,"%s: can not find memory for array...\n", myname);
	    exit(1);
    }

    /* strings reading */
    i=0;
    while (fgets(str, 160, fp) != NULL) {
	char *cs, s1[30], cat[10], name[30],
	     sh_ra[7],  sm_ra[7],  ss_ra[10],
	     sd_dec[7], sm_dec[7], ss_dec[10];
	double p = M_PI/180.;
	int hr, mr, dd, md, sg;
	double sr, sd, ra, dec;
	double lon, lat;
	int n_par;

	if (*str == '\n' || *str == '\r' || *str == '#'|| *str == '%')
		continue;

	cs = str;
	while (*cs != '\0' && *cs != '#') cs++;
	if (*cs == '#') *cs = '\0';

	/* number of separated parameters in a string */
	n_par = n_skip_wd(str);

	if (n_par == 3 || n_par == 4) {
	    /* read CATS string */
	    if (n_par == 3)
		sscanf(str, "%s%s", ss_ra, ss_dec);
	    else
		sscanf(str, "%s%s%s", name, ss_ra, ss_dec);
	    ra  = str2deg(ss_ra);
	    dec = str2deg(ss_dec);
	} else {
	    /* read CATS string */
	    if (n_par == 7)
		sscanf(str, "%s%s%s%s%s%s",
		    sh_ra, sm_ra, ss_ra, sd_dec, sm_dec, ss_dec);
	    else if (n_par == 8)
		sscanf(str, "%s%s%s%s%s%s%s", name,
		    sh_ra, sm_ra, ss_ra, sd_dec, sm_dec, ss_dec);
	    else
		sscanf(str, "%s%s%s%s%s%s%s%s%s", cat, name,
		    sh_ra, sm_ra, ss_ra, s1, sd_dec, sm_dec, ss_dec);

	    /* right ascension or galactic latitude */
	    hr = atoi(sh_ra);
	    if (*sh_ra=='-')  { hr = -hr; sg = -1; } else sg = 1;
	    mr = atoi(sm_ra);
	    sr = atof(ss_ra);
	    ra = hr + mr/60. + sr/3600.;
	    ra *= sg;

	    /* declination or galactic longitude */
	    dd = atoi(sd_dec);
	    if (*sd_dec=='-')  { dd = -dd; sg = -1; } else sg = 1;
	    md = atoi(sm_dec);
	    sd = atof(ss_dec);
	    dec = dd + md/60. + sd/3600.;
	    dec *= sg;
	}
	switch(flag_coor) {
	case 0:
	default :
	    /* convert (alp(hrs),del(degs)) -> (lat(degs), lon(degs)) */

	    equat2gal(ra, dec, &lat, &lon);
	    (*cr)[i].lon = lon * p; /* deg -> rad */
	    (*cr)[i].lat = lat * p;
	    break;
	case 1:
	    (*cr)[i].lat = ra * p;  /* deg -> rad */
	    (*cr)[i].lon = dec * p;
	    break;
	}
	i++;
    }
    fclose(fp);


}

void  cats_str2sph(str, theta, phi, t, gal_flag)
char   *str;
double *theta, *phi, *t;
int gal_flag;
{
	double lat, lon;
	char s1[30], cat[10], name[30], st[10],
	     sh_ra[7],  sm_ra[7],  ss_ra[10],
	     sd_dec[7], sm_dec[7], ss_dec[10];
	double p = M_PI/180.;
	int hr, mr, dd, md, sg;
	double sr, sd, ra, dec;

	/* read CATS string */

	sscanf(str, "%s%s %s%s%s %s %s%s%s %s %s%s", cat, name,
		sh_ra, sm_ra, ss_ra, s1, sd_dec, sm_dec, ss_dec, s1, s1, st);


	/* right ascension or galactic latitude */
	hr = atoi(sh_ra);
	if (*sh_ra=='-')  { hr = -hr; sg = -1; } else sg = 1;
	mr = atoi(sm_ra);
	sr = atof(ss_ra);
	ra = hr + mr/60. + sr/3600.;
	ra *= sg;

	/* declination or galactic longitude */
	dd = atoi(sd_dec);
	if (*sd_dec=='-')  { dd = -dd; sg = -1; } else sg = 1;
	md = atoi(sm_dec);
	sd = atof(ss_dec);
	dec = dd + md/60. + sd/3600.;
	dec *= sg;

	*t = atof(st);

	if ( gal_flag )  {
		lat = ra;
		lon = dec;
	} else {
		equat2gal(ra, dec, &lat, &lon);
	}
	lat *= p;
	lon *= p;

	gal2pol(lat, lon, theta, phi);

	return;
}
