#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "ff.h"
#include "glesp.h"

#define  PR     fprintf
#define  ER     stderr

/* detect file type, read map and check size of arrays     */
/* if arrays are short, then free old and catch new memory */

extern char tmpcat[161];
extern char *myname;
extern int flag_take_memory;  /* flag to take memory */

HEAD hd, *ph;

int read_map1(f, nx, nphi, nph, x, inp_file)
double **f;
int *nx, *nphi;
double **x;
int **nph;
char *inp_file;
{
	register int i,j;
	register FILE *fp_in;
	int l_nx;
	int nptot=0;

	ph = &hd;

	ph->tform1[0] = ph->tform2[0] = ph->tform3[0] =
	ph->tdim1[0] = ph->tdim2[0] = ph->tdim3[0] = '\0';

	if ((fp_in = fopen(inp_file, "r")) == (FILE *)0) {
	    PR(ER, "%s: can not open file '%s'\n", myname, inp_file);
	    exit(1);
	}

	/****************************************/
	/* read basic keywords from the 1st head */
	/****************************************/
	Hinit(fp_in);

	if (EVget("EXTEND") != NULL)
	    strcpy(ph->extend, EVget("EXTEND"));

	if (strcmp(ph->extend, "T") != 0) {
	    PR(ER, "%s: bad extension in file '%s'\n", myname, inp_file);
	    exit(1);
	}

	Hfree();

	/****************************************/
	/* read basic keywords from the 2nd head */
	/****************************************/
	Hinit(fp_in);

	if (EVget("XTENSION") != NULL)
	    strcpy(ph->xtens, EVget("XTENSION"));

	if (strcmp(ph->xtens, "BINTABLE") != 0) {
	    PR(ER, "%s: can work only with BINTABLE extension\n", myname);
	    exit(1);
	}
	if (EVget("NAXIS") != NULL)
		ph->naxis  = Skey_get("NAXIS");
	if (EVget("NAXIS1") != NULL)
		ph->naxis1  = Lkey_get("NAXIS1");
	if (EVget("NAXIS2") != NULL)
		ph->naxis2  = Lkey_get("NAXIS2");

	if (EVget("TFIELDS") != NULL)
		ph->tfields = Lkey_get("TFIELDS");

	if( ph->tfields != 3) {
		PR(ER, "%s: bad number of fields (%d) for function ",
			myname, ph->tfields);
		PR(ER, "'ff_read_ext2_map'. It must be 3.\n");
		exit(1);
	}

	if (EVget("TFORM1") != NULL)
		strcpy(ph->tform1, EVget("TFORM1"));
	if (EVget("TDIM1") != NULL)
		strcpy(ph->tdim1, EVget("TDIM1"));
	if (EVget("TTYPE1") != NULL)
		strcpy(ph->ttype1, EVget("TTYPE1"));
	if (EVget("TUNIT1") != NULL)
		strcpy(ph->tunit1, EVget("TUNIT1"));

	if (EVget("TFORM2") != NULL)
		strcpy(ph->tform2, EVget("TFORM2"));
	if (EVget("TDIM2") != NULL)
		strcpy(ph->tdim2, EVget("TDIM2"));
	if (EVget("TTYPE2") != NULL)
		strcpy(ph->ttype2, EVget("TTYPE2"));
	if (EVget("TUNIT2") != NULL)
		strcpy(ph->tunit2, EVget("TUNIT2"));

	if (EVget("TFORM3") != NULL)
		strcpy(ph->tform3, EVget("TFORM3"));
	if (EVget("TDIM3") != NULL)
		strcpy(ph->tdim3, EVget("TDIM3"));
	if (EVget("TTYPE3") != NULL)
		strcpy(ph->ttype3, EVget("TTYPE3"));
	if (EVget("TUNIT3") != NULL)
		strcpy(ph->tunit3, EVget("TUNIT3"));

	if (EVget("OBJECT") != NULL)
		strcpy(ph->object, EVget("OBJECT"));
	if (EVget("ORIGIN") != NULL)
		strcpy(ph->origin, EVget("ORIGIN"));

	/* X-axis grid type */
	if (EVget("GRID") != NULL) {
		char str[51];
		strcpy(str, EVget("GRID"));
		switch(*str) {
		case 'P':
		case 'p':
			if (strcmp("PHI(X)+X", str) == 0)
				ph->grid = GRID_PHI_X;
			else
				ph->grid = GRID_PHI_X_LEGENDRE;
			break;
		case 'L':
		case 'l':
			ph->grid = GRID_LEGENDRE;
			break;
		case 'U':
		case 'u':
			ph->grid = GRID_UNIFORM_COS;
			break;
		case 'T':
		case 't':
			ph->grid = GRID_UNIFORM_THETA;
			break;
		default :
			ph->grid = GRID_NON_SET;
			break;
		}
	}

	if (*(ph->tdim1)=='\0' || *(ph->tdim2)=='\0' || *(ph->tdim3)=='\0') {
	     PR(ER, "%s: bad value of dimensions. Check TDIM# keywords.\n",
		myname);
	     exit(1);
	}

	/* extract dimension of the 1st array */
	i = atoi(ph->tdim1+1);

	/* extract dimensions of the 2nd array */
	l_nx = atoi(ph->tdim2+1);
	if (l_nx != i) {
		PR(ER, "%s: bad dimension by X-axis...\n", myname);
		exit(1);
	}
	*nx = l_nx;
	if (flag_take_memory) {
	    if ((*x = (double *)calloc((unsigned)(*nx), sizeof **x)) ==
		 (double *)0) {
nomem:
		      PR(ER,"%s: no memory for arrays...\n", myname);
		      exit(1);
	    }
	}
	if (flag_take_memory) {
	    if ((*nph = (int *)calloc((unsigned)(*nx), sizeof **nph)) ==
		 (int *)0)
		    goto nomem;
	}
	/******************/
	/* read 1st field */
	/******************/
	/* read 1-dim array of  X-argument vector */
	bitpix = ph->bitpix1 = atoi(ph->tform1) / *nx * 8;
	for(i=0; i < *nx; i++) {
/*
		(*x)[i] = get_pix_d(fp_in);
*/
		(*x)[i] = get_pix(fp_in);
	}

	/**********************************/
	/* read 2nd field */
	/**********************************/
	/* read 1-dim array of Nph-vector */
	bitpix = ph->bitpix2 = atoi(ph->tform2) / *nx * 8;
	*nphi = 0;
	for(j=0; j < *nx; j++) {
		(*nph)[j] = get_pix_d(fp_in) + 0.5;
		nptot += (*nph)[j];
		if (*nphi < (*nph)[j])  *nphi = (*nph)[j];
	}

	ph->nx = *nx;

	/* memory under 3d field */
	if (flag_take_memory) {
	    if ((*f = (double *)calloc((unsigned)(*nx * *nphi), sizeof **f))==
		 (double *)0)
		    goto nomem;
	}
	/*****************/
	/* read 3d field */
	/*****************/
	/* read 2-dim array */
	bitpix = ph->bitpix3 = atoll(ph->tform3) / nptot * 8;
	for(i=0; i < *nx; i++) {
		register int k = i * *nphi;
		for(j=0; j< (*nph)[i]; j++) {
			double v;
			if (feof(fp_in)) {
			    PR(ER, "%s: unexpected end of file '%s'\n",
				myname, inp_file);
			    exit(1);
			}
			v = get_pix_d(fp_in);
			(*f)[k+j] = v;
		}
	}
	if (fp_in != stdin && fp_in != (FILE *)0)  fclose(fp_in);
	Hfree();

	return nptot;
}

int read_map2(f, nx, nphi, nph, x, inp_file)
double **f;
int *nx, *nphi;
double **x;
int **nph;
char *inp_file;
{
	int tr_flag = 0;            /* transposing flag */
	int file_type_flag;
	double *phi = (double *)0;

	/* free memory always */
	if (*f  != (double *)0)  free((void *)*f);
	if (*x  != (double *)0)  free((void *)*x);
	if (*nph!= (int *)   0)  free((void *)*nph);
	*f   = (double *)0;
	*x   = (double *)0;
	*nph = (int *)0;

	/* pointer to head structure */
	ph = &hd;
	ph->grid = GRID_NON_SET;

	/* detection of file type */
	/* flag: F_BIN_TAB_PHI  - Binary Table, x = cos theta, phi variable */
	/* flag: F_BIN_TAB  - Binary Table, x = cos theta */
	/* flag: F_BFc      - basic F-file, x = cos theta */
	/* flag: F_BFc_TR   - basic F-file, x = cos theta */
	/* flag: F_BF       - basic F-file, x = theta     */
	/* flag: F_BF_TR    - basic F-file, x = theta     */
	/* flag: F_other    - other type  */
	/* flag: F_notF     - not F-file  */
	file_type_flag = ffile_type(inp_file);

	if ( file_type_flag == F_BF_TR && file_type_flag == F_BFc_TR)
		tr_flag = 1;
/*
PR(ER,"file_type_flag=%d, tr_flag=%d\n", file_type_flag, tr_flag);
*/
	/* in memory we keep transposed image:  */
	/*      f(nphi,nx)                      */
	/*      by X (in strings): cos theta,   */
	/*      by Y (in columns): phi.         */
	/* It gives a possibility to change     */
	/* step by X-axis with interpolation    */

	switch(file_type_flag) {
	case  F_BIN_TAB_PHI: /* Binary Table, x = cos theta, phi variable */
		ff_read_ext2_map(f, x, nx, nph, nphi, inp_file);
		break;
	case  F_BIN_TAB:
		/* read F-BinTab format, create f-, x-, phi-arrays */
		/* and catch memory                                */
		ff_read_ext_map(f, x, nx, &phi, nphi, inp_file);
		break;
	case F_BFc:
	case F_BFc_TR:  /* cos(theta) by X */
		/* read basic-format, create f-, x-, phi-arrays */
		/* and catch memory                                */
		ff_read_stand_map(f,x,nx, &phi,nphi, tr_flag, inp_file);
		break;
	case F_BF:
	case F_BF_TR:   /* theta by X */
		/* read basic-format, create f-, x-, phi-arrays */
		/* and catch memory                                */
		ff_read_stand_map(f,x,nx, &phi,nphi, tr_flag, inp_file);
		theta_grid = 2;
		break;
	case F_other : /* other type F-file */
		PR(ER, "%s: it is F-file of other type.\n", myname);
		exit(1);
	case F_notF : /* it is not F-file at all */
		PR(ER, "%s: it is not F-file.\n", myname);
		exit(1);
	}
	if (*nph == (int *)0) {
	    register int i;
	    if((*nph =(int *)calloc((unsigned)*nx, sizeof **nph))==(int *)0) {
		PR(ER, "%s: read_map can not find memory...\n", myname);
		exit(1);
	    }
	    for(i=0; i < *nx; i++) (*nph)[i] = *nphi;
	}

	if (phi != (double *)0) free((void *)phi);

	return ph->grid;
}

int read_map0(f, nx, nphi, x, phi, dphi, inp_file)
double **f;
int *nx, *nphi;
double **x, **phi;
double *dphi;
char *inp_file;
{
	int tr_flag = 0;            /* transposing flag */
	int file_type_flag;

	/* free memory always */
	if (*f  != (double *)0)  free((void *)*f);
	if (*x  != (double *)0)  free((void *)*x);
	if (*phi!= (double *)0)  free((void *)*phi);
	*f   = (double *)0;
	*x   = (double *)0;
	*phi = (double *)0;

	/* pointer to head structure */
	ph = &hd;
	ph->grid = GRID_NON_SET;

	/* detection of file type */
	/* flag: F_BIN_TAB  - Binary Table, x = cos theta */
	/* flag: F_BFc      - basic F-file, x = cos theta */
	/* flag: F_BFc_TR   - basic F-file, x = cos theta */
	/* flag: F_BF       - basic F-file, x = theta     */
	/* flag: F_BF_TR    - basic F-file, x = theta     */
	/* flag: F_other    - other type  */
	/* flag: F_notF     - not F-file  */
	file_type_flag = ffile_type(inp_file);

	if ( file_type_flag == F_BF_TR && file_type_flag == F_BFc_TR)
		tr_flag = 1;
/*
PR(ER,"file_type_flag=%d, tr_flag=%d\n", file_type_flag, tr_flag);
*/
	/* in memory we keep transposed image:  */
	/*      f(nphi,nx)                      */
	/*      by X (in strings):  cos theta,  */
	/*      by Y (in columns): phi.         */
	/* It gives a possibility to change     */
	/* step by X-axis with interpolation    */

	switch(file_type_flag) {
	case  F_BIN_TAB:
		/* read F-BinTab format, create f-, x-, phi-arrays */
		/* and catch memory                                */
		ff_read_ext_map(f, x, nx, phi, nphi, inp_file);
		break;
	case F_BFc:
	case F_BFc_TR:  /* cos(theta) by X */
		/* read basic-format, create f-, x-, phi-arrays */
		/* and catch memory                                */
		ff_read_stand_map(f,x,nx, phi,nphi, tr_flag, inp_file);
		break;
	case F_BF:
	case F_BF_TR:   /* theta by X */
		/* read basic-format, create f-, x-, phi-arrays */
		/* and catch memory                                */
		ff_read_stand_map(f,x,nx, phi,nphi, tr_flag, inp_file);
		theta_grid = 2;
		break;
	case F_other : /* other type F-file */
		PR(ER, "%s: it is F-file of other type.\n", myname);
		exit(1);
	case F_notF : /* it is not F-file at all */
		PR(ER, "%s: it is not F-file.\n", myname);
		exit(1);
	}

	return ph->grid;
}

void read_glesp_map(f, x, nx, np, nph, inp_file)
double **f;
int *nx, *np;
double **x;
int **nph;
char *inp_file;
{
	int map_type_flag = test_glesp_map(inp_file);

	if (map_type_flag < 0) {
		PR(ER, "%s: it is not a GLESP FITS file...\n", myname);
		exit(1);
	}
	if (map_type_flag == 2) {
		PR(ER, "%s: it is probably a file with a_lm-s...\n", myname);
		exit(1);
	}


	if (map_type_flag) { /* FITS     */
		in_glesp_map(f, x, nx, np, nph, inp_file);
	} else  {              /* F-format */
		read_map1(f, nx, np, nph, x, inp_file);
	}
	return;
}


void read_glesp_tmap(t, x, nx, np, nph, inp_file)
double **t;
int *nx, *np;
double **x;
int **nph;
char *inp_file;
{
	double *u = (double *)0, *q = (double *)0;
	int map_type_flag = test_glesp_map(inp_file);

	if (map_type_flag < 0) {
		PR(ER, "%s: it is not a GLESP FITS file...\n", myname);
		exit(1);
	}
	if (map_type_flag == 2) {
		PR(ER, "%s: it is probably a file with a_lm-s...\n", myname);
		exit(1);
	}

	if (map_type_flag == 3) { /* FITS-file 3 maps (T,Q,U) */
	    in_glesp_3map(t, &q, &u, x, nx, np, nph, inp_file);
	    free((void *)q);
	    free((void *)u);
	} else {
	    if (map_type_flag) { /* FITS     */
		    in_glesp_map(t, x, nx, np, nph, inp_file);
	    } else  {              /* F-format */
		    read_map1(t, nx, np, nph, x, inp_file);
	    }
	}
	return;
}


void read_glesp_qmap(q, x, nx, np, nph, inp_file)
double **q;
int *nx, *np;
double **x;
int **nph;
char *inp_file;
{
	double *t = (double *)0, *u = (double *)0;
	int map_type_flag = test_glesp_map(inp_file);

	if (map_type_flag < 0) {
		PR(ER, "%s: it is not a GLESP FITS file...\n", myname);
		exit(1);
	}
	if (map_type_flag == 2) {
		PR(ER, "%s: it is probably a file with a_lm-s...\n", myname);
		exit(1);
	}

	if (map_type_flag == 3) { /* FITS-file 3 maps (T,Q,U) */
	    in_glesp_3map(&t, q, &u, x, nx, np, nph, inp_file);
	    free((void *)t);
	    free((void *)u);
	} else {
	    if (map_type_flag) { /* FITS     */
		    in_glesp_map(q, x, nx, np, nph, inp_file);
	    } else  {              /* F-format */
		    read_map1(q, nx, np, nph, x, inp_file);
	    }
	}
	return;
}

void read_glesp_umap(u, x, nx, np, nph, inp_file)
double **u;
int *nx, *np;
double **x;
int **nph;
char *inp_file;
{
	double *t = (double *)0, *q = (double *)0;
	int map_type_flag = test_glesp_map(inp_file);

	if (map_type_flag < 0) {
		PR(ER, "%s: it is not a GLESP FITS file...\n", myname);
		exit(1);
	}
	if (map_type_flag == 2) {
		PR(ER, "%s: it is probably a file with a_lm-s...\n", myname);
		exit(1);
	}

	if (map_type_flag == 3) { /* FITS-file 3 maps (T,Q,U) */
	    in_glesp_3map(&t, &q, u, x, nx, np, nph, inp_file);
	    free((void *)t);
	    free((void *)q);
	} else {
	    if (map_type_flag) { /* FITS     */
		    in_glesp_map(u, x, nx, np, nph, inp_file);
	    } else  {              /* F-format */
		    read_map1(u, nx, np, nph, x, inp_file);
	    }
	}
	return;
}


void read_glesp_3map(f, q, u, x, nx, np, nph, inp_file)
double **f, **q, **u;
int *nx, *np;
double **x;
int **nph;
char *inp_file;
{
	int map_type_flag = test_glesp_map(inp_file);

	if (map_type_flag < 0) {
		PR(ER, "%s: it is not a GLESP FITS file...\n", myname);
		exit(1);
	}
	if (map_type_flag == 2) {
		PR(ER, "%s: it is probably a file with a_lm-s...\n", myname);
		exit(1);
	}

	if (map_type_flag == 3) { /* FITS-file 3 maps (T,Q,U) */
	    in_glesp_3map(f, q, u, x, nx, np, nph, inp_file);
	} else {
	    if (map_type_flag) { /* FITS     */
		    in_glesp_map(f, x, nx, np, nph, inp_file);
	    } else  {              /* F-format */
		    read_map1(f, nx, np, nph, x, inp_file);
	    }
	    *u = (double *)0;
	    *q = (double *)0;
	}
	return;
}

