#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "glesp.h"
#include "fits.h"
#include "ff.h"

#define  PR     fprintf
#define  ER     stderr

/* detect file type, read map and check size of arrays     */
/* if arrays are short, then free old and catch new memory */
extern void short_file_name(char *);

char *myname;
extern int sun_data_key ;
static int map_flag;
HEAD hd, *ph;
int flag_take_memory = 1;  /* flag to take memory */

void read_map_(f, nx, nphi, nph, x, inp_file)
double *f;
int *nx, *nphi;
double *x;
int *nph;
char *inp_file;
{
	flag_take_memory = 0;  /* flag to take memory */

	myname = "read_map";
	ph = &hd;

	ph->tform1[0] = ph->tform2[0] = ph->tform3[0] =
	ph->tdim1[0] = ph->tdim2[0] = ph->tdim3[0] = '\0';

	short_file_name(inp_file);

	read_glesp_map(&f, &x, nx, nphi, &nph, inp_file);

	return;
}

extern void read_map_size_fits(int *, int *, char *);
extern void read_map_size_ff(int *, int *, char *);
void read_map_size_(nx, nphi_m, filename)
int *nx;
int *nphi_m;
char *filename;
{

    map_flag = test_glesp_map(filename);
    if (map_flag == 1)
	    read_map_size_fits(nx, nphi_m, filename);
    else if (map_flag == 0)
	    read_map_size_ff(nx, nphi_m, filename);
    else {
	    PR(ER, "read_map_size: file '%s' is no a GLESP map file...\n",
		    filename);
    }
    return;
}

void read_map_size_fits(nx, nphi_m, filename)
int *nx;
int *nphi_m;
char *filename;
{
	register int i,j;
	register FILE *fp_in;
	int nphi=0, nph_max=0;
	int *nph = (int*)0;

	myname = "read_map_size";
	ph = &hd;

	sun_data_key = 0;

	ph->tform1[0] = ph->tform2[0] = ph->tform3[0] =
	ph->tdim1[0] = ph->tdim2[0] = ph->tdim3[0] = '\0';

	short_file_name(filename);

	if ((fp_in = fopen(filename, "r")) == (FILE *)0) {
	    PR(ER, "%s: can not open file '%s'\n", myname, filename);
	    exit(1);
	}

	/****************************************/
	/* read basic keywords from the 1st head */
	/****************************************/
	FHinit(fp_in);

	if (FEVget("EXTEND") != NULL)
	    strcpy(ph->extend, FEVget("EXTEND"));

	if (strcmp(ph->extend, "T") != 0) {
	    PR(ER, "%s: bad extension in file '%s'\n", myname, filename);
	    exit(1);
	}

	FHfree();

	/****************************************/
	/* read basic keywords from the 2nd head */
	/****************************************/
	FHinit(fp_in);

	if (FEVget("XTENSION") != NULL)
	    strcpy(ph->xtens, FEVget("XTENSION"));

	if (strcmp(ph->xtens, "BINTABLE") != 0) {
	    PR(ER, "%s: can work only with BINTABLE extension\n", myname);
	    exit(1);
	}
	if (FEVget("NAXIS") != NULL)
		ph->naxis  = FSkey_get("NAXIS");
	if (FEVget("NAXIS1") != NULL)
		ph->naxis1  = FLkey_get("NAXIS1");
	if (FEVget("NAXIS2") != NULL)
		ph->naxis2  = FLkey_get("NAXIS2");

	if (FEVget("TFIELDS") != NULL)
		ph->tfields = FLkey_get("TFIELDS");

	if( ph->tfields != 3) {
		PR(ER, "%s: bad number of fields (%d) for function ",
			myname, ph->tfields);
		PR(ER, "'ff_read_ext2_map'. It must be 3.\n");
		exit(1);
	}

	if (FEVget("TFORM1") != NULL)
		strcpy(ph->tform1, FEVget("TFORM1"));

	if (FEVget("TFORM2") != NULL)
		strcpy(ph->tform2, FEVget("TFORM2"));

	if (FEVget("TFORM3") != NULL)
		strcpy(ph->tform3, FEVget("TFORM3"));

	if(*(ph->tform1)=='\0'|| *(ph->tform2)=='\0'|| *(ph->tform3)=='\0') {
	     PR(ER, "%s: bad value of dimensions. Check TFORM# keywords.\n",
		myname);
	     exit(1);
	}

	/* extract dimension of the 1st array */
	i = atoi(ph->tform1);

	/* extract dimensions of the 2nd array */
	*nx = atoi(ph->tform2);
	if (*nx != i ) {
		PR(ER, "%s: bad dimension by X-axis...\n", myname);
		exit(1);
	}

	/*************/
	/* 1st field */
	/*************/
	/* read 1-dim array of  X-argument vector */
	/* bitpix = ph->bitpix1 = atoi(ph->tform1) / *nx * 8; */
	i = ph->tform1[strlen(ph->tform1)-1];
	if ( i == 'D' )
		bitpix = ph->bitpix1 = DOUBLEPIX;
	else /* if ( i == 'E' )*/
		bitpix = ph->bitpix1 = FLOATPIX;
	for(j=0; j < *nx; j++)
		get_pix_fits(fp_in);

	/**********************************/
	/* take memory and read 2nd field */
	/**********************************/
	if ((nph = (int *)calloc((unsigned)*nx, sizeof *nph)) == (int *)0){
		PR(ER, "%s: no memory for temporal file.\n", myname);
		exit(1);
	}
	/* read 1-dim array of Nph-vector */
	/* bitpix = ph->bitpix2 = atoi(ph->tform2) / *nx * 8; */
	i = ph->tform2[strlen(ph->tform2)-1];
	switch (i) {
	case 'I':
		ph->bitpix2 = SHORTPIX;
		break;
	case 'J':
		ph->bitpix2 = LONGPIX;
		break;
	default:
	case 'E':
		ph->bitpix2 = FLOATPIX;
		break;
	}
	bitpix = ph->bitpix2;
	for(j=0; j < *nx; j++) {
		double v = get_pix_fits(fp_in) + 0.5;
		nph[j] = v;
		if ( nph[j] > nph_max )  nph_max = nph[j];
		nphi += nph[j];
	}

	ph->nx = *nx;

	if (fp_in != stdin && fp_in != (FILE *)0)  fclose(fp_in);
	FHfree();

	*nphi_m = nph_max;

	if (nph != (int *)0)  free((void *)nph);
	return;
}

void read_map_size_ff(nx, nphi_m, filename)
int *nx;
int *nphi_m;
char *filename;
{
	register int i,j;
	register FILE *fp_in;
	int nphi=0, nph_max=0;
	int *nph = (int*)0;

	myname = "read_map_size";
	ph = &hd;


	sun_data_key = 0;

	ph->tform1[0] = ph->tform2[0] = ph->tform3[0] =
	ph->tdim1[0] = ph->tdim2[0] = ph->tdim3[0] = '\0';

	short_file_name(filename);

	if ((fp_in = fopen(filename, "r")) == (FILE *)0) {
	    PR(ER, "%s: can not open file '%s'\n", myname, filename);
	    exit(1);
	}

	/****************************************/
	/* read basic keywords from the 1st head */
	/****************************************/
	Hinit(fp_in);

	if (EVget("EXTEND") != NULL)
	    strcpy(ph->extend, EVget("EXTEND"));

	if (strcmp(ph->extend, "T") != 0) {
	    PR(ER, "%s: bad extension in file '%s'\n", myname, filename);
	    exit(1);
	}

	Hfree();

	/****************************************/
	/* read basic keywords from the 2nd head */
	/****************************************/
	Hinit(fp_in);

	if (EVget("XTENSION") != NULL)
	    strcpy(ph->xtens, EVget("XTENSION"));

	if (strcmp(ph->xtens, "BINTABLE") != 0) {
	    PR(ER, "%s: can work only with BINTABLE extension\n", myname);
	    exit(1);
	}
	if (EVget("NAXIS") != NULL)
		ph->naxis  = Skey_get("NAXIS");
	if (EVget("NAXIS1") != NULL)
		ph->naxis1  = Lkey_get("NAXIS1");
	if (EVget("NAXIS2") != NULL)
		ph->naxis2  = Lkey_get("NAXIS2");

	if (EVget("TFIELDS") != NULL)
		ph->tfields = Lkey_get("TFIELDS");

	if( ph->tfields != 3) {
		PR(ER, "%s: bad number of fields (%d) for function ",
			myname, ph->tfields);
		PR(ER, "'ff_read_ext2_map'. It must be 3.\n");
		exit(1);
	}

	if (EVget("TFORM1") != NULL)
		strcpy(ph->tform1, EVget("TFORM1"));
	if (EVget("TDIM1") != NULL)
		strcpy(ph->tdim1, EVget("TDIM1"));
	if (EVget("TTYPE1") != NULL)
		strcpy(ph->ttype1, EVget("TTYPE1"));
	if (EVget("TUNIT1") != NULL)
		strcpy(ph->tunit1, EVget("TUNIT1"));

	if (EVget("TFORM2") != NULL)
		strcpy(ph->tform2, EVget("TFORM2"));
	if (EVget("TDIM2") != NULL)
		strcpy(ph->tdim2, EVget("TDIM2"));
	if (EVget("TTYPE2") != NULL)
		strcpy(ph->ttype2, EVget("TTYPE2"));
	if (EVget("TUNIT2") != NULL)
		strcpy(ph->tunit2, EVget("TUNIT2"));

	if (EVget("TFORM3") != NULL)
		strcpy(ph->tform3, EVget("TFORM3"));
	if (EVget("TDIM3") != NULL)
		strcpy(ph->tdim3, EVget("TDIM3"));
	if (EVget("TTYPE3") != NULL)
		strcpy(ph->ttype3, EVget("TTYPE3"));
	if (EVget("TUNIT3") != NULL)
		strcpy(ph->tunit3, EVget("TUNIT3"));

	if (EVget("OBJECT") != NULL)
		strcpy(ph->object, EVget("OBJECT"));
	if (EVget("ORIGIN") != NULL)
		strcpy(ph->origin, EVget("ORIGIN"));

	/* X-axis grid type */
	if (EVget("GRID") != NULL) {
		char str[51];
		strcpy(str, EVget("GRID"));
		switch(*str) {
		case 'P':
		case 'p':
			if (strcmp("PHI(X)+X", str) == 0)
				ph->grid = GRID_PHI_X;
			else
				ph->grid = GRID_PHI_X_LEGENDRE;
			break;
		case 'L':
		case 'l':
			ph->grid = GRID_LEGENDRE;
			break;
		case 'U':
		case 'u':
			ph->grid = GRID_UNIFORM_COS;
			break;
		case 'T':
		case 't':
			ph->grid = GRID_UNIFORM_THETA;
			break;
		default :
			ph->grid = GRID_NON_SET;
			break;
		}
	}

	if (*(ph->tdim1)=='\0' || *(ph->tdim2)=='\0' || *(ph->tdim3)=='\0') {
	     PR(ER, "%s: bad value of dimensions. Check TDIM# keywords.\n",
		myname);
	     exit(1);
	}

	/* extract dimension of the 1st array */
	i = atoi(ph->tdim1+1);

	/* extract dimensions of the 2nd array */
	*nx = atoi(ph->tdim2+1);
	if (*nx != i ) {
		PR(ER, "%s: bad dimension by X-axis...\n", myname);
		exit(1);
	}
	/*************/
	/* 1st field */
	/*************/
	/* read 1-dim array of  X-argument vector */
	bitpix = ph->bitpix1 = atoi(ph->tform1) / *nx * 8;
	for(j=0; j < *nx; j++)
		get_pix(fp_in);

	/**********************************/
	/* take memory and read 2nd field */
	/**********************************/
	if ((nph = (int  *)calloc((unsigned)*nx, sizeof *nph)) == (int  *)0){
		PR(ER, "%s: no memory for temporal file.\n", myname);
		exit(1);
	}
	/* read 1-dim array of Nph-vector */
	bitpix = ph->bitpix2 = atoi(ph->tform2) / *nx * 8;
	for(j=0; j < *nx; j++) {
		double v = get_pix(fp_in) + 0.5;
		nph[j] = v;
		if ( nph[j] > nph_max )  nph_max = nph[j];
		nphi += nph[j];
	}

	ph->nx = *nx;

	if (fp_in != stdin && fp_in != (FILE *)0)  fclose(fp_in);
	Hfree();

	*nphi_m = nph_max;

	if (nph != (int *)0)  free((void *)nph);
	return;
}


int ffile_type_(ffile)
char *ffile;
{
	int flag = -1;

	myname = "ffile_type";

	short_file_name(ffile);

	map_flag = test_glesp_map(ffile);

	if (map_flag == 2) {
		PR(ER, "%s: file '%s' probably contains a_lm-s...\n",
		    myname, ffile);
	}

	if (map_flag == 0 || map_flag == 1)
		 flag = 1;
	else if (map_flag != -1)
		 flag = 0;

	return flag;
}

void short_file_name(fname)
char *fname;
{
	register int i = 0;
	register char *cs = fname;

	/* find '\0' at the end */
	while (i<1025 && *cs)  { i++; cs++; }
	if (i==1025) {
		PR(ER,"name for file is too long: '%s' (%d symbols)\n",
			fname, i);
		exit(1);
	}
	if (*cs == '\0') {
		cs--;
		while (*cs == ' ' || *cs == '\t') { *cs='\0'; cs--; }
	}
	return;
}
