# include <stdio.h>
# include <stdlib.h>
# include <math.h>

# include "glesp.h"
# include "gd/plot_gd.h"

# define PR     fprintf
# define ER     stderr


extern void read_marker(struct _crm_ **, int *, char *, int);
extern void marker_str(struct _crm_ *, char *, int);
extern void read_mark_str(struct _crm_ **, int *, char *, int);
extern void read_mark_pol(struct _crm_ **, int *, char *, int);
extern void marker_str_pol(struct _crm_ *, char *, int);

void read_marker(crm, n, filename, coord_flag)
struct _crm_  **crm;
int *n;
char *filename;
int coord_flag;
{
	register int i=0;
	char str[161];
	FILE *fp;

	/* open file */
	if ((fp = fopen(filename, "r")) == (FILE *)0) {
		PR(ER, "%s: can not open file '%s'\n", myname, filename);
		exit(1);
	}

	/* strings count */
	while (fgets(str, 160, fp) != NULL) {
		if (*str =='\n' || *str == '\r' || *str=='#'|| *str =='%')
			continue;
		i++;
	}
	fseek(fp, 0l, 0);
	*n = i;

	/* find memory for arrays */
	if ((*crm = (struct _crm_ *)calloc((unsigned)(*n), sizeof **crm)) ==
	    (struct _crm_ *)0) {
		PR(ER,"%s: can not find memory for array...\n", myname);
		exit(1);
	}

	/* strings reading */
	i=0;
	while (fgets(str, 160, fp) != NULL) {
	    if (*str == '\n' || *str == '\r' || *str == '#'|| *str == '%')
		    continue;

	    marker_str((*crm)+i, str, coord_flag);
	    i++;
	}

	fclose(fp);

	return;
}

void marker_str(crm, str, coord_flag)
struct _crm_  *crm;
char *str;
int coord_flag;
{
     double cr1, cr2;
     char st[16], sp[16], sm[16], ss[16], sc[16], spol[16];
     st[0] = sp[0] = sm[0] = ss[0] = sc[0] = spol[0] = '\0';

     skip_wd(0, str, st);
     skip_wd(1, str, sp);
     skip_wd(2, str, sm);
     skip_wd(3, str, ss);
     skip_wd(4, str, sc);


     switch(coord_flag) {
     case 1: /* galactic coordinates */
	    crm->theta = cr1 = deg2rad(st);
	    crm->phi   = cr2 = deg2rad(sp);
	    gal2pol(cr1, cr2, &crm->theta, &crm->phi);
	    break;
     case 2: /* equatorial coordinates */
	    {
	       double ra, dec, lat, lon, p= M_PI/180.;
	       ra  = str2deg(st);
	       dec = str2deg(sp);
	       equat2gal(ra, dec, &lat, &lon);

	       lat *= p; lon *= p;
	       crm->theta =  M_PI/2 - lat;
	       crm->phi   =  lon;
	    }
	    break;
     default: /* polar coordinates */
	    crm->theta = cr1 = deg2rad(st);
	    crm->phi   = cr2 = deg2rad(sp);
	    break;
     }

     if (*sm != 0)  {
	    if (*sm == 'p')  {
		crm->mark = -10;
		skip_wd(5, str, spol);
		crm->vpa = deg2rad(spol);
	    } else
		crm->mark = st2mark(sm);
     } else
	    crm->mark = -1;

     if (*ss != 0)
	    crm->size = atoi(ss);
     else
	    crm->size = -1;

     if (*sc != 0)
	    crm->color = atoi(sc);
     else
	    crm->color = -1;

     return;
}

void read_mark_str(crm, n, str, coord_flag)
struct _crm_  **crm;
int *n;
char *str;
int coord_flag;
{
	*n = 1;
	/* find memory for arrays */
	if ((*crm = (struct _crm_ *)calloc((unsigned)(*n), sizeof **crm)) ==
	    (struct _crm_ *)0) {
		PR(ER,"%s: can not find memory for array...\n", myname);
		exit(1);
	}
	marker_str(*crm, str, coord_flag);

	return;
}

void read_mark_pol(crm, n, filename, coord_flag)
struct _crm_  **crm;
int *n;
char *filename;
int coord_flag;
{
	register int i=0;
	char str[161];
	FILE *fp;

	/* open file */
	if ((fp = fopen(filename, "r")) == (FILE *)0) {
		PR(ER, "%s: can not open file '%s'\n", myname, filename);
		exit(1);
	}

	/* strings count */
	while (fgets(str, 160, fp) != NULL) {
		if (*str =='\n' || *str == '\r' || *str=='#'|| *str =='%')
			continue;
		i++;
	}
	fseek(fp, 0l, 0);
	*n = i;

	/* find memory for arrays */
	if ((*crm = (struct _crm_ *)calloc((unsigned)(*n), sizeof **crm)) ==
	    (struct _crm_ *)0) {
		PR(ER,"%s: can not find memory for array...\n", myname);
		exit(1);
	}

	/* strings reading */
	i=0;
	while (fgets(str, 160, fp) != NULL) {
	    if (*str == '\n' || *str == '\r' || *str == '#'|| *str == '%')
		    continue;

	    marker_str_pol((*crm)+i, str, coord_flag);
	    i++;
	}

	fclose(fp);

	return;
}

void marker_str_pol(crm, str, coord_flag)
struct _crm_  *crm;
char *str;
int coord_flag;
{
     double cr1, cr2;
     char st[16], sp[16], spol[16];
     st[0] = sp[0] = spol[0] = '\0';

     skip_wd(0, str, st);
     skip_wd(1, str, sp);
     skip_wd(2, str, spol);


     switch(coord_flag) {
     case 1: /* galactic coordinates */
	    crm->theta = cr1 = deg2rad(st);
	    crm->phi   = cr2 = deg2rad(sp);
	    gal2pol(cr1, cr2, &crm->theta, &crm->phi);
	    break;
     case 2: /* equatorial coordinates */
	    {
	       double ra, dec, lat, lon, p= M_PI/180.;
	       ra  = str2deg(st);
	       dec = str2deg(sp);
	       equat2gal(ra, dec, &lat, &lon);

	       lat *= p; lon *= p;
	       crm->theta =  M_PI/2 - lat;
	       crm->phi   =  lon;
	    }
	    break;
     default: /* polar coordinates */
	    crm->theta = cr1 = deg2rad(st);
	    crm->phi   = cr2 = deg2rad(sp);
	    break;
     }

     crm->mark  = -10;
     crm->vpa   = deg2rad(spol);
     crm->size  = -1;
     crm->color = -1;

     return;
}
