#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "glesp.h"

#define PR      fprintf
#define ER      stderr

extern double (*func_weight)();
extern int    read_obs_n(char *);
extern void   read_obs_data(char *, struct _obs_ **, int *);
extern double weight_p(int, int, int, double *, int, int, double, double,
		 double (*)());
extern int    ind_x(int, double *, double);

/* read observational data */
void read_obs(f, nx, np, x, nph, obs_file, pix_size)
double *f;
int nx, np;
int *nph;
double *x;
char *obs_file;
double pix_size;
{
	register int i, j;
	int n_pnt;
	struct _obs_ *ot;
	float  *w_array = (float *)0;
	int    *count_array = (int *)0;
	double pi2 = 2. * M_PI;


	/* open file with observations */
	/* read number of points and observational data*/
	read_obs_data(obs_file, &ot, &n_pnt);

	i = nx * np;
	/* take memory for temporal-arrays */
	if ((w_array = (float *)calloc(i, sizeof *w_array)) == (float *)0) {
nomem:
		PR(ER, "%s: no memory for temporal arrays...\n", myname);
		exit(1);
	}
	if ((count_array = (int *)calloc(i, sizeof *count_array)) == (int *)0)
		goto nomem;

	/*
	 *
	 *
	 *
	 *
	 */

	/* copy pixels to map */
	for(i=0; i< n_pnt; i++) {
		/* index-s of map */
		int i_map = ind_x(nx, x, cos(ot[i].theta));/* index by X */
		int j_map = ot[i].phi / pi2 * nph[i_map];

		double w_p = weight_p (nx, np, nph[i_map], x,
			i_map, j_map, (double)ot[i].theta, (double)ot[i].phi,
			func_weight);
		int ii = i_map * np + j_map; /* index in array */

		w_array[ii] += w_p;        /* sum of weights */
		f[ii] = w_p * ot[i].t;     /* sum of pixels  */
		count_array[ii]++;
	}
	/* weighting of pixels */
	for(i=0; i< nx; i++) {
		int ii = i * np;
		for(j=0; j< np; j++) {
			if (count_array[ii+j])
			     f[ii+j] /=  w_array[ii+j];
		}
	}

	if (w_array != (float *)0)     free((void *)w_array);
	if (count_array != (int *)0)   free((void *)count_array);
	if (ot != (struct _obs_ *)0)   free((void *)ot);

	return;
}

/* detect index by X */
int ind_x_old(nx, x, x_in)
int nx;
double *x, x_in;
{
	register int i;
	int i_min=0;
	double d_min = 1.e+30;

	for(i=0;i<nx;i++) {
		double dx = x[i] - x_in;
		if (fabs(dx) <  d_min) {
			i_min = i;
			d_min = dx;
		}
	}

	return i_min;
}

/*
 * weighting function
 *
 */
double weight_p (nx, np, nph, x, i_map, j_map, theta, phi, func_w)
int nx, np, nph;    /* configure grid               */
double *x;          /* input x=cos(theta) array     */
int i_map, j_map;   /* index of corresponding pixel */
double theta, phi;
double (*func_w)();
{
    double pi2 = 2 * M_PI;
    double w;
    /* detect shift of input point position from a pixel center */
    double dp = pi2 / nph;                 /* step by phi */
    double d_theta=acos(x[i_map]) - theta, /* shift from center by theta */
	   d_phi = dp * j_map - phi;       /* shift from center by phi   */
    double x1,x2, p1,p2; /* pixel bounds */

    /* detect pixel boundaries */
    x1 = x[i_map];
    x2 = x[i_map+1];
    p1 = dp * j_map;
    p2 = dp * (j_map+1);

    /* weight calculation */
    w = (*func_w)(d_theta, d_phi);

    return w;
}

