#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "glesp.h"

#define PR      fprintf
#define ER      stderr
#define ST      stdout
#define ABS(x)  ((x) < 0 ? -(x) : (x))
#define FMAX    1.e+30

HEAD *ph;

/*
 *  Simple smoothing and Tikhonov restoration procedures
 *  over a_[lm}-s.
 *
 *  Oleg Verkhodabov (2003-2004)
 *      Last modification 19-Oct-2004, SAO
 */
char alm_file[161], beam_file[161],
     outp_file[161];
int  print_alm_flag = 0;
int  lmin_print = -1;
int  lmax_print = -1;
int  cut_lmin = -1;     /* cut value for a_lm        */
int  cut_lmax = 0;      /* cut value for a_lm        */
int  lmax = 0;
double alp = 0.0;       /* coefiicient for restoration */
int smooth_flag = 1;
int  outp_fflag = 0;    /* default output binary data in FITS format */
double fwhm = -1.;      /* beam width            */
double cf_factor = -1.; /* just multiply by exp(-l(l+1)/cf) */
int pol3alm_flag= 0;    /* polarization alm-file       */

extern void setopt(int, char **), usage(void), usage2(void);
extern void get_val_str(char *, double *, double *);

int    naxis2;
char *myname="rsalm";

int main(argc, argv)
int argc;
char **argv;
{
	register int l, m;
	double *alm = (double *)0, *almp1 = (double *)0, *almp2 = (double *)0;
	double *beam, *f;
	int lm_beam = 0;
	int nalm=0, nbuf=0;
	int file_type;
	FILE *fp_out = (FILE *)0;

	alm_file[0] = outp_file[0] = beam_file[0] = '\0';

	/* loop over external parameters */
	setopt(argc, argv);

       /* test of input GLESP file type:
	* flag = -1       it is not GLESP file
	* flag = 0        F-file
	* flag = -2       FITS-file, but not GLESP map
	* flag = -3       ASCII-file
	* flag = 1        FITS-file
	* flag = 2        FITS-file, alm-file
	* flag = 3        FITS-file, 3 maps (T,Q,U)
	* flag = 4        FITS-file, 3 alm-s (A,E,B)
	* flag = 5        FITS-file, Hpix map
	* flag = 6        FITS-file, Hpix 3-map
	*/
	file_type = test_glesp_map(alm_file);
	switch(file_type) {
	default:
	case 3 :
		PR(ER, "%s: '%s' is not GLESP alm-file. It contains a map\n",
		    myname, alm_file);
		exit(1);
	case -1:
	case  0:  /* F-file    */
	case -2:  /* FITS-file */
	case  1:  /* FITS-file */
		PR(ER, "%s: '%s' is not a GLESP file.\n",  myname, alm_file);
		exit(1);
	case  4:  /* FITS-file 3 alm-s (A,E,B)  */
		if (!pol3alm_flag)  pol3alm_flag = 't';
		break;
	case  -3: /* ASCII - may be alm ? */
	case   2: /* FITS  alm-file       */
		;  /* Ok */
		break;
	}
	/* read input file */
	if (pol3alm_flag) {
	   /* read 3-alm file */
	   read_hp_palm(&alm, &almp1, &almp2, &lmax, &nbuf, alm_file);
	}  else {
	   /* read ASCII or 1-alm file */
	   if (lmax ==0 && lmax_print > 0) {
		lmax = -lmax_print;
		nbuf = -2 * (lmax+1);
		nalm = -(lmax + 1);
	   }
	   read_alm_table(alm_file, &alm, &nalm, &nbuf, &lmax);
	   if (lmax < 0 ) lmax = -lmax;
	}

	/* read beam from file */
	if (*beam_file != '\0') {
	    double *qq = (double *)0; /* organize temporal parameter   */
				      /* for function,just for nothing */
	    read_cl_table (beam_file, &beam, &qq, &lm_beam);
	    free((void *)qq);
	}

	if (cut_lmin > 0) {
		for(l=0; l<=cut_lmin; l++) {
		    int ll = l * nbuf;
		    alm[ll] = 0.0;
		    for(m=1; m<=l; m++) {
			alm[ll + m] = alm[ll + m + lmax+1] = 0.0;
			if (pol3alm_flag) {
			    almp1[ll + m] = almp1[ll + m + lmax+1] = 0.0;
			    almp2[ll + m] = almp2[ll + m + lmax+1] = 0.0;
			}
		    }
		}
	}

	/* reconfiguration of a_{lm}-array */
	if ( lm_beam > 0 && lmax != lm_beam && lmax > lm_beam) {
		alm2alm(&alm, &lmax, &nbuf, lm_beam, nbuf);
		lmax = lm_beam;
	}

	/* same operations for E,B-modes */
	f = alm;
	switch (pol3alm_flag) {
	case 'E':
	case 'e':
		f = almp1;
		break;
	case 'B':
	case 'b':
		f = almp2;
		break;
	default :
		break;
	}

	if (smooth_flag) {
	    /* smooth  */
	    extern void beam_cfg_smooth(double *, int, int, double);

	    if (fwhm > 0.) { /* beam width */
		  if ( cf_factor > 0.)
			beam_cfg_smooth(f, lmax, nbuf, fwhm);
		  else
			beam_gauss_smooth(f, lmax, nbuf, fwhm);
	    } else
		  beam_smooth(f, lmax, nbuf, beam);
	} else {
	    /* restore a_{lm} */
	    if (fwhm > 0.) /* beam width */
		beam_tikh_gauss(f, lmax, nbuf, fwhm, alp);
	    else
		beam_tikh(f, lmax, nbuf, beam, alp);
	}

	if (lmax_print < 0 || lmax_print > lmax)
		lmax_print = lmax;
	if (lmin_print < 0 || lmin_print > lmax)
		lmin_print = 0;

	if (cut_lmax <= 0)  cut_lmax = lmax;
	if (cut_lmax > lmax_print)  cut_lmax = lmax_print;

	if (outp_file[0] == '\0' || strcmp(outp_file, "stdout") == 0)
	    fp_out = stdout;
	else
	    if ((fp_out = fopen(outp_file, "w")) == (FILE *)0) {
		PR(ER, "%s: can not create file '%s'\n", myname, outp_file);
		exit(1);
	    }

	if ((fp_out = fopen(outp_file, "w")) == (FILE *)0) {
	    PR(ER, "%s: can not create file '%s'\n", myname, outp_file);
	    exit(1);
	}
	/* write alm */
	if (pol3alm_flag) {
	     write_ieb_alm(alm, almp1, almp2, lmax, nbuf, cut_lmax,fp_out);
	} else {
	    if (outp_fflag)
		ff_write_alm(alm, nbuf, lmax, cut_lmax, fp_out);
	    else
		out_glesp_alm(alm, lmax, nbuf, cut_lmax, fp_out);
	}

	/* close file */
	if(fp_out != (FILE *)0 && fp_out != stdout) fclose(fp_out);

	/* free memory */
	if (alm != (double *)0)  free((void *)alm);

	exit(0);
}

void setopt(argc, argv)
int argc;
char **argv;
{
	register int i;
	myname = argv[0];
	if (argc == 1)
		usage();
	for(i=1; i<argc; i++) {
	    switch(argv[i][0]) {
	    case '-':
		switch(argv[i][1]) {
		case 'a':
			if(strncmp(argv[i]+1, "al", 2) == 0) {
			    alp = atof(argv[++i]);
			    smooth_flag = 0;
			} else
			    strcpy(alm_file, argv[++i]);
			break;
		case 'b':
			strcpy(beam_file, argv[++i]);
			break;
		case 'B':
			pol3alm_flag = 'b';
			break;
		case 'c':
			if (strncmp(argv[i]+1, "clma", 4) == 0) {
				cut_lmax = atoi(argv[++i]);
			} else if (strncmp(argv[i]+1, "clmi", 4) == 0) {
				cut_lmin = atoi(argv[++i]);
			} else if (strncmp(argv[i]+1, "cf", 2) == 0) {
				/* multiply by exp(-l(l+1)/cf) */
				fwhm = cf_factor = atoi(argv[++i]);
			}
			break;
		case 'e':
		case 'E':
			pol3alm_flag = 'e';
			break;
		case 'f':
			if (strncmp("fw", argv[i]+1, 2) == 0) {
			    fwhm = atof(argv[++i]);
			    if (fwhm < 1.e-8) {
				PR(ER, "%s: you can not smooth data with ",
					   myname);
				PR(ER, "such beam size: '-fw %g'\n", fwhm);
				PR(ER, "Please, set positive value.\n");
				exit(1);
			    }
			    if (argv[i-1][strlen(argv[i-1])-1] == '2') {
				/* last symbol after 'fw' is '2' */
				fwhm *= M_SQRT2;
			    }
			    smooth_flag = 1;
			} else {
			    PR(ER,"%s: illegal flag '%s'\n",myname,argv[i]);
			    usage();
			}
			break;
		case 'h':
			if (strncmp(argv[i]+1, "h2",2) == 0)
			    usage2();
			else
			    usage();
			break;
		case 'l':
			if (strncmp(argv[i]+1, "lmn",3) == 0 ||
			    strncmp(argv[i]+1, "lmi",3) == 0 )
				lmin_print = atoi(argv[++i]);
			else
				lmax_print = atoi(argv[++i]);
			break;
		case 'L':
			lmax = atoi(argv[++i]);
			break;
		case 'o':
			if (strncmp(argv[i]+1, "of", 2) == 0) {
			    outp_fflag = 1; /* output in F-format */
			}
			strcpy(outp_file, argv[++i]);
			break;
		case 's':
			if (strncmp(argv[i]+1, "sm",2) == 0)
			    smooth_flag = 1;
			else {
			    PR(ER,"%s: illegal flag '%s'\n",myname,argv[i]);
			    usage();
			}
			break;
		case 't':
		case 'T':
			pol3alm_flag = 't';
			break;
		case 'v':
			PR(ER, "%s: version 0.2 (30/04/2009), GLESP 2.0.\n",
			    myname);
			exit(0);
		case 'w':
			fwhm = atof(argv[++i]);
			smooth_flag = 1;
			break;
		default :
			usage();
			break;
		}
		break;
	    case '+':
		switch(argv[i][1]) {
		case 'h':
			usage();
		default :
			usage();
			break;
		}
		break;
	    default :
		strcpy(alm_file, argv[i]);
		break;
	    }
	}

	return;
}

void usage()
{
PR(ER,"usage: %s -a alm_file [-o out_file] [-b beam_file] [-alp a-coeff.]\n",
myname);
PR(ER,"[-sm(smoothing)] [-l lmax_print] [-lmin lmin_print] [-fw{2} beam_width(')]\n");
PR(ER,"[-clmin cut_lmin] [-clmax cut_lmax] [-cf factor] [-{T,E,B}] [-v(vers.)]\n");
exit(2);
}


void usage2()
{
PR(ST, "usage: %s -a alm_file [FLAGS]\n", myname);
PR(ST, "-----------------------------------------------------------------\n");
//PR(ST,"-cf factor\n");
PR(ST,"-a alm_file - input alm-file, if it's 3alm use '-T','-B' or '-E'\n");
PR(ST,"-alp a-coeff - sets alpha-coef. for Tikhonov regularization resotration\n");
PR(ST,"-b beam_file - harmonic beam file: Beam_l = exp(-sigma^2 l(l+1)/2\n");
PR(ST,"-B - selects B-pol. data from 3-file (see flags '-T','-E')\n");
PR(ST,"-clmin cut_lmin - appropriates 0 for alm-s with l<=lmin\n");
PR(ST,"-clmax cut_lmax - appropriates 0 for alm-s with l>=lmax\n");
PR(ST,"-E - selects E-pol. data from 3-file (see flags '-T','-B')\n");
PR(ST,"-fw{2} fwhm(') - smooth a map with gaussian beam of 'fwhm' arcmin\n");
PR(ST,"-l lmax_print - sets lmax for output stream\n");
PR(ST,"-lmin lmin_print - sets lmin for output stream\n");
PR(ST,"-o out_file - output alm-file\n");
PR(ST,"-sm - smooths a_lm with harmonic beam file (see '-b' flag)\n");
PR(ST,"-T - selects temperature anistropy data from 3-file (see flags '-E','-B')\n");
PR(ST,"-v - prints version of '%s'\n", myname);
exit(0);
}


