#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define PR      fprintf
#define ER      stderr

extern void     rffti_(int *, double *);
extern void     rfftb_(int *, double *, double *);
extern void     set_phi_(int *, double *, double *, double *);
extern void     fftw_f(int, double *, double *);

extern int  sym_grid;    /* symmetric grid over x */
extern int  fill_0;      /* fill arrays by 0      */
/*
 *  fast loop over m - SUM_m[a_{lm}e^{i m phi}]
 *
 *
 *    IMPLICIT REAL*8(a-h,o-y),COMPLEX*16(z)
 *    DIMENSION f(Nmx,Nmp),Nph(Nmx)
 *    DIMENSION c(Nmp),wsave(2*Nmp+15)
 *    DIMENSION apm(0:Jmx,Nmx),bpm(0:Jmx,Nmx)
 *
 */
void sig_fast3_(Nmx,Nmp, Lmx, Nx, Nph, apm, bpm, c, f, wsave)
int *Nmx, *Nmp;
int *Lmx, *Nx,
    *Nph;
double *f;
double *apm, *bpm;
double *c, *wsave;
{
	register int ix, ip, ic;
	double sq2 = M_SQRT2; /* 1.41421356237309504880 */
	int nphi_max = 0, nx1 = *Nx - 1;
	int Kx = sym_grid > 0 ? (*Nx + 1)/2 :  *Nx;

	if (fill_0) {
		for(ix=0; ix < *Nx; ix++)
			if (nphi_max < Nph[ix])
				nphi_max = Nph[ix];
	}

	for(ix=0; ix < Kx; ix++) {  /*  loop over x */
		register int ll = ix * *Nmp, ipl;
		int nphi_work = fill_0 ? nphi_max : Nph[ix];
		c[0] = sq2 * apm[ix];
		ic = 0;

		wsave[0]=0.0;
		for(ip=1, ipl=*Nmx+ix; ip<= *Lmx; ip++, ipl+=*Nmx) {
			c[++ic] = apm[ipl];
			c[++ic] = bpm[ipl];
			wsave[ip+ip]=wsave[ip+ip-1]=0.0;

		}
		++ic;

		if (ic < nphi_work)
			for (ip=ic; ip<nphi_work; ip++)   c[ip] = 0.0;

		fftw_f(nphi_work, c, wsave);
		for(ip=0; ip<nphi_work; ip++)   f[ll+ip] = 0.50 * c[ip];

		if (sym_grid > 0) {  /* then calculation for south pole */
		       register int ix_s = nx1 - ix, ipl;
		       if (ix_s == ix)
				continue;

		       ll = ix_s * *Nmp;
		       c[0] = sq2 * apm[ix_s];
		       ic = 0;
		       wsave[0]=0.0;
		       for(ip=1, ipl=*Nmx+ix_s; ip<= *Lmx; ip++, ipl+=*Nmx) {
			     c[++ic] = apm[ipl];
			     c[++ic] = bpm[ipl];
			     wsave[ip+ip]=wsave[ip+ip-1]=0.0;
		       }
		       ++ic;
		       if (ic < nphi_work)
			     for (ip=ic; ip<nphi_work; ip++)   c[ip] = 0.0;

		       fftw_f(nphi_work, c, wsave);

		       for(ip=0; ip<nphi_work; ip++) f[ll+ip] = 0.50 * c[ip];
		}
	}
	return;
}

/*
 *   loop over m - SUM_m[a_{lm}e^{i m phi}]
 *
 *   IMPLICIT REAL*8(a-h,o-y)
 *   DIMENSION f(Nmx,Nmp),Nph(Nx)
 *   DIMENSION apm(0:Jmx,Nmx),bpm(0:Jmx,Nmx)
 *   DIMENSION phi(Nmp)
 *   DIMENSION cmphi(2*Nmp)
 */
void  sig_phi_(Nmx,Nmp, Lmx,Nx, Nph,phi,phi2,apm,bpm,cmphi,f)
int *Nmx, *Nmp;
int *Lmx, *Nx,
    *Nph;
double *phi, *phi2;
double *apm, *bpm;
double *cmphi;
double *f;
{
	register int iq, ix, m;
	double *smphi = cmphi + *Nmp;
	int Nphi = Nph[0];
	double dphi, s;
	double sum=0.0;

	set_phi_(&Nphi, phi2, phi, &dphi);

	for(iq=0; iq < Nphi; iq++) {  /* loop over phi */
		double phi1 = phi[iq];
		for(m=1; m <= *Lmx; m++) {
			cmphi[m] = cos(m*phi1);
			smphi[m] = sin(m*phi1);
		}
		for(ix=0; ix < *Nx; ix++) { /* loop over x */
			register int ll = ix * *Nph;
			s = 0.70710678118654752440 * apm[ix];
			for(m=1; m <= *Lmx; m++) {
			   int mm = m * *Nmx;
			   s += apm[mm+ix]*cmphi[m] - bpm[mm+ix]*smphi[m];
			}
			f[ll+iq] = s;
			sum += s*s;
		}  /*  end of the loop over x   */
	}          /*  end of the loop over phi */
/*
	printf("sum^2=%15.8e\n", sum/(Nphi+0.));
*/
	return;
}
