#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#define PR      fprintf
#define ER      stderr
struct _ph_st_ {
     int l, m;
     double ph;
};

extern void  setopt(int, char **), usage(void);
extern void  sort_phase(struct _ph_st_ *, int);

char ph_file[161];
int lmax = 2000;
double ph0 = 0.0, eps = 0.0;

char *myname="sort_phase";

int main(argc, argv)
int argc;
char **argv;
{
	register int l, m;
	double *ph = (double *)0;
	struct _ph_st_ *ph_st;
	double pi2 = 2 *M_PI;
	unsigned int nl;
	FILE *fp_out = stdout;

	ph_file[0] = '\0';

	/* loop over external parameters */
	setopt(argc, argv);

	read_ph_table(ph_file,  &ph,  &lmax);

	nl = (lmax+1)*(lmax+1)/2;

	if ((ph_st=(struct _ph_st_ *)calloc(nl, sizeof *ph_st)) ==
	     (struct _ph_st_ *)0) {
		PR(ER, "%s: no memory for arrays...\n", myname);
		exit(1);
	}

	/* fill in array */
	nl=0;
	for(l=2; l<=lmax; l++)
	    for(m=1; m<=l; m++) {
		ph_st[nl].l  = l;
		ph_st[nl].m  = m;
		ph_st[nl].ph = ph[l*(lmax+1)+m];
		nl++;
	    }

	/* sort array */
	sort_phase(ph_st, nl);

	/* select all phases near given one */
	m = 0;
	for(l=0; l<nl; l++) {
	    if (ph0 > eps) {
		if (ph_st[l].ph < ph0 - eps)
			continue;
		if (ph_st[l].ph > ph0 + eps)
			break;
	    } else {
		if (ph_st[l].ph > ph0 + eps && ph_st[l].ph < ph0+pi2-eps)
			continue;
	    }
	    m++;
	    PR(fp_out, "%4d  %4d  %12.5e\n",
		    ph_st[l].l, ph_st[l].m, ph_st[l].ph);
	}
	PR(ER, "m = %d\n", m);

	exit(0);
}

void setopt(argc, argv)
int argc;
char **argv;
{
    register int i;
    myname = argv[0];

    for(i=1; i<argc; i++) {
	switch(argv[i][0]) {
	case '-':
		switch(argv[i][1]) {
		case 'e':
			eps = atof(argv[++i]);
			break;
		case 'p':
			ph0 = atof(argv[++i]);
			break;
		case 'h':
			usage();
			break;
		default:
			break;
		}
		break;
	case '+':
		switch(argv[i][1]) {
		case 'h':
		default:
			usage();
			break;
		}
		break;
	default:
		strcpy(ph_file, argv[i]);
		break;
	}
    }

    return;
}

void usage()
{
PR(ER, "usage: %s file -p phi0 -e eps\n", myname);
exit(2);
}

/* "Quicker-sort algorithm" from system library         */
int i_cmp(p1, p2)
void  *p1, *p2;
{
	struct _ph_st_ *ph1 = (struct _ph_st_ *)p1,
		       *ph2 = (struct _ph_st_ *)p2;
	return (ph1->ph < ph2->ph ? -1 : (ph1->ph > ph2->ph ? 1 : 0));
}
void sort_phase(ph, n)
struct _ph_st_ *ph;
int n;
{
	qsort(ph, n, sizeof(*ph), i_cmp);
	return;
}
