#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "glesp.h"

#define PR            fprintf
#define ER            stderr
#define MAX(x,y)      ((x) > (y) ?  (x) : (y))
#define ABS(x)        ((x) < 0   ? -(x) : (x))

/* create GD-image and copy map to output file */
void t2mol(f, nx, nphi, x, nph, img, xs, ys, lon_0, mask)
double *f;
int nx, nphi;
double *x;
int    *nph;
double  **img;
int xs, ys;
double lon_0;   /* longitude to be added for rotation */
int **mask;
{
    register int i, j;
    unsigned int n = xs*ys;
    double u, v, theta, phi;
    double pi = M_PI;
    double x_c, y_c, x_r, y_r;

    /* find memorry */
    if ( (*img = (double *)calloc(n, sizeof **img)) == (double *)0 ) {
nomem:
	PR(ER, "t2mol: no memory for temporal array...\n");
	exit(1);
    }
    if ( (*mask = (int *)calloc(n, sizeof **mask)) == (int *)0 )
	goto nomem;

    /* center of image */
    y_c = ys / 2;
    x_c = xs / 2;
    x_r = xs / 2;
    y_r = ys / 2;

    /* fill in image */
    for(i=0; i<ys; i++) {
	register int ii = i * xs;

	v = (i - y_c) / (y_r/1.02);

	for(j=0; j<xs; j++) {
	    u = 2. * (j - x_c) / (x_r/1.02);
	    (*img)[ii+j] = 0.0;
	    theta = phi = 1.e+30;

	    /* select inside ellipse */
	    if ( u*u/4. + v*v <= 1.) {
		double sqv = sqrt(1.-v*v);

		(*mask)[ii+j] = 1;

		/* move theta int [0, pi] */
		theta = pi/2. - asin(2./pi * (asin(v) + v*sqv));
		/* lon in [-pi,pi], longitude lon_0 added */
		phi = -lon_0 - pi/2 * u / MAX(sqv, 1.e-6);
		/* move phi into [0, 2pi] */
		if ( phi < 0.0 )    phi += 2. * pi;
		if ( phi > 2.*pi )  phi -= 2. * pi;
	    }
	    /* keep only meaningful points */
	    if (ABS(theta) < 1.e5) {
		int n = ang2n(theta, phi, x, nx, nphi, nph);
//              if ( (*img)[ii+j] > 1.e-29)
		     (*img)[ii+j] = f[n];
	    }
	}
    }

    return;
}


void ht2mol(f, nx, nphi, x, offs, nph, img, xs, ys, lon_0, mask)
double *f;
int nx, nphi;
double *x, *offs;
int    *nph;
double  **img;
int xs, ys;
double lon_0;   /* longitude to be added for rotation */
int **mask;
{
    register int i, j;
    unsigned int n = xs*ys;
    double u, v, theta, phi;
    double pi = M_PI;
    double x_c, y_c, x_r, y_r;

    /* find memorry */
    if ( (*img = (double *)calloc(n, sizeof **img)) == (double *)0 ) {
nomem:
	PR(ER, "t2mol: no memory for temporal array...\n");
	exit(1);
    }
    if ( (*mask = (int *)calloc(n, sizeof **mask)) == (int *)0 )
	goto nomem;

    /* center of image */
    y_c = ys / 2;
    x_c = xs / 2;
    x_r = xs / 2;
    y_r = ys / 2;

    /* fill in image */
    for(i=0; i<ys; i++) {
	register int ii = i * xs;

	v = (i - y_c) / (y_r/1.02);

	for(j=0; j<xs; j++) {
	    u = 2. * (j - x_c) / (x_r/1.02);
	    (*img)[ii+j] = 0.0;
	    theta = phi = 1.e+30;

	    /* select inside ellipse */
	    if ( u*u/4. + v*v <= 1.) {
		double sqv = sqrt(1.-v*v);

		(*mask)[ii+j] = 1;

		/* move theta int [0, pi] */
		theta = pi/2. - asin(2./pi * (asin(v) + v*sqv));
		/* lon in [-pi,pi], longitude lon_0 added */
		phi = -lon_0 - pi/2 * u / MAX(sqv, 1.e-6);
		/* move phi into [0, 2pi] */
		if ( phi < 0.0 )      phi += 2. * pi;
		if ( phi > 2.*pi )    phi -= 2. * pi;
	    }
	    /* keep only meaningful points */
	    if (ABS(theta) < 1.e5) {
		int n = hang2n(theta, phi, x, nx, nphi, nph, offs);
//                if ( (*img)[ii+j] > 1.e-29)
		     (*img)[ii+j] = f[n];
	    }
	}
    }

    return;
}


/* create GD-image, output overlaid mask map */
void overlay2mol(overlay_col, overlay_wid, lon_0, f,x,nx,np,nph, flag)
int overlay_col, overlay_wid;
double lon_0;
double *f, *x;
int nx, np, *nph;
int flag;           /* mask for v=1, or for v=0 */
{
    extern int get_nx_gd(), get_ny_gd();
    extern void color_gd(), move_gd();
    extern void marker_gd();
    register int i, j, k;
    struct _cr_  {double theta, phi; int v; int x, y;} *cr;
    int nx_gd = get_nx_gd(), ny_gd = get_ny_gd();
    int nn = nx_gd*ny_gd;
    int y_c, x_c, x_r, y_r;
    double theta=0., phi=0.;
    double pi = M_PI;
    int overlay_color = overlay_col;
    int overlay_width = overlay_wid;
    int i_x, j_phi;

    cr = (struct _cr_ *)calloc((unsigned)nn, sizeof *cr);
    if (cr == (struct _cr_ *)0) {
	PR(ER, "overlay2mol: no memory for temporal array...\n");
	exit(1);
    }

    y_c = ny_gd / 2;
    x_c = nx_gd / 2;
    x_r = nx_gd / 2;
    y_r = ny_gd / 2;

    /* put copy of base map to memory */
    for(k=i=0; i<ny_gd; i++) {
	double u, v = (i - y_c) / (y_r/1.02);

	for(j=0; j<nx_gd; j++) {
	    theta = phi = 1.e+30;
	    u = 2. * (j - x_c) / (x_r/1.02);
	    if ( u*u/4. + v*v <= 1.) {
		double sqv = sqrt(1.-v*v);
		/* theta in [0, pi] */
		theta = pi/2. - asin(2./pi * (asin(v) + v*sqv));
		/* lon in [-pi,pi], the minus sign for astro convention */
		phi = -lon_0 - pi/2 * u / MAX(sqv, 1.e-6);
		/* phi in [0, 2pi] */
		if ( phi < 0.0 )    phi += 2. * pi;
		if ( phi > 2.*pi )  phi -= 2. * pi;
		angle2pixel(x, nx, nph, &i_x, &j_phi, theta, phi);

		cr[k].theta = theta;
		cr[k].phi   = phi;
		cr[k].y     = i;
		cr[k].x     = j;

		if (flag) {
		    if (f[i_x * np + j_phi] > 0.5)  cr[k].v = 1;
		} else {
		    if (f[i_x * np + j_phi] < 0.5)  cr[k].v = 1;
		}
		if (cr[k].v) {  /* put mask */
		    /* put marker */
		    int marker=0;
		    color_gd(overlay_color);
		    move_gd(cr[k].x, cr[k].y);
		    marker_gd(overlay_width, marker); /* point */
		}
	    }
	    k++;
	}
    }

    free((void *)cr);

    return;
}


#define ABS(x)  ((x) < 0 ? -(x) : (x))
/* create GD-image, output grid on map and copy map to output file */
/* despite too many loops, it works fast                           */
void grid2mol(grid_type, grid_col, lon_0, d1_step, d2_step)
int grid_type;  /* 1-galactic, 2-equatorial, 3-ecliptic */
int grid_col;
double lon_0;
double d1_step, d2_step;
{
    extern int get_nx_gd(), get_ny_gd();
    register int i, j, k, pa, pb;
    struct _cr_  {double theta, phi; double a, b; int x, y;} *cr;
    int nx_gd = get_nx_gd(), ny_gd = get_ny_gd();
    int nn = nx_gd*ny_gd, na=0, nb=0;
    int y_c, x_c, x_r, y_r, x_m, y_m;
    double theta=0., phi=0., a, b, a0=0., b0=0., da=15., db=15.;
    double pi = M_PI;
    double pmin= 1.e+30, pmax= -1.e+30;
    int gr_eq = 0, gr_ga = 0, gr_ec = 0;
/*
    double a0_eq = 0., b0_eq = 0., da_eq = 15., db_eq = 15.;
    double a0_ec = 0., b0_ec = 0., da_ec = 15., db_ec = 15.;
    double a0_ga = 0., b0_ga = 0., da_ga = 15., db_ga = 15.;
*/
    int grid_color = grid_col;
    int grid_width = 1;
    extern void pol2gal(double, double, double *, double *);
    extern void gal2equat(double, double, double*, double*);
    extern void equat2eclip(double, double, double*, double*);
    extern void color_gd(), move_gd();
    extern void marker_gd();

    cr = (struct _cr_ *)calloc((unsigned)nn, sizeof *cr);
    if (cr == (struct _cr_ *)0) {
	PR(ER, "grid2mol: no memory for temporal array...\n");
	exit(1);
    }

    switch (grid_type) {
    case 0:
    case 1: /* galactic */
	gr_ga = 1;
	break;
    case 2: /* equatorial */
	gr_eq = 1;
	break;
    case 3: /* ecliptic */
	gr_ec = 1;
	break;
    case 12:
    case 21: /* galactic & equatorial*/
	gr_ga = 1;
	gr_eq = 1;
	break;
    case 13:
    case 31:/* galactic & ecliptic*/
	gr_ga = 1;
	gr_ec = 1;
	break;
    case 23:
    case 32:/* equatorial & ecliptic*/
	gr_eq = 1;
	gr_ec = 1;
	break;
    case 123: /* galactic, equatorial & ecliptic */
	gr_ga = 1;
	gr_eq = 1;
	gr_ec = 1;
	break;
    }

    y_c = ny_gd / 2;
    x_c = nx_gd / 2;
    x_r = nx_gd / 2;
    y_r = ny_gd / 2;

    /* put copy of base map to memory */
    /* put each cooreponding pixel coordinates to memory */
    for(k=i=0; i<ny_gd; i++) {
	double u, v = (i - y_c) / (y_r/1.02);

	for(j=0; j<nx_gd; j++) {
	    cr[k].a = cr[k].b = 1e+30;
	    theta = phi = 1.e+30;
	    u = 2. * (j - x_c) / (x_r/1.02);
	    if ( u*u/4. + v*v <= 1.) {
		double sqv = sqrt(1.-v*v);
		/* theta in [0, pi] */
		theta = pi/2. - asin(2./pi * (asin(v) + v*sqv));
		/* lon in [-pi,pi], the minus sign for astro convention */
		phi = -lon_0 - pi/2 * u / MAX(sqv, 1.e-6);
		/* phi in [0, 2pi] */
		if ( phi < 0.0 )    phi += 2. * pi;
		if ( phi > 2.*pi )  phi -= 2. * pi;

		/* coordinates of a grid point */
		if (grid_type == 0 || grid_type == 1) {
		    /* convert polar coordinates to galactic coordinates */
		    pol2gal(theta,phi, &a, &b);
		    /* radians -> degrees */
		    cr[k].a = a*180/M_PI;
		    cr[k].b = b*180/M_PI;
		} else if (grid_type == 2) {
		    pol2gal(theta,phi, &a, &b);
		    cr[k].a = a*180/M_PI;
		    cr[k].b = b*180/M_PI;
		    gal2equat(cr[k].a, cr[k].b, &a, &b);
		    cr[k].a = a; /* right ascension */
		    cr[k].b = b; /* declination     */
		} else if (grid_type == 3) {
		    pol2gal(theta,phi, &a, &b);
		    cr[k].a = a*180/M_PI; /* gal.latitude  */
		    cr[k].b = b*180/M_PI; /* gal.longitude */
//if (ABS(phi) < 1.e-3 && ABS(theta-M_PI/2) < 1.e-3)
//PR(ER, "lat=%g, lon=%g\n");
		    gal2equat(cr[k].a, cr[k].b, &a, &b);
		    cr[k].a = a; /* right ascension, hrs */
		    cr[k].b = b; /* declination, degs    */
		    equat2eclip(cr[k].a, cr[k].b, &a, &b);
//if (ABS(phi) < 1.e-3 && ABS(theta-M_PI/2) < 1.e-3)
//PR(ER, "%d, %d,    %g,%g  ecl_lat=%g,ecl_lon=%g\n",
//i,j, cr[k].a,cr[k].b,  a,b);
		    cr[k].a = a; /* ecliptic latitude, degs  */
		    cr[k].b = b; /* ecliptic longitude, degs */
		} else {
		    pol2gal(theta,phi, &a, &b);
		    cr[k].a = a*180/M_PI;
		    cr[k].b = b*180/M_PI;
		    grid_type = 0; /* then galactic coordinates */
		}
	    }
	    if (phi < 1.e+3){
		if (pmin > phi) pmin = phi;
		if (pmax < phi) pmax = phi;
	    }
	    cr[k].theta = theta;
	    cr[k].phi   = phi;
	    cr[k].y     = i;
	    cr[k].x     = j;
	    k++;
	}
    }

    if (grid_type == 0 || grid_type == 1) {
	/* galactic coordinates */
	a0 = -90;        /* gal. latitude  */
	b0 = 0;          /* gal. longitude */
	da = d1_step<0.?15.:d1_step; /* step by spher.coor. a (lat.),degs */
	db = d2_step<0.?15.:d2_step; /* step by spher.coor. b (lon.),degs */
	na = 180. / da;  /* number of parallels */
	nb = 360. / db;  /* number of meridians */
    } else if (grid_type == 2) {
	a0 = 0;          /* right ascension  */
	b0 = -90;        /* declination      */
	da = d1_step<0.? 2.:d1_step;/* step by spher.coor. a (R.A.),hours */
	db = d2_step<0.?15.:d2_step;/* step by spher.coor. b (Dec.), degs */
	na = 24.  / da;  /* number of parallels */
	nb = 180. / db;  /* number of meridians */
    } else if (grid_type == 3) {
	/* ecliptic coordinates */
	a0 = -90;        /* eclip. latitude  */
	b0 = 0;          /* eclip. longitude */
	da = d1_step<0.?15.:d1_step; /* step by spher.coor. a (lat.),degs */
	db = d2_step<0.?15.:d2_step; /* step by spher.coor. b (lon.),degs */
	na = 180. / da;  /* number of parallels */
	nb = 360. / db;  /* number of meridians */
    }

    /* loop over grid a-coordinate */
    /*******************************/
    for(pa=0; pa<=na; pa++) {
	a = a0 + pa * da;
	x_m = y_m = -1;

	/* find closest pixel on image */
	for(i=0; i<ny_gd; i++) {
	    int ii = i * nx_gd;        /* beginning of the row */
	    int ii1 = (i+1) * nx_gd;
	    int ic = i * nx_gd + x_c;  /* middle of the row    */

	    if (i+1 >= ny_gd)   /* last line */
		   break;

	    /* we are still out of picture ellipse */
	    if (cr[ic].theta > 1.e5)  continue;

	    for(j=0; j<nx_gd; j++) {
		int jj  = ii+j;
		int jj1 = ii+j+1;
		int ij  = ii1+j;

		if (cr[jj].phi > 1.e5 || cr[jj1].phi > 1.e5)
		    continue;

		/* check conditions */
		if (
		    ((cr[jj].a <= a && a < cr[ij].a) && cr[ij].a<1.e5) ||
		     (cr[jj].a >= a && a > cr[ij].a) ||
		     (cr[jj].a >= a && a > cr[jj1].a) ||
		     (cr[jj].a <= a && a < cr[jj1].a) ||
		     ABS(cr[jj].a-a) < 1.e-6
		   ) {
		    /* then set point */
		    y_m = i;
		    if (ABS(cr[jj].a-a) < ABS(cr[jj1].a-a))
			x_m = j;
		    else
			x_m = j+1;
		    /* put marker */
		    color_gd(grid_color);
		    move_gd(x_m, y_m);
		    marker_gd(grid_width, grid_color); /* point */
		}
	    }
	}
    }

    /* loop over grid b-coordinate */
    /*******************************/
    for(pb=0; pb<=nb; pb++) {
	b = b0 + pb * db;
	x_m = y_m = -1;

	/* find closest pixel on image */
	for(i=0; i<ny_gd; i++) {
	    int ii = i * nx_gd;        /* beginning of the row */
	    int ii1 = (i+1) * nx_gd;
	    int ic = i * nx_gd + x_c;  /* middle of the row    */

	    if (i+1 >= ny_gd)   /* last line */
		   break;

	    /* we are still out of picture ellipse */
	    if (cr[ic].theta > 1.e5)  continue;

	    for(j=0; j<nx_gd; j++) {
		int jj  = ii+j;
		int jj1 = ii+j+1;
		int ij  = ii1+j;

		if (cr[jj].phi > 1.e5 || cr[jj1].phi > 1.e5)
		    continue;

		/* check conditions */
		if (
		    ((cr[jj].b <= b && b < cr[ij].b) && cr[ij].b<1.e5) ||
		     (cr[jj].b >= b && b > cr[ij].b) ||
		     (cr[jj].b >= b && b > cr[jj1].b) ||
		     (cr[jj].b <= b && b < cr[jj1].b) ||
		     ABS(cr[jj].b-b) < 1.e-6
		   ) {
		    /* then set point */
		    y_m = i;
		    if (ABS(cr[jj].b-b) < ABS(cr[jj1].b-b))
			x_m = j;
		    else
			x_m = j+1;
			/* put marker */
			color_gd(grid_color);
			move_gd(x_m, y_m);
			marker_gd(grid_width, grid_color); /* point */
		}
	    }
	}
    }
    free((void *)cr);

    return;
}
